#!/usr/bin/perl -w

#
#  Copyright 2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

#
#  Isaias Sanchez-Cortina                                     
#                                                                            
# Parses STDIN input text for later Language Model training.                 

#BEGIN {
#    eval { require Lingua::EN::Numbers  };
#    if (!$@) {
#        use Lingua::EN::Numbers qw(num2en num2en_ordinal);
#        $BOOL_Lingua_EN_Numbers=1;
#    }
#}

binmode(STDIN,':utf8');
binmode(STDOUT,':utf8');
use utf8;
use strict; 

#CONTROL VARS:                                                               
our $BOOL_Lingua_EN_Numbers=0;
our $BOOL_numbers2text=0;
our $str_Apostrophe='uncommon';


our $bRemoveDash=1;
our $sSil='Do not interpret <SIL> as anything special';
our $sSpellID='There are no tokens to specify the pronunctiation';
our $sCurlyTokens='Keep';
our $sRegex='Ignore /verbatim/correct/ syntax.';
our $sRegexSF='keep';

#help                                                                        
sub help {
    print STDERR "
Parses text from the standard input in order to meet the specified format requirements.              
The different parsing methods are thought to be useful for Language Model and for Acoustic Models.                           
                                                                                                     
SYNTAX:                                                                                              
cat file.tr | $0 [OPTIONS]                     

USEFUL OPTIONS FOR ACOUSTIC MODELING: 
   +numbers2text --Regex acoustic --SIL keep  --CurlyTokens keep --SpellID keep -dash --Apostrophe uncommon
USEFUL OPTIONS FOR LM MODELING: 
   +numbers2text --Regex acoustic --SIL remove  --CurlyTokens remove --SpellID remove -dash --Apostrophe destokenize
 OR
   +numbers2text --Regex ortographic --SIL remove  --CurlyTokens remove --SpellID remove +dash --Apostrophe uncommon

   
OPTIONS  (the default values are highlited)                                                                                  
[--Apostrophe METHOD] [\+numbers2text] [+acoustic|+ortographic]

+numbers2text : Converts integers to text.        Example: 'HI! 34'  => hi thirty-four
\e[1m-numbers2text\e[0m : Do not Convert numbers to text.   Example: 'HI! 34' => hi 34

+dash         : Do not remove dash marks.         Example: 34  => thirty-four
\e[1m-dash\e[0m         : Remove all dash marks.            Example: 34  => thirty four

--SIL ACTION   :
        remove (or del or delete) : Remove <SIL> tags.  Example: '<<x>> IS <SIL> > THAN 3' => x is than three
        keep                      : keep <SIL> tags  .  Example: '<<x>> IS <SIL> > THAN 3' => x is <SIL> than three
        \e[1mignore\e[0m                   : $sSpellID.
                                                        Example: '<<x>> IS <SIL> > THAN 3' => x is sil than three


        
--SpellID ACTION : Number in parenthesis after a word are used 
                    to specify the pronunctiation ID in the phonetic dictionary.
        remove (or del or delete) : Remove (NUMBER) in WORD(NUMBER) : Example 'AND(3) 3' => and three
        keep                      : Keep (NUMBER) in WORD(NUMBER)   : Example 'AND(3) 3' => and(3) three
        \e[1mignore\e[0m                   : $sSpellID
                                                        Example: AND(3) 3' => and three three

--CurlyTokens ACTION: Assumes the text contain special curly-bracketed tokens
        remove (or del or delete) : Remove {TOKEN}                  : Example 'HI! {{ {COUGH} ' => hi
        \e[1mkeep\e[0m                      : Keep   {TOKEN}                  : Example 'HI! {{ {COUGH} ' => hi {COUGH}
        'map to STRING'           : Convert all {TOKEN}s into the specified STRING.
        'map to {SF}'             : Convert all {TOKEN}s into the noise token {SF}
                                                                    : Example 'HI! {{ {COUGH} ' => hi {SF}
        ignore                    : Do not interpret {TOKEN} as anything special 
                                                                    : Example 'HI! {{ {COUGH} ' => hi cough
--regex PART  : Assumes the text may contain the following word transcription convention :
                /Verbatim/Ortographic/ . This was used to transcibe Polimedia corpus.
                For instance:   /eh// /um// /wo// what we /af/have/ here is 
        acoustic (or ac or left): Remove the right part of regex-like transcription expressions. 
        ortographic (or right)  : Remove the right part of regex-like transcription expressions. 
        
        The first is is usually intended to train acoustic models (prior to the transliterations), 
        while the ortographic, is useful for language model training or end user reading.
        
        \e[1mignore\e[0m                  : $sRegex

--Apostrophe METHOD:
    destokenize (or destokenise , destoken , untoken, untokenize , untokenise) :                 
        Joins two words if the first one have a trailing aposthrophe mark (') or the second a leading mark.
        example: echo \"Ey(2)! How are you?. You 're my 'best' firend amongst the 56 friend I 've ever got.\" | $0
            ey how are you you're my best firend amongst the fifty-six friend i've ever got
    tokenize (or tokenise , token ) :                                                            
        Splits words with ' mark into two separete words, but those ' marks used for highlithing.
    \e[1muncommon\e[0m :                                                                                   
        Splits words with ' mark into two separete words, but those of very common use like:     
        o'clock, ain't, AREN'T, CAN'T, COULDN'T, DIDN'T, DOESN'T, DON'T, HADN'T, HASN'T, HAVEN'T,
        ISN'T, MUSTN'T, NEEDN'T, SHOULDN'T, WASN'T, WEREN'T, WOULDN'T                            
        {i, you, he, she, it, we, they, there, that, what}'{m, s, ve, ll, d, re}                 

\[e1mCOMMON BEHAVIOUR\e[0m:
The following actions will be always performed regardless of the specified options:
 # Remove groups of issolated non-letters.
    
    - The ' marks not used as genetive will be removed.  Example: 'thing' => thing ; people's => people 's
    - All isolated symbols not in a word or 'special token' will be removed:  Ex.: This will remove ' ~;+ ' ; but not 'hello-bye' if +dash
    - Blindly Remove all symbols which are not letters, inequality simbols, blanks, aposthrophees, slashes,  braces, dashes, digits, nor middle dot ( l·l)
    - Letters are lowercased, even for UTF-8 non-common letters.

                            ----------------------------------------------------------------
                            2013-3-27 v0.1 by Isaias Sanchez-Cortina (isanchez\@dsic.upv.es)
";
exit;
}


###########
# PARSE ARGUMENTS 
while (@ARGV) {
    $_=shift(@ARGV);
    /^\--?h(elp)?/ && (help, exit(0)); 
    /^\--Apostrophe$/ && ($str_Apostrophe=shift, next); 
    /^\+numbers2text$/ && ($BOOL_numbers2text=1,next); 
    /^\-numbers2text$/ && ($BOOL_numbers2text=0,next); 
    /^--Regex$/ && ($sRegex=shift,next); 
    /^\-dash$/ && ($bRemoveDash=1,next); 
    /^\+dash$/ && ($bRemoveDash=0,next); 
    /^--SIL$/ && ($sSil=shift,next); 
    /^--SpellID$/ && ($sSpellID=shift, next) ;
    /^--CurlyTokens$/ && ($sCurlyTokens=shift, next) ;

    die "$0: [WW] Unrecognized option $_ . See help (-h)\n"
}
help() if (-t STDIN) ; 
# Check 
if ($BOOL_numbers2text && ($BOOL_Lingua_EN_Numbers=0) ) {
    print STDERR "$0: [WW] Ignoring +numbers2text: Required external package not found. To install use  'cpan install  Lingua::EN::Numbers' \n"  ;
    $BOOL_numbers2text=0;
}

##################################################################################################
# Precompile common regexps
my $REGEX_polimedia_SF;
my $REGEX_polimedia;
my $polimedia=0;
my $polimedia_SF=1;
my @REGEX_common=( qr/\/+/ );

# If polimedia conventions hold, select left (phonetic-like) or right part (orthographucally correct)
#  from tokens like :          /verbatim/ortographically correct/ 
if ($sRegex =~ /^orto(graphic)?|right/ ) {
    $polimedia=2; 
    $REGEX_polimedia_SF=qr/\/\s?SF\s?\/\s*\//i;
    $REGEX_polimedia=qr/\/(?!=SF\/)[^\/]*\/([^\/]*)\//i;
} elsif ($sRegex =~ /^ac(oustic)?|left/) {
    $polimedia=1; 
    $REGEX_polimedia_SF=qr/\/\s*SF\s*\/\s*\//i;
    $REGEX_polimedia=qr/\/(?!=SF\/)([^\/]*)\/[^\/]*\//i;
}

# <SIL> tags and < > symbols:
if ($sSil =~ /^remove|del(ete)?/i) {
    ###### Delete <SIL> tags and all <> 
    push @REGEX_common,  qr/<SIL>/i ,  qr/[\<\>]+/ ; 
} elsif ($sSil =~ /^keep/i) {
    ###### Delete <> which are not part of <SIL> tag
    push @REGEX_common,  qr/<(?!SIL>)/i ,  qr/(?<!<SIL)>/i ; 
} else {
    ###### Delete all <> caresly. 
    push @REGEX_common, qr/[\<\>]+/;
}

# {TOKENS}  and curly symbols:
my $bMapTo=0;
my $sMapTo="";
my $REGEX_TokenToMap=qr/\{\w++\}/;
if ($sCurlyTokens =~  /^map\s*to\s*(.*)/i ) {    
    # Map all {TOKENS} tp $1 , then erase the remaining {} symbols not in the resulting {$1}
    $sMapTo=$1;
    $bMapTo=1;
    push @REGEX_common , qr/\{(?!\w+\})/ , qr/(\{\w++)?+\}(?(1)\K(*SKIP))/ ;
} elsif ($sCurlyTokens =~  /^remove|del(ete)?/i ) {    
    # Erase all {TOKENS} , then erase the remaining {} symbols 
    push @REGEX_common , qr/\{\w+\}/ , qr/[\{\}]+/ ;
} elsif ($sCurlyTokens  =~ /^keep/i )  {     # Erase all {}
    # Erase {} symbols which are not part of a token like {COUGH}
    push @REGEX_common , qr/\{(?!\w+\})/ , qr/(\{\w++)?+\}(?(1)\K(*SKIP))/ ;
} else {
    # Erase {} symbols caresly
    push @REGEX_common , qr/[\{\}]+/;

}

# Pronunctiation number ID:
our $REGEX_Digits; # used by number2text
our $REGEX_Digits00=qr/^00$/;
our $bSpellID_keep;
our $REGEX_SpellID;
if ($sSpellID =~  /^remove|del(ete)?/i ) {    
    # Erase all (NUMBER) in word(NUMBER)
    $REGEX_Digits=qr/(\d(*SKIP)(?(?<=\w\(\d)\d*+(*SKIP)(?(?=\))(*FAIL))|\d*+(*SKIP)))/o;
    push @REGEX_common , qr/\w\K\(\d+\)/ ;
    $bSpellID_keep=0;
} elsif ($sSpellID =~ /^keep/i )  {     # Erase all {}
    # Keep word(NUMBER) and ensure not to transliterate them if using number2text
    $REGEX_Digits=qr/(\d(*SKIP)(?(?<=\w\(\d)\d*+(*SKIP)(?(?=\))(*FAIL))|\d*+(*SKIP)))/o;
    push @REGEX_common , qr/\((?!\d+\))/ , qr/(\(\d++)?+\)(?(1)\K(*SKIP))/ ;
    $bSpellID_keep=1;
    $REGEX_SpellID=qr/\(\d+\)$/;
} else {
    # Erase {} symbols caresly
    $REGEX_Digits=qr/(\d++)/o;
    push @REGEX_common , qr/[\(\)]+/ ;
    $bSpellID_keep=0;

}

# Dashes
push  @REGEX_common , qr/\-+/ if $bRemoveDash;
# Common substitutions:
push @REGEX_common , 
    qr/(^|\s)\K[^\p{L}]+(?=\s|$)/               , # Remove groups of issolated non-letters. Ex.: This will remove " ~;+ " ; but not "wiki+wiki"
    qr/[^\p{L}<> \x{027}\(\)\{\}\-\d·]+/        ; # Blindly Remove all symbols which are not letters, inequality simbols, blanks, aposthrophees, slashes,  braces, dashes, digits, nor middle dot ( l·l)
    # deprecated  qr/(^|\s)\K[\<\>\(\)\x{027}\{\}]+(?=\s|$)/  ;
    




############################################################################################


# Apostrophe  
my @REGEX_Apostrophe;
my $nApostrophe;
my @UNCOMMON;
if ($str_Apostrophe =~ /^(des|un)token(i[sz]e)?$/i ) {
    $nApostrophe=1;
    @REGEX_Apostrophe=( 
    qr/(^|\s)\K\x{027}+([^\x{027}\s]+)\x{027}+/,
    qr/([^s\x{027}]) +([\x{027}]\p{L})/,
    );

} elsif ($str_Apostrophe =~ /^token(i[sz]e)?$/i ) {
    $nApostrophe=2;
    @REGEX_Apostrophe=(
    qr/(^|\s)\K\x{027}+([^\x{027}\s]+)\x{027}+/,
    qr/(\p{L})([\x{027}]\p{L})/,
    );

} else {
    $nApostrophe=3;
    @REGEX_Apostrophe=(
    qr/^\x{027}([^\x{027}]*)\x{027}$/ ,
    qr/^([^\s\x{027}]++)(\x{027}[^\s\x{027}]++)$/ ,
    qr/^(o\x{027}clock|ain\x{027}t|AREN\x{027}T|CAN\x{027}T|COULDN\x{027}T|DIDN\x{027}T|DOESN\x{027}T|DON\x{027}T|HADN\x{027}T|HASN\x{027}T|HAVEN\x{027}T|ISN\x{027}T|MUSTN\x{027}T|NEEDN\x{027}T|SHOULDN\x{027}T|WASN\x{027}T|WEREN\x{027}T|WOULDN\x{027}T)$/i,
    qr/^(i|you|he|she|it|we|they|there|that|what)$/i,
    qr/^\x{027}(m|s|ve|ll|d|re)$/i
    );
    @UNCOMMON=(
    qr/(^| )o\K +(?=\x{027}clock)/i , 
    qr/(^| )(ain|AREN|CAN|COULDN|DIDN|DOESN|DON|HADN|HASN|HAVEN|ISN|MUSTN|NEEDN|SHOULDN|WASN|WEREN|WOULDN)\K (?=\x{027}T )/i , 
    qr/(^| )(i|you|he|she|it|we|they|there|that|what)\K (?=\x{027}(m|s|ve|ll|d|re) )/i 
    ) ; 

}





   my ( $bef, $aft, $m , $A , $B ); 
my $phrase;
my @w;

while(<STDIN>) {
    study; # SPEED-UP 
    $_=lc; # Work with lowercase

    # Polimedia transcription conventions:
    if ($polimedia) {
        if ($polimedia == 1 ) {
            s/$REGEX_polimedia_SF/\{SF\}/g ;
        } else {
            s/$REGEX_polimedia_SF//g ;
        }
        s/$REGEX_polimedia/$1/g
    }
    
    # NUmbers to text but for those between parenthesis if $sSpellID
    #  The notation word(NUMBER) is used in corpora like TEDLIUM to indentify the proper pronunctiation in a numbered dictionary
    $_=numbers2text($_)    if $BOOL_numbers2text ;
    
    # If map {Tokens} ...
    $_ =~ s/$REGEX_TokenToMap/$sMapTo/g  if ($bMapTo) ;
    
    # Common regex 
    for my $r (0..$#REGEX_common) {
        s/$REGEX_common[$r]/ /g;
    }

    if ($nApostrophe==1) {
        # DESTOKENIZE
        # Remove aposthrophees used for highlithing                      
        s/$REGEX_Apostrophe[0]/$2/g           ;
        # destokenize all aposthrophees but <*>s' <word>                 
        s/$REGEX_Apostrophe[1]/$1$2/g     ;
    } elsif ($nApostrophe==2) {
        #TOKENIZE
        print STDERR "Tokenizing ... \n";
        # Remove aposthrophees used for highlithing                      
        s/$REGEX_Apostrophe[0]/$2/g;
        # tokenize all aposthrophees or dashes                           
        s/$REGEX_Apostrophe[1]/$1 $2/g;
    } else {
        #UNCOMMON:
        # Tokenize by ', but for common slangs like it's, etc.
        my $s="" ; 
        for  ( split ) {
            s/$REGEX_Apostrophe[0]/$1/; # remove leading and trailing from 'word'
            unless ( $bSpellID_keep && ($_ =~ $REGEX_SpellID) ) { # do not separate if it has a spell variant ID
                if ( $_ =~ $REGEX_Apostrophe[1] ) {
                    $bef=$1 ; my $aft=$2; my $m=$bef.$aft; 
                    $_=$bef.' '.$aft  unless ( ($m =~ $REGEX_Apostrophe[2]) || ( ($bef =~ $REGEX_Apostrophe[3])&&($aft =~ $REGEX_Apostrophe[4]) ) ) ;
                }
            }
            $s.="$_ ";
        }
        $_="$s";
        # Destokenize common 
        for my $r (0 ..$#UNCOMMON)  {  s/$UNCOMMON[$r]//ig  }; 
        
    }


    s/\s*$//;
    s/^\s*+//g;
    print STDOUT "$_\n"; 
}



  # Converts all the integer in a sentence to words                   
sub numbers2text {    
    # Wrapper for Lingua       plus one rule spectific to TED corpus 
    my $STR_phrase = shift;
    while ($STR_phrase =~ /$REGEX_Digits/gc) {
        my $INT_replaceAtInitpos=$-[0];
        my $INT_replaceAtEndpos=$+[0];
        my $STR_CapturedDigits=$1;
        my $STR_TextNumber; 
        if ( $STR_CapturedDigits =~ $REGEX_Digits00 ) { 
            $STR_TextNumber="double zero";
        } else {
            $STR_TextNumber=num2en($STR_CapturedDigits );
        }
        substr( $STR_phrase,$INT_replaceAtInitpos, $INT_replaceAtEndpos-$INT_replaceAtInitpos," $STR_TextNumber ");
        my $N_substitutions++;
    }  
    return $STR_phrase;
}

