#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# Takes as imput a wav file and a trs file, and outputs wav segments according to trs syncs.
# Each segment has de following filename: [start_time]_[end_time]-[video_name].{wav|tr}, where
# tr file is its corresponding trs transcription.
# Segments will be placed in subdirs speech | background according to BG_TOKEN. It also outputs a file
# named feats_sequential.lst in which all wav filepaths are shown ordered by time.

# Changelog
# 2013-04-02 Adrià Giménez Pastor
# - More robust endTime parsing.
# 2013-03-21 Joan Albert Silvestre
# - Exits if sox failed.
# 2013-02-25 Adrià Giménez Pastor
# - Generates spks.lst
# 2013-02-22 Adrià Giménez Pastor
# - Normalized error messages
# 2013-01-16 Joan Albert Silvestre
# - Added Event tag handling: discards segments of foreign languages. Option -e
# 2012-06-12 Joan Albert Silvestre
# - Sample ID can be manually specified with option -i
# - Formatted time range for segment filenames
# 2012-02-28 Joan Albert Silvestre
# - Outputs ordered akbin filepath list 
# 2012-02-24 Joan Albert Silvestre
# - Added new cases when looking for sync text
# 2012-02-21 Joan Albert Silvestre
# - First Version

import sys
import os
import commands
from optparse import OptionParser

def check_event(s):
 
  if s.find("<Event ") > -1: # It is an event tag?
    if s.find('type="language"') > -1: # it is a language event? 
      if s.find('desc="%s"'%(LANG)) == -1: # if it is not the target language, we discard all syncs
        if s.find('extent="begin"') > -1:
          return True
        elif s.find('extent="end"') > -1:
          return False
  
	return None

# Cmdline

usage = "usage: %prog [options] <i_wav> <i_trs> <o_dir>"
parser = OptionParser(usage)
parser.add_option("-i", "--video_id", dest="video_name", type="string", help="manually specify the current sample ID")
parser.add_option("-e", "--event_handle", action="store_false", dest="eh_flag", default=True, help="Do not handle Event tags: do not discard all syncs under language tags different than target language")
parser.add_option("-b", "--bg_token", dest="bg_token", type="string", default="[sonido de fondo]", help="Defines background token. Default: '[sonido de fondo]'")
parser.add_option("-l", "--lang", dest="lang", type="string", default="_", help="Sets target language. Default: None (Discard all events)")

(options, args) = parser.parse_args()

if len(args) != 3:
	parser.error("[EE]: Incorrect number of arguments.")

wav_fn=args[0]
trs_fn=args[1]
out_dir=args[2]

if options.video_name == None:
	video_name, ext = os.path.splitext(os.path.basename(trs_fn)) # video file name without extension
else:
	video_name=options.video_name

BG_TOKEN = options.bg_token
LANG = options.lang

# Read trs file and gather syncs

trs_fd=open(trs_fn)
trs=trs_fd.readlines()
trs_fd.close()

global_end_time=0
syncs=[]; i=0;
speakers=[]
last_spk= 'UNK'

discard_flag=False;

while i < len(trs):
 
  res=check_event(trs[i])
  if res != None: discard_flag=res
  if trs[i].find("<Sync time=") > -1: # <Sync> tag
    
    time=float(trs[i].split('"')[1]) # get sync time
    tr=""
    i+=1

    if i >= len(trs): 
      sys.stderr.write('[WW]: possibly bad transcription for %s at sync time %s.\n'%(trs_fn,time))
      break; # end of file (bad transcription?)

    # now we look for current sync transcription
    while trs[i].find("<Sync time=") == -1: # until we see next <sync> tag
      res=check_event(trs[i]) # check if current line is an event tag
      if res != None: # is an event tag
        discard_flag=res
        i+=1
      elif trs[i][0] == '<': # another tag different than <Event>
        i+=1 # skip it
      elif len(trs[i].split()) == 0: # empty line
        i+=1 # skip it
      else: # I hope text :)
        tr=trs[i] # save it
        break; # exit loop, we have found the text
    
      if i >= len(trs): 
        sys.stderr.write('[WW]: possibly bad transcription for %s at sync time %s.\n'%(trs_fn,time))
        break; # end of file (bad transcription?)
		
    if tr != "": # only if transcription text found
      syncs.append((time,trs[i],discard_flag))
      speakers.append(last_spk)
      i+=1
    else:
      sys.stderr.write('[WW]: possibly bad transcription (empty text) for %s at sync time %s.\n'%(trs_fn,time))

  elif trs[i].find("<Section") > -1:
    aux= trs[i].split('"')
    length= len(aux)
    for j in xrange(0,length):
      if aux[j].find("endTime=") > -1: break
    global_end_time= float(aux[j+1]) if j<length-1 else None
    if global_end_time==None: sys.exit('[EE]: section without endTime')
    i+=1
    
  elif trs[i].find("<Turn") > -1:
    aux= trs[i].split('"')
    length= len(aux)
    for j in xrange(0,length):
      if aux[j].find("speaker=") > -1: break
    last_spk= aux[j+1] if j<length-1 else 'UNK'
    i+=1

  else:
    i+=1

# Create output dirs; segments will be placed in speech/background subdirs according to BG_TOKEN

if os.path.isdir('%s/sp'%(out_dir)) == False:
  os.makedirs('%s/sp'%(out_dir))
  os.makedirs('%s/bg'%(out_dir))

# File in which we will print wav filepaths ordered in time
wavlist_fd=open('%s/wavs_sequential.lst'%(out_dir),'w')

# Loop over all syncs
fspk=open('%s/spks.lst'%out_dir,'w')
for si in xrange(len(syncs)):
  
  if si+1 < len(syncs): # check if current sync is not the last one

    if (syncs[si][2] == False and options.eh_flag == True) or options.eh_flag == False: # discard flag was not enabled for this sync (Event in other language)

      seg_start_time=syncs[si][0] # starting time of current sync/segment
      seg_end_time=syncs[si+1][0] # ending time of current sync/segment
      seg_length=seg_end_time-seg_start_time

      # check if current sync/segment is non-speech (background)
      stype="sp"
      if syncs[si][1].find(BG_TOKEN) > -1:
        stype="bg"

			# Trim wav file
      seg_wav_fn=out_dir+"/"+stype+"/"+str("{0:.3f}".format(seg_start_time).zfill(9))+"_"+str("{0:.3f}".format(seg_end_time).zfill(9))+"-"+video_name+".wav" # wav segment filename
      res = commands.getstatusoutput('sox %s %s trim %s %s'%(wav_fn,seg_wav_fn,str("{0:.3f}".format(seg_start_time).zfill(9)),str("{0:.3f}".format(seg_length).zfill(9))))
      if res[0] != 0: 
        sys.stderr.write("[EE]: Failed to trim wav: %s\n"%(res[1]))
        sys.exit(1)

      # output transcription of current sync/segment
      seg_tr_fn=out_dir+"/"+stype+"/"+str("{0:.3f}".format(seg_start_time).zfill(9))+"_"+str("{0:.3f}".format(seg_end_time).zfill(9))+"-"+video_name+".tr" # transcription filename
      seg_tr_fd=open(seg_tr_fn,'w')
      seg_tr_fd.write(syncs[si][1]) # write transcription
      seg_tr_fd.close()
      
      # Output speakers
      if stype=='sp':
        fspk.write((u'%s\n'%speakers[si]).encode('utf-8'))
        fspk.write("%s\n"%(str("{0:.3f}".format(seg_start_time).zfill(9))+"_"+str("{0:.3f}".format(seg_end_time).zfill(9))+"-"+video_name))
      
			# output wav filename
      wavlist_fd.write('%s\n'%(os.path.abspath(seg_wav_fn)))
      
  else: # special case for last sync

    # Case for last sync/segment

    if (syncs[si][2] == False and options.eh_flag == True) or options.eh_flag == False: # discard flag was not enabled for this sync (Event in other language)

      seg_start_time=syncs[si][0] # starting time of last segment
      seg_end_time=global_end_time # ending time of last segment
      seg_length=seg_end_time-seg_start_time

      # check if current sync/segment is non-speech (background)
      stype="sp"
      if syncs[si][1].find(BG_TOKEN) > -1:
        stype="bg"

			# Trim wav file
      seg_wav_fn=out_dir+"/"+stype+"/"+str("{0:.3f}".format(seg_start_time).zfill(9))+"_"+str("{0:.3f}".format(seg_end_time).zfill(9))+"-"+video_name+".wav" # wav segment filename
      res = commands.getstatusoutput('sox %s %s trim %s %s'%(wav_fn,seg_wav_fn,str("{0:.3f}".format(seg_start_time).zfill(9)),str("{0:.3f}".format(seg_length).zfill(9))))
      if res[0] != 0: 
        sys.stderr.write("[EE]: Failed to trim wav: %s\n"%(res[1]))
        sys.exit(1)

      # output transcription of current sync/segment
      seg_tr_fn=out_dir+"/"+stype+"/"+str("{0:.3f}".format(seg_start_time).zfill(9))+"_"+str("{0:.3f}".format(seg_end_time).zfill(9))+"-"+video_name+".tr" # transcription filename
      seg_tr_fd=open(seg_tr_fn,'w')
      seg_tr_fd.write(syncs[si][1]) # write transcription
      seg_tr_fd.close() 
      		
      # Output speakers
      if stype=='sp':
        fspk.write((u'%s\n'%speakers[si]).encode('utf-8'))
        fspk.write("%s\n"%(str("{0:.3f}".format(seg_start_time).zfill(9))+"_"+str("{0:.3f}".format(seg_end_time).zfill(9))+"-"+video_name))

			# output wav filename
      wavlist_fd.write('%s\n'%(os.path.abspath(seg_wav_fn)))
  
wavlist_fd.close()
fspk.close()
