#!/bin/bash

#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

READLINK=`which greadlink`; READLINK=${READLINK:-readlink}; READLINK="$READLINK -f"

set -e

if [[ -z "$BPATH" ]]
then
    BPATH=`$READLINK $0`
    BPATH=`dirname $BPATH`
fi

# Changelog
# 2013-03-15 Joan Albert Silvestre
# - option "-aachen" renamed as "-hes"
# - Error & exit when "-hes" option enabled on phonetic transcription mode.
# - Modified usage string.
# 2013-03-15 Adrià Giménez Pastor
# - HES -> HESITATION
# 2013-02-22 Adrià Giménez Pastor
# - HESITATION -> HES
# 2012-10-23 Joan Albert
# - Now tolerates non-isolated substitutions, like "que/e//que"
# 2012-10-22 Joan Albert
# - Added translit option
# - Added aachen option
# 2012-10-21 Jesús Andrés
# - Several improvements
# 2012-05-30 Joan Albert Silvestre
# - Text selection on substitutions is not a mandatory argument (moved to -sub option), 
#   this decision make this script suitable for other purposes.
# 2012-05-29 Joan Albert Silvestre
# - Speed improvements (approx. 2x faster)
# 2012-05-17 Joan Albert Silvestre
# - First Version

USAGE="Usage: `basename $0` {txt|pho} [options] < input_string\n\n
\tA tool for parsing transcriptions, based on poliMedia transcription Guidelines.\n\n
\tMandatory arguments:\n
\t\t{txt|pho}: Set output mode: text or phonetic transcription.\n\n
\tGeneral Options:\n
\t\t-sub {am|lm}: Get acoustic text or corrected text on substitutions /am/lm/ .\n
\t\t\tIf this option is not enabled, substitutions will be preserved.\n
\t\t-conv: Convert from ISO-8859-1 to UTF-8.\n
\t\t-hes: Uses [HESITATION] in the acoustic part. Not compatible in phonetic transcription mode.\n
\t\t-translit: Transliterates numbers.\n
\t\t-h: Show this help.\n\n
\t{pho} Options:\n
\t\t-sp: Add Short Pause (SP) at the end of every word, and at the beginning and the end of the sentence.\n
\t\t-sil: Silence (SIL) at the beginning and the end of the sentence, and Short Pause (SP) at the end of every word.\n\n
\t{txt} Options:\n
\t\t-lc: Lowercase text.\n
\t\t-noalpha: Remove non-alphanumerical chars.\n
\t\t-notilde: Remove tildes.\n
\t\t-nodiaeresis: Remove diaeresis.\n
"

export LC_ALL="C.UTF-8" 2>/dev/null # Macintosh tweak

if [ $# -lt 1 ]; then
    echo -e $USAGE 1>&2;
    exit
fi

if [ $1 != "txt" -a $1 != "pho" ]; then
    echo -e $USAGE 1>&2;
    exit
fi

tt=$1
part=$2

phonetic_transcriptor="$BPATH/Syllables.perl -sp -sil -longsil +SIMPLE -t +c"

sbns="SF" # short background noise symbol
sp="SP"; sil="SIL"

sp_opt=0; sil_opt=0; lc_opt=0; noalpha_opt=0; notilde_opt=0; conv_opt=0; sub_opt=0;
nodieresis_opt=0; hes_opt=0; translit_opt=0;

while [ "${2:0:1}" = "-" ]; do
    case $2 in
	-sub)
	    sub_opt="1"
	    sub=$3
	    shift 2
	    ;;
	-sp)
	    sp_opt="1"
	    shift 1
	    ;;
	-sil)
	    sil_opt="1"
	    shift 1
	    ;;
	-lc) # lowercase
	    lc_opt="1"
	    shift 1
	    ;;
	-noalpha) # elimina signes puntuació
	    noalpha_opt="1"
	    shift 1
	    ;;
	-notilde) # elimina accents
	    notilde_opt="1"
	    shift 1
	    ;;
	-nodiaeresis) # elimina dieresies
	    nodieresis_opt="1"
	    shift 1
	    ;;
	-conv) 
	    conv_opt="1"
	    shift 1
	    ;;
    -hes) # hesitations
	    hes_opt="1"
			shift 1
			;;
    -translit) # translitera
	    translit_opt="1"
			shift 1
			;;
	-h)
	    echo -e $USAGE 1>&2;
	    exit
	    ;;
	*)
	    echo "`basename $0`: Unknown option $1." 1>&2;
	    echo -e $USAGE 1>&2;
	    exit
	    ;;
    esac
done

[ $tt = "pho" -a $hes_opt -eq 1 ] && { echo "[EE] Cannot output [HESITATION] token on phonetic transcription mode" 1>&2; exit 1; }

txt=$(cat /dev/stdin | perl -pe 's/\r\n/\n/')

###### Pre-conversion from ISO to UTF-8, if applicable

if [ $conv_opt -eq 1 ]; then

    TMP=${TMPDIR:-/tmp}/$(basename $0 .sh).$$; mkdir $TMP 2>/dev/null;

    echo "$txt" > $TMP/intxt;

    #if [ `file $TMP/intxt | grep ISO-8859 | wc -l` -gt 0 ]; then # si l'entrada és ISO-8859
	txt=`cat $TMP/intxt | iconv -f ISO-8859-1 -t UTF-8`
    #fi

    rm -r $TMP

fi

if [[ $sil_opt -eq 0 ]]
then
    txt=$(echo "$txt" | sed "s|/$sbns//||g") # Remove SF
fi

txt=$(echo "$txt" | sed -e 's|\(/[^/]*/[^/]*/\)| \1 |g') # isolate subs, i.e. "que/e//que" -> "que /e// que"

if [ $translit_opt -eq 1 ]; then # translitera fins 999.999.999.999.999.999 (bilions)

  # elimina zeros a l'esquerra
  txt=$(echo "$txt" |
	 sed -e 's/ 0\{1,\}\([0-9]*\)\([ ]\{1,\}\|$\)/ \1 /g')
   
  # bilions
	txt=$(echo "$txt" |
	 sed -e 's/ 1\([0-9]\{12\}\)\([ ]\{1,\}\|$\)/ un billón \1 /g' \
   -e 's/ \([1-9]\)\([0-9]\{0,5\}\)\([0-9]\{12\}\)\([ ]\{1,\}\|$\)/ \1\2 billones \3 /g')

	# elimina zeros a l'esquerra
  txt=$(echo "$txt" |
	 sed -e 's/ 0\{1,\}\([0-9]*\)\([ ]\{1,\}\|$\)/ \1 /g')

  # milions
	txt=$(echo "$txt" |
	 sed -e 's/ 1\([0-9]\{6\}\)\([ ]\{1,\}\|$\)/ un millón \1 /g' \
   -e 's/ \([1-9]\)\([0-9]\{0,5\}\)\([0-9]\{6\}\)\([ ]\{1,\}\|$\)/ \1\2 millones \3 /g')

	# elimina zeros a l'esquerra
  txt=$(echo "$txt" |
	 sed -e 's/ 0\{1,\}\([0-9]*\)\([ ]\{1,\}\|$\)/ \1 /g')

  # milers
  txt=$(echo "$txt" |
	 sed -e 's/ 1\([0-9]\{3\}\)\([ ]\{1,\}\|$\)/ mil \1 /g' \
	 -e 's/ \([1-9]\)\([0-9]\{0,2\}\)\([0-9]\{3\}\)\([ ]\{1,\}\|$\)/ \1\2 mil \3 /g')

  # centenes
	txt=$(echo "$txt" |
	 sed -e 's/ 100\([ ]\{1,\}\|$\)/ cien /g' \
	 -e 's/ 1\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ ciento \1 /g' \
	 -e 's/ 2\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ doscientos \1 /g' \
	 -e 's/ 3\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ trescientos \1 /g' \
	 -e 's/ 4\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ cuatrocientos \1 /g' \
	 -e 's/ 5\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ quinientos \1 /g' \
	 -e 's/ 6\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ seiscientos \1 /g' \
	 -e 's/ 7\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ setecientos \1 /g' \
	 -e 's/ 8\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ ochocientos \1 /g' \
	 -e 's/ 9\([0-9][0-9]\)\([ ]\{1,\}\|$\)/ novecientos \1 /g')

	# elimina zeros a l'esquerra
  txt=$(echo "$txt" |
	 sed -e 's/ 0\{1,\}\([0-9]*\)\([ ]\{1,\}\|$\)/ \1 /g')
  
	# desenes
  txt=$(echo "$txt" |
	 sed -e 's/ 10\([ ]\{1,\}\|$\)/ diez /g' \
	 -e 's/ 11\([ ]\{1,\}\|$\)/ once /g' \
	 -e 's/ 12\([ ]\{1,\}\|$\)/ doce /g' \
	 -e 's/ 13\([ ]\{1,\}\|$\)/ trece /g' \
	 -e 's/ 14\([ ]\{1,\}\|$\)/ catorce /g' \
	 -e 's/ 15\([ ]\{1,\}\|$\)/ quince /g' \
	 -e 's/ 16\([ ]\{1,\}\|$\)/ dieciséis  /g' \
	 -e 's/ 17\([ ]\{1,\}\|$\)/ diecisiete /g' \
	 -e 's/ 18\([ ]\{1,\}\|$\)/ dieciocho /g' \
	 -e 's/ 19\([ ]\{1,\}\|$\)/ diecinueve /g' \
	 -e 's/ 20\([ ]\{1,\}\|$\)/ veinte /g' \
	 -e 's/ 21\([ ]\{1,\}\|$\)/ veintiuno /g' \
	 -e 's/ 22\([ ]\{1,\}\|$\)/ veintidós /g' \
	 -e 's/ 23\([ ]\{1,\}\|$\)/ veintitrés /g' \
	 -e 's/ 24\([ ]\{1,\}\|$\)/ veinticuatro /g' \
	 -e 's/ 25\([ ]\{1,\}\|$\)/ veinticinco /g' \
	 -e 's/ 26\([ ]\{1,\}\|$\)/ veintiséis /g' \
	 -e 's/ 27\([ ]\{1,\}\|$\)/ veintisiete /g' \
	 -e 's/ 28\([ ]\{1,\}\|$\)/ veintiocho /g' \
	 -e 's/ 29\([ ]\{1,\}\|$\)/ veintinueve /g' \
	 -e 's/ 30\([ ]\{1,\}\|$\)/ treinta /g' \
	 -e 's/ 31\([ ]\{1,\}\|$\)/ treinta y uno /g' \
	 -e 's/ 32\([ ]\{1,\}\|$\)/ treinta y dos /g' \
	 -e 's/ 33\([ ]\{1,\}\|$\)/ treinta y tres /g' \
	 -e 's/ 34\([ ]\{1,\}\|$\)/ treinta y cuatro /g' \
	 -e 's/ 35\([ ]\{1,\}\|$\)/ treinta y cinco /g' \
	 -e 's/ 36\([ ]\{1,\}\|$\)/ treinta y seis  /g' \
	 -e 's/ 37\([ ]\{1,\}\|$\)/ treinta y siete /g' \
	 -e 's/ 38\([ ]\{1,\}\|$\)/ treinta y ocho /g' \
	 -e 's/ 39\([ ]\{1,\}\|$\)/ treinta y nueve /g' \
	 -e 's/ 40\([ ]\{1,\}\|$\)/ cuarenta /g' \
	 -e 's/ 41\([ ]\{1,\}\|$\)/ cuarenta y uno /g' \
	 -e 's/ 42\([ ]\{1,\}\|$\)/ cuarenta y dos /g' \
	 -e 's/ 43\([ ]\{1,\}\|$\)/ cuarenta y tres /g' \
	 -e 's/ 44\([ ]\{1,\}\|$\)/ cuarenta y cuatro /g' \
	 -e 's/ 45\([ ]\{1,\}\|$\)/ cuarenta y cinco /g' \
	 -e 's/ 46\([ ]\{1,\}\|$\)/ cuarenta y seis  /g' \
	 -e 's/ 47\([ ]\{1,\}\|$\)/ cuarenta y siete /g' \
	 -e 's/ 48\([ ]\{1,\}\|$\)/ cuarenta y ocho /g' \
	 -e 's/ 49\([ ]\{1,\}\|$\)/ cuarenta y nueve /g' \
	 -e 's/ 50\([ ]\{1,\}\|$\)/ cincuenta /g' \
	 -e 's/ 51\([ ]\{1,\}\|$\)/ cincuenta y uno /g' \
	 -e 's/ 52\([ ]\{1,\}\|$\)/ cincuenta y dos /g' \
	 -e 's/ 53\([ ]\{1,\}\|$\)/ cincuenta y tres /g' \
	 -e 's/ 54\([ ]\{1,\}\|$\)/ cincuenta y cuatro /g' \
	 -e 's/ 55\([ ]\{1,\}\|$\)/ cincuenta y cinco /g' \
	 -e 's/ 56\([ ]\{1,\}\|$\)/ cincuenta y seis  /g' \
	 -e 's/ 57\([ ]\{1,\}\|$\)/ cincuenta y siete /g' \
	 -e 's/ 58\([ ]\{1,\}\|$\)/ cincuenta y ocho /g' \
	 -e 's/ 59\([ ]\{1,\}\|$\)/ cincuenta y nueve /g' \
	 -e 's/ 60\([ ]\{1,\}\|$\)/ sesenta /g' \
	 -e 's/ 61\([ ]\{1,\}\|$\)/ sesenta y uno /g' \
	 -e 's/ 62\([ ]\{1,\}\|$\)/ sesenta y dos /g' \
	 -e 's/ 63\([ ]\{1,\}\|$\)/ sesenta y tres /g' \
	 -e 's/ 64\([ ]\{1,\}\|$\)/ sesenta y cuatro /g' \
	 -e 's/ 65\([ ]\{1,\}\|$\)/ sesenta y cinco /g' \
	 -e 's/ 66\([ ]\{1,\}\|$\)/ sesenta y seis  /g' \
	 -e 's/ 67\([ ]\{1,\}\|$\)/ sesenta y siete /g' \
	 -e 's/ 68\([ ]\{1,\}\|$\)/ sesenta y ocho /g' \
	 -e 's/ 69\([ ]\{1,\}\|$\)/ sesenta y nueve /g' \
	 -e 's/ 70\([ ]\{1,\}\|$\)/ setenta /g' \
	 -e 's/ 71\([ ]\{1,\}\|$\)/ setenta y uno /g' \
	 -e 's/ 72\([ ]\{1,\}\|$\)/ setenta y dos /g' \
	 -e 's/ 73\([ ]\{1,\}\|$\)/ setenta y tres /g' \
	 -e 's/ 74\([ ]\{1,\}\|$\)/ setenta y cuatro /g' \
	 -e 's/ 75\([ ]\{1,\}\|$\)/ setenta y cinco /g' \
	 -e 's/ 76\([ ]\{1,\}\|$\)/ setenta y seis  /g' \
	 -e 's/ 77\([ ]\{1,\}\|$\)/ setenta y siete /g' \
	 -e 's/ 78\([ ]\{1,\}\|$\)/ setenta y ocho /g' \
	 -e 's/ 79\([ ]\{1,\}\|$\)/ setenta y nueve /g' \
	 -e 's/ 80\([ ]\{1,\}\|$\)/ ochenta /g' \
	 -e 's/ 81\([ ]\{1,\}\|$\)/ ochenta y uno /g' \
	 -e 's/ 82\([ ]\{1,\}\|$\)/ ochenta y dos /g' \
	 -e 's/ 83\([ ]\{1,\}\|$\)/ ochenta y tres /g' \
	 -e 's/ 84\([ ]\{1,\}\|$\)/ ochenta y cuatro /g' \
	 -e 's/ 85\([ ]\{1,\}\|$\)/ ochenta y cinco /g' \
	 -e 's/ 86\([ ]\{1,\}\|$\)/ ochenta y seis  /g' \
	 -e 's/ 87\([ ]\{1,\}\|$\)/ ochenta y siete /g' \
	 -e 's/ 88\([ ]\{1,\}\|$\)/ ochenta y ocho /g' \
	 -e 's/ 89\([ ]\{1,\}\|$\)/ ochenta y nueve /g' \
	 -e 's/ 90\([ ]\{1,\}\|$\)/ noventa /g' \
	 -e 's/ 91\([ ]\{1,\}\|$\)/ noventa y uno /g' \
	 -e 's/ 92\([ ]\{1,\}\|$\)/ noventa y dos /g' \
	 -e 's/ 93\([ ]\{1,\}\|$\)/ noventa y tres /g' \
	 -e 's/ 94\([ ]\{1,\}\|$\)/ noventa y cuatro /g' \
	 -e 's/ 95\([ ]\{1,\}\|$\)/ noventa y cinco /g' \
	 -e 's/ 96\([ ]\{1,\}\|$\)/ noventa y seis  /g' \
	 -e 's/ 97\([ ]\{1,\}\|$\)/ noventa y siete /g' \
	 -e 's/ 98\([ ]\{1,\}\|$\)/ noventa y ocho /g' \
	 -e 's/ 99\([ ]\{1,\}\|$\)/ noventa y nueve /g')
	
	# unitats
	txt=$(echo "$txt" |
	 sed -e 's/ 1\([ ]\{1,\}\|$\)/ uno /g' \
	 -e 's/ 2\([ ]\{1,\}\|$\)/ dos /g' \
	 -e 's/ 3\([ ]\{1,\}\|$\)/ tres /g' \
	 -e 's/ 4\([ ]\{1,\}\|$\)/ cuatro /g' \
	 -e 's/ 5\([ ]\{1,\}\|$\)/ cinco /g' \
	 -e 's/ 6\([ ]\{1,\}\|$\)/ seis /g' \
	 -e 's/ 7\([ ]\{1,\}\|$\)/ siete /g' \
	 -e 's/ 8\([ ]\{1,\}\|$\)/ ocho /g' \
	 -e 's/ 9\([ ]\{1,\}\|$\)/ nueve /g')

fi

####### Acoustic/Corrected

if [ $sub_opt -eq 1 ]; then
    
    if [ $sub == "am" ]; then
			
	    if [ $hes_opt -eq 1 ]; then
          txt=$(echo "$txt" | sed -e 's@/[em]//@ [HESITATION] @g')
		fi
			
		txt=$(echo "$txt" | sed -e 's@/\([^/]*\)/\([^/]*\)/@ \1 @g')

    elif [ $sub == "lm" ]; then
		  txt=$(echo "$txt" | sed -e 's@/[em]//@ [HESITATION] @g' | sed -e 's@/\([^/]*\)/\([^/]*\)/@ \2 @g')
    else
    	echo "`basename $0`: Unknown argument $sub for -sub option." 1>&2;
    	echo -e $USAGE 1>&2;
	    exit
    fi

fi

####### Text

if [ $tt = "txt" ]; then
    
    if [ $noalpha_opt -eq 1 ]; then
        txt=$(echo "$txt" | sed 's/\[HESITATION\]/ltHESITATIONgt/g' | sed 's/\[hesitation\]/ltHESITATIONgt/g' | sed 's/[[:punct:]]/ /g' | sed 's/ltHESITATIONgt/[HESITATION]/g' | sed 's/lthesitationgt/[HESITATION]/g');
    fi
    
    if [ $notilde_opt -eq 1 ]; then
	txt=$(echo "$txt" |sed -e 's/á/a/g' -e 's/é/e/g' -e 's/í/i/g' -e 's/ó/o/g' -e 's/ú/u/g' -e 's/à/a/g' -e 's/è/e/g' -e 's/ì/i/g' -e 's/ò/o/g' -e 's/ù/u/g' \
	    -e 's/Á/A/g' -e 's/É/E/g' -e 's/Í/I/g' -e 's/Ó/O/g' -e 's/Ú/U/g' -e 's/À/A/g' -e 's/È/E/g' -e 's/Ì/I/g' -e 's/Ò/O/g' -e 's/Ù/U/g');
    fi
    if [ $nodieresis_opt -eq 1 ]; then
	txt=$(echo "$txt" |sed -e 's/ä/a/g' -e 's/ë/e/g' -e 's/ï/i/g' -e 's/ö/o/g' -e 's/ü/u/g' -e 's/Ä/A/g' -e 's/Ë/E/g' -e 's/Ï/I/g' -e 's/Ö/O/g' -e 's/Ü/U/g')
    fi;

    if [ $lc_opt -eq 1 ]; then # lowercase
	txt=$(echo "$txt" | sed 's/./\l&/g') # text original. no funciona al mac
	#txt=$(echo "$txt" | tr '[:upper:]' '[:lower:]') # version modificada fent servir tr per al mac
    fi

####### Phonetic

else

    txt=$(echo "$txt" | sed 's/[[:punct:]]/ /g') 

    if [ $sil_opt -eq 1 ]; then # transcripció valida per a experiments amb silenci llarg (sil) i curt (sp) [ sil a principi i final de frase, i substituint SF; sp després de cada paraula. ]

		#txt=`echo "$txt" | sed "s/ $sbns / /g"` # Eliminar SF
	
	txt=$(echo $txt |
	    gawk -v pt="$phonetic_transcriptor" -v SF="$sbns" '
{
for(i=1;i<=NF;++i)
  {if($i==SF) {print "SIL"}
   else { system("echo "$i" | "pt) }
  }
}' |
	    gawk '
{aux=$1;for(i=2;i<=NF;++i)aux=aux" "$i;lines[NR]= aux}
END {
if (lines[1]=="SIL") ret= "SIL"
else ret= "SIL "lines[1]
prev= lines[1]
for ( i= 2; i <= NR; ++i )
{
  if ( lines[i] == "SIL" )
   {
   if ( prev != "SIL" ) ret= ret" SIL"
   }
  else if ( prev != "SIL" ) ret= ret" SP "lines[i]
  else                      ret= ret" "lines[i]
  prev= lines[i]
}
if ( prev != "SIL" ) ret= ret" SIL"
print ret
}')
	
    elif [ $sp_opt -eq 1 ]; then # transcripció valida per a experiments amb únicament curt (sp) [ sp a principi i final de frase, i entre paraules. Es descarten SF's. ]

		#txt=`echo $txt | sed "s/ $sbns / /g"` # Eliminar SF

	txt=`echo $txt | gawk -v pt="$phonetic_transcriptor" '{for(i=1;i<=NF;++i) print $i | pt}' | gawk '{printf("SP %s ",$0)} END{print "SP"}'`

    else # transcripció estàndar sense silencis

		#txt=`echo $txt | sed "s/ $sbns / /g"` # Eliminar SF

	txt=`echo $txt | gawk -v pt="$phonetic_transcriptor" '{for(i=1;i<=NF;++i) print $i | pt}' | gawk '{printf("%s ",$0)} END{print ""}'`
	
    fi		

fi

txt=$(echo "$txt" | sed -e 's/\([[:space:]]\)[[:space:]]*/\1/g' -e 's/^[[:space:]]\{1,\}//g' -e 's/[[:space:]]\{1,\}$//g') # remove extra spaces

echo "$txt"

