#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tLtask-preprocess.py: Performs the preprocess of videos into TLK fileformats
# Copyright 2013,2014 transLectures-UPV Team
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#include i18n
import gettext
_ = gettext.gettext

import argparse
import ConfigParser
from sys import argv, exit, stdin, stdout, stderr
import os
import io
import subprocess
import shlex
import datetime
import locale
import re

def fatal_err(message):
  stderr.write("[EE] "+ message + "\n")
  exit(1)

def debug_msg(message):
  stderr.write("[II] "+ message + "\n")


#####################################################################
#####################################################################
#####################################################################
def main():
  #i18n support
  locale.setlocale(locale.LC_MESSAGES, '')
  locale.setlocale(locale.LC_CTYPE, '')
  PACKAGE = "tlk"
  LOCALEDIR = ""
  if "PACKAGE" in os.environ:
    PACKAGE = os.environ["PACKAGE"]
  if "LOCALEDIR" in os.environ:
    LOCALEDIR = os.environ["LOCALEDIR"]
  gettext.bindtextdomain(PACKAGE, LOCALEDIR)
  gettext.textdomain(PACKAGE)

  parser = argparse.ArgumentParser(
      prog="tLtask-preprocess",
      description=_('Preprocess media and transcription files for use with TLK tools'),
      epilog=_("""tLtask-preprocess prepares data for it to be used with the TLK tools. Given a
set of media files and their annotation, this tool performs the feature extraction of
samples and their corresponding phonetic annotation. Additionally, it also
extracts clusters from the input audio, which can be used for speaker or video
adaptation steps."""))

  parser.add_argument(dest='language',
    help=_("Sets the language of the documents to extract."),
    choices=['es', 'ca', 'it'])
  parser.add_argument(dest='input_format',
    help=_("Format of the input transcripts"),
    choices=['dfxp', 'trs', 'stm', 'srt', 'rt'])
  parser.add_argument(dest='input_directory',
    help=_("Input directory or file list with the transcriptions and media files"))
  parser.add_argument(dest='output',
    help=_("Output directory"))

  parser.add_argument('--norm',
    help=_('Normalize each sample by the mean and variance of the cluster of speaker or\
 media file it belongs to (by default: video)' ),
      choices=['video', 'spk', 'non'], default='video')
  parser.add_argument('-v',
    help=_('Output additional verbose information'),
    action='store_true')
  parser.add_argument('--version',
    help=_('Output version information and exit'),
    action='store_true')

  args = parser.parse_args(argv[1:])

  #wrap tLtask-preprocess.sh
  dirname = os.path.dirname(os.path.realpath(argv[0]))

  os.putenv('TLLANG', args.language)
  os.putenv('INPUTFORMAT', args.input_format)
  os.putenv('TL_PREPROCESS_LANG', "%s/%s" %(dirname, args.language))
  os.putenv('TL_PREPROCESS_COMMON', "%s/common" %(dirname))
  os.putenv('TL_PREPROCESS_EXTERNAL', "%s/external" %(dirname))
  os.putenv('TLUTILS', "%s/common/Utils.bash" %(dirname))
  os.putenv('CMVN', args.norm)
  os.putenv('PRGNAME', 'tLtask-preprocess')

  if args.norm == 'video':
    os.putenv('CLUSTERS_heuristic', 'byProps')
    os.putenv('CLUSTERS_heuristici_select', 'MEDIA_NAME')
  elif args.norm == 'spk':
    os.putenv('CLUSTERS_heuristic', 'byProps')
    os.putenv('CLUSTERS_heuristici_select', 'MEDIA_NAME SPEAKER')

  command = "bash %s/%s/tLinitprj.sh %s %s" %(
      dirname,
      args.language,
      args.input_directory,
      args.output)

  cmd = shlex.split(command)
  try:
    ret = subprocess.call(cmd)
  except Exception as e:
    fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(command, e) ) )
  if ret != 0:
    fatal_err( _( "Failed when executing: '%s' with exit code: %d" ) %(command, ret))


####################################################################################################
if __name__ == "__main__":
  main()

