# -*- coding: utf-8 -*-

#
#  Copyright 2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

""" Python utils for TLK """


__all__= [ 'Bernoulli', 'HMM', 'AModel', 'LogLinear', 'DGaussian',
           'Mixture', 'TiedStates', 'Trans', 'WrongAModel', 'Hybrid' ]


from array import array
from math import exp, log, sqrt




#############
# UTILITATS #
#############

_INF= float('-inf')

class ParseFile :
    
    def __init__ ( self, fd ) :
        self.fd= fd
        self.buffer= fd.readline ()
        if self.buffer == "" : self.eof= True
        else                 : self.eof= False
    
    def gtoken ( self ) :
        token= ""
        while not self.eof and token == "" :
            aux= self.buffer.split ( None, 1 )
            if len(aux) > 1 :
                token, self.buffer= aux
            elif len(aux) == 1 :
                token= aux[0]
                self.buffer= ""
            else :
                self.buffer= self.fd.readline ()
                if self.buffer == "" : self.eof= True
        return token


def atop ( s ) :
    if s == '-INF' : return 0.0
    else           : return exp(float(s))

def atolp ( s ) :
    if s == '-INF' : return _INF
    else           : return float(s)


def ptoa ( s ) :
    if s == 0 : return '-INF'
    else      : return str(log(s))

def lptoa ( s ) :
    if s == _INF : return '-INF'
    else         : return str(s)


class Empty :
    def __init__ ( self ) :
        pass




##############
# EXCEPCIONS #
##############

class WrongAModel(Exception):
    """Invalid acoustic model"""
    def __init__ ( self, value, state= None ) :
        if state == None :
            self.value= value
        else:
            self.value= '%s %s' % (str(state),value)
    def __str__ ( self ) :
        return self.value




##################
# TIPUS (CMODEL) #
##################

class AModelType:
    def __init__ ( self, copy_global, write_global, read_global, STATE ) :
        self.copy_global= copy_global
        self.write_global= write_global
        self.read_global= read_global
        self.STATE= STATE

_cmodel_types= {}


## Bernoulli ##

def Bernoulli_copy_global ( fro, to ) :
    to.D= fro.D


def Bernoulli_write_global ( self, to ) :
    to.write ( 'D %d\n' % self.D )

    
def Bernoulli_read_global ( self, p ) :
    if p.gtoken () != 'D' :
        raise WrongAModel("unable to read the feature vector dimensio D")
    token= p.gtoken ()
    if token == '' :
        raise WrongAModel("unable to read the value of D")
    self.D= int(token)
    if self.D <= 0 :
        raise WrongAModel("D must be positive")


class Bernoulli:
    """Bernoulli"""

    def __init__ ( self, cmodel= None, p= None, q= None ) :
        if cmodel == None :
            return
        if p.gtoken () != 'P' :
            raise WrongAModel("unable to read emission probabilities",q)
        self.p= array ( 'd' )
        for d in xrange(0,cmodel.D):
            token= p.gtoken ()
            if token == '' :
                raise WrongAModel("unable to read P[%d]" % d+1,q)
            self.p.append( atop(token) )
    

    def copy ( self ) :
        """Copy Bernoulli"""
        berno= Bernoulli ()
        berno.p= array ( 'd', list(self.p) )
        return berno
    
    
    def write ( self, to, cmodel ) :
        """Write Bernoulli"""
        to.write ( 'P' )
        for p in self.p :
            to.write ( ' %s' % ptoa(p) )
        to.write ( '\n' )


_cmodel_types['Bernoulli']= AModelType ( Bernoulli_copy_global,
                                         Bernoulli_write_global,
                                         Bernoulli_read_global,
                                         Bernoulli )


## Hybrid ##

def Hybrid_copy_global ( fro, to ) :
    to.D= fro.D

def Hybrid_write_global ( self, to ) :
    to.write ( 'D %d\n' % self.D )

def Hybrid_read_global ( self, p ) :
    if p.gtoken () != 'D' :
        raise WrongAModel("unable to read the feature vector dimension D")
    token= p.gtoken ()
    if token == '' :
        raise WrongAModel("unable to read the value of D")
    self.D= int(token)
    if self.D <= 0 :
        raise WrongAModel("D must be positive")

class Hybrid:
    """A hybrid state"""
    
    def __init__(self,cmodel=None,p=None,q=None):
        if cmodel==None: return
        token= p.gtoken()
        if token=='':
            raise WrongAModel("unable to read the index of the state",q)
        self.ind= int(token)
        if self.ind<0 or self.ind>=cmodel.D:
            raise WrongAModel("the index is out of range",q)
        
    def copy(self):
        """Copy a hybrid state"""
        h= Hybrid()
        h.ind= self.ind
        return h
        
    def write(self,to,cmodel):
        """Writes the hybrid state"""
        to.write('%d\n'%self.ind)

_cmodel_types['Hybrid']= AModelType ( Hybrid_copy_global,
                                      Hybrid_write_global,
                                      Hybrid_read_global,
                                      Hybrid )


## LogLinear ##

def LogLinear_copy_global ( fro, to ) :
    to.D= fro.D


def LogLinear_write_global ( self, to ) :
    to.write ( 'D %d\n' % self.D )

    
def LogLinear_read_global ( self, p ) :
    if p.gtoken () != 'D' :
        raise WrongAModel("unable to read the feature vector dimension D")
    token= p.gtoken ()
    if token == '' :
        raise WrongAModel("unable to read the value of D")
    self.D= int(token)
    if self.D <= 0 :
        raise WrongAModel("D must be positive")


class LogLinear:
    """Log-linear model"""

    def __init__ ( self, cmodel= None, p= None, q= None ) :
        if cmodel == None :
            return
        if p.gtoken () != 'L' :
            raise WrongAModel("unable to read the weights L",q)
        self.l= array ( 'd' )
        for d in xrange(0,cmodel.D):
            token= p.gtoken ()
            if token == '' :
                raise WrongAModel("unable to read L[%d]" % d+1,q)
            self.l.append( float(token) )
    

    def copy ( self ) :
        """Copy a log-linear model"""
        ll= LogLinear ()
        ll.l= array ( 'd', list(self.l) )
        return ll
    
    
    def write ( self, to, cmodel ) :
        """Escriu el model log-linear en el fitxer TO"""
        to.write ( 'L' )
        for l in self.l :
            to.write ( ' %g' % l )
        to.write ( '\n' )


_cmodel_types['LogLinear']= AModelType ( LogLinear_copy_global,
                                         LogLinear_write_global,
                                         LogLinear_read_global,
                                         LogLinear )


## Mixture ##

def Mixture_copy_global ( fro, to ) :
    to.members_args= Empty ()
    to.members_args.tname= fro.members_args.tname
    typ= to.members_args.typ= fro.members_args.typ
    typ.copy_global ( fro.members_args, to.members_args )


def Mixture_write_global ( self, to ):
    to.write ( '%s\n' % self.members_args.tname )
    self.members_args.typ.write_global ( self.members_args, to )


def Mixture_read_global ( self, p ) :
    token= p.gtoken ()
    if token == '' :
        raise WrongAModel("unable to read the mixture component type")
    typ= _cmodel_types.get ( token )
    if typ == None :
        raise WrongAModel("tipus de model (%s) desconegut"
                          % token)
    self.members_args= Empty ()
    self.members_args.tname= token
    self.members_args.typ= typ
    typ.read_global ( self.members_args, p )


class Mixture :
    """Mixture"""

    def __init__ ( self, cmodel= None, p= None, q= None ) :

        if cmodel == None :
            self.I= 0
            self.p= array ( 'd' )
            self.members= []
            return
        
        if p.gtoken () != 'I' :
            raise WrongAModel(("unable to read the number of"+
                               " mixture components I"),q)
        token= p.gtoken ()
        if token == '' :
            raise WrongAModel("unable to read the value of I",q)
        self.I= int(token)
        if self.I <= 0 :
            raise WrongAModel("I must be positive",q)

        if p.gtoken () != 'PMembers' :
            raise WrongAModel(("unable to read the mixture"+
                               " coefficients PMembers"),(c,q))
        self.p= array ( 'd' )
        for i in xrange(0,self.I):
            token= p.gtoken ()
            if token == '' :
                raise WrongAModel("unable to read PMembers[%d]" % i,q)
            self.p.append ( float(token) )

        if p.gtoken () != 'Members' :
            raise WrongAModel("unable to read the mixture components Members",q)
        self.members= []
        for i in xrange(0,self.I):
            self.members.append (
                cmodel.members_args.typ.STATE ( cmodel.members_args,
                                                p, (q,i) ) )
        
    # end
    

    def copy ( self ) :
        """Copy a mixture"""
        mix= Mixture ()
        mix.I= self.I
        mix.p= array ( 'd', list(self.p) )
        mix.members= []
        for m in self.members:
            mix.members.append ( m.copy () )
        return mix

    
    def write ( self, to, cmodel ) :
        """Write a mixture"""
        to.write ( 'I %d\n' % self.I )
        to.write ( 'PMembers' )
        for prob in self.p:
            to.write ( ' %g' % prob )
        to.write ( '\n' )
        to.write ( 'Members\n' )
        for m in self.members:
            m.write ( to, cmodel.members_args )


_cmodel_types['Mixture']= AModelType ( Mixture_copy_global,
                                       Mixture_write_global,
                                       Mixture_read_global,
                                       Mixture )


## DGaussian ##

def DGaussian_copy_global ( fro, to ):
    to.D= fro.D
    to.smooth= array ( 'd', list(fro.smooth) )


def DGaussian_write_global ( self, to ):
    to.write ( 'D %d\n'%self.D )
    to.write ( 'SMOOTH' )
    for i in self.smooth:
        to.write ( ' %g'%i )
    to.write ( '\n' )


def DGaussian_read_global ( self, p ):
    if p.gtoken() != 'D' :
        raise WrongAModel("unable to read the feature vector dimension D")
    token= p.gtoken()
    if token == '' :
        raise WrongAModel("unable to read the value of D")
    self.D= int(token)
    if self.D <= 0 :
        raise WrongAModel("D must be positive")
    if p.gtoken() != 'SMOOTH' :
        raise WrongAModel("unable to read the smooth vector SMOOTH")
    self.smooth= array ( 'd' )
    for d in xrange(0,self.D):
        token= p.gtoken()
        if token == '' :
            raise WrongAModel("unable to read SMOOTH[%d]"%(d+1))
        self.smooth.append ( float(token) )


class DGaussian:
    """Diagonal Gaussian"""
    
    def __init__ ( self, cmodel= None, p= None, q= None ):
        if cmodel == None: return
        if p.gtoken() != 'MU' :
            raise WrongAModel("unable to read means MU",q)
        self.mu= array ( 'd' )
        for d in xrange(0,cmodel.D):
            token= p.gtoken()
            if token == '' :
                raise WrongAModel("unable to read MU[%d]"%(d+1),q)
            self.mu.append ( float(token) )
        if p.gtoken() != 'VAR' :
            raise WrongAModel("unable to read variances VAR",q)
        self.var= array ( 'd' )
        for d in xrange(0,cmodel.D):
            token= p.gtoken()
            if token == '' :
                raise WrongAModel("unable to read VAR[%d]"%(d+1),q)
            self.var.append ( float(token) )
        
            
    def copy ( self ):
        """Copy diagonal Gaussian"""
        dg= DGaussian()
        dg.mu= array ( 'd', list(self.mu) )
        dg.var= array ( 'd', list(self.var) )
        return dg
    
    
    def write ( self, to, cmodel ):
        """Write diagonal Gaussian"""
        to.write ( 'MU' )
        for d in xrange(0,cmodel.D):
            to.write ( ' %g'%self.mu[d] )
        to.write ( '\n' )
        to.write ( 'VAR' )
        for d in xrange(0,cmodel.D):
            to.write ( ' %g'%self.var[d] )
        to.write ( '\n' )
 

_cmodel_types['DGaussian']= AModelType ( DGaussian_copy_global,
                                         DGaussian_write_global,
                                         DGaussian_read_global,
                                         DGaussian )


## TiedStates ##

def TiedStates_copy_global ( fro, to ) :
    to.members_args= Empty ()
    to.members_args.tname= fro.members_args.tname
    typ= to.members_args.typ= fro.members_args.typ
    typ.copy_global ( fro.members_args, to.members_args )
    to.members_args.N= fro.members_args.N
    to.members_args.names= list(fro.members_args.names)
    to.members_args.states= []
    to.members_args.states_dict= {}
    for n in xrange(0,to.members_args.N):
        aux= fro.members_args.states[n].copy ()
        to.members_args.states.append ( aux )
        to.members_args.states_dict[fro.members_args.names[n]]= aux


def TiedStates_write_global ( self, to ):
    to.write ( '%s\n' % self.members_args.tname )
    self.members_args.typ.write_global ( self.members_args, to )
    to.write ( 'N %d\n'%self.members_args.N )
    to.write ( 'States\n' )
    for n in xrange(0,self.members_args.N):
        to.write('%s\n'%self.members_args.names[n])
        self.members_args.states[n].write ( to, self.members_args )


def TiedStates_read_global ( self, p ) :
    # Subtipus i global del subtipus
    token= p.gtoken ()
    if token == '' :
        raise WrongAModel("unable to read the type of states")
    typ= _cmodel_types.get ( token )
    if typ == None :
        raise WrongAModel("unknown state type: %s"%token)
    self.members_args= Empty ()
    self.members_args.tname= token
    self.members_args.typ= typ
    typ.read_global ( self.members_args, p )
    # Llig els estats
    if p.gtoken () != 'N' :
        raise WrongAModel("unable to read the number of states N")
    token= p.gtoken ()
    if token == '' :
        raise WrongAModel("unable to read the value of N")
    self.members_args.N= int(token)
    if self.members_args.N <= 0 :
        raise WrongAModel("N must be positive")
    if p.gtoken () != 'States' :
        raise WrongAModel("unable to read the states States")
    self.members_args.states= []
    self.members_args.states_dict= {}
    self.members_args.names= []
    for n in xrange(0,self.members_args.N):
        token= p.gtoken ()
        if token == '' :
            raise WrongAModel("Unable to read the state's name",n)
        name= token
        if self.members_args.states_dict.get(name)!=None:
            raise WrongAModel("State %s already exists"%name)
        aux= typ.STATE ( self.members_args, p, (n) )
        self.members_args.states.append ( aux )
        self.members_args.names.append ( name )
        self.members_args.states_dict[name]= aux


class TiedStates:
    """State's name"""

    def __init__ ( self, cmodel= None, p= None, q= None ) :

        if cmodel == None :
            self.name= ''
            return
        
        token= p.gtoken ()
        if token == '' :
            raise WrongAModel("unable to read the state's name",q)
        self.name= token
    

    def copy ( self ) :
        """Copy tied state"""
        ts= TiedStates ()
        ts.name= self.name
        return ts

    
    def write ( self, to, cmodel ) :
        """Write tied state"""
        to.write('%s\n'%self.name)


_cmodel_types['TiedStates']= AModelType ( TiedStates_copy_global,
                                          TiedStates_write_global,
                                          TiedStates_read_global,
                                          TiedStates )




##########
# CMODEL #
##########

class Trans:
    """Transitions. Probabilities are stored using logarithm"""
    
    LINEAL= 0
    LIST= 1
    POINTER= 2

    
    class ListEdge:

        def __init__(self,arg1,arg2=None):
            if isinstance(arg1,Trans.ListEdge) :
                self.q= arg1.q
                self.p= arg1.p
            else:
                self.q= arg1
                self.p= arg2

        def __str__ ( self ):
            return '%d %s'%(self.q+1,lptoa(self.p))
        
        
    class ListState:
        
        def __init__(self,arg1=None):
            if isinstance(arg1,Trans.ListState):
                self.edges= [ Trans.ListEdge(e) for e in arg1.edges ]
                self.fp= arg1.fp
            else:
                self.edges= []
                self.fp= 0.0

        def write ( self, to ):
            for e in self.edges:
                to.write ( '  %s\n'%str(e) )
            if self.fp != 0.0 :
                to.write ( '  F %s\n'%lptoa(self.fp) )
            to.write ( '  .\n' )

    
    @staticmethod
    def __read_list ( p, origin, c, Q ):
        ret= Trans.ListState()
        if p.gtoken() != origin:
            raise WrongAModel(("unable to read the edges"+
                               " of state %s")%origin,c)
        token= p.gtoken()
        while token != '.':
            if token == '':
                raise WrongAModel(("Waiting for the '.' in the edge list"+
                                   " of state %s")%origin,c)
            elif token == 'F' :
                token= p.gtoken()
                if token == '' :
                    raise WrongAModel(("There is an edge without probability"+
                                       " in state %s"%origin),c)
                ret.fp= atolp(token)
            else:
                q= int(token)
                if q < 1 or q > Q :
                    raise WrongAModel(("out of range edge"+
                                       " in state %s")%origin,c)
                token= p.gtoken()
                if token == '' :
                    raise WrongAModel(("There is an edge without probability"+
                                       " in state %s"%origin),c)
                prob= atolp(token)
                ret.edges.append ( Trans.ListEdge ( q-1, prob ) )
            token= p.gtoken()
        return ret
    
    
    def __init_fd ( self, p, c, Q, model ):
        token= p.gtoken()

        # Lineal
        if token == 'Trans' :
            self.type= Trans.LINEAL
            self.isp01= False
            self.p= array ( 'd' )
            for q in xrange(0,Q):
                token= p.gtoken ()
                if token == '' :
                    raise WrongAModel("unable to read Trans[%d]"%q,c)
                self.p.append ( atolp(token) )
        elif token == 'TransP01' :
            self.type= Trans.LINEAL
            self.isp01= True
            self.p0= array ( 'd' )
            self.p1= array ( 'd' )
            for q in xrange(0,Q):
                token= p.gtoken ()
                if token == '' :
                    raise WrongAModel("unable to read TransP01[%d][0]"%q,c)
                self.p0.append ( atolp(token) )
                token= p.gtoken ()
                if token == '' :
                    raise WrongAModel("unable to read TransP01[%d][1]"%q,c)
                self.p1.append ( atolp(token) )
        
        # Llista
        elif token == 'TransL' :
            self.type= Trans.LIST
            self.init= Trans.__read_list ( p, 'I', c, Q )
            self.states= []
            for q in xrange(0,Q):
                self.states.append ( Trans.__read_list ( p, str(q+1), c, Q ) )
                
        # Punter
        elif token == 'TransP':
            self.type= Trans.POINTER
            token= p.gtoken()
            self.ptr= model.ilabels.get(token)
            if self.ptr == None:
                raise WrongAModel("Pointer to unexiting HMM: %s"%token)
            if model.hmms[self.ptr].trans.type == Trans.POINTER:
                self.ptr= model.hmms[self.ptr].trans.ptr
            self.lbl= model.labels[self.ptr]
        
        else: raise WrongAModel("unable to read transitions",c)


    def __init_trans ( self, trans ):
        self.type= trans.type
        if self.type == Trans.LINEAL:
            self.isp01= trans.isp01
            if self.isp01:
                self.p0= array ( 'd', list(trans.p0) )
                self.p1= array ( 'd', list(trans.p1) )
            else:
                self.p= array ( 'd', list(trans.p) )
        elif self.type == Trans.LIST:
            self.init= Trans.ListState(trans.init)
            self.states= [ Trans.ListState(s) for s in trans.states ]
        elif self.type == Trans.POINTER:
            self.ptr= trans.ptr
            self.lbl= trans.lbl
    
    
    def __init__ ( self, arg1, arg2=None, arg3=None, model=None ):
        if isinstance(arg1,ParseFile) :
            self.__init_fd ( arg1, arg2, arg3, model )
        else: self.__init_trans ( arg1 )
        

    def write ( self, to ):
        """Write transitions"""
        if self.type == Trans.LINEAL :
            if self.isp01 :
                to.write ( 'TransP01\n' )
                for i in xrange(0,len(self.p0)):
                    to.write ( '%s %s\n'%(lptoa(self.p0[i]),lptoa(self.p1[i])) )
            else:
                to.write ( 'Trans\n' )
                to.write ( '%s'%lptoa(self.p[0]))
                for p in self.p[1:] :
                    to.write ( ' %s' % lptoa(p) )
                to.write ( '\n' )
        elif self.type == Trans.LIST:
            to.write ( 'TransL\n' )
            to.write ( 'I\n' )
            self.init.write ( to )
            q= 1
            for s in self.states:
                to.write ( '%d\n'%q )
                s.write ( to )
                q+= 1
        elif self.type == Trans.POINTER:
            to.write ( 'TransP %s\n'%self.lbl )
    
    def tolist ( self, Q ):
        """Converts transitions from LINEAL to LIST"""
        if self.type == Trans.LINEAL :
            self.type= Trans.LIST
            self.init= Trans.ListState()
            self.init.edges.append ( Trans.ListEdge ( 0, 0.0 ) )
            self.states= []
            for q in xrange(0,Q-1):
                aux= Trans.ListState()
                if self.isp01:
                    aux.edges.append ( Trans.ListEdge ( q, self.p0[q] ) )
                    aux.edges.append ( Trans.ListEdge ( q+1, self.p1[q] ) )
                else:
                    aux.edges.append (
                        Trans.ListEdge ( q, log(1.0-exp(self.p[q])) ) )
                    aux.edges.append ( Trans.ListEdge ( q+1, self.p[q] ) )
                self.states.append ( aux )
            aux= Trans.ListState()
            if self.isp01:
                aux.edges.append ( Trans.ListEdge ( Q-1, self.p0[Q-1] ) )
                aux.fp= self.p1[Q-1]
            else:
                aux.edges.append (
                    Trans.ListEdge ( Q-1, log(1.0-exp(self.p[Q-1])) ) )
                aux.fp= self.p[Q-1]
            self.states.append ( aux )
            
    
class HMM:
    """HMM"""

    def __init_fd ( self, p, cmodel, STATE, c ) :

        # Capçalera
        if p.gtoken () != 'Q' :
            raise WrongAModel("unable to read the number of states Q",c)
        token= p.gtoken ()
        if token == '' :
            raise WrongAModel("unable to read the value of Q",c)
        self.Q= int(token)
        if self.Q <= 0 :
            raise WrongAModel("Q must be positive",c)

        # Transicions
        self.trans= Trans ( p, c, self.Q, cmodel )
        
        # Estats.
        self.states= []
        for q in xrange(0,self.Q):
            self.states.append ( STATE ( cmodel, p, (c,q+1) ) )
        
    # end


    def __init_char ( self, char ) :
        self.Q= char.Q
        self.trans= Trans(char.trans)
        self.states= []
        for s in char.states:
            self.states.append ( s.copy () )
            
    
    def __init__ ( self, arg1, arg2, arg3= None, arg4= None ) :
        if isinstance(arg1,ParseFile) :
            self.__init_fd ( arg1, arg2, arg3, arg4 )
        else:
            self.__init_char ( arg1 )

    
    def write ( self, to, cmodel ) :
        """Write HMM"""
        to.write ( 'Q %d\n' % self.Q )
        self.trans.write ( to )
        for s in self.states :
            s.write ( to, cmodel )
        
        
    def tomix ( self ) :
        """Transform HMM states into equivalent mixture states"""
        for q in xrange(0,self.Q):
            s= Mixture ()
            s.I= 1
            s.p.append ( 1.0 )
            s.members.append ( self.states[q] )
            self.states[q]= s


class AModel:
    """Acoustic model"""

    def __init_fd ( self, fd ) :
        
        p= ParseFile ( fd )

        # Capçalera
        token= p.gtoken()
        if token != 'CMODEL' and token != 'AMODEL':
            WrongAModel("unable to read 'token' AMODEL or CMODEL")
        token= p.gtoken ()
        if token == '' :
            raise WrongAModel("unable to read the type of acoustic model")
        self._typ= _cmodel_types.get ( token )
        if self._typ == None :
            raise WrongAModel("unknown model type: %s"%token)
        self.tname= token
        self._typ.read_global ( self, p )
        if p.gtoken () != 'N' :
            raise WrongAModel("unable to read the number of HMMs N")
        token= p.gtoken ()
        if token == '' :
            raise WrongAModel("unable to read the value of N")
        self.N= int(token)
        if self.N <= 0 :
            raise WrongAModel("N must be positive")
        
        # Caràcters
        self.hmms= []
        self.labels= []
        self.ilabels= {}
        for n in xrange(0,self.N):
            token= p.gtoken ()
            if token == '' :
                raise WrongAModel("unable to read the name of the HMM")
            if len(token) < 3 or token[0] != "'" or token[-1] != "'" :
                raise WrongAModel("HMM's name format must be '<STRING>'")
            lbl= token[1:-1]
            self.labels.append ( lbl )
            self.ilabels[lbl]= n
            self.hmms.append ( HMM ( p, self, self._typ.STATE, n+1 ) )
        
    # end


    def __init_cmodel ( self, cmodel ) :
        self._typ= cmodel._typ
        self.tname= cmodel.tname
        self._typ.copy_global ( cmodel, self )
        self.N= cmodel.N
        self.labels= list(cmodel.labels)
        self.ilabels= dict(cmodel.ilabels)
        self.hmms= []
        for c in cmodel.hmms:
            self.hmms.append ( HMM(c,None,self._typ.STATE) )
    
    
    def __init__ ( self, value ) :
        """
        AModel(file) -> Reads acoustic model from file
        AModel(AModel) -> Copies an acoustic model
        """
        if type(value) == file :
            self.__init_fd ( value )
        else:
            self.__init_cmodel ( value )

    
    def write ( self, to ) :
        """Write acoustic model"""
        to.write ( 'AMODEL\n' )
        to.write ( '%s\n' % self.tname )
        self._typ.write_global ( self, to )
        to.write ( 'N %d\n' % self.N )
        for n in xrange(0,self.N):
            to.write ( "'%s'\n" % self.labels[n] )
            self.hmms[n].write ( to, self )


    def tomix ( self ) :
        """Acoustic model is transformed into an equivalent model of
        mixture type
        """
        members_args= Empty ()
        members_args.tname= self.tname
        members_args.typ= self._typ
        self._typ.copy_global ( self, members_args )
        self.members_args= members_args
        self.tname= 'Mixture'
        self._typ= _cmodel_types['Mixture']
        for c in self.hmms:
            c.tomix ()
