#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  wgraph.py: Operations over word graphs
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from sys import argv,stdin,stdout,stderr
from optparse import OptionParser
from math import log,exp

#####################################################################
#  Levensthein edit distance from wer++
#####################################################################
def string_equal(str1,str2):
  return (str1 == str2)

def lev_changes(str1, str2, i_cost, d_cost, d_sub,vocab={}, eq_func=string_equal):
  d={}; sub={};
  for i in range(len(str1)+1):
    d[i]=dict()
    d[i][0]=i*d_cost
    sub[i]={}
    sub[i][0]='D'
  for i in range(len(str2)+1):
    d[0][i] = i*i_cost
    sub[0][i]='I'
  for i in range(1, len(str1)+1):
    for j in range(1, len(str2)+1):
      if d[i][j-1]+i_cost < d[i-1][j]+d_cost and d[i][j-1] < d[i-1][j-1]+(not eq_func(str1[i-1],str2[j-1]))*d_sub:
        if vocab=={} or (str2[j-1] in vocab):
          sub[i][j] = 'I';
        else:
          sub[i][j] = 'O'; #Oov insertion
      elif d[i-1][j]+d_cost < d[i][j-1]+i_cost and d[i-1][j] < d[i-1][j-1]+(not eq_func(str1[i-1],str2[j-1]))*d_sub:
        sub[i][j] = 'D';
      else:
        if eq_func(str1[i-1],str2[j-1]):
          sub[i][j] = 'E';
        else:
          if vocab=={} or (str2[j-1] in vocab):
            sub[i][j] = 'S';
          else:
            sub[i][j] = 'A'; #Oov Substitution
      d[i][j] = min(d[i][j-1]+i_cost, d[i-1][j]+d_cost, d[i-1][j-1]+(not eq_func(str1[i-1],str2[j-1]))*d_sub)

  i=len(str1); j=len(str2); path=[]
  while(i > 0 or j > 0):
    path.append([sub[i][j],i-1,j-1])
    if(sub[i][j] == 'I' or sub[i][j] == 'O'):
      j-=1
    elif(sub[i][j] == 'D'):
      i-=1
    else:
      j-=1; i-=1;
  path.reverse()
  return path


#####################################################################
#  Robust sum of logarithms
#####################################################################
def sum_logs(a,b):
  if a == float("Inf"):
    return b
  elif b == float("Inf"):
    return a
  elif a > b:
    return log(exp(b-a) + 1.0)+a
  else:
    return log(exp(a-b) + 1.0)+b

#######################################################
##  Structures to read the graph   ####################
#######################################################

# State
class State:
    def __init__ ( self, t ):
        self.t= t
        self.iedges= []
        self.oedges= []
    def __repr__ ( self ):
        return '<t:%d;input:%s;output:%s>'%(self.t,self.iedges,self.oedges)

# Edge
class Edge:
    def __init__ ( self, src, dst, word, ac, lm, phos=None ):
        self.src= src
        self.dst= dst
        self.word= word
        self.lm= lm
        self.ac= ac
        self.phos= phos
    def __repr__ ( self ):
        return '(s:%d;e:%d;a:%g;l:%g;w:%s)'%(self.src,self.dst,self.ac,
                                             self.lm,self.word)

# Hypothesis
class Hyp:
    def __init__ ( self, prev= -1, prob= 0.0, word= "", edge= None ):
        self.prev= prev
        self.prob= prob
        self.word= word
        self.edge= edge
    def __repr__ ( self ):
      return '(w:%s;f:%f)\n' %(self.word,self.prob)

# Graph structure and HTK parse
class WGraph:
  def __init__( self ):
    self.gsf = 1.0
    self.wip = 0.0
    self.NS = 0
    self.NE = 0
    self.states = [ None ]
  def __init__( self, f_in ):
    l= f_in.readline().split()
    if len(l) != 1 and l[0] != 'VERSION=1.0' :
        exit ( "Incorrect format" )

    # Sample
    l= f_in.readline().split()
    if len(l) != 1 :
        exit ( "Incorrect format" )

    # GSF and WIP
    l= f_in.readline().split()
    if len(l) != 2 :
        exit ( "Incorrect format" )
    aux= l[0].split('=')
    if len(aux) != 2 and aux[0] != 'lmscale' :
        exit ( "Incorrect format" )
    self.gsf= float(aux[1])
    aux= l[1].split('=')
    if len(aux) != 2 and aux[0] != 'wdpenalty' :
        exit ( "Incorrect format" )
    self.wip= float(aux[1])

    # Scale
    l= f_in.readline().split()
    if len(l) != 1 :
        exit ( "Incorrect format" )
    aux= l[0].split('=')
    if len(aux) != 2 and aux[0] != 'acscale' :
        exit ( "Incorrect format" )
    scale= float(aux[1])

    # GSF and WIP
    l= f_in.readline().split()
    if len(l) != 2 :
        exit ( "Incorrect format" )
    aux= l[0].split('=')
    if len(aux) != 2 and aux[0] != 'N' :
        exit ( "Incorrect format" )
    self.NS= int(aux[1])
    aux= l[1].split('=')
    if len(aux) != 2 and aux[0] != 'L' :
        exit ( "Incorrect format" )
    self.NE= int(aux[1])

    def parse_token(token,name,f):
        aux= token.split('=')
        if len(aux) != 2 and aux[0] != name :
            exit ( "Incorrect format" )
        return f(aux[1])

    # States
    self.states= [None]*self.NS
    self.end_time = 0
    for i in range(0,self.NS):
        l= f_in.readline().split()
        if len(l) != 2 :
            exit ( "Incorrect format" )
        state= parse_token ( l[0], 'I', int )
        self.states[state]= State ( parse_token ( l[1], 't', int ) )
        if self.states[state].t > self.end_time:
          self.end_time = self.states[state].t

    # Edges
    for i in xrange(0,self.NE):
        l= f_in.readline().split()
        if len(l) < 6 :
            exit ( "Incorrect format" )

        name= parse_token ( l[0], 'J', int )
        src= parse_token ( l[1], 'S', int )
        dst= parse_token ( l[2], 'E', int )
        if len(l[3]) < 3 or l[3][:2] != 'W=' :
            exit ( "Incorrect format" )
        word= l[3][2:]
        ac= parse_token ( l[4], 'a', float )
        lm= parse_token ( l[5], 'l', float )

        phonemes= None
        if len(l) > 6:
          phonemes= l[6][1:-1].split(':')

        edge= Edge(src,dst,word,ac*log(10),lm*log(10), phonemes)
        #output edges
        self.states[src].oedges.append ( edge )
        #input edges
        self.states[dst].iedges.append ( edge )

#####################################################################
#  Calculates the probability of an edge
#####################################################################
  def edge_prob(self, e):
    return self.asf*(e.lm*self.gsf + self.wip + e.ac)

##########################################################################
#  Return a list of edges order in topologically order from the beginning
##########################################################################
  def top_order(self):
    stack= []
    counters= []
    for s in xrange(0,self.NS):
        aux= len(self.states[s].iedges)
        counters.append ( aux )
        if aux == 0 : stack.append ( s )
    #print counters

    top= []
    while stack != []:
        s = stack.pop()
        top.append ( s )
        for e in self.states[s].oedges:
            counters[e.dst]-= 1
            if counters[e.dst] == 0 :
                stack.append ( e.dst )
    #print top,len(top)
    return top

#####################################################################
#  Return a list of edges order in inverse topologically order
#####################################################################
  def inv_top_order(self):
    stack= []
    counters= []
    for s in xrange(0,self.NS):
        aux= len(self.states[s].oedges)
        counters.append ( aux )
        if aux == 0 : stack.append ( s )
    #print counters

    inv_top= []
    while stack != []:
        s = stack.pop()
        inv_top.append ( s )
        for e in self.states[s].iedges:
            counters[e.src]-= 1
            if counters[e.src] == 0 :
              stack.append ( e.src )
    #print top,len(top)
    return inv_top

#####################################################################
# Obtain the viterbi probabity for each state
#####################################################################
  def viterbi_probs(self):
    top = self.top_order()
    hyps= [None]*self.NS
    for i in xrange(0,self.NS):
        s= top[i]
        edges= self.states[s].iedges
        if edges == [] :
            hyps[s]= Hyp()
            continue
        maxp = float("-Inf")
        for ed in edges:
            aux= hyps[ed.src].prob + self.edge_prob(ed)
            if aux > maxp :
                maxp,prev,word= aux,ed.src,ed.word
        hyps[s]= Hyp(prev,maxp,word,ed)

    return hyps

#####################################################################
# Gets the most probable path from viterbi probabilities
#####################################################################
  def best_path(self):
  # Obtain the best path arriving to the end
    hyps = self.viterbi_probs()
    m_prob = float("-Inf"); m_ret = ""
    ends = [x for x in hyps if x.word == "</s>"]
    for x in ends:
      s = hyps.index(x)
      prob= hyps[s].prob
      ret= ''
      while s != 0:
          ret= hyps[s].word + ' ' + ret
          s= hyps[s].prev
      if prob > m_prob:
        m_prob = prob
        m_ret = ret

    return (m_prob,m_ret)

#####################################################################
# Gets the most probable path along with its delimitation 
#####################################################################
  def best_path_limits(self):
    hyps = self.viterbi_probs()
    m_prob = float("-Inf"); m_ret = ""
    ends = [x for x in hyps if x.word == "</s>"]
    if len(ends) > 1:
      stderr.write("[WW] There are multiple end nodes!\n")
    x = ends[0]
    s = hyps.index(x)
    v = []
    # start_state, end_state, start_time, start_end, word
    while s != 0:
      v.append((hyps[s].prev, s, self.states[hyps[s].prev].t,
            self.states[s].t, hyps[s].word, hyps[s].edge))
      s= hyps[s].prev

    v.reverse()
    return v

#####################################################################
# look for the graph root
#####################################################################
  def root(self):
    for s in xrange(0,self.NS):
        aux= len(self.states[s].iedges)
        if aux == 0 :
          return s

#####################################################################
# look for end node
#####################################################################
  def end_node(self):
    ends = [ x for x in xrange(len(self.states)) if len(self.states[x].oedges) == 0 ]
    return ends[0]

#####################################################################
####  Performs the forward algorithm in the graphs and return the probability
###   at each node
#####################################################################
  def forward_probs(self):
    top = self.top_order()
    state_prob = []
    for i in xrange(0,len(top)):
      state_prob.append(float("Inf"))
    state_prob[top[0]] = 0.0

    for i in xrange(0,len(top)):
      s = top[i]
      for ed in self.states[s].oedges:
        nxt, word = ed.dst,ed.word
        p = state_prob[s] + self.edge_prob(ed)
        state_prob[nxt] = sum_logs(state_prob[nxt], p)

    return state_prob

#####################################################################
####  Performs the backward algorithm in the graphs and return the probability
###   at each node
#####################################################################
  def backward_probs(self):
    top = self.inv_top_order()
    state_prob = []
    for i in xrange(0,len(top)):
      state_prob.append(float("Inf"))
    state_prob[top[0]] = 0.0

    for i in xrange(0,len(top)):
      s = top[i]
      for ed in self.states[s].iedges:
        prv, word = ed.src,ed.word
        p = state_prob[s] + self.edge_prob(ed)
        state_prob[prv] = sum_logs(state_prob[prv],p)

    return state_prob

######################################################################################
#### Get the which states are active at each time, i.e. have edges during that time
######################################################################################
  def states_active_at_each_time(self):
    v = {}
    for i in xrange(0, self.states[self.end_node()].t):
      v[i]=[]

    for i in xrange(0, self.NS):
      for e in self.states[i].oedges:
        src, dst = e.src, e.dst
        start = self.states[src].t
        end   = self.states[dst].t

        for j in xrange(start,end):
          if src not in v[j]:
            v[j].append(src)

    return v

###############################################################################
#### Compute the posterior probability of a word (based on Wessel et al 2001)
###############################################################################
  def posterior_word_CM(self, start_node, end_node, fwd_probs, bck_probs, denominator):
    edges = [ x for x in self.states[start_node].oedges if x.dst == end_node ]

    p = float("-Inf")
    for e in edges:
      prob = fwd_probs[e.src]+self.edge_prob(e)+bck_probs[e.dst]-denominator
      if p < prob:
        p = prob

    return p

#####################################################################################
#### Compute the maximum posterior probability of a word (based on Wessel et al 2001)
#####################################################################################
  def posteriorMax_word_CM(self, start_node, end_node, start_time, end_time, word, \
                                fwd_probs, bck_probs, denominator, indexes):
    max_sum = float("-Inf")
    last_indexes = []
    for i in xrange(start_time, end_time):
      prob = float("Inf")
      if indexes[i] == last_indexes:
        continue
      last_indexes = indexes[i]
      for st in indexes[i]:
        #select only intersecting edges
        edges = [x for x in self.states[st].oedges if x.word == word and \
                self.states[x.src].t < end_time and self.states[x.dst].t > start_time and \
                self.states[x.src].t <= i and self.states[x.dst].t > i]
        if edges == []:
          continue
        for e in edges:
          p = self.edge_prob(e)
          p = fwd_probs[e.src]+p+bck_probs[e.dst]-denominator
          prob = sum_logs(prob,p)

        if prob > max_sum:
          max_sum = prob

    return max_sum


#######################################################################
#### Return the best acoustic probability between two states
#######################################################################
  def best_acoustic_score(self, start_node, end_node):
    edges = [ x for x in self.states[start_node].oedges if x.dst == end_node ]
    max_ac = float("-Inf")
    for e in edges:
      if e.ac > max_ac:
        max_ac = e.ac

    return max_ac

#######################################################################
#### Return the best language model probability between two states
#######################################################################
  def best_language_score(self, start_node, end_node):
    edges = [ x for x in self.states[start_node].oedges if x.dst == end_node ]
    max_lm = float("-Inf")
    for e in edges:
      if e.lm > max_lm:
        max_lm = e.lm

    return max_lm

#######################################################################
#### Compute confidence measures
#######################################################################
  def posterior_word_based_CM(self,asf,gsf,wip,reference=None):
    self.asf = 1.0
    best_limits = self.best_path_limits()
    res = []

    self.asf = 1.0/asf
    if gsf != float("-Inf"):
      self.gsf = gsf
    if wip != float("-Inf"):
      self.wip = wip

    fwd = self.forward_probs()
    bck = self.backward_probs()

    indexes = self.states_active_at_each_time()
    denominator = bck[self.root()]

    best_words = [x[4] for x in best_limits[1:-1]]
    changes = None
    if reference != None:
      changes = lev_changes(best_words,reference,1,1,1)

    n = 0
    for start_node, end_node, start, end, word, edge in best_limits[1:-1]:
      posterior_max = self.posteriorMax_word_CM(start_node, end_node, start, end, \
                word, fwd, bck, denominator, indexes)

      if reference != None:
        editions = [ x[0] for x in changes if x[1] == n ]
        ref = 1 if 'E' in editions else 0
        res.append(PosteriorResult(edge, posterior_max, start, end, ref))
      else:
        res.append(PosteriorResult(edge, posterior_max, start, end))
      n+=1

    return res

#######################################################################
#### Generate a dot file with the word graph
#######################################################################
  def generate_dot(self, f, prune, asf, gsf, wip):
      self.asf = 1.0/asf
      if gsf != float("-Inf"):
        self.gsf = gsf
      if wip != float("-Inf"):
        self.wip = wip

      if prune != 0.0:
        prune = log(prune)

      fwd = self.forward_probs()
      bck = self.backward_probs()

      indexes = self.states_active_at_each_time()
      denominator = bck[self.root()]

      oedges_not_prunned={}
      for i in xrange(self.NS):
        for j in self.states[i].oedges:
          p = self.edge_prob(j)
          p = fwd[j.src]+p+bck[j.dst]-denominator
          if prune == 0.0 or p >= prune:
            if i not in oedges_not_prunned:
              oedges_not_prunned[i] = []
            oedges_not_prunned[i].append(j)

      f.write("digraph \"word-graph\" {\n\n")
      f.write("rankdir=\"LR\"\n")
      rank={}

      #Nodes
      for j in oedges_not_prunned.keys():
        t = self.states[j].t
        if t not in rank:
          rank[t] = []
        rank[t].append(j)
      for i in rank:
        f.write("{rank=same\n")
        for j in rank[i]:
          t = self.states[j].t
          f.write("Node%s [label=\"%s\"]\n" %(j, t))
        f.write("}\n")

      #edges
      for i in oedges_not_prunned:
        for j in oedges_not_prunned[i]:
          p = self.edge_prob(j)
          p = fwd[j.src]+p+bck[j.dst]-denominator
          f.write("Node%s->Node%s [label=\"%s\\n%.1f\\n%.1f\"]\n" %(j.src, j.dst, j.word, j.ac, j.lm))

      f.write("}\n")

################################################################33
### Class for results
################################################################33
class PosteriorResult:
  def __init__( self, edge, posterior, start, end, correct=None ):
    self.edge = edge
    self.posterior = posterior
    self.correct = correct
    self.start = start
    self.end = end
  def __repr__( self ):
    if self.correct != None:
      return "%s %d %g %d %d" %(self.edge.word, self.correct, exp(self.posterior), self.start, self.end)
    else:
      return "%s %g %d %d" %(self.edge.word, exp(self.posterior), self.start, self.end)



#######################################################################
#######################################################################
#######################################################################
def main():
  cmd_parser = OptionParser(usage="usage: %prog [options]")
  cmd_parser.add_option('-v',
      action="store_true",dest="v",
      help='Verbose power information')
  cmd_parser.add_option('-b',
      action="store_true",dest="b",
      help='Search the best path')
  cmd_parser.add_option('-c',
      action="store_true",dest="c",
      help='Extract confidence measures: #WORD PosteriorMax Start End')
  cmd_parser.add_option('-a', '--asf',
      action="store",type="float", default=1.0, dest="asf",
      help='Scale the whole arc prob')
  cmd_parser.add_option('-g', '--gsf',
      action="store",type="float", default=float("-Inf"), dest="gsf",
      help='Grammar Scale Factor')
  cmd_parser.add_option('-w', '--wip',
      action="store",type="float", default=float("-Inf"), dest="wip",
      help='Word Insertion Penalty')
  cmd_parser.add_option('-d', '--dot',
      action="store",type="string", dest="dot_file",
      help='Generate a dot archive from the wordgraph')
  cmd_parser.add_option('-p', '--prune',
      action="store",type="float", dest="prune", default=0.0,
      help='ONLY for dot generation. Prune all edges below the given posterior probability')
  cmd_parser.add_option('-r', '--reference',
      action="store",type="string", default=None, dest="ref",
      help='Include an indicator wether the words is correct (1) or incorrect (0) according to a reference')
  cmd_parser.add_option('--tLalign',
      action="store_true",dest="tLalign",
      help='Format the output like tLalign')


  cmd_parser.parse_args(argv)
  (opts, args)= cmd_parser.parse_args()

  g = WGraph(stdin)
  if opts.v:
    stdout.write("[II] Readed word graph with %d states and %d arcs\n" %(g.NS,g.NE))

  if opts.b:
    g.asf = 1.0 # In recognition there is no asf
    if opts.gsf != float("-Inf"):
      g.gsf = opts.gsf
    if opts.wip != float("-Inf"):
      g.wip = opts.wip
    prob, words = g.best_path()
    stdout.write("Log-probability: %f Sentence: %s\n" %(prob, words))

  if opts.c:
    reference = None
    if opts.ref != None:
      for i in open(opts.ref,"r"):
        reference = i.split()
    res = g.posterior_word_based_CM(opts.asf, opts.gsf, opts.wip, reference)
    if opts.tLalign:
      for i in res:
        j = 0
        for k in i.edge.phos:
          pho, start, end = k.split(',')
          stdout.write("%d %d %s %d %s %g\n"
              %(int(start), int(end)-int(start)+1, pho, j, i.edge.word, exp(i.posterior)))
          j+=1
      stdout.write(".\n")
    else:
      for i in res:
        stdout.write("%s\n" %i)

  if opts.dot_file != None:
    f = open(opts.dot_file, "w")
    g.generate_dot(f, opts.prune, opts.asf,opts.gsf,opts.wip)

if __name__ == "__main__":
  main()
