#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tLtask-recognise: Executes the recognition using transLectures tools
# Copyright 2013,2014 transLectures-UPV Team
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#include i18n
import gettext
_ = gettext.gettext

import argparse
import ConfigParser
from sys import argv, exit, stdin, stdout, stderr, path
import os
import io
import subprocess
import shlex
import datetime
import locale
import re
from time import sleep

path.append(os.path.dirname(os.path.realpath(__file__))+"/scripts")
from tLprogressbar import tLProgressBar

__version__ = filter(str.isdigit, "$Revision: 9628 $")

def show_version(prog):
  stdout.write("%s version: %s\n" %(prog, __version__))

config_file_executables_defaults = """
[Executables]
bindir                   = """+(os.environ['TLK_PATH']+"/" if 'TLK_PATH' in os.environ else "")+"""
tLrecognise_ex           = %(bindir)stLrecognise
tLcmllr_ex               = %(bindir)stLcmllr
tLcmllrfeas_ex           = %(bindir)stLcmllrfeas
tLbin2txt_ex             = %(bindir)stLbin2txt
tLalign_ex               = %(bindir)stLalign
tLdnn-classify_ex        = %(bindir)stLdnn-classify
tLdnn-adapt_ex           = %(bindir)stLdnn-adapt
tLdnn-combine_ex     = %(bindir)stLdnn-combine
submit                   = submit.sh
get_cache                = get_cache.sh
"""

config_file_lists_defaults = """
[Lists]
set-name               = lists/samples
standard-samples       = %(set-name)s.lst
clusters               = %(set-name)s-clusters.lst
clusters-samples       = %(set-name)s-clusters-samples.lst
"""

config_file_general_defaults = """
[General]
prefix-name                 = recognition
tLtask-recognise-version    = """+__version__+"""
submit-options-default      =
submit-options-recognise    =
submit-options-cmllrfeas    =
submit-options-dnn          = -g
submit-options-dnn-adapt    = -g
splits                      = 1
output-dir                  = %(prefix-name)s
cmllr-second-pass           = yes
cmllr-pre-segment-samples   = no
cmllr-pre-segment-use-tLtask-segment = no
adaptation-third-pass       = no
recognition-third-pass      = yes
cache-files                 = no
"""

config_file_hmm_defaults = """
[HMM]
models-dir            = models
prefix-name           = tlk
mixture-components    = 128
feature-vector-size   = 39
standard-model        = %(models-dir)s/%(prefix-name)s.standard.tiedphoneme_I%(mixture-components)s.model
target-model          = %(models-dir)s/%(prefix-name)s.standard.tiedphoneme_I01.model
cmllr-model           = %(models-dir)s/%(prefix-name)s.cmllr.tiedphoneme_I%(mixture-components)s.model
tiedlist              = %(models-dir)s/%(prefix-name)s.standard.tlist
pre-cmllr-tiedlist    = 
cmllr-tiedlist        = %(models-dir)s/%(prefix-name)s.cmllr.tlist
"""

config_file_dnn_defaults = """
[DNN]
apply-on              =
standard-net          =
cmllr-net             =
"""

config_file_dnn_adaptation_defaults = """
[DNN-adaptation]
align-on                = generative
epochs                  = 1
learning-rate           = 0.05
momentum                = 0.3
smooth-samples          = yes
interpolation-weight    = 0.8
"""

config_file_lm_defaults = """
[LM]
language-model        =
lexicon               =
"""

config_file_searchparameters_defaults = """
[Search-parameters]
grammar-scale-factor   = 11
word-insertion-penalty = 0
histogram-pruning      = 10000
beam-pruning           = 140
word-end-pruning       = 90
look-ahead-capacity    = 1000
"""

config_file_wordgraph_defaults = """
[Wordgraph]
create-wordgraphs           = no
wordgraph-density           = 30
extract-confidence-measures = no
confidence-measures-asf     = 10
"""

config_file_defaults = \
[
  config_file_executables_defaults,
  config_file_lists_defaults,
  config_file_general_defaults,
  config_file_hmm_defaults,
  config_file_lm_defaults,
  config_file_dnn_defaults,
  config_file_dnn_adaptation_defaults,
  config_file_searchparameters_defaults,
  config_file_wordgraph_defaults
]

def fatal_err(message):
  stderr.write("[EE] "+ message + "\n")
  exit(1)

def debug_msg(message):
  stderr.write("[II] "+ message + "\n")

class tLrecogniser:
  def __init__(self, conf, verbose=False, filecheck=True, not_launch=False,
      log_folder=None):
    self.conf = conf
    self.last_submitted = None
    self.segment_outdir = None
    self.v = verbose
    self.filecheck = filecheck
    self.not_launch = not_launch
    self.log_folder = log_folder
    self.log_folder_submit_op = []
    if self.log_folder != None:
      self.log_folder_submit_op = ("-o %s" %log_folder).split()
    self.launched_info_file = None
    self.launched_info_filename = None
    if self.conf.get('General','submit-options-default') != "":
      self.submit_ops_default = self.conf.get('General','submit-options-default')
    else: self.submit_ops_default= None

    #Initialise some parameters
    self.prefix = ""
    if self.conf.get('General','prefix-name') != "":
      self.prefix = self.conf.get('General','prefix-name') + "."

  def info_msg(self, message):
    if self.not_launch:
      stderr.write("\n\n# [II] "+ message + "\n\n")
    else:
      stderr.write("[II] "+ message + "\n")

  def submit_using_bashscript(self):
    dirname = os.path.dirname(os.path.realpath(argv[0]))
    command= "bash %s/tLtask-recognise.sh " %dirname
    #put executables path
    command+= "--TLK_RECOGNISE=\"%s\" "  %self.conf.get('Executables','tLrecognise_ex')
    command+= "--TLK_CMLLR=\"%s\" "      %self.conf.get('Executables','tLcmllr_ex')
    command+= "--TLK_CMLLRFEAS=\"%s\" "  %self.conf.get('Executables','tLcmllrfeas_ex')
    command+= "--SUBMIT=\"%s\" "        %self.conf.get('Executables','submit')
    command+= "--GET_CACHE=\'%s\' "     %self.conf.get('Executables','get_cache')
    #List files
    command+= "--LIST=\"%s\" " %self.conf.get('Lists','standard-samples')
    if self.conf.get('General','cmllr-second-pass') == 'yes':
      command+= "--CLUSTERS_list=\"%s\" " %self.conf.get('Lists','clusters')
      command+= "--CLUSTERS_samples=\"%s\" " %self.conf.get('Lists','clusters-samples')
    #Search parameters
    command+= "--GSF=\"%s\" " %self.conf.get('Search-parameters','grammar-scale-factor')
    command+= "--WIP=\"%s\" " %self.conf.get('Search-parameters','word-insertion-penalty')
    command+= "--BEAM=\"%s\" " %self.conf.get('Search-parameters','beam-pruning')
    command+= "--HP=\"%s\" " %self.conf.get('Search-parameters','histogram-pruning')
    command+= "--WEP=\"%s\" " %self.conf.get('Search-parameters','word-end-pruning')
    command+= "--LAC=\"%s\" " %self.conf.get('Search-parameters','look-ahead-capacity')
    #General parameters
    command+= "--EXP_PREFIX=\"%s\" "        %self.conf.get('General','prefix-name')
    command+= "--SUBMIT_OPS=\"%s\" "        %self.conf.get('General','submit-options')
    command+= "--SAMPLESperPROCESS=\"%s\" " %self.conf.get('General','samples-per-process')
    command+= "--OUT_PATH=\"%s\" "          %self.conf.get('General','output-dir')
    #HMMs parameters
    command+= "--HMMS=\"%s\" "          %self.conf.get('HMM','standard-model')
    command+= "--TIEDLIST=\"%s\" "      %self.conf.get('HMM','tiedlist')
    if self.conf.get('General','cmllr-second-pass') == 'yes':
      command+= "--CMLLR_HMMS=\"%s\" "          %self.conf.get('HMM','cmllr-model')
      command+= "--CMLLR_TIEDLIST=\"%s\" "      %self.conf.get('HMM','cmllr-tiedlist')
      command+= "--CMLLR_TARGET=\"%s\" " %self.conf.get('HMM','target-model')
    #LM parameters
    command+= "--LM=\"%s\" "  %self.conf.get('LM','language-model')
    command+= "--LEX=\"%s\" " %self.conf.get('LM','lexicon')
    #Wordgraph parameters
    if self.conf.get('Wordgraph','create-wordgraphs') == 'yes':
      command+= "+WORDGRAPH "
      command+= "--WORDGRAPH_DENS=\"%s\" " %self.conf.get('Wordgraph','wordgraph-density')
    #CMLLR parameters
    if self.conf.get('General','cmllr-second-pass') == 'yes':
      command+= "+CMLLR "

    #Additional options
    command+= "--LOG_FOLDER=\"%s\" " %self.log_folder
    if self.not_launch:
      command+= "+NOT_SUBMIT "

    if not self.v: command+= '-q '
    if not self.filecheck: command+= '-f '
    cmd = shlex.split(command)
    ret = subprocess.call(cmd)
    if ret != 0:
      fatal_err( _( "Failed when executing: '%s' with exit code: %d" ) %(cmd, ret))



  #####################################################################
  ##  submit a process given a name and a command line
  #####################################################################
  def submit(self, submission_name, command_line,
    array=None, depend_on=None, submit_ops=None,
    bgworker=None #launch process in parallel and execute this function
                           #periodically
    ): #None = depend on the launched proc

    submission_name = self.prefix + submission_name
    #submit options
    cmd = [self.conf.get('Executables', 'submit'), "-n ", submission_name]
    depend = None
    if self.last_submitted:
      depend = self.last_submitted
    if depend_on:
      depend = depend_on
    if depend != None and depend != "":
      cmd += ['-w ', depend]
    if array:
      cmd += ['-j ', array]
    if submit_ops:
      cmd += submit_ops.split()
    elif self.submit_ops_default != None:
      cmd += self.submit_ops_default.split()
    if self.log_folder_submit_op != []:
      cmd += self.log_folder_submit_op

    #process
    cmd += command_line

    #Execute
    #Workaround subprocess " " arguments insertions
    string = " ".join(cmd)
    if self.v:
      debug_msg( _( "Submitting: %s\n" ) %string )

    #write to launched info file
    self.launched_info_file.write("%s\n" %string)

    #prepare cmd to execute the process
    cmd = shlex.split(string)
    if self.not_launch:
      stdout.write("%s\n" %string)
    else:
      try:
        if bgworker == None:
          proc = subprocess.Popen(cmd)
          ret = proc.wait()
        else:
          bgworker_function, bgworker_params = bgworker
          proc = subprocess.Popen(cmd)
          ret = bgworker_function(proc, bgworker_params)
        if ret != 0:
          fatal_err( _( "Failed when executing: '%s' with exit code: %d" ) %(string, ret))
      except Exception as e:
        fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(string, e) ) )

    self.last_submitted = submission_name


  ###############################################################################
  #### Writes information about the recognition launched
  ###############################################################################
  def init_launched_info(self):
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M")
    self.launched_info_filename = "launched-recognition.%s.log" %timestamp

    f = open(self.launched_info_filename,"w")
    self.launched_info_file = f


  ###############################################################################
  #### Writes information about the recognition launched
  ###############################################################################
  def print_config_on_launched_info(self):
    debug_msg( _( "Writing information about the recognition in %s" ) %self.launched_info_filename)
    self.launched_info_file.write("### CONFIGURATION FILE USED ###\n\n")
    self.conf.write(self.launched_info_file)
    self.launched_info_file.write("\n### COMMANDS LAUNCHED ###\n\n")

  #####################################################################
  ##  Wrap of checkpath in order to ignore it if necessary
  #####################################################################
  def checkfile(self, path):
    if self.filecheck:
      return os.path.exists(path)
    else:
      return False

  ###########################################################################
  #### Performs the recognition of a list of files using a given model
  ###########################################################################
  def recognise(self,
    name,
    model,
    samples_lst,
    output_dir,
    lexicon,
    tiedlist,
    dnn_network,
    dnn_enabled
    ):

    # Generate DNN files
    if dnn_enabled:
      hybrid_model = "%s/tmp/hybrid.am" %(output_dir)
      if not self.checkfile(hybrid_model):
        self.generate_hybrid_model(
            name,
            model,
            hybrid_model)
      model = hybrid_model

      senones_lst = "%s/tmp/senones.lst" %(output_dir)
      if not self.checkfile(senones_lst):
        self.generate_dnn_senones(
            name,
            samples_lst,
            dnn_network,
            output_dir)
      samples_lst = senones_lst

    splits = int(self.conf.get('General','splits'))
    if splits < 1:
      fatal_err( _( "The number of splits must be at least 1" ) )

    # First generate the sublists and wordgraphs if necessary
    mhrecognise_init_ops = []
    if self.conf.get('Wordgraph','create-wordgraphs') == "yes":
      mhrecognise_init_ops += [ "-w" ]

    if not self.checkfile(output_dir+"/tmp/feas_1") or \
      ( self.conf.get('Wordgraph','create-wordgraphs') == "yes" and
        not self.checkfile(output_dir+"/tmp/wgs_1") ):
      self.submit(name+".recognise-init",
          ["mhrecognise_init.sh"] +
            mhrecognise_init_ops + [
              output_dir,
              samples_lst,
              str(splits)])

    # Create the tiedphoneme lexicon from the monophoneme lexicon and the tied list
    tied_lexicon = output_dir+"/tmp/lexicon-tied.lex"
    if not self.checkfile(tied_lexicon):
      self.submit(name+".create-tied-lex",
          ["lexmono2tri.py",
            "-i", lexicon,
            "-o", tied_lexicon,
            tiedlist])

    # Launch the recognition 
    # Launch only the missing ones
    array = []; last = None
    for i in range(1,splits+1):
      if not self.checkfile(output_dir+("/tmp/outs_%d.done" %i)):
        if last == None:
          last = i
      else:
        if last != None:
          array += ["%d-%d" %(last,i-1)]
          last = None
    if last != None:
      array += ["%d-%d" %(last,splits)]

    if array != []:
      #options for tLrecognise
      recognition_ops = [
        "-g", str(self.conf.get('Search-parameters','grammar-scale-factor')),
        "-w", str(self.conf.get('Search-parameters','word-insertion-penalty')),
        "-m", str(self.conf.get('Search-parameters','histogram-pruning')),
        "-b", str(self.conf.get('Search-parameters','beam-pruning')),
        "-e", str(self.conf.get('Search-parameters','word-end-pruning')),
        "--lookahead-capacity",
              str(self.conf.get('Search-parameters','look-ahead-capacity')),
        "-l", tied_lexicon,
        "--sym-align",
        "-s", "SP",
        "-v" , "-v" ]

      mhrecognise_body_ops = []
      if self.conf.get('Wordgraph','create-wordgraphs') == "yes":
        recognition_ops += [
          "--htk-wgs",
          "-N", str(self.conf.get('Wordgraph', 'wordgraph-density'))
          ]
        mhrecognise_body_ops += ["-w"]

      if self.conf.get('General','cache-files') == "yes":
        mhrecognise_body_ops += [ "-t" ]

      if self.conf.get('Executables','get_cache') != '':
        mhrecognise_body_ops += [ "-g", self.conf.get('Executables', 'get_cache') ]

      depend_on = self.last_submitted
      #set progress bar if launched locally
      bgworker_ex = None
      if self.conf.get('Executables','submit') == "submit.sh":
        bgworker_ex = ( self.bgworker_mhrecognise_body, output_dir+"/tmp" )
      for array_jobs in array:
        self.submit(name+".recognise-body",
            ["mhrecognise_body.sh",
              "-r", self.conf.get('Executables', 'tLrecognise_ex') ] +
              mhrecognise_body_ops +
            [
              output_dir,
              self.conf.get('LM','language-model'),
              model
            ] + recognition_ops,
            array = array_jobs,
            submit_ops = self.conf.get('General','submit-options-recognise'),
            depend_on = depend_on,
            bgworker = bgworker_ex
            )
        self.last_submitted = depend_on
      #sane the last submitted if there were multiple submit
      self.last_submitted = self.prefix + name + ".recognise-body"

    # Parse the output
    if not self.checkfile(output_dir+"/transcription"):
      mhrecognise_finish_ops = []
      if self.conf.get('Wordgraph','create-wordgraphs') == "yes" and \
            self.conf.get('Wordgraph','extract-confidence-measures') == "yes":
        mhrecognise_finish_ops += ["-c"]
        mhrecognise_finish_ops += ["-a" , self.conf.get('Wordgraph','confidence-measures-asf')]
      self.submit(name+".recognise-finish",
          ["mhrecognise_finish.sh" ] +
            mhrecognise_finish_ops + [
            "-p",
            output_dir,
            str(splits),
            output_dir+"/transcription"])

  #####################################################################
  ##  Train a tied phoneme model using CART from an estimated triphoneme model
  #####################################################################
  def bgworker_mhrecognise_body(self, proc, tmp_dir):
    pbar = tLProgressBar()

    init=False
    ret=None
    while True:
      #Check if the process is finished
      ret = proc.poll()
      if ret != None:
        break
      output = subprocess.check_output(['mhrecognise_body_bgworker.sh',tmp_dir])
      output = output.strip()
      v = output.split()
      rec = int(v[0])
      total = int(v[1]) if int(v[1]) > 0 else 1

      #Init bar to show until a samples is recognised
      if init == False:
        pbar.maxval = total
        init = True

      pbar.update(rec)
      sleep(5)

    pbar.finish()
    return ret

  #####################################################################
  ##  Execute tLtask-segment to segment a video file
  #####################################################################
  def launch_segmentation(self, outdir, overr_opts='', segm_trs=''):
    command = "%s " %self.conf.get('Executables','tltask_segment_ex') + \
                ( "--no-launch " if self.not_launch else " " ) + \
                "--log-folder %s " %(self.log_folder) + \
                "--depend-on %s " %self.last_submitted + \
                ("--segmentation-trs %s " % segm_trs if segm_trs != '' else ' ') + \
                "--no-filecheck " + \
                "--print-output-info " + \
                "%s " % self.conf.get('Segmenter','config-file') + \
                "Lists_videos-list=%s/../../video.lst " %( self.conf.get('General', 'output-dir') ) + \
                "General_output-dir=%s " %(outdir+"/segment/out") + \
                "General_tmp-dir=%s " %(outdir+"/segment/tmp") + \
                "General_prefix-name=%s.tLtask-segment " %(self.prefix) + \
                ("General_submit-options-default='%s' " %self.conf.get('General','submit-options-default') if self.conf.get('General','submit-options-default') != '' else '') + \
                ("Executables_submit=%s " %(self.conf.get('Executables','submit')) if self.conf.get('Executables','submit') != '' else '') + \
                "%s " % overr_opts

    cmd = shlex.split(command)
    try:
      ret = subprocess.check_output(cmd)
    except Exception as e:
      fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(command, e) ) )
  
    self.segment_outdir, self.last_submitted = self.parse_output_info(ret)

  #####################################################################
  ##  Parse tLtask output info
  #####################################################################
  def parse_output_info(self, tLtask_output):
    outdir = ''; b1 = False
    last_sub = ''; b2 = False
    for l in tLtask_output.split('\n'):
        if b1 and b2: break
        if len(l.split()) < 2: continue
        if l.split()[0] == '__OUTPUT_DIR:':
            outdir = ' '.join(l.split()[1:]).strip()
            b1 = True
        elif l.split()[0] == '__LAST_SUBMITTED:':
            last_sub = ' '.join(l.split()[1:]).strip()
            b2 = True
    if not(b1 and b2):
        fatal_err( _( "Could not retrieve __OUTPUT_DIR and __LAST_SUBMMITED tokens from the recogniser output." ))
    return outdir, last_sub


  #####################################################################
  ##  Train a tied phoneme model using CART from an estimated triphoneme model
  #####################################################################
  def generate_CMLLR_samples(self, name):
    output_dir = self.conf.get('General', 'output-dir')

    samples_lst = self.conf.get('Lists','standard-samples')
    phos_lst = "%s/standard_step1/tmp/phos.lst" % self.conf.get('General','output-dir')
    monophos_lst = "%s/standard_step1/tmp/monophos.lst" %self.conf.get('General','output-dir')
    samples_clusters_samples_lst = self.conf.get('Lists','clusters-samples')

    #Shorten standard samples heuristicaly

    if self.conf.get('General','cmllr-pre-segment-samples') == 'yes':
      if not self.checkfile("%s/cmllr_step2/tmp/pre-segment/samples.lst"
          %(output_dir)):
        args = ['segment-pre-cmllr.py',
            '--tLbin2txt', self.conf.get('Executables','tLbin2txt_ex'),
            '--feature-vector-size', self.conf.get('HMM','feature-vector-size'),
            "%s/standard_step1/transcription.align" %(output_dir),
            self.conf.get('Lists','standard-samples'),
            self.conf.get('Lists','clusters-samples'),
            "%s/cmllr_step2/tmp/pre-segment" %(output_dir)]
        if self.conf.get('General','cmllr-pre-segment-use-tLtask-segment') == 'yes':
          args.append('--fulla-comb-CNN-DNN')
        self.submit(name+".pre-segment-samples", args)
        if self.conf.get('General','cmllr-pre-segment-use-tLtask-segment') == 'yes':
          self.launch_segmentation("%s/cmllr_step2/tmp/pre-segment" %(output_dir), segm_trs="%s/cmllr_step2/tmp/pre-segment/segmentation.trs" %(output_dir))
          samples_lst = "%s/cmllr_step2/tmp/pre-segment/segment/out/allfeas.lst" %(output_dir)
        else:
          samples_lst = "%s/cmllr_step2/tmp/pre-segment/samples.lst" %(output_dir)
      phos_lst = "%s/cmllr_step2/tmp/pre-segment/samples-phos.lst" %(output_dir)
      monophos_lst = "%s/cmllr_step2/tmp/pre-segment/samples-monophos.lst" %(output_dir)
      samples_clusters_samples_lst = \
              "%s/cmllr_step2/tmp/pre-segment/samples-clusters-samples.lst" %(output_dir)

    if self.conf.get('HMM','pre-cmllr-tiedlist') != None and self.conf.get('HMM','pre-cmllr-tiedlist').strip() != "":
      self.change_tiedphos(
         "recognition.cmllr", 
         monophos_lst,
         self.conf.get('HMM','pre-cmllr-tiedlist'), 
         "%s/cmllr_step2" % self.conf.get('General','output-dir'))
      phos_lst = "%s/cmllr_step2/tmp/change_tiedphos/new_tiedphos.lst" % self.conf.get('General','output-dir')

    step2_dir = "%s/cmllr_step2" %(output_dir)
    #Generate the clusters lists
    cmllrfeas_lst = "%s/tmp/cmllrfeas.lst" %(step2_dir)
    if not self.checkfile(cmllrfeas_lst):
      self.submit(name+".create-clusters",
          ['generate-clusters.sh',
          samples_lst,
          phos_lst,
          samples_clusters_samples_lst,
          self.conf.get('Lists','clusters'),
          step2_dir+"/tmp",
          cmllrfeas_lst
          ])

    cluster_name_dependency = self.last_submitted
    #Launch CMLLR estimation and sample generation
    allprocs = ""
    for i in open(self.conf.get('Lists', 'clusters'),"r"):
      cluster = i[:-1]
      matrix_file = "%s/tmp/matrices/%s.matrix" %(step2_dir, cluster)
      if not self.checkfile(matrix_file):
        self.submit(name+".generate-features-cluster-%s" %cluster,
            ['cmllr-features.sh',
            '-e', self.conf.get('Executables','tLcmllr_ex'),
            '-g', self.conf.get('Executables','tLcmllrfeas_ex'),
            cluster,
            self.conf.get('HMM','target-model'),
            "%s/tmp/clusters/%s-phos_tied.lst" %(step2_dir, cluster),
            "%s/tmp/clusters/%s.lst" %(step2_dir, cluster),
            step2_dir+"/tmp",
            "%s/tmp/clusters/%s-cmllrfeas.lst" %(step2_dir, cluster)
            ],
            depend_on=cluster_name_dependency,
            submit_ops = self.conf.get('General','submit-options-cmllrfeas')
            )
        allprocs += self.last_submitted + ","

    self.last_submitted = allprocs[:-1]


  #####################################################################
  ##  Change tiedphos
  #####################################################################
  def change_tiedphos(self, name, in_monophos_lst, tlist, output_dir):
    self.info_msg( _( "Change recognition output tiedphos" ) )
    self.submit(name+".change_tiedphos", 
        ["generate-tiedphos-from-phos.sh", 
         in_monophos_lst, 
         tlist, 
         output_dir+"/tmp/change_tiedphos/new_tiedphos.lst",
         output_dir+"/tmp/change_tiedphos",
         "SP"])

  #####################################################################
  ##  Generate hybrid model
  #####################################################################
  def generate_hybrid_model(self, name, input_model, output_model):
    self.info_msg( _( "Generate the hybrid model" ) )
    self.submit(name+".generate-hybrid-model",
        ["generate-hybrid-model.sh",
          "-b", self.conf.get("Executables","tLbin2txt_ex"),
          input_model,
          output_model])

  #####################################################################
  ##  Generate hybrid model
  #####################################################################
  def generate_dnn_senones(self, name, samples_lst, dnn_network,
    output_dir):

    self.info_msg( _( "Generate the DNN senones" ) )
    self.submit(name+".generate-senones",
        ["generate-senones.sh",
          "-c", self.conf.get("Executables","tLdnn-classify_ex"),
          samples_lst,
          dnn_network,
          output_dir],
          submit_ops = self.conf.get('General','submit-options-dnn')
          )

  ##################################################################################
  ##  Generate state alignment for DNN adaptation
  ##################################################################################
  def dnn_align_samples(self, name, transcriptions_list, samples_list, model, dest_dir, rename_samples_lst=""):

    if not self.checkfile(dest_dir+"/model_0"):
      #Copy model and split lists
      self.submit(name+".align-init",
          ["mhtrain_init.sh", dest_dir,
          model, transcriptions_list, samples_list, self.conf.get('General','splits')])

    if not self.checkfile(dest_dir+"/state.labels"):
      #align training at state model
      self.submit(name+".align",
          ["mhdnntrain_align.sh",
            "-e", self.conf.get('Executables','tLalign_ex'),
            dest_dir],
            array=("1-%d" %int(self.conf.get('General','splits'))))

      #puts all together the aligns
      cmd_args = [ "mhdnntrain_align_finish.sh", "-e", self.conf.get('Executables','tLbin2txt_ex') ]
      if rename_samples_lst != "":
        cmd_args.append("-r")
        cmd_args.append(rename_samples_lst)
      cmd_args = cmd_args + [ dest_dir, self.conf.get('General','splits') ]
      self.submit(name+".align-finish", cmd_args)


  ##################################################################################
  ##  Generate adapted DNN samples
  ##################################################################################
  def dnn_generate_adapted_samples(self, name, samples_list, net, dest_dir, previous_dir, comb_feas=''):

    if not self.checkfile(dest_dir+"/dnnfeas-combined.lst"):

      cmd_args = [ "generate-adapted-combined-senones.sh",
        "-c", self.conf.get('Executables','tLdnn-classify_ex'),
        "-a", self.conf.get('Executables','tLdnn-adapt_ex'),
        "-o", self.conf.get('Executables','tLdnn-combine_ex') ]

      if self.conf.get('DNN-adaptation', 'smooth-samples') == 'no':
        cmd_args.append("-n")

      if comb_feas != '':
        cmd_args.append("-l")
        cmd_args.append(comb_feas)

      cmd_args = cmd_args + [ dest_dir,
        previous_dir,
        samples_list,
        net,
        self.conf.get('DNN-adaptation','epochs'),
        self.conf.get('DNN-adaptation','learning-rate'),
        self.conf.get('DNN-adaptation','momentum'),
        self.conf.get('DNN-adaptation','interpolation-weight') ]

      self.submit(name+".generate-combined-senones", cmd_args,
        submit_ops = self.conf.get('General','submit-options-dnn-adapt'))

  #####################################################################
  ##  Combine Systems Senones
  #####################################################################
  def combine_systems_senones(self, name, senones_lst1, senones_lst2, output_dir, c_lambda=0.5):
    self.info_msg( _( "System senones combination" ) )
    self.submit(name+".combine-senones", 
        ["generate-combined-senones.sh", 
         "-o", self.conf.get('Executables','tLdnn-combine_ex'),
         senones_lst1, 
         senones_lst2, 
         output_dir, 
         str(c_lambda)],
         submit_ops = self.conf.get('General','submit-options-dnn'))

  #####################################################################
  ##  Delete tmp
  #####################################################################
  def delete_tmp(self):
    self.submit("delete-tmp",
        ['rm', '-rf', self.conf.get('General','output-dir')+'/*/tmp'])


#####################################################################
#####################################################################
#####################################################################
def main():
  #i18n support
  locale.setlocale(locale.LC_MESSAGES, '')
  locale.setlocale(locale.LC_CTYPE, '')
  PACKAGE = "tlk"
  LOCALEDIR = ""
  if "PACKAGE" in os.environ:
    PACKAGE = os.environ["PACKAGE"]
  if "LOCALEDIR" in os.environ:
    LOCALEDIR = os.environ["LOCALEDIR"]
  gettext.bindtextdomain(PACKAGE, LOCALEDIR)
  gettext.textdomain(PACKAGE)

  config = ConfigParser.ConfigParser(allow_no_value=True)
  #Load defaults
  for i in config_file_defaults:
    config.readfp(io.BytesIO(i))
#config.set('General','tLtask-recognise-version',0)

  parser = argparse.ArgumentParser(
      prog="tLtask-recognise",
      description=_('Recognize an input list of samples using TLK tools'),
      epilog=_("""tLtask-recognise is used to perform the recognition of a list
of audio samples in TLK format using TLK. This tool simplifies the
recognition process, as it only requires the definition of a task via a
configuration file, rather than than being very well acquainted with the
details of Automatic Speech Recognition (ASR) and executing several sequential
TLK core utilities, such as tLrecognise or tLcmllr"""))

  parser.add_argument(dest='config_file',
    help=_("Load configuration from this file"), nargs='?')
  parser.add_argument("--write-example-config-file", action='store_true',dest='write_example',
    help=_("Write the default configuration to the standard output and exit"))
  parser.add_argument("-S", "--scripts-dir", dest='scripts_path',
    help=( _( "Directory containing the recognition scripts." ) ), metavar="DIRNAME")
  parser.add_argument("--no-launch", dest='not_launch',
    help=_("Do not submit any command, only print them"), action="store_true")
  parser.add_argument("--no-filecheck", dest='filecheck',
    help=_("Disable file checking. Using this option forces the recognition to be performed from scratch"), action="store_false")
  parser.add_argument('--log-folder',
    help=_('Store all logs in this folder'), metavar="DIRNAME",
    dest='log_folder', default='log')
  parser.add_argument('--delete-tmp',
    help=_('Delete the tmp folder once the recognition is finished'),
    action='store_true')
  parser.add_argument('--depend-on',
    help=_('The first process depends on this if defined. Only available in grid environments.'),
    dest='depend_on')
  parser.add_argument('--print-output-info', dest='print_output_info',
    help=_('Prints to stdout the output directory and the last process name.'),
    action='store_true')
  parser.add_argument('-v',
    help=_('Output additional verbose information'),
    action='store_true')
  parser.add_argument('--version',
    help=_('Output version information and exit'),
    action='store_true')
  parser.add_argument('--comb-senones-lst', dest='comb_feas', default='',
    help=_('Combine current senones with this senones list after dnn adaptation.'))

  parser.add_argument(dest='overwrite',
    help=_("Overwrite variables from config. Format -> Section_Variable=Value"), nargs='*')

  args = parser.parse_args(argv[1:])

  # Assigns this enviorement variable to set the name of all log messages
  os.putenv('PRGNAME', 'tLtask-recognise')

  #write a sample config file
  if args.write_example:
    config.write(stdout)
    exit(0)

  #show version (according to translectures-revision SVN)
  if args.version:
    show_version("tLtask-recognise")
    exit(0)

  #Modify scripts PATH if necessary
  if args.scripts_path:
    os.putenv('PATH', os.environ['PATH']+':'+args.scripts_path)
  else: #add the scripts folder on this executable path
    dirname = os.path.dirname(os.path.realpath(argv[0]))
    os.environ['PATH'] = os.environ['PATH']+':'+dirname+'/scripts'

  if args.not_launch:
    stdout.write("export PATH=%s\n" %os.environ['PATH'])

  #read an external config file and check version number
  if not args.config_file:
    parser.error( _( """A config_file must be given. Use --write-example-config-file
to generate a template of the configuration file.""") )
  else:
    config.read(args.config_file)

    if not config.has_option('General','tLtask-recognise-version') or \
        ( config.get('General','tLtask-recognise-version') != '_RELEASE_' and \
        config.get('General','tLtask-recognise-version') != __version__ ):
      fatal_err( _(
"Wrong version number in loaded configuration file. Some options may have\n"
"changed or not be supported anymore. Check your configuration, and\n"
"update the tLtask-recognise-version to %d if everything is correct."
        ) %int(__version__) )

  #Overwrite variables with command line
  if args.overwrite != None:
    for o in args.overwrite:
      m = re.search("^(.*)_(.*)=(.*)$", o)
      if m == None:
        fatal_err( _( "Incorrect overwrite in element %s" ) %o )
      debug_msg("Setting [%s] %s = %s" %(m.group(1), m.group(2), m.group(3)))
      if config.has_option(m.group(1),m.group(2)):
        config.set(m.group(1), m.group(2), m.group(3))
      else:
        fatal_err( _( "Section and options of overwrite command %s do not exists in config" ) %o )

  #Filechecks
  if args.filecheck:
    if not os.path.exists(config.get('Lists','standard-samples')):
      fatal_err( _( "Standard samples list: %s not found\n" ) %config.get('Lists','standard-samples'))

    if not os.path.exists(config.get('LM','language-model')):
      fatal_err( _( "Language model file: %s not found\n" ) %config.get('LM','language-model'))

    if not os.path.exists(config.get('LM','lexicon')):
      fatal_err( _( "Lexicon file: %s not found\n" ) %config.get('LM','lexicon'))

    if config.get('General','cmllr-second-pass') == 'yes':
      if not os.path.exists(config.get('Lists','clusters')):
        fatal_err( _( "Clusters list: %s not found\n" ) %config.get('Lists','clusters'))

      if not os.path.exists(config.get('Lists','clusters-samples')):
        fatal_err( _( "Cluster samples list: %s not found\n" ) %config.get('Lists','clusters-samples'))

  if not os.path.exists(args.log_folder):
    os.mkdir(args.log_folder)
  elif not os.path.isdir(args.log_folder):
    debug_msg( _( "Log folder is an archive. All logs will be appended to this file" ) )

  #Prepare command to be submitted to tLtask-recognise.sh
  recogniser = tLrecogniser(config, verbose=args.v, filecheck=args.filecheck,
      not_launch=args.not_launch, log_folder=args.log_folder)

  # Add an external dependency to the first submit
  if args.depend_on:
    recogniser.last_submitted = args.depend_on

  #if args.submit_using_bashscript:
  #  recogniser.submit_using_bashscript()
  #  exit(0)

  #HMM filechecks
  if args.filecheck:
    #standard step hmm files
    if not os.path.exists(config.get('HMM','standard-model')):
      fatal_err( _( "Standard model: %s not found\n" ) %config.get('HMM','standard-model'))

    if not os.path.exists(config.get('HMM','tiedlist')):
      fatal_err( _( "Tiedlist: %s not found\n" ) %config.get('HMM','tiedlist'))

    if "standard" in config.get('DNN','apply-on').split(",") and \
        not os.path.exists(config.get('DNN','standard-net')):
      fatal_err( _( "Standard DNN: %s not found\n" ) %config.get('DNN','standard-net'))

    if config.get('General','cmllr-second-pass') == 'yes':
    #cmllr step hmm files
      if not os.path.exists(config.get('HMM','target-model')):
        fatal_err( _( "Target model: %s not found\n" ) %config.get('HMM','target-model'))

      if not os.path.exists(config.get('HMM','cmllr-tiedlist')):
        fatal_err( _( "CMLLR Tiedlist: %s not found\n" ) %config.get('HMM','cmllr-tiedlist'))

      if not os.path.exists(config.get('HMM','cmllr-model')):
        fatal_err( _( "CMLLR model: %s not found\n" ) %config.get('HMM','cmllr-model'))

      if "cmllr" in config.get('DNN','apply-on').split(",") and \
          not os.path.exists(config.get('DNN','cmllr-net')):
        fatal_err( _( "CMLLR DNN: %s not found\n" ) %config.get('DNN','cmllr-net'))

  ### Prepare for submitting the recognition
  ### Store configuration employed in a file
  recogniser.init_launched_info()
  recogniser.print_config_on_launched_info()

  debug_msg("Submitting the recognition")
   
  output_dir = ''

  recogniser.info_msg( _( "Recognising the samples using the standard model" ) )
  recogniser.recognise(
    "recognition.standard-step1",
    config.get('HMM','standard-model'),
    config.get('Lists','standard-samples'),
    config.get('General','output-dir')+"/standard_step1",
    config.get('LM','Lexicon'),
    config.get('HMM','tiedlist'),
    config.get('DNN','standard-net'),
    "standard" in [x.strip() for x in config.get('DNN','apply-on').split(",")]
    )
  output_dir = config.get('General','output-dir')+"/standard_step1"

  # CMLLR second pass
  if config.get('General','cmllr-second-pass') == 'yes':

    recogniser.info_msg( _( "Generating CMLLR samples" ) )
    recogniser.generate_CMLLR_samples(
      "recognition.cmllr"
      )

    recogniser.info_msg( _( "Recognising the samples using the CMLLR model" ) )
    recogniser.recognise(
      "recognition.cmllr-step2",
      config.get('HMM','cmllr-model'),
      config.get('General','output-dir')+"/cmllr_step2/tmp/cmllrfeas.lst",
      config.get('General','output-dir')+"/cmllr_step2",
      config.get('LM','Lexicon'),
      config.get('HMM','cmllr-tiedlist'),
      config.get('DNN','cmllr-net'),
      "cmllr" in [x.strip() for x in config.get('DNN','apply-on').split(",")]
      )
    output_dir = config.get('General','output-dir')+"/cmllr_step2"

  # DNN adaptation third pass
  
  if config.get('General','adaptation-third-pass') == 'yes' and \
    "cmllr" in [x.strip() for x in config.get('DNN','apply-on').split(",")]:

    recogniser.info_msg( _( "Adapting DNN for CMLLR" ) )

    if config.get('DNN-adaptation', 'align-on') == 'generative':
      model = "%s" %config.get('HMM','cmllr-model')
      feas_lst = "%s/cmllr_step2/tmp/cmllrfeas.lst" %config.get('General','output-dir')
      rename_lst = ""
    elif config.get('DNN-adaptation', 'align-on') == 'hybrid':
      model = "%s/cmllr_step2/tmp/hybrid.am" %config.get('General','output-dir')
      feas_lst = "%s/cmllr_step2/tmp/senones.lst" %config.get('General','output-dir')
      rename_lst = config.get('General','output-dir')+"/cmllr_step2/tmp/cmllrfeas.lst"
    else:
      fatal_err( _( "Un-allowed 'DNN-adaptation' -> 'align-on' option in config file." ) )

    phos_lst = "%s/cmllr_step2/tmp/phos.lst" %config.get('General','output-dir')

    recogniser.dnn_align_samples (
        "recognition.adaptation-step3", phos_lst, feas_lst, model,
        "%s/adaptation_step3/tmp/" %config.get('General','output-dir'), rename_samples_lst=rename_lst)

    recogniser.info_msg( _( "Generating the adapted DNN senones" ) )
    recogniser.dnn_generate_adapted_samples(
        "recognition.adaptation-step3",
        config.get('General','output-dir')+"/cmllr_step2/tmp/cmllrfeas.lst",
        config.get('DNN','cmllr-net'),
        "%s/adaptation_step3/tmp/" %config.get('General','output-dir'),
        "%s/cmllr_step2/tmp/" %config.get('General','output-dir'),
        comb_feas=args.comb_feas
        )
    if config.get('General','recognition-third-pass') == 'yes': 
      recogniser.info_msg( _( "Recognising the adapted samples using the DNN-CMLLR model" ) )
      recogniser.recognise(
        "recognition.adaptation-step3",
        config.get('General','output-dir')+"/cmllr_step2/tmp/hybrid.am",
        config.get('General','output-dir')+"/adaptation_step3/tmp/dnnfeas-combined.lst",
        config.get('General','output-dir')+"/adaptation_step3",
        config.get('LM','Lexicon'),
        config.get('HMM','cmllr-tiedlist'),
        config.get('DNN','cmllr-net'),
        False
        )
    output_dir = config.get('General','output-dir')+"/adaptation_step3"

  if args.print_output_info:
    stdout.write("__OUTPUT_DIR: %s\n" % os.path.abspath(output_dir))
    if recogniser.last_submitted != None:
      stdout.write("__LAST_SUBMITTED: %s\n" % recogniser.last_submitted)
    else:
       stdout.write("\n")

  #####################   Common steps   ############################
  if args.delete_tmp:
    #8 delete tmp-folder
    recogniser.delete_tmp()

####################################################################################################
if __name__ == "__main__":
  main()

