#!/bin/bash
#
#  Copyright 2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

set -eu

function print_help {
    cat <<EOF
$0 <wavs-list> <trs-list> <output-dir>

  Generate TLK samples from the speech segments defined in the .trs files.
  

  <wavs-list>       List of wavs to extract the TLK samples
  <trs-list>        List with the trs files marking the segmentation
  <output-dir>      Otput directory to store TLK samples

Options:

  -h                Shows this help
  -C <num>          Number of cepstrals
  -o <str>          tLextract options

EOF
}

TLEXTRACTEX=tLextract
TLEXTRACTOPTS=""
NUMARGS=4
FEAS=15
while getopts "hC:o:" OPT
do
    case $OPT in
        h) print_help; exit 0;;
        C) FEAS=$OPTARG;;
        o) TLEXTRACTOPTS="$OPTARG";;
    esac
done
[ $[$#-OPTIND+1] -eq $NUMARGS ] || { print_help >&2; exit 1; }
shift $[OPTIND-1]

set -x

WAV_LST=$1
TRS_LST=$2
TMPDIR=$3
OUTDIR=$4

function log {
    echo -e "$0: [II] $@" >&2
}

TMP_WAV_LST=`mktemp wavXXXlst`
paste -d ' ' $WAV_LST $TRS_LST | while read WAV TRS; do
  WAV_BASENAME=`echo $WAV | sed -e 's+^tmp/++' -e 's+\.wav$++'`
  segment_trs_wav.py -b B $WAV $TRS $TMPDIR/segment/$WAV_BASENAME

  #Listing all wavs
  find $TMPDIR/segment/$WAV_BASENAME/sp/ -name '*.wav' > $TMP_WAV_LST

  #Making the feature list and creating their directories
  TMP_TLFEA_LST="$OUTDIR/$WAV_BASENAME.lst"
  mkdir -p `dirname $TMP_TLFEA_LST`
  cat $TMP_WAV_LST | sed -e "s+$TMPDIR/segment+$OUTDIR+" -e 's+\.wav$+.tLfea+' \
    > $TMP_TLFEA_LST
  mkdir -p `cat $TMP_TLFEA_LST | sed -e 's+\(.*/\).*$+\1+' | sort -u`

  #Generating samples
  log $"Generating samples"
  if [ "$TLEXTRACTOPTS" != "" ]; then
    COMMAND="$TLEXTRACTEX $TLEXTRACTOPTS -n $TMP_WAV_LST $TMP_TLFEA_LST"
  else
    COMMAND="$TLEXTRACTEX -C $FEAS -n $TMP_WAV_LST $TMP_TLFEA_LST"
  fi
  log $"Executing: $COMMAND"
  $COMMAND
done

#Listing all feature files created
rm -f $OUTDIR/generated-videos.lst $OUTDIR/allfeas.lst
for i in `< $WAV_LST`; do
  LIST=`echo $i | sed -e 's+^tmp/++' -e 's+\.wav$+.lst+'`
  echo $OUTDIR/$LIST >> $OUTDIR/generated-videos.lst
  sort $OUTDIR/$LIST >> $OUTDIR/allfeas.lst
done 

rm -r $TMP_WAV_LST

