#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tLtask-segment: Segments an video file 
# Copyright 2014 transLectures-UPV Team
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#include i18n
import gettext
_ = gettext.gettext

import argparse
import ConfigParser
from sys import argv, exit, stdin, stdout, stderr
import os
import io
import subprocess
import shlex
import datetime
import locale
import re

__version__ = filter(str.isdigit, "$Revision: 9571 $")

def show_version(prog):
  stdout.write("%s version: %s\n" %(prog, __version__))

config_file_executables_defaults = """
[Executables]
bindir                   = """+(os.environ['TLK_PATH']+"/" if 'TLK_PATH' in os.environ else "")+"""
tLrecognise_ex           = %(bindir)stLrecognise
tLextract_ex             = %(bindir)stLextract
submit                   = submit.sh
get_cache                = get_cache.sh
"""

config_file_lists_defaults = """
[Lists]
set-name               = lists/samples
videos-list       = %(set-name)s.lst
"""

config_file_general_defaults = """
[General]
prefix-name               = tlk
tLtask-segment-version  = """+__version__+"""
submit-options-default    =
output-dir                = %(prefix-name)s
tmp-dir                   = tmp
"""

config_file_segmenter_defaults = """
[Segmenter]
acoustic-model            =
language-model            =
grammar-scale-factor      = 11
histogram-prunning        = 10000
beam-prunning             = 140
cepstrals-extracted       = 15
tLextract-segment-options =
tLextract-speech-options  =
"""

config_file_defaults = \
[
  config_file_executables_defaults,
  config_file_lists_defaults,
  config_file_general_defaults,
  config_file_segmenter_defaults
]

def fatal_err(message):
  stderr.write("[EE] "+ message + "\n")
  exit(1)

def debug_msg(message):
  stderr.write("[II] "+ message + "\n")

class tLsegmenter:
  def __init__(self, conf, verbose=False, filecheck=True, not_launch=False,
      log_folder=None):
    self.conf = conf
    self.last_submitted = None
    self.v = verbose
    self.filecheck = filecheck
    self.not_launch = not_launch
    self.log_folder = log_folder
    self.log_folder_submit_op = []
    if self.log_folder != None:
      self.log_folder_submit_op = ("-o %s" %log_folder).split()
    self.launched_info_file = None
    self.launched_info_filename = None
    if self.conf.get('General','submit-options-default') != "":
      self.submit_ops_default = self.conf.get('General','submit-options-default')
    else: self.submit_ops_default= None

    #Initialise some parameters
    self.prefix = ""
    if self.conf.get('General','prefix-name') != "":
      self.prefix = self.conf.get('General','prefix-name') + "."

  def info_msg(self, message):
    if self.not_launch:
      stdout.write("\n\n# [II] "+ message + "\n\n")
    else:
      stdout.write("[II] "+ message + "\n")

  #####################################################################
  ##  submit a process given a name and a command line
  #####################################################################
  def submit(self, submission_name, command_line,
    array=None, depend_on=None, submit_ops=None): #None = depend on the launched proc

    submission_name = self.prefix + submission_name
    #submit options
    cmd = [self.conf.get('Executables', 'submit'), "-n ", submission_name]
    depend = None
    if self.last_submitted:
      depend = self.last_submitted
    if depend_on:
      depend = depend_on
    if depend != None and depend != "":
      cmd += ['-w ', depend]
    if array:
      cmd += ['-j ', array]
    if submit_ops:
      cmd += submit_ops.split()
    elif self.submit_ops_default != None:
      cmd += self.submit_ops_default.split()
    if self.log_folder_submit_op != []:
      cmd += self.log_folder_submit_op

    #process
    cmd += command_line

    #Execute
    #Workaround subprocess " " arguments insertions
    string = " ".join(cmd)
    if self.v:
      debug_msg( _( "Submitting: %s\n" ) %string )

    #write to launched info file
    self.launched_info_file.write("%s\n" %string)

    #prepare cmd to execute the process
    cmd = shlex.split(string)
    if self.not_launch:
      stdout.write("%s\n" %string)
    else:
      try:
        ret = subprocess.call(cmd)
        if ret != 0:
          fatal_err( _( "Failed when executing: '%s' with exit code: %d" ) %(string, ret))
      except Exception as e:
        fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(string, e) ) )

    self.last_submitted = submission_name


  ###############################################################################
  #### Writes information about the training launched
  ###############################################################################
  def init_launched_info(self):
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M")
    self.launched_info_filename = "launched-segmentation.%s.log" %timestamp

    f = open(self.launched_info_filename,"w")
    self.launched_info_file = f


  ###############################################################################
  #### Writes information about the training launched
  ###############################################################################
  def print_config_on_launched_info(self):
    debug_msg( _( "Writing information about the segmentation in %s" ) %self.launched_info_filename)
    self.launched_info_file.write("### CONFIGURATION FILE USED ###\n\n")
    self.conf.write(self.launched_info_file)
    self.launched_info_file.write("\n### COMMANDS LAUNCHED ###\n\n")

  #####################################################################
  ##  Wrap of checkpath in order to ignore it if necessary
  #####################################################################
  def checkfile(self, path):
    if self.filecheck:
      return os.path.exists(path)
    else:
      return False

  ###########################################################################
  #### Performs the extraction of TLK features from a list of videos
  ###########################################################################
  def extract_audio(self, name):
    if not os.path.exists(self.conf.get('General','tmp-dir')+"/tLfeas.lst"):
      cmd_args = ['extract-audio.sh']
      tLex_opts = self.conf.get('Segmenter','tLextract-segment-options')
      if tLex_opts  not in ("", None):
        cmd_args.append("-o") 
        cmd_args.append("'\"%s\"'" % tLex_opts)
      cmd_args = cmd_args + [self.conf.get('Lists','videos-list'), self.conf.get('General','tmp-dir')]
      self.submit(name+".extract-audio", cmd_args)

  #################################################################################
  #### Recognises a lists of TLK samples to divide them into speech and non-speech
  #################################################################################
  def segment_audio(self, name):

    tmpdir = self.conf.get('General','tmp-dir')
    if not os.path.exists(tmpdir+"/frametags.align"):
      self.submit(name+".audio-segmentation",
          ['audio-segmentation.sh',
            "-r", self.conf.get('Executables','tLrecognise_ex'),
            "-b", self.conf.get('Segmenter','beam-prunning'),
            "-g", self.conf.get('Segmenter','grammar-scale-factor'),
            "-m", self.conf.get('Segmenter','histogram-prunning'),
            tmpdir+"/tLfeas.lst",
            tmpdir,
            self.conf.get('Segmenter','language-model'),
            self.conf.get('Segmenter','acoustic-model')
            ])

  #################################################################################
  #### Generate speech samples from the segmentation
  #################################################################################
  def generate_speech_samples(self, name, segm_trs=''):

    tmpdir = self.conf.get('General','tmp-dir')
    outputdir = self.conf.get('General','output-dir')
    if not os.path.exists(outputdir+"/allfeas.lst"):
      cmd_args = ['generate-speech-samples.sh']
      tLex_opts = self.conf.get('Segmenter','tLextract-speech-options')
      if tLex_opts not in ("", None):
        cmd_args.append("-o")
        cmd_args.append("'\"%s\"'" % tLex_opts)
      else:
        cmd_args.append('-C')
        cmd_args.append(self.conf.get('Segmenter','cepstrals-extracted'))
      if segm_trs != '':
        if not(os.path.exists(tmpdir)):
          os.makedirs(tmpdir)
        with open(tmpdir+"/trs.lst", 'w') as fd:
          fd.write("%s\n" % segm_trs)
      cmd_args = cmd_args + [tmpdir+"/wavs.lst", tmpdir+"/trs.lst", tmpdir, outputdir]
      
      self.submit(name+".generate-speech-samples", cmd_args)

  #####################################################################
  ##  Delete tmp
  #####################################################################
  def delete_tmp(self):
    self.submit("delete-tmp",
        ['rm', '-rf', self.conf.get('General','tmp-dir')])

#####################################################################
#####################################################################
#####################################################################
def main():
  #i18n support
  locale.setlocale(locale.LC_MESSAGES, '')
  locale.setlocale(locale.LC_CTYPE, '')
  PACKAGE = "tlk"
  LOCALEDIR = ""
  if "PACKAGE" in os.environ:
    PACKAGE = os.environ["PACKAGE"]
  if "LOCALEDIR" in os.environ:
    LOCALEDIR = os.environ["LOCALEDIR"]
  gettext.bindtextdomain(PACKAGE, LOCALEDIR)
  gettext.textdomain(PACKAGE)

  config = ConfigParser.ConfigParser(allow_no_value=True)
  #Load defaults
  for i in config_file_defaults:
    config.readfp(io.BytesIO(i))
#config.set('General','tLtask-recognise-version',0)

  parser = argparse.ArgumentParser(
      prog="tLtask-segment",
      description=_('Segments an input list of videos using TLK tools'),
      epilog=_("""tLtask-segment is used to perform the segmentation in speech
fragments of a list of videos. This tool simplifies the
segmentation process, as it only requires the definition of a task via a
configuration file, rather than being very well acquainted with the
details of Automatic Speech Recognition (ASR) and executing several sequential
TLK core utilities."""))

  parser.add_argument(dest='config_file',
    help=_("Load configuration from this file"), nargs='?')
  parser.add_argument("--write-example-config-file", action='store_true',dest='write_example',
    help=_("Write the default configuration to the standard output and exit"))
  parser.add_argument("-S", "--scripts-dir", dest='scripts_path',
    help=( _( "Directory containing the training scripts" ) ), metavar="DIRNAME")
  parser.add_argument("--no-launch", dest='not_launch',
    help=_("Do not submit any command, only print them"), action="store_true")
  parser.add_argument("--no-filecheck", dest='filecheck',
    help=_("Disable file checking. Use of this option forces the recognition to be performed from scratch"), action="store_false")
  parser.add_argument('--log-folder',
    help=_('All logs will be stored in this folder'), metavar="DIRNAME",
    dest='log_folder', default='log')
  parser.add_argument('--delete-tmp',
    help=_('Delete the tmp folder once the training is finished'),
    action='store_true')
  parser.add_argument('--depend-on',
    help=_('If defined, the first process depends on this. Only available in grid environments.'),
    dest='depend_on')
  parser.add_argument('--segmentation-trs', dest='segmentation_trs', default='',
    help=_('Use the segmentation of the provided trs file.'))
  parser.add_argument('--print-output-info', dest='print_output_info',
    help=_('Prints to stdout the output directory and the last process name.'),
    action='store_true')
  parser.add_argument('-v',
    help=_('Output additional verbose information'),
    action='store_true')
  parser.add_argument('--version',
    help=_('Output version information and exit'),
    action='store_true')

  parser.add_argument(dest='overwrite',
    help=_("Overwrite variables from config. Format -> Section_Variable=Value"), nargs='*')

  args = parser.parse_args(argv[1:])

  # Assigns this enviorement variable to set the name of all log messages
  os.putenv('PRGNAME', 'tLtask-segment')

  #write a sample config file
  if args.write_example:
    config.write(stdout)
    exit(0)

  #show version (according to translectures-revision SVN)
  if args.version:
    show_version("tLtask-segment")
    exit(0)

  #Modify scripts PATH if necessary
  if args.scripts_path:
    os.putenv('PATH', os.environ['PATH']+':'+args.scripts_path)
  else: #add the scripts folder on this executable path
    dirname = os.path.dirname(os.path.realpath(argv[0]))
    os.environ['PATH'] = os.environ['PATH']+':'+dirname+'/scripts'

  if args.not_launch:
    stdout.write("export PATH=%s\n" %os.environ['PATH'])

  #read an external config file and check version number
  if not args.config_file:
    parser.error( _( """A config_file must be given. Use --write-example-config-file
to generate a template of the configuration file.""") )
  else:
    config.read(args.config_file)

    if not config.has_option('General','tLtask-segment-version') or \
        ( config.get('General','tLtask-segment-version') != '_RELEASE_' and \
        config.get('General','tLtask-segment-version') != __version__ ):
      fatal_err( _(
"Wrong version number in loaded configuration file. Some options may have\n"
"changed or not be supported anymore. Check your configuration, and\n"
"update the tLtask-segment-version to %d if everything is correct."
        ) %int(__version__) )

  #Overwrite variables with command line
  if args.overwrite != None:
    for o in args.overwrite:
      m = re.search("^(.*)_(.*)=(.*)$", o)
      if m == None:
        fatal_err( _( "Incorrect overwrite in element %s" ) %o )
      debug_msg("Setting [%s] %s = %s" %(m.group(1), m.group(2), m.group(3)))
      if config.has_option(m.group(1),m.group(2)):
        config.set(m.group(1), m.group(2), m.group(3))
      else:
        fatal_err( _( "Section and options of overwrite command %s do not exist in config" ) %o )

  #Filechecks
  if args.filecheck:
    if not os.path.exists(config.get('Lists','videos-list')):
      fatal_err( _( "Input video list: %s not found\n" ) %config.get('Lists','videos-list'))

    if not os.path.exists(config.get('Segmenter','acoustic-model')):
      fatal_err( _( "Acoustic model file: %s not found\n" ) %config.get('Segmenter','acoustic-model'))

    if not os.path.exists(config.get('Segmenter','language-model')):
      fatal_err( _( "Language model file: %s not found\n" ) %config.get('Segmenter','language-model'))

  #Prepare command to be submitted to tLtask-recognise.sh
  segmenter = tLsegmenter(config, verbose=args.v, filecheck=args.filecheck,
      not_launch=args.not_launch, log_folder=args.log_folder)

  # Add an external dependency to the first submit
  if args.depend_on:
    segmenter.last_submitted = args.depend_on

  ### Prepare for submitting the recognition
  ### Store configuration employed in a file
  segmenter.init_launched_info()
  segmenter.print_config_on_launched_info()

  debug_msg("Submitting the segmentation")


  segmenter.info_msg( _( "Extracting audio from videos" ) )
  segmenter.extract_audio(
    "segmentation"
    )

  if args.segmentation_trs == '':

    segmenter.info_msg( _( "Segmenting audio into speech and non-speech segments" ) )
    segmenter.segment_audio(
      "segmentation"
      )

  segmenter.info_msg( _( "Generating samples in TLK format" ) )
  segmenter.generate_speech_samples(
    "segmentation", segm_trs=args.segmentation_trs
    )

  if args.print_output_info:
    stdout.write("__OUTPUT_DIR: %s\n" % os.path.abspath(segmenter.conf.get('General','output-dir')))
    if segmenter.last_submitted != None:
      stdout.write("__LAST_SUBMITTED: %s\n" % segmenter.last_submitted)
    else:
       stdout.write("\n")


  #####################   Common steps   ############################
  if args.delete_tmp:
    #8 delete tmp-folder
    segmenter.delete_tmp()

####################################################################################################
if __name__ == "__main__":
  main()

