# -*- coding: utf-8 -*-

# Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


""" Performs CART using TLK.py """

__all__= ['cart']
__copyright__= """Adrià Giménez Pastor, 2012"""


from array import array
from TLK import *
from math import exp,log,pi
import re
import sys




######################
# CONSTANTS PRIVADES #
######################

_INF= float('-inf')
LOG2PI= log(2*pi)




####################
# CLASSES PRIVADES #
####################

class Log:
    
    def __init__(self,level,fn=None):
        self.f= file(fn,'w') if fn!=None else sys.stderr
        self.level= level
        
    def w(self,level,msg):
        if level <= self.level:
            self.f.write(msg+'\n')


# Conté per a un trifonema la informació necessària per a calcular el
# CART.
class Tri:
    
    # Es supossa que la correspondència entre els arguments és
    # correcta.
    def __init__(self,name,occs,amodel):
        self.name= name
        char= amodel.hmms[amodel.ilabels[name]]
        self.Q= char.Q
        self.occs= occs[name]
        self.x= []; self.x2= []
        for q in xrange(0,self.Q):
            occ= self.occs[q]
            mu= char.states[q].mu
            tmp= array('d',[x*occ for x in mu])
            self.x.append(tmp)
            var= char.states[q].var
            tmp= array('d',[0]*amodel.D)
            for d in xrange(0,amodel.D):
                tmp[d]= (var[d]+mu[d]*mu[d])*occ
            self.x2.append(tmp)

    def __repr__(self):
        return '<NAME:%s;Q:%d;OCCS:%s;X:%s;X2:%s>'%(self.name,self.Q,
                                                    repr(self.occs),
                                                    repr(self.x),repr(self.x2))

# Representa un conjunt de trifonemes, per a un estat donat, amb la
# seva loglikelihood (score). També es guarda la mitja, la variança i
# l'ocupació. Si la loglikelihood és -inf (occ=0) la mitja i la
# variança no estan definides. També està l'estat.
class Cluster:
    
    @staticmethod
    def __get_stats(tris,q):
        if tris == []: return None,0,None,None
        D= len(tris[0].x[0])
        occ= 0
        x= array('d',[0]*D)
        x2= array('d',[0]*D)
        for tri in tris:
            occ+= tri.occs[q]
            tri_x= tri.x[q]
            tri_x2= tri.x2[q]
            for d in xrange(0,D):
                x[d]+= tri_x[d]
                x2[d]+= tri_x2[d]
        return D,occ,x,x2
    
    # Es supossa que tots els trifonemes tenen l'estat indicat. El
    # primer estat és el 0.
    def __init__(self,tris,q):
        self.q= q
        self.tris= list(tris)
        D,occ,x,x2= Cluster.__get_stats ( tris, q )
        self.occ= occ
        if occ == 0 :
            self.ll= _INF
            return
        for d in xrange(0,D): x[d]/= occ
        self.mu= x
        for d in xrange(0,D):
            x2[d]= x2[d]/occ - x[d]*x[d]
        self.var= x2
        aux= 0
        for d in xrange(0,D): aux+= log(self.var[d])
        aux+= D*LOG2PI + D
        self.ll= -0.5*occ*aux
        
    def __repr__(self):
        tmp= [t.name for t in self.tris]
        return '<Q:%d;LL:%g;OCC:%g;MU:%s;VAR:%s;TRIS:%s>'%(self.q,
                                                           self.ll,self.occ,
                                                           repr(self.mu),
                                                           repr(self.var),
                                                           repr(tmp))


# Representa una pregunta.
class Query:
    
    def __init__(self,name,expr):
        self.name= name
        self.expr= expr
        self.__pattern= re.compile(expr)
        self.__cache= {}
        
    def __repr__(self):
        return '<%s:/%s/>'%(self.name,self.expr)
    
    def eval(self,name):
        ret= self.__cache.get(name)
        if ret != None: return ret
        self.__cache[name]= ret= (re.search(self.__pattern,name)!=None)
        return ret
    
    # Torna els trifonemes que cumplixen amb la pregunta.
    def subset(self,tris):
        ret= []
        for t in tris:
            if self.eval ( t.name ) : ret.append ( t )
        return ret
    
    # Dividix el conjunt de trifonemes en dos subconjunts (yes,no).
    def split(self,tris):
        yes= []; no= []
        for t in tris:
            if self.eval ( t.name ) : yes.append ( t )
            else                    : no.append ( t )
        return yes,no


# L'arbre de CART.
class Tree:
    
    class Node:
        # name: Nom
        # bname: Nom base
        # cluster: Cluster associat
        # queries: Preguntes pendents.
        # query: pregunta utilitzada per a baixar. None indica fulla.
        # q: Estat 
        # yes: Node si ix sí.
        # no: Node si ix no.
        def __init__(self,name,cluster,queries,bname=None):
            self.name= name
            self.bname= name if bname==None else bname
            self.cluster= cluster
            self.queries= list(queries)
            
    def __init__(self,amodel,counts_fn,queries,min_occ,min_inc,log):
        self.log= log
        self.tris= get_tris ( amodel, counts_fn, log )
        log.w(1,"Initializing CART")
        self.min_occ= min_occ
        log.w(1,"  - Minimum occupation is %g"%self.min_occ)
        self.min_inc= min_inc
        log.w(1,"  - Minimum increment is %g"%self.min_inc)
        self.queries= [Query(i[0],i[1]) for i in queries]
        if log.level >= 4:
            log.f.write('  - Queries:\n')
            for q in self.queries:
                log.f.write('    - %s : /%s/\n'%(q.name,q.expr))
        self.init_nodes= {}
        self.leaves= []
        
    def __release_leave(self,node):
        self.__tmp_leaves.append(node)
        self.__counter+= 1
        node.counter= self.__counter
        node.query= node.queries= None
        self.log.w(3,'  CONSOLIDATED "%s" [%s_%d] => OCC:%g'%(node.name,
                                                              node.bname,
                                                              node.counter,
                                                              node.cluster.occ))
        
    def __split_node(self,node,stack):
        if self.log.level >= 3:
            self.log.f.write('\n  SPLIT "%s"\n'%node.name)
            if self.log.level >= 4:
                self.log.f.write('  %d symbols:'%len(node.cluster.tris))
            else:
                self.log.f.write('  %d symbols\n'%len(node.cluster.tris))
            if self.log.level >= 4:
                for t in node.cluster.tris:
                    self.log.f.write(' %s'%t.name)
                self.log.f.write('\n')
        if node.queries==[]:
            self.__release_leave ( node )
            return
        maxq= None
        for i in xrange(0,len(node.queries)):
            q= node.queries[i]
            self.log.w(3,'      %s : /%s/ ?'%(q.name,q.expr))
            yes,no= q.split(node.cluster.tris)
            yes= Cluster(yes,node.cluster.q)
            if self.log.level >= 3:
                self.log.f.write('        Yes => OCC: %g N:%d\n'%(yes.occ,
                                                                 len(yes.tris)))
            if yes.occ < self.min_occ: continue
            no= Cluster(no,node.cluster.q)
            if self.log.level >= 3:
                self.log.f.write('        No => OCC: %g N:%d\n'%(no.occ,
                                                                 len(no.tris)))
            if no.occ < self.min_occ: continue
            inc= (yes.ll+no.ll) - node.cluster.ll
            if self.log.level >= 3:
                self.log.f.write(('        (%g'
                                  +' + %g) - %g = %g\n')%(yes.ll,no.ll,
                                                          node.cluster.ll,inc))
            if inc < self.min_inc: continue
            if maxq == None or inc>maxinc:
                maxq,maxy,maxn,maxinc,maxi= q,yes,no,inc,i
        if maxq == None:
            self.__release_leave ( node )
            return
        self.log.w(3,'      MAX => %s : %s   INC: %g'%(maxq.name,
                                                       maxq.expr,maxinc))
        node.query= maxq
        node.cluster= None
        queries= list(node.queries)
        del queries[maxi]
        node.yes= Tree.Node('%s:%s'%(node.name,maxq.name),
                            maxy,queries,node.bname)
        stack.append(node.yes)
        node.no= Tree.Node('%s:!%s'%(node.name,maxq.name),
                           maxn,queries,node.bname)
        stack.append(node.no)
        
    def add_init_cluster(self,name,query,q):
        # Crea el node inicial
        self.__tmp_leaves= []
        self.__counter= 0
        query= Query(name,query)
        cluster= Cluster(query.subset(self.tris),q)
        if self.log.level >= 2: N= len(cluster.tris)
        node= Tree.Node('s%d:%s'%(q+1,name),cluster,self.queries)
        aux= self.init_nodes.get(q)
        if aux == None:
            aux= self.init_nodes[q]= []
        aux.append ( (query,node) )
        # Crea l'arbre
        stack= [node]
        while stack != []:
            node= stack.pop()
            self.__split_node ( node, stack )
        # Acumula fulles
        if self.log.level >= 2:
            N2= len(self.__tmp_leaves)
            self.log.f.write('  Reduction %d -> %d\n'%(N,N2))
        self.leaves+= self.__tmp_leaves
        
    # Torna la fulla associada a un nom i un estat.
    def get_leaf(self,name,q):
        iqueries= self.init_nodes.get(q)
        if iqueries==None:
            raise Exception(("Unable to find a CART leaf related to"+
                             " a state in position %d")%(q+1))
        init_node= None
        for iquery,node in iqueries:
            if iquery.eval(name):
                init_node= node; break
        if init_node==None:
            raise Exception(("Unable to find a CART leaf related to the"+
                             " symbol %s for the %dth state")%(name,q+1))
        node= init_node
        while node.query!=None:
            node= node.yes if node.query.eval(name) else node.no
        return node




#####################
# FUNCIONS PRIVADES #
#####################

# Llig les ocupacions
def get_occs(amodel,counts_fn):
    f= file(counts_fn,'r')
    l= f.readline()
    counter= 0
    occs= {}
    while l != "":
        line= l.split()
        l= f.readline()
        counter+= 1
        lline= len(line)
        if lline==0 or line[0]== '#' : continue
        if lline<2 :
            raise Exception('%s: line %d only has 1 field'%(counts_fn,
                                                            counter))
        sym= line[0]
        Q= int(line[1])
        n= amodel.ilabels.get(sym)
        if n == None:
            raise Exception("%s: symbol '%s' is out of model"%(counts_fn,
                                                                   sym))
        char= amodel.hmms[n]
        if char.Q != Q:
            raise Exception(("%s: symbol '%s' has %d states"+
                             " instead of %d")%(counts_fn,
                                                sym,
                                                char.Q,Q))
        if lline != Q+2:
            raise Exception(("%s: %d fields were expected"+
                             " in line %d")%(counts_fn,Q+2,counter))
        tmp= occs.get(sym)
        if tmp != None:
            raise Exception(("%s: symbol '%s' is repeated"+
                             " in line %d")%(counts_fn,sym,counter))
        tmp= array('d',[float(x) for x in line[2:]])
        occs[sym]= tmp
    f.close()
    if len(occs) != amodel.N:
        raise Exception(("%s: there are no occupations"+
                         " for all models")%counts_fn)
    return occs


# Obté el conjunt de tots els trifonemes a partir del model i del
# fitxer d'ocupacions.
def get_tris(amodel,counts_fn,log):
    log.w(1,"Reading state occupations")
    occs= get_occs ( amodel, counts_fn )
    tris= []
    for name in amodel.ilabels:
        tris.append ( Tri(name,occs,amodel) )
    if log.level >= 4:
        log.f.write('State occupations:\n')
        for t in tris:
            log.f.write('%s\t\t%d'%(t.name,t.Q))
            for q in xrange(0,t.Q):
                log.f.write('\t\t%g'%t.occs[q])
            log.f.write('\n')
            if log.level >= 5:
                for q in xrange(0,t.Q):
                    log.f.write('    X[%d]: %s\n'%(q,repr(list(t.x[q]))))
                    log.f.write('    X2[%d]: %s\n'%(q,repr(list(t.x2[q]))))
    return tris


def mget_center(name):
    aux=name.split('-')
    aux= aux[1] if len(aux)>1 else aux[0]
    return aux.split('+')[0]


def calc_fulllist(freephos,get_center,amodel):
    # Get phonem list
    phos= set([])
    for n in xrange(0,amodel.N):
        pho= get_center(amodel.labels[n])
        if not pho in freephos:
            phos.add(pho)
    # Generate fulllist
    ret= list(freephos)
    for p in phos:
        ret.append(p)
        for p2 in phos:
            ret.append('%s-%s'%(p2,p))
            ret.append('%s+%s'%(p,p2))
            for p3 in phos:
                ret.append('%s-%s+%s'%(p2,p,p3))
    return ret


# Crea una llista amb els nous hmms i les transicions
# disponibles. Torna trans,hmms. Trans és un diccionari de
# centers->hmms(amb transicions). hmms és una llista de tuples
# (NAME,(CENTER,NODE1,NODE2...))
def get_hmms(amodel,tree,fulllist,get_center,tiedlist_fn,freephos):
    # Fixa variables
    if fulllist==None: fulllist= calc_fulllist(freephos,get_center,amodel)
    # Crea llista de transicions disponibles
    trans= {}
    for n in xrange(0,amodel.N):
        center= get_center(amodel.labels[n])
        if trans.get(center) != None: continue
        char= amodel.hmms[n]
        if char.trans.type==Trans.POINTER: char= amodel.hmms[char.trans.ptr]
        trans[center]= char
    # Crea hmms i tiedlist
    hmms= []; itiedlist= {}; tiedlist= {}
    for s in fulllist:
        center= get_center(s)
        char= trans.get(center)
        if char == None:
            raise Exception(("There are not transitions"+
                             " for symbol '%s'")%s)
        key= [center]
        for q in xrange(0,char.Q):
            key.append(tree.get_leaf(s,q))
        key= tuple(key)
        aux= itiedlist.get(key)
        if aux == None:
            hmms.append((s,key))
            tiedlist[s]= s
            itiedlist[key]= s
        else: tiedlist[s]= aux
    # Escriu tiedlist
    f= file(tiedlist_fn,'w')
    for k,v in tiedlist.iteritems():
        if k==v : f.write('%s\n'%k)
        else    : f.write('%s %s\n'%(k,v))
    f.close()
    return trans,hmms


def get_node_name(node,use_long_names):
    if use_long_names: return '%s'%node.name
    else             : return '%s_%d'%(node.bname,node.counter)


# Escriu el model de caràcters
def write_amodel(amodel,tree,out_fn,hmms,trans,use_long_names):
    f= sys.stdout if out_fn==None else file(out_fn,'w')
    # Capçalera.
    f.write('AMODEL\n')
    f.write('TiedStates\n')
    f.write('DGaussian\n')
    f.write('D %d\n'%amodel.D)
    f.write('SMOOTH')
    for v in amodel.smooth: f.write(' %g'%v)
    f.write('\n')
    # Estats (Capçalera)
    f.write('N %d\n'%len(tree.leaves))
    f.write('States\n')
    for node in tree.leaves:
        f.write('%s\n'%get_node_name(node,use_long_names))
        f.write('MU')
        for v in node.cluster.mu: f.write(' %g'%v)
        f.write('\n')
        f.write('VAR')
        for v in node.cluster.var: f.write(' %g'%v)
        f.write('\n')
    # HMMs.
    actives= {}
    f.write('N %d\n'%len(hmms))
    for hmm in hmms:
        name= hmm[0]
        f.write("'%s'\n"%name)
        key= hmm[1]
        center= key[0]
        char= trans[center]
        f.write('Q %d\n'%char.Q)
        aux= actives.get(center)
        if aux == None:
            actives[center]= name
            char.trans.write(f)
        else: f.write('TransP %s\n'%aux)
        f.write('%s'%get_node_name(key[1],use_long_names))
        for node in key[2:]:
            f.write(' %s'%get_node_name(node,use_long_names))
        f.write('\n')
    if f!=sys.stdout: f.close()


def make_init_clusters(amodel,get_center,freephos):
    ret= []
    aux= set([])
    for n in xrange(0,amodel.N):
        pho= get_center(amodel.labels[n])
        if pho in aux: continue
        if pho in freephos: pattern= '(^%s$)'%pho
        else:               pattern= '(^%s$)|(-%s[+])|(-%s$)|(^%s[+])'%(pho,pho,
                                                                        pho,pho)
        for q in xrange(0,amodel.hmms[n].Q):
            ret.append ( (pho,pattern,q) )
        aux.add(pho)
    return ret


################
# PART PÚBLICA #
################

def cart(**args):
    """ CART
        
        OBLIGATORIS:
        amodel_fn: Nom del fitxer amb el model de caràcters.
        occs_fn: Nom del fitxer amb les ocupacions del estats.
        queries: Una llista amb les preguntes. Cada pregunta és una
                 tupla de la forma: (NAME,EXPR)
                   - NAME: el nom de la pregunta.
                   - EXPR: un 'string' amb l'expressió regular de la
                     pregunta.
        tiedlist_fn: Fitxer on s'escriu la correspondència entre els
                     símbols per als que es volien generar HMMs i els
                     símbols que realment s'han generat.
        freephos: A set with the free context phonemes.

        OPCIONALS:
        amodel_out_fn: nom del fitxer on guardar el model de caràcters
                       convertit. Per defecte l'eixida estàndard.
        fulllist: Llista amb tots els símbols que es volen modelar. Si
                  dos símbols generen exactament el mateix HMM
                  aleshores sols el primer s'escriu en el model de
                  caràcters. Per defecte totes les combinacions que es
                  poden fer L-C+R dels caràcters del model. On C
                  s'obté amb get_center.
        get_center: Quan es genere un símbol nou A el primer que es
                    farà és buscar una matriu de transicions del model
                    antic per a eixe símbol. Concretament es
                    seleccionarà la matriu de transicions del primer
                    símbol B que complisca
                    get_center(A)==get_center(B). On 'get_center' és
                    una matriu que rep el nom d'un símbol i torna un
                    altre nom. Per defecte, suposant un format de
                    símbol L-C+R torna C.
        log_fn: Fitxer on escriure el log. Per defecte l'eixida d'error.
        min_occ: L'ocupació mínima que ha de tindre un cluster. Per
                 defecte 200.
        min_inc: Increment mínim de la logversemblança per a poder
                 dividir un cluster. Per defecte 350.
        use_long_names: Un valor distint a 0 indica que per als estats
                        utilitza noms que descriuen les decisions
                        preses. Per defecte utilitza números.
        verbose: Nivell de verbositat. Per defecte 0.
        init_cluster: Una llista amb els clusters inicials. Els
                      clusters no han de solapar-se i han de cobrir
                      tots els trifonemes i estats. En cas contrari es
                      poden produir efectes no desitjats. Cada cluster
                      és una tupla de la forma (NAME,EXPR,Q)
                        - NAME: el nom del cluster.
                        - EXPR: un 'string' amb una expressió regular.
                        - Q: l'estat per al que es va a fer el
                          cluster. Comença a comptar per 0.
                      Per defecte crea un per cada símbol central i
                      estat assumint L-C+R. El central s'obté amb
                      get_center.
    """
    
    # Inicialitza Log
    log_fn= args.get('log_fn')
    verbose= args.get('verbose')
    log= Log(0 if verbose==None else int(verbose),log_fn)
    log.w(1,"Reading acoustic model")
    
    # Carrega el model
    amodel= AModel(file(args['amodel_fn'],'r'))
    if amodel.tname != "DGaussian" :
        raise Exception("acoustic model must be of type DGaussians")
    if log.level >= 5 :
        log.f.write('Acoustic model:\n')
        amodel.write(log.f)
    
    # Inicialitza l'arbre de derivació.
    get_center= args.get('get_center')
    if get_center==None: get_center= mget_center
    init_clusters= args.get('init_clusters')
    if init_clusters==None:
        init_clusters= make_init_clusters(amodel,get_center,args['freephos'])
    min_occ= args.get('min_occ')
    min_occ= 200 if min_occ==None else float(min_occ)
    min_inc= args.get('min_inc')
    min_inc= 350 if min_inc==None else float(min_inc)
    tree= Tree(amodel,args['occs_fn'],args['queries'],min_occ,min_inc,log)
    log.w(1,"Making CART")
    for c in init_clusters:
        log.w(2,"  Making subtree (%s : /%s/,%d)"%(c[0],c[1],c[2]+1))
        tree.add_init_cluster ( c[0], c[1], c[2] ) 
    log.w(2,"S'han creat %d estats"%len(tree.leaves))
    
    # Crea el model.
    log.w(1,"Making acoustic model")
    fulllist= args.get('fulllist')
    trans,hmms= get_hmms ( amodel, tree, args.get('fulllist'),
                           get_center, args['tiedlist_fn'],
                           args['freephos'] )
    log.w(1,"Writing acoustic model")
    use_long_names= args.get('use_long_names')
    use_long_names= 0 if use_long_names==None else int(use_long_names)
    write_amodel ( amodel, tree, args.get('amodel_out_fn'),
                   hmms,trans,use_long_names)
