#!/bin/bash

# Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

PROGRAM_NAME=`basename "$0"`
# helper functions
function error {
    echo -e "[EE] $PROGRAM_NAME: $@" >&2
    exit 1
}

if [ $# -lt 5 ]; then
  cat <<EOF
$PROGRAM_NAME <list-of-phonemes> <tlist> <list-of-tied-phos> <tmpdir> [options]

ARGUMENTS
  <list-of-phonemes>       List of files containing the phoneme annotation
  <tlist>                  List of how to build phos->tiedphos
  <list-of-tied-phos>      List of files containing the tiedphoneme annotation
  <tmpdir>                 Temporal directory of the estimation
  <freephos>               list of context free phoenems separated by comma

OPTIONS
  <1|0>       (def. 0): Do or do not generate crossword triphonemes. 
  If 1:
  <Real>      (def. 0): Generate crossword triphonemes only if the duration of the shortpause between words is longer than <Real> frames (def. 0.0)
                      #  Note: if <Real>=0 the crossword will always be generate when possible.
                      #  Otherwise, an additional translitertion files are needed to tell which is the frame duration of the short pauses.
  <extension> (def. '.SPlength')  : Additional extension to the .pho supplied files.
                      #  the  .pho.SPlength files should contain the same exact phonetic transliteration as in their .pho counterpart,
                      #  except for the 'SP' phonemes, which has been changed by 'SP=duration' , with duration a real number.
  
EOF
    error "SYNTAX ERROR. Expected at least 5 arguments" 
fi

# ARGUMENTS
PHOS=$1
TLIST=$2
TIEDPHOS=$3
TMPDIR=$4
FREEPHOS=${5//[[:space:]]/}

# OPTIONS
CROSSWORDS=${6:-"no"}
SPthreshold=${7:-0}
PHOSPlength_ext='.SPlength'


# CHECK ARGUMENTS :
[ -s "$PHOS" ]  || error "1st argument: '$PHOS' file is emtpy or does not exist"
nPHOS=`wc -l < "$PHOS"`
[ -s "$TLIST" ]  || error "2nd argument: '$TLIST' file is emtpy or does not exist"

set -eu
# If Crossword depends on the durarion of the short pause (ie. SPthreshold>0,)
#   Use the next code, which expects for addional transliteration files with extension : .pho.SPlength
#   NOTE: the  .pho.SPlength files should contain the same exact phonetic transliteration as in their .pho counterpart,
#         except for the 'SP' phonemes, which has been changed by 'SP=duration' , with duration a real number
if [[ "$CROSSWORDS" == "yes" ]]
then
    while read l ;   do
        [ -f "$l"  ] || error "Expected non-emtpy phonetic transliteration for file '$l'"
        j=`dirname "$l"`
        k=${l##*/};
        DIR=$TMPDIR/phos_tied/$j
        mkdir -p "$DIR" || error "Cannot create/access temporary path '$DIR'"
        echo "$l$PHOSPlength_ext" $DIR/$k.txt
    done < "$PHOS"  | gawk -v SPthreshold="$SPthreshold" -v TLIST=$TLIST -v FREEPHOS="$FREEPHOS" '
    BEGIN {
        N= split(FREEPHOS,aux,",")
        for ( n= 1; n <= N; ++n ) freephos[aux[n]]= 1
        while ( getline < TLIST )
        if ( NF == 1 ) tlist[$1]=$1
        else           tlist[$1]=$2
        for ( p in freephos ) tlist[p]= p
    }

{
  input= $1
  output= $2
  N= 0
  while ( getline < input )
    for ( i= 1; i <= NF; ++i )
    {
      N+= 1
      phos[N]= $i
      if ( $i in freephos ) lengths[N]= SPthreshold+1
      else if ( $i ~ /.=./ )
      {
        split($i,aux,"=")
        if ( aux[1] in freephos )
        {
           phos[N]= aux[1]
           lengths[N]= aux[2]
        }
        else lengths[N]= -1
      }
      else lengths[N]= -1
    }
  close(input)
  ret= ""
  for ( i= 1; i <= N; ++i )
  {
    if ( phos[i] in freephos ) tri= phos[i]
    else
    {
      if ( i > 1 && !(phos[i-1] in freephos) )
         tri= phos[i-1]"-"
      else if ( (i>2) && lengths[i-1]<=SPthreshold && (phos[i-1] in freephos) )
         tri= phos[i-2]"-"
      else
         tri= ""
      tri= tri $i
      if ( i < N && !(phos[i+1] in freephos) )
         tri= tri"+"phos[i+1]
      else if ( (i<(N-1)) && lengths[i+1]<=SPthreshold \
                 && (phos[i+1] in freephos) )
         tri= tri"+"phos[i+2]
    }
    ret= ret" "tlist[tri]
  }
  if (ret == "")  {
    print "[WW] Void transcription, a single SP will be the output transcription!" >> /dev/stderr;
    print "SP" > output
  } else print ret > output
  close ( output )
}'

else # Word internal
    while read l ;   do
        if [ ! -s $l  ] ; then
          echo "[WW] Void transcription $l" >&2
          continue
        fi
        j=`dirname $l`
        k=${l##*/};
        DIR=$TMPDIR/phos_tied/$j
        mkdir -p "$DIR" 
        [ -d "$DIR" ]  || error "Cannot create/access temporary path '$DIR'"
        echo $l $DIR/$k.txt
    done < $PHOS  | gawk -v crosswords="$CROSSWORDS" -v TLIST=$TLIST -v FREEPHOS="$FREEPHOS" '
    BEGIN {
        N= split(FREEPHOS,aux,",")
        for ( n= 1; n <= N; ++n ) freephos[aux[n]]= 1
        while ( getline < TLIST )
        if ( NF == 1 ) tlist[$1]=$1
        else           tlist[$1]=$2
        for ( p in freephos ) tlist[p]= p
    }

    {
        input= $1
        output= $2
        ret= ""
        while ( getline < input )
        for ( i= 1; i <= NF; ++i )
        {
            if ( $i in freephos ) tri= $i
            else
            {
                if ( i > 1 && !($(i-1) in freephos) )                        tri= $(i-1)"-"
                else if ( (crosswords==1) && (i>2) && ($(i-1) in freephos) ) tri=$(i-2)"-"
                else                                                         tri= ""
                tri= tri $i
                if ( i < NF && !($(i+1) in freephos) )                            tri= tri"+"$(i+1)
                else if ( (crosswords==1) && (i<(NF-1)) && ($(i+1) in freephos) ) tri= tri"+"$(i+2)
            }
            ret= ret" "tlist[tri]
        }
        if (ret == "")  {
            print "[WW] Void transcription, a single SP will be the output transcription!" >> /dev/stderr;
            print "SP" > output
        } else print ret > output
        close ( output )
        close ( input )
    }'

fi

while read l ;   do
    j=`dirname $l`
    k=${l##*/};
    DIR=$TMPDIR/phos_tied/$j
    echo $DIR/$k.txt
done < "$PHOS" > "$TIEDPHOS"

[ -s "$TIEDPHOS" ]  || error "Output list '$TIEDPHOS' is emtpy or does not exist"
nTIEDPHOS=`wc -l < "$TIEDPHOS"`
[ "$nPHOS" -eq "$nTIEDPHOS" ] || error "Number of samples MISMATCH: '$PHOS' list constains $nPHOS samples while output list '$TIEDPHOS' contains $nTIEDPHOS samples"

echo "Done!"
