#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tLtask-train: Executes the training using transLectures tools
# Copyright 2012,2013,2014 transLectures-UPV Team
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#include i18n
import gettext
_ = gettext.gettext

import argparse
import ConfigParser
from sys import argv, exit, stdin, stdout, stderr
import os
import io
import subprocess
import shlex
import datetime
import locale
import re

__version__ = filter(str.isdigit, "$Revision: 9571 $")

def show_version(prog):
  stdout.write("%s version: %s\n" %(prog, __version__))

config_file_executables_defaults = """
[Executables]
bindir                = """+(os.environ['TLKPATH']+"/" if 'TLKPATH' in os.environ else "")+"""
tLtrain_ex            = %(bindir)stLtrain
tLupdate_ex           = %(bindir)stLupdate
tLcmllr_ex            = %(bindir)stLcmllr
tLcmllrfeas_ex        = %(bindir)stLcmllrfeas
tLmkproto_ex          = %(bindir)stLmkproto
tLtomix_ex            = %(bindir)stLtomix
tLmumix_ex            = %(bindir)stLmumix
tLbin2txt_ex          = %(bindir)stLbin2txt
submit                = submit.sh
get_cache             = get_cache.sh
"""

config_file_lists_defaults = """
[Lists]
set-name               = train
standard-samples       = %(set-name)s.lst
clusters               = %(set-name)s-clusters.lst
clusters-samples       = %(set-name)s-clusters-samples.lst
phonemes               = %(set-name)s-phos.lst
unique-monophonemes    = %(set-name)s-uniq-monophos.lst
"""

config_file_general_defaults = """
[General]
tar                   = no
tar-feas-split        = no
mutex-on-write        = no
train-tmp-dir         = tmp
models-dir            = models
store-counts          = no
prefix-name           = tlk
tLtask-train-version  = """+__version__+"""
get_cache-tmp         = /tmp
"""

config_file_hmm_defaults = """
[HMM]
feature-vector-size   = 39
states                = 3
variance-smooth       = 0.001
context-free-phonemes = SP
"""

config_file_training = """
[Training]
splits                            = 10
monophone-iters                   = 8
monophone-model-name              = monophone
triphoneme-iters                  = 4
triphoneme-model-name             = triphoneme
tiedphoneme-iters                 = 4
tiedphoneme-mixture-components    = 128
tiedphoneme-model-name            = tied
tiedphoneme-list                  = tlist
submit-general-ops                =
submit-iter-ops                   =
submit-up-ops                     =
"""

config_file_cart= """
[CART]
minimum-occupation   = 400
minimum-ll-increment = 750
queries = cart_queries_es.py
"""

config_file_incremental_training = """
[Incremental-Training]
#the program expects a ${NAME}.model and ${NAME}.counts
#apart from a ${NAME}.tlist for standard and cmllr
base-target-model-path                    = target
base-standard-tiedphoneme-model-path      = standard
base-cmllr-tiedphoneme-model-path         = cmllr
"""

config_file_defaults = \
[
  config_file_executables_defaults,
  config_file_lists_defaults,
  config_file_general_defaults,
  config_file_hmm_defaults,
  config_file_training,
  config_file_cart
]

def fatal_err(message):
  stderr.write("[EE] "+ message + "\n")
  exit(1)

def debug_msg(message):
  stderr.write("[II] "+ message + "\n")

#####################################################################
#####################################################################
##
##  Class to submit training according to configuration file
##
#####################################################################
#####################################################################
class tLtrainer:
  def __init__(self, conf, verbose=False, filecheck=True, not_launch=False,
      log_folder=None):
    self.conf = conf
    self.last_submitted = None
    self.v = verbose
    self.filecheck = filecheck
    self.not_launch = not_launch
    self.log_folder_submit_op = []
    if log_folder != None:
      self.log_folder_submit_op = ("-o %s" %log_folder).split()
    self.launched_info_file = None
    self.launched_info_filename = None

    #Initialise some parameters
    self.prefix = ""
    if self.conf.get('General','prefix-name') != "":
      self.prefix = self.conf.get('General','prefix-name') + "."
    if self.conf.get('Training','submit-general-ops') != "":
      self.submit_ops_default = self.conf.get('Training','submit-general-ops')
    else: self.submit_ops_default= None

####################################
## Print information 
####################################
  def info_msg(self, message):
    if self.not_launch:
      stdout.write("\n\n# [II] "+ message + "\n\n")
    else:
      stdout.write("[II] "+ message + "\n")

####################################
## Returns the context-free phonemes
####################################
  def get_freephos(self):
    ret= re.sub(r'\s','',self.conf.get('HMM','context-free-phonemes')).strip()
    if ret=='': ret= 'SP'
    return ret

#####################################################################
##  Returns the name of the models folder plus the model name prefix
#####################################################################
  def model_name_prefix(self,name):
    return self.conf.get('General','models-dir') + "/" + \
      self.prefix + name + "."

#####################################################################
##  Returns the name of the tmp folder plus the model name prefix
#####################################################################
  def tmp_name_prefix(self, name):
    return self.conf.get('General','train-tmp-dir') + "/" + \
      self.prefix + name + "."

#####################################################################
##  Wrap of checkpath in order to ignore it if necessary
#####################################################################
  def checkfile(self, path):
    if self.filecheck:
      return os.path.exists(path)
    else:
      return False

#####################################################################
##  submit a process given a name and a command line
#####################################################################
  def submit(self, submission_name, command_line,
    array=None, depend_on=None, submit_ops=None): #None = depend on the launched proc

    submission_name = self.prefix + submission_name
    #submit options
    cmd = [self.conf.get('Executables', 'submit'), "-n ", submission_name]
    depend = None
    if self.last_submitted:
      depend = self.last_submitted
    if depend_on:
      depend = depend_on
    if depend != None and depend != "":
      cmd += ['-w ', depend]
    if array:
      cmd += ['-j ', array]
    if submit_ops:
      cmd += submit_ops.split()
    elif self.submit_ops_default != None:
      cmd += self.submit_ops_default.split()
    if self.log_folder_submit_op != []:
      cmd += self.log_folder_submit_op

    #process
    cmd += command_line

    #Execute
    #Workaround subprocess " " arguments insertions
    string = " ".join(cmd)
    if self.v:
      debug_msg( _( "Submitting: %s\n" ) %string )

    #write to launched info file
    self.launched_info_file.write("%s\n" %string)

    #prepare cmd to execute the process
    cmd = shlex.split(string)
    if self.not_launch:
      stdout.write("%s\n" %string)
    else:
      try:
        ret = subprocess.call(cmd)
      except Exception as e:
        fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(string, e) ) )
      if ret != 0:
        fatal_err( _( "Failed when executing: '%s' with exit code: %d" ) %(string, ret))

    self.last_submitted = submission_name

  ###########################################################################
  #### Submit the tar of features and transcriptions files   ################
  ###########################################################################
  def submit_tar(self,
      name,
      samples_lst,
      word_lst,
      iter_opts #options to be passed to mhtrain_em_iter
      ):
    #tar features
    tar_file = "%s/%s.tar" %( self.conf.get('General', 'train-tmp-dir'),
        os.path.basename(samples_lst) )
    if self.conf.get('General','tar-feas-split') == 'yes':
      #split as many tar-feas as splits in training
      if not self.checkfile("%s_1" %tar_file):
        self.submit(name+".create-feature-tar-file",
          ["tar-list.sh", "-s %s" %self.conf.get('Training','splits'),
            tar_file, samples_lst ],
            submit_ops=self.conf.get('Training','submit-up-ops')
          )
      iter_opts.append("-t %s -f" %tar_file)
    else:
      #there is only one tar file
      iter_opts.append("-t %s" %tar_file)
      if not self.checkfile(tar_file):
        self.submit(name+".create-feature-tar-file",
          ["tar-list.sh", tar_file, samples_lst ],
            submit_ops=self.conf.get('Training','submit-up-ops')
          )

    #tar transcriptions
    tar_trans_file = "%s/%s.tar" %( self.conf.get('General', 'train-tmp-dir'),
        os.path.basename(word_lst) )
    iter_opts.append("-s %s" %tar_trans_file)
    if not self.checkfile(tar_trans_file):
      self.submit(name+".create-transcription-tar-file",
        ["tar-list.sh", tar_trans_file, word_lst ],
          submit_ops=self.conf.get('Training','submit-up-ops')
        )



  ###########################################################################
  #### Standard train of a HMM. Performs I iterations per split   ###########
  #### a new mixture is generated splitting each component in two ###########
  ###########################################################################
  def train_model(self,
      tmp_estimation_dir,   #directory to store intermediate calculations
      name,                 #training name
      model,                #input model
      word_lst,             #list of transcriptions
      samples_lst,          #list of samples
      splits,               #splits per iteration
      n_mixes,              #number of components of the final model
      iters,                #iterations per step
      store_occs = False
      ):

    dest_dir = tmp_estimation_dir+"/"+self.prefix + name

    if not self.checkfile(dest_dir+"/model_0"):
    #Submit initial preparation: copy model and split lists
      self.submit(name+".train-init",
          ["mhtrain_init.sh", dest_dir,
          model, word_lst, samples_lst, str(splits)])

    iter_opts = []
    up_opts = []
    finish_opts = []

    if self.conf.get('General','tar') == 'yes':
      iter_opts.append("-g '\"%s\"'" %self.conf.get('Executables','get_cache'))
      self.submit_tar(name, samples_lst, word_lst, iter_opts)

    if store_occs:
      iter_opts.append("-o")
      finish_opts.append("-o -s %d" %splits)

    if self.conf.get('General','store-counts') == 'yes':
      up_opts.append("-c")
      finish_opts.append("-c")

    if self.conf.get('General','mutex-on-write') == "yes":
      iter_opts.append("-m")

    aux= self.conf.get('General','get_cache-tmp')
    if aux != "": iter_opts.append('-T %s'%aux)

    #Launch for all mixtures
    mix = 1
    model_name_last = ""
    while mix <= n_mixes:
      model_name = "%s/%s_I%.2d" \
        %(self.conf.get('General', 'models-dir'), self.prefix + name, mix)

      #launch from the last generated model
      if not self.checkfile(model_name+".model"):
        if mix != 1:
          #Split last model
          self.submit(name+(".train-em-MIX%.2d-init" %(mix)),
            ["mhtrain_em_init.sh",
            "-e", self.conf.get('Executables','tLmumix_ex'),
            dest_dir, model_name_last])

        for i in xrange(1, iters+1):
          #Launch iteration
          self.submit(name+(".train-em-MIX%.2d-ITER%.d-estimate" %(mix, i)),
            ["mhtrain_em_iter.sh"] +
            iter_opts +
            ["-e", self.conf.get('Executables','tLtrain_ex'),
            dest_dir, str(i)],
            array=("1-%d" %splits),
            submit_ops=self.conf.get('Training','submit-iter-ops'))

          #Join counts
          self.submit(name+(".train-em-MIX%.2d-ITER%.d-update" %(mix, i)),
            ["mhtrain_em_up.sh"] +
            up_opts +
            (['-m'] if i == iters and n_mixes != 1 else []) + #mixcount only on the last
            ["-e", self.conf.get('Executables','tLupdate_ex'),
            dest_dir, str(splits), str(i)],
            submit_ops=self.conf.get('Training','submit-up-ops')
            )

        #Store mixture model
        self.submit(name+(".train-em-MIX%.2d-finish" %(mix)),
          ["mhtrain_em_finish.sh"] +
          finish_opts +
          (['-m'] if n_mixes != 1 else []) + #mixcount only on the last
          [dest_dir, str(iters), model_name])

      model_name_last = model_name
      newmix=mix*2
      if ( (mix < n_mixes) and (newmix>n_mixes)  ) :
          newmix=n_mixes
      mix=newmix

#####################################################################
##  submits a monophone training using the standard samples
#####################################################################
  def train_monophonemes(self, name, transcriptions_list, samples_list):
    #Submit creation of prototype model
    if not self.checkfile(self.conf.get('General', 'train-tmp-dir')+"/model_0"):
      self.submit(name+".make-proto", ['create-initial-model.py',
          '--tLmkproto', self.conf.get('Executables','tLmkproto_ex'),
          '--add-sp',
          self.conf.get('General', 'train-tmp-dir'),
          self.conf.get('HMM', 'feature-vector-size'),
          self.conf.get('HMM', 'states'),
          self.conf.get('Lists', 'unique-monophonemes'),
          samples_list])

    monophoneme_model = self.model_name_prefix(name) + \
                        self.conf.get('Training','monophone-model-name')
    #Submit the training of the monophone model
    if not self.checkfile(monophoneme_model+"_I01.model"):
      self.train_model(self.conf.get('General', 'train-tmp-dir'),
          name+".monophone",
          self.conf.get('General', 'train-tmp-dir')+"/model_0",
          transcriptions_list,
          samples_list,
          int(self.conf.get('Training', 'splits')),
          1,
          int(self.conf.get('Training', 'monophone-iters')))

#####################################################################
##  Generates for each transcription phoneme a triphone transcription
##  and creates a list of them
#####################################################################
  def train_triphonemes(self, name, transcriptions_list, samples_list):
    #generate the triphones annotation
    triphos_lst = self.tmp_name_prefix(name)+"triphos.lst"
    if not self.checkfile(triphos_lst):
      self.submit(name+".generate-triphonemes",
          ['generate-triphos-from-phos.sh',
           transcriptions_list,
           triphos_lst,
           self.tmp_name_prefix(name)+"triphos",
           self.get_freephos(),'no','0'
           #,self.conf.get('HMM', 'crosswords'),
           #self.conf.get('HMM', 'crosswords-threshold')
           ])

    #Generate an initial triphone model from monophone
    monophone_model = self.model_name_prefix(name) + \
                      self.conf.get('Training','monophone-model-name')
    triphoneme_model_tmp = self.tmp_name_prefix(name) + \
                           self.conf.get('Training','triphoneme-model-name')
    if not self.checkfile(triphoneme_model_tmp+"_I01.model"):
      self.submit(name+".convert-monophoneme-to-triphoneme",
          ['generate-triphoneme-model-from-mononophone.sh',
          "-e", self.conf.get('Executables','tLbin2txt_ex'),
          monophone_model+"_I01.model",
          self.tmp_name_prefix(name)+"triphos/triphos.lst",
          triphoneme_model_tmp+"_I01.model"])

    #train the triphone model
    triphoneme_model = self.model_name_prefix(name) + \
                       self.conf.get('Training','triphoneme-model-name')
    if not self.checkfile(triphoneme_model+"_I01.model"):
      self.train_model(self.conf.get('General', 'train-tmp-dir'),
          name+".triphoneme",
          triphoneme_model_tmp+"_I01.model",
          self.tmp_name_prefix(name)+"triphos.lst",
          samples_list,
          int(self.conf.get('Training', 'splits')),
          1,
          int(self.conf.get('Training', 'triphoneme-iters')),
          store_occs=True
          )

#####################################################################
##  Train a tied phoneme model using CART from an estimated triphone model
#####################################################################
  def train_tiedphonemes(self, name, transcriptions_list, samples_list):
    #generate first tied phoneme model from triphoneme using CART
    triphoneme_model = self.model_name_prefix(name) + \
                       self.conf.get('Training','triphoneme-model-name')
    tiedphoneme_init_model = self.model_name_prefix(name) + \
                        self.conf.get('Training','tiedphoneme-model-name')+ \
                        "_init"
    tlist              = self.model_name_prefix(name) + \
                        self.conf.get('Training','tiedphoneme-list')
    if not self.checkfile(tiedphoneme_init_model+"_I01.model"):
      self.submit(name+".generate-tied-init-model",
          ['generate-tiedphoneme-model-from-CART.sh',
           '-e', self.conf.get('Executables','tLbin2txt_ex'),
           '-m', self.conf.get('Executables','tLtomix_ex'),
           '-O', self.conf.get('CART','minimum-occupation'),
           '-I', self.conf.get('CART','minimum-ll-increment'),
           triphoneme_model+"_I01.model",
           triphoneme_model+"_I01.occs",
           tiedphoneme_init_model+"_I01.model",
           tlist,
           self.get_freephos(),
           self.conf.get('CART','queries')
          ],
          submit_ops=self.conf.get('Training', 'submit-iter-ops')
          )

    #generate the triphones annotation
    tiedphos_lst = self.tmp_name_prefix(name)+"tiedphos.lst"
    if not self.checkfile(tiedphos_lst):
      self.submit(name+".generate-tiedphonemes",
          ['generate-tiedphos-from-phos.sh',
           transcriptions_list,
           tlist,
           tiedphos_lst,
           self.tmp_name_prefix(name)+"tiedphonemes",
           self.get_freephos(),'no','0'
           #self.conf.get('HMM', 'crosswords', 'no'),
           #self.conf.get('HMM', 'crosswords-threshold', '0')
           ])

    #train the tied model
    tiedphoneme_model = self.model_name_prefix(name) + \
                       self.conf.get('Training','triphoneme-model-name')
    mixes = int(self.conf.get('Training','tiedphoneme-mixture-components'))
    if not self.checkfile(tiedphoneme_model+("_I%.d.model" %mixes)):
      self.train_model(self.conf.get('General', 'train-tmp-dir'),
          name+".tiedphoneme",
          tiedphoneme_init_model+"_I01.model",
          tiedphos_lst,
          samples_list,
          int(self.conf.get('Training', 'splits')),
          mixes,
          int(self.conf.get('Training', 'tiedphoneme-iters'))
          )

#####################################################################
##  Train a tied phoneme model using CART from an estimated triphone model
#####################################################################
  def generate_CMLLR_samples(self,
      name,
      samples_lst,
      tiedphos_lst,
      clusters_samples_lst,
      clusters_lst,
      target_model,
      out_dir):

    #Generate the clusters lists
    cmllrfeas_lst = "%s/cmllrfeas.lst" %(out_dir)
    if not self.checkfile(cmllrfeas_lst):
      self.submit(name+".create-clusters",
          ['generate-clusters.sh',
          samples_lst,
          tiedphos_lst,
          clusters_samples_lst,
          clusters_lst,
          out_dir,
          cmllrfeas_lst
          ])

    cluster_name_dependency = self.last_submitted
    #Launch CMLLR estimation and sample generation only of not generated matrices

    clusters = []
    for i in open(clusters_lst, "r"):
      clusters.append(i[:-1])

    #put in array only those that have not been generated
    array = []; last = None
    for i in range(1,len(clusters)+1):
      matrix_file = "%s/matrices/%s.matrix" %(
          out_dir, clusters[i-1])
      if not self.checkfile(matrix_file):
        if last == None:
          last = i
      else:
        if last != None:
          array += ["%d-%d" %(last,i-1)]
          last = None
    if last != None:
      array += ["%d-%d" %(last,len(clusters))]

    for array_jobs in array:
      self.submit(name+".generate-cmllr-samples",
          ['array-cmllr-features.sh',
          '-e', self.conf.get('Executables','tLcmllr_ex'),
          '-g', self.conf.get('Executables','tLcmllrfeas_ex'),
          clusters_lst,
          target_model,
          out_dir
          ],
          array = array_jobs,
          depend_on=cluster_name_dependency)

    self.last_submitted = self.prefix + name + ".generate-cmllr-samples"

#####################################################################
##  Delete tmp
#####################################################################
  def delete_tmp(self):
    self.submit("delete-tmp",
        ['rm', '-rf', self.conf.get('General', 'train-tmp-dir')+"/*"])

###################################################################################
##  Incremental train a tied phoneme model using the counts of a previous ones
###################################################################################
  def train_incremental_tiedphonemes(self, name, transcriptions_list, samples_list,
      base_tiedphoneme_model_name):
    #generate the tiedphonemes annotation
    tiedphos_lst = self.tmp_name_prefix(name)+"tiedphos.lst"
    tlist = "%s.tlist" %(base_tiedphoneme_model_name)
    if not self.checkfile(tiedphos_lst):
      self.submit(name+".generate-tiedphonemes",
                  ['generate-tiedphos-from-phos.sh',
                   transcriptions_list,
                   tlist,
                   tiedphos_lst,
                   self.tmp_name_prefix(name)+"tiedphonemes",
                   self.get_freephos(),'no','0'
                   #self.conf.get('HMM', 'crosswords', 'no'),
                   #self.conf.get('HMM', 'crosswords-threshold', '0')
                   ])

    #train the tied model
    tiedphoneme_model = self.model_name_prefix(name) + \
                       self.conf.get('Training','triphoneme-model-name')
    mixes = int(self.conf.get('Training','tiedphoneme-mixture-components'))
    base_model = "%s.model" %(base_tiedphoneme_model_name)
    base_counts = "%s.counts" %(base_tiedphoneme_model_name)
    if not self.checkfile(tiedphoneme_model+("_I%.d.model" %mixes)):
      self.incremental_train_model(self.conf.get('General', 'train-tmp-dir'),
          name+".tiedphoneme",
          base_model,
          base_counts,
          tiedphos_lst,
          samples_list,
          int(self.conf.get('Training', 'splits')),
          int(self.conf.get('Training', 'tiedphoneme-iters'))
          )

  ###############################################################################
  #### Incremental train of a HMM. Performs I iterations to the given model #####
  ###############################################################################
  def incremental_train_model(self,
      tmp_estimation_dir,   #directory to store intermediate calculations
      name,                 #training name
      base_model,           #input model
      base_counts,          #input model counts
      word_lst,             #list of transcriptions
      samples_lst,          #list of samples
      splits,               #splits per iteration
      iters,                #iterations per step
      store_occs = False
      ):

    dest_dir = tmp_estimation_dir+"/"+self.prefix + name
    nmixes = int(self.conf.get('Training','tiedphoneme-mixture-components'))

    if not self.checkfile(dest_dir+"/model_0"):
    #Submit initial preparation: copy model and split lists
      self.submit(name+".train-init",
          ["mhtrain_init.sh", dest_dir,
          base_model, word_lst, samples_lst, str(splits)])

    iter_opts = []
    up_opts = []
    finish_opts = []

    if self.conf.get('General','tar') == 'yes':
      self.submit_tar(name, samples_lst, word_lst, iter_opts)

    if store_occs:
      iter_opts.append("-o")
      finish_opts.append("-o -s %d" %splits)

    if self.conf.get('General','store-counts') == 'yes':
      up_opts.append("-c")
      finish_opts.append("-c")

    if self.conf.get('General','mutex-on-write') == "yes":
      iter_opts.append("-m")

    aux= self.conf.get('General','get_cache-tmp')
    if aux != "": iter_opts.append('-T %s'%aux)

    #Launch for all mixtures
    model_name_last = ""
    model_name = "%s/%s_I%.2d" \
      %(self.conf.get('General', 'models-dir'), self.prefix + name, nmixes)

    if not self.checkfile(model_name+".model"):
      #Launch iteration
      for i in xrange(1, iters+1):
        self.submit(name+(".train-em-MIX%.2d-ITER%.d-estimate" %(nmixes, i)),
          ["mhtrain_em_iter.sh"] +
          iter_opts +
          ["-e", self.conf.get('Executables','tLtrain_ex'),
          dest_dir, str(i)],
          array=("1-%d" %splits),
          submit_ops=self.conf.get('Training','submit-iter-ops'))

        #Join counts
        self.submit(name+(".train-em-MIX%.2d-ITER%.d-update" %(nmixes, i)),
          ["mhtrain_em_up.sh"] +
          up_opts +
          ["-e", self.conf.get('Executables','tLupdate_ex'),
          "-A", base_counts,
          dest_dir, str(splits), str(i)],
          submit_ops=self.conf.get('Training','submit-up-ops')
          )

      #Store mixture model
      self.submit(name+(".train-em-MIX%.2d-finish" %(nmixes)),
        ["mhtrain_em_finish.sh", "-d"] +
        finish_opts +
        [dest_dir, str(iters), model_name])

  ###############################################################################
  #### Writes information about the training launched
  ###############################################################################
  def init_launched_info(self):
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M")
    self.launched_info_filename = "launched-train.%s.log" %timestamp

    f = open(self.launched_info_filename,"w")
    self.launched_info_file = f

  ###############################################################################
  #### Writes information about the training launched
  ###############################################################################
  def print_config_on_launched_info(self):
    debug_msg( _( "Writing information about the training in %s" ) %self.launched_info_filename)
    self.launched_info_file.write("### CONFIGURATION FILE USED ###\n\n")
    self.conf.write(self.launched_info_file)
    self.launched_info_file.write("\n### COMMANDS LAUNCHED ###\n\n")

  ##################################################################################
  ##  Submits the generation of tiedphoneme transcriptions
  ##################################################################################
  def generate_tiedphonemes(self, name,
      transcriptions,
      tlist):

      tiedphos_lst = self.tmp_name_prefix(name)+"tiedphos.lst"
      if not self.checkfile(tiedphos_lst):
        self.submit(name+".generate-tiedphonemes",
            ['generate-tiedphos-from-phos.sh',
             transcriptions,
             tlist,
             tiedphos_lst,
             self.tmp_name_prefix(name)+"tiedphonemes",
             self.get_freephos(),'no','0'
             ])


  ##################################################################################
  ##  Train a deep neural network from the samples, transcription and model passed
  ##################################################################################
  def train_deep_neural_network(self, name,
      train_samples,
      train_transcriptions,
      model,
      validation_samples,
      validation_transcriptions):


    dest_dir_align_tr = "%s/%s%s" %(self.conf.get('General', 'train-tmp-dir'),
                              self.prefix, name+".align-train")
    dest_dir_align_val = "%s/%s%s" %(self.conf.get('General', 'train-tmp-dir'),
                              self.prefix, name+".align-validation")
    dest_dir = "%s/%s%s" %(self.conf.get('General', 'train-tmp-dir'),
                              self.prefix, name)

    #Align train
    self.dnn_align_samples(name,
        train_transcriptions,
        train_samples,
        model,
        dest_dir_align_tr
        )

    #align development in order to calculate the validation error
    if self.conf.get('DNN-pretraining','validation') == 'yes' or \
        self.conf.get('DNN-fine-tunning1','validation') == 'yes':

      tiedphos_lst = self.tmp_name_prefix(name)+".validation.tiedphos.lst"
      self.dnn_align_samples(name+".validation",
          tiedphos_lst,
          validation_samples,
          model,
          dest_dir_align_val
          )

    #Get Net topology 5x3000
    net_topology = []
    for hiddden_layer_batch in self.conf.get('DNN-general','topology').split("-"):
      n_hidden_layers=int(hiddden_layer_batch.split("x")[0])
      hidden_units=int(hiddden_layer_batch.split("x")[1])
      for n in xrange(0,n_hidden_layers):
        net_topology.append(hidden_units)

    #Prototype creation
    if not self.checkfile(dest_dir+"/proto"):
      self.submit(name+".generate-proto",
          ['mhdnntrain_init.sh',
           dest_dir,
           model,
           str(net_topology[0])
           ])

    #Pretraining
    for count in range(1,len(net_topology)+1):
      name = "%sPRx%d" %(self.prefix, count)
      net_name = "%s/%s.net" %(dest_dir, name)
      semi_net_name = "%s/%s.semi.net" %(dest_dir, name)

      if not self.checkfile(net_name):
        self.submit(name+".pretraining-%d" %count,
            [ self.conf.get('Executables', 'dnntrain_ex') ] + \
                ( [ '-proto', dest_dir+"/proto" ] if count == 1 else \
                [ '-net',   old_semi_net_name ] ) + \
            [
              '-out',       net_name,
              '-tsamples',  "%s/align.labels.lst" %(dest_dir_align_tr),
              '-tlabels',   "%s/align.labels" %(dest_dir_align_tr),
              '-vsamples',  "%s/align.labels.lst" %(dest_dir_align_val),
              '-vlabels',   "%s/align.labels" %(dest_dir_align_val),
              '-ws',        self.conf.get('DNN-pretraining','window-size'),
              '-epochs',    self.conf.get('DNN-pretraining','epochs'),
              '-batch',     self.conf.get('DNN-pretraining','batch'),
              '-mom',       self.conf.get('DNN-pretraining','momentum'),
              '-wdec',      self.conf.get('DNN-pretraining','weight-decay'),
              '-lr',        self.conf.get('DNN-pretraining','learning-rate'),
              '-seed',      '7',
              '-llb',
              '-lle',
              '-erre' ],
            submit_ops=self.conf.get('Training','submit-dnn-ops')
              )

      if count < len(net_topology):
        if not self.checkfile(semi_net_name):
          self.submit(name+".addLayer-%d" %count,
              [ 'addLayer.sh',
                net_name,
                semi_net_name,
                str(net_topology[count])])

      old_semi_net_name = semi_net_name
      count=count+1

    #Fine tunning 1
    name = "%sFT1" %(self.prefix)
    previous_net = "%s/%sPRx%d.net" %(dest_dir, self.prefix, len(net_topology))
    net_name = "%s/%s.net" %(dest_dir, name)
    if not self.checkfile(net_name):
      self.submit(name+".fine-tunning-1",
          [ self.conf.get('Executables', 'dnntrain_ex'),
            '-net',       previous_net,
            '-out',       net_name,
            '-tsamples',  train_samples,
            '-tlabels',   "%s/state.labels" %(dest_dir_align_tr),
            '-vsamples',  validation_samples,
            '-vlabels',   "%s/state.labels" %(dest_dir_align_val),
            '-ws',        self.conf.get('DNN-fine-tunning-1','window-size'),
            '-epochs',    self.conf.get('DNN-fine-tunning-1','epochs'),
            '-batch',     self.conf.get('DNN-fine-tunning-1','batch'),
            '-mom',       self.conf.get('DNN-fine-tunning-1','momentum'),
            '-wdec',      self.conf.get('DNN-fine-tunning-1','weight-decay'),
            '-lr',        self.conf.get('DNN-fine-tunning-1','learning-rate'),
            '-seed',      '7',
            '-llb',
            '-lle',
            '-erre' ],
          submit_ops=self.conf.get('Training','submit-dnn-ops')
            )



  ##################################################################################
  ##  Generate state alignment for DNN training
  ##################################################################################
  def dnn_align_samples(self, name, transcriptions_list, samples_list, model, dest_dir):

    if not self.checkfile(dest_dir+"/model_0"):
      #Copy model and split lists
      self.submit(name+".train-init",
          ["mhtrain_init.sh", dest_dir,
          model, transcriptions_list, samples_list, self.conf.get('Training','splits')])

    if not self.checkfile(dest_dir+"/state.labels"):
      #align training at state model
      self.submit(name+".align",
          ["mhdnntrain_align.sh",
            "-e", self.conf.get('Executables','tLalign_ex'),
            dest_dir],
            array=("1-%d" %int(self.conf.get('Training','splits'))))

      #puts all together the aligns
      self.submit(name+".align-finish",
          ["mhdnntrain_align_finish.sh",
            "-e", self.conf.get('Executables','tLbin2txt_ex'),
            dest_dir, self.conf.get('Training','splits')])


#####################################################################
#####################################################################
#####################################################################
def main():
  #i18n support
  locale.setlocale(locale.LC_MESSAGES, '')
  locale.setlocale(locale.LC_CTYPE, '')
  PACKAGE = "tlk"
  LOCALEDIR = ""
  if "PACKAGE" in os.environ:
    PACKAGE = os.environ["PACKAGE"]
  if "LOCALEDIR" in os.environ:
    LOCALEDIR = os.environ["LOCALEDIR"]
  gettext.bindtextdomain(PACKAGE, LOCALEDIR)
  gettext.textdomain(PACKAGE)

  config = ConfigParser.ConfigParser(allow_no_value=True)
  #Load defaults
  for i in config_file_defaults:
    config.readfp(io.BytesIO(i))

  parser = argparse.ArgumentParser(
      prog="tLtask-train",
      description=_('Train a standard acoustic model using TLK'),
      epilog=_("""This program performs the training of an acoustic model using TLK.
All necessary parameters are loaded from a configuration file. Please make sure that
all necessary files are ready before training. The training process might take long,
and some errors due to empty files might go undetected. For more details, please
read the help at http://translectures.eu/tlk/help."""))

  parser.add_argument(dest='config_file',
    help=_("Load configuration from this file"), nargs="?")
  parser.add_argument("--write-example-config-file", action='store_true',dest='write_example',
    help=_("Write the default configuration to the standard output and exit"))
  parser.add_argument("-S", "--scripts-dir", dest='scripts_path',
    help=( _( "Directory containing the training scripts" ) ), metavar="DIRNAME")
  parser.add_argument("--no-filecheck", dest='filecheck',
    help=_("Disable file checking. Using this option forces the training to be performed from scratch"), action="store_false")
  parser.add_argument("--no-launch", dest='not_launch',
    help=_("Do not submit any command, only print them"), action="store_true")
  parser.add_argument('--log-folder',
    help=_('Store all logs in this folder'), metavar="DIRNAME",
    dest='log_folder')
  parser.add_argument('--delete-tmp',
    help=_('Delete the tmp folder once the training is finished'),
    action='store_true')
  parser.add_argument('--training-mode',
    help=_('Select the training model: Batch (training from scratch)'
        ', Incremental (iterating over a given model)'),
      dest='training_mode', choices=['Batch', 'Incremental'], default='Batch')
  parser.add_argument('--depend-on',
    help=_('The first process depends on this if defined. Only available in grid environments.'),
    dest='depend_on')
  parser.add_argument('-v',
    help=_('Output additional verbose information'),
    action='store_true')
  parser.add_argument('--version',
    help=_('Output version information and exit'),
    action='store_true')

  parser.add_argument(dest='overwrite',
    help=_("Overwrite variables from config. Format -> Section_Variable=Value"), nargs='*')

  args = parser.parse_args(argv[1:])

  #write a sample config file
  if args.write_example:
    config.write(stdout)
    exit(0)

  #show version (according to translectures-revision SVN)
  if args.version:
    show_version("tLtask-train")
    exit(0)

  #read an external config file and check version number
  if not args.config_file:
    parser.error( _( "A config_file must be given. Use the option --write-example-config-file \
to generate a template." ) )
  else:
    ret = config.read(args.config_file)
    if args.config_file not in ret:
      fatal_err( _( "There has been a problem reading the configuration file: %s"
            %args.config_file ) )

    if not config.has_option('General','tLtask-train-version') or \
        ( config.get('General','tLtask-train-version') != '_RELEASE_' and \
        config.get('General','tLtask-train-version') != __version__ ):
      fatal_err( _(
"Wrong version number in loaded configuration file. Some options may have\n"
"changed or not be supported anymore. Check your configuration, and\n"
"update the tLtask-train-version to %d if everything is correct."
        ) %int(__version__) )

  #Overwrite variables with command line
  if args.overwrite != None:
    for o in args.overwrite:
      m = re.search("^(.*)_(.*)=(.*)$", o)
      if m == None:
        fatal_err( _( "Incorrect overwrite in element %s" ) %o )
      debug_msg("Setting [%s] %s = %s" %(m.group(1), m.group(2), m.group(3)))
      if config.has_option(m.group(1),m.group(2)):
        config.set(m.group(1), m.group(2), m.group(3))
      else:
        fatal_err( _( "Section and options of overwrite command %s do not exist in config" ) %o )

  if args.scripts_path:
    os.putenv('PATH', os.environ['PATH']+':'+args.scripts_path)
  else: #add the scripts folder on this executable path
    dirname = os.path.dirname(os.path.realpath(argv[0]))
    os.environ['PATH'] = os.environ['PATH']+':'+dirname+'/scripts'

  if args.not_launch:
    stdout.write("export PATH=%s\n" %os.environ['PATH'])

  #create log folder if necessary
  if args.log_folder != None:
    if not os.path.exists(args.log_folder):
      os.mkdir(args.log_folder)
    elif not os.path.isdir(args.log_folder):
      debug_msg( _( "Log folder is an archive. All logs will be appended to this file" ) )

  trainer = tLtrainer(config, verbose=args.v, filecheck=args.filecheck,
      not_launch=args.not_launch, log_folder=args.log_folder)

  if args.depend_on:
    trainer.last_submitted = args.depend_on

  #check if the files and directories necessary exists
  if args.filecheck:
    if not os.path.exists(config.get('Lists','standard-samples')):
      fatal_err( _( "Standard samples list: %s not found\n" ) %config.get('Lists','standard-samples'))

    if not os.path.exists(config.get('Lists','clusters')):
      fatal_err( _( "Cluster list: %s not found\n" ) %config.get('Lists','clusters'))

    if not os.path.exists(config.get('Lists','clusters-samples')):
      fatal_err( _( "Samples cluster list: %s not found\n" ) %config.get('Lists','clusters-samples'))

    if not os.path.exists(config.get('Lists','phonemes')):
      fatal_err( _( "Phoneme annotation list: %s not found\n" ) %config.get('Lists','phonemes'))

    if not os.path.exists(config.get('Lists','unique-monophonemes')):
      fatal_err( _( "Phonemes list: %s not found\n" ) %config.get('Lists','unique-monophonemes'))

  if not os.path.isdir(config.get('General','train-tmp-dir')):
    os.mkdir(config.get('General','train-tmp-dir'))

  if not os.path.isdir(config.get('General','models-dir')):
    os.mkdir(config.get('General','models-dir'))

  #print in a log file the configuration employed
  trainer.init_launched_info()
  trainer.print_config_on_launched_info()

 ####################    Train from scratch  #####################
  if args.training_mode == 'Batch':
    #1.1 generate monophone model
    trainer.info_msg( _( "Training standard monophone model" ) )
    trainer.train_monophonemes("standard",
        config.get('Lists', 'phonemes'),
        config.get('Lists', 'standard-samples'))

    #1.2 generate triphone model
    trainer.info_msg( _( "Training standard triphone model" ) )
    trainer.train_triphonemes("standard",
        config.get('Lists', 'phonemes'),
        config.get('Lists', 'standard-samples')
        )

    #1.3 generate tiedphoneme model
    trainer.info_msg( _( "Training standard tiedphoneme model" ) )
    trainer.train_tiedphonemes("standard",
        config.get('Lists', 'phonemes'),
        config.get('Lists', 'standard-samples'))

    #1.4 train deep neural network
    #if "standard" in [ x.strip() for x in config.get('DNN-general','apply-on').split(",")]:
    if False:
      trainer.info_msg( _( "Training standard deep neural network" ) )

      trainer.generate_tiedphonemes("standard-dnn.validation",
          config.get('DNN-validation-lists', 'phonemes'),
          "%s/%s.tlist" %(config.get('General','models-dir'), trainer.prefix + "standard"))

      trainer.generate_tiedphonemes("standard-dnn.validation",
          config.get('DNN-validation-lists', 'phonemes'),
          "%s/%s.tlist" %(config.get('General','models-dir'), trainer.prefix + "standard"))

      trainer.train_deep_neural_network("standard-dnn",
          config.get('Lists', 'standard-samples'),
          "%s/%s.standard.tiedphos.lst" %(config.get('General', 'train-tmp-dir'),
                                config.get('General', 'prefix-name')),
          "%s/%s.standard.tiedphoneme_I%s.model" %(config.get('General', 'models-dir'),
                                                config.get('General', 'prefix-name'),
                                config.get('Training', 'tiedphoneme-mixture-components')),
          config.get('DNN-validation-lists', 'standard-samples'),
          "%s/%s.standard-dnn.validation.tiedphos.lst" %(config.get('General', 'models-dir'),
                                                config.get('General', 'prefix-name')),
         )

    #4 generate CMLLR features
    trainer.info_msg( _( "Generating CMLLR samples" ) )
    trainer.generate_CMLLR_samples("standard",
        config.get('Lists','standard-samples'),
        "%s/%s.standard.tiedphos.lst" %(config.get('General', 'train-tmp-dir'),
                              config.get('General', 'prefix-name')),
        config.get('Lists','clusters-samples'),
        config.get('Lists','clusters'),
        "%s/%s.standard.tiedphoneme_I01.model"%(config.get('General', 'models-dir'),
                                              config.get('General', 'prefix-name')),
        config.get('General', 'train-tmp-dir')
        )

    cmllr_samples = "%s/cmllrfeas.lst" %(config.get('General', 'train-tmp-dir'))
    #5 generate monophone model
    trainer.info_msg( _( "Training CMLLR monophones" ) )
    trainer.train_monophonemes("cmllr",
        config.get('Lists', 'phonemes'),
        cmllr_samples)

    #6 generate triphone model
    trainer.info_msg( _( "Training CMLLR triphone model" ) )
    trainer.train_triphonemes("cmllr",
        config.get('Lists', 'phonemes'),
        cmllr_samples,
        )

    #7 generate tiedphoneme model
    trainer.info_msg( _( "Training CMLLR tiedphoneme model" ) )
    trainer.train_tiedphonemes("cmllr",
        config.get('Lists', 'phonemes'),
        cmllr_samples)

    #1.4 train deep neural network
    #if "cmllr" in [ x.strip() for x in config.get('DNN-general','apply-on').split(",")]:
    if False:
      trainer.info_msg( _( "Training CMLLR deep neural network" ) )

      trainer.generate_tiedphonemes("cmllr-dnn.validation",
          config.get('DNN-validation-lists', 'phonemes'),
          "%s/%s.tlist" %(config.get('General','models-dir'), trainer.prefix + "cmllr"))

      trainer.generate_CMLLR_samples("cmllr-dnn.validation",
          config.get('DNN-validation-lists', 'standard-samples'),
          "%s/%s.cmllr.tiedphos.lst" %(config.get('General', 'train-tmp-dir'),
                      config.get('General', 'prefix-name')),
          config.get('DNN-validation-lists', 'clusters-samples'),
          config.get('DNN-validation-lists', 'clusters'),
          "%s/%s.cmllr.tiedphoneme_I01.model" %(config.get('General', 'models-dir'),
                                                config.get('General', 'prefix-name')),
          config.get('General', 'train-tmp-dir')+"/validation/"
          )

      trainer.train_deep_neural_network("cmllr-dnn",
          "%s/cmllrfeas.lst" %(config.get('General', 'train-tmp-dir')),
          "%s/%s.cmllr.tiedphos.lst" %(config.get('General', 'train-tmp-dir'),
                      config.get('General', 'prefix-name')),
          "%s/%s.cmllr.tiedphoneme_I%s.model" %(config.get('General', 'models-dir'),
                                                config.get('General', 'prefix-name'),
                                config.get('Training', 'tiedphoneme-mixture-components')),
          "tmp/validation/cmllrfeas.lst",
          "%s/%s.cmllr-dnn.validation.tiedphos.lst" %(config.get('General', 'models-dir'),
                                                config.get('General', 'prefix-name')),
         )

  ####################    Incremental training   #####################
  elif args.training_mode == 'Incremental':
    #check config file
    if not config.has_section('Incremental-Training'):
      fatal_err( _( "Missing Incremental-Training section in configuration file\n" ) )

    #check required models
    if args.filecheck:
      if not os.path.exists("%s.model" %(config.get('Incremental-Training','base-target-model-path'))):
        fatal_err( _( "Base target model does not exist" ) )
      if not os.path.exists("%s.model" %config.get('Incremental-Training','base-standard-tiedphoneme-model-path')):
        fatal_err( _( "Base standard model does not exist" ) )
      if not os.path.exists("%s.counts" %(config.get('Incremental-Training','base-standard-tiedphoneme-model-path'))):
        fatal_err( _( "Base standard counts do not exist" ) )
      if not os.path.exists("%s.model" %(config.get('Incremental-Training','base-cmllr-tiedphoneme-model-path'))):
        fatal_err( _( "Base CMLLR model does not exist" ) )
      if not os.path.exists("%s.counts" %(config.get('Incremental-Training','base-cmllr-tiedphoneme-model-path'))):
        fatal_err( _( "Base CMLLR counts do not exist" ) )

    #1.1 generate tiedphoneme model
    trainer.train_incremental_tiedphonemes("standard",
        config.get('Lists', 'phonemes'),
        config.get('Lists', 'standard_samples'),
        config.get('Incremental-Training','base-standard-tiedphoneme-model-path'))

    #1.2 generate CMLLR features
    trainer.generate_CMLLR_samples("standard",
        config.get('Lists','standard-samples'),
        "%s/%s.standard.tiedphos.lst" %(config.get('General', 'train-tmp-dir'),
                              config.get('General', 'prefix-name')),
        config.get('Lists','clusters-samples'),
        config.get('Lists','clusters'),
        "%s.model" %config.get('Incremental-Training','base-target-model-path'),
        config.get('General', 'train-tmp-dir')
        )
    cmllr_samples = "%s/cmllrfeas.lst" \
            %(config.get('General', 'train-tmp-dir'))

    #2 generate tiedphoneme model
    trainer.train_incremental_tiedphonemes("cmllr",
        config.get('Lists', 'phonemes'),
        cmllr_samples,
        config.get('Incremental-Training','base-cmllr-tiedphoneme-model-path'))


  #####################   Common steps   ############################
  if args.delete_tmp:
    #8 delete tmp-folder
    trainer.delete_tmp()

####################################################################################################
if __name__ == "__main__":
  main()
