#!/usr/bin/env python
# -*- coding: utf-8 -*-

# tLtranscribe: Automatically transcribe a video using TLK
# Copyright 2014 transLectures-UPV Team
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#include i18n
import gettext
_ = gettext.gettext

import argparse
import ConfigParser
from sys import argv, exit, stdin, stdout, stderr
import os
import io
import subprocess
import shlex
import locale
import re
import tempfile
import shutil

__version__ = filter(str.isdigit, "$Revision: 6979 $")

def show_version(prog):
  stdout.write("%s version: %s\n" %(prog, __version__))

config_file_general_defaults = """
[General]
bindir                   = """+(os.environ['TLK_PATH']+"/" if 'TLK_PATH' in os.environ else "")+"""
submit-ex = submit.sh
tLtask-segment-ex = %(bindir)stLtask-segment
tLtask-recognise-ex = %(bindir)stLtask-recognise
submit-options-default =
prefix-name =
"""

config_file_segmenter_defaults = """
[Segmenter]
config-file =
"""

config_file_recogniser_defaults = """
[Recogniser]
config-file =
"""

config_file_defaults = \
[
  config_file_general_defaults,
  config_file_segmenter_defaults,
  config_file_recogniser_defaults
]

def fatal_err(message):
  stderr.write("[EE] "+ message + "\n")
  exit(1)

def debug_msg(message):
  stderr.write("[II] "+ message + "\n")

class tLtranscribe:
  def __init__(self, conf, verbose=False, filecheck=True, not_launch=False,
      log_folder=None, tmp_folder=None):
    self.conf = conf
    self.last_submitted = None
    self.recognise_outdir = None
    self.segment_outdir = None
    self.v = verbose
    self.tmp_folder = tmp_folder
    self.filecheck = filecheck
    self.not_launch = not_launch
    self.log_folder = log_folder
    self.log_folder_submit_op = []
    if self.log_folder != None:
      self.log_folder_submit_op = ("-o %s" %log_folder).split()
    if self.conf.get('General','submit-options-default') != "":
      self.submit_ops_default = self.conf.get('General','submit-options-default')
    else: self.submit_ops_default= None

    #Initialise some parameters
    if self.conf.get('General','prefix-name') == "":
      self.prefix = ""
    else:
      self.prefix = self.conf.get('General','prefix-name') + "."

  def info_msg(self, message):
    if self.not_launch:
      stdout.write("\n\n# [II] "+ message + "\n\n")
    else:
      stdout.write("[II] "+ message + "\n")

  #####################################################################
  ##  submit a process given a name and a command line
  #####################################################################
  def submit(self, submission_name, command_line,
    array=None, depend_on=None, submit_ops=None,
    bgworker=None #launch process in parallel and execute this function
                           #periodically
    ): #None = depend on the launched proc

    submission_name = self.prefix + submission_name
    #submit options
    cmd = [self.conf.get('General', 'submit-ex'), "-n ", submission_name]
    depend = None
    if self.last_submitted:
      depend = self.last_submitted
    if depend_on:
      depend = depend_on
    if depend != None and depend != "":
      cmd += ['-w ', depend]
    if array:
      cmd += ['-j ', array]
    if submit_ops:
      cmd += submit_ops.split()
    elif self.submit_ops_default != None:
      cmd += self.submit_ops_default.split()
    if self.log_folder_submit_op != []:
      cmd += self.log_folder_submit_op

    #process
    cmd += command_line

    #Execute
    #Workaround subprocess " " arguments insertions
    string = " ".join(cmd)
    if self.v:
      debug_msg( _( "Submitting: %s\n" ) %string )

    #prepare cmd to execute the process
    cmd = shlex.split(string)
    if self.not_launch:
      stdout.write("%s\n" %string)
    else:
      try:
        if bgworker == None:
          proc = subprocess.Popen(cmd)
          ret = proc.wait()
        else:
          bgworker_function, bgworker_params = bgworker
          proc = subprocess.Popen(cmd)
          ret = bgworker_function(proc, bgworker_params)
        if ret != 0:
          fatal_err( _( "Failed when executing: '%s' with exit code: %d" ) %(string, ret))
      except Exception as e:
        fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(string, e) ) )

    self.last_submitted = submission_name

  #####################################################################
  ##  Prepare file list for segmentation
  #####################################################################
  def prepare_segmentation(self, name, video_file, depend=''):
    if depend != '':
      self.submit(name,
        ["prepare-segmentation.sh",
          video_file,
          self.tmp_folder],
        depend_on=depend)
    else:
      self.submit(name,
        ["prepare-segmentation.sh",
          video_file,
          self.tmp_folder])

  #####################################################################
  ##  Execute tLtask-segment to segment a video file
  #####################################################################
  def launch_segmentation(self, data_dir, video_id, overr_opts='', segm_trs=''):
    sub_dir = "segmenter"
    command = ("%s " %self.conf.get('General','tLtask-segment-ex') +
               ( "--no-launch " if self.not_launch else " " ) +
               "--log-folder %s " %(self.log_folder) +
               "--depend-on %s " %self.last_submitted +
               ("--segmentation-trs %s " % segm_trs if segm_trs != '' else ' ') +
               "--no-filecheck " +
               "--print-output-info " +
               "%s " % self.get_real_path(data_dir, 'segmenter', self.conf.get('Segmenter','config-file')) +
               "Lists_videos-list=%s/video.lst " %(self.tmp_folder) +
               "General_output-dir=%s " %(self.tmp_folder+"/segment/out") +
               "General_tmp-dir=%s " %(self.tmp_folder+"/segment/tmp") + 
               "General_prefix-name=%s%s " %(self.prefix, video_id) + 
               ("General_submit-options-default='%s' " %self.conf.get('General','submit-options-default') if self.conf.get('General','submit-options-default') != '' else '') + 
               ("Executables_submit=%s " %(self.conf.get('General','submit-ex')) if self.conf.get('General','submit-ex') != '' else '') +
               "%s " % overr_opts)

    cmd = shlex.split(command)
    try:
      ret = subprocess.check_output(cmd)
    except Exception as e:
      fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(command, e) ) )
  
    self.segment_outdir, self.last_submitted = self.parse_output_info(ret)

  #####################################################################
  ##  Prepare file list for recognition
  #####################################################################
  def prepare_recognition(self, name, video_file):
    self.submit(name,
        ["prepare-recognition.sh",
          video_file,
          self.tmp_folder])

  #####################################################################
  ##  Execute tLtask-recognise to recogniser the automatically segmented video
  #####################################################################
  def launch_recognition(self, data_dir, video_id, parallel_execution = False, overr_opts='', comb_opt=''):
    cores = "General_splits=1" if parallel_execution else ""

    samples_lst_name = "%s/segment/out/allfeas.lst" %self.tmp_folder
    clusters_lst_name = "%s/clusters.lst" %self.tmp_folder
    clusters_samples_lst_name = "%s/clusters_samples.lst" %self.tmp_folder

    sub_dir = "recogniser"
    command = "%s " %self.conf.get('General','tLtask-recognise-ex') + \
                ( "--no-launch " if self.not_launch else " " ) + \
                "--log-folder %s " %(self.log_folder) + \
                "--depend-on %s " %(self.last_submitted) + \
                ( "--comb-senones-lst %s " % comb_opt if comb_opt != '' else " ") + \
                "--no-filecheck " + \
                "--print-output-info " + \
                "%s " % self.get_real_path(data_dir, 'recogniser', self.conf.get('Recogniser','config-file')) + \
                "Lists_standard-samples=%s " %(samples_lst_name) + \
                "Lists_clusters=%s " %(clusters_lst_name) + \
                "Lists_clusters-samples=%s " %(clusters_samples_lst_name) + \
                "General_output-dir=%s " %(self.tmp_folder+"/recognition/out") + \
                "General_prefix-name=%s%s " %(self.prefix, video_id) + \
                ( "General_submit-options-default='%s' " %self.conf.get('General','submit-options-default') if self.conf.get('General','submit-options-default') != '' else '') + \
                ("Executables_submit=%s " %(self.conf.get('General','submit-ex')) if self.conf.get('General','submit-ex') != '' else '') + \
                "%s " % cores + \
                "%s " % overr_opts
    cmd = shlex.split(command)
    try:
      ret = subprocess.check_output(cmd)
    except Exception as e:
      fatal_err( _( "Could not execute command: \"%s\" . Exception: %s" %(command, e) ) )

    self.recognise_outdir, self.last_submitted = self.parse_output_info(ret)

  #####################################################################
  ##  Parse tLtask output info
  #####################################################################
  def parse_output_info(self, tLtask_output):
    outdir = ''; b1 = False
    last_sub = ''; b2 = False
    for l in tLtask_output.split('\n'):
        if b1 and b2: break
        if len(l.split()) < 2: continue
        if l.split()[0] == '__OUTPUT_DIR:':
            outdir = ' '.join(l.split()[1:]).strip()
            b1 = True
        elif l.split()[0] == '__LAST_SUBMITTED:':
            last_sub = ' '.join(l.split()[1:]).strip()
            b2 = True
    if not(b1 and b2):
        fatal_err( _( "Could not retrieve __OUTPUT_DIR and __LAST_SUBMMITED tokens from the recogniser output." ))
    return outdir, last_sub

  #####################################################################
  ##  Copy back srts
  #####################################################################
  def copy_back_results(self, name, dest):
    self.submit(name,
        ["copy-back-srt.sh",
          self.recognise_outdir,
          dest])

  #####################################################################
  ##  Delete tmp
  #####################################################################
  def delete_tmp(self, name):
    self.submit(name,
        ['rm', '-rf', self.tmp_folder])

  #####################################################################
  ##  Misc
  #####################################################################
  def get_real_path(self, data_dir, sub_dir, file_path):
      return file_path if os.path.isabs(file_path) else data_dir+"/"+sub_dir+"/"+file_path  

#####################################################################
#####################################################################
#####################################################################
def main():
  #i18n support
  locale.setlocale(locale.LC_MESSAGES, '')
  locale.setlocale(locale.LC_CTYPE, '')
  PACKAGE = "tlk"
  LOCALEDIR = ""
  if "PACKAGE" in os.environ:
    PACKAGE = os.environ["PACKAGE"]
  if "LOCALEDIR" in os.environ:
    LOCALEDIR = os.environ["LOCALEDIR"]
  gettext.bindtextdomain(PACKAGE, LOCALEDIR)
  gettext.textdomain(PACKAGE)

  #auxiliary scripts path
  dirname = os.path.dirname(os.path.realpath(argv[0]))
  os.environ['PATH'] = os.environ['PATH']+':'+dirname+'/scripts'

  #Path to systems
  tLtranscribe_data_path = "."
  if 'TLK_SYSTEMS_DATADIR' in os.environ:
    tLtranscribe_data_path = os.environ['TLK_SYSTEMS_DATADIR']

  config = ConfigParser.ConfigParser(allow_no_value=True)
  #Load defaults
  for i in config_file_defaults:
    config.readfp(io.BytesIO(i))
#config.set('General','tLtask-recognise-version',0)

  parser = argparse.ArgumentParser(
      prog="tLtranscribe",
      description=_('Automatically transcribe a video using installed TLK models'),
      epilog=_("""tLtranscribe reads a video the from command line and automatically
transcribes it using TLK tools. All the process is performed using installed
preconfigured models."""))

  parser.add_argument(dest='video_file',
    help=_("Video file to be transcribed."))
  parser.add_argument('--config-file',
    help=_('Load configuration from this file'), dest='config_file', default='', metavar="CONFIG")
  class WriteConfigAction(argparse.Action):
    def __init__(self,option_strings,dest,nargs=0,**kwargs):
      super(WriteConfigAction, self).__init__(option_strings,dest,0,
                                              **kwargs)
    def __call__(self,parser,namespace,values,option_string=None):
      config.write(stdout)
      exit(0)
  parser.add_argument("--write-example-config-file", action=WriteConfigAction,
                      help=_("Write the default configuration to the"+
                             " standard output and exit"))
  parser.add_argument('--language',
    help=_('Language of the speech in the video. Currently, Spanish (es) is supported'),
      choices=['es'], default='es')
  parser.add_argument('--video-id', dest='video_id', default='',
    help=_('Video ID.'))
  parser.add_argument('--output-file', dest='output_file', default='',
    help=_('Path to store the resulting transcription file.'))
  parser.add_argument('--tmp-folder',
    help=_('Path to the folder for temporary file storage. By default, a temporary folder will be created using mktemp'))
  parser.add_argument('--not-delete-tmp',
    help=_('Do not delete the tmp folder after finishing the transcription'),
    action='store_true')
  parser.add_argument('--override-segmenter-config-opts',
    help=_('tLtask-segment options'), metavar="OPTS",
    dest='overr_segmenter_opts', default='')
  parser.add_argument('--override-recogniser-config-opts',
    help=_('tLtask-recognise options'), metavar="OPTS",
    dest='overr_recogniser_opts', default='')
  parser.add_argument("--no-launch", dest='not_launch',
    help=_("Do not submit any command, only print them"), action="store_true")
  parser.add_argument('--log-folder',
    help=_('All logs will be stored in this folder'), metavar="DIRNAME",
    dest='log_folder', default='log')
  parser.add_argument('--comb-senones-lst', dest='comb_feas', default='',
    help=_('Combine current system with another system senones list after dnn adaptation.'))
  parser.add_argument('--segmentation-trs', dest='segmentation_trs', default='',
    help=_('Use the segmentation of the provided trs file.'))
  parser.add_argument('--systems-folder',
    help=_('Path to a directory containing the tLtranscribe systems. By default, it is assumed to be installed in the location of TLK'), metavar="DIRNAME")
  parser.add_argument('--only-one-core',
    help=_('Do not execute processes paralelly'),
    action='store_true')
  parser.add_argument('--depend-on',
    help=_('The first process depends on this if defined. Only available in grid environments.'),
    dest='depend_on', default='')
  parser.add_argument('--version',
    help=_('Output version information and exit'),
    action='store_true')
  parser.add_argument('-v',
    help=_('Output additional verbose information'),
    action='store_true')

  args = parser.parse_args(argv[1:])

  # Assigns this enviorement variable to set the name of all log messages
  os.putenv('PRGNAME', 'tLtranscribe')

  #show version (according to translectures-revision SVN)
  if args.version:
    show_version("tLtranscribe")
    exit(0)

  if args.systems_folder:
    tLtranscribe_data_path = args.systems_folder

  data_dir = tLtranscribe_data_path+"/"+args.language
  video_id = os.path.basename(args.video_file)
  if args.video_id != '': 
    video_id = args.video_id

  if args.config_file != '':
    tLtranscribe_conf_filename = args.config_file
    if not os.path.exists(tLtranscribe_conf_filename):
      fatal_err( _( """Configuration file %s does not exist."""
            %tLtranscribe_conf_filename ) )
  else:
    tLtranscribe_conf_filename = data_dir+"/tLtranscribe.conf"
    #Load configuration file from standard route
    if not os.path.exists(tLtranscribe_conf_filename):
      fatal_err( _( """Configuration file %s does not exist. Check the models directory or
change it using the --systems-folder option or the environment variable
TLK_SYSTEMS_DATADIR."""
            %tLtranscribe_conf_filename ) )

  if  not config.read(tLtranscribe_conf_filename):
    fatal_err( _( "There has been a problem reading the configuration file: %s"
          %tLtranscribe_conf_filename ) )

  # Create a temporary file
  if args.tmp_folder == None:
    args.tmp_folder = tempfile.mkdtemp(dir=".")
  else:
    if not os.path.exists(args.tmp_folder):
      os.makedirs(args.tmp_folder)

  if not os.path.exists(args.tmp_folder+"/segment"):
    os.makedirs(args.tmp_folder+"/segment")
  
  if not os.path.exists(args.tmp_folder+"/recognition"):
    os.makedirs(args.tmp_folder+"/recognition")

  #WARNING!!! clusters.lst is still needed at the start!
  clusters_lst = open(args.tmp_folder+"/clusters.lst", "w")
  clusters_lst.write("%s\n" %os.path.basename(args.video_file))
  clusters_lst.close()

  #create log folder if necessary
  if args.log_folder != None:
    if not os.path.exists(args.log_folder):
      os.mkdir(args.log_folder)
    elif not os.path.isdir(args.log_folder):
      debug_msg( _( "Log folder is an archive. All logs will be appended to this file" ) )

  tLtranscriber = tLtranscribe(config, verbose=args.v,
      not_launch=args.not_launch, log_folder=args.log_folder, tmp_folder=args.tmp_folder)

  # Execute segmenter
  debug_msg( _( "Executing tLtask-segment\n" ) )
  tLtranscriber.prepare_segmentation("%s.prepare-segmentation" % video_id, args.video_file, depend=args.depend_on)
  tLtranscriber.launch_segmentation(data_dir, video_id, overr_opts=args.overr_segmenter_opts, segm_trs=args.segmentation_trs)

  # Execute recogniser
  debug_msg( _( "Executing tLtask-recognise\n" ) )
  tLtranscriber.prepare_recognition("%s.prepare-recognition" % video_id, args.video_file)
  tLtranscriber.launch_recognition(data_dir, video_id, 
      parallel_execution = args.only_one_core, overr_opts=args.overr_recogniser_opts, comb_opt=args.comb_feas)

  #copy back the generated srt file
  debug_msg( _( "Copying back the results\n" ) )
  dest = "%s.srt" % args.video_file
  if args.output_file != '':
      dest = args.output_file 
  tLtranscriber.copy_back_results("%s.copy-back-results" % video_id, dest)

  if not args.not_delete_tmp:
    # delete tmp-folder
    tLtranscriber.delete_tmp("%s.delete-tmp" % video_id)

####################################################################################################
if __name__ == "__main__":
  main()

