/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  align.c - Realitza aliniament forçat.
 *
 *  Adrià Giménez Pastor, 2008,2010,2011,2012,2013.
 *
 */

#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "get_pbm_lst.h"
#include "get_words.h"
#include "macros.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 3




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  const char *output_fn;
  tLBool      nowords;
  int         statealign;
  int         printprobs;
  
};


struct args
{
  
  const char *amodel_fn;
  const char *wdlst_fn;
  const char *pbmlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{

  fprintf ( to, "%s [%s] <amodel> <transc-lst> <sample-lst>\n\n",
	    prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Acoustic model") );
  fprintf ( to, "    <transc-lst> %s\n",
	    _("List of the files containing the transcriptions") );
  fprintf ( to, "    <sample-lst> %s\n\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the initialized model will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      --state-align: %s\n", _("State-level alignment") );
  fprintf ( to, "      --print-probs: %s\n",
	    _("Print the probability for each sample") );
  fprintf ( to, "      --words: %s\n",
	    _("Transcriptions will be expected to be words") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, cwords, n;
  const char *prog;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "state-align", &(opts->statealign) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "print-probs", &(opts->printprobs) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "words", &cwords );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  opts->nowords= (cwords == 0);
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->amodel_fn= argv[0];
  args->wdlst_fn= argv[1];
  args->pbmlst_fn= argv[2];
  
} /* end usage */


static void
align_sample (
	      tLAligner   *aligner,
	      tLDict      *syms,
	      tLSeqs      *words,
	      tLFea       *fea,
	      FILE        *out,
	      const char  *pbm,
	      size_t       word,
	      tLSegment  **alignment,
	      size_t      *asize,
	      int          statealign,
	      int          printprobs,
	      int          verbose
	      )
{
  
  tLSeq cs;
  char *err;
  tLSegment *p;
  const tLAModel *amodel;
  tLFloat prob;
  
  
  /* Prepara. */
  amodel= aligner->amodel;
  if ( tL_fea_load ( fea, pbm, &err ) == -1 ) tL_error ( err );
  if ( !tL_amodel_test ( amodel, fea, &err ) )
    {
      tL_warning ( _("Sample '%s' was not aligned: %s"), pbm, err );
      free ( err );
      return;
    }
  cs= *(tL_seqs_get_seq ( words, word ));
  
  if ( verbose > 1 )
    {
      fprintf ( stderr,
		_("Aligning sample '%s' with the transcription '"),
		pbm );
      tL_seq_print ( &cs, stderr, syms, ' ' );
      fprintf ( stderr, "'\n" );
    }
  
  /* Alinea. */
  prob= -HUGE_VAL;
  if ( tL_align ( aligner, fea, word, alignment, asize, &prob ) == -1 )
    {
      tL_warning ( _("Sample '%s' cannot be aligned"), pbm );
      return;
    }
  if ( printprobs ) fprintf ( out, "PROB %g %s\n", _DB(prob), pbm );
  for ( p= *alignment; p->length != -1; ++p )
    if ( statealign )
      fprintf ( out, "%s %u %d %d %s\n",
		tL_dict_get_token ( syms, p->c ), p->q,
		p->pos, p->length, pbm );
    else
      fprintf ( out, "%s %d %d %s\n",
		tL_dict_get_token ( syms, p->c ),
		p->pos, p->length, pbm );
  
} /* end align_sample */


static void
align (
       tLAModel   *amodel,
       tLDict     *syms,
       tLFileList *pbm_lst,
       tLSeqs     *words,
       const char *output,
       int         statealign,
       int         printprobs,
       int         verbose
       )
{
  
  tLAligner *aligner;
  tLFea *fea;
  FILE *to;
  tLSegment *alignment;
  size_t asize, n;
  char *err;
  
  
  if ( verbose )
    fprintf ( stderr, _("Aligning\n") );
  
  /* Reserva memòria. */
  aligner= tL_aligner_new ( amodel, words, statealign, &err );
  if ( aligner == NULL ) tL_error ( err );
  fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  if ( output == NULL ) to= stdout;
  else
    {
      to= fopen ( output, "w" );
      if ( to == NULL )
        tL_error ( _("Could not open '%s'"), output );
    }
  alignment= NULL;
  
  /* Alinea. */
  for ( n= 0; n < pbm_lst->N; ++n )
    align_sample ( aligner, syms, words, fea,
		   to, pbm_lst->files[n], words->seqs[n],
		   &alignment, &asize, statealign, printprobs,
		   verbose );
  
  /* Allibera memòria. */
  if ( alignment != NULL ) free ( alignment );
  if ( to != stdout ) fclose ( to );
  tL_fea_free ( fea );
  tL_aligner_free ( aligner );
  
} /* end align */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel;
  tLFileList *pbm_lst;
  tLSeqs *words;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  words= get_words ( args.wdlst_fn, syms, opts.nowords, NULL, opts.verbose );
  if ( words->N != pbm_lst->N )
    tL_error ( _("The sample list contains %lu entries,"
		 " but the word list contains %lu entries"),
	       _PS(pbm_lst->N), _PS(words->N) );
  
  /* Cos. */
  align ( amodel, syms, pbm_lst, words,
	  opts.output_fn, opts.statealign,
	  opts.printprobs, opts.verbose );
  
  /* Allibera memòria. */
  tL_seqs_free ( words );
  tL_filelist_free ( pbm_lst );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
