/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  bin2txt.c - Programa que transforma de format binari a format text
 *              i viceversa. Cal recordar que el format binari és
 *              depenent de la màquina pero permet mantindre el valors
 *              exactes que es representen en memòria.
 *
 *  Adrià Giménez Pastor, 2010,2011,2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 1




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  const char *input_fn;
  const char *output_fn;
  int         reverse;
  
};


struct args
{
  
  const char *format;
  
};



/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] [amodel|lm|pbm|realfea]\n\n",
	    prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    amodel: %s\n", _("acoustic model") );
  fprintf ( to, "    lm: %s\n", _("language model") );
  fprintf ( to, "    pbm: %s\n",
	    _("file containing a sequence of binary feature vectors"
	      " represented as a binary image") );
  fprintf ( to, "    realfea: %s\n\n",
	    _("file containing a sequence of real feature vectors") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -i,--input <string>: %s\n",
	    _("File from which the data to be transformed will be read."
	      " Defaults to the standard input") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the transformed input data will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -r,--reverse: %s\n",
	    _("Convert from text to binary format") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          *argc,
       const char  **argv[],
       struct args  *args, 
       struct opts  *opts
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n;
  const char *prog;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'i', "input", &(opts->input_fn) );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_counter ( cmdl, 'r', "reverse", &(opts->reverse) );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, *argc, *argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  
  /* Arguments. */
  if ( n < NUM_ARGS ) goto error;
  args->format= (*argv)[0];
  
  return;
  
 error:
  help ( stderr, prog, FALSE );
  tL_error ( _("Unknown option or wrong number of arguments;"
	       " type --help for help") );
  
} /* end usage */


static gzFile
open_input (
	    const char   *input_fn,
	    const tLBool  mode_text
	    )
{
  
  gzFile f;
  int fd;
  
  
  if ( input_fn == NULL )
    {
      fd= dup ( fileno ( stdin ) );
      return gzdopen ( fd,"r" );
    }
  f= gzopen ( input_fn, mode_text?"rt":"rb" );
  if ( f == NULL )
    tL_error ( _("Could not open file '%s' for reading"), input_fn );
  
  return f;
  
} /* end open_input */


static void
close_input (
	     gzFile      f,
	     const char *fn
	     )
{
  gzclose ( f );
} /* end close_input */


static FILE *
open_output (
	     const char   *output_fn,
	     const tLBool  mode_text
	     )
{
  
  FILE *f;
  
  
  if ( output_fn == NULL ) return stdout;
  f= fopen ( output_fn, mode_text?"wt":"wb" );
  if ( f == NULL )
    tL_error ( _("Could not open file '%s' for writing"),
	       output_fn );
  
  return f;
  
} /* end open_output */


static void
close_output (
	      FILE *f
	      )
{
  if ( f != stdout ) fclose ( f );
} /* end close_output */


static void
format_amodel (
	       const struct opts *opts
	       )
{
  
  gzFile f;
  FILE *fout;
  tLDict *syms;
  tLAModel *amodel;
  char *err;
  
  
  if ( opts->verbose )
    {
      if ( opts->reverse )
	fprintf ( stderr, _("Converting AMODEL from text to binary format\n") );
      else
	fprintf ( stderr, _("Converting AMODEL from binary to text format\n") );
    }
  
  syms= tL_dict_new ();
  f= open_input ( opts->input_fn, opts->reverse!=0 );
  amodel= tL_amodel_new_from_file ( f, syms, &err );
  if ( amodel == NULL ) tL_error ( "AMODEL: %s", err );
  close_input ( f, opts->input_fn );
  fout= open_output ( opts->output_fn, opts->reverse==0 );
  tL_amodel_print ( amodel, fout, syms, (opts->reverse!=0) );
  close_output ( fout );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  if ( opts->verbose )
    fprintf ( stderr, _("Conversion complete\n") );
  
} /* end format_amodel */


static void
format_lmmodel (
	       const struct opts *opts
	       )
{
  
  gzFile f;
  FILE *fout;
  tLLexicon *lexicon;
  tLLM *lm;
  char *err;
  
  
  if ( opts->verbose )
    {
      if ( opts->reverse )
	fprintf ( stderr, _("Converting LM from text to binary format\n") );
      else
	fprintf ( stderr, _("Converting LM from binary to text format\n") );
    }
  
  lexicon= tL_lexicon_new ( TL_LEXICON_CONV );
  f= open_input ( opts->input_fn, opts->reverse!=0 );
  lm= tL_lm_new_from_file ( f, lexicon, TRUE, NULL, &err );
  if ( lm == NULL ) tL_error ( "LM: %s", err );
  close_input ( f, opts->input_fn );
  fout= open_output ( opts->output_fn, opts->reverse==0 );
  tL_lm_print ( lm, fout, lexicon->words, (opts->reverse!=0) );
  close_output ( fout );
  tL_lm_free ( lm );
  tL_lexicon_free ( lexicon );
  
  if ( opts->verbose )
    fprintf ( stderr, _("Conversion complete\n") );
  
} /* end format_amodel */


static void
format_pbm (
	    const struct opts *opts
	    )
{
  
  gzFile f;
  FILE *fout;
  tLFea *fea;
  char *err;
  
  if ( opts->verbose )
    {
      if ( opts->reverse )
	fprintf ( stderr,
		  _("Converting binary feature sequence"
		    " from text to binary format\n") );
      else
	fprintf ( stderr,
		  _("Converting binary feature sequence"
		    " from binary to text format\n") );
    }
  
  f= open_input ( opts->input_fn, opts->reverse!=0 );
  fea= tL_fea_new ( TL_FEA_BINARY );
  if ( tL_fea_load_fd ( fea, f, &err ) == -1 )
    tL_error ( _("An error occurred while reading the sample: %s"), err );
  close_input ( f, opts->input_fn );
  fout= open_output ( opts->output_fn, opts->reverse==0 );
  if ( tL_fea_write_fd ( fea, fout, opts->reverse, &err ) != 0 )
    tL_error ( _("An error occurred while writing the sample: %s"), err );
  close_output ( fout );
  tL_fea_free ( fea );
  
  if ( opts->verbose )
    fprintf ( stderr, _("Conversion complete\n") );
  
  return;
  
} /* end format_pbm */


static void
format_realfea (
		const struct opts *opts
		)
{
  
  gzFile f;
  FILE *fout;
  tLFea *fea;
  char *err;
  
  if ( opts->verbose )
    {
      if ( opts->reverse )
	fprintf ( stderr,
		  _("Converting real feature sequence"
		    " from text to binary format\n") );
      else
	fprintf ( stderr,
		  _("Converting real feature sequence"
		    " from binary to text format\n") );
    }
  
  f= open_input ( opts->input_fn, opts->reverse!=0 );
  fea= tL_fea_new ( TL_FEA_REAL );
  if ( tL_fea_load_fd ( fea, f, &err ) == -1 )
    tL_error ( _("An error occurred while reading the sample: %s"), err );
  close_input ( f, opts->input_fn );
  fout= open_output ( opts->output_fn, opts->reverse==0 );
  if ( tL_fea_write_fd ( fea, fout, opts->reverse, &err ) != 0 )
    tL_error ( _("An error occurred while writing the sample: %s"), err );
  close_output ( fout );
  tL_fea_free ( fea );
  
  if ( opts->verbose )
    fprintf ( stderr, _("Conversion complete\n") );
  
  return;
  
} /* end format_realfea */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct args args;
  struct opts opts;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( &argc, &argv, &args, &opts );
  
  if ( !strcmp ( args.format, "amodel" ) )
    format_amodel ( &opts );
  else if ( !strcmp ( args.format, "pbm" ) )
    format_pbm ( &opts );
  else if ( !strcmp ( args.format, "realfea" ) )
    format_realfea ( &opts );
  else if ( !strcmp ( args.format, "lm" ) )
    format_lmmodel ( &opts );
  else tL_error ( _("The format '%s' is unknown"), args.format );
  
  return EXIT_SUCCESS;
  
}
