/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  classify.c - Classificador de paraules.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */

#include <assert.h>
#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "get_lexicon.h"
#include "get_pbm_lst.h"
#include "get_priors.h"
#include "macros.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 3

#define UD(ptr) ((udata_t *) (ptr))




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  const char *output_fn;
  tLFloat     gsf;
  const char *lexicon_fn;
  int         nhyp;
  int         output_list;
  int         print_chars;
  
};


struct args
{
  
  const char *priors_fn;
  const char *amodel_fn;
  const char *pbmlst_fn;
  
};


typedef struct
{
  
  const tLDict *words;
  const tLDict *syms;
  const char   *file;
  
} udata_t;




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <priors> <amodel> <sample-lst>\n\n",
	    prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <priors> %s\n", _("Prior probability model") );
  fprintf ( to, "    <amodel> %s\n", _("Acoustic model") );
  fprintf ( to, "    <sample-lst> %s\n\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -g,--grammar-scale-factor <float>: %s\n",
	    _("Grammar Scale Factor. Default is 1.0") );
  fprintf ( to, "      -l,--lexicon <string>: %s\n",
	    _("File containing the lexicon."
	      " By default, it is created from <priors>,"
	      " assuming that each class is a UTF-8 word") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the result will be written") );
  fprintf ( to, "      -N,--number-hypotheses <int>: %s\n",
	    _("Number of hypotheses obtained per sample. Default is 1") );
  fprintf ( to, "      --output-is-list: %s\n",
	    _("The output file will be a list of file names."
	      " Each name will correspond to a test sample and"
	      " will identify the file to which the result will be written") );
  fprintf ( to, "      --print-syms: %s\n\n",
	    _("Print the transcription in characters instead of the class") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n;
  const char *prog, *gsf, *nhyp;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'g', "grammar-scale-factor", &gsf );
  add_cmdl_value ( cmdl, 'l', "lexicon", &(opts->lexicon_fn) );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 'N', "number-hypotheses", &nhyp );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "output-is-list",
		     &(opts->output_list) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "print-syms",
		     &(opts->print_chars) );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( gsf == NULL ) opts->gsf= 1.0;
  else
    {
      opts->gsf= atof ( gsf );
      if ( opts->gsf == 0.0 )
	tL_error ( _("The Grammar Scale Factor cannot be 0.0") );
    }
  if ( nhyp == NULL ) opts->nhyp= 1;
  else
    {
      opts->nhyp= atoi ( nhyp );
      if ( opts->nhyp < 1 )
	tL_error ( _("The number of hypotheses must be positive") );
    }
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->priors_fn= argv[0];
  args->amodel_fn= argv[1];
  args->pbmlst_fn= argv[2];
  
} /* end usage */


static void
callback (
	  void        *udata,
	  tLProb       prob_class,
	  tLProb       prob_lexicon,
	  tLProb       prob_sample,
	  size_t       class,
	  const tLSeq *cs
	  )
{
  
  fprintf ( stderr, "-> p('%s')[%g] * p('",
	    tL_dict_get_token ( UD(udata)->words, class), _DB(prob_class) );
  tL_seq_print ( cs, stderr, UD(udata)->syms, ' ' );
  fprintf ( stderr, "'|'%s')[%g] * p('%s'|'",
	    tL_dict_get_token ( UD(udata)->words, class ),
	    _DB(prob_lexicon), UD(udata)->file );
  tL_seq_print ( cs, stderr, UD(udata)->syms, ' ' );
  fprintf ( stderr, "')[%g] = [%g]\n", _DB(prob_sample),
	    _DB(prob_class+prob_lexicon+prob_sample) );
  
} /* end callback */


static void
classify_sample (
		 tLClassifier *cls,
		 const tLDict *syms,
		 tLFea        *fea,
		 FILE         *out,
		 const char   *pbm,
		 tLClassHyp   *hyps,
		 const int     nhyp,
		 const int     output_list,
		 const int     print_chars,
		 const int     verbose
		 )
{
  
  char *err;
  tLDict *labels;
  int i;
  
  
  /* Prepara. */
  if ( tL_fea_load ( fea, pbm, &err ) == -1 ) tL_error ( err );
  if ( !tL_amodel_test ( cls->amodel, fea, &err ) )
    {
      tL_warning ( _("The sample '%s' will not be classified: %s"), pbm, err );
      free ( err );
      return;
    }
  
  /* Classifica. */
  tL_classify ( cls, fea, hyps, nhyp );
  
  /* Imprimeix. */
  labels= cls->lexicon->words;
  if ( hyps[0].p == -HUGE_VAL )
    tL_warning ( _("'%s' could not be classified."
		 " It will be classified as '%s'"),
		 pbm, tL_dict_get_token ( labels, 0 ) );
  
  if ( nhyp == 1 )
    {
      if ( !output_list )
	fprintf ( out, "\"%s\": ", pbm );
      if ( print_chars )
	{
	  tL_seq_print ( hyps[0].cs, out, syms, ' ' );
	  putc ( '\n', out );
	}
      else fprintf ( out, "%s\n", tL_dict_get_token ( labels, hyps[0].c ) );
    }
  else
    {
      if ( !output_list )
	fprintf ( out, "%s", pbm );
      if ( print_chars )
	{
	  if ( !output_list ) putc ( '\n', out );
	  for ( i= 0; i < nhyp; ++i )
	    {
	      fprintf ( out, "%g ", _DB(hyps[i].p) );
	      if ( hyps[i].cs == NULL ) fprintf ( out, "none\n" );
	      else
		{
		  tL_seq_print ( hyps[i].cs, out, syms, ' ' );
		  putc ( '\n', out );
		}
	    }
	}
      else
	{
	  for ( i= 0; i < nhyp; ++i )
	    fprintf ( out, " %s %g",
		      tL_dict_get_token ( labels, hyps[i].c ),
		      _DB(hyps[i].p) );
	  putc ( '\n', out );
	}
    }
  
} /* end classify_sample */


static void
classify (
	  const tLPriors   *priors,
	  const tLLexicon  *lexicon,
	  const tLAModel   *amodel,
	  const tLDict     *syms,
	  const tLFileList *pbm_lst,
	  const char       *output,
	  const int         nhyp,
	  const int         output_list_op,
	  const int         print_chars,
	  const int         verbose
	  )
{
  
  size_t n;
  tLFea *fea;
  FILE *to;
  tLClassHyp *hyps;
  tLFileList *tolist;
  char *err;
  tLClassifier *cls;
  udata_t udata;
  int output_list;
  
  
  if ( verbose )
    fprintf ( stderr, _("Classifying\n") );
  
  output_list= output_list_op && output!=NULL;
  tolist= NULL; to= NULL; /* CALLA */
  
  /* Reserva memòria. */
  udata.words= lexicon->words;
  udata.syms= syms;
  udata.file= NULL;
  cls= tL_classifier_new ( amodel, lexicon, priors,
			   (verbose>1)?callback:NULL,
			   &udata, &err );
  if ( cls == NULL ) tL_error ( err );
  fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  if ( output == NULL ) to= stdout;
  else
    {
      if ( output_list )
	{
	  tolist= tL_filelist_new_from_file ( output );
	  if ( tolist == NULL )
	    tL_error ( _("Could not open '%s'"), output );
	  if ( tolist->N != pbm_lst->N )
	    tL_error ( _("The output file list contains %lu entries,"
			 " but the sample list contains %lu entries"),
			 tolist->N, pbm_lst->N );
	}
      else
	{
	  to= fopen ( output, "w" );
	  if ( to == NULL )
	    tL_error ( _("Could not open '%s'"), output );
	}
    }
  hyps= mem_alloc ( tLClassHyp, nhyp );
  
  /* Classifica. */
  for ( n= 0; n < pbm_lst->N; ++n )
    {
      if ( output_list )
	{
	  to= fopen ( tolist->files[n], "w" );
	  if ( to == NULL )
	    tL_error ( _("Could not open '%s'"), tolist->files[n] );
	}
      udata.file= pbm_lst->files[n];
      classify_sample ( cls, syms, fea, to,
			pbm_lst->files[n], hyps, nhyp, output_list,
			print_chars, verbose );
      if ( output_list ) fclose ( to );
    }
  
  /* Allibera memòria. */
  free ( hyps );
  if ( output_list ) tL_filelist_free ( tolist );
  else if ( to != stdout ) fclose ( to );
  tL_fea_free ( fea );
  tL_classifier_free ( cls );
  
} /* end classify */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel;
  tLPriors *priors;
  tLLexicon *lexicon;
  tLFileList *pbm_lst;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  lexicon= get_lexicon ( opts.lexicon_fn, syms, opts.verbose );
  assert ( lexicon->type == TL_LEXICON_CONV );
  priors= get_priors ( args.priors_fn, lexicon,
		       (opts.lexicon_fn == NULL),
		       syms, opts.verbose );
  if ( opts.gsf != 1.0  )
    tL_priors_set_gsf ( priors, lexicon->words->N, opts.gsf );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  
  /* Classifica. */
  classify ( priors, lexicon, amodel, syms, pbm_lst,
	     opts.output_fn, opts.nhyp, opts.output_list,
	     opts.print_chars, opts.verbose );
  
  /* Allibera memòria. */
  tL_filelist_free ( pbm_lst );
  tL_priors_free ( priors );
  tL_lexicon_free ( lexicon );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
