/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  cmllr.c - Adapta un model amb CMLLR.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "get_pbm_lst.h"
#include "get_words.h"
#include "macros.h"
#include "write_amodel.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 3

#define abs(A) (((A)<0.0) ? -(A) : A)




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int           verbose;
  const char   *output_fn;
  tLFloat       cfactor;
  unsigned int  max_iters;
  tLBool        nowords;
  int           skiplll;
  unsigned int  niters;
  
};


struct args
{
  
  const char *amodel_fn;
  const char *wdlst_fn;
  const char *pbmlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <amodel> <transc-lst> <sample-lst>\n\n",
	    prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Initial model") );
  fprintf ( to, "    <transc-lst> %s\n",
	    _("List of the files containing the transcriptions") );
  fprintf ( to, "    <sample-lst> %s\n\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -c,--cfactor <double>: %s\n",
	    _("Convergence factor. Default is 0.0001") );
  fprintf ( to, "      -m,--max-iters <uint>: %s\n",
	    _("Maximum number of iterations. Default is 20") );
  fprintf ( to, "      -M,--cmllr-iters <uint>: %s\n",
	    _("Number of iterations for the CMLLR algorithm."
	      " 0 means diagonal matrix computation. Default is 0") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the resulting matrices will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      --show-last-ll: %s\n",
	    _("Log probability will be shown for the last iteration") );
  fprintf ( to, "      --words: %s\n",
	    _("Transcriptions will be expected to be words") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, cwords, n, showlll;
  const char *prog, *cfactor, *max, *niters;


  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'c', "cfactor", &cfactor );
  add_cmdl_value ( cmdl, 'm', "max-iters", &max );
  add_cmdl_value ( cmdl, 'M', "cmllr-iters", &niters );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "show-last-ll", &showlll );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "words", &cwords );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( cfactor != NULL )
    {
      opts->cfactor= atof ( cfactor );
      if ( opts->cfactor <= 0.0 )
        tL_error ( _("The convergence factor must be greater than 0") );
    }
  else opts->cfactor= 0.0001;
  if ( max != NULL )
    {
      opts->max_iters= 
        (unsigned int) strtol ( max, (char **) NULL, 10 );
    }
  else opts->max_iters= 20;
  if ( niters != NULL )
    opts->niters= (unsigned int) strtol ( niters, (char **) NULL, 10 );
  else opts->niters= 0;
  opts->nowords= (cwords == 0);
  opts->skiplll= !showlll;
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->amodel_fn= argv[0];
  args->wdlst_fn= argv[1];
  args->pbmlst_fn= argv[2];
  
} /* end usage */


static tLProb
itertrain_sample (
		  tLCMLLR    *cmllr,
                  tLTrainer  *trainer,
                  size_t      seq,
                  const char *pbm,
                  size_t     *N,
                  tLFea      *fea,
		  tLFea      *fea_aux,
                  int         verbose
                  )
{
  
  char *err;
  tLProb prob;
  
  
  /* Prepara. */
  if ( tL_fea_load ( fea, pbm, &err ) == -1 )
    tL_error ( _("Error while reading '%s': %s"), pbm, err );
  if ( !tL_amodel_test ( trainer->amodel, fea, &err ) )
    {
      tL_warning ( _("The sample '%s' will not be used for training: %s"),
                   pbm, err );
      free ( err );
      return 0.0;
    }
  tL_cmllr_adapt_fea ( cmllr, fea, fea_aux );
  if ( verbose > 1 )
    fprintf ( stderr,
              _("Processing '%s'. Rows:%d, columns:%d"),
              pbm, fea->rows, fea->cols );
  
  /* Acumula. */
  prob= tL_train ( trainer, fea_aux, seq );
  if ( prob == -HUGE_VAL )
    {
      tL_warning ( _("The sample '%s' is too short"), pbm );
      return 0.0;
    }
  ++(*N);
  
  if ( verbose > 1 )
    fprintf ( stderr, "  lprob:%g\n", _DB(prob) );
  
  return prob;
  
} /* end itertrain_sample */


static tLProb
itertrain (
	   tLCMLLR    *cmllr,
           tLTrainer  *trainer,
           tLFea      *fea,
	   tLFea      *fea_aux,
           tLFileList *pbm_lst,
           int         verbose
           )
{
  
  tLProb lv;
  int n;
  size_t N_samples;
  
  
  /* Inicialitza. */
  tL_amlogcounts_init ( trainer->counts, trainer->amodel );
  if ( trainer->occs != NULL )
    tL_occupancies_init ( trainer->occs, trainer->amodel );
  lv= 0.0;
  
  N_samples= 0;
  for ( n= 0; n < trainer->seqs->N; ++n )
    lv+= itertrain_sample ( cmllr, trainer, trainer->seqs->seqs[n],
                            pbm_lst->files[n], &N_samples,
                            fea, fea_aux, verbose );
  lv+= tL_cmllr_get_lv_constant ( cmllr, trainer->counts );
  
  return (N_samples > 0) ? (lv / trainer->seqs->N) : -HUGE_VAL;
  
} /* end itertrain */


static void
cmllr_alg (
	   tLCMLLR            *cmllr,
	   tLTrainer          *trainer,
	   tLAMlogCounts      *counts,
	   tLSeqs             *words,
	   tLFileList         *pbm_lst,
	   unsigned int        max_iters,
	   tLFloat             cfactor,
	   int                 skiplll,
	   const tLDict       *syms,
	   const unsigned int  niters,
	   int                 verbose
	   )
{
  
  int i;
  tLProb lv, lvp;
  tLFloat inc;
  tLBool conv;
  tLFea *fea, *fea_aux;
  
  
  conv= FALSE; /* CALLA!! */
  
  if ( verbose )
    fprintf ( stderr, _("Starting CMLLR\n") );
  
  fea= tL_fea_new ( tL_amodel_get_fea_type ( cmllr->amodel ) );
  fea_aux= tL_fea_new ( tL_amodel_get_fea_type ( cmllr->amodel ) );
  inc= HUGE_VAL;
  i= 0;
  lv= itertrain ( cmllr, trainer, fea, fea_aux, pbm_lst, verbose );
  if ( verbose )
    {
      fprintf ( stderr, "  NºIter\t\tLL\t\tINC\n" );
      fprintf ( stderr, "%c %d\t\t%g\t\t%g\n",
		lv==0.0?'.':'*', i, _DB(lv), _DB(inc) );
    }
  if ( lv == 0 ) goto quit; /* No es toca el model. */
  while ( lv != -HUGE_VAL && i < max_iters && (i == 0 || !conv) )
    {
      
      /* Trainer té un punter al model que torna get_adapated_model. */
      tL_cmllr_update_matrix ( cmllr, counts, niters );
      ++i;
      if ( skiplll && i == max_iters ) break;
      
      lvp= lv;
      lv= itertrain ( cmllr, trainer, fea, fea_aux, pbm_lst, verbose );
      inc= (lv - lvp) / abs(lvp);
      conv= (inc < cfactor);
      if ( verbose )
        fprintf ( stderr, "%c %d\t\t%g\t\t%g\n",
                  conv ? '.' : '*',
                  i, _DB(lv), _DB(inc) );
      
    }
 quit:
  tL_fea_free ( fea );
  tL_fea_free ( fea_aux );
  
} /* end cmllr_alg */


static void
write_cmllr_matrix (
		    const char    *fn,
		    const tLCMLLR *cmllr
		    )
{
  
  FILE *f;


  if ( fn != NULL )
    {
      f= fopen ( fn, "w" );
      if ( f == NULL )
        tL_error ( _("Could not open file '%s'"), fn );
      tL_cmllr_print_matrix ( cmllr, f );
      fclose ( f );
    }
  else tL_cmllr_print_matrix ( cmllr, stdout );
  
} /* end write_cmllr_matrix */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel;
  tLFileList *pbm_lst;
  tLSeqs *words;
  tLAMlogCounts *counts;
  tLTrainer *trainer;
  tLCMLLR *cmllr;
  char *err;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  words= get_words ( args.wdlst_fn, syms, opts.nowords, NULL, opts.verbose );
  if ( words->N != pbm_lst->N )
    tL_error ( _("The sample list contains %lu entries,"
		 " but the word list contains %lu entries"),
               _PS(pbm_lst->N), _PS(words->N) );
  
  /* Cos. */
  cmllr= tL_cmllr_new ( amodel, &err );
  if ( cmllr == NULL ) tL_error ( err );
  counts= (opts.niters==0) ?
    tL_amlogcounts_new ( amodel ) : tL_amlogcounts_new_cmllr ( amodel );
  trainer= tL_trainer_new ( amodel, counts, words, NULL, &err );
  if ( trainer == NULL ) tL_error ( err );
  cmllr_alg ( cmllr, trainer, counts, words, pbm_lst,
	      opts.max_iters, opts.cfactor, opts.skiplll,
	      syms, opts.niters, opts.verbose );
  write_cmllr_matrix ( opts.output_fn, cmllr );
  
  /* Allibera memòria. */
  tL_trainer_free ( trainer );
  tL_amlogcounts_free ( counts, amodel );
  tL_cmllr_free ( cmllr );
  tL_seqs_free ( words );
  tL_filelist_free ( pbm_lst );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
