/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  cmllrfeas.c - A partir de la transcripció fonètica, un model i un
 *                conjunt de matrius de transformació CMLLR adapta les
 *                mostres d'entrada.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */

#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "get_pbm_lst.h"
#include "get_words.h"
#include "macros.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 4




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  
};


struct args
{
  
  const char *amodel_fn;
  const char *pbmlst_fn;
  const char *mats_fn;
  const char *outlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{

  fprintf ( to, "%s [%s] <amodel> <sample-lst> <matrix>"
	    " <out-lst>\n\n", prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Acoustic model") );
  fprintf ( to, "    <sample-lst> %s\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    <matrix> %s\n",
	    _("CMLLR transformation matrix") );
  fprintf ( to, "    <out-lst> %s\n\n",
	    _("List of the files to which"
	      " the transformed samples will be written") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n;
  const char *prog;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->amodel_fn= argv[0];
  args->pbmlst_fn= argv[1];
  args->mats_fn= argv[2];
  args->outlst_fn= argv[3];
  
} /* end usage */


static void
transform_sample (
		  tLCMLLR    *cmllr,
		  tLFea      *in_fea,
		  tLFea      *out_fea,
		  const char *pbm,
		  const char *out_fn,
		  int         verbose
		  )
{
  
  char *err;
  const tLAModel *amodel;
  
  
  /* Prepara. */
  amodel= cmllr->amodel;
  if ( tL_fea_load ( in_fea, pbm, &err ) == -1 ) tL_error ( err );
  if ( !tL_amodel_test ( amodel, in_fea, &err ) )
    tL_error ( _("The sample '%s' cannot be transformed: %s"), pbm, err );
  
  /* Transforma. */
  if ( verbose > 1 )
    fprintf ( stderr, _("Transforming the sample '%s'\n"), pbm );
  tL_cmllr_adapt_fea ( cmllr, in_fea, out_fea );
  if ( tL_fea_write ( out_fea, out_fn, TRUE, &err ) == -1 )
    tL_error ( err );
  
} /* end align_sample */


static void
transform (
	   tLAModel   *amodel,
	   tLCMLLR    *cmllr,
	   tLFileList *pbm_lst,
	   tLFileList *out_lst,
	   int         verbose
	   )
{
  
  tLFea *in_fea, *out_fea;
  size_t n;
  
  
  if ( verbose )
    fprintf ( stderr, _("Transforming\n") );
  
  /* Reserva memòria. */
  in_fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  out_fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  
  /* Transforma. */
  for ( n= 0; n < pbm_lst->N; ++n )
    transform_sample ( cmllr, in_fea, out_fea, pbm_lst->files[n],
		       out_lst->files[n], verbose );
  
  /* Allibera memòria. */
  tL_fea_free ( out_fea );
  tL_fea_free ( in_fea );
  
} /* end align */


static tLFileList *
get_out_lst (
             const char       *name,
	     const tLFileList *in_lst,
             const int         verbose
             )
{
  
  tLFileList *fl;
  size_t n;
  
  
  fl= tL_filelist_new_from_file ( name );
  if ( fl == NULL ) tL_error ( _("Could not open '%s'"), name );
  if ( fl->N != in_lst->N )
    tL_error ( _("The number of samples in '%s' is different from"
		 " the number of samples to be transformed") );
  if ( verbose )
    {
      fprintf ( stderr, _("The otput list has been read\n") );
      if ( verbose > 2 )
        {
          for ( n= 0; n < fl->N; ++n )
            fprintf ( stderr, "  - %lu: \"%s\"\n",
                      _PS(n), fl->files[n] );
        }
    }
  
  return fl;
  
} /* end get_out_lst */


static void
read_matrix (
	     const char    *fn,
	     const tLCMLLR *cmllr,
	     const int      verbose
	     )
{
  
  gzFile f;
  char *err;
  
  
  f= gzopen ( fn, "r" );
  if ( f == NULL ) tL_error ( _("Could not open '%s'"), fn );
  if ( tL_cmllr_read_matrix ( cmllr, f, &err ) == -1 )
    tL_error ( err );
  gzclose ( f );
  
  if ( verbose )
    {
      fprintf ( stderr,
		_("The CMLLR transformation matrices have been read\n") );
      if ( verbose > 2 )
	tL_cmllr_print_matrix ( cmllr, stderr );
    }
  
} /* end read_matrix */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel;
  tLFileList *pbm_lst, *out_lst;
  tLCMLLR *cmllr;
  char *err;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  out_lst= get_out_lst ( args.outlst_fn, pbm_lst, opts.verbose );
  cmllr= tL_cmllr_new ( amodel, &err );
  if ( cmllr == NULL ) tL_error ( err );
  read_matrix ( args.mats_fn, cmllr, opts.verbose );
  
  /* Cos. */
  transform ( amodel, cmllr, pbm_lst, out_lst, opts.verbose );
  
  /* Allibera memòria. */
  tL_cmllr_free ( cmllr );
  tL_filelist_free ( out_lst );
  tL_filelist_free ( pbm_lst );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
