/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  Miguel Ángel del Agua Teba - mdelagua@dsic.upv.es, madelagua@gmail.com
 *  Jesús Andrés Ferrer - jandres@dsic.upv.es
 *  Adrià Gimenez Pastor - agimenez@dsic.upv.es
 */

#include <cmath>
#include <cuda_runtime.h>
#include <curand_kernel.h>

#define TILE_WIDTH 16

#define IDX2C(i,j,ld) (((j)*(ld))+(i))

template<class T>
__global__ void AdaptationUpdateWithMomentum_kernel(T * adapt_matrix, T * update_matrix, T * old_update_matrix,
                                                    const T * gradient, const T weight_norm, const T bias_norm,
                                                    const int nhidden, const int nvisible, const T lr, const T mom) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if (i < nhidden && j < nvisible) {
    T adapt_matrix_value = adapt_matrix[IDX2C(i,j,nhidden)];
    T value = 0;

    if (j!=0)
      value = lr/weight_norm * gradient[IDX2C(i,j,nhidden)]
        + mom * old_update_matrix[IDX2C(i,j,nhidden)];
    else
      value = lr/bias_norm * gradient[IDX2C(i,j,nhidden)]
        + mom * old_update_matrix[IDX2C(i,j,nhidden)];

    update_matrix[IDX2C(i,j,nhidden)] = value;
    adapt_matrix[IDX2C(i,j,nhidden)] = adapt_matrix_value + value;
  }
}

template<class T>
__global__ void AdaptationUpdate_kernel(T * adapt_matrix, const T * gradient,
                                        const T weight_norm, const T bias_norm,
                                        int nhidden, int nvisible, T lr) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if (i < nhidden && j < nvisible) {
    T adapt_matrix_value = adapt_matrix[IDX2C(i,j,nhidden)];
    T value = 0;

    if (j!=0)
      value = lr/weight_norm * gradient[IDX2C(i,j,nhidden)];
    else
      value = lr/bias_norm * gradient[IDX2C(i,j,nhidden)];

    adapt_matrix[IDX2C(i,j,nhidden)] = adapt_matrix_value + value;
  }
}

template<class T>
__global__ void MergeDeltas_kernel(T * delta, T * dest, int sample_size, int num_samples,
                                   const int current_lang, const int * langs) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  //We must take into account the bias
  if (i < sample_size && j < num_samples)
    if (current_lang == langs[j])
      dest[IDX2C(i+1,j,sample_size+1)] = delta[IDX2C(i+1,j,sample_size+1)];
}

template<class T>
__global__ void Sigmoid_kernel(T * A, int rows, int columns) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if (i < rows && j < columns) {
    if (i == 0)
      A[j * rows + i] = 1;
    else
      A[j * rows + i] = 1.0 / (1.0 + expf(-A[j * (rows) + i]));
  }
}

template <class T>
__global__ void DerivativeOfSigmoid_kernel( T * __restrict__ delta, int rows, int columns, const T * __restrict__ Z ) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if( i < rows && j < columns ) {
    if ( i == 0 )
      delta[j*rows+i]=0;
    else {
      T localZ = Z[j*rows+i];
      delta[j*rows+i] = delta[j*rows+i] * ( localZ * ( 1.0f - localZ ) );
    }
  }
}

template <class T>
__global__ void RectifiedLinear_kernel( T * A, int rows, int columns ){
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if( i < rows && j < columns ){
    if ( i == 0 )
      A[j*rows+i] = 1;
    else{
      T localA = A[j*rows+i];
      A[j*rows+i] = localA < 0.0f ? 0.0f : localA;
    }
  }
}

template <class T>
__global__ void DerivativeOfRectifiedLinear_kernel( T * __restrict__ delta, int rows, int columns, const T * __restrict__ Z){
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if( i < rows && j < columns ){
    if ( i == 0 )
      delta[j*rows+i]=0;
    else
      delta[j*rows+i] = delta[j*rows+i] * (Z[j*rows+i] > 0.0f);
  }
}

template <class T>
__global__ void SoftRectifiedLinear_kernel( T * A, int rows, int columns ){
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if( i < rows && j < columns ){
    if ( i == 0 )
      A[j*rows+i] = 1;
    else{
      //A[j*rows+i] = log(1+exp(A[j*rows+i]));
      T localA = A[j*rows+i];
      //A[j*rows+i] = localA > 4.0f ? localA : log(1+expf(localA));
      A[j*rows+i] = localA > 4.0f ? localA : log1pf(expf(localA));
    }
  }
}

template <class T>
__global__ void DerivativeOfSoftRectifiedLinear_kernel(T * delta, int rows, int columns, const T * A, const T * Z) {
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if( i < rows && j < columns ){
    if ( i == 0 )
      delta[j*rows+i]=0;
    else{
      T localZ=Z[j*rows+i];
      T localA=A[j*rows+i];
      if (localZ <= 4.0f){
        const float f = __expf(localA);
        delta[j*rows+i] = delta[j*rows+i] * __fdividef(f, 1.0f + f);
      }
    }
  }
}

template <class T>
__global__ void ElementWiseProduct_kernel( T * __restrict__ A, int rows, int columns, const int * B ){
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  int j = blockIdx.y * blockDim.y + threadIdx.y;
  if( i < rows && j < columns && i > 0 ) //Excluding bias
    A[IDX2C(i,j,rows)]=A[IDX2C(i,j,rows)]*B[IDX2C(i,j,rows)];
}

template <class T>
__global__ void Substract_kernel( T * __restrict__ A, int rows, int columns, int sample, const T * W, const T * op1, const int * labels ){
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  //int j = blockIdx.y * blockDim.y + threadIdx.y;
  if( i < columns /*&& j == 0*/ )
    A[i]=W[IDX2C(labels[sample],i,rows)]-op1[i];
}

template<class T> 
void ElementWiseProduct_k(T * A, int rows, int columns, const int * B, cudaStream_t stream) {
  dim3 dimGrid(ceil((float) rows / TILE_WIDTH),
               ceil((float) columns / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  ElementWiseProduct_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( A, rows, columns, B );
}

template<class T> 
void Substract_k(T * A, int rows, int columns, int sample, const T * B, const T * C, const int * labels, cudaStream_t stream) {
  dim3 dimGrid(ceil((float) rows / TILE_WIDTH), 1
               /*ceil((float) columns / TILE_WIDTH)*/, 1);
  dim3 dimBlock(TILE_WIDTH*TILE_WIDTH, 1, 1);
  Substract_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( A, rows, columns, sample, B, C, labels );
}

template<class T> 
void Sigmoid_k(T * H, int num_samples, int dim, cudaStream_t stream) {
  dim3 dimGrid(ceil((float) num_samples / TILE_WIDTH),
               ceil((float) (dim) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  Sigmoid_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( H, num_samples, dim );
}

template<class T> 
void DerivativeOfSigmoid_k(T * A, int num_samples, int dim, T * B,
                           cudaStream_t stream) {
  dim3 dimGrid(ceil((float) num_samples / TILE_WIDTH),
               ceil((float) (dim) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  DerivativeOfSigmoid_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( A, num_samples, dim, B );
}

template<class T> 
void RectifiedLinear_k(T * H, int num_samples, int dim, cudaStream_t stream) {
  dim3 dimGrid(ceil((float) num_samples / TILE_WIDTH),
               ceil((float) (dim) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  RectifiedLinear_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( H, num_samples, dim );
}

template<class T> 
void DerivativeOfRectifiedLinear_k(T * A, int num_samples, int dim, T * B,
                                   cudaStream_t stream) {
  dim3 dimGrid(ceil((float) num_samples / TILE_WIDTH),
               ceil((float) (dim) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  DerivativeOfRectifiedLinear_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( A, num_samples, dim, B );
}

template<class T> 
void SoftRectifiedLinear_k(T * H, int num_samples, int dim, cudaStream_t stream) {
  dim3 dimGrid(ceil((float) num_samples / TILE_WIDTH),
               ceil((float) (dim) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  SoftRectifiedLinear_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( H, num_samples, dim );
}

template<class T> 
void DerivativeOfSoftRectifiedLinear_k(T * A, int num_samples, int dim, T * B, T * C,
                                       cudaStream_t stream) {
  dim3 dimGrid(ceil((float) num_samples / TILE_WIDTH),
               ceil((float) (dim) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  DerivativeOfSoftRectifiedLinear_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>( A, num_samples, dim, B, C );
}


template<class T> 
void MergeDeltas_k(T * delta, T * dest, int sample_size, int num_samples, const int current_lang,
                   const int * langs, cudaStream_t stream) {
  dim3 dimGrid(ceil((float) sample_size / TILE_WIDTH),
               ceil((float) (num_samples) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  MergeDeltas_kernel<T> <<< dimGrid, dimBlock, 0, stream>>>(delta, dest, sample_size, num_samples,
                                                            current_lang, langs);
}

template<class T>
void AdaptationUpdateWithMomentum_k(T * adapt_matrix, T * update_matrix, T * old_update_matrix,
                                    const T * gradient, const T weight_norm, const T bias_norm,
                                    const int nhidden, const int nvisible, const T lr, const T mom, cudaStream_t stream) {
  dim3 dimGrid(ceil((float) nhidden / TILE_WIDTH),
               ceil((float) (nvisible) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  AdaptationUpdateWithMomentum_kernel<T> <<< dimGrid, dimBlock, 0, stream >>>(adapt_matrix, update_matrix, old_update_matrix, gradient,
                                                                              weight_norm, bias_norm, nhidden, nvisible, lr, mom);
  cudaMemcpy(old_update_matrix, update_matrix, nhidden * nvisible * sizeof(T), cudaMemcpyDeviceToDevice);
}

template<class T> 
void AdaptationUpdate_k(T * adapt_matrix, const T * gradient, const T weight_norm, const T bias_norm,
                        int nhidden, int nvisible, T lr, cudaStream_t stream){
  dim3 dimGrid(ceil((float) nhidden / TILE_WIDTH),
               ceil((float) (nvisible) / TILE_WIDTH), 1);
  dim3 dimBlock(TILE_WIDTH, TILE_WIDTH, 1);
  AdaptationUpdate_kernel<T> <<< dimGrid, dimBlock, 0, stream >>>(adapt_matrix, gradient, weight_norm, bias_norm, nhidden, nvisible, lr);
}

template void ElementWiseProduct_k<double>(double * A, int rows, int columns, const int * B, cudaStream_t stream);
template void Substract_k<double>(double * A, int rows, int columns, int sample, const double * B, const double * C, const int * labels, cudaStream_t stream);
template void Sigmoid_k<double>(double * H, int sample_size, int num_samples,
                                cudaStream_t stream);
template void DerivativeOfSigmoid_k<double>(double * A, int sample_size,
                                            int num_samples, double * B, cudaStream_t stream);
template void RectifiedLinear_k<double>(double * H, int sample_size, int num_samples,
                                        cudaStream_t stream);
template void DerivativeOfRectifiedLinear_k<double>(double * A, int sample_size,
                                                    int num_samples, double * B, cudaStream_t stream);
template void SoftRectifiedLinear_k<double>(double * H, int sample_size, int num_samples,
                                            cudaStream_t stream);
template void DerivativeOfSoftRectifiedLinear_k<double>(double * A, int sample_size,
                                                        int num_samples, double * B, double * C, cudaStream_t stream);
template void MergeDeltas_k<double>(double * delta, double * dest, int sample_size,
                                    int num_samples, const int current_lang, const int * langs,
                                    cudaStream_t stream);
template void AdaptationUpdateWithMomentum_k<double>(double * adapt_matrix, double * update_matrix, double * old_update_matrix,
                                                     const double * gradient, const double weight_norm, const double bias_norm,
                                                     const int nhidden, const int nvisible, const double lr, const double mom, cudaStream_t stream);
template void AdaptationUpdate_k<double>(double * adapt_matrix, const double * gradient, const double weight_norm, const double bias_norm, int nhidden, int nvisible, double lr, cudaStream_t stream);

template void ElementWiseProduct_k<float>(float * A, int rows, int columns, const int * B, cudaStream_t stream);
template void Substract_k<float>(float * A, int rows, int columns, int sample, const float * B, const float * C, const int * labels, cudaStream_t stream);
template void Sigmoid_k<float>(float * H, int sample_size, int num_samples,
                               cudaStream_t stream);
template void DerivativeOfSigmoid_k<float>(float * A, int sample_size,
                                           int num_samples, float * B, cudaStream_t stream);
template void RectifiedLinear_k<float>(float * H, int sample_size, int num_samples,
                                       cudaStream_t stream);
template void DerivativeOfRectifiedLinear_k<float>(float * A, int sample_size,
                                                   int num_samples, float * B, cudaStream_t stream);
template void SoftRectifiedLinear_k<float>(float * H, int sample_size, int num_samples,
                                           cudaStream_t stream);
template void DerivativeOfSoftRectifiedLinear_k<float>(float * A, int sample_size,
                                                       int num_samples, float * B, float * C, cudaStream_t stream);
template void MergeDeltas_k<float>(float * delta, float * dest, int sample_size,
                                   int num_samples, const int current_lang, const int * langs,
                                   cudaStream_t stream);
template void AdaptationUpdateWithMomentum_k<float>(float * adapt_matrix, float * update_matrix, float * old_update_matrix,
                                                    const float * gradient, const float weight_norm, const float bias_norm,
                                                    const int nhidden, const int nvisible, const float lr, const float mom, cudaStream_t stream);
template void AdaptationUpdate_k<float>(float * adapt_matrix, const float * gradient, const float weight_norm, const float bias_norm, int nhidden, int nvisible, float lr, cudaStream_t stream);
