/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  Miguel Ángel del Agua Teba - mdelagua@dsic.upv.es, madelagua@gmail.com
 *  Jesús Andrés Ferrer - jandres@dsic.upv.es
 *  Adrià Giménez Pastor - agimenez@dsic.upv.es
 */

#include <random>
#include <iostream>
#include <cmath>
#include <fstream>
#include <sstream>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <string>
#include <algorithm>
#include <vector>
#include "./msg.h"
#include "../intl.h"

#ifdef GPU
#include "./mmathcublas.h"
#elif CBLAS
#include "./mmathcblas.h"
#else
#include "./mmath.h"
#endif

#include "./utils.h"

#ifndef DNN_H_
#define DNN_H_

#define IDX2R(i, j, ld) (((i)*(ld))+(j))
#define IDX2C(i, j, ld) (((j)*(ld))+(i))

/*
 * =============================================================================
 *        Class:  DNN
 *  Description:  Deep Neural Network
 * =============================================================================
 */
template<class T> class DNN {
  public:
  vector<vector<T>> weight_matrix_; ///< Weights matrices
  vector<vector<T>> update_matrix_; ///< Weights updates
  vector<vector<T>> old_update_matrix_; ///< Previous matrix weight update
  vector<vector<T>> gradient_matrix_; ///< Current gradient for each param
  vector<vector<T>> Z_matrix_; ///< Sigmoid outputs for each layer
  vector<vector<T>> A_matrix_; ///< Outputs for each layer
  vector<vector<T>> delta_matrix_; ///< Error estimations
  vector<int> units_per_layer_; ///< Number of units per layer. (last layer = sum over all langs)
  vector<int> output_units_per_lang_; ///< Number of output units per Softmax (multilingual)
  vector<int> output_parameter_offset_;
  int num_layers_; ///< Total number of layrs vis+hid
  my_random_engine generator_; ///< Global random number generator
  unsigned long long num_generator_calls_;
  int seed_;
  MMathInterface<T> * math_operations_; ///< Class of Math operations
  vector<vector<T>> priori_log_probs_;
  MESSAGE msg_; ///< Message provider
  kActivationFunction af_; ///< Activation function used through the network
  class DNNError { }; ///< Exception handling

  ///Constructor that initialises the NN from a vector containing the number
  ///of units per layer, and an integer indicating the seed for the random
  ///number generator
  DNN(const vector<int>& units_per_layer, const vector<int>& output_units,
      const kActivationFunction af, int seed);
  ///Constructor that initializes the NN from a DNN or RBMStack
  DNN(const string & file_name, int seed);
  ///Constructor that initializes the NN from a DNN and load a previous state
  DNN(const string & file_name, const string & state_file, int seed);
  ///Default destructor
  ~DNN() { delete math_operations_; if(indexes_ != NULL) delete[] indexes_; }
  ///Loads the weights from a previously trained DNN
  int LoadDNN(const string & in_file);
  ///Save to a file the current weights and biases
  int WriteToFile(ofstream & out_file);
  ///Save current network state to a file (Designed to continue the training in
  ///the future
  int WriteStateToFile(const string & file_name);
  int LoadStateFromFile(const string & file_name);

  ///Performs the adaptation of a Neural Network
  void Adapt(T * adaptation_data, const vector<sample<T>> & adapt_frames,
             int number_of_samples, T learning_rate, T mom, int epochs,
             int batch_size, int window_size, T * validation_data,
             const vector<sample<T>> & validation_frames,
             int number_of_validation_samples,
             bool likelihood_per_batch, bool likelihood_per_epoch, bool error_per_batch,
             bool error_per_epoch, int verbose);
  ///Performs classification of a given test set and returns the posterior
  ///log probabilities (plain=1)
  ///or a sequence of 0s and 1 to in the predicted class (plain=0)
  int Classify(T * test_data, int num_samples, vector<sample<T>> & frames_info,
               int batch_size, int window_size, kClassifierOutput clout,
               vector<string> & out_file_names);
  private:
  int * indexes_;
  int num_samples_;
  ///Initializes W matrices with a uniform distribution
  void InitializeWeightMatrixUniform();
  ///Initializes W matrices with a normal distribution
  void InitializeWeightMatrixNormal();
  ///Show info for debug
  void ShowInfo(int num_samples);
  ///Restores Z, A, delta_matrix_, update_matrix_ and gradient_matrix_ matrices to its default values
  void RestoreParams(int num_samples);
  ///Weights update
  void WeightsUpdate(T learning_rate, T momentum, T weight_decay);
};
/* -----  end of class DNN  ----- */

/* Constructors and Destructors */
template<class T>
DNN<T>::DNN(const vector<int>& units_per_layer, const vector<int>& output_units_per_lang,
            const kActivationFunction af, int seed) {
  af_ = af;
  seed_ = seed;
  generator_.seed(seed_);
  num_generator_calls_ = 0;
  indexes_ = NULL;

  int output_parameter_mem = 0;

  math_operations_ = new MMath<T>(&generator_);

  /* Consideraremos que units_per_layer tiene al principio la capa visible */
  units_per_layer_ = units_per_layer;
  output_units_per_lang_ = output_units_per_lang;
  num_layers_ = units_per_layer_.size();

  output_parameter_offset_.clear();
  output_parameter_offset_.push_back(0);
  //Take into account the bias
  for (int output_layer = 0; output_layer < output_units_per_lang_.size(); ++output_layer) {
    if (output_layer < output_units_per_lang.size()-1)
      output_parameter_offset_.push_back(output_parameter_offset_[output_parameter_offset_.size()-1] +
                                        (units_per_layer_[units_per_layer.size()-2]+1) *
                                        output_units_per_lang_[output_layer]);
    output_parameter_mem += output_units_per_lang_[output_layer] * (units_per_layer_[num_layers_-2]+1);
  }

  weight_matrix_.resize(num_layers_);
  update_matrix_.resize(num_layers_);
  old_update_matrix_.resize(num_layers_);
  gradient_matrix_.resize(num_layers_);
  A_matrix_.resize(num_layers_);
  Z_matrix_.resize(num_layers_);
  delta_matrix_.resize(num_layers_);

  /* No existirá un weight_matrix_[0] porque solo se tendrán aquellos que estén
   * asociados a una capa oculta */
  for (int ly = 1; ly < num_layers_; ++ly) {
    if (ly < (num_layers_-1)){
      weight_matrix_[ly].resize(
          units_per_layer_[ly] * (units_per_layer_[ly - 1] + 1));
      update_matrix_[ly].resize(
          units_per_layer_[ly] * (units_per_layer_[ly - 1] + 1));
      old_update_matrix_[ly].resize(
          units_per_layer_[ly] * (units_per_layer_[ly - 1] + 1));
      gradient_matrix_[ly].resize(
          units_per_layer_[ly] * (units_per_layer_[ly - 1] + 1));
    } else {
      weight_matrix_[ly].resize(output_parameter_mem);
      update_matrix_[ly].resize(output_parameter_mem);
      old_update_matrix_[ly].resize(output_parameter_mem);
      gradient_matrix_[ly].resize(output_parameter_mem);
    }
  }

  InitializeWeightMatrixUniform();

  //Initializing priori_log_probs
  priori_log_probs_.resize(output_units_per_lang_.size());
  for (int i = 0; i < output_units_per_lang_.size(); ++i ){
    priori_log_probs_[i].resize( output_units_per_lang_[i] );
    for (int j = 0; j < output_units_per_lang_[i]; ++j){
      priori_log_probs_[i][j] = 0;
    }
  }
}

template<class T>
DNN<T>::DNN(const string & file_name, int seed = 123) {
  /* The seed must be correctly initialized in case we need to train a
   * semi-defined net */
  /* semi-defined nets are those which are not fully set on its corresponding
   * file and thus contain some W matrices that require an initialization */
  /* The seed used for shuffling the samples will be that specified on the
   * network file if exists, in other case will be this seed */
  seed_ = seed;
  generator_.seed(seed_);
  num_generator_calls_ = 0;
  indexes_ = NULL;

  math_operations_ = new MMath<T>(&generator_);

  if (LoadDNN(file_name)) {
    msg_ << _("Loading DNN...") << MESSAGE::info;
  } else {
    msg_ << _("Could not load weights") << MESSAGE::error;
    throw DNNError();
  }

  //Initializing priori_log_probs
  /* priori_log_probs_.resize(output_units_per_lang_.size());
  for (int i = 0; i < output_units_per_lang_.size(); ++i ){
    priori_log_probs_.resize( output_units_per_lang_[i] );
    for (int j = 0; j < output_units_per_lang_[i]; ++j)
      priori_log_probs_[i][j] = 0;
  } */
}

template<class T>
DNN<T>::DNN(const string & file_name, const string & state_file, int seed = 123) {
  /* The seed must be correctly initialized in case we need to train a
   * semi-defined net */
  /* semi-defined nets are those which are not fully set on its corresponding
   * file and thus contain some W matrices that require an initialization */
  /* The seed used for shuffling the samples will be that specified on the
   * network file if exists, in other case will be this seed */
  seed_ = seed;
  generator_.seed(seed_);
  num_generator_calls_ = 0;
  indexes_ = NULL;

  math_operations_ = new MMath<T>(&generator_);

  if (LoadDNN(file_name)) {
    msg_ << _("Loading DNN...") << MESSAGE::info;
  } else {
    msg_ << _("Could not load weights") << MESSAGE::error;
    throw DNNError();
  }

  if (!LoadStateFromFile(state_file)){
    throw DNNError();
  }

  generator_.seed(seed_);
  generator_.discard(num_generator_calls_);
}

template<class T>
void DNN<T>::InitializeWeightMatrixUniform() {
  int ly;
  uniform_real_distribution<T> distribution(-0.1, 0.1);
  //for (int ly = 1; ly < units_per_layer_.size(); ++ly) {
    /* From "Understanding the difficulty of training deep
     * feedforward neural networks" */
    /* uniform_real_distribution<T>
     distribution( -sqrt(6)/sqrt(units_per_layer_[ly-1] + units_per_layer_[ly]),
     sqrt(6)/sqrt(units_per_layer_[ly-1] + units_per_layer_[ly]) ); */
   // for(typename vector<T>::iterator it = weight_matrix_[ly].begin();
    //    it != weight_matrix_[ly].end(); ++it)
    //  *it = distribution(generator_);
  //}
  for (ly = 1; ly < units_per_layer_.size()-1; ++ly)
    for (int i = 0; i < units_per_layer_[ly]; ++i)
      for (int j = 0; j < units_per_layer_[ly - 1] + 1; ++j)
        weight_matrix_[ly][IDX2C(i, j, units_per_layer_[ly])] = distribution(
            generator_);

  ly = units_per_layer_.size()-1;
  for (int lang = 0; lang < output_units_per_lang_.size(); ++lang)
    for (int i = 0; i < output_units_per_lang_[lang]; ++i)
      for (int j = 0; j < units_per_layer_[ly - 1] + 1; ++j)
        weight_matrix_[ly][IDX2C(i, j, output_units_per_lang_[lang])
            +output_parameter_offset_[lang]] = distribution(generator_);


}

template<class T>
void DNN<T>::InitializeWeightMatrixNormal() {
  normal_distribution<T> distribution(0, 1);
  for (int ly = 1; ly < units_per_layer_.size(); ++ly) {
    for(typename vector<T>::iterator it = weight_matrix_[ly].begin();
        it != weight_matrix_[ly].end(); ++it)
      *it = distribution(generator_);
  }
}

template<class T>
void DNN<T>::Adapt(T * adaptation_data, const vector<sample<T>> & adaptation_frames,
                   int num_samples, T learning_rate, T mom, int epochs,
                   int batch_size, int window_size, T * validation_data,
                   const vector<sample<T>> & validation_frames,
                   int num_validation_samples,
                   bool likelihood_per_batch, bool likelihood_per_epoch, bool error_per_batch,
                   bool error_per_epoch, int verbose){

  int e, num_batch, ly;
  int num_batches = num_samples / batch_size;
  int res = num_samples % batch_size;
  T likelihood_estimation = 0, likelihood_estimation_per_batch = 0;
  int last_data_matrix_size = 0;
  num_samples_ = num_samples;
  vector<T> gradient_matrix;
  gradient_matrix.resize((units_per_layer_[num_layers_-2]+1)*(units_per_layer_[num_layers_-2]+1));

  if (indexes_ == NULL){
    indexes_ = new int[num_samples_];
    for (int i = 0; i < num_samples_; ++i) indexes_[i] = i;
  }

  for (ly = 0; ly < num_layers_; ++ly) {
    if (ly == 0) {
      A_matrix_[ly].resize(
          batch_size * (units_per_layer_[ly] * window_size + 1));
      Z_matrix_[ly].resize(
          batch_size * (units_per_layer_[ly] * window_size + 1));
    } else if (ly < (num_layers_ - 1)) {
      A_matrix_[ly].resize(batch_size * (units_per_layer_[ly] + 1));
      Z_matrix_[ly].resize(batch_size * (units_per_layer_[ly] + 1));
    } else {
      for(int i = 0; i < output_units_per_lang_.size(); ++i)
        last_data_matrix_size += batch_size * (output_units_per_lang_[i] + 1);
      A_matrix_[ly].resize(last_data_matrix_size);
      Z_matrix_[ly].resize(last_data_matrix_size);
    }
  }

#if (defined GPU) || (defined GPULM) || (defined GPULMSTREAMS)
  math_operations_->Adapt(adaptation_data,
                          adaptation_frames,
                          indexes_,
                          num_samples,
                          learning_rate,
                          mom,
                          epochs,
                          batch_size,
                          window_size,
                          validation_data,
                          validation_frames,
                          num_validation_samples,
                          weight_matrix_, 
                          Z_matrix_, A_matrix_, 
                          gradient_matrix,
                          units_per_layer_,
                          output_units_per_lang_,
                          likelihood_per_batch,
                          likelihood_per_epoch,
                          error_per_batch,
                          error_per_epoch,
                          af_,
                          verbose);
#endif
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  DNN
 *      Method:  DNN :: restoreParams
 * Description:  Restaura a sus valores por defecto las matrices Z, A y Delta
 *--------------------------------------------------------------------------------------
 */
template<class T>
void DNN<T>::RestoreParams(int num_samples) {
  int ly, i, ns, j;

  /* Restauro valores de Zs y As y deltas */
  for (ly = 0; ly < num_layers_; ++ly) {
    for (i = 0; i <= units_per_layer_[ly]; ++i)
      for (ns = 0; ns < num_samples; ++ns) {
        Z_matrix_[ly][IDX2C(i, ns, units_per_layer_[ly]+1)] = 0;
        A_matrix_[ly][IDX2C(i, ns, units_per_layer_[ly]+1)] = 0;
      }
    for (i = 0; i <= units_per_layer_[ly]; ++i)
      for (ns = 0; ns < num_samples; ++ns)
        delta_matrix_[ly][IDX2C(i, ns, units_per_layer_[ly]+1)] = 0;
    if (ly > 0)
      for (i = 0; i < units_per_layer_[ly]; ++i)
        for (j = 0; j <= units_per_layer_[ly - 1]; ++j) {
          update_matrix_[ly][IDX2C(i, j, units_per_layer_[ly])] = 0;
          gradient_matrix_[ly][IDX2C(i, j, units_per_layer_[ly])] = 0;
        }
  }
}

template<class T>
int DNN<T>::Classify(T * data, int num_samples, vector<sample<T>> & frames,
                     int batch_size, int window_size,
                     kClassifierOutput clout, vector<string> & out_file_names) {
  int last_data_matrix_size = 0; int ly;
  for (ly = 0; ly < num_layers_; ++ly) {
    if (ly == 0) {
      A_matrix_[ly].resize(
          batch_size * (units_per_layer_[ly] * window_size + 1));
      Z_matrix_[ly].resize(
          batch_size * (units_per_layer_[ly] * window_size + 1));
    } else if (ly < (num_layers_ - 1)) {
      A_matrix_[ly].resize(batch_size * (units_per_layer_[ly] + 1));
      Z_matrix_[ly].resize(batch_size * (units_per_layer_[ly] + 1));
    } else {
      for(uint i = 0; i < output_units_per_lang_.size(); ++i)
        last_data_matrix_size += batch_size * (output_units_per_lang_[i] + 1);
      A_matrix_[ly].resize(last_data_matrix_size);
      Z_matrix_[ly].resize(last_data_matrix_size);
    }
  }

  return math_operations_->Classify(data, num_samples, frames, priori_log_probs_, batch_size,
                                    window_size, weight_matrix_, Z_matrix_, A_matrix_, units_per_layer_,
                                    output_units_per_lang_, af_, clout, out_file_names);
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  DNN
 *      Method:  DNN :: writeToFile
 * Description:  Writes the current network parameters to a file.
 *--------------------------------------------------------------------------------------
 */
template<class T>
int DNN<T>::WriteToFile(ofstream & out_file) {
  int i, j;

  if (out_file.fail())
    return 0;
  out_file << "DNN " << num_layers_ - 1 << endl;
  out_file.precision(10);
  for (int ly = 1; ly < num_layers_; ly++) {
    if (ly < (num_layers_-1)){
      out_file << "W " << units_per_layer_[ly] << " " << units_per_layer_[ly - 1]
          << endl;

      for (i = 0; i < units_per_layer_[ly]; ++i) { /* Top H */
        for (j = 0; j <= units_per_layer_[ly - 1]; ++j) /* Bot V */
          out_file << weight_matrix_[ly][IDX2C(i, j, units_per_layer_[ly])]
              << " ";
        out_file << endl;
      }
    } else {
      for(int lang = 0; lang < output_units_per_lang_.size(); ++lang ){
        if (lang == 0)
          out_file << "W " << output_units_per_lang_[lang] << " " << units_per_layer_[ly - 1]
              << endl;
        else
          out_file << "[W] " << output_units_per_lang_[lang] << " " << units_per_layer_[ly - 1]
              << endl;
        for (i = 0; i < output_units_per_lang_[lang]; ++i) { /* Top H */
          for (j = 0; j <= units_per_layer_[ly - 1]; ++j) /* Bot V */
            out_file << weight_matrix_[ly][IDX2C(i, j, output_units_per_lang_[lang])
                +output_parameter_offset_[lang]] << " ";
          out_file << endl;
        } 
      }
    }
  }
  return 1;
}

template<class T>
int DNN<T>::WriteStateToFile(const string & file_name){
  string header = "DNNSTATE";
  int layers = units_per_layer_.size();
  ofstream out_file;
  out_file.open(file_name, ios::out | ios::binary);
  
  if (!out_file.is_open()){
    msg_ << "Could not open for writing the file " << file_name << MESSAGE::error;
    return 0;
  }

  num_generator_calls_ = generator_.num_random_calls_;
  ///Saving:
  ///DNNSTATE SEED SEED_offset num_samples_ indexes[0]-indexes[num_samples-1]
  ///layers units_per_layer[0]-units_per_layer_[layers-1]
  ///old_update_matrix_[1]-old_update_matrix_[layers-1]

  out_file.write(header.c_str(), header.length());

  out_file.write(reinterpret_cast<char*>(&seed_), sizeof(int));
  out_file.write(reinterpret_cast<char*>(&num_generator_calls_), sizeof(unsigned long long));

  out_file.write(reinterpret_cast<char*>(&num_samples_), sizeof(int));
  for (int i = 0; i < num_samples_; ++i)
    out_file.write(reinterpret_cast<char*>(&indexes_[i]), sizeof(int));

  out_file.write(reinterpret_cast<char*>(&layers), sizeof(int));
  for (int ly = 0; ly < units_per_layer_.size(); ++ly)
    out_file.write(reinterpret_cast<char*>(&units_per_layer_[ly]), sizeof(int));

  for (int ly = 1; ly < units_per_layer_.size()-1; ++ly){
    for (int i = 0; i < units_per_layer_[ly]; ++i)
      for (int j = 0; j <= units_per_layer_[ly-1]; ++j)
        out_file.write(reinterpret_cast<char*>(&weight_matrix_[ly][IDX2C(i,j,units_per_layer_[ly])]),
                       sizeof(weight_matrix_[ly][0]));
  }
  for (int lang = 0; lang < output_units_per_lang_.size(); ++lang){
    for (int i = 0; i < output_units_per_lang_[lang]; ++i)
      for (int j = 0; j <= units_per_layer_[layers-2]; ++j)
        out_file.write(reinterpret_cast<char*>(&weight_matrix_[layers-1][IDX2C(i,j,output_units_per_lang_[lang])+
                                               output_parameter_offset_[lang]]), sizeof(weight_matrix_[1][0]));

  }

  for (int ly = 1; ly < units_per_layer_.size()-1; ++ly){
    for (int i = 0; i < units_per_layer_[ly]; ++i)
      for (int j = 0; j <= units_per_layer_[ly-1]; ++j)
        out_file.write(reinterpret_cast<char*>(&old_update_matrix_[ly][IDX2C(i,j,units_per_layer_[ly])]),
                       sizeof(old_update_matrix_[ly][0]));
  }
  for (int lang = 0; lang < output_units_per_lang_.size(); ++lang){
    for (int i = 0; i < output_units_per_lang_[lang]; ++i)
      for (int j = 0; j <= units_per_layer_[layers-2]; ++j)
        out_file.write(reinterpret_cast<char*>(&old_update_matrix_[layers-1][IDX2C(i,j,output_units_per_lang_[lang])+
                                               output_parameter_offset_[lang]]), sizeof(old_update_matrix_[1][0]));

  }

  out_file.close();
  return 1;
}

template<class T>
int DNN<T>::LoadStateFromFile(const string & file_name){
  string header = "DNNSTATE";
  int layers = units_per_layer_.size();
  char * memblock = new char[header.length()];
  ifstream in_file;
  in_file.open(file_name, ios::in | ios::binary);
  
  if (!in_file.is_open()){
    msg_ << _("The file") << " " << file_name << " " << _("could not be read") << MESSAGE::error;
    return 0;
  }
  ///Loading:
  ///DNNSTATE SEED SEED_offset num_samples_ indexes[0]-indexes[num_samples-1]
  ///layers units_per_layer[0]-units_per_layer_[layers-1]
  ///weight_matrix_[1]-weight_matrix_[layers-1]
  ///old_update_matrix_[1]-old_update_matrix_[layers-1]

  in_file.read(memblock, sizeof(char)*header.length());
  string * aux = new string(memblock, sizeof(char)*header.length());

  if ((*aux).compare(header)){
    msg_ << _("Could not load network state") << MESSAGE::error;
    return 0;
  }

  in_file.read((char*) &seed_, sizeof(int));
  in_file.read((char*) &num_generator_calls_, sizeof(unsigned long long));

  in_file.read((char*) &num_samples_, sizeof(int));
  indexes_ = new int [num_samples_];
  for (int i = 0; i < num_samples_; ++i)
    in_file.read((char*)(&indexes_[i]), sizeof(int));

  vector<int> units_per_layer;
  in_file.read((char*) &layers, sizeof(int));
  units_per_layer.resize(layers);
  for (int ly = 0; ly < units_per_layer.size(); ++ly){
    in_file.read((char*)&units_per_layer[ly], sizeof(int));
  }

  for (int ly = 1; ly < units_per_layer.size()-1; ++ly){
    for (int i = 0; i < units_per_layer[ly]; ++i)
      for (int j = 0; j <= units_per_layer[ly-1]; ++j)
        in_file.read((char*) &(weight_matrix_[ly][IDX2C(i,j,units_per_layer[ly])]),
                     sizeof(weight_matrix_[ly][0]));
  }
  for (int lang = 0; lang < output_units_per_lang_.size(); ++lang){
    for (int i = 0; i < output_units_per_lang_[lang]; ++i)
      for (int j = 0; j <= units_per_layer[layers-2]; ++j)
        in_file.read((char*)&(weight_matrix_[layers-1][IDX2C(i,j,output_units_per_lang_[lang])+
                              output_parameter_offset_[lang]]), sizeof(weight_matrix_[layers-1][0]));
  }

  for (int ly = 1; ly < units_per_layer.size()-1; ++ly){
    for (int i = 0; i < units_per_layer[ly]; ++i)
      for (int j = 0; j <= units_per_layer[ly-1]; ++j){
        in_file.read((char*) &(old_update_matrix_[ly][IDX2C(i,j,units_per_layer[ly])]),
                     sizeof(old_update_matrix_[ly][IDX2C(i,j,units_per_layer[ly])]));
      }
  }
  for (int lang = 0; lang < output_units_per_lang_.size(); ++lang){
    for (int i = 0; i < output_units_per_lang_[lang]; ++i)
      for (int j = 0; j <= units_per_layer[layers-2]; ++j)
        in_file.read((char*)&(old_update_matrix_[layers-1][IDX2C(i,j,output_units_per_lang_[lang])+
                              output_parameter_offset_[lang]]), sizeof(old_update_matrix_[layers-1][0]));
  }

  in_file.close();
  return 1;
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  DNN
 *      Method:  DNN :: LoadDNN
 * Description:  Initializes a new DNN based on the content of a given file
 *--------------------------------------------------------------------------------------
 */
template<class T>
int DNN<T>::LoadDNN(const string & file_name) {
  string line;
  string word;
  string delimiter = " ";
  vector<string> tokens;
  vector<string>::iterator it;
  stringstream ss;
  int i, j, nhidden, nvisible, wInd;
  //normal_distribution<T> distribution (0,1);
  uniform_real_distribution<T> distribution(-0.1, 0.1);
  ifstream in_file(file_name);
  bool reading_softmax = false;
  bool header_readed = false;
  bool read_priori = false;
  bool last_W_readed = false;
  int lang = 0;
  output_units_per_lang_.clear();
  output_parameter_offset_.clear();
  output_parameter_offset_.push_back(0);

  if (in_file.fail()) {
    msg_ << _("Loading DNN: Bad file descriptor") << MESSAGE::error;
    in_file.close();
    return 0;
  }

  if (in_file.is_open()) {
    wInd = 0;
    tokens.clear();
    getline(in_file, line);
    utils::Tokenize(line, tokens, delimiter);
    ss << tokens[0];
    ss >> word;
    ss.clear();
    if (tokens[0].compare("DNN") != 0) {
      in_file.close();
      return 0;
    }
    ss << tokens[1];
    ss >> num_layers_;
    ss.clear();
    num_layers_++;
    /* Anyadimos capa visible */
    weight_matrix_.resize(num_layers_);
    update_matrix_.resize(num_layers_);
    old_update_matrix_.resize(num_layers_);
    gradient_matrix_.resize(num_layers_);
    A_matrix_.resize(num_layers_);
    Z_matrix_.resize(num_layers_);
    delta_matrix_.resize(num_layers_);
    while (in_file.good() && wInd < num_layers_ - 1) {
      header_readed = false;
      /* ****** */
      /* Read W */
      /* ****** */
      tokens.clear();
      getline(in_file, line);
      if (line.length() == 0) {
        msg_ << _("The provided net does not meet the format: An empty line has been found")
             << MESSAGE::error;
        in_file.close();
        return 0;
      }
      utils::Tokenize(line, tokens, delimiter);
      ss << tokens[0];
      ss >> word;
      ss.clear();
      if (word.compare("W") != 0 && word.compare("[W]") != 0 && tokens.size() != 3) {
        msg_
            << _("The provided net does not meet the format: The expected keyword was W and found")
            << " " << tokens[0] << MESSAGE::error;
        in_file.close();
        return 0;
      }

      reading_softmax = false;
      if (word.compare("W") == 0){
        /* Reading W dimensions */
        ss << tokens[1];
        ss >> nhidden;
        ss.clear();
        ss << tokens[2];
        ss >> nvisible;
        ss.clear();
        if (wInd == 0) {
          units_per_layer_.push_back(nvisible);
          units_per_layer_.push_back(nhidden);
        } else {
          if (*(units_per_layer_.end() - 1) != nvisible) {
            msg_ << _("Problem loading network units: ");
            msg_
                << _("In the current W, the neural units does not match those specified in the previous W")
                << MESSAGE::error;
            in_file.close();
            return 0;
          }
          units_per_layer_.push_back(nhidden);
        }
        tokens.clear();

        weight_matrix_[wInd + 1].resize((nhidden) * (nvisible + 1));
        update_matrix_[wInd + 1].resize((nhidden) * (nvisible + 1));
        old_update_matrix_[wInd + 1].resize((nhidden) * (nvisible + 1));
        gradient_matrix_[wInd + 1].resize((nhidden) * (nvisible + 1));

        //memset( &old_update_matrix_[wInd+1], 0, nhidden*(nvisible+1)*sizeof(T) );
        for (int e = 0; e < nhidden * (nvisible + 1); ++e)
          old_update_matrix_[wInd + 1][e] = 0;

        header_readed = true;
        last_W_readed = true;
        reading_softmax = false;
      } else if (word.compare("[W]") == 0) {
        /* Reading W dimensions */
        reading_softmax = true;
        header_readed = true;
        lang += 1;
        ss << tokens[1];
        ss >> nhidden;
        ss.clear();
        ss << tokens[2];
        ss >> nvisible;
        ss.clear();
        if (units_per_layer_[units_per_layer_.size()-2] != nvisible) {
          msg_ << _("Problem loading network units: ");
          msg_
              << _("In [W], the neural units does not match those specified in the previous W")
              << MESSAGE::error;
          in_file.close();
          return 0;
        }

        int last_layer=units_per_layer_.size()-1;
        if (last_W_readed)
          output_units_per_lang_.push_back(units_per_layer_[last_layer]);

        units_per_layer_[last_layer] += nhidden;
        tokens.clear();


        output_units_per_lang_.push_back(nhidden);

        output_parameter_offset_.push_back(
            output_parameter_offset_[output_parameter_offset_.size()-1]+
            output_units_per_lang_[output_units_per_lang_.size()-2]*(units_per_layer_[last_layer-1]+1));

        weight_matrix_[weight_matrix_.size()-1].resize(
            weight_matrix_[weight_matrix_.size()-1].size() + nhidden * (units_per_layer_[last_layer-1]+1));
        update_matrix_[update_matrix_.size()-1].resize(
            update_matrix_[update_matrix_.size()-1].size() + nhidden * (units_per_layer_[last_layer-1]+1));
        old_update_matrix_[old_update_matrix_.size()-1].resize(
            old_update_matrix_[old_update_matrix_.size()-1].size() + nhidden * (units_per_layer_[last_layer-1]+1));
        gradient_matrix_[gradient_matrix_.size()-1].resize(
            gradient_matrix_[gradient_matrix_.size()-1].size() + nhidden * (units_per_layer_[last_layer-1]+1));

        for(typename vector<T>::iterator it = old_update_matrix_[old_update_matrix_.size()-1].begin();
            it != old_update_matrix_[old_update_matrix_.size()-1].end(); ++it)
          *it = 0;
        last_W_readed = false;
      }
      i = 0;
      while (in_file.good() && i < nhidden && header_readed) {
        int begin = in_file.tellg();
        getline(in_file, line);
        utils::Tokenize(line, tokens, delimiter);
        /* In case not a full net is specified -> W data will not be set */
        if (tokens[0].compare("LPriori") == 0 || tokens[0].compare("W") == 0 || tokens[0].compare("[W]") == 0) {
          /* Initialize the W matrix as it would for a new DNN */
          if ( ! reading_softmax ){
            for (typename vector<T>::iterator it = weight_matrix_[wInd+1].begin();
                 it != weight_matrix_[wInd+1].end(); ++it)
              *it = distribution(generator_);
            /* In case we have read W or LPriori, we must go back in order to read the header again and load its values */
            //if (tokens[0].compare("LPriori") == 0 || tokens[0].compare("W") == 0)
          } else {
            for (typename vector<T>::iterator it = weight_matrix_[weight_matrix_.size()-1].begin();
                 it != weight_matrix_[weight_matrix_.size()-1].end(); ++it)
              *it = distribution(generator_);
          }
          in_file.seekg(begin);
          break;
        /* } else if (tokens[0].compare("[W]") == 0) {
          for (typename vector<T>::iterator it = weight_matrix_[weight_matrix_.size()-1].begin();
               it != weight_matrix_[weight_matrix_.size()-1].end(); ++it)
            *it = distribution(generator_);
          if (tokens[0].compare("[W]") == 0)
            in_file.seekg(begin);
          break; */
        } else {
          j = 0;
          if( ! reading_softmax ) {
            for (it = tokens.begin(); it < tokens.end() && j <= nvisible; ++it, ++j) {
              ss << *it;
              ss >> weight_matrix_[wInd + 1][IDX2C(i, j, nhidden)];
              ss.clear();
            }
          }
          else {
            for (it = tokens.begin(); it < tokens.end() && j <= nvisible; ++it, ++j) {
              ss << *it;
              ss >> weight_matrix_[weight_matrix_.size()-1][IDX2C(i, j, output_units_per_lang_[lang])+output_parameter_offset_[lang]];
              ss.clear();
            }
          }
          ++i;
          tokens.clear();
        }
      }

      header_readed = false;

      /* ******************* */
      /* Reading prior probs */
      /* ******************* */

      int begin = in_file.tellg();
      getline(in_file, line);
      tokens.clear();
      utils::Tokenize(line, tokens, delimiter);
      ss << tokens[0];
      ss >> word;
      ss.clear();
      if (word.compare("LPriori") == 0){
        read_priori = true;
        priori_log_probs_.resize(lang+1);
      } else if ( word.compare("[W]") == 0){
        wInd--;
      }
      in_file.seekg(begin);
      
      if (read_priori) {
        if( output_units_per_lang_.size() == 0 )
          output_units_per_lang_.push_back(units_per_layer_[units_per_layer_.size()-1]);
        tokens.clear();
        getline(in_file, line);
        utils::Tokenize(line, tokens, delimiter);
        ss << tokens[0];
        ss >> word;
        ss.clear();
        if (word.compare("LPriori") != 0)
          msg_ << _("LPriori token could not be found, a priori probabilities will not be loaded")
              << MESSAGE::warning;
        while(in_file.is_open() && in_file.good() && (!in_file.eof()) && word.compare("LPriori") == 0) {
          if (tokens.size() > 1){
            ss << tokens[1];
            ss >> lang;
            ss.clear();
          } else {
            lang = 0;
          }
          tokens.clear();
          getline(in_file, line);
          utils::Tokenize(line, tokens, delimiter);
          j = 0;
          T aux;
          for (it = tokens.begin(); it < tokens.end() && j < nhidden; ++it, ++j) {
            if ((*it).compare("-inf") == 0) {
              msg_
                  << _("A -inf value has been found while reading a priori log probabilities")
                  << MESSAGE::error;
              in_file.close();
              return 0;
            }
            ss << *it;
            ss >> aux;
            ss.clear();
            priori_log_probs_[lang].push_back(aux);
          }
          tokens.clear();
          getline(in_file, line);
          utils::Tokenize(line, tokens, delimiter);
          ss << tokens[0];
          ss >> word;
          ss.clear();
        }
        if (in_file.is_open() && in_file.good() && (!in_file.eof())) {
          if (word.compare("TANH") == 0)
            af_ = TANH;
          else if (word.compare("RELU") == 0)
            af_ = RELU;
          else if (word.compare("SOFTRELU") == 0)
            af_ = SOFTRELU;
          else if (word.compare("SIGMOID") == 0)
            af_ = SIGMOID;
        }
        else {
          msg_ << 
              _("The activation function has not been specified. Therefore, SIGMOID will be used.") 
              << MESSAGE::warning;
          af_ = SIGMOID;
        }
        /* if (in_file.is_open() && in_file.good() && (!in_file.eof())) {
          tokens.clear();
          getline(in_file, line);
          if (line.length() > 0){
            utils::Tokenize(line, tokens, delimiter);
            ss << tokens[0];
            ss >> word;
            ss.clear();
            cout << "OJO: " << word << endl;
            if ((word.compare("SEED") != 0) || (word.compare("SEED") == 0 && tokens.size() != 3))
              msg_ << "Problem loading the network seed and its random number generator calls"
                  << MESSAGE::warning;
            else {
              ss << tokens[1];
              ss >> seed_;
              ss.clear();
              ss << tokens[2];
              ss >> num_generator_calls_;
              ss.clear();
            }
          } else {
            msg_ << "The seed has not been specified" << MESSAGE::warning;
          }
        } */
        tokens.clear();
      }
      wInd++;
    }
  }

  in_file.close();

  return 1;
}

template<class T>
void DNN<T>::ShowInfo(int num_samples) {
  int ly, i, j, ne;

  cout << "-------------------------------" << endl;
  for (ly = 0; ly < num_layers_; ++ly) {
    cout << _("---> Layer ") << ly << endl;

    if (num_samples > 0) {
      cout << "A:" << endl;
      for (i = 0; i <= units_per_layer_[ly]; ++i) {
        for (ne = 0; ne < num_samples; ++ne)
          cout << A_matrix_[ly][IDX2C(i, ne, units_per_layer_[ly]+1)] << " ";
        cout << endl;
      }

      cout << "Z:" << endl;
      for (i = 0; i <= units_per_layer_[ly]; ++i) {
        for (ne = 0; ne < num_samples; ++ne)
          cout << Z_matrix_[ly][IDX2C(i, ne, units_per_layer_[ly]+1)] << " ";
        cout << endl;
      }

      cout << "Delta: " << endl;
      for (i = 0; i <= units_per_layer_[ly]; ++i) {
        for (ne = 0; ne < num_samples; ++ne)
          cout << delta_matrix_[ly][IDX2C(i, ne, units_per_layer_[ly]+1)]
              << " ";
        cout << endl;
      }
    }

    if (ly > 0) {
      cout << _("Gradient: ") << endl;
      for (i = 0; i < units_per_layer_[ly]; ++i) {
        for (j = 0; j <= units_per_layer_[ly - 1]; ++j)
          cout << gradient_matrix_[ly][IDX2C(i, j, units_per_layer_[ly])]
              << " ";
        cout << endl;
      }
      cout << "W:" << endl;
      cout << units_per_layer_[ly - 1] << " x " << units_per_layer_[ly] << endl;
      for (i = 0; i < units_per_layer_[ly]; ++i) { /* Top H */
        for (j = 0; j <= units_per_layer_[ly - 1]; ++j) /* Bot V */
          cout << weight_matrix_[ly][IDX2C(i, j, units_per_layer_[ly])] << " ";
        cout << endl;
      }

      cout << "dW" << endl;
      for (i = 0; i < units_per_layer_[ly]; ++i) { /* Top H */
        for (j = 0; j <= units_per_layer_[ly - 1]; ++j) /* Bot V */
          cout << update_matrix_[ly][IDX2C(i, j , units_per_layer_[ly])] << " ";
        cout << endl;
      }
      cout << "old_update_matrix_" << endl;
      for (i = 0; i < units_per_layer_[ly]; ++i) { /* Top H */
        for (j = 0; j <= units_per_layer_[ly - 1]; ++j) /* Bot V */
          cout << old_update_matrix_[ly][IDX2C(i, j , units_per_layer_[ly])]
              << " ";
        cout << endl;
      }
    }
  }
  cout << "+++++++++++++++++++++++++++++++++" << endl;
  cout << endl;
}

#endif  // DNN_H_
