/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  Miguel Ángel del Agua Teba - mdelagua@dsic.upv.es, madelagua@gmail.com
 *  Jesús Andrés Ferrer - jandres@dsic.upv.es
 *  Adrià Giménez Pastor - agimenez@dsic.upv.es
 */

#include <iostream>
#include <fstream>
#include <vector>
#include <sstream>
#include <cmath>
#include "dnn.h"
#include "Options/options.h"
#include "msg.h"
#include "../intl.h"

#ifdef USE_DOUBLE
typedef double floatingpoint;
#else
typedef float floatingpoint;
#endif

using namespace std;

int main(int argc, char *argv[]) {
  vector<string> file_names;
  vector<string> file_output_names;
  int width = 0, height = 0, count = 0, num_samples = 0, batch_size;
  bool just_classify = false;
  stringstream ss;
  ofstream file;
  string f_list, f_labels, f_net, f_output, f_samples_list = "", f_sigmoid_list = "";
  vector<floatingpoint> content;
  vector<int> num_labels_per_lang;
  vector<int> num_samples_per_lang;
  DNN<floatingpoint> * d1;
  kFileType fT;
  JAF::Options prog(argv[0]);
  MESSAGE msg;
  kClassifierOutput clout = POSTERIOR_OUTPUT;
  int lang = 0;

  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );

  prog.addClass(_("Classification Options"), _("Options concerning classification"));
  prog.add<string>("samples", _("Path to the file containing the samples to classify"), f_list, true);
  prog.add<string>("labels", _("Path to the file containing the labels"), f_labels);
  prog.add<string>("net", _("Path to the net"), f_net, true);
  prog.add<int>("batch", _("Batch size used for classification"), batch_size, true);
  prog.add<string>("osamples", _("List containing the paths to the output samples"), f_samples_list);
  prog.add<string>("osigmoid", _("List containing the paths to the output sigmoids"), f_sigmoid_list);
  prog.addTrue("oclassify",
      _("If specified, the predicted labels and classification error will be output (labels are needed)"),
      just_classify);
  prog.add<int>("lang", _("Index of the softmax used to classify (0 by default). This parameter is not needed if just classifying (lang extracted from the labels file)."), lang, false);

  if (prog.read(argc, argv) < 0) {
    prog.syntax(cerr);
    return 1;
  }

  if (f_samples_list == "" and f_sigmoid_list == "" and just_classify == false) {
    msg << "-osamples, -oclassify " << _("or") << " -osigmoid " << _("must be specified") << MESSAGE::error;
    return 1;
  }

  if ( lang < 0 ){
    msg << _("Lang index must be greater than or equal to 0") << MESSAGE::error;
    return 1;
  }

  if (just_classify)
    clout = NONE_OUTPUT;
  else if (f_sigmoid_list != "")
    clout = SIGMOID_OUTPUT;
  else if (f_samples_list != "")
    clout = POSTERIOR_OUTPUT;
  else
    clout = NONE_OUTPUT;

  /* Reading output file list */
  if (clout == POSTERIOR_OUTPUT) {
    msg << _("Reading output file list ... ") << MESSAGE::info;
    if (!utils::ReadFileList(f_samples_list, file_output_names)) {
      msg << _("Unable to load output file list") << MESSAGE::error;
      return 1;
    }
  } else if (clout == SIGMOID_OUTPUT) {
    msg << _("Reading output file list ... ") << MESSAGE::info;
    if (!utils::ReadFileList(f_sigmoid_list, file_output_names)) {
      msg << _("Unable to load output file list") << MESSAGE::error;
      return 1;
    }
  }

  /* Reading file list */
  if (!utils::ReadFileList(f_list, file_names)) {
    msg << _("Unable to load file list") << MESSAGE::error;
    return 1;
  }

  /* Check files exists */
  if (!utils::CheckFileExists(file_names))
    return 1;

  if (clout == SIGMOID_OUTPUT or clout == POSTERIOR_OUTPUT) {
    if (file_names.size() != file_output_names.size()) {
      msg << _("The number of features to classify does not match the number of output file samples: ")
          << file_names.size() << _(" vs ") << file_output_names.size()
          << MESSAGE::error;
      return 1;
    }
  }

  if (!utils::ReadFileType(file_names[0], fT, height, width)) {
    msg << _("Unable to check file type") << MESSAGE::error;
    return 1;
  }

  vector<sample<floatingpoint>> frames;

  if (just_classify) {
    if (f_labels == "") {
      msg << _("The labels file must be specified") << MESSAGE::error;
      return 1;
    }
    msg << _("Reading labels...") << MESSAGE::info;
    if (!utils::ReadLabels(f_labels, frames, num_labels_per_lang, num_samples_per_lang )) {
      msg << _("Unable to load labels") << MESSAGE::error;
      return 1;
    }
    num_samples = frames.size();
  }

  msg << _("Reading samples...") << MESSAGE::info;

  content.resize(file_names.size() * height * width);

  if (fT == AKREALTF or fT == AKREALBF) {
    //READ AKREALTF
    ifstream file;
    vector<floatingpoint> auxContent;
    int num_samples_prev = 0, num_samples_aux = 0;
    int sample_size_aux = 0;
    content.clear();

    for (vector<string>::iterator it = file_names.begin(); it != file_names.end(); ++it) {
      file.open(*it);
      if (!utils::LoadRealFea(file, auxContent, num_samples_aux, sample_size_aux)) {
        msg << _("Could not read tlRealFea from classification samples")
            << MESSAGE::error;
        return 1;
      }

      for (int j = num_samples_prev; j < num_samples_prev + num_samples_aux; ++j)
        for (int i = 0; i < sample_size_aux; ++i)
          content.push_back(auxContent[IDX2R(j-num_samples_prev,i,sample_size_aux)]);

      if (num_samples == 0) //Si frames estaba vacia (no se leyeron las etiquetas...)
        frames.resize(num_samples_prev + num_samples_aux);

      for (int i = num_samples_prev; i < num_samples_prev + num_samples_aux; ++i) {
        frames[i].segmentLength = num_samples_aux;
        frames[i].position = i - num_samples_prev;
        frames[i].dim = sample_size_aux;
        frames[i].content = &content[IDX2C(0, i, sample_size_aux )];
        frames[i].segmentPosition = num_samples_prev;
        frames[i].fileName = *it;
        if (!just_classify)
          frames[i].lang = lang;
      }

      file.close();
      auxContent.clear();
      num_samples_prev += num_samples_aux;
    }

    height = 1;
    width = sample_size_aux;
  } else {
    msg << _("Incorrect sample file type for classification") << MESSAGE::error;
    return 1;
  }

  num_samples = frames.size();

  msg << _("Loading Neural Network...") << MESSAGE::info;
  /* During classification the seed it is not used, so it will be hardcoded to 0 */
  try {
    d1 = new DNN<floatingpoint>(f_net, 0);
  } catch (DNN<floatingpoint>::DNNError ) {
    msg << _("The DNN could not be correctly loaded") << MESSAGE::error;
    return 1;
  }

  if ((lang+1) > d1->output_units_per_lang_.size()) {
    msg << _("Lang index is greater than the available number of languages. It should be less than or equal to ")
        << d1->output_units_per_lang_.size() << MESSAGE::error;
    return 1;
  }

  /* if (d1->units_per_layer_[d1->units_per_layer_.size() - 1] < num_labels) {
    msg
        << "The number of output units can not be less than the number of validation labels"
        << MESSAGE::error;
    return 1;
  } */

  if (d1->units_per_layer_[0] % (height*width) != 0) {
    msg << _("The input dimension of the data does not match with the input layer dimension.") << MESSAGE::error;
    return 1;
  }

  msg << _("Window size: ") << d1->units_per_layer_[0]/(height*width) << MESSAGE::info;
  msg << _("Classifying...") << MESSAGE::info;
  
  if (!d1->Classify(content.data(), num_samples, frames, batch_size, 
                    d1->units_per_layer_[0]/(height * width), clout, file_output_names))
    msg << _("An error ocurred while classifying") << MESSAGE::error;

  file_names.clear();
  file_output_names.clear();
  frames.clear();
  content.clear();
  delete d1;

  return 0;
}
