/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  Miguel Ángel del Agua Teba - mdelagua@dsic.upv.es, madelagua@gmail.com
 *  Jesús Andrés Ferrer - jandres@dsic.upv.es
 *  Adrià Giménez Pastor - agimenez@dsic.upv.es
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits>
#include <string>
#include <vector>
#include <cmath>
#include <ctime>
#include "./mmathinterface.h"
#include "./cudautils.h"
#include "./devicememory.h"
#include "../intl.h"

#ifndef MMATH_H_
#define MMATH_H_

template<class T>
void ElementWiseProduct_k(T * A, int rows, int columns, const int * B, cudaStream_t stream);

template<class T>
void Substract_k(T * A, int rows, int columns, int sample, const T * B,
                 const T * C, const int * labels, cudaStream_t stream);
template<class T>
void Sigmoid_k(T * h, int sample_size, int num_samples, cudaStream_t stream);
template<class T>
void DerivativeOfSigmoid_k(T * A_matrix, int sample_size, int num_samples,
                           T * B, cudaStream_t stream);
template<class T>
void MergeDeltas_k(T * delta_matrix, T * dest_matrix, int sample_size, int num_samples,
                   const int current_lang, const int * langs, cudaStream_t stream);
template<class T>
void AdaptationUpdateWithMomentum_k(T * adapt_matrix, T * update_matrix, T * old_update_matrix,
                                    const T * gradient, const T weight_norm, const T bias_norm,
                                    const int nhidden, const int nvisible, const T lr, const T mom, cudaStream_t stream);
template<class T>
void AdaptationUpdate_k(T * adaptation_matrix, const T * gradient_matrix, const T weight_norm,
                        const T bias_norm,
                        int num_hidden, int num_visible, T lr, cudaStream_t stream);
template<class T>
void RectifiedLinear_k(T * h, int sample_size, int num_samples, cudaStream_t stream);
template<class T>
void DerivativeOfRectifiedLinear_k(T * A_matrix, int sample_size, int num_samples,
                                   T * B, cudaStream_t stream);
template<class T>
void SoftRectifiedLinear_k(T * h, int sample_size, int num_samples, cudaStream_t stream);
template<class T>
void DerivativeOfSoftRectifiedLinear_k(T * A_matrix, int sample_size, int num_samples,
                                       T * B, T * C, cudaStream_t stream);



template<class T>
class MMath: public MMathInterface<T> {
public:
  explicit MMath(my_random_engine * generator):MMathInterface<T>(generator) {}
  ~MMath() {}

  int Forward(vector<vector<T>> & Z_matrix, vector<vector<T>> & A_matrix,
              T * d_weight_matrix, T * d_Z_matrix, T * d_A_matrix, int *
              d_mask_matrix, T * d_data, int num_samples, T dropout, const
              vector<int> & parameter_offset, const vector<int> &
              data_offset, const vector<int> & output_parameter_offset,
              const vector<int> & output_data_offset, const vector<int> & units_per_layer,
              const vector<int> & output_units_per_lang,
              kActivationFunction ac_fun,
              cublasHandle_t * handle, cudaStream_t s0);

  int AdaptationForward (vector<vector<T>> & Z_matrix, vector<vector<T>> & A_matrix,
                         T * d_weight_matrix, T * d_Z_matrix,
                         T * d_A_matrix, const T * d_adapt_matrix, T * d_new_weight_matrix,
                         T * d_data, int num_samples,
                         const vector<int> & parameter_offset, const vector<int> & data_offset,
                         const vector<int> & output_parameter_offset,
                         const vector<int> & output_data_offset,
                         const vector<int> & units_per_layer, const vector<int> & output_units_per_lang,
                         kActivationFunction ac_fun, cublasHandle_t * handle, cudaStream_t s0);

  int AdaptationGradient (T * d_gradient, T * d_op1, T * d_op2, const vector<T> & Z, const T * d_W,
                          vector <T> & W, const T * d_input, vector<T> & input, 
                          const int * d_labels, const int num_samples, 
                          const vector<int> & units_per_layer, cublasHandle_t * handle,
                          cudaStream_t s0);

  int Adapt(const T * training_data,
            const vector<sample<T>> & training_frames, int * indexes, int num_samples,
            T lr, T mom, int epochs, int batch_size,
            int window_size, const T * validation,
            const vector<sample<T>> & validation_frames,
            int num_validation_samples, vector<vector<T>> & weight_matrix,
            vector<vector<T>> & Z_matrix,
            vector<vector<T>> & A_matrix,
            vector<T> & gradient_matrix,
            vector<int> & units_per_layer,
            vector<int> & output_units_per_lang, 
            bool likelihood_per_batch, bool likelihood_per_epoch,
            bool error_per_batch, bool error_per_epoch,
            kActivationFunction ac_fun,
            int verbose);

  int Classify(T * data, int num_samples, vector<sample<T>> & frames,
               vector<vector<T>> & priori_log_probs, int batch_size, int window_size,
               vector<vector<T>> & weight_matrix, vector<vector<T>> & Z_matrix,
               vector<vector<T>> & A_matrix, vector<int> & units_per_layer,
               vector<int> & output_units_per_lang,
               kActivationFunction ac_fun,
               kClassifierOutput clout, vector<string> & output_file_names);

  int LoadData(T * data,
               int num_samples,
               const vector<sample<T>> & frames,
               int * indexes,
               int num_batch,
               const vector<int> & data_offset,
               const vector<int> & units_per_layer,
               T * d_Z_matrix,
               vector<vector<T>> & Z_matrix,
               int * d_meta_data,
               cudaStream_t stream);

  int LoadData(T * data, int num_samples, const vector<int> & data_offset,
               const vector<int> & units_per_layer, T * d_Z_matrix,
               vector<vector<T>> & Z_matrix);

  int LoadData(T * data, int num_samples,
               const vector<int> & parameter_offset, const vector<int> & data_offset,
               const vector<int> & units_per_layer, T * d_Z_matrix,
               vector<vector<T>> & Z_matrix, T * d_weight_matrix,
               const vector<vector<T>> & weight_matrix, cudaStream_t stream);

  private:
  MESSAGE msg_;
  void ShowInfo(const vector<vector<T>> & weight_matrix,
                const vector<vector<T>> & update_matrix,
                const vector<vector<T>> & old_update_matrix,
                const vector<vector<T>> & Z_matrix,
                const vector<vector<T>> & A_matrix,
                const vector<vector<T>> & gradient_matrix,
                const vector<vector<T>> & delta_matrix,
                const vector<vector<int>> & mask_matrix,
                const vector<int> & units_per_layer,
                const vector<int> & output_units_per_lang,
                const vector<int> & output_parameter_offset,
                const vector<int> & output_data_offset,
                int num_samples);

  int MoveDataFromDeviceToHost(vector<vector<T>> & weight_matrix,
                               vector<vector<T>> & update_matrix,
                               vector<vector<T>> & old_update_matrix,
                               vector<vector<T>> & Z_matrix,
                               vector<vector<T>> & A_matrix,
                               vector<vector<T>> & gradient_matrix,
                               vector<vector<T>> & delta_matrix,
                               vector<vector<int>> & mask_matrix,
                               vector<int> & units_per_layer,
                               vector<int> & output_units_per_lang,
                               T * d_weight_matrix,
                               T * d_update_matrix,
                               T * d_old_update_matrix,
                               T * d_Z_matrix,
                               T * d_A_matrix,
                               T * d_gradient_matrix,
                               T * d_delta_matrix,
                               int * d_mask_matrix,
                               const vector<int> & parameter_offset,
                               const vector<int> & data_offset,
                               const vector<int> & output_parameter_offset,
                               const vector<int> & output_data_offset,
                               int batch_size);

  void CreateOffsets(int num_samples,
                     const vector<int> & units_per_layer,
                     const vector<int> & output_units_per_lang,
                     vector<int> & parameter_offset,
                     vector<int> & data_offset,
                     vector<int> & output_parameter_offset,
                     vector<int> & output_data_offset,
                     int * parameter_mem,
                     int * data_mem);

  T AdaptationLikelihood(const vector<vector<T>> & Z_matrix,
                         const vector<sample<T>> & frames,
                         const vector<int> & output_units_per_lang,
                         const vector<int> & output_data_offset,
                         int num_samples, const int * indexes, int num_batch);

  int AdaptationValidate(vector<vector<T>> & Z_matrix,
                         vector<vector<T>> & A_matrix, const T * validation_data,
                         const T * d_adapt_matrix, T * d_new_weight_matrix,
                         int num_validation_samples,
                         const vector<sample<T>> & validation_frames,
                         const vector<int> & parameter_offset,
                         const vector<int> & data_offset,
                         const vector<int> & output_parameter_offset,
                         const vector<int> & output_data_offset,
                         int batch_size, const int window_size,
                         const vector<int> & units_per_layer,
                         const vector<int> & output_units_per_lang,
                         T * d_weight_matrix, kActivationFunction ac_fun,
                         cublasHandle_t * handle, cudaStream_t s1);

  int ShowCudaInfo(T * d_V, int rows, int cols, int ly,
                   vector<int> & data_offset) {
    T * aux = new T[rows * cols];
    CUBLAS_CALL(
        cublasGetMatrix(rows, cols, sizeof(T), &d_V[data_offset[ly]], rows, &aux[0], rows));
    for (int i = 0; i < rows; ++i) {
      for (int j = 0; j < cols; ++j)
        cout << aux[IDX2C(i, j, rows)] << " ";
      cout << endl;
    }
    cout << endl;
    delete[] aux;
    return 1;
  }
};

/* *********************************************************************
 * *************** Deep Neural Networks implementation *****************
 * *********************************************************************/

template<class T>
int MMath<T>::AdaptationForward (vector<vector<T>> & Z_matrix, vector<vector<T>> & A_matrix,
                                 T * d_weight_matrix, T * d_Z_matrix,
                                 T * d_A_matrix, const T * d_adapt_matrix, T * d_new_weight_matrix,
                                 T * d_data, int num_samples,
                                 const vector<int> & parameter_offset, const vector<int> & data_offset,
                                 const vector<int> & output_parameter_offset, 
                                 const vector<int> & output_data_offset,
                                 const vector<int> & units_per_layer, const vector<int> & output_units_per_lang,
                                 kActivationFunction ac_fun, cublasHandle_t * handle, cudaStream_t s0) {
  /* Host */
  int i, ns, ly;
  const T kAlpha = 1;
  const T kBeta = 0;
  int num_langs = output_units_per_lang.size();

  CUBLAS_CALL(
      cublascopy(*handle, (units_per_layer[0]+1) * num_samples,
                 &d_data[0], 1, &d_Z_matrix[data_offset[0]], 1));

  for (ly = 1; ly < units_per_layer.size(); ++ly) {
    if (ly < (units_per_layer.size() - 1)) {
      CUBLAS_CALL(
          cublasgemm(*handle, CUBLAS_OP_N, CUBLAS_OP_N,
                     units_per_layer[ly], num_samples, units_per_layer[ly-1]+1, &kAlpha,
                     &d_weight_matrix[parameter_offset[ly]], units_per_layer[ly],
                     &d_Z_matrix[data_offset[ly-1]], units_per_layer[ly-1]+1,
                     &kBeta, &d_A_matrix[data_offset[ly]+1], units_per_layer[ly]+1));

      CUBLAS_CALL(
          cublascopy(*handle, (units_per_layer[ly]+1) * num_samples,
                     &d_A_matrix[data_offset[ly]], 1, &d_Z_matrix[data_offset[ly]], 1));

      if (ac_fun == SIGMOID)
        Sigmoid_k(&d_Z_matrix[data_offset[ly]], (units_per_layer[ly] + 1), num_samples, s0);
      else
        RectifiedLinear_k(&d_Z_matrix[data_offset[ly]], (units_per_layer[ly] + 1), num_samples, s0);

    } else {
      //Last Layer
      for (int lang = 0; lang < num_langs; ++lang) {
        CUBLAS_CALL(
            cublasgemm(*handle, CUBLAS_OP_N, CUBLAS_OP_N,
                       output_units_per_lang[lang], units_per_layer[ly-1]+1, units_per_layer[ly-1]+1, &kAlpha,
                       &d_weight_matrix[parameter_offset[ly]+output_parameter_offset[lang]],
                       output_units_per_lang[lang],
                       d_adapt_matrix, units_per_layer[ly-1]+1,
                       &kBeta, d_new_weight_matrix, output_units_per_lang[lang]));
       CUBLAS_CALL(
            cublasgemm(*handle, CUBLAS_OP_N, CUBLAS_OP_N,
                       output_units_per_lang[lang], num_samples, units_per_layer[ly-1]+1, &kAlpha,
                       d_new_weight_matrix,
                       output_units_per_lang[lang],
                       &d_Z_matrix[data_offset[ly-1]], units_per_layer[ly-1]+1,
                       &kBeta, &d_A_matrix[data_offset[ly]+1+output_data_offset[lang]],
                       output_units_per_lang[lang]+1));

        CUBLAS_CALL(
            cublasGetMatrix(output_units_per_lang[lang]+1, num_samples, sizeof(T),
                            &d_A_matrix[data_offset[ly]+output_data_offset[lang]], 
                            output_units_per_lang[lang]+1,
                            &(A_matrix[ly][output_data_offset[lang]]), output_units_per_lang[lang]+1));

        this->Softmax(&(A_matrix[ly][output_data_offset[lang]]), num_samples, output_units_per_lang[lang]+1,
                      &(Z_matrix[ly][output_data_offset[lang]]));

        for (ns = 0; ns < num_samples; ++ns)
          Z_matrix[ly][IDX2C(0, ns, output_units_per_lang[lang]+1) + output_data_offset[lang]] = 1;

        for (ns = 0; ns < num_samples; ++ns)
          for (i = 1; i < output_units_per_lang[lang]+1; ++i){
            Z_matrix[ly][IDX2C(i, ns, output_units_per_lang[lang]+1) + output_data_offset[lang]] =
                exp(Z_matrix[ly][IDX2C(i, ns, output_units_per_lang[lang]+1) + output_data_offset[lang]]);
          }

        CUBLAS_CALL(
            cublasSetMatrixAsync((output_units_per_lang[lang]+1), num_samples, sizeof(T),
                                 &Z_matrix[ly][output_data_offset[lang]], output_units_per_lang[lang]+1,
                                 &d_Z_matrix[data_offset[ly]+output_data_offset[lang]],
                                 output_units_per_lang[lang]+1, s0));
      }
    }
  }

  return 1;
}

template<class T>
int MMath<T>::AdaptationGradient (T * d_gradient, T * d_op1, T * d_op2, const vector<T> & Z,
                                  const T * d_W, vector<T> & W, const T * d_input, vector<T> & input, 
                                  const int * d_labels, const int num_samples, 
                                  const vector<int> & units_per_layer, cublasHandle_t * handle,
                                  cudaStream_t s0){

  int i, j, t, k, sum;
  int last_layer = units_per_layer.size()-1;
  const T kAlpha = 1;
  const T kBeta = 1;
  vector<T> op1; op1.resize(units_per_layer[last_layer-1]+1);
  vector<T> op2; op2.resize(units_per_layer[last_layer-1]+1);
  vector<int> labels; labels.resize(num_samples);
  

  CUDA_CALL(cudaMemset(d_gradient, 0, (units_per_layer[last_layer-1]+1)*
                       (units_per_layer[last_layer-1]+1) * sizeof(T)));
  CUDA_CALL(cudaMemset(d_op2, 0, (units_per_layer[last_layer-1]+1)*sizeof(T)));

  for (t = 0; t < num_samples; ++t){
    CUDA_CALL(cudaMemset(d_op1, 0, (units_per_layer[last_layer-1]+1)*sizeof(T)));
    for (j = 0; j < units_per_layer[last_layer]; ++j){
      if( Z[IDX2C(j+1, t, units_per_layer[last_layer]+1)] > 1e-4 )
        cublasSaxpy(*handle, units_per_layer[last_layer-1]+1, &Z[IDX2C(j+1, t, units_per_layer[last_layer]+1)],
                    &d_W[IDX2C(j, 0, units_per_layer[last_layer])], units_per_layer[last_layer],
                    d_op1, 1);
    }

    Substract_k(d_op2, units_per_layer[last_layer], units_per_layer[last_layer-1]+1,
                t, d_W, d_op1, d_labels, s0);

    CUBLAS_CALL(
        cublasgemm(*handle, CUBLAS_OP_N, CUBLAS_OP_T,
                   units_per_layer[last_layer-1]+1/*m*/, units_per_layer[last_layer-1]+1/*n*/, 1/*k*/,
                   &kAlpha, d_op2, units_per_layer[last_layer-1]+1,
                   &d_input[IDX2C(0, t, units_per_layer[last_layer-1]+1)], units_per_layer[last_layer-1]+1,
                   &kBeta, d_gradient, units_per_layer[last_layer-1]+1));
  }

  return 1;
}

template<class T>
int MMath<T>::Forward(vector<vector<T>> & Z_matrix, vector<vector<T>> & A_matrix, T * d_weight_matrix, 
                      T * d_Z_matrix,
                      T * d_A_matrix, int * d_mask_matrix, T * d_data, int num_samples, T dropout,
                      const vector<int> & parameter_offset, const vector<int> & data_offset,
                      const vector<int> & output_parameter_offset, const vector<int> & output_data_offset,
                      const vector<int> & units_per_layer, const vector<int> & output_units_per_lang,
                      kActivationFunction ac_fun, cublasHandle_t * handle,
                      cudaStream_t s0) {
  /* Host */
  int i, ns, ly;
  const T kAlpha = 1/(1-dropout);
  const T kBeta = 0;
  int num_langs = output_units_per_lang.size();

  CUBLAS_CALL(
      cublascopy(*handle, (units_per_layer[0]+1) * num_samples,
                 &d_data[0], 1, &d_Z_matrix[data_offset[0]], 1));

  //Applying dropout to the inputs (excluding bias)
  if( dropout != 0 )
    ElementWiseProduct_k(&d_Z_matrix[data_offset[0]], units_per_layer[0]+1, num_samples,
                         &d_mask_matrix[data_offset[0]], s0);

  for (ly = 1; ly < units_per_layer.size(); ++ly) {
    if (ly < (units_per_layer.size() - 1)) {
      CUBLAS_CALL(
          cublasgemm(*handle, CUBLAS_OP_N, CUBLAS_OP_N,
                     units_per_layer[ly], num_samples, units_per_layer[ly-1]+1, &kAlpha,
                     &d_weight_matrix[parameter_offset[ly]], units_per_layer[ly],
                     &d_Z_matrix[data_offset[ly-1]], units_per_layer[ly-1]+1,
                     &kBeta, &d_A_matrix[data_offset[ly]+1], units_per_layer[ly]+1));

      CUBLAS_CALL(
          cublascopy(*handle, (units_per_layer[ly]+1) * num_samples,
                     &d_A_matrix[data_offset[ly]], 1, &d_Z_matrix[data_offset[ly]], 1));

      //Applying dropout to the hidden layers (excluding bias)
      if( dropout != 0 )
        ElementWiseProduct_k(&d_Z_matrix[data_offset[ly]], units_per_layer[ly]+1,
                             num_samples, &d_mask_matrix[data_offset[ly]], s0); 

      if (ac_fun == SIGMOID)
        Sigmoid_k(&d_Z_matrix[data_offset[ly]], (units_per_layer[ly] + 1), num_samples, s0);
      else
        RectifiedLinear_k(&d_Z_matrix[data_offset[ly]], (units_per_layer[ly] + 1), num_samples, s0);
      //SoftRectifiedLinear_k(&d_Z_matrix[data_offset[ly]], (units_per_layer[ly] + 1), num_samples, s0);

    } else { 
      //Last Layer
      for (int lang = 0; lang < num_langs; ++lang) {
        CUBLAS_CALL(
            cublasgemm(*handle, CUBLAS_OP_N, CUBLAS_OP_N,
                       output_units_per_lang[lang], num_samples, units_per_layer[ly-1]+1, &kAlpha,
                       &d_weight_matrix[parameter_offset[ly]+output_parameter_offset[lang]],
                       output_units_per_lang[lang],
                       &d_Z_matrix[data_offset[ly-1]], units_per_layer[ly-1]+1,
                       &kBeta, &d_A_matrix[data_offset[ly]+1+output_data_offset[lang]],
                       output_units_per_lang[lang]+1));

        CUBLAS_CALL(
            cublasGetMatrix(output_units_per_lang[lang]+1, num_samples, sizeof(T),
                            &d_A_matrix[data_offset[ly]+output_data_offset[lang]], 
                            output_units_per_lang[lang]+1,
                            &(A_matrix[ly][output_data_offset[lang]]), output_units_per_lang[lang]+1));

        this->Softmax(&(A_matrix[ly][output_data_offset[lang]]), num_samples, output_units_per_lang[lang]+1,
                      &(Z_matrix[ly][output_data_offset[lang]]));

        for (ns = 0; ns < num_samples; ++ns)
          Z_matrix[ly][IDX2C(0, ns, output_units_per_lang[lang]+1) + output_data_offset[lang]] = 1;

        CUBLAS_CALL(
            cublasSetMatrixAsync((output_units_per_lang[lang]+1), num_samples, sizeof(T),
                                 &Z_matrix[ly][output_data_offset[lang]], output_units_per_lang[lang]+1,
                                 &d_Z_matrix[data_offset[ly]+output_data_offset[lang]],
                                 output_units_per_lang[lang]+1, s0));
      }
    }
  }

  return 1;
}

template<class T>
int MMath<T>::Classify(T * data, int num_samples, vector<sample<T>> & frames,
                       vector<vector<T>> & priori_log_probs, int batch_size,
                       int window_size, vector<vector<T>> & weight_matrix,
                       vector<vector<T>> & Z_matrix, vector<vector<T>> & A_matrix,
                       vector<int> & units_per_layer, vector<int> & output_units_per_lang,
                       kActivationFunction ac_fun, kClassifierOutput clout,
                       vector<string> & file_output_name) {

  int num_layers = units_per_layer.size();
  int * new_label = new int[batch_size];
  const T kDropout = 0;
  int * d_mask_matrix;

  T count, wrong;
  T * d_Z_matrix;
  T * d_A_matrix;
  T * d_weight_matrix;
  T * d_data;
  T * windowed_data;
  stringstream ss;
  string hypothesis, reference;
  cublasHandle_t handle;
  cudaStream_t s1;
  string current_file, previous_file = "";
  vector<vector<int>> mask_matrix;
  windowed_data = new T[batch_size * units_per_layer[0]];
  wrong = count = 0;
  int * indexes = new int[num_samples];
  for (int i = 0; i < num_samples; ++i)
    indexes[i] = i;

  vector<int> parameter_offset;
  vector<int> data_offset;
  vector<int> output_parameter_offset;
  vector<int> output_data_offset;

  int aux_memory; ///Memory needed to store one batch of Z, A, etc of the whole network
  int parameter_memory; ///Memory needed to store the parameters W, gW, etc of the whole network

  DeviceMemory<T> d_memory_T;
  DeviceMemory<int> d_memory_int;

  T max;
  int max_index;

  int original_batch_size = batch_size;
  int num_batches = num_samples / batch_size;
  int res = num_samples % batch_size;

  if (res > 0)
    num_batches++;

  CUDA_CALL(cudaStreamCreate(&s1));
  CUBLAS_CALL(cublasCreate(&handle));
  CUBLAS_CALL(cublasSetStream(handle, s1));

  CreateOffsets(batch_size, units_per_layer, output_units_per_lang, parameter_offset, data_offset,
                output_parameter_offset, output_data_offset, &parameter_memory, &aux_memory);

  d_memory_T.Add(DeviceElement<T>(&d_weight_matrix, parameter_memory, true));
  d_memory_T.Add(DeviceElement<T>(&d_data, batch_size*(units_per_layer[0]+1), false));
  d_memory_T.Add(DeviceElement<T>(&d_Z_matrix, aux_memory, true));
  d_memory_T.Add(DeviceElement<T>(&d_A_matrix, aux_memory, true));
  d_memory_int.Add(DeviceElement<int>(&d_mask_matrix, aux_memory, false));

  d_memory_T.Alloc();
  d_memory_int.Alloc();

  if (clout == NONE_OUTPUT) {
    cerr << _("# Hypothesis\tReference") << endl;
    cerr << "# ---------------------------------" << endl;
  }

  int segment_length = 0;
  int file_count = -1;
  ofstream out_file;

  for (int num_batch = 0; num_batch < num_batches; ++num_batch) {
    if ((num_batch == num_batches - 1) && res > 0) {
      batch_size = res;
      delete[] windowed_data;
      delete[] new_label;
      windowed_data = new T[batch_size * units_per_layer[0]];
      new_label = new int[batch_size];

      CreateOffsets(batch_size, units_per_layer, output_units_per_lang,
                    parameter_offset, data_offset,
                    output_parameter_offset, output_data_offset,
                    &parameter_memory, &aux_memory);

      d_memory_T.Dealloc();
      d_memory_int.Dealloc();
      d_memory_T.Add(DeviceElement<T>(&d_weight_matrix, parameter_memory, true));
      d_memory_T.Add(DeviceElement<T>(&d_data, batch_size*(units_per_layer[0]+1), false));
      d_memory_T.Add(DeviceElement<T>(&d_Z_matrix, aux_memory, true));
      d_memory_T.Add(DeviceElement<T>(&d_A_matrix, aux_memory, true));
      d_memory_int.Add(DeviceElement<int>(&d_mask_matrix, aux_memory, false));
      d_memory_T.Alloc();
      d_memory_int.Alloc();
    }

    utils::GenerateWindowedSamples<T>(data, frames, indexes,
                                      units_per_layer[0]/window_size,
                                      original_batch_size, batch_size,
                                      num_batch, window_size, windowed_data);

    LoadData(windowed_data, batch_size, parameter_offset, data_offset,
             units_per_layer, d_data, Z_matrix, d_weight_matrix, weight_matrix, s1);

    Forward(Z_matrix, A_matrix, d_weight_matrix, d_Z_matrix, d_A_matrix, d_mask_matrix,
            d_data, batch_size, kDropout, parameter_offset, data_offset,
            output_parameter_offset, output_data_offset, units_per_layer,
            output_units_per_lang, ac_fun, &handle, s1);

    /* Obtengo la ultima Z_matrix */
    if (clout == NONE_OUTPUT || clout == POSTERIOR_OUTPUT) {
      for (int lang = 0; lang < output_units_per_lang.size(); ++lang) { 
        CUBLAS_CALL(
            cublasGetMatrix(output_units_per_lang[lang]+1, batch_size, sizeof(T),
                            &d_Z_matrix[data_offset[num_layers-1]+output_data_offset[lang]],
                            output_units_per_lang[lang]+1,
                            &(Z_matrix[num_layers-1][output_data_offset[lang]]),
                            output_units_per_lang[lang]+1));
      }

      for (int ne = 0; ne < batch_size; ++ne) {
        int lang = frames[num_batch*original_batch_size+ne].lang;
        max = Z_matrix[num_layers - 1][IDX2C(1, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]];
        max_index = 0;
        for (int i = 2; i <=  output_units_per_lang[lang]; ++i)
          if (Z_matrix[num_layers - 1][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]]
              > max) {
            max =
                Z_matrix[num_layers - 1][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]];
            max_index = i - 1;
          }
        new_label[ne] = max_index;
      }

      for (int ne = 0; ne < batch_size; ++ne) {
        if (new_label[ne] != frames[num_batch * original_batch_size + ne].label)
          wrong++;
        count++;
      }

    } else if (clout == SIGMOID_OUTPUT) {
      CUBLAS_CALL(
          cublasGetMatrix(units_per_layer[num_layers-2]+1, batch_size, sizeof(T),
                          &d_Z_matrix[data_offset[num_layers-2]],
                          units_per_layer[num_layers-2]+1,
                          Z_matrix[num_layers-2].data(),
                          units_per_layer[num_layers-2]+1));
    }

    string header = "AKREALBF";

    /* *********************** */
    if (clout == NONE_OUTPUT) {
      /* *********************** */
      for (int ne = 0; ne < batch_size; ++ne)
        cerr << new_label[ne] << " "
            << frames[num_batch * original_batch_size + ne].label << endl;
    /* *********************** */
    } else if (clout == POSTERIOR_OUTPUT) {
      /* *********************** */
      for (int ne = num_batch * original_batch_size;
          ne < num_batch * original_batch_size + batch_size; ++ne) {
        int lang = frames[ne].lang;
        current_file = frames[ne].fileName;
        if (ne == num_batch * original_batch_size) {
          //Both samples belong to the same segment
          //It is suposed that segmentLength > 0
          if (current_file.compare(previous_file) == 0) {
            out_file.close();
            out_file.open(file_output_name[file_count],
                ios::out | ios::app | ios::binary);
          } else {
            file_count++;
            out_file.close();
            out_file.open(file_output_name[file_count]);

            if (!out_file.is_open()) {
              msg_ << _("Could not write file") << " " << file_output_name[file_count]
                  << MESSAGE::error;
              return 0;
            }
            segment_length = frames[ne].segmentLength;

            out_file.write(header.c_str(), header.length());
            out_file.write(reinterpret_cast<char*>(&output_units_per_lang[lang]), sizeof(int));
            out_file.write(reinterpret_cast<char*>(&frames[ne].segmentLength), sizeof(int));
            /* out_file << "AKREALTF";
             out_file << units_per_layer[num_layers-1] << " " << frames[ne].segmentLength << endl; */
          }
        } else {
          if (segment_length == 0) {
            file_count++;
            out_file.close();
            out_file.open(file_output_name[file_count]);

            if (!out_file.is_open()) {
              msg_ << _("Could not write file") << " " << file_output_name[file_count]
                  << MESSAGE::error;
              return 0;
            }
            segment_length = frames[ne].segmentLength;

            out_file.write(header.c_str(), header.length());
            out_file.write(reinterpret_cast<char*>(&output_units_per_lang[lang]), sizeof(int));
            out_file.write(reinterpret_cast<char*>(&frames[ne].segmentLength), sizeof(int));
            /* out_file << "AKREALTF" << endl;
             out_file << units_per_layer[num_layers-1] << " " << frames[ne].segmentLength << endl; */
          }
        }

        if (!out_file.is_open()) {
          msg_ << _("Could not append to file") << " " << MESSAGE::error;
          return 0;
        }

        float aux;
        lang = frames[ne].lang;
        for (int i = 1; i <= output_units_per_lang[lang]; ++i) {
          aux =
              Z_matrix[num_layers - 1][IDX2C(i, ne-num_batch*original_batch_size, output_units_per_lang[lang]+1) + output_data_offset[lang]]
              - priori_log_probs[lang][i - 1];
          out_file.write(reinterpret_cast<char*>(&aux), sizeof(float));
        }

        --segment_length;
        previous_file = current_file;
      }
      out_file.close();
    /* *********************** */
    } else if (clout == SIGMOID_OUTPUT) {
    /* *********************** */
      for (int ns = num_batch * original_batch_size;
          ns < num_batch * original_batch_size + batch_size; ++ns) {
        current_file = frames[ns].fileName;
        if (ns == num_batch * original_batch_size) {
          //Both samples belong to the same segment
          //It is suposed that segmentLength > 0
          if (current_file.compare(previous_file) == 0) {
            out_file.close();
            out_file.open(file_output_name[file_count], ios::out | ios::app);
          } else {
            file_count++;
            out_file.close();
            out_file.open(file_output_name[file_count]);

            if (!out_file.is_open()) {
              msg_ << _("Could not write file") << " " << file_output_name[file_count]
                  << MESSAGE::error;
              return 0;
            }
            segment_length = frames[ns].segmentLength;

            out_file.write(header.c_str(), header.length());
            out_file.write(reinterpret_cast<char*>(&units_per_layer[num_layers - 2]), sizeof(int));
            out_file.write(reinterpret_cast<char*>(&frames[ns].segmentLength), sizeof(int));
            /* out_file << "AKREALTF" << endl;
             out_file << units_per_layer[num_layers-2] << " " << frames[ns].segmentLength << endl; */
          }
        } else {
          if (segment_length == 0) {
            file_count++;
            out_file.close();
            out_file.open(file_output_name[file_count]);

            if (!out_file.is_open()) {
              msg_ << _("Could not write file") << " " << file_output_name[file_count]
                  << MESSAGE::error;
              return 0;
            }
            segment_length = frames[ns].segmentLength;

            out_file.write(header.c_str(), header.length());
            out_file.write(reinterpret_cast<char*>(&units_per_layer[num_layers - 2]), sizeof(int));
            out_file.write(reinterpret_cast<char*>(&frames[ns].segmentLength), sizeof(int));
            /* out_file << "AKREALTF" << endl;
             out_file << units_per_layer[num_layers-2] << " " << frames[ns].segmentLength << endl; */
          }
        }

        if (!out_file.is_open()) {
          msg_ << _("Could not append to file") << " " << MESSAGE::error;
          return 0;
        }

        float aux;
        for (int i = 1; i <= units_per_layer[num_layers - 2]; ++i) {
          aux =
              Z_matrix[num_layers - 2][IDX2C(i, ns-num_batch*original_batch_size, units_per_layer[num_layers-2]+1)];
          out_file.write(reinterpret_cast<char*>(&aux), sizeof(float));
        }

        /* for(int i = 1; i <= units_per_layer[num_layers-2]; ++i)
         out_file << Z_matrix[num_layers-2][IDX2C(i, ns-num_batch*original_batch_size, units_per_layer[num_layers-2]+1)] << " ";
         out_file << endl; */
        --segment_length;
        previous_file = current_file;
      }
      out_file.close();
    }
  }

  d_memory_T.Dealloc();
  d_memory_int.Dealloc();
  delete[] new_label;
  delete[] windowed_data;

  if (clout == NONE_OUTPUT) {
    cerr << " # " << _("Classification_Err=") << " " << wrong / count * 100 << " %" << endl;
    cerr << " # " << _("Total Errors=") << " " << wrong << ", " << _("Total Elements=") << " " << count << endl;
  }
  return 1;
}

template<class T>
int MMath<T>::Adapt(const T * training_data,
                    const vector<sample<T>> & training_frames, int * indexes, int num_samples,
                    T lr, T mom, int epochs, int batch_size,
                    int window_size, const T * validation,
                    const vector<sample<T>> & validation_frames,
                    int num_validation_samples, vector<vector<T>> & weight_matrix,
                    vector<vector<T>> & Z_matrix,
                    vector<vector<T>> & A_matrix,
                    vector<T> & gradient_matrix,
                    vector<int> & units_per_layer,
                    vector<int> & output_units_per_lang, 
                    bool likelihood_per_batch, bool likelihood_per_epoch,
                    bool error_per_batch, bool error_per_epoch,
                    kActivationFunction ac_fun,
                    int verbose) {
  const int kStreamsNumber = 4;
  cudaEvent_t event[kStreamsNumber];
  cudaStream_t stream[kStreamsNumber];
  for (int i = 0; i < kStreamsNumber; ++i) {
    CUDA_CALL(cudaStreamCreate(&stream[i]));
    CUDA_CALL(cudaEventCreateWithFlags(&event[i], cudaEventDisableTiming));
  }
  cublasHandle_t handle;
  int e, stream_index, batch_index_to_load;
  int num_batches = num_samples / batch_size;
  int res = num_samples % batch_size;
  int old_batch_size = batch_size;
  T ll = 0, llBatch = 0;

  int last_layer = units_per_layer.size() - 1;
  output_units_per_lang.clear();
  output_units_per_lang.push_back(units_per_layer[last_layer]);

  T * d_weight_matrix;
  T * d_new_weight_matrix;
  T * d_adapt_matrix;
  T * d_Z_matrix;
  T * d_A_matrix;
  T * d_gradient_matrix;
  T * d_update_matrix;
  T * d_old_update_matrix;
  T * d_delta_matrix;
  T * d_data; ///Used to store the data of several batches, as many as streams
  T * d_op1;
  T * d_op2;
  //T * d_l2norm;
  int * d_metadata; ///Used to store the labels of several batches, as many as streams
  int * d_mask_matrix;
  T weight_norm;
  T bias_norm;
  T * windowed_data;
  vector<int> parameter_offset;
  vector<int> data_offset;
  vector<int> output_parameter_offset;
  vector<int> output_data_offset;
  int aux_memory; ///Memory needed to store one batch of Z, A, etc of the whole network
  int parameter_memory; ///Memory needed to store the parameters W, gW, etc of the whole network
  DeviceMemory<T> d_memory_T;
  DeviceMemory<int> d_memory_int;
  vector<T> adapt_matrix;

  int num_langs = output_units_per_lang.size();

  CreateOffsets(batch_size, units_per_layer, output_units_per_lang,
                parameter_offset, data_offset,
                output_parameter_offset, output_data_offset,
                &parameter_memory, &aux_memory);

  d_memory_T.Add(DeviceElement<T>(&d_weight_matrix, parameter_memory, true));
  d_memory_T.Add(DeviceElement<T>(&d_new_weight_matrix,
                                  (units_per_layer[last_layer])*(units_per_layer[last_layer-1]+1), true));
  //Matrix for estimating the transformation of the last weight_matrix 
  d_memory_T.Add(DeviceElement<T>(&d_adapt_matrix,
                                  (units_per_layer[last_layer-1]+1)*(units_per_layer[last_layer-1]+1), true));
  //Gradient matrix that will try to maximize the objective function by
  //modifying the adaptation matrix
  d_memory_T.Add(DeviceElement<T>(&d_gradient_matrix,
                                  (units_per_layer[last_layer-1]+1)*(units_per_layer[last_layer-1]+1), true));
  d_memory_T.Add(DeviceElement<T>(&d_update_matrix,
                                  (units_per_layer[last_layer-1]+1)*(units_per_layer[last_layer-1]+1), true));
  d_memory_T.Add(DeviceElement<T>(&d_old_update_matrix,
                                  (units_per_layer[last_layer-1]+1)*(units_per_layer[last_layer-1]+1), true));
  d_memory_T.Add(DeviceElement<T>(&d_op1, units_per_layer[last_layer-1]+1, true));
  d_memory_T.Add(DeviceElement<T>(&d_op2, units_per_layer[last_layer-1]+1, true));
  //d_memory_T.Add(DeviceElement<T>(&d_l2norm, units_per_layer[last_layer-1]+1, true));
  d_memory_T.Add(DeviceElement<T>(&d_Z_matrix, aux_memory, true));
  d_memory_T.Add(DeviceElement<T>(&d_A_matrix, aux_memory, true));
  d_memory_T.Add(DeviceElement<T>(&d_delta_matrix, aux_memory, true));
  d_memory_T.Add(DeviceElement<T>(&d_data, kStreamsNumber*batch_size*(units_per_layer[0]+1), false));
  d_memory_int.Add(DeviceElement<int>(&d_metadata, kStreamsNumber*batch_size*COUNT, false));
  
  d_memory_T.Alloc();
  d_memory_int.Alloc();

  adapt_matrix.resize((units_per_layer[last_layer-1]+1)*(units_per_layer[last_layer-1]+1));
  for(int i = 0; i < units_per_layer[last_layer-1]+1; ++i)
    adapt_matrix[IDX2C(i,i,units_per_layer[last_layer-1]+1)] = 1;

  CUDA_CALL(
      cudaHostRegister(Z_matrix[0].data(),
                       batch_size*(units_per_layer[0]+1)*sizeof(T),
                       cudaHostRegisterPortable));

  CUDA_CALL(
      cudaHostAlloc(&windowed_data,
                    kStreamsNumber * batch_size * units_per_layer[0] * sizeof(T),
                    cudaHostAllocDefault));
  
  /* Copy W to device */
  for (int ly = 1; ly < units_per_layer.size(); ++ly) {
    if (ly < (units_per_layer.size()-1)){
      CUBLAS_CALL(
          cublasSetMatrixAsync(units_per_layer[ly], units_per_layer[ly-1]+1, sizeof(T),
                               weight_matrix[ly].data(), units_per_layer[ly],
                               &d_weight_matrix[parameter_offset[ly]],
                               units_per_layer[ly], stream[0]));
    } else {
      for (int lang = 0; lang < output_units_per_lang.size(); ++lang) {
        CUBLAS_CALL(
            cublasSetMatrixAsync(output_units_per_lang[lang], units_per_layer[ly-1]+1, sizeof(T),
                                 &(weight_matrix[ly][output_parameter_offset[lang]]), output_units_per_lang[lang],
                                 &d_weight_matrix[parameter_offset[ly]+output_parameter_offset[lang]],
                                 output_units_per_lang[lang], stream[0]));

      }
      CUBLAS_CALL(
          cublasSetMatrixAsync(units_per_layer[ly-1]+1, units_per_layer[ly-1]+1, sizeof(T),
                               adapt_matrix.data(), units_per_layer[ly-1]+1,
                               d_adapt_matrix, units_per_layer[ly-1]+1, stream[0]));
    }
  }

  /* Initialiasing CUBLAS */
  CUBLAS_CALL(cublasCreate(&handle));

  if ((num_batches == 0 && res > 0) || (num_batches == 1 && res == 0)) {
    num_batches = 1;
    res = 0;
    batch_size = num_samples;
  }

  for (e = 0; e < epochs; e++) {
    this->Shuffle(indexes, num_samples);
    ll = 0;
    CUBLAS_CALL(cublasSetStream(handle, stream[0]));

    for (int num_batch = 0; num_batch < num_batches; ++num_batch) {
      //batch_index_to_load = num_batch + kStreamsNumber - 1;
      batch_index_to_load = num_batch;
      stream_index = num_batch % kStreamsNumber;
      //Continue loading batches while the previous are processed
      if (batch_index_to_load < num_batches) {
        //cout << "Loading Batch: " << batch_index_to_load << " streamIndex: " << batch_index_to_load % kStreamsNumber << endl;
        utils::GenerateWindowedSamples<T>(training_data, training_frames,
                                          indexes, units_per_layer[0]/window_size,
                                          batch_size, batch_size, batch_index_to_load,
                                          window_size,
                                          &windowed_data[(batch_index_to_load % kStreamsNumber) * (units_per_layer[0] * batch_size)]);
        CUBLAS_CALL(cublasSetStream(handle, stream[batch_index_to_load%kStreamsNumber]));
        LoadData(
            &windowed_data[(batch_index_to_load % kStreamsNumber) * (units_per_layer[0] * batch_size)],
            batch_size, training_frames, indexes, batch_index_to_load, data_offset,
            units_per_layer,
            &d_data[(batch_index_to_load % kStreamsNumber) * ((units_per_layer[0] + 1) * batch_size)], Z_matrix,
            &d_metadata[(batch_index_to_load % kStreamsNumber) * batch_size * COUNT],
            stream[batch_index_to_load % kStreamsNumber]);
      }

      //Afecta a las funciones de CUBLAS
      CUBLAS_CALL(cublasSetStream(handle, stream[stream_index]));

      if (num_batch > 0)
        CUDA_CALL(cudaStreamWaitEvent(stream[stream_index], event[(num_batch-1)%kStreamsNumber], 0));

      AdaptationForward (Z_matrix, A_matrix, d_weight_matrix, d_Z_matrix,
                         d_A_matrix, d_adapt_matrix, d_new_weight_matrix,
                         &d_data[(num_batch % kStreamsNumber) * ((units_per_layer[0] + 1) * batch_size)],
                         batch_size, parameter_offset, data_offset,
                         output_parameter_offset, output_data_offset, units_per_layer, 
                         output_units_per_lang, ac_fun, &handle, stream[stream_index]);

      AdaptationGradient (d_gradient_matrix, d_op1, d_op2, Z_matrix[last_layer],
                          &d_weight_matrix[parameter_offset[last_layer]], weight_matrix[last_layer],
                          &d_Z_matrix[data_offset[last_layer-1]], Z_matrix[last_layer-1],
                          &d_metadata[(num_batch % kStreamsNumber) * batch_size * COUNT + batch_size * LABELS],
                          batch_size, units_per_layer, &handle, stream[stream_index]);

      cublasSnrm2(handle, units_per_layer[last_layer-1],
                  &d_gradient_matrix[IDX2C(0,0,units_per_layer[last_layer-1]+1)],
                  units_per_layer[last_layer-1]+1, &bias_norm);

      cublasSnrm2(handle, (units_per_layer[last_layer-1]+1)*(units_per_layer[last_layer-1]+1),
                  &d_gradient_matrix[IDX2C(0,0,units_per_layer[last_layer-1]+1)],
                  1, &weight_norm);

      weight_norm = sqrt(weight_norm*weight_norm - bias_norm*bias_norm);

      AdaptationUpdateWithMomentum_k(d_adapt_matrix, d_update_matrix, d_old_update_matrix,
                                     d_gradient_matrix, weight_norm, bias_norm, units_per_layer[last_layer-1]+1,
                                     units_per_layer[last_layer-1]+1, lr, mom, stream[stream_index]);

      CUDA_CALL(cudaEventRecord(event[stream_index], stream[stream_index]));

      if (likelihood_per_epoch || likelihood_per_batch) {
        llBatch = AdaptationLikelihood(Z_matrix, training_frames, output_units_per_lang,
                                       output_data_offset, batch_size, indexes, num_batch);
        ll += llBatch;
        if (likelihood_per_batch)
          cout << "\tBatch: " << num_batch << " LL: " << llBatch / batch_size << endl;
      }
      if (error_per_batch) {
        AdaptationValidate(Z_matrix, A_matrix, validation, d_adapt_matrix, d_new_weight_matrix,
                           num_validation_samples, validation_frames,
                           parameter_offset, data_offset, output_parameter_offset, output_data_offset,
                           old_batch_size, window_size,
                           units_per_layer, output_units_per_lang, d_weight_matrix, ac_fun,
                           &handle, stream[stream_index]);
        if (verbose) {
          cout << _("Training") << endl;
          AdaptationValidate(Z_matrix, A_matrix, training_data, d_adapt_matrix, d_new_weight_matrix,
                             num_samples, training_frames,
                             parameter_offset, data_offset, output_parameter_offset, output_data_offset,
                             old_batch_size, window_size,
                             units_per_layer, output_units_per_lang, d_weight_matrix, ac_fun,
                             &handle, stream[stream_index]);
        }
      }
    }
    //Finish all preceding operations on streams
    for (int i = 0; i < kStreamsNumber; ++i) CUDA_CALL(cudaEventSynchronize(event[i]));
    if (likelihood_per_epoch) {
      cout << "Epoch: " << e << " LL: " << ll / num_samples << " LR: " << lr
          << endl;
    } else {
      cout << "Epochs: " << e << endl;
    }
    if (error_per_epoch) {
      AdaptationValidate(Z_matrix, A_matrix, validation, d_adapt_matrix, d_new_weight_matrix,
                         num_validation_samples, validation_frames,
                         parameter_offset, data_offset, output_parameter_offset, output_data_offset,
                         old_batch_size, window_size,
                         units_per_layer, output_units_per_lang, d_weight_matrix, ac_fun,
                         &handle, stream[stream_index]);
      if (verbose) {
        cout << _("Training") << endl;
        AdaptationValidate(Z_matrix, A_matrix, training_data, d_adapt_matrix, d_new_weight_matrix,
                           num_samples, training_frames,
                           parameter_offset, data_offset, output_parameter_offset, output_data_offset,
                           old_batch_size, window_size,
                           units_per_layer, output_units_per_lang, d_weight_matrix, ac_fun,
                           &handle, stream[stream_index]);
      }
    }
  }

  for (unsigned ly = 1; ly < units_per_layer.size()-1; ++ly) {
    CUBLAS_CALL(
        cublasGetMatrix(units_per_layer[ly], units_per_layer[ly-1]+1,
                        sizeof(T), &d_weight_matrix[parameter_offset[ly]],
                        units_per_layer[ly], weight_matrix[ly].data(),
                        units_per_layer[ly]));
  }

  int lang = 0;
  const T kAlpha = 1;
  const T kBeta = 0;

  CUBLAS_CALL(
      cublasgemm(handle, CUBLAS_OP_N, CUBLAS_OP_N,
                 output_units_per_lang[lang], units_per_layer[last_layer-1]+1,
                 units_per_layer[last_layer-1]+1, &kAlpha,
                 &d_weight_matrix[parameter_offset[last_layer]+output_parameter_offset[lang]],
                 output_units_per_lang[lang],
                 d_adapt_matrix, units_per_layer[last_layer-1]+1,
                 &kBeta, d_new_weight_matrix, output_units_per_lang[lang]));

  CUBLAS_CALL(
      cublasGetMatrix(output_units_per_lang[lang], units_per_layer[last_layer-1]+1, sizeof(T), 
                      d_new_weight_matrix, output_units_per_lang[lang],
                      &(weight_matrix[last_layer][output_parameter_offset[lang]]),
                      output_units_per_lang[lang]));

  d_memory_T.Dealloc();
  d_memory_int.Dealloc();

  CUDA_CALL(cudaFreeHost(windowed_data));
  CUDA_CALL(cudaHostUnregister(Z_matrix[0].data()));

  for (int i = 0; i < kStreamsNumber; ++i) {
    CUDA_CALL(cudaStreamDestroy(stream[i]));
    CUDA_CALL(cudaEventDestroy(event[i]));
  }

  CUBLAS_CALL(cublasDestroy(handle));
  CUDA_CALL(cudaDeviceReset());

  return 1;
}


///Creates the parameter and data offsets needed to allocate memory sequentially
///on device memory. Moreover it returns the total memory required for
///the memory of a parameter and the memory of a batch
template<class T>
void MMath<T>::CreateOffsets(int num_samples,
                             const vector<int> & units_per_layer,
                             const vector<int> & output_units_per_lang,
                             vector<int> & parameter_offset,
                             vector<int> & data_offset,
                             vector<int> & output_parameter_offset,
                             vector<int> & output_data_offset,
                             int * parameter_mem,
                             int * data_mem ) {
  parameter_offset.clear();
  data_offset.clear();
  output_parameter_offset.clear();
  output_data_offset.clear();
  *parameter_mem = 0;
  *data_mem = 0;
  int output_parameter_mem = 0;
  int output_data_mem = 0;

  parameter_offset.push_back(0);
  parameter_offset.push_back(0);
  data_offset.push_back(0);
  output_parameter_offset.push_back(0); //Take into account the bias
  output_data_offset.push_back(0); //Take into account the bias

  //Offset estimated with bias
  for (int output_layer = 0; output_layer < output_units_per_lang.size(); ++output_layer) {
    if (output_layer < output_units_per_lang.size()-1) {
      output_data_offset.push_back(output_data_offset[output_data_offset.size()-1] +
                                   num_samples * (output_units_per_lang[output_layer]+1));
      output_parameter_offset.push_back(output_parameter_offset[output_parameter_offset.size()-1] +
                                        (units_per_layer[units_per_layer.size()-2]+1) *
                                        output_units_per_lang[output_layer]);
    }
    output_data_mem += num_samples * (output_units_per_lang[output_layer] + 1);
    output_parameter_mem += (units_per_layer[units_per_layer.size()-2]+1) *
        output_units_per_lang[output_layer];
  }

  for (int ly = 0; ly < units_per_layer.size(); ++ly) {
    if ( ly == 0 ) {
      *data_mem += num_samples * (units_per_layer[ly] + 1);
    } else if ( ly > 0 && ly < (units_per_layer.size()-1) ){
      data_offset.push_back(
          data_offset[data_offset.size()-1] + num_samples * (units_per_layer[ly-1] + 1));
      parameter_offset.push_back(
          parameter_offset[parameter_offset.size() - 1]
          + (units_per_layer[ly - 1] + 1) * (units_per_layer[ly]));
      *parameter_mem += (units_per_layer[ly - 1] + 1) * (units_per_layer[ly]);
      *data_mem += num_samples * (units_per_layer[ly] + 1);
    } else { //Last Layer
      data_offset.push_back(
          data_offset[data_offset.size()-1] + num_samples * (units_per_layer[ly-1] + 1));
      *parameter_mem += output_parameter_mem;
      *data_mem += output_data_mem;
    }
  }

  /* cout << "Output_data_offset" << endl;
  for(int i = 0; i < output_data_offset.size(); ++i)
    cout << output_data_offset[i] << " ";
  cout << endl;
  cout << "Output_data_mem: " << output_data_mem << endl;

  cout << "Output_parameter_offset" << endl;
  for(int i = 0; i < output_parameter_offset.size(); ++i)
    cout << output_parameter_offset[i] << " ";
  cout << endl;
  cout << "Output_parameter_mem: " << output_parameter_mem << endl;

  cout << "--------------" << endl;

  cout << "Data_offset: " << endl;
  for(int i = 0; i < data_offset.size(); ++i)
    cout << data_offset[i] << " ";
  cout << endl;
  cout << "data_mem: " << *data_mem << endl;

  cout << "Parameter_offset: " << endl;
  for(int i= 0; i < parameter_offset.size(); ++i)
    cout << parameter_offset[i] << " ";
  cout << endl;
  cout << "parameter_mem: " << *parameter_mem << endl; */
}

template<class T>
int MMath<T>::LoadData(T * data, int num_samples,
                       const vector<int> & data_offset, const vector<int> & units_per_layer,
                       T * d_Z_matrix, vector<vector<T>> & Z_matrix) {

  for (int ne = 0; ne < num_samples; ++ne)
    Z_matrix[0][IDX2C(0, ne, units_per_layer[0]+1)] = 1;

  for (int i = 1; i <= units_per_layer[0]; ++i)
    for (int ne = 0; ne < num_samples; ++ne)
      Z_matrix[0][IDX2C(i, ne, units_per_layer[0]+1)] =
          data[IDX2C(i-1, ne, units_per_layer[0])];

  CUBLAS_CALL(
      cublasSetMatrix(units_per_layer[0]+1, num_samples, sizeof(T),
                      Z_matrix[0].data(), units_per_layer[0]+1,
                      &d_Z_matrix[data_offset[0]], units_per_layer[0]+1));

  return 1;
}

template<class T>
int MMath<T>::LoadData(T * data, int num_samples,
                       const vector<int> & parameter_offset, const vector<int> & data_offset,
                       const vector<int> & units_per_layer, T * d_Z_matrix,
                       vector<vector<T>> & Z_matrix, T * d_weight_matrix,
                       const vector<vector<T>> & weight_matrix, cudaStream_t stream) {
  for (int ne = 0; ne < num_samples; ++ne)
    Z_matrix[0][IDX2C(0, ne, units_per_layer[0]+1)] = 1;

  for (int i = 1; i <= units_per_layer[0]; ++i)
    for (int ne = 0; ne < num_samples; ++ne)
      Z_matrix[0][IDX2C(i, ne, units_per_layer[0]+1)] =
          data[IDX2C(i-1, ne, units_per_layer[0])];

  CUBLAS_CALL(
      cublasSetMatrixAsync(units_per_layer[0]+1, num_samples, sizeof(T),
                           Z_matrix[0].data(), units_per_layer[0]+1,
                           &d_Z_matrix[data_offset[0]], units_per_layer[0]+1,
                           stream));

  for (unsigned ly = 1; ly < units_per_layer.size(); ++ly) {
    CUBLAS_CALL(
        cublasSetMatrixAsync(units_per_layer[ly], units_per_layer[ly-1]+1,
                             sizeof(T), weight_matrix[ly].data(),
                             units_per_layer[ly],
                             &d_weight_matrix[parameter_offset[ly]],
                             units_per_layer[ly], stream));
  }

  return 1;
}

template<class T>
int MMath<T>::LoadData(T * data, int num_samples,
                       const vector<sample<T>> & frames, int * indexes, int num_batch,
                       const vector<int> & data_offset, const vector<int> & units_per_layer,
                       T * d_data_matrix, vector<vector<T>> & Z_matrix, int * d_metadata,
                       cudaStream_t stream) {

  int * metadata = new int[num_samples * COUNT];

  for (int i = 0; i < num_samples; ++i) {
    metadata[LABELS * num_samples + i] =
        frames[indexes[num_samples * num_batch + i]].label;
    metadata[LANGS * num_samples + i] =
        frames[indexes[num_samples * num_batch + i]].lang;
  }

  for (int ne = 0; ne < num_samples; ++ne)
    Z_matrix[0][IDX2C(0, ne, units_per_layer[0]+1)] = 1;

  for (int i = 1; i <= units_per_layer[0]; ++i)
    for (int ne = 0; ne < num_samples; ++ne)
      Z_matrix[0][IDX2C(i, ne, units_per_layer[0]+1)] =
          data[IDX2C(i-1, ne, units_per_layer[0])];

  CUBLAS_CALL(
      cublasSetMatrixAsync(units_per_layer[0]+1, num_samples, sizeof(T),
                           Z_matrix[0].data(), units_per_layer[0]+1,
                           &d_data_matrix[data_offset[0]], units_per_layer[0]+1,
                           stream));

  CUBLAS_CALL(
      cublasSetMatrixAsync(1, num_samples*COUNT, sizeof(int),
                           metadata, 1, d_metadata, 1, stream));

  delete[] metadata;

  return 1;
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  DNN
 *      Method:  DNN :: Likelihood
 * Description:  Estimates the likelihood of the current minibatch. It is suposed that
 * the last layer has provided an output, which is necessary for this method.
 *--------------------------------------------------------------------------------------
 */

template<class T>
T MMath<T>::AdaptationLikelihood(const vector<vector<T>> & Z_matrix,
                                 const vector<sample<T>> & frames,
                                 const vector<int> & output_units_per_lang,
                                 const vector<int> & output_data_offset,
                                 int num_samples, const int * indexes, int num_batch) {
  T res = 0;
  int index;

  for (int ne = 0; ne < num_samples; ++ne) {
    index = indexes[num_samples * num_batch + ne];
    res +=
        log(Z_matrix[Z_matrix.size()-1][IDX2C(frames[index].label+1, ne, output_units_per_lang[frames[index].lang]+1)
        + output_data_offset[frames[index].lang]]);
  }

  return res;
}


template<class T>
int MMath<T>::AdaptationValidate(vector<vector<T>> & Z_matrix,
                                 vector<vector<T>> & A_matrix, const T * validation_data,
                                 const T * d_adapt_matrix, T * d_new_weight_matrix,
                                 int num_validation_samples,
                                 const vector<sample<T>> & validation_frames,
                                 const vector<int> & parameter_offset,
                                 const vector<int> & data_offset,
                                 const vector<int> & output_parameter_offset,
                                 const vector<int> & output_data_offset,
                                 int batch_size, const int window_size,
                                 const vector<int> & units_per_layer,
                                 const vector<int> & output_units_per_lang,
                                 T * d_weight_matrix, kActivationFunction ac_fun,
                                 cublasHandle_t * handle, cudaStream_t s1) {
  T max;
  int num_layers = units_per_layer.size();
  int * new_label = new int[batch_size];
  T count, wrong;
  T * d_auxA_matrix;
  T * d_auxZ_matrix;
  T * d_auxData;
  T * windowed_data;
  int * d_auxMask_matrix;
  vector<vector<int>> auxMask_matrix;
  stringstream ss;
  string hypothesis, reference;
  int max_index;
  const T kDropout = 0;

  int mem = 0;
  int output_mem = 0;

  A_matrix.clear();
  A_matrix.resize(units_per_layer.size());
  
  for( int ly = 0; ly < units_per_layer.size(); ++ly){
    A_matrix[ly].resize(batch_size*(units_per_layer[ly]+1));
  }

  ///Estimating output_layer offset for data matrices: Z,A,etc
  for (int output_layer = 0; output_layer < output_units_per_lang.size(); ++output_layer)
    output_mem += batch_size * (output_units_per_lang[output_layer] + 1);

  ///Estimating offset for data matrices: Z, A, etc
  for (int ly = 0; ly < units_per_layer.size()-1; ++ly)
    mem += batch_size * (units_per_layer[ly] + 1);
  mem+=output_mem;

  int original_batch_size = batch_size;
  int num_batches = num_validation_samples / batch_size;
  int res = num_validation_samples % batch_size;

  if (res > 0)
    num_batches++;

  CUDA_CALL(cudaMalloc(&d_auxZ_matrix, mem * sizeof(T)));
  CUDA_CALL(cudaMalloc(&d_auxA_matrix, mem * sizeof(T)));
  CUDA_CALL(cudaMalloc(&d_auxData, batch_size * (units_per_layer[0]+1) * sizeof(T)));

  windowed_data = new T[batch_size * units_per_layer[0]];

  wrong = count = 0;

  int * indexes = new int[num_validation_samples];
  for (int i = 0; i < num_validation_samples; ++i)
    indexes[i] = i;

  for (int num_batch = 0; num_batch < num_batches; ++num_batch) {
    if ((num_batch == num_batches - 1) && res > 0) {
      batch_size = res;
      delete[] windowed_data;
      delete[] new_label;
      new_label = new int[batch_size];
      windowed_data = new T[batch_size * units_per_layer[0]];
    }
    utils::GenerateWindowedSamples(validation_data, validation_frames, indexes,
                                   units_per_layer[0] / window_size, original_batch_size, batch_size,
                                   num_batch, window_size, windowed_data);

    CUDA_CALL(cudaMemset(d_auxZ_matrix, 0, mem * sizeof(T)));
    CUDA_CALL(cudaMemset(d_auxA_matrix, 0, mem * sizeof(T)));
    CUDA_CALL(cudaMemset(d_auxData, 0, batch_size * (units_per_layer[0]+1) * sizeof(T)));

    for (int ne = 0; ne < batch_size; ++ne)
      Z_matrix[0][IDX2C(0, ne, units_per_layer[0]+1)] = 1;

    for (int i = 1; i <= units_per_layer[0]; ++i)
      for (int ne = 0; ne < batch_size; ++ne)
        Z_matrix[0][IDX2C(i, ne, units_per_layer[0]+1)] =
            windowed_data[IDX2C(i-1, ne, units_per_layer[0])];

    CUBLAS_CALL(
        cublasSetMatrix(units_per_layer[0]+1, batch_size, sizeof(T),
                        Z_matrix[0].data(), units_per_layer[0]+1,
                        &d_auxData[0], units_per_layer[0]+1));

    AdaptationForward(Z_matrix, A_matrix, d_weight_matrix, d_auxZ_matrix, d_auxA_matrix,
                      d_adapt_matrix, d_new_weight_matrix, d_auxData, batch_size,
                      parameter_offset, data_offset, output_parameter_offset, output_data_offset,
                      units_per_layer, output_units_per_lang, ac_fun, handle, s1);

    for (int ne = 0; ne < batch_size; ++ne) {
      int lang = validation_frames[num_batch*original_batch_size+ne].lang;
      max = Z_matrix[num_layers - 1][IDX2C(1, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]];
      max_index = 0;
      for (int i = 2; i <=  output_units_per_lang[lang]; ++i)
        if (Z_matrix[num_layers - 1][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]]
            > max) {
          max =
              Z_matrix[num_layers - 1][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]];
          max_index = i - 1;
        }
      new_label[ne] = max_index;
    }

    for (int ne = 0; ne < batch_size; ++ne) {
      if (new_label[ne] != validation_frames[num_batch * original_batch_size + ne].label)
        wrong++;
      count++;
    }
  }

  CUDA_CALL(cudaFree(d_auxZ_matrix));
  CUDA_CALL(cudaFree(d_auxA_matrix));
  CUDA_CALL(cudaFree(d_auxData));

  cout << " " << _("Validation_Err=") << " " << wrong / count * 100 << " %";
  cout << " " << _("Total Errors=") << " " << wrong << ", " << _("Total Elements=") << " " << count << endl;
  delete[] indexes;
  delete[] new_label;
  delete[] windowed_data;

  return 1;
}

template<class T>
int MMath<T>::MoveDataFromDeviceToHost(vector<vector<T>> & weight_matrix,
                                       vector<vector<T>> & update_matrix,
                                       vector<vector<T>> & old_update_matrix,
                                       vector<vector<T>> & Z_matrix,
                                       vector<vector<T>> & A_matrix,
                                       vector<vector<T>> & gradient_matrix,
                                       vector<vector<T>> & delta_matrix,
                                       vector<vector<int>> & mask_matrix,
                                       vector<int> & units_per_layer,
                                       vector<int> & output_units_per_lang,
                                       T * d_weight_matrix, T * d_update_matrix,
                                       T * d_old_update_matrix, T * d_Z_matrix,
                                       T * d_A_matrix, T * d_gradient_matrix,
                                       T * d_delta_matrix,
                                       int * d_mask_matrix,
                                       const vector<int> & parameter_offset,
                                       const vector<int> & data_offset,
                                       const vector<int> & output_parameter_offset,
                                       const vector<int> & output_data_offset,
                                       int batch_size) {
  for (int ly = 0; ly < units_per_layer.size(); ++ly) {
    if (ly == 0) {
      CUBLAS_CALL(
          cublasGetMatrix((units_per_layer[ly]+1), batch_size, sizeof(T),
                          &d_Z_matrix[data_offset[ly]], units_per_layer[ly]+1,
                          Z_matrix[ly].data(), units_per_layer[ly]+1));
      CUBLAS_CALL(
          cublasGetMatrix((units_per_layer[ly]+1), batch_size, sizeof(T),
                          &d_A_matrix[data_offset[ly]], units_per_layer[ly]+1,
                          A_matrix[ly].data(), units_per_layer[ly]+1));
      CUBLAS_CALL(
          cublasGetMatrix((units_per_layer[ly]+1), batch_size, sizeof(int),
                          &d_mask_matrix[data_offset[ly]], units_per_layer[ly]+1,
                          mask_matrix[ly].data(), units_per_layer[ly]+1));

    } else if (ly < units_per_layer.size()-1) {
      CUBLAS_CALL(
          cublasGetMatrix(units_per_layer[ly], units_per_layer[ly-1]+1, sizeof(T),
                          &d_weight_matrix[parameter_offset[ly]],
                          units_per_layer[ly], weight_matrix[ly].data(),
                          units_per_layer[ly]));
      CUBLAS_CALL(
          cublasGetMatrix(units_per_layer[ly], units_per_layer[ly-1]+1, sizeof(T),
                          &d_update_matrix[parameter_offset[ly]],
                          units_per_layer[ly], update_matrix[ly].data(),
                          units_per_layer[ly]));
      CUBLAS_CALL(
          cublasGetMatrix(units_per_layer[ly], units_per_layer[ly-1]+1, sizeof(T),
                          &d_old_update_matrix[parameter_offset[ly]],
                          units_per_layer[ly], old_update_matrix[ly].data(),
                          units_per_layer[ly]));
      
      CUBLAS_CALL(
          cublasGetMatrix((units_per_layer[ly]+1), batch_size, sizeof(T),
                          &d_Z_matrix[data_offset[ly]], units_per_layer[ly]+1,
                          Z_matrix[ly].data(), units_per_layer[ly]+1));
      CUBLAS_CALL(
          cublasGetMatrix((units_per_layer[ly]+1), batch_size, sizeof(T),
                          &d_A_matrix[data_offset[ly]], units_per_layer[ly]+1,
                          A_matrix[ly].data(), units_per_layer[ly]+1));
      CUBLAS_CALL(
          cublasGetMatrix((units_per_layer[ly]+1), batch_size, sizeof(int),
                          &d_mask_matrix[data_offset[ly]], units_per_layer[ly]+1,
                          mask_matrix[ly].data(), units_per_layer[ly]+1));
      CUBLAS_CALL(
          cublasGetMatrix(units_per_layer[ly], units_per_layer[ly-1]+1, sizeof(T),
                          &d_gradient_matrix[parameter_offset[ly]],
                          units_per_layer[ly], gradient_matrix[ly].data(),
                          units_per_layer[ly]));
      CUBLAS_CALL(
          cublasGetMatrix((units_per_layer[ly]+1), batch_size, sizeof(T),
                          &d_delta_matrix[data_offset[ly]], units_per_layer[ly]+1,
                          delta_matrix[ly].data(), units_per_layer[ly]+1));
    } else {
      for (int lang = 0; lang < output_units_per_lang.size(); ++lang) {
        cout << "LANG: " << lang << endl;
        CUBLAS_CALL(
            cublasGetMatrix(output_units_per_lang[lang], units_per_layer[ly-1]+1, sizeof(T),
                            &d_weight_matrix[parameter_offset[ly]+output_parameter_offset[lang]],
                            output_units_per_lang[lang], &(weight_matrix[ly][output_parameter_offset[lang]]),
                            output_units_per_lang[lang]));
        CUBLAS_CALL(
            cublasGetMatrix(output_units_per_lang[lang], units_per_layer[ly-1]+1, sizeof(T),
                            &d_update_matrix[parameter_offset[ly]+output_parameter_offset[lang]],
                            output_units_per_lang[lang], &(update_matrix[ly][output_parameter_offset[lang]]),
                            output_units_per_lang[lang]));
        CUBLAS_CALL(
            cublasGetMatrix(output_units_per_lang[lang], units_per_layer[ly-1]+1, sizeof(T),
                            &d_old_update_matrix[parameter_offset[ly]+output_parameter_offset[lang]],
                            output_units_per_lang[lang], &(old_update_matrix[ly][output_parameter_offset[lang]]),
                            output_units_per_lang[lang]));
        CUBLAS_CALL(
            cublasGetMatrix(output_units_per_lang[lang], units_per_layer[ly-1]+1, sizeof(T),
                            &d_gradient_matrix[parameter_offset[ly]+output_parameter_offset[lang]],
                            output_units_per_lang[lang], &(gradient_matrix[ly][output_parameter_offset[lang]]),
                            output_units_per_lang[lang]));
        CUBLAS_CALL(
            cublasGetMatrix((output_units_per_lang[lang]+1), batch_size, sizeof(T),
                            &d_Z_matrix[data_offset[ly]+output_data_offset[lang]], output_units_per_lang[lang]+1,
                            &(Z_matrix[ly][output_data_offset[lang]]), output_units_per_lang[lang]+1));
        CUBLAS_CALL(
            cublasGetMatrix((output_units_per_lang[lang]+1), batch_size, sizeof(T),
                            &d_A_matrix[data_offset[ly]+output_data_offset[lang]], output_units_per_lang[lang]+1,
                            &(A_matrix[ly][output_data_offset[lang]]), output_units_per_lang[lang]+1));
        CUBLAS_CALL(
            cublasGetMatrix((output_units_per_lang[lang]+1), batch_size, sizeof(T),
                            &d_mask_matrix[data_offset[ly]+output_data_offset[lang]], output_units_per_lang[lang]+1,
                            &(mask_matrix[ly][output_data_offset[lang]]), output_units_per_lang[lang]+1));
        CUBLAS_CALL(
            cublasGetMatrix((output_units_per_lang[lang]+1), batch_size, sizeof(T),
                            &d_delta_matrix[data_offset[ly]+output_data_offset[lang]], output_units_per_lang[lang]+1,
                            &(delta_matrix[ly][output_data_offset[lang]]), output_units_per_lang[lang]+1));
      }
    }
  }
}

template<class T>
void MMath<T>::ShowInfo(const vector<vector<T>> & weight_matrix,
                        const vector<vector<T>> & update_matrix,
                        const vector<vector<T>> & old_update_matrix,
                        const vector<vector<T>> & Z_matrix,
                        const vector<vector<T>> & A_matrix,
                        const vector<vector<T>> & gradient_matrix,
                        const vector<vector<T>> & delta_matrix,
                        const vector<vector<int>> & mask_matrix,
                        const vector<int> & units_per_layer,
                        const vector<int> & output_units_per_lang,
                        const vector<int> & output_parameter_offset,
                        const vector<int> & output_data_offset,
                        int num_samples) {
  int ly, i, j, ne;

  cout << "-------------------------------" << endl;
  for (ly = 0; ly < units_per_layer.size()-1; ++ly) {
    cout << "---> " << _("Layer") << " " << ly << endl;

    cout << "A:" << endl;
    for (i = 0; i <= units_per_layer[ly]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << A_matrix[ly][IDX2C(i, ne, units_per_layer[ly]+1)] << " ";
      cout << endl;
    }

    cout << "Z:" << endl;
    for (i = 0; i <= units_per_layer[ly]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << Z_matrix[ly][IDX2C(i, ne, units_per_layer[ly]+1)] << " ";
      cout << endl;
    }

    cout << "Z_mask:" << endl;
    cout << units_per_layer[ly]+1 << " x " << num_samples << endl;
    for (i = 0; i <= units_per_layer[ly]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << mask_matrix[ly][IDX2C(i, ne, units_per_layer[ly]+1)] << " ";
      cout << endl;
    }

    cout << "Delta: " << endl;
    for (i = 0; i <= units_per_layer[ly]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << delta_matrix[ly][IDX2C(i, ne, units_per_layer[ly]+1)]
            << " ";
      cout << endl;
    }

    if (ly > 0) {
      cout << _("Gradient:") << " " << endl;
      for (i = 0; i < units_per_layer[ly]; ++i) {
        for (j = 0; j <= units_per_layer[ly - 1]; ++j)
          cout << gradient_matrix[ly][IDX2C(i, j, units_per_layer[ly])] << " ";
        cout << endl;
      }
      cout << "W:" << endl;
      cout << units_per_layer[ly - 1] << " x " << units_per_layer[ly] << endl;
      for (i = 0; i < units_per_layer[ly]; ++i) { /* Top H */
        for (j = 0; j <= units_per_layer[ly - 1]; ++j) /* Bot V */
          cout << weight_matrix[ly][IDX2C(i, j, units_per_layer[ly])] << " ";
        cout << endl;
      }

      cout << "dW" << endl;
      for (i = 0; i < units_per_layer[ly]; ++i) { /* Top H */
        for (j = 0; j <= units_per_layer[ly - 1]; ++j) /* Bot V */
          cout << update_matrix[ly][IDX2C(i, j, units_per_layer[ly])] << " ";
        cout << endl;
      }
      cout << "old_update_matrix" << endl;
      for (i = 0; i < units_per_layer[ly]; ++i) { /* Top H */
        for (j = 0; j <= units_per_layer[ly - 1]; ++j) /* Bot V */
          cout << old_update_matrix[ly][IDX2C(i, j, units_per_layer[ly])]
              << " ";
        cout << endl;
      }
    }
  }
  cout << _("Layer:") << " " << units_per_layer.size() << endl;
  for (int lang = 0; lang < output_units_per_lang.size(); ++lang) {
    cout << "Lang: " << lang << endl;
    cout << "A:" << endl;
    for (i = 0; i <= output_units_per_lang[lang]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << A_matrix[ly][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]] << " ";
      cout << endl;
    }

    cout << "Z:" << endl;
    for (i = 0; i <= output_units_per_lang[lang]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << Z_matrix[ly][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]] << " ";
      cout << endl;
    }

    cout << "Z_mask:" << endl;
    cout << output_units_per_lang[lang]+1 << " x " << num_samples << endl;
    for (i = 0; i <= output_units_per_lang[lang]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << mask_matrix[ly][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]] << " ";
      cout << endl;
    }

    cout << "Delta: " << endl;
    for (i = 0; i <= output_units_per_lang[lang]; ++i) {
      for (ne = 0; ne < num_samples; ++ne)
        cout << delta_matrix[ly][IDX2C(i, ne, output_units_per_lang[lang]+1)+output_data_offset[lang]]
            << " ";
      cout << endl;
    }

    cout << _("Gradient:") << " " << endl;
    for (i = 0; i < output_units_per_lang[lang]; ++i) {
      for (j = 0; j <= units_per_layer[ly - 1]; ++j)
        cout << gradient_matrix[ly][IDX2C(i, j, output_units_per_lang[lang])+output_parameter_offset[lang]] << " ";
      cout << endl;
    }
    cout << "W:" << endl;
    cout << units_per_layer[ly - 1] << " x " << output_units_per_lang[lang] << endl;
    for (i = 0; i < output_units_per_lang[lang]; ++i) { /* Top H */
      for (j = 0; j <= units_per_layer[ly - 1]; ++j) /* Bot V */
        cout << weight_matrix[ly][IDX2C(i, j, output_units_per_lang[lang])+output_parameter_offset[lang]] << " ";
      cout << endl;
    }
    cout << "dW" << endl;
    for (i = 0; i < output_units_per_lang[lang]; ++i) { /* Top H */
      for (j = 0; j <= units_per_layer[ly - 1]; ++j) /* Bot V */
        cout << update_matrix[ly][IDX2C(i, j, output_units_per_lang[lang])+output_parameter_offset[lang]] << " ";
      cout << endl;
    }
    cout << "old_update_matrix" << endl;
    for (i = 0; i < output_units_per_lang[lang]; ++i) { /* Top H */
      for (j = 0; j <= units_per_layer[ly - 1]; ++j) /* Bot V */
        cout << old_update_matrix[ly][IDX2C(i, j, output_units_per_lang[lang])+output_parameter_offset[lang]]
            << " ";
      cout << endl;
    }
  }
  cout << "+++++++++++++++++++++++++++++++++" << endl;
  cout << endl;
}

#endif // MMATH_H_
