/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  Miguel Ángel del Agua Teba - mdelagua@dsic.upv.es, madelagua@gmail.com
 *  Jesús Andrés Ferrer - jandres@dsic.upv.es
 *  Adrià Giménez Pastor - agimenez@dsic.upv.es
 */

#include <iostream>
#include <cmath>
#include <cstring>
#include <random>
#include <algorithm>
#include "utils.h"
#include "msg.h"
#include <typeinfo>
#include <limits>
#include "../intl.h"

#define IDX2C(i,j,ld) (((j)*(ld))+(i))
#define IDX2R(i,j,ld) (((i)*(ld))+(j))

using namespace std;

#ifndef MMATH_INTERFACE_H_
#define MMATH_INTERFACE_H_
template<class T>
class MMathInterface
{
public:
  MMathInterface(my_random_engine * g) {
    generator_ = g;
    min_eps_ = log(numeric_limits<double>::epsilon());
  };
  virtual ~MMathInterface() {
  };
  /* **************************************************************************** */

  virtual int Adapt(const T * training_data,
                    const vector<sample<T>> & training_frames, int * indexes, int num_samples,
                    T lr, T mom, int epochs, int batch_size,
                    int window_size, const T * validation,
                    const vector<sample<T>> & validation_frames,
                    int num_validation_samples, vector<vector<T>> & weight_matrix,
                    vector<vector<T>> & Z_matrix,
                    vector<vector<T>> & A_matrix,
                    vector<T> & gradient_matrix,
                    vector<int> & units_per_layer,
                    vector<int> & output_units_per_lang, 
                    bool likelihood_per_batch, bool likelihood_per_epoch,
                    bool error_per_batch, bool error_per_epoch,
                    kActivationFunction ac_fun,
                    int verbose) = 0;

  virtual int Classify(T * data, int num_samples, vector<sample<T>> & frames,
                       vector<vector<T>> & priori_log_probs, int bacth_size,
                       int window_size, vector<vector<T>> & weight_matrix,
                       vector<vector<T>> & Z_matrix, vector<vector<T>> & A_matrix,
                       vector<int> & units_per_layer, vector<int> & output_units_per_lang,
                       kActivationFunction ac_fun, kClassifierOutput clout,
                       vector<string> & out_file_names) = 0;

protected:
  my_random_engine * generator_;
  MESSAGE msg_;
  T min_eps_;

  void Softmax (T * A, const int num_samples, const int sample_size, T * Z) {
    int i, ns;
    T denominator, aux;
    int denominator_max_index;

    // Se supone que el bias de A será 0
    for (ns = 0; ns < num_samples; ++ns) {
      denominator = 0; denominator_max_index = 0;
      denominator_max_index = utils::MaxIndexTCol <T> (A, ns, num_samples, sample_size);

      for (i = 1; i < sample_size; ++i) {
        aux = A[IDX2C(i, ns, sample_size)] - A[denominator_max_index];
        if (aux > min_eps_)
          denominator += exp(aux);
      }
      denominator = log(denominator) + A[denominator_max_index];

      for (i = 1; i < sample_size; ++i) {
        // Me quedo el logaritmo (quito exponencial)
        Z[IDX2C(i, ns, sample_size)] =
            A[IDX2C(i, ns, sample_size)] - denominator;
      }
    }
  }

  void AdaptationGradient (vector<T> & gradient, const vector<T> & Z, const vector<T> & W,
                           const vector<T> & input, const int num_samples, 
                           const vector<int> & units_per_layer) {
    int i, j, t, k, sum;
    int last_layer = units_per_layer.size()-1;
    vector<T> op1; op1.resize(units_per_layer[last_layer-1]+1);
    vector<T> op2; op2.resize(units_per_layer[last_layer-1]+1);
    T aux;

    for (t = 0; t < num_samples; ++t){
      //op1_t = SUM_j p(j|x(t),M,c)W_j
      for (j = 0; j < units_per_layer[last_layer]; ++j){
        aux = exp(Z[IDX2C(j+1, t, num_samples)]);
        for (i = 0; i < units_per_layer[last_layer-1]+1; ++i){
          op1[i] += aux * W[IDX2C(j, i, units_per_layer[last_layer])];
        }
      }
      //op2_t = W_t - op1_t
      for (i = 0; i < units_per_layer[last_layer-1]+1; ++i){
        op2[i] = W[IDX2C(t, i, units_per_layer[last_layer])] -
            op1[IDX2C(i, 0, units_per_layer[last_layer-1]+1)];
      }
      //gradient+=op2*input
      for (i = 0; i < units_per_layer[last_layer-1]+1; ++i){
        for (k = 0; k < units_per_layer[last_layer-1]+1; ++k){
          gradient[IDX2C(i, k, units_per_layer[last_layer-1]+1)] +=
              op2[IDX2C(i, 0, units_per_layer[last_layer-1]+1)] *
              input[IDX2C(k, t, units_per_layer[last_layer-1]+1)];
        }
      }
    }

    op1.clear();
    op2.clear();
  }

  void Shuffle(int * indexes, int num_samples) {
    ul ns, random_num = 0;
    uniform_real_distribution<T> distribution(0, num_samples - 1);

    for (ns = 0; ns < num_samples; ++ns) {
      random_num = distribution(*generator_);
      if (random_num < 0) {
        msg_ << _("From shuffle in mmathinterface.h: Random index is negative ")
            << random_num << MESSAGE::error;
        msg_ << _("generator= ") << (*generator_)() << MESSAGE::error;
        abort();
      }
      swap(indexes[ns], indexes[random_num]);
    }
  }

};
#endif
