/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  Miguel Ángel del Agua Teba - mdelagua@dsic.upv.es, madelagua@gmail.com
 *  Jesús Andrés Ferrer - jandres@dsic.upv.es
 *  Adrià Giménez Pastor - agimenez@dsic.upv.es
 */

#include "utils.h"

namespace utils {
void Tokenize(string& str, vector<string>& tokens, const string& delimiters) {
  // Skip delimiters at beginning.
  string::size_type last_pos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  string::size_type pos = str.find_first_of(delimiters, last_pos);

  while (string::npos != pos || string::npos != last_pos) {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(last_pos, pos - last_pos));
    // Skip delimiters.  Note the "not_of"
    last_pos = str.find_first_not_of(delimiters, pos);
    // Find next "non-delimiter"
    pos = str.find_first_of(delimiters, last_pos);
  }
}

int ReadFileList(const string &name, vector<string> &file_name,
                 int & num_samples) {
  string line, word;
  ifstream file(name);
  vector<string> tokens;
  stringstream ss;
  if (file.is_open()) {
    getline(file, line);
    if (line.compare("")){
      msg << _("The file list is empty") << MESSAGE::error;
      return 0;
    }
    Tokenize(line, tokens, " ");
    ss << tokens[0];
    ss >> word;
    ss.clear();
    if (word.compare("FILELIST") != 0) {
      cerr << _("Incompatible file list format") << endl;
      return 0;
    }
    ss << tokens[1];
    ss >> num_samples;
    ss.clear();
    tokens.clear();
    while (file.good()) {
      getline(file, line);
      if (line.length() > 0) {
        file_name.push_back(line);
      }
    }
    file.close();
  } else {
    msg << _("Unable to open file ") << name << MESSAGE::error;
    return 0;
  }

  return 1;
}

int ReadFileList(const string &name, vector<string> &file_name) {
  string line, word;
  ifstream file(name);
  vector<string> tokens;
  stringstream ss;
  if (file.is_open()) {
    getline(file, line);
    if (line.empty()){
      msg << _("The file list is empty") << MESSAGE::error;
      return 0;
    }
    Tokenize(line, tokens, " ");
    ss << tokens[0];
    ss >> word;
    ss.clear();
    if (!word.compare("FILELIST") == 0)
      file.seekg(0);
    while (file.good()) {
      getline(file, line);
      if (line.length() > 0) {
        file_name.push_back(line);
      }
    }
    file.close();
  } else {
    msg << _("Unable to open file ") << name << MESSAGE::error;
    return 0;
  }
  return 1;
}

int CheckFileExists(vector<string> & file_names) {
  ifstream file;
  for (unsigned int i = 0; i < file_names.size(); ++i) {
    file.open(file_names[i]);
    //file.read(0, sizeof(char));
    if (!file) {
      msg << _("Unable to open file ") << file_names[i] << MESSAGE::error;
      return 0;
    }
    file.close();
  }
  return 1;
}

int ReadFileType(ifstream & file, kFileType& fT, int& height, int& width) {
  stringstream ss;
  string line;
  vector<string> tokens;
  const int max_header_size = 8;
  int count = 0;
  char * memblock = new char[max_header_size];
  string * aux;

  while (file.good() and count < max_header_size) {
    file.read(memblock + count, sizeof(char));
    aux = new string(memblock, sizeof(char) * (count + 1));
    if ((*aux).compare("P1") == 0) {
      fT = P1;
      delete aux;
      break;
    } else if ((*aux).compare("RAW") == 0) {
      fT = RAW;
      delete aux;
      break;
    } else if ((*aux).compare("AKREALTF") == 0) {
      fT = AKREALTF;
      delete aux;
      break;
    } else if ((*aux).compare("AKREALBF") == 0) {
      fT = AKREALBF;
      delete aux;
      break;
    } else if ((count + 1) == max_header_size) {
      fT = UNK;
      delete aux;
      break;
    }
    count++;
    delete aux;
  }

  delete[] memblock;

  if (file.good() and fT != AKREALBF) {
    file.seekg(0);
    getline(file, line);
    Tokenize(line, tokens, " ");

    switch (fT) {
    case RAW:
      ss << tokens[1];
      ss >> width;
      ss.clear();
      height = 1;
      tokens.clear();
      break;
    case P1:
      tokens.clear();
      getline(file, line);
      Tokenize(line, tokens, " ");
      ss << tokens[0];
      ss >> height;
      ss.clear();
      ss << tokens[1];
      ss >> width;
      ss.clear();
      tokens.clear();
      break;
    case AKREALTF:
      tokens.clear();
      getline(file, line);
      Tokenize(line, tokens, " ");
      ss << tokens[0];
      ss >> height;
      ss.clear();
      ss << tokens[1];
      ss >> width;
      ss.clear();
      tokens.clear();
      break;
    case UNK:
      height = 0;
      width = 0;
      break;
    case AKREALBF:
      break;
    }
  } else if (file.good() and fT == AKREALBF) {
    file.seekg(sizeof(char) * 8);
    file.read((char *) &height, sizeof(int));
    file.read((char *) &width, sizeof(int));
  } else
    return 0;
  return 1;
}

int ReadFileType(const string& file_name, kFileType& fT, int& height,
                 int& width) {
  ifstream file;
  file.open(file_name);
  ReadFileType(file, fT, height, width);
  file.close();
  return 1;
}
}
