/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  Miguel Ángel del Agua Teba - mdelagua@dsic.upv.es, madelagua@gmail.com
 *  Jesús Andrés Ferrer - jandres@dsic.upv.es
 *  Adrià Giménez Pastor - agimenez@dsic.upv.es
 */

#include <iostream>
#include <vector>
#include <string>
#include <fstream>
#include <sstream>
#include "msg.h"
#include <random>
#include "../intl.h"

using namespace std;

#ifndef UTILS
#define UTILS

#define MAX(X,Y) ((X) > (Y) ? (X):(Y))
#define IDX2C(i,j,ld) (((j)*(ld))+(i))
#define IDX2R(i,j,ld) (((i)*(ld))+(j))

typedef unsigned long ul;

enum kMDOffset {
  LABELS, LANGS, POSITIONS, SEGMENTPOSITIONS, SEGMENTLENGTHS, COUNT = 2
};
enum kClassifierOutput {
  NONE_OUTPUT, POSTERIOR_OUTPUT, SIGMOID_OUTPUT
};
enum kFileType {
  P1, RAW, AKREALTF, AKREALBF, UNK
};
enum kActivationFunction {
  SIGMOID, TANH, RELU, SOFTRELU, UNKNOWN
};

class my_random_engine: public default_random_engine{
public:
  unsigned long long num_random_calls_;
  my_random_engine ():default_random_engine(){
    num_random_calls_ = 0;
  }

  my_random_engine (int seed):default_random_engine(seed){
    num_random_calls_ = 0;
  }

  my_random_engine (int seed, unsigned long long z):default_random_engine(seed){
    this->discard(z);
  }

  void seed (typename default_random_engine::result_type val){
    num_random_calls_ = 0;
    return default_random_engine::seed(val);
  }

  void discard (unsigned long long z){
    num_random_calls_ = z;
    return default_random_engine::discard(z);
  }

  typename default_random_engine::result_type operator()(){
    num_random_calls_++;
    return default_random_engine::operator()();
  }
};

template<class T>
struct sample {
  int segmentPosition; ///< Initial position of the segment within a global Sample matrix
  int segmentLength; ///< Length of the segment which the sample belongs to
  int position; ///< Index of the sample within the segment
  int label; ///< Index indicating the label of the sample
  T * content; ///< Pointer to content of the sample
  int dim; ///< Dimension of content
  string fileName; ///< file_name to which it belongs
  int lang; ///< Language of the sample
};

namespace utils {
static MESSAGE msg;
void Tokenize(string&, vector<string>&, const string&); /* Tokenize a given string based on an specified delimiter */
int MaxIndex(double *, int);
/// Each of the specified files could contain several samples
int ReadFileList(const string&, vector<string>&, int & num_samples);
int ReadFileList(const string&, vector<string>&);
int ReadFileType(ifstream&, kFileType&, int&, int&);
int ReadFileType(const string&, kFileType&, int&, int&);
int CheckFileExists(vector<string> &);
template<class T>
int LoadRawFileContent(vector<string>&, T *, int, int);
template<class T>
int LoadRealFea(ifstream & file_name, vector<T>& content, int & rows,
                int & cols);
template<class T>
int ReadLabels(string & file_name, vector<sample<T>> & frames,
               vector<int> & num_labels_per_lang, vector<int> & num_samples_per_lang);
template<class T>
int GenerateWindowedSamples(const T * data, const vector<sample<T>> & frames,
                            int * indexes, int sample_size, int batchSize,
                            int num_samples, int batch, int ws, T * result);
template<class T>
int MaxIndex(T *, int);
template<class T>
int MaxIndexT(T *, int, int, int);
template<class T>
int MaxIndexTCol(T *, int, int, int);
}

template<class T> int utils::MaxIndex(T * v, int num_samples) {
  int index = 0;
  T value = v[0];
  for (int i = 1; i < num_samples; ++i) {
    if (v[i] > value) {
      value = v[i];
      index = i;
    }
  }
  return index;
}

/* This method does not take into account the first value (bias) */
template<class T> int utils::MaxIndexT(T * v, int col, int num_samples,
    int dim) {
  int index = num_samples + col;
  T value = v[num_samples + col];
  for (int i = 1; i < dim; ++i)
    if (v[i * num_samples + col] > value) {
      value = v[i * num_samples + col];
      index = i * num_samples + col;
    }
  return index;
}

template<class T> int utils::MaxIndexTCol(T * v, int col, int num_samples,
    int dim) {
  int index = col * dim;
  T value = v[col * dim];
  for (int i = 1; i < dim; ++i) {
    if (v[col * dim + i] > value) {
      value = v[col * dim + i];
      index = col * dim + i;
    }
  }
  return index;
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  utils
 *      Method:  utils :: generateWindowedSamples
 * Description:  Extracts windows from the data
 *--------------------------------------------------------------------------------------
 */
template<class T> int utils::GenerateWindowedSamples(const T * data,
    const vector<sample<T>> & frames, int * indexes, int sample_size,
    int batchSize, int num_samples, int mbatch, int ws, T * result) {

  int i, m, w, dim, sampleInd, index, sample;
  int margin = ws / 2;
  /* Empty first frames are filled copying the first non empty frame */
  /* Empty final frames are filled copying the final non empty frame */

  index = 0;
  for (sample = mbatch * batchSize; sample < mbatch * batchSize + num_samples;
      ++sample) {
    i = indexes[sample];
    sampleInd = frames[i].position;
    for (w = 0, m = -margin; m <= margin; ++m, ++w) {
      if (((sampleInd + m) < 0))
        for (dim = 0; dim < sample_size; ++dim)
          result[IDX2C(sample_size*w+dim, index, sample_size*ws)] =
              data[IDX2C((ul)dim, (ul)frames[i].segmentPosition, (ul)sample_size)];
      else if ((sampleInd + m) >= frames[i].segmentLength)
        for (int dim = 0; dim < sample_size; ++dim)
          result[IDX2C(sample_size*w+dim, index, sample_size*ws)] =
              data[IDX2C((ul)dim, (ul)frames[i].segmentPosition+frames[i].segmentLength-1, (ul)sample_size)];
      else
        for (int dim = 0; dim < sample_size; ++dim)
          result[IDX2C(sample_size*w+dim, index, sample_size*ws)] =
              data[IDX2C((ul)dim, (ul)i+(ul)m, (ul)sample_size)];
    }
    ++index;
  }
  return 1;
}

/*
 *--------------------------------------------------------------------------------------
 *       Class:  utils
 *      Method:  utils :: ReadLabels
 * Description:  Loads labels from a file. It is expected that the order is the same as
 * the order in the training sample file list.
 *--------------------------------------------------------------------------------------
 */
template<class T>
int utils::ReadLabels(string & file_name, vector<sample<T>> & frames,
                      vector<int> & num_labels_per_lang, vector<int> & num_samples_per_lang) {
  string line, fT;
  ifstream file;
  stringstream ss;
  vector<string> tokens;
  int times;
  int index;
  int language_id = 0;
  vector<int> readed_langs;
  readed_langs.push_back(0);
  num_samples_per_lang.clear();
  num_labels_per_lang.clear();
  num_labels_per_lang.push_back(0);
  num_samples_per_lang.push_back(0);

  file.open(file_name);
  getline(file, line);
  Tokenize(line, tokens, " ");
  ss << tokens[0];
  ss >> fT;
  ss.clear();
  tokens.clear();

  if (fT.compare("LABELS") != 0)
    file.seekg(0);

  if (file.is_open()) {
    while (file.good()) {
      getline(file, line);
      if (line.compare("") == 0)
        continue;
      Tokenize(line, tokens, " ");
      ss << tokens[1];
      ss >> times;
      ss.clear();
      if( tokens.size() > 5 ) { ///Multilingual labels file
        ss << tokens[4];
        ss >> index;
        ss.clear();
        ss << tokens[3];
        ss >> language_id;
        ss.clear();
        while((language_id + 1) > (int)readed_langs.size()){ //Language not yet read
          readed_langs.push_back(readed_langs.size());
          num_labels_per_lang.push_back(0);
          num_samples_per_lang.push_back(0);
        }
        if (num_labels_per_lang[language_id] < (index + 1))
          num_labels_per_lang[language_id] = index + 1;
        num_samples_per_lang[language_id] += times;
      } else {
        ss << tokens[3];
        ss >> index;
        ss.clear();
        if (num_labels_per_lang[0] < (index + 1))
          num_labels_per_lang[0] = index + 1;
        num_samples_per_lang[0] += times;
      }

      for (int i = 0; i < times; ++i) {
        sample<T> temp;
        temp.label = index;
        temp.lang = language_id;
        frames.push_back(temp);
      }
      tokens.clear();
    }
    file.close();
  } else {
    cerr << _("Could not open file") << " " << file_name << endl;
    return 0;
  }
  return 1;
}

template<class T>
int utils::LoadRawFileContent(vector<string>& file_names, T * content, int h,
    int w) {
  string stream, line;
  ifstream file;
  stringstream ss;
  int count, nline;
  count = nline = 0;
  vector<string> tokens;
  for (vector<string>::iterator it = file_names.begin(); it != file_names.end();
      ++it) {
    file.open(*it);
    nline = 0;
    while (file.good()) {
      if (nline == 0) {
        getline(file, line);
        nline++;
        tokens.clear();
        Tokenize(line, tokens, " ");
        if (!tokens[0].compare("RAW") == 0) {
          cout << _("File is not RAW") << endl;
          return 0;
        }
      } else {
        getline(file, line);
        nline++;
        if (line.compare("") == 0)
          continue;
        tokens.clear();
        Tokenize(line, tokens, " ");
        for (unsigned i = 0; i < tokens.size(); ++i) {
          ss << tokens[i];
          ss >> content[count];
          ss.clear();
          count++;
        }
      }
    }
    file.close();
  }

  return 1;
}

template<class T>
int utils::LoadRealFea(ifstream & file, vector<T> & content, int & num_samples,
    int & sample_size) {
  T aux;
  string line;
  stringstream ss;
  int nsample;
  nsample = 0;
  vector<string> tokens;
  kFileType fT;

  /* A tener en cuenta que están anotadas al revés */
  if (!ReadFileType(file, fT, sample_size, num_samples)) {
    cerr << _("Could not load file type from loadRealFea") << endl;
    return 0;
  }
  if (fT != AKREALTF and fT != AKREALBF) {
    cerr << _("Unknown file type") << endl;
    return 0;
  }

  content.resize(sample_size * num_samples);

  if (fT == AKREALTF) {
    nsample = 0;
    while (file.good()) {
      getline(file, line);
      if (line.compare("") == 0)
        continue;
      tokens.clear();
      Tokenize(line, tokens, " ");
      for (unsigned i = 0; i < tokens.size(); ++i) {
        ss << tokens[i];
        ss >> aux;
        ss.clear();
        content[IDX2C( i, nsample, sample_size)] = aux;
      }
      nsample++;
    }
  } else if (fT == AKREALBF) {
    file.seekg(sizeof(char) * 8 + sizeof(int) * 2);

    file.read((char*) &content[IDX2C(0,0,sample_size)],
        sizeof(float) * num_samples * sample_size);
    /* for( int ne = 0; ne < num_samples; ++ne )
     for( int i = 0; i < sample_size; ++i )
     file.read( (char*)&content[IDX2C(i, ne, sample_size)], sizeof(float) ); */

  }

  return 1;
}

#endif
