/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  extract.c - Feature extraction.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */


#include <assert.h>
#include <libTLK/TLK.h>
#include <limits.h>
#include <locale.h>
#include <sox.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmdl.h"
#include "intl.h"
#include "macros.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 2




/*********/
/* TYPES */
/*********/

struct opts
{
  
  int    verbose;
  int    outfreq;
  double wduration;
  int    numchans;
  int    lofreq;
  int    hifreq;
  tLBool binary;
  int    ftype;
  int    numceps;
  tLBool dzeromean;
  double preemcoef;
  tLBool dhamming;
  int    ceplifter;
  tLBool dpower;
  tLBool eenergy;
  int    deltawindow;
  int    accwindow;
  tLBool dcep0;
  tLBool dmeannorm;
  tLBool mustdnorm;
  
};


struct args
{
  
  const char *input_fn;
  const char *output_fn;
  
};


typedef struct
{
  
  size_t        size;
  size_t        N;
  double       *v;
  sox_sample_t  min;
  sox_sample_t  max;
  
} buffer_t;




/*************/
/* FUNCTIONS */
/*************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{

  fprintf ( to, "%s [%s] <input-lst> <output-lst>\n", prog, _("options") );
  fprintf ( to, "\n" );
  if ( !details ) return;
  fprintf ( to, "    <input-lst> %s\n",
	    _("List of the input audio files") );
  fprintf ( to, "    <output-lst> %s\n\n",
	    _("List of the files to which"
	      " the extracted samples will be written") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -a,--acceleration-window <int>: %s\n",
	    _("Window size that will be used"
	      " to calculate the acceleration coefficients (second"
	      " order derivatives). 0 means disabled. Default is 2") );
  fprintf ( to, "      -c,--num-channels <int>: %s\n",
	    _("Number of channels used"
	      " in filterbank analysis. Default is 26") );
  fprintf ( to, "      -C,--num-cepstrals <int>: %s\n",
	    _("Number of cepstral coefficients"
	      " to be extracted. Default is 12") );
  fprintf ( to, "      -d,--deltas-window <int>: %s\n",
	    _("Window size that will be used"
	      " to calculate the delta coefficients (first order"
	      " derivatives). 0 means disabled. Default is 2") );
  fprintf ( to, "      -f,--frequency <int>: %s\n",
	    _("Number of frames per second"
	      " to be extracted. Default is 100") );
  fprintf ( to, "      -F,--cepstra-lifter <int>: %s\n",
	    _("Value used to lifter the cepstra."
	      " 0 means disabled. Default is 22") );
  fprintf ( to, "      -n,--norm: %s\n",
	    _("Mean and variance normalization of all samples") );
  fprintf ( to, "      -p,--pre-emphasis-coefficient <double>: %s\n",
	    _("Coefficient used to pre-emphasize the signal."
	      " 0 means disabled. Default is 0.97") );
  fprintf ( to, "      -t,--type [melspec|fbank|mfcc]: %s\n",
	    _("Type of the extracted features."
	      " Default is mfcc") );
  fprintf ( to, "      -w,--window-duration <double>: %s\n",
	    _("Duration in seconds of the window used in extraction."
	      " Default is 0.025") );
  fprintf ( to, "      --disable-cep0: %s\n",
	    _("Don't calculate cepstrum 0") );
  fprintf ( to, "      --disable-hamming: %s\n",
	    _("Don't apply Hamming windowing") );
  fprintf ( to, "      --disable-mean-norm: %s\n",
	    _("Disable mean normalization at the sample level") );
  fprintf ( to, "      --disable-power: %s\n",
	    _("Don't use the power in filterbank analysis") );
  fprintf ( to, "      --disable-zeromean: %s\n",
	    _("DC mean will not be removed from the input audio") );
  fprintf ( to, "      --enable-energy: %s\n", _("Calculate the energy") );
  fprintf ( to, "      --hifreq <int>: %s\n",
	    _("The high frequency cut-off used in filterbank analysis."
	      " Default value is the maximum possible") );
  fprintf ( to, "      --lofreq <int>: %s\n",
	    _("The low frequency cut-off used in filterbank analysis."
	      " Default value is the minimum possible") );
  fprintf ( to, "      --noraw: %s\n",
	    _("Store features in text format") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n, noraw, dzeromean, dhamming, dpower, eenergy,
    dcep0, dmeannorm, cnorm;
  const char *prog, *nchans, *freq, *wduration, *lofreq, *hifreq,
    *ftype, *nceps, *preemcoef, *ceplifter, *accwindow, *deltawindow;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'a', "acceleration-window", &accwindow );
  add_cmdl_value ( cmdl, 'c', "num-channels", &nchans );
  add_cmdl_value ( cmdl, 'C', "num-cepstrals", &nceps );
  add_cmdl_value ( cmdl, 'd', "deltas-window", &deltawindow );
  add_cmdl_value ( cmdl, 'f', "frequency", &freq );
  add_cmdl_value ( cmdl, 'F', "cepstra-lifter", &ceplifter );
  add_cmdl_counter ( cmdl, 'n', "norm", &cnorm );
  add_cmdl_value ( cmdl, 'p', "pre-emphasis-coefficient", &preemcoef );
  add_cmdl_value ( cmdl, 't', "ftype", &ftype );
  add_cmdl_value ( cmdl, 'w', "window-duration", &wduration );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "disable-cep0", &dcep0 );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "disable-hamming", &dhamming );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "disable-mean-norm", &dmeannorm );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "disable-power", &dpower );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "disable-zeromean", &dzeromean );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "enable-energy", &eenergy );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "hifreq", &hifreq );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "lofreq", &lofreq );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "noraw", &noraw );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( nchans == NULL ) opts->numchans= 26;
  else
    {
      opts->numchans= atoi ( nchans );
      if ( opts->numchans < 1 )
	tL_error ( _("num-channels is less than 1") );
    }
  if ( nceps == NULL ) opts->numceps= 12;
  else
    {
      opts->numceps= atoi ( nceps );
      if ( opts->numceps < 1 )
	tL_error ( _("num-cepstrals is less than 1") );
    }
  if ( freq == NULL ) opts->outfreq= 100;
  else
    {
      opts->outfreq= atoi ( freq );
      if ( opts->outfreq < 1 )
	tL_error ( _("frequency is negative") );
    }
  if ( wduration == NULL ) opts->wduration= 0.025;
  else
    {
      opts->wduration= atof ( wduration );
      if ( opts->wduration <= 0.0 )
	tL_error ( _("window-duration is less than or equal to 0") );
    }
  if ( hifreq == NULL ) opts->hifreq= INT_MAX;
  else
    {
      opts->hifreq= atoi ( hifreq );
      if ( opts->hifreq < 1 )
	tL_error ( _("hifreq is less than 1") );
    }
  if ( lofreq == NULL ) opts->lofreq= -1;
  else
    {
      opts->lofreq= atoi ( lofreq );
      if ( opts->lofreq < 1 )
	tL_error ( _("lofreq is less than 1") );
    }
  if ( ftype == NULL ) opts->ftype= TL_MFCC;
  else
    {
      if ( !strcmp ( ftype, "mfcc" ) ) opts->ftype= TL_MFCC;
      else if ( !strcmp ( ftype, "melspec" ) ) opts->ftype= TL_MELSPEC;
      else if ( !strcmp ( ftype, "fbank" ) ) opts->ftype= TL_FBANK;
      else tL_error ( _("Unknown feature type: %s"), ftype );
    }
  opts->binary= (noraw==0);
  opts->dzeromean= (dzeromean!=0);
  if ( preemcoef == NULL ) opts->preemcoef= 0.97;
  else
    {
      opts->preemcoef= atof ( preemcoef );
      if ( opts->preemcoef < 0 || opts->preemcoef >= 1.0 )
	tL_error ( _("pre-emphasis coefficient must be in the range [0,1[") );
    }
  opts->dhamming= (dhamming!=0);
  opts->ceplifter= (ceplifter==NULL) ? 22 : atoi ( ceplifter );
  opts->dpower= (dpower!=0);
  opts->eenergy= (eenergy!=0);
  opts->deltawindow= (deltawindow==NULL) ? 2 : atoi ( deltawindow );
  opts->accwindow= (accwindow==NULL) ? 2 : atoi ( accwindow );
  opts->dcep0= (dcep0!=0);
  opts->dmeannorm= (dmeannorm!=0);
  opts->mustdnorm= (cnorm!=0);
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->input_fn= argv[0];
  args->output_fn= argv[1];
  
} /* end usage */


static tLFileList *
load_list (
	   const char   *fname,
	   const tLBool  is_input,
	   const int     verbose
	   )
{
  
  tLFileList *fl;
  size_t n;
  
  
  if ( verbose )
    fprintf ( stderr,
	      is_input ? 
	      _("Loading the input sample list\n") :
	      _("Loading the output sample list\n") );
  fl= tL_filelist_new_from_file ( fname );
  if ( fl == NULL ) tL_error ( _("Unable to open '%s'"), fname );
  
  if ( verbose > 1 )
    {
      fprintf ( stderr,
		is_input ? _("Input samples:\n") : _("Output samples:\n") );
      for ( n= 0; n < fl->N; ++n )
	fprintf ( stderr, "  - %s\n", fl->files[n] );
    }
  
  return fl;
  
} /* end load_list */


static buffer_t *
create_buffer (void)
{
  
  buffer_t *new;
  
  
  new= mem_alloc ( buffer_t, 1 );
  new->v= mem_alloc ( double, 1 );
  new->N= 0;
  new->size= 1;
  
  return new;
  
} /* end create_buffer */


static void
free_buffer (
	     buffer_t *buffer
	     )
{
  
  free ( buffer->v );
  free ( buffer );
  
} /* end free_buffer */


static void
resize_buffer (
	       buffer_t     *buffer,
	       const size_t  size
	       )
{
  
  buffer->size= size;
  buffer->v= mem_realloc ( double, buffer->v, size );
  
} /* end resize_buffer */


static void
init_buffer (
	     buffer_t           *buffer,
	     const sox_format_t *f
	     )
{
  
  size_t length;
  double ratio;
  
  
  assert ( f->signal.channels > 0 );
  ratio=  f->signal.rate/16000;
  length= (size_t) (f->signal.length/f->signal.channels)/ratio;
  if ( length > buffer->size ) resize_buffer ( buffer, length );
  buffer->max= SOX_SAMPLE_MIN; buffer->min= SOX_SAMPLE_MAX;
  buffer->N= 0;
  
} /* end init_buffer */


static int
output_flow (
	     sox_effect_t *       effp,
	     sox_sample_t const * ibuf,
	     sox_sample_t *       obuf,
	     size_t *             isamp,
	     size_t *             osamp
	     )
{
  
  buffer_t *buffer;
  size_t aux, n;
  double *p;
  sox_sample_t val;
  
  
  buffer= *((buffer_t **) (effp->priv));
  if ( (aux= buffer->N+*isamp) > buffer->size ) resize_buffer ( buffer, aux );
  for ( n= 0, p= &(buffer->v[buffer->N]); n < *isamp; ++n )
    {
      val= ibuf[n];
      p[n]= (double) val;
      /*
      p[n]= (val>=0) ?
	(val/(double) SOX_SAMPLE_MAX) :
	-(val/(double) SOX_SAMPLE_MIN);
      */
	/*
      if ( val > buffer->max ) buffer->max= val;
      if ( val < buffer->min) buffer->min= val;
	*/
    }
  buffer->N+= n;
  *osamp= 0;
  
  return SOX_SUCCESS;
  
} /* end output_flow */


static sox_effect_handler_t const *
output_handler ( void )
{
  
  static sox_effect_handler_t handler = {
    "output", NULL, SOX_EFF_MCHAN, NULL, NULL,
    output_flow, NULL, NULL, NULL, sizeof(buffer_t *)
  };
  
  return &handler;
  
} /* end output_handler */


/* sox_format_t *out; */
static sox_effects_chain_t *
create_chain (
	      sox_format_t     *f,
	      sox_signalinfo_t *si_info,
	      buffer_t         *buffer
	      )
{
  
  static char * rate[] = {"-v","16k"};
  static char * mixer[]= {"0.5,0.5"};
  sox_effect_t *e;
  const sox_effect_handler_t *h;
  sox_effects_chain_t *chain;
  char * params[1];
  sox_signalinfo_t aux;
  
  
  chain= sox_create_effects_chain ( &f->encoding, &f->encoding );
  
  /* Input. */
  h= sox_find_effect ( "input" ); if ( h == NULL ) goto error;
  e= sox_create_effect ( h ); if ( e == NULL ) goto error;
  params[0]= (char *) f;
  if ( sox_effect_options ( e, 1, params ) != SOX_SUCCESS ) goto error;
  if ( sox_add_effect ( chain, e, si_info, si_info ) != SOX_SUCCESS )
    goto error;
  free ( e );
  
  /* Polyphase. */
  /*
  aux= *si_info; aux.rate= 16000;
  h= sox_find_effect ( "polyphase" );
  if ( h == NULL ) goto error;
  e = sox_create_effect ( h );
  if ( sox_effect_options ( e, 0, NULL ) != SOX_SUCCESS ) goto error;
  if ( sox_add_effect ( chain, e, &(f->signal), &aux ) != SOX_SUCCESS )
    goto error;
  */
  /* Mixer. */
  if ( si_info->channels != 1 )
    {
      aux= *si_info; aux.channels= 1;
      h= sox_find_effect ( "mixer" ); if ( h == NULL ) goto error;
      e= sox_create_effect ( h ); if ( e == NULL ) goto error;
      if ( sox_effect_options ( e, 1, mixer ) != SOX_SUCCESS ) goto error;
      if ( sox_add_effect ( chain, e, si_info, &aux ) != SOX_SUCCESS )
	goto error;
      free ( e );
    }
  
  /* Rate. */
  if ( si_info->rate != 16000 )
    {
      h= sox_find_effect ( "rate" ); if ( h == NULL ) goto error;
      e= sox_create_effect ( h ); if ( e == NULL ) goto error;
      if ( sox_effect_options ( e, 2, rate ) != SOX_SUCCESS ) goto error;
      if ( sox_add_effect ( chain, e, si_info, si_info ) != SOX_SUCCESS )
	goto error;
      free ( e );
    }
  
  /* Output. */
  /*
  h= sox_find_effect ( "output" );
  if ( h == NULL ) goto error;
  e = sox_create_effect ( h );
  params[0]= (char *) out;
  if ( sox_effect_options ( e, 1, params ) != SOX_SUCCESS ) goto error;
  if ( sox_add_effect ( chain, e, si_info, si_info ) != SOX_SUCCESS )
    goto error;
  */
  e= sox_create_effect ( output_handler () ); if ( e == NULL ) goto error;
  *((buffer_t **) e->priv)= buffer;
  if ( sox_add_effect ( chain, e, si_info, si_info ) != SOX_SUCCESS )
    goto error;
  free ( e );
  
  return chain;
  
 error:
  tL_error ( _("Error while creating the SoX chain") );
  return NULL;
  
} /* end create_chain */


static void
extract_sample_read (
		     const char *fn,
		     buffer_t   *buffer
		     )
{
  
  sox_format_t *f;
  sox_effects_chain_t *chain;
  sox_signalinfo_t si_chain;
  /*
  int channels;
  double rate;
  */
  
  
  f= sox_open_read ( fn, NULL, NULL, NULL );
  if ( f == NULL ) tL_error ( _("An error occurred while opening '%s'"), fn );
  init_buffer ( buffer, f );
  /*
  channels= f->signal.channels; rate= f->signal.rate;
  f->signal.channels= 1;
  f->signal.rate= 16000;
  assert(out = sox_open_write("kk.wav", &(f->signal), NULL, NULL, NULL, NULL));
  f->signal.channels= channels; f->signal.rate= rate;
  */
  si_chain= f->signal; si_chain.mult= NULL;
  chain= create_chain ( f, &si_chain, buffer );
  sox_flow_effects ( chain, NULL, NULL );
  sox_delete_effects_chain ( chain );
  sox_close ( f );
  /* sox_close(out);*/
  
} /* end extract_sample_read */


static void
extract_sample (
		buffer_t             *buffer,
		tLFeatureExtractor   *fext,
		tLFea                *fea,
		const char           *infn,
		const char           *outfn,
		const tLBool          binary,
		tLMuStd             **mustd,
		const int             verbose
		)
{
  
  int ret;
  char *err;
  
  
  /* Read file. */
  if ( verbose > 1 ) fprintf ( stderr, "  < '%s'\n", infn );
  extract_sample_read ( infn, buffer );
  if ( buffer->N == 0 )
    {
      tL_warning ( _("'%s' is empty, unable to extract features"), infn );
      return;
    }
  
  /* Extract features. */
  tL_extract_features ( fext, buffer->v, buffer->N, 16000, fea );
  if ( mustd != NULL )
    {
      if ( *mustd == NULL ) *mustd= tL_mustd_new ( fea->rows );
      else if ( fea->rows != (*mustd)->D )
	tL_error ( _("An unexpected error occurred while estimating"
		     " the mean and variance normalization for sample"
		     " '%s': dimensions differ"), infn );
      tL_mustd_accum ( *mustd, fea );
    }
  
  /* Write file. */
  if ( verbose > 1 ) fprintf ( stderr, "  > '%s'\n", outfn );
  ret= tL_fea_write ( fea, outfn, binary, &err );
  if ( ret ) tL_error ( _("An error occurred while writing '%s': %s"),
			outfn, err );
  
} /* end extract_sample */


static void
norm_sample (
	     const char    *fn,
	     tLFea         *fea,
	     const tLMuStd *mustd,
	     const tLBool   binary,
	     const int      verbose
	     )
{
  
  char *err;
  int ret, t;
  unsigned int d;
  float **frames, *frame;
  
  
  if ( verbose > 1 ) fprintf ( stderr, _("  Normalizing '%s'\n"), fn );
  ret= tL_fea_load ( fea, fn, &err );
  if ( ret )
    {
      tL_warning ( _("'%s' will not be normalized, an error occurred"
		     " while reading it: %s"), fn, err );
      free ( err );
      return;
    }
  frames= tL_fea_get_real_data ( fea );
  for ( t= 0; t < fea->cols; ++t )
    {
      frame= frames[t];
      for ( d= 0; d < mustd->D; ++d )
        frame[d]= (frame[d]-mustd->mu[d])/mustd->std[d];
    }
  ret= tL_fea_write ( fea, fn, binary, &err );
  if ( ret ) tL_error ( _("An error occurred while writing '%s': %s"),
			fn, err );
  
} /* end norm_sample */


/* Returns a tLMuStd struct if mean and variance normalization is
   applied. */
static void
extract (
	 const tLFileList  *in_lst,
	 const tLFileList  *out_lst,
	 const struct opts *opts
	 )
{
  
  size_t n;
  buffer_t *buffer;
  tLFeatureExtractor *fext;
  tLFea *fea;
  tLMuStd *mustd, **mustdptr;
  
  
  if ( opts->verbose ) fprintf ( stderr, _("Extracting samples\n") );
  
  /* Prepare. */
  mustd= NULL;
  mustdptr= opts->mustdnorm ? &mustd : NULL;
  buffer= create_buffer ();
  fea= tL_fea_new ( TL_FEA_REAL );
  fext= tL_feature_extractor_new ();
  fext->ftype= opts->ftype;
  fext->outfreq= opts->outfreq;
  fext->wduration= opts->wduration;
  fext->numchans= opts->numchans;
  fext->numceps= opts->numceps;
  fext->lofreq= opts->lofreq;
  fext->hifreq= opts->hifreq;
  fext->zeromean= !(opts->dzeromean);
  fext->preemcoef= opts->preemcoef;
  fext->usehamming= !(opts->dhamming);
  fext->ceplifter= opts->ceplifter;
  fext->usepower= !(opts->dpower);
  fext->energy= opts->eenergy;
  fext->deltawindow= opts->deltawindow;
  fext->accwindow= opts->accwindow;
  fext->cep0= !(opts->dcep0);
  fext->meannorm= !(opts->dmeannorm);
  
  /* Extract. */
  for ( n= 0; n < in_lst->N; ++n )
    extract_sample ( buffer, fext, fea, in_lst->files[n],
		     out_lst->files[n], opts->binary,
		     mustdptr, opts->verbose );
  
  /* Mu-std normalization. */
  if ( mustd != NULL )
    {
      if ( opts->verbose ) fprintf ( stderr, _("Normalizing samples\n") );
      tL_mustd_calc ( mustd );
      for ( n= 0; n < out_lst->N; ++n )
	norm_sample ( out_lst->files[n], fea, mustd,
		      opts->binary, opts->verbose );
    }
    
  /* Free memory. */
  if ( mustd != NULL ) tL_mustd_free ( mustd );
  tL_fea_free ( fea );
  tL_feature_extractor_free ( fext );
  free_buffer ( buffer );
  
} /* end extract */




/********/
/* MAIN */
/********/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLFileList *in_lst, *out_lst;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  if ( opts.verbose )
    {
      fprintf ( stderr, _("Type of features: ") );
      switch ( opts.ftype )
	{
	case TL_MELSPEC: fprintf ( stderr, "melspec" ); break;
	case TL_FBANK: fprintf ( stderr, "fbank" ); break;
	case TL_MFCC: fprintf ( stderr, "mfcc" ); break;
	}
      fprintf ( stderr, _("\nNumber of frames per second: %d\n"),
		opts.outfreq );
      fprintf ( stderr, _("Window duration in seconds: %g\n"), opts.wduration );
      fprintf ( stderr, _("Remove DC mean: %s\n"),
		opts.dzeromean?_("No"):_("Yes") );
      fprintf ( stderr, _("Pre-emphasis coefficient: %g\n"), opts.preemcoef );
      fprintf ( stderr, _("Apply Hamming windowing: %s\n"),
		opts.dhamming?_("No"):_("Yes") );
      fprintf ( stderr, _("Number of channels: %d\n"), opts.numchans );
      fprintf ( stderr, _("Use power: %s\n"), opts.dpower?_("No"):_("Yes") );
      if ( opts.ftype == TL_MFCC )
	{
	  fprintf ( stderr, _("Number of cepstral coefficients: %d\n"),
		    opts.numceps );
	  fprintf ( stderr, _("Cepstral lifter value: %d\n"), opts.ceplifter );
	}
      fprintf ( stderr, _("Calculate energy: %s\n"),
		opts.eenergy?_("Yes"):_("No") );
      fprintf ( stderr, _("Calculate cepstrum 0: %s\n"),
		opts.dcep0?_("No"):_("Yes") );
      fprintf ( stderr, _("Mean normalization: %s\n"),
		opts.dmeannorm?_("No"):_("Yes") );
      fprintf ( stderr, _("Delta coefficients window: %d\n"),
		opts.deltawindow );
      fprintf ( stderr, _("Acceleration coefficients window: %d\n"),
		opts.accwindow );
      fprintf ( stderr, _("High frequency: %d\n"), opts.hifreq );
      fprintf ( stderr, _("Low frequency: %d\n"), opts.lofreq );
      fprintf ( stderr, _("Mean-variance normalization (all samples): %s"),
		opts.mustdnorm ? _("Yes") : _("No") );
    }
  
  if ( sox_init () != SOX_SUCCESS ) tL_error ( _("Unable to initialize SoX") );
  sox_globals.repeatable= sox_true;
  sox_globals.verbosity= opts.verbose;
  in_lst= load_list ( args.input_fn, TRUE, opts.verbose );
  out_lst= load_list ( args.output_fn, FALSE, opts.verbose );
  if ( in_lst->N != out_lst->N )
    tL_error ( _("The number of samples in '%s' and '%s' differs"),
	       args.input_fn, args.output_fn );
  extract ( in_lst, out_lst, &opts );
  tL_filelist_free ( in_lst );
  tL_filelist_free ( out_lst );
  sox_quit ();
  
  return EXIT_SUCCESS;
  
}
