/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  hinit_new.c - Inicialització a la HTK (HInit).
 *
 *  Adrià Gimenez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "macros.h"
#include "write_amodel.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 2


#define ERROR_LINE						\
  tL_error ( _("'%s' [line %lu]: 4 fields were expected"),	\
	     file_name, _PS(line) )


#define free_list(LIST) free ( (LIST).v )




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int           verbose;
  const char   *output_fn;
  tLFloat       smooth;
  tLFloat       cfactor;
  unsigned int  max_iters;
  tLBool        smooth_trans;
  tLBool        binary;
  
};


struct args
{
  
  const char *amodel_fn;
  const char *aligns_fn;
  
};


struct seg
{
  
  unsigned int c;
  int          pos;
  int          length;
  
};


struct sfile
{
  
  struct seg *segs;
  size_t      N;
  
};


struct sfilelist
{
  
  tLDict       *fnames;
  struct sfile *sfiles;
  size_t        N;
  
};


struct list
{
  
  size_t *v;
  size_t  N;
  
};


struct mlv
{
  
  tLFloat current;
  tLFloat prev;
  tLBool  mod;
  int     N;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <amodel> <aligns>\n\n", prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Initial acoustic model") );
  fprintf ( to, "    <aligns> %s\n\n",
	    _("Alignment file with the format returned by 'tLalign'") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -B,--binary: %s\n",
	    _("Print the acoustic model in binary format") );
  fprintf ( to, "      -c,--cfactor <double>: %s\n",
	    _("Convergence factor. Default is 0.0001") );
  fprintf ( to, "      -m,--max-iters <uint>: %s\n",
	    _("Maximum number of iterations. Default is 20") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the initialized acoustic model will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -s,--smooth <double>: %s\n",
	    _("Smoothing factor. Default is 1e-6") );
  fprintf ( to, "      --raw-transitions: %s\n",
	    _("Smoothing will not be applied to transition probabilities") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{

  cmdl_t cmdl;
  int chelp, cversion, n, csmooth_trans, cbinary;
  const char *smooth, *prog, *cfactor, *max;


  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_counter ( cmdl, 'B', "binary", &(cbinary) );
  add_cmdl_value ( cmdl, 'c', "cfactor", &cfactor );
  add_cmdl_value ( cmdl, 'm', "max-iters", &max );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 's', "smooth", &smooth );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "raw-transitions",
                     &csmooth_trans );

  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( cfactor != NULL )
    {
      opts->cfactor= atof ( cfactor );
      if ( opts->cfactor <= 0.0 )
        tL_error ( _("The convergence factor must be greater than 0") );
    }
  else opts->cfactor= 0.0001;
  if ( max != NULL )
    {
      opts->max_iters=
        (unsigned int) strtol ( max, (char **) NULL, 10 );
    }
  else opts->max_iters= 20;
  if ( smooth != NULL )
    {
      opts->smooth= atof ( smooth );
      if ( opts->smooth < 0.0 || opts->smooth > 1.0 )
        tL_error ( _("The smoothing factor must be in the range [0,1]") );
    }
  else opts->smooth= 1e-6;
  opts->smooth_trans= (csmooth_trans == 0);
  opts->binary= (cbinary != 0);
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->amodel_fn= argv[0];
  args->aligns_fn= argv[1];
  
} /* end usage */


static struct sfilelist *
create_sfilelist (
		  const char *file_name,
		  tLDict     *syms,
		  int         verbose
		  )
{
  
  gzFile f;
  struct sfilelist *new;
  size_t size, *sizes, n, line, id, nsize;
  tLBuffer *buffer;
  tLDict *fnames;
  char *begin, *end, *name;
  struct seg seg;
  
  
  if ( verbose )
    fprintf ( stderr, _("Reading samples\n") );
  
  /* Inicialitza. */
  new= mem_alloc ( struct sfilelist, 1 );
  new->fnames= fnames= tL_dict_new ();
  new->N= 0;
  new->sfiles= mem_alloc ( struct sfile, 1 );
  new->sfiles[0].N= 0;
  new->sfiles[0].segs= mem_alloc ( struct seg, 1 );
  sizes= mem_alloc ( size_t, 1 );
  size= 1;
  sizes[0]= 1;
  
  /* Llig. */
  f= gzopen ( file_name, "r" );
  if ( f == NULL )
    tL_error ( _("Could not open file '%s'"), file_name );
  
  buffer= tL_buffer_new ();
  line= 1;
  while ( tL_gline ( f, buffer ) )
    {
      
      /* Caràcter. */
      if ( tL_gtokenstr ( buffer->v, &begin, &end ) )
        ERROR_LINE;
      *end= '\0';
      seg.c= tL_dict_find ( syms, begin );
      if ( seg.c == syms->N )
        tL_error ( _("'%s' [line %lu]: '%s' is not in the dictionary"),
		   file_name, _PS(line), begin );
      
      /* Posició inicial. */
      if ( tL_gtokenstr ( end+1, &begin, &end ) )
        ERROR_LINE;
      *end= '\0';
      seg.pos= atoi ( begin );
      if ( seg.pos < 0 )
        tL_error ( _("'%s' [line %lu]: the position is negative"),
		   file_name, _PS(line) );
      
      /* Llargària. */
      if ( tL_gtokenstr ( end+1, &begin, &end ) )
        ERROR_LINE;
      *end= '\0';
      seg.length= atoi ( begin );
      if ( seg.length < 1 )
        tL_error ( _("'%s' [line %lu]: length is less than 1"),
		   file_name, _PS(line) );
      
      /* Fitxer i inserta. */
      name= ++end;
      if ( tL_dict_insert ( fnames, end, &id ) )
	{
	  if ( new->N == size )
	    {
	      nsize= size*2;
	      new->sfiles=
		   mem_realloc ( struct sfile, new->sfiles, nsize );
	      sizes= mem_realloc ( size_t, sizes, nsize );
	      for ( n= size; n < nsize; ++n )
		{
		  new->sfiles[n].N= 0;
		  new->sfiles[n].segs= mem_alloc ( struct seg, 1 );
		  sizes[n]= 1;
		}
	      size= nsize;
	    }
	  ++(new->N);
	}
      if ( new->sfiles[id].N == sizes[id] )
	{
	  nsize= sizes[id]*2;
	  new->sfiles[id].segs=
	       mem_realloc ( struct seg, new->sfiles[id].segs, nsize );
	  sizes[id]= nsize;
	}
      new->sfiles[id].segs[new->sfiles[id].N++]= seg;
      
      ++line;
      
      if ( verbose > 1 )
        fprintf ( stderr, "  - '%s' %d %d '%s'\n",
                  tL_dict_get_token ( syms, seg.c ), seg.pos,
                  seg.length, name );
      
    }
  tL_buffer_free ( buffer );
  
  gzclose ( f );
  
  /* Ajusta memòria. */
  if ( new->N < size )
    {
      for ( n= new->N; n < size; ++n )
	free ( new->sfiles[n].segs );
      new->sfiles= mem_realloc ( struct sfile, new->sfiles, new->N );
    }
  for ( n= 0; n < new->N; ++n )
    if ( new->sfiles[n].N < sizes[n] )
      new->sfiles[n].segs=
	   mem_realloc ( struct seg, new->sfiles[n].segs, sizes[n] );
  free ( sizes );
  
  return new;
  
} /* end create_sfilelist */


static void
free_sfilelist (
		struct sfilelist *sfilelist
		)
{
  
  size_t n;
  
  
  for ( n= 0; n < sfilelist->N; ++n )
    free ( sfilelist->sfiles[n].segs );
  free ( sfilelist->sfiles );
  tL_dict_free ( sfilelist->fnames );
  free ( sfilelist );
  
} /* end free_sfilelist */


static void
init_list (
	   struct list *list,
	   size_t       size
	   )
{
  
  list->v= mem_alloc ( size_t, size );
  list->N= 0;
  
} /* end init_list */


static void
init_suniform (
	       tLAModel               *amodel,
	       const tLDict           *syms,
	       tLAMiCounts            *counts,
	       const struct sfilelist *sflist,
	       tLFea                  *fea,
	       tLBool                 *actives,
	       tLHMMInitializer       *init,
	       int                     verbose
	       )
{
  
  tLUniHMMInitializer *iuc_aux;
  char *err;
  unsigned int c;
  const struct sfile *sfiles;
  struct sfile sfile;
  size_t n, n2;
  struct seg seg;
  const char *fname;
  
  
  if ( verbose )
    fprintf ( stderr, _("Initializing through uniform segmentation:\n") );
  
  /* Prepara. */
  iuc_aux= tL_uni_hmm_initializer_new ( amodel, &err );
  if ( iuc_aux == NULL ) tL_error ( err );
  tL_amicounts_init ( counts, amodel );
  for ( c= 0; c < amodel->N; ++c ) actives[c]= FALSE;
  
  /* Processa mostres. */
  sfiles= sflist->sfiles;
  for ( n= 0; n < sflist->N; ++n )
    {
      sfile= sfiles[n];
      fname= tL_dict_get_token ( sflist->fnames, n );
      if ( tL_fea_load ( fea, fname, &err ) == -1 ) tL_error ( err );
      if ( !tL_amodel_test ( amodel, fea, &err ) )
        {
          tL_warning ( _("'%s': will not be used for initialization,"
			 " because the image doesn't pass the global test: %s"),
			 fname, err );
          free ( err );
          continue;
        }
      for ( n2= 0; n2 < sfile.N; ++n2 )
	{
	  seg= sfile.segs[n2];
	  if ( (seg.pos + seg.length) > fea->cols )
	    tL_error ( _("[%d,%d] '%s': incorrect segment"),
		       seg.pos, seg.pos + seg.length - 1, fname );
	  /* NOTA: que siga un subsegment no importa perquè el test
	   * sempre està relacionat en la grandària dels vectors.
	   */
	  if ( seg.length < amodel->hmms[seg.c].Q )
	    {
	      tL_warning ( _("[%d,%d] '%s': will not be used for"
			     " initialization, it has %d columns"
			     " but at least %d were expected"),
			   seg.pos, seg.pos + seg.length - 1,
			   fname, seg.length, amodel->hmms[seg.c].Q );
	      continue;
	    }
	  actives[seg.c]= TRUE;
	  tL_init_hmm_seg ( init, counts, seg.c, fea, seg.pos, seg.length );
	}
    }
  
  /* Actualitza i llibera memòria. */
  for ( c= 0; c < amodel->N; ++c )
    if ( !actives[c] )
      {
	tL_warning ( _("The symbol '%s' has no segments and will be uniformly"
		       " initialized"), tL_dict_get_token ( syms, c ) );
	tL_init_uniform_hmm ( iuc_aux, counts, c );
      }
  tL_amicounts_update_amodel ( counts, amodel );
  tL_uni_hmm_initializer_free ( iuc_aux );
  
} /* init_suniform */


static void
train_viterbi_iter (
		    tLHMMVTrainer     *cvtrainer,
		    tLAMiCounts       *counts,
		    tLFea             *fea,
		    struct sfilelist  *sflist,
		    struct list       *files,
		    const struct list *as,
		    const tLBool      *actives,
		    struct mlv        *mlv
		    )
{
  
  const struct sfile *sfiles;
  struct sfile sfile;
  size_t n, id, n2;
  const tLDict *fnames;
  const char *fname;
  tLBool loaded;
  struct seg seg;
  char *err;
  tLFloat ret;
  
  
  for ( n= 0; n < as->N; ++n )
    {
      n2= as->v[n];
      mlv[n2].mod= FALSE;
      mlv[n2].current= 0.0;
      mlv[n2].N= 0;
      tL_amicounts_init_hmm ( counts, cvtrainer->amodel, (unsigned int) n2 );
    }
  sfiles= sflist->sfiles;
  fnames= sflist->fnames;
  n= 0;
  while ( n < files->N )
    {
      id= files->v[n];
      sfile= sfiles[id];
      fname= tL_dict_get_token ( fnames, id );
      loaded= FALSE;
      for ( n2= 0; n2 < sfile.N; ++n2 )
	{
	  seg= sfile.segs[n2];
	  if ( !actives[seg.c] ) continue;
	  if ( seg.length < cvtrainer->amodel->hmms[seg.c].Q )
	    {
	      tL_warning ( _("[%d,%d] '%s': will not be used for"
			     " initialization, it has %d columns"
			     " but at least %d were expected"),
			   seg.pos, seg.pos + seg.length - 1,
			   fname, seg.length,
			   cvtrainer->amodel->hmms[seg.c].Q );
	      continue;
	    }
	  if ( !loaded )
	    {
	      if ( tL_fea_load ( fea, fname, &err ) == -1 ) tL_error ( err );
	      if ( !tL_amodel_test ( cvtrainer->amodel, fea, &err ) )
		{
		  tL_warning ( _("'%s': will not be used for initialization,"
				 " because the image doesn't pass"
				 " the global test: %s"),
			       fname, err );
		  free ( err );
		  goto next;
		}
	      loaded= TRUE;
	    }
	  if ( (seg.pos + seg.length) > fea->cols )
	    tL_error ( _("[%d,%d] '%s': incorrect segment"),
		       seg.pos, seg.pos + seg.length - 1, fname );
	  ret= tL_train_viterbi_hmm_seg ( cvtrainer, counts, seg.c, fea,
					  seg.pos, seg.length );
	  if ( ret != -HUGE_VAL )
	    {
	      mlv[seg.c].current+= ret;
	      mlv[seg.c].mod= TRUE;
	      ++(mlv[seg.c].N);
	    }
	  else
	    tL_warning ( _("[%d,%d] '%s': too short, will be ignored"),
			 seg.pos, seg.pos + seg.length - 1, fname );
	}
    next:
      if ( loaded ) ++n;
      else          files->v[n]= files->v[--(files->N)];
    }
  for ( n= 0; n < as->N; ++n )
    {
      n2= as->v[n];
      mlv[n2].current= mlv[n2].mod ?
	(mlv[n2].current / (tLFloat) mlv[n2].N) : -HUGE_VAL;
    }
  
} /* end train_viterbi_iter */


static void
update_hmms (
	     tLAModel          *amodel,
	     tLAMiCounts       *counts,
	     const struct list *as
	     )
{
  
  size_t n;
  unsigned int c;
  
  
  for ( n= 0; n < as->N; ++n )
    {
      c= (unsigned int) (as->v[n]);
      tL_amicounts_update_hmm ( counts, amodel, c );
    }
  
} /* end update_hmms */


static void
train_viterbi (
	       tLAModel         *amodel,
	       tLAMiCounts      *counts,
	       const tLDict     *syms,
	       struct sfilelist *sflist,
	       tLFea            *fea,
	       tLBool           *actives,
	       unsigned int      max_iters,
	       tLFloat           cfactor,
	       int               verbose
	       )
{
  
  tLHMMVTrainer *tvc_aux;
  char *err;
  struct list files, as;
  size_t n;
  struct mlv *mlv;
  unsigned int c;
  int i;
  tLFloat aux, inc;
  tLBool conv;
  
  
  if ( verbose )
    fprintf ( stderr, _("Viterbi training:\n") );
  
  /* Inicialitza. */
  tvc_aux= tL_hmm_vtrainer_new ( amodel, &err );
  if ( tvc_aux == NULL ) tL_error ( err );
  init_list ( &files, sflist->N );
  for ( n= 0; n < sflist->N; ++n ) files.v[n]= n;
  files.N= sflist->N;
  init_list ( &as, amodel->N );
  for ( c= 0; c < amodel->N; ++c )
    if ( actives[c] ) as.v[as.N++]= c;
  mlv= mem_alloc ( struct mlv, amodel->N );
  if ( verbose )
    fprintf ( stderr, "# SYM\t\tNºIter\t\tMLV\t\tINC\n" );
  
  /* Cos. */
  i= 0;
  train_viterbi_iter ( tvc_aux, counts, fea, sflist,
		       &files, &as, actives, mlv );
  n= 0;
  while ( n < as.N )
    {
      c= (unsigned int) as.v[n];
      aux= mlv[c].prev= mlv[c].current;
      if ( verbose )
	fprintf ( stderr, "%c %s\t\t%d\t\t%g\t\t%g\n",
		  aux==0.0?'.':'*',
		  tL_dict_get_token ( syms, c ),
		  i, _DB(aux), _DB(HUGE_VAL) );
      if ( aux != 0.0 && aux != -HUGE_VAL ) ++n;
      else
	{
	  actives[c]= FALSE;
	  as.v[n]= as.v[--as.N];
	}
    }
  while ( as.N > 0 && i < max_iters )
    {
      update_hmms ( amodel, counts, &as );
      train_viterbi_iter ( tvc_aux, counts, fea, sflist,
			   &files, &as, actives, mlv );
      ++i;
      n= 0;
      while ( n < as.N )
	{
	  c= (unsigned int) as.v[n];
	  aux= mlv[c].current;
	  inc= (aux - mlv[c].prev) / -mlv[c].prev;
	  mlv[c].prev= aux;
	  conv= (inc < cfactor);
	  if ( verbose )
	    fprintf ( stderr, "%c %s\t\t%d\t\t%g\t\t%g\n",
		      conv?'.':'*',
		      tL_dict_get_token ( syms, c ),
		      i, _DB(aux), _DB(inc) );
	  if ( !conv && aux != -HUGE_VAL ) ++n;
	  else
	    {
	      actives[c]= FALSE;
	      as.v[n]= as.v[--as.N];
	    }
	}
    }
  
  /* Allibera memòria. */
  free ( mlv );
  free_list ( as );
  free_list ( files );
  tL_hmm_vtrainer_free ( tvc_aux );
  
} /* end train_viterbi */


static void
hinit (
       tLAModel         *amodel,
       tLDict           *syms,
       struct sfilelist *sflist,
       unsigned int      max_iters,
       tLFloat           cfactor,
       int               verbose
       )
{
  
  tLAMiCounts *counts;
  tLFea *fea;
  tLBool *actives;
  tLHMMInitializer *ic_aux;
  char *err;
  
  
  if ( verbose )
    fprintf ( stderr, _("Initializing\n") );
  
  /* Inicialitza. */
  counts= tL_amicounts_new ( amodel );
  fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  actives= mem_alloc ( tLBool, amodel->N );
  ic_aux= tL_hmm_initializer_new ( amodel, &err );
  if ( ic_aux == NULL ) tL_error ( err );
  
  /* Cos. */
  init_suniform ( amodel, syms, counts, sflist, fea, actives, ic_aux, verbose );
  if ( max_iters > 0 )
    train_viterbi ( amodel, counts, syms, sflist, fea,
		    actives, max_iters, cfactor, verbose );
  
  /* Allibera memòria. */
  free ( actives );
  tL_fea_free ( fea );
  tL_hmm_initializer_free ( ic_aux );
  tL_amicounts_free ( counts, amodel );
  
} /* end hinit */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel;
  struct sfilelist *sflist;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  sflist= create_sfilelist ( args.aligns_fn, syms, opts.verbose );
  
  /* Cos. */
  hinit ( amodel, syms, sflist, opts.max_iters, opts.cfactor, opts.verbose );
  tL_amodel_smooth ( amodel, opts.smooth, opts.smooth_trans );
  write_amodel ( opts.output_fn, amodel, syms, opts.binary );
  
  /* Allibera memòria. */
  free_sfilelist ( sflist );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
