/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  init_amodel.c - Inicialitza el model de caràcters.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */

#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "get_pbm_lst.h"
#include "get_words.h"
#include "macros.h"
#include "write_amodel.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 3




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  const char *output_fn;
  tLFloat     smooth;
  tLBool      nowords;
  tLBool      smooth_trans;
  tLBool      binary;
  
};


struct args
{
  
  const char *amodel_fn;
  const char *wdlst_fn;
  const char *pbmlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <amodel> <transc-lst> <sample-lst>\n\n",
	    prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Initial acoustic model") );
  fprintf ( to, "    <transc-lst> %s\n",
	    _("List of the files containing the transcriptions") );
  fprintf ( to, "    <sample-lst> %s\n\n", 
	    _("List of the files containing the samples") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -B,--binary: %s\n",
	    _("Print the model in binary format") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the initialized model will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -s,--smooth <double>: %s\n",
	    _("Smoothing factor. Default is 1e-6") );
  fprintf ( to, "      --words: %s\n",
	    _("Transcriptions will be expected to be words") );
  fprintf ( to, "      --raw-transitions: %s\n",
	    _("Smoothing will not be applied to transition probabilities") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, cwords, n, csmooth_trans, cbinary;
  const char *smooth, *prog;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_counter ( cmdl, 'B', "binary", &(cbinary) );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 's', "smooth", &smooth );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "raw-transitions", &csmooth_trans );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "words", &cwords );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( smooth != NULL )
    {
      opts->smooth= atof ( smooth );
      if ( opts->smooth < 0.0 || opts->smooth > 1.0 )
        tL_error ( _("The smoothing factor must be in the range [0,1]") );
    }
  else opts->smooth= 1e-6;
  opts->nowords= (cwords == 0);
  opts->smooth_trans= (csmooth_trans == 0);
  opts->binary= (cbinary != 0);
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->amodel_fn= argv[0];
  args->wdlst_fn= argv[1];
  args->pbmlst_fn= argv[2];
  
} /* end usage */


static void
init_sample (
	     tLInitializer *ini,
	     tLAMiCounts   *counts,
	     tLBool        *wsamples,
	     size_t         word,
	     tLFea         *fea,
	     tLSeq          cs,
	     const char    *pbm
	     )
{
  
  size_t ret;
  int i;
  
  
  ret= tL_init ( ini, counts, fea, word );
  if ( ret == 0 )
    for ( i= 0; i < cs.length; ++i )
      wsamples[cs.v[i]]= TRUE;
  else
    tL_warning ( _("The sample '%s' will not be used for initialization,"
		   " it has %d columns but at least %lu were expected"),
		   pbm, fea->cols, _PS(ret) );
  
} /* end init_sample */


static void
init (
      tLAModel   *amodel,
      tLDict     *syms,
      tLSeqs     *words,
      tLFileList *pbm_lst,
      int         verbose
      )
{
  
  size_t n, word;
  tLAMiCounts *counts;
  tLFea *fea;
  tLSeq cs;
  tLBool *wsamples;
  tLInitializer *ini;
  tLUniHMMInitializer *iuc_aux;
  char *err;
  
  
  if ( verbose )
    fprintf ( stderr, _("Initializing\n") );
  
  /* Reserva memòria. */
  ini= tL_initializer_new ( amodel, words, &err );
  if ( ini == NULL ) tL_error ( err );
  iuc_aux= tL_uni_hmm_initializer_new ( amodel, &err );
  if ( iuc_aux == NULL ) tL_error ( err );
  counts= tL_amicounts_new ( amodel );
  tL_amicounts_init ( counts, amodel );
  fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  wsamples= mem_alloc ( tLBool, syms->N );
  for ( n= 0; n < syms->N; ++n )
    wsamples[n]= FALSE;
  
  /* Recorre les mostres. */
  for ( n= 0; n < words->N; ++n )
    {
      
      if ( tL_fea_load ( fea, pbm_lst->files[n], &err ) != 0 )
	tL_error ( err );
      word= words->seqs[n];
      cs= *(tL_seqs_get_seq ( words, word ));
      if ( verbose > 1 )
	{
	  fprintf ( stderr, "  - '" );
	  tL_seq_print ( &cs, stderr, syms, ' ' );
	  fprintf ( stderr, "' \"%s\"\n", pbm_lst->files[n] );
	  if ( verbose > 2 ) tL_fea_print ( fea, stderr );
	}
      
      if ( !tL_amodel_test ( amodel, fea, &err ) )
	{
	  tL_warning ( _("The sample '%s' will not be used"
			 " for initialization: %s"),
		       pbm_lst->files[n], err );
	  free ( err );
	  continue;
	}
      
      init_sample ( ini, counts, wsamples, word, fea,
		    cs, pbm_lst->files[n] );
      
    }
  
  /* Inicialització de caràcters sense mostres. */
  for ( n= 0; n < syms->N; ++n )
    if ( !wsamples[n] )
      {
	tL_warning ( _("The symbol '%s' has no samples"
		       " and will be uniformly initialized"),
		      tL_dict_get_token ( syms, n ) );
	 tL_init_uniform_hmm ( iuc_aux, counts, n );
      }
  
  tL_amicounts_update_amodel ( counts, amodel );
  
  /* Allibera memòria. */
  tL_initializer_free ( ini );
  tL_uni_hmm_initializer_free ( iuc_aux );
  free ( wsamples );
  tL_amicounts_free ( counts, amodel );
  tL_fea_free ( fea );
  
} /* end init */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLFileList *pbm_lst;
  tLSeqs *words;
  tLAModel *amodel;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  words= get_words ( args.wdlst_fn, syms, opts.nowords, NULL, opts.verbose );
  if ( words->N != pbm_lst->N )
    tL_error ( _("The sample list contains %lu entries,"
		 " but the transcription list contains %lu entries"),
	       _PS(pbm_lst->N), _PS(words->N) );
  
  /* Cos. */
  init ( amodel, syms, words, pbm_lst, opts.verbose );
  tL_amodel_smooth ( amodel, opts.smooth, opts.smooth_trans );
  write_amodel ( opts.output_fn, amodel, syms, opts.binary );
  
  /* Allibera memòria. */
  tL_seqs_free ( words );
  tL_filelist_free ( pbm_lst );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
