/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  itertrain.c - Implementació de 'itertrain.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "itertrain.h"
#include "macros.h"




static tLProb
itertrain_sample (
		  tLTrainer  *trainer,
		  size_t      seq,
		  const char *pbm,
		  size_t     *N,
		  tLFea      *fea,
		  int         verbose
		  )
{
  
  char *err;
  tLProb prob;
  
  
  /* Prepara. */
  if ( !tL_amodel_test ( trainer->amodel, fea, &err ) )
    {
      tL_warning ( _("The sample '%s' will not be used for training: %s"),
		   pbm, err );
      free ( err );
      return 0.0;
    }
  if ( verbose > 1 )
    fprintf ( stderr,
	      _("Processing '%s'. Rows:%d, columns:%d"),
	      pbm, fea->rows, fea->cols );
  
  /* Acumula. */
  prob= tL_train ( trainer, fea, seq );
  if ( prob == -HUGE_VAL )
    {
      tL_warning ( _("'%s': too short, will be ignored"), pbm );
      return 0.0;
    }
  ++(*N);
  
  if ( verbose > 1 )
    fprintf ( stderr, "  lprob:%g\n", _DB(prob) );
  
  return prob;
  
} /* end itertrain_sample */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

feablock_t *
create_feablock (
		 const unsigned int  size,
		 const tLAModel     *amodel
		 )
{
  
  feablock_t *new;
  unsigned int n;
  
  
  new= mem_alloc ( feablock_t, 1 );
  new->size= size;
  new->feas= mem_alloc ( tLFea *, size );
  for ( n= 0; n < size; ++n )
    new->feas[n]= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  new->index= mem_alloc ( int, size );
  
  return new;
  
} /* end create_feablock */


void
free_feablock (
	       feablock_t *feablock
	       )
{
  
  unsigned int n;
  
  
  for ( n= 0; n < feablock->size; ++n )
    tL_fea_free ( feablock->feas[n] );
  free ( feablock->feas );
  free ( feablock->index );
  free ( feablock );
  
} /* end free_feablock */


tLProb
itertrain (
	   tLTrainer  *trainer,
	   feablock_t *feablock,
	   tLFileList *pbm_lst,
	   tLTar      *tar,
	   int         verbose
	   )
{
  
  tLProb lv;
  int n, ind;
  unsigned int i;
  size_t N_samples;
  char *err;
  gzFile f;
  
  
  /* Inicialitza. */
  tL_amlogcounts_init ( trainer->counts, trainer->amodel );
  if ( trainer->occs != NULL )
    tL_occupancies_init ( trainer->occs, trainer->amodel );
  lv= 0.0;
  
  N_samples= 0;
  n= 0;
  do {
    feablock->N= 0;
    for ( i= 0; i < feablock->size && n < trainer->seqs->N; ++i, ++n )
      {
	if ( tar == NULL )
	  {
	    if ( tL_fea_load ( feablock->feas[i],
			       pbm_lst->files[n], &err ) == -1 )
	      goto error;
	  }
	else
	  {
	    f= tL_tar_get_file ( tar, pbm_lst->files[n], NULL, &err );
	    if ( f == NULL ) goto error;
	    if ( tL_fea_load_fd ( feablock->feas[i], f, &err ) == -1 )
	      goto error;
	  }
	feablock->index[i]= n;
	++(feablock->N);
      }
    for ( i= 0; i < feablock->N; ++i )
      {
	ind= feablock->index[i];
	lv+= itertrain_sample ( trainer, trainer->seqs->seqs[ind],
				pbm_lst->files[ind], &N_samples,
				feablock->feas[i], verbose );
      }
  } while ( n != trainer->seqs->N );
  
  return (N_samples > 0) ? (lv / trainer->seqs->N) : -HUGE_VAL;
  
 error:
  tL_error ( _("Error while reading '%s': %s"), pbm_lst->files[n], err );
  return -HUGE_VAL; /* CALLA! */
  
} /* end itertrain */
