/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  abmat.c - Implementació de 'abmat.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012.
 *
 */


#include <stddef.h>
#include <stdlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "abmat.h"
#include "amodel_itrans.h"
#include "error.h"
#include "mem.h"




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static tLFloat
calc_alpha_sum (
                const tLFloat         IF,
                const tLFloat        *a,
                const itrans_state_t  state,
                tLFloat              *calc
                )
{
  
  const itrans_edge_t *ed;
  tLFloat max, aux;
  volatile tLFloat fix;
  unsigned int N;
  
  
  N= 0; max= -HUGE_VAL;
  for ( ed= state.begin; ed != state.end; ++ed )
    {
      aux= fix= *(ed->p) + a[ed->q];
      if ( aux != -HUGE_VAL )
        {
          calc[N++]= aux;
          if ( aux > max ) max= aux;
        }
    }
  if ( state.ip != NULL )
    {
      aux= fix= *(state.ip) + IF;
      if ( aux != -HUGE_VAL )
        {
          calc[N++]= aux;
          if ( aux > max ) max= aux;
        }
    }
  
  return (tLFloat) ((N > 1) ? tL_radd_vals ( calc, max, N ) : max);
  
} /* end calc_alpha_sum */


static void
calc_alpha_t0 (
	       const fb_meths_t *fb_meths,
               tLFloat          *aI,
               abmat_sym_t      *a_t0,
               tLFloat          *calc,
               const tLFea      *fea,
	       void             *udata
               )
{
  
  tLFloat aIF, aIF2, *a_ts, aux;
  unsigned int q;
  const tLAModelHMM *cchar;
  const tLAModelITrans *itrans;
  const tLAModelTrans *trans;
  size_t s, last_s, length;
  const tLProb *ip;
  
  
  length= fb_meths->get_length ( udata );
  last_s= length-1;
  aIF= 0.0; aIF2= -HUGE_VAL;
  for ( s= 0; s <= last_s; ++s )
    {
      cchar= fb_meths->a_set_sym ( s, udata );
      a_ts= a_t0[s].v;
      trans= &(cchar->trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
        {
        case TLLINEAL: /* LINEAL. */
          if ( aIF != -HUGE_VAL )
            {
              a_ts[0]= aIF + fb_meths->a_get_prob ( fea, 0, cchar->states,
						    0, udata );
              aIF= -HUGE_VAL;
              a_t0[s].IF=
                (cchar->Q == 1 && s != last_s) ?
                a_ts[0] + trans->lineal.p[0].p1 :
                -HUGE_VAL;
            }
	  else a_ts[0]= a_t0[s].IF= -HUGE_VAL;
          for ( q= 1; q < cchar->Q; ++q ) a_ts[q]= -HUGE_VAL;
          break;
        case TLLIST: /* LLISTA. */
          itrans= trans->list.itrans;
          if ( aIF != -HUGE_VAL )
            {
              for ( q= 0; q < cchar->Q; ++q )
                {
                  if ( (ip= itrans->states[q].ip) != NULL &&
                       (aux= aIF + *ip) != -HUGE_VAL )
                    a_ts[q]= aux + fb_meths->a_get_prob ( fea, 0,
							  cchar->states,
							  q, udata );
                  else a_ts[q]= -HUGE_VAL;
                }
              if ( s != last_s && (ip= itrans->final.ip) != NULL ) aIF+= *ip;
              else aIF= -HUGE_VAL;
            }
          else for ( q= 0; q < cchar->Q; ++q ) a_ts[q]= -HUGE_VAL;
          a_t0[s].IF=
            ( s != last_s ) ?
            calc_alpha_sum ( aIF2, a_ts, itrans->final, calc ) :
            -HUGE_VAL;
          break;
        default:
          tL_error ( _( "%s - calc_alpha_t0: This should not happen!" ), __FILE__ );
        }
      aIF2= a_t0[s].IF;
      aI[s]= aIF;
    }
  
} /* end calc_alpha_t0 */


/* NOTA: 'remain' és el número de frames que falten contant al frame
   actual. */
static void
calc_alpha_t (
	      const fb_meths_t *fb_meths,
              abmat_sym_t      *a_t0,
              abmat_sym_t      *a_t1,
              tLFloat          *calc,
              const tLFea      *fea,
              int               t,
              const int        *dist,
              int               remain,
	      void             *udata
              )
{
  
  tLFloat aIF, aIF2, *a_t0s, *a_t1s, aux, a, b;
  unsigned int q;
  const tLAModelHMM *cchar;
  const tLAModelITrans *itrans;
  const tLAModelTrans *trans;
  size_t s, last_s, i, length;
  const tLProbPair *p;
  const itrans_state_t *states;  
  
  
  length= fb_meths->get_length ( udata );
  i= 0;
  last_s= length-1;
  aIF= aIF2= -HUGE_VAL;
  for ( s= 0; s <= last_s; ++s )
    {
      cchar= fb_meths->a_set_sym ( s, udata );
      a_t0s= a_t0[s].v;
      a_t1s= a_t1[s].v;
      trans= &(cchar->trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
        {
        case TLLINEAL: /* LINEAL. */
          p= trans->lineal.p;
          if ( dist[i++] < remain )
            {
              b= a_t0s[0] + p[0].p0;
              aux= RADD ( aIF, b );
              a_t1s[0]=
                (aux == -HUGE_VAL) ? -HUGE_VAL :
                aux + fb_meths->a_get_prob ( fea, t, cchar->states, 0, udata );
            }
          else a_t1s[0]= -HUGE_VAL;
          for ( q= 1; q < cchar->Q; ++q )
            if ( dist[i++] < remain )
              {
                a= a_t0s[q-1] + p[q-1].p1;
                b= a_t0s[q] + p[q].p0;
                aux= RADD ( a, b );
                a_t1s[q]=
                  (aux == -HUGE_VAL) ? -HUGE_VAL :
                  aux + fb_meths->a_get_prob ( fea, t, cchar->states,
					       q, udata );
              }
            else a_t1s[q]= -HUGE_VAL;
          a_t1[s].IF=
            (s != last_s) ?
            a_t1s[cchar->Q-1] + p[cchar->Q-1].p1 :
            -HUGE_VAL;
          break;
        case TLLIST: /* LLISTA. */
          itrans= trans->list.itrans;
          states= itrans->states;
          for ( q= 0; q < cchar->Q; ++q )
            if ( dist[i++] < remain )
              {
                aux= calc_alpha_sum ( aIF, a_t0s, states[q], calc );
                a_t1s[q]=
                  (aux == -HUGE_VAL) ? -HUGE_VAL :
                  aux + fb_meths->a_get_prob ( fea, t, cchar->states,
					       q, udata );
              }
            else a_t1s[q]= -HUGE_VAL;
          a_t1[s].IF= 
            (s != last_s) ?
            calc_alpha_sum ( aIF2, a_t1s, itrans->final, calc ) :
            -HUGE_VAL;
          break;
        default:
          tL_error ( _( "%s - calc_alpha_t: This should not happen" ), __FILE__ );
        }
      aIF= a_t0[s].IF; aIF2= a_t1[s].IF;
    }
  
} /* end calc_alpha_t */


static void
calc_beta_tT (
	      const fb_meths_t  *fb_meths,
              abmat_sym_t       *a_t1,
              abmat_sym_t       *b_t1,
              const int          T,
	      void              *udata
              )
{
  
  size_t s;
  unsigned int q;
  const tLAModelHMM *cchar;
  tLFloat *a_ts, *b_ts, bIF, alpha, aux;
  const tLAModelTrans *trans;
  const tLAModelTransListState *states;
  
  
  bIF= 0.0;
  s= fb_meths->get_length ( udata );
  do {
  start:
    cchar= fb_meths->b_set_sym ( --s, udata );
    b_ts= b_t1[s].v; b_t1[s].IF= bIF;
    if ( bIF == -HUGE_VAL )
      {
        for ( q= 0; q < cchar->Q; ++q ) b_ts[q]= -HUGE_VAL;
        if ( s == 0 ) return; else goto start;
      }
    a_ts= a_t1[s].v;
    trans= &(cchar->trans);
    if ( trans->type == TLPOINTER )
      {
        /* Sols m'interesen les transicions. */
	fb_meths->b_reset_counts ( trans->pointer.c, udata );
        trans= trans->pointer.ptr;
      }
    switch ( trans->type )
      {
      case TLLINEAL:
        for ( q= 0; q < cchar->Q-1; ++q ) b_ts[q]= -HUGE_VAL;
        if ( (alpha= a_ts[q]) == -HUGE_VAL ) b_ts[q]= -HUGE_VAL;
        else
          {
            b_ts[q]= aux= trans->lineal.p[q].p1 + bIF;
            aux+= alpha;
	    fb_meths->b_accum_weight ( q, T, aux, udata );
	    if ( aux != -HUGE_VAL )
	      fb_meths->b_accum_trans_p1 ( aux, q, udata );
          }
        bIF= -HUGE_VAL; /* Ací sempre és així. */
        break;
      case TLLIST:
        states= trans->list.states;
        for ( q= 0; q < cchar->Q; ++q )
          if ( (alpha= a_ts[q]) == -HUGE_VAL ) b_ts[q]= -HUGE_VAL;
          else
            {
              b_ts[q]= aux= states[q].fp + bIF;
              aux+= alpha;
	      fb_meths->b_accum_weight ( q, T, aux, udata );
	      if ( aux != -HUGE_VAL )
		fb_meths->b_accum_trans_fp ( aux, q+1, udata );
            }
        if ( s != 0 )
          {
            if ( (alpha= a_t1[s-1].IF) == -HUGE_VAL ) bIF= -HUGE_VAL;
            else
              {
                bIF+= trans->list.init.fp;
		aux= alpha + bIF;
		if ( aux != -HUGE_VAL )
		fb_meths->b_accum_trans_fp ( aux, 0, udata );
              }
          }
        break;
      default:
        tL_error ( _( "%s - calc_beta_tT: This should not happen!" ), __FILE__ );
      }
  } while ( s != 0 );
  
} /* end calc_beta_tT */


static tLFloat
calc_beta_sum_lineal (
		      const fb_meths_t   *fb_meths,
                      const tLFloat       a,
                      const tLFloat       b,
                      const tLFloat       alpha,
                      const unsigned int  q,
		      void               *udata
                      )
{
  
  tLFloat aux;
  
  if ( (aux= alpha+a) != -HUGE_VAL )
    fb_meths->b_accum_trans_p0 ( aux, q, udata );
  if ( (aux= alpha+b) != -HUGE_VAL )
    fb_meths->b_accum_trans_p1 ( aux, q, udata );
  
  return RADD ( a, b );
  
} /* end calc_beta_sum_lineal */


static tLFloat
calc_beta_sum_list (
		    const fb_meths_t             *fb_meths,
                    const tLFloat                 IF,
                    const tLFloat                *b,
                    const tLFloat                 alpha,
                    const tLAModelTransListState  state,
                    tLFloat                      *calc,
                    const int                     t_next,
                    const unsigned int            q_inc,
                    const tLAModelTransListEdge  *base,
		    void                         *udata
                    )
{
  
  const tLAModelTransListEdge *ed;
  tLFloat max, aux;
  volatile tLFloat fix;
  unsigned int N;
  ptrdiff_t pos;
  
  
  N= 0; max= -HUGE_VAL;
  for ( ed= state.begin, pos= state.begin-base; ed != state.end; ++ed, ++pos )
    {
      aux= fix=
	ed->p + fb_meths->b_get_prob ( ed->q, t_next, udata ) + b[ed->q];
      if ( aux != -HUGE_VAL )
        {
          calc[N++]= aux;
          if ( aux > max ) max= aux;
	  fb_meths->b_accum_trans_edge ( alpha+aux, pos, udata );
        }
    }
  aux= fix= state.fp + IF;
  if ( aux != -HUGE_VAL )
    {
      calc[N++]= aux;
      if ( aux > max ) max= aux;
      fb_meths->b_accum_trans_fp ( alpha+aux, q_inc, udata );
    }
  
  return (tLFloat) ((N > 1) ? tL_radd_vals ( calc, max, N ) : max);
  
} /* end calc_beta_sum_list */


static void
calc_beta_t (
	     const fb_meths_t *fb_meths,
             abmat_sym_t      *a_t0,
             abmat_sym_t      *b_t0,
             abmat_sym_t      *b_t1,
             tLFloat          *calc,
             const int         t,
	     void             *udata
             )
{
  
  size_t s;
  unsigned int q;
  const tLAModelHMM *cchar;
  tLFloat *a_t0s, *b_t0s, *b_t1s, bIF, alpha, a, b, beta;
  const tLAModelTrans *trans;
  const tLAModelTransListState *states;
  const tLProbPair *p;
  const tLAModelTransListEdge *base_edges;
  
  
  bIF= -HUGE_VAL;
  s= fb_meths->get_length ( udata );
  do {
    cchar= fb_meths->b_set_sym ( --s, udata );
    b_t0s= b_t0[s].v; b_t0[s].IF= bIF;
    a_t0s= a_t0[s].v;
    b_t1s= b_t1[s].v;
    trans= &(cchar->trans);
    if ( trans->type == TLPOINTER )
      {
        /* Sols m'interesen les transicions. */
	fb_meths->b_reset_counts ( trans->pointer.c, udata );
        trans= trans->pointer.ptr;
      }
    switch ( trans->type )
      {
      case TLLINEAL:
        p= trans->lineal.p;
        for ( q= 0; q < cchar->Q-1; ++q )
          if ( (alpha= a_t0s[q]) == -HUGE_VAL ) b_t0s[q]= -HUGE_VAL;
          else
            { 
              a=
		p[q].p0 + fb_meths->b_get_prob ( q, t+1, udata ) + b_t1s[q];
              b=
		p[q].p1 + fb_meths->b_get_prob ( q+1, t+1, udata ) + b_t1s[q+1];
              b_t0s[q]= beta=
                calc_beta_sum_lineal ( fb_meths, a, b, alpha, q, udata );
	      fb_meths->b_accum_weight ( q, t, alpha+beta, udata );
            }
        if ( (alpha= a_t0s[q]) == -HUGE_VAL ) b_t0s[q]= -HUGE_VAL;
        else
          {
            a= p[q].p0 + fb_meths->b_get_prob ( q, t+1, udata ) + b_t1s[q];
            b= p[q].p1 + bIF;
            b_t0s[q]= beta=
              calc_beta_sum_lineal ( fb_meths, a, b, alpha, q, udata );
	    fb_meths->b_accum_weight ( q, t, alpha+beta, udata );
          }
        if ( s != 0 )
          {
            if ( a_t0[s-1].IF == -HUGE_VAL ) bIF= -HUGE_VAL;
            else bIF= fb_meths->b_get_prob ( 0, t+1, udata ) + b_t1s[0];
            /* Esta transició sempre és 0.0, no cal recalcular-la. */
          }
        break;
      case TLLIST:
        states= trans->list.states;
        base_edges= trans->list.edges;
        for ( q= 0; q < cchar->Q; ++q )
          if ( (alpha= a_t0s[q]) == -HUGE_VAL ) b_t0s[q]= -HUGE_VAL;
          else
            {
              b_t0s[q]= beta=
                calc_beta_sum_list ( fb_meths, bIF, b_t1s, alpha,
				     states[q], calc, t+1, q+1,
                                     base_edges, udata );
	      fb_meths->b_accum_weight ( q, t, alpha+beta, udata );
            }
        if ( s != 0 )
          {
            if ( (alpha= a_t0[s-1].IF) == -HUGE_VAL ) bIF= -HUGE_VAL;
            else
              bIF= calc_beta_sum_list ( fb_meths, bIF, b_t1s, alpha,
                                        trans->list.init, calc, t+1,
					0, base_edges, udata );
          }
        break;
      default:
        tL_error ( _( "%s - calc_beta_t: This should not happen" ), __FILE__ );
      }
  } while ( s != 0 );
  
} /* end calc_beta_t */


static void
calc_beta_tI (
	      const fb_meths_t *fb_meths,
              tLFloat          *aI,
              abmat_sym_t      *b_t1,
              tLFloat          *calc,
	      void             *udata
              )
{
  
  size_t s;
  const tLAModelHMM *cchar;
  tLFloat *b_t1s, bIF, alpha;
  const tLAModelTrans *trans;
  
  
  bIF= -HUGE_VAL;
  s= fb_meths->get_length ( udata );
  do {
    cchar= fb_meths->b_set_sym ( --s, udata );
    b_t1s= b_t1[s].v;
    trans= &(cchar->trans);
    if ( trans->type == TLPOINTER )
      {
        /* Sols m'interesen les transicions. */
	fb_meths->b_reset_counts ( trans->pointer.c, udata );
        trans= trans->pointer.ptr;
      }
    switch ( trans->type )
      {
      case TLLINEAL:
        bIF= 
          ((s == 0) || (aI[s-1] != -HUGE_VAL)) ?
	  fb_meths->b_get_prob ( 0, 0, udata ) + b_t1s[0] : -HUGE_VAL;
        break;
      case TLLIST:
        alpha= (s==0) ? 0.0 : aI[s-1];
        bIF= 
          (alpha != -HUGE_VAL) ?
          calc_beta_sum_list ( fb_meths, bIF, b_t1s, alpha,
                               trans->list.init, calc, 0, 0,
                               trans->list.edges, udata ):
          -HUGE_VAL;
        break;
      default:
        tL_error ( _( "%s - calc_beta_tI: This should not happen" ), __FILE__ );
      }
  } while ( s != 0 );
  
} /* end calc_beta_tI */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_abmat_free (
	       abmat_t *m
	       )
{
  
  free ( m->aI );
  free ( m->a[0][0].v );
  free ( m->a[0] );
  free ( m->a );
  free ( m->b[0][0].v );
  free ( m->b[0] );
  free ( m->b );
  free ( m->calc );
  free ( m );
  
} /* end tL_abmat_free */


abmat_t *
tL_abmat_new (
	      const tLAModel *amodel,
	      unsigned int    max_length
	      )
{
  
  abmat_t *new;
  abmat_sym_t *a_t, *b_t;
  unsigned int c, max_Q;
  tLFloat *mem;
  int t;
  
  
  max_Q= amodel->max_Q;
  mt ( new= mem_alloc ( abmat_t, 1 ) );
  mt ( new->aI= mem_alloc ( tLFloat, max_length ) );
  mt ( new->a= mem_alloc ( abmat_sym_t *, 1 ) );
  mt ( new->a[0]= a_t= mem_alloc ( abmat_sym_t, max_length ) );
  mt ( a_t[0].v= mem= mem_alloc ( tLFloat, max_Q*max_length ) );
  for ( c= 0; c < max_length; ++c, mem+= max_Q )
    a_t[c].v= mem;
  mt ( new->b= mem_alloc ( abmat_sym_t *, 2 ) );
  mt ( new->b[0]= b_t= mem_alloc ( abmat_sym_t, max_length*2 ) );
  new->b[1]= b_t+max_length;
  mt ( b_t[0].v= mem= mem_alloc ( tLFloat, max_Q*max_length*2 ) );
  for ( t= 0; t < 2; ++t )
    {
      b_t= new->b[t];
      for ( c= 0; c < max_length; ++c, mem+= max_Q )
	b_t[c].v= mem;
    }
  mt ( new->calc= mem_alloc ( tLFloat, max_Q+1 ) );
  new->max_width= 1;
  
  return new;
  
} /* end tL_abmat_new */


void
tL_abmat_resize (
		 abmat_t      *m,
		 unsigned int  max_length,
		 unsigned int  max_Q,
		 int           width
		 )
{
  
  abmat_sym_t **a, *a_t;
  int t;
  tLFloat *mem;
  unsigned int c;
  
  
  mt ( m->a= a= mem_realloc ( abmat_sym_t *, m->a, width ) );
  mt ( a[0]= a_t= mem_realloc ( abmat_sym_t, a[0], width*max_length ) );
  for ( t= 0; t < width; ++t, a_t+= max_length )
    a[t]= a_t;
  a_t= a[0];
  mt ( a_t[0].v= mem=
       mem_realloc ( tLFloat, a_t[0].v, width*max_length*max_Q ) );
  for ( t= 0; t < width; ++t )
    {
      a_t= a[t];
      for ( c= 0; c < max_length; ++c, mem+= max_Q )
	a_t[c].v= mem;
    }
  m->max_width= width;
  
} /* end tL_abmat_resize */


tLFloat
tL_calc_alpha (
	       abmat_t          *m,
	       const fb_meths_t *fb_meths,
	       const tLFea      *fea,
	       const int        *dist,
	       void             *udata
	       )
{
  
  abmat_sym_t **a, *a_t0, *a_t1;
  tLFloat *calc_m, *aI;
  int t, remain;
  const tLAModelHMM *cchar;
  const tLAModelTrans *trans;
  size_t last_s, length;
  tLFloat ret;
  
  
  /* Prepara. */
  a= m->a;
  aI= m->aI;
  calc_m= m->calc;
  length= fb_meths->get_length ( udata );
  
  /* Calcula. */
  a_t1= a[0];
  calc_alpha_t0 ( fb_meths, aI, a_t1, calc_m, fea, udata );
  for ( t= 1, remain= fea->cols-1; t < fea->cols; ++t, --remain )
    {
      a_t0= a_t1; a_t1= a[t];
      calc_alpha_t ( fb_meths, a_t0, a_t1, calc_m, fea,
		     t, dist, remain, udata );
    }
  
  /* Calcula probabilitat. */
  last_s= length-1;
  cchar= fb_meths->a_set_sym ( last_s, udata );
  trans= &(cchar->trans);
  if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
  switch ( trans->type )
    {
    case TLLINEAL:
      a_t1[last_s].IF= ret=
        a_t1[last_s].v[cchar->Q-1] +
        trans->lineal.p[cchar->Q-1].p1;
      break;
    case TLLIST:
      a_t1[last_s].IF= ret=
        calc_alpha_sum ( (last_s>0)?a_t1[last_s-1].IF:-HUGE_VAL,
                         a_t1[last_s].v, trans->list.itrans->final,
                         calc_m );
      break;
    default: tL_error ( _( "%s - calc_alpha: This should not happen" ), __FILE__ );
      ret= 0.0; /* CALLAT!!!! */
    }
  
  return ret;
  
} /* end tL_calc_alpha */


void
tL_calc_beta (
	      abmat_t          *m,
	      const fb_meths_t *fb_meths,
	      const int         T,
	      void             *udata
	      )
{
  
  abmat_sym_t **a, **b, *b_t0, *b_t1, *tmp;
  tLFloat *calc_m;
  int t;
  
  
  /* Prepara. */
  a= m->a;
  b= m->b;
  calc_m= m->calc;
  
  /* Calcula. */
  t= T-1;
  b_t0= b[0]; b_t1= b[1];
  calc_beta_tT ( fb_meths, a[t], b_t1, t, udata );
  for ( --t; t >= 0; --t )
    {
      calc_beta_t ( fb_meths, a[t], b_t0, b_t1, calc_m, t, udata );
      tmp= b_t0; b_t0= b_t1; b_t1= tmp;
    }
  calc_beta_tI ( fb_meths, m->aI, b_t1, calc_m, udata );
  
} /* end tL_calc_beta */
