/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  align.c - Realitza aliniament forçat.
 *
 *  Adrià Giménez Pastor, 2008,2010,2011,2012,2013.
 *
 */

#include <assert.h>
#include <limits.h>
#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
  
/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "align.h"
#include "calcelprob.h"
#include "amodel_itrans.h"
#include "common.h"
#include "create_nstates.h"
#include "error.h"
#include "eprobs.h"
#include "eprobst.h"
#include "hypmng.h"
#include "mem.h"
#include "samples.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define CALC_VITERBI_BODY                                    \
  							     \
  for ( q= 1; q < Q; ++q )					\
    {								\
      ++begin; ++end;						\
      tr1[begin].p=						\
        tr0[begin-1].p + pc[q-1].p1 +				\
        tL_eprobs_lprob ( eprobs, amodel, fea, c, q, begin );   \
      if ( statealign )						\
	{							\
          hyp= tr0[begin-1].hyp;				\
	  tr1[begin].hyp= tL_hypmng_add ( hypmng, hyp, begin ); \
	  release_hyp ( hypmng, hyp );				\
	}							\
      else tr1[begin].hyp= tr0[begin-1].hyp;			\
      /* take_hyp(); release_hyp(); */				\
      for ( t= begin+1; t < end; ++t )				\
        {							\
          a= tr0[t-1].p + pc[q-1].p1;				\
          b= tr1[t-1].p + pc[q].p0;				\
          if ( a > b )						\
            {							\
              tr1[t].p= a;					\
	      if ( statealign )					\
		{						\
		  hyp= tr0[t-1].hyp;				\
		  tr1[t].hyp= tL_hypmng_add ( hypmng, hyp, t ); \
		  release_hyp ( hypmng, hyp );			\
		}						\
	      else tr1[t].hyp= tr0[t-1].hyp;			\
              /* take_hyp(); release_hyp(); */			\
            }							\
          else							\
            {							\
              tr1[t].p= b;					\
              hyp= tr1[t].hyp= tr1[t-1].hyp;			\
              take_hyp ( hypmng, hyp );				\
              release_hyp ( hypmng, tr0[t-1].hyp );		\
            }							\
          tr1[t].p+=						\
            tL_eprobs_lprob ( eprobs, amodel, fea, c, q, t );   \
        }							\
      								\
      aux= tr0;							\
      tr0= tr1;							\
      tr1= aux;							\
    }								\
  prob_f= pc[Q-1].p1




/*********/
/* TIPUS */
/*********/

struct tr_node
{
  
  tLProb p;
  size_t hyp;
  
};


struct tr
{
  
  struct tr_node *tr0;
  struct tr_node *tr1;
  int             length;
  
};


typedef struct
{
  
  struct tr_node **tr0;    /* |S|*|Q|+1. Q és el final. */
  struct tr_node **tr1;    /* |S|*|Q|+1. Q és el final. */
  
} tr_gen_t;


struct _tLAligner
{
  
  const tLAModel    *amodel;
  const tLSeqs      *seqs;
  enum {
    LINEAR,
    GENERIC
  }                  type;
  union {
    struct
    {
      struct tr *tr;
      eprobs_t  *eprobs;
      size_t    *nstates;
    } lin;
    struct
    {
      tr_gen_t   *tr;
      eprobst_t  *eprobs;
      int       **dists;
    } gen;
  }                  tr;
  hypmng_t          *hypmng;
  unsigned long int *buffer;
  size_t             bsize;
  tLBool             statealign;
  
};




/************/
/* CONSTANT */
/************/

static const struct tr_node TRN0= { -HUGE_VAL, NULLHYP };




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static int **
create_dists (
	      const tLAModel *amodel,
	      const tLSeqs   *seqs
	      )
{
  
  const tLDict *cseqs;
  int **new, *dist, length, offset;
  const tLAModelHMM *chars, *cchar;
  const tLSeq *cs;
  size_t s, n;
  
  
  cseqs= seqs->cseqs;
  chars= amodel->hmms;
  mt ( new= mem_alloc ( int *, cseqs->N ) );
  for ( n= 0; n < cseqs->N; ++n )
    {
      cs= dict_GCSEQ ( cseqs, n );
      length= 0;
      for ( s= 0; s < cs->length; ++s )
	length+= chars[cs->v[s]].Q;
      mt ( new[n]= dist= mem_alloc ( int, length ) );
      offset= 0;
      s= cs->length;
      do {
	cchar= &(chars[cs->v[--s]]);
	length-= cchar->Q;
	offset= tL_init_state_dists ( &(dist[length]), cchar, offset );
      } while ( s != 0 );
    }
  
  return new;
  
} /* end create_dists */


static void
free_dists (
	    int          **dists,
	    const tLSeqs  *seqs
	    )
{
  
  size_t n;
  
  
  for ( n= 0; n < seqs->cseqs->N; ++n )
    free ( dists[n] );
  free ( dists );
  
} /* end free_dists */


static struct tr *
create_tr ()
{
  
  struct tr *new;
  
  
  mt ( new= mem_alloc ( struct tr, 1 ) );
  mt ( new->tr0= mem_alloc ( struct tr_node, 2 ) );
  new->tr1= new->tr0 + 1;
  new->length= 1;
  
  return new;
  
} /* end create_tr */


static void
free_tr (
	 struct tr *tr
	 )
{
  
  free ( tr->tr0 );
  free ( tr );
  
} /* end free_tr */


/* Sols si és necessari. */
static void
resize_tr (
	   struct tr* tr,
	   int        length
	   )
{
  
  mt ( tr->tr0= mem_realloc ( struct tr_node,
			      tr->tr0, length<<1 ) );
  tr->tr1= tr->tr0 + length;
  tr->length= length;
  
} /* end resize_tr */


static tr_gen_t *
create_tr_gen (
	       const tLAModel *amodel,
	       unsigned int    max_length
	       )
{
  
  unsigned int c, max_Q, aux;
  tr_gen_t *new;
  struct tr_node **mem, *mem2;
  
  
  max_Q= amodel->max_Q+1; /* Per a l'estat final. */
  mt ( new= mem_alloc ( tr_gen_t, 1 ) );
  aux= 2*max_length;
  mt ( new->tr0= mem= mem_alloc ( struct tr_node *, aux ) );
  mt ( mem[0]= mem2= mem_alloc ( struct tr_node, aux*max_Q ) );
  for ( c= 0; c < aux; ++c, mem2+= max_Q )
    mem[c]= mem2;
  new->tr1= new->tr0 + max_length;
  
  return new;
  
} /* end create_tr_gen */


static void
free_tr_gen (
	     tr_gen_t *tr
	     )
{
  
  free ( tr->tr0[0] );
  free ( tr->tr0 );
  free ( tr );
  
} /* end free_tr_gen */


static struct tr_node
viterbi_gen_max (
		 const struct tr_node  I,
		 struct tr_node       *tr,
		 const itrans_state_t  state,
		 hypmng_t             *hypmng,
		 const int             statealign,
		 const int             t,
		 const unsigned int    next_q
		 )
{
  
  tLBool prev_is_regular;
  unsigned int prev;
  tLProb aux;
  struct tr_node ret;
  const itrans_edge_t *ed;
  size_t h;
  
  
  prev= 0; /*CALLA!!!*/
  prev_is_regular= TRUE;
  ret= TRN0;
  for ( ed= state.begin; ed != state.end; ++ed )
    {
      aux= *(ed->p) + tr[ed->q].p;
      if ( aux > ret.p )
	{
	  ret.p= aux;
	  prev= ed->q;
	  ret.hyp= tr[prev].hyp;
	}
    }
  if ( state.ip != NULL )
    {
      aux= *(state.ip) + I.p;
      if ( aux > ret.p )
        {
	  ret.p= aux;
	  ret.hyp= I.hyp;
	  prev_is_regular= FALSE;
	}
    }
  if ( ret.hyp != NULLHYP )
    {
      if ( statealign && prev_is_regular && prev != next_q )
	{
	  h= tL_hypmng_add ( hypmng, ret.hyp, t );
	  ret.hyp= tL_hypmng_add ( hypmng, h, prev );
	  release_hyp ( hypmng, h );
	}
      else take_hyp ( hypmng, ret.hyp );
    }
  
  return ret;
  
} /* end viterbi_gen_max */


static void
viterbi_gen_t0 (
		const tLAModel  *amodel,
		eprobst_t       *eprobs,
		hypmng_t        *hypmng,
		struct tr_node **tr0,
		const tLFea     *fea,
		tLSeq           cs,
		int             statealign
		)
{
  
  const tLAModelHMM *chars;
  size_t s, last_s, h, h2;
  unsigned int c, q;
  tLAModelHMM cchar;
  struct tr_node *tr_s, trI, trI2;
  const tLAModelTrans *trans;
  const tLAModelITrans *itrans;
  const tLProb *ip;
  tLProb aux;
  
  
  /* NOTA: Ninguna hipòtesis activa pot ser NULLHYP. */
  chars= amodel->hmms;
  last_s= cs.length-1;
  trI.p= 0.0; trI.hyp= tL_hypmng_add ( hypmng, NULLHYP,
				       statealign ? ULONG_MAX : 0 );
  trI2= TRN0;
  for ( s= 0; s <= last_s; ++s )
    {
      c= cs.v[s];
      cchar= chars[c];
      tr_s= tr0[s];
      trans= &(cchar.trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
        {
	case TLLINEAL: /* LINEAL. */
	  if ( trI.hyp != NULLHYP /*Podria haver consultat trI.p*/ )
	    {
	      tr_s[0].p= trI.p + tL_eprobst_lprob ( eprobs, amodel,
						    fea, c, 0, 0 );
	      tr_s[0].hyp= trI.hyp;
	      /*take_hyp ( hypmng, trI.hyp );
		release_hyp ( hypmng, trI.hyp ); */
	    }
	  else tr_s[0]= TRN0;
          for ( q= 1; q < cchar.Q; ++q ) tr_s[q]= TRN0;
	  trI= TRN0;
	  if ( cchar.Q == 1 && s != last_s )
	    {
	      trI2.p= tr_s[0].p + trans->lineal.p[0].p1;
	      if ( trI2.p != -HUGE_VAL )
		{
		  if ( statealign )
		    {
		      h= tL_hypmng_add ( hypmng, tr_s[0].hyp, 0 ); /*t*/
		      h2= tL_hypmng_add ( hypmng, h, 0 ); /*q*/
		      release_hyp ( hypmng, h );
		      trI2.hyp= tL_hypmng_add ( hypmng, h2, ULONG_MAX );
		      release_hyp ( hypmng, h2 );
		    }
		  else
		    trI2.hyp= tL_hypmng_add ( hypmng, tr_s[0].hyp, 0 ); /*t*/
		}
	      else trI2.hyp= NULLHYP;
	      tr_s[cchar.Q]= trI2;
	    }
	  else tr_s[cchar.Q]= trI2= TRN0;
	  break;
	case TLLIST: /* LLISTA. */
	  itrans= trans->list.itrans;
	  if ( trI.hyp != NULLHYP )
	    {
	      for ( q= 0; q < cchar.Q; ++q )
		{
		  if ( (ip= itrans->states[q].ip) != NULL &&
		       (aux= trI.p + *ip) != -HUGE_VAL )
		    {
		      tr_s[q].p= aux + tL_eprobst_lprob ( eprobs, amodel,
							  fea, c, q, 0 );
		      tr_s[q].hyp= trI.hyp; take_hyp ( hypmng, trI.hyp );
		    }
		  else tr_s[q]= TRN0;
		}
	      h= trI.hyp;
	      if ( s != last_s &&
		   (ip= itrans->final.ip) != NULL &&
		   (trI.p+= *ip) != -HUGE_VAL )
		/* Valid per a statealign i no statealign. Pot
		   significar t=-1 o inici de nou segment. */
		trI.hyp= tL_hypmng_add ( hypmng, h, ULONG_MAX );
	      else trI= TRN0;
	      release_hyp ( hypmng, h );
	    }
	  else for ( q= 0; q < cchar.Q; ++q ) tr_s[q]= TRN0;
	  if ( s != last_s )
	    {
	      trI2= viterbi_gen_max ( trI2, tr_s, itrans->final,
				      hypmng, statealign, 0, cchar.Q );
	      if ( (h= trI2.hyp) != NULLHYP )
		{
		  trI2.hyp= tL_hypmng_add ( hypmng, h,
					    statealign ? ULONG_MAX : 0 );
		  release_hyp ( hypmng, h );
		}
	      tr_s[cchar.Q]= trI2;
	    }
	  break;
	default:
	  tL_error ( _( "%s - viterbi_gen_t0: This should not happen" ), __FILE__ );
	}
    }
  
} /* end viterbi_gen_t0 */


static void
viterbi_gen_t (
	       const tLAModel  *amodel,
	       eprobst_t       *eprobs,
	       hypmng_t        *hypmng,
	       struct tr_node **tr0,
	       struct tr_node **tr1,
	       const tLFea     *fea,
	       const int        t,
	       tLSeq           cs,
	       int              statealign,
	       const int       *dist,
	       int              remain
	       )
{
  
  const tLAModelHMM *chars;
  size_t s, last_s, h, h2, i;
  unsigned int c, q;
  tLAModelHMM cchar;
  struct tr_node *tr0_s, *tr1_s, trI, trI2, anode;
  const tLAModelTrans *trans;
  const tLAModelITrans *itrans;
  tLProb a, b;
  const tLProbPair *p;
  const itrans_state_t *states;
  
  
  /* NOTA: Ninguna hipòtesis activa pot ser NULLHYP. */
  i= 0;
  chars= amodel->hmms;
  last_s= cs.length-1;
  trI= trI2= TRN0;
  for ( s= 0; s <= last_s; ++s )
    {
      c= cs.v[s];
      cchar= chars[c];
      tr0_s= tr0[s];
      tr1_s= tr1[s];
      trans= &(cchar.trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
        {
	case TLLINEAL: /* LINEAL. */
	  p= trans->lineal.p;
	  if ( dist[i++] < remain )
	    {
	      b= tr0_s[0].p + p[0].p0;
	      if ( trI.p > b ) { anode.p= trI.p; anode.hyp= trI.hyp; }
	      else             { anode.p= b; anode.hyp= tr0_s[0].hyp; }
	      if ( anode.p != -HUGE_VAL )
		{
		  anode.p+= tL_eprobst_lprob ( eprobs, amodel,
					       fea, c, 0, t );
		  take_hyp ( hypmng, anode.hyp );
		  tr1_s[0]= anode;
		}
	      else tr1_s[0]= TRN0;
	    }
	  else tr1_s[0]= TRN0;
	  if ( trI.hyp != NULLHYP ) release_hyp ( hypmng, trI.hyp );
	  for ( q= 1; q < cchar.Q; ++q )
	    {
	      if ( dist[i++] < remain )
		{
		  a= tr0_s[q-1].p + p[q-1].p1;
		  b= tr0_s[q].p + p[q].p0;
		  if ( a > b )
		    {
		      tr1_s[q].p= a + tL_eprobst_lprob ( eprobs, amodel,
							 fea, c, q, t );
		      if ( statealign )
			{
			  h= tL_hypmng_add ( hypmng, tr0_s[q-1].hyp, t-1 );
			  tr1_s[q].hyp= tL_hypmng_add ( hypmng, h, q-1 );
			  release_hyp ( hypmng, h );
			}
		      else
			{
			  tr1_s[q].hyp= h= tr0_s[q-1].hyp;
			  take_hyp ( hypmng, h );
			}
		    }
		  else if ( b != -HUGE_VAL )
		    {
		      tr1_s[q].p= b + tL_eprobst_lprob ( eprobs, amodel,
							 fea, c, q, t );
		      tr1_s[q].hyp= h= tr0_s[q].hyp;
		      take_hyp ( hypmng, h );
		    }
		  else tr1_s[q]= TRN0;
		}
	      else tr1_s[q]= TRN0;
	      if ( (h= tr0_s[q-1].hyp) != NULLHYP ) release_hyp ( hypmng, h );
	    }
	  if ( (h= tr0_s[cchar.Q-1].hyp) != NULLHYP ) release_hyp ( hypmng, h );
	  if ( s != last_s )
	    {
	      anode.p= tr1_s[cchar.Q-1].p + p[cchar.Q-1].p1;
	      if ( anode.p != -HUGE_VAL )
		{
		  if ( statealign )
		    {
		      h= tL_hypmng_add ( hypmng, tr1_s[cchar.Q-1].hyp, t );
		      h2= tL_hypmng_add ( hypmng, h, cchar.Q-1 );
		      release_hyp ( hypmng, h );
		      anode.hyp= tL_hypmng_add ( hypmng, h2, ULONG_MAX );
		      release_hyp ( hypmng, h2 );
		    }
		  else
		    anode.hyp=
		      tL_hypmng_add ( hypmng, tr1_s[cchar.Q-1].hyp, t );
		  tr1_s[cchar.Q]= anode;
		}
	      else tr1_s[cchar.Q]= TRN0;
	    }
	  break;
	case TLLIST: /* LLISTA. */
	  itrans= trans->list.itrans;
	  states= itrans->states;
	  for ( q= 0; q < cchar.Q; ++q )
	    if ( dist[i++] < remain )
	      {
		anode= viterbi_gen_max ( trI, tr0_s, states[q],
					 hypmng, statealign, t-1, q );
		if ( anode.hyp != NULLHYP )
		  anode.p+= tL_eprobst_lprob ( eprobs, amodel,
					       fea, c, q, t );
		tr1_s[q]= anode;
	      }
	    else tr1_s[q]= TRN0;
	  if ( trI.hyp != NULLHYP ) release_hyp ( hypmng, trI.hyp );
	  for ( q= 0; q < cchar.Q; ++q )
	    if ( (h= tr0_s[q].hyp) != NULLHYP ) release_hyp ( hypmng, h );
	  if ( s != last_s )
	    {
	      anode= viterbi_gen_max ( trI2, tr1_s, itrans->final,
				       hypmng, statealign, t, cchar.Q );
	      if ( (h= anode.hyp) != NULLHYP )
		{
		  anode.hyp= tL_hypmng_add ( hypmng, h,
					     statealign ? ULONG_MAX : t );
		  release_hyp ( hypmng, h );
		}
	      tr1_s[cchar.Q]= anode;
	    }
	  break;
	default:
	  tL_error ( _( "%s - viterbi_gen_t: This should not happen" ), __FILE__ );
	}
      trI= tr0_s[cchar.Q]; trI2= tr1_s[cchar.Q];
    }
  
} /* end viterbi_gen_t */


/* Neteja i calcula l'últim trI. */
static size_t
viterbi_gen_tF (
		const tLAModel  *amodel,
		hypmng_t        *hypmng,
		struct tr_node **tr1,
		const int        last_t,
		tLSeq           cs,
		int              statealign,
		tLFloat         *prob
		)
{
  
  const tLAModelHMM *chars;
  size_t s, last_s, h;
  unsigned int c, q;
  tLAModelHMM cchar;
  struct tr_node *tr1_s, trI2, anode;
  const tLAModelTrans *trans;
  const tLAModelITrans *itrans;
  
  
  /* NOTA: Ninguna hipòtesis activa pot ser NULLHYP. */
  chars= amodel->hmms;
  last_s= cs.length-1;
  trI2= TRN0;
  for ( s= 0; s <= last_s; ++s )
    {
      c= cs.v[s];
      cchar= chars[c];
      tr1_s= tr1[s];
      trans= &(cchar.trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
        {
	case TLLINEAL: /* LINEAL. */
	  if ( s == last_s )
	    {
	      anode.p= tr1_s[cchar.Q-1].p + trans->lineal.p[cchar.Q-1].p1;
	      if ( anode.p != -HUGE_VAL )
		{
		  if ( statealign )
		    {
		      h= tL_hypmng_add ( hypmng, tr1_s[cchar.Q-1].hyp, last_t );
		      anode.hyp= tL_hypmng_add ( hypmng, h, cchar.Q-1 );
		      release_hyp ( hypmng, h );
		    }
		  else
		    anode.hyp=
		      tL_hypmng_add ( hypmng, tr1_s[cchar.Q-1].hyp, last_t );
		  
		}
	      else anode.hyp= NULLHYP;
	      tr1_s[cchar.Q]= anode;
	    }
	  break;
	case TLLIST: /* LLISTA. */
	  itrans= trans->list.itrans;
	  if ( s == last_s )
	    {
	      anode= viterbi_gen_max ( trI2, tr1_s, itrans->final,
				       hypmng, statealign, last_t, cchar.Q );
	      if ( (h= anode.hyp) != NULLHYP && !statealign )
		{
		  anode.hyp= tL_hypmng_add ( hypmng, h, last_t );
		  release_hyp ( hypmng, h );
		}
	      tr1_s[cchar.Q]= anode;
	    }
	  break;
	default:
	  tL_error ( _( "%s - viterbi_gen_tF: This should not happen" ), __FILE__ );
	}
      if ( trI2.hyp != NULLHYP ) release_hyp ( hypmng, trI2.hyp );
      for ( q= 0; q < cchar.Q; ++q )
	if ( (h= tr1_s[q].hyp) != NULLHYP ) release_hyp ( hypmng, h );
      trI2= tr1_s[cchar.Q];
    }
  *prob= trI2.p;
  
  return trI2.hyp;
  
} /* end viterbi_gen_tF */


/* NOTA:
   !statealign: Totes les hipòtesis comencen amb un 0 seguit
   dels instants on es va aplegar a l'estat final de cada caràcter. ULONG_MAX
   representa l'instant -1.
   statealign: Cada segment comença per un ULONG_MAX seguit de parelles (t,q),
   q és l'estat dins del segment, t l'instant on es va eixir d'eixe
   estat (l'últim instant generat en eixe estat). */
static size_t
viterbi_gen (
	     const tLAModel *amodel,
	     tr_gen_t       *tr,
	     eprobst_t      *eprobs,
	     hypmng_t       *hypmng,
	     const tLFea    *fea,
	     tLSeq          cs,
	     int             statealign,
	     const int      *dist,
	     tLFloat        *prob
	     )
{
  
  struct tr_node **tr0, **tr1, **tmp;
  int t, remain;
  
  
  /* Prepara. */
  init_eprobst ( eprobs, amodel, fea );
  tr0= tr->tr0;
  tr1= tr->tr1;
  
  /* Cos. */
  tL_reset_eprobst ( eprobs );
  viterbi_gen_t0 ( amodel, eprobs, hypmng, tr0, fea, cs, statealign );
  for ( t= 1, remain= fea->cols-1; t < fea->cols; ++t, --remain )
    {
      tL_reset_eprobst ( eprobs );
      viterbi_gen_t ( amodel, eprobs, hypmng, tr0, tr1,
		      fea, t, cs, statealign, dist, remain );
      tmp= tr0; tr0= tr1; tr1= tmp;
    }
  
  return viterbi_gen_tF ( amodel, hypmng, tr0, fea->cols-1,
			  cs, statealign, prob );
  
} /* end viterbi_gen */


static size_t
viterbi (
         const tLAModel *amodel,
         struct tr      *tr,
         eprobs_t       *eprobs,
	 hypmng_t       *hypmng,
         const tLFea    *fea,
	 tLSeq          cs,
         size_t          nstates,
	 int             statealign,
	 tLFloat        *prob
         )
{

  int diff, begin, end, t;
  struct tr_node *tr0, *tr1, *aux;
  tLProb a, b, prob_f;
  tLProbPair *pc;
  tLAModelHMM *cchar;
  size_t c, s, hyp;
  unsigned int q, Q;


  /* Prepara. */
  if ( fea->cols < nstates ) return NULLHYP;
  diff= fea->cols - nstates + 1;
  tr0= tr->tr0; tr1= tr->tr1;
  begin= 0; end= diff;

  /* s=0 */
  c= cs.v[0];
  cchar= &(amodel->hmms[c]);
  pc= cchar->trans.lineal.p;
  Q= cchar->Q;
  
  /* --> q=0 */
  tr0[0].p= tL_eprobs_lprob ( eprobs, amodel,
			      fea, c, 0, 0 );
  tr0[0].hyp= NULLHYP;
  take_hyp ( hypmng, NULLHYP );
  for ( t= 1; t < diff; ++t )
    {
      tr0[t].p=
	tr0[t-1].p + pc[0].p0 +
	tL_eprobs_lprob ( eprobs, amodel, fea, c, 0, t );
      tr0[t].hyp= NULLHYP;
      take_hyp ( hypmng, NULLHYP );
    }
  
  /* --> q>0 */
  CALC_VITERBI_BODY;
  
  /* s>0 */
  for ( s= 1; s < cs.length; ++s )
    {

      c= cs.v[s];
      cchar= &(amodel->hmms[c]);
      pc= cchar->trans.lineal.p;
      Q= cchar->Q;
      
      /* --> q=0 */
      ++begin; ++end;
      tr1[begin].p=
        tr0[begin-1].p + prob_f +
        tL_eprobs_lprob ( eprobs, amodel, fea, c, 0, begin );
      hyp= tr0[begin-1].hyp;
      tr1[begin].hyp= tL_hypmng_add ( hypmng, hyp, begin );
      release_hyp ( hypmng, hyp );
      for ( t= begin+1; t < end; ++t )
        {
          a= tr0[t-1].p + prob_f;
          b= tr1[t-1].p + pc[0].p0;
	  if ( a > b )
	    {
	      tr1[t].p= a;
	      tr1[t].hyp= tL_hypmng_add ( hypmng,
				       tr0[t-1].hyp, t );
	    }
	  else
	    {
	      tr1[t].p= b;
	      hyp= tr1[t].hyp= tr1[t-1].hyp;
	      take_hyp ( hypmng, hyp );
	    }
	  release_hyp ( hypmng, tr0[t-1].hyp );
	  tr1[t].p+=
	    tL_eprobs_lprob ( eprobs, amodel, fea, c, 0, t );
        }
      
      aux= tr0;
      tr0= tr1;
      tr1= aux;
      
      /* --> q>0 */
      CALC_VITERBI_BODY;
      
    }
  
  for ( t= begin; t < (fea->cols-1); ++t )
    release_hyp ( hypmng, tr0[t].hyp );
  hyp= tr0[t].hyp;
  if ( (*prob= (tr0[t].p + prob_f)) != -HUGE_VAL ) return hyp;
  else
    {
      release_hyp ( hypmng, hyp );
      return NULLHYP;
    }
  
} /* end viterbi */


/* N no inclou l'elemnt buit. Sempre fica el sentinela. */
static void
alloc_segments (
		tLSegment    **segments,
		size_t        *size,
		const size_t   N
		)
{
  
  size_t aux;
  
  
  aux= N+1;
  if ( *segments == NULL )
    {
      *size= aux;
      mt( *segments= mem_alloc ( tLSegment, aux ) );
    }
  else if ( aux > *size )
    {
      *size= aux;
      mt ( *segments= mem_realloc ( tLSegment, *segments, aux ) );
    }
  (*segments)[N].length= -1;
  
} /* end alloc_segments */


static int
align_lin (
	   tLAligner     *aligner,
	   const tLFea   *fea,
	   const size_t   seq,
	   tLSegment    **alignment,
	   size_t        *size,
	   tLFloat       *prob
	   )
{
  
  size_t hyp, length, i, s;
  unsigned int q, maxQ;
  unsigned long prev;
  tLSeq cs;
  tLBool uniseg;
  
  
  /* Prepara. */
  if ( fea->cols > aligner->tr.lin.tr->length )
    resize_tr ( aligner->tr.lin.tr, fea->cols );
  if ( fea->cols > aligner->tr.lin.eprobs->length )
    tL_resize_eprobs ( aligner->tr.lin.eprobs, aligner->amodel, fea->cols );
  cs= *(tL_seqs_get_seq ( aligner->seqs, seq ));
  
  /* Alinea. */
  uniseg= (cs.length == 1 &&
	   (!aligner->statealign || aligner->amodel->hmms[cs.v[0]].Q == 1));
  init_eprobs ( aligner->tr.lin.eprobs, aligner->amodel, fea );
  hyp= viterbi ( aligner->amodel, aligner->tr.lin.tr,
		 aligner->tr.lin.eprobs, aligner->hypmng,
		 fea, cs, aligner->tr.lin.nstates[seq],
		 aligner->statealign, prob );
  if ( (uniseg && *prob == -HUGE_VAL) ||
       (!uniseg && hyp == NULLHYP) ) return -1;
  else if ( uniseg ) /* Ací hyp és NULLHYP. */
    {
      release_hyp ( aligner->hypmng, hyp );
      alloc_segments ( alignment, size, 1 );
      (*alignment)[0].c= cs.v[0];
      (*alignment)[0].q= 0;
      (*alignment)[0].pos= 0;
      (*alignment)[0].length= fea->cols;
    }
  else
    {
      length= tL_get_hyp ( aligner->hypmng, hyp, aligner->buffer );
      release_hyp ( aligner->hypmng, hyp );
      alloc_segments ( alignment, size, length+1 );
      prev= 0;
      if ( !aligner->statealign )
	{
	  for ( i= 0; i < length; ++i )
	    {
	      (*alignment)[i].c= cs.v[i];
	      (*alignment)[i].q= 0;
	      (*alignment)[i].pos= (int) prev;
	      (*alignment)[i].length= (int) (aligner->buffer[i]-prev);
	      prev= aligner->buffer[i];
	    }
	  (*alignment)[i].c= cs.v[i];
	  (*alignment)[i].q= 0;
	  (*alignment)[i].pos= (int) prev;
	  (*alignment)[i].length= (int) ((unsigned long) fea->cols - prev);
	}
      else
	{
	  s= 0;
	  maxQ= aligner->amodel->hmms[cs.v[s]].Q;
	  q= 0;
	  for ( i= 0; i < length; ++i )
	    {
	      (*alignment)[i].c= cs.v[s];
	      (*alignment)[i].q= q;
	      (*alignment)[i].pos= (int) prev;
	      (*alignment)[i].length= (int) (aligner->buffer[i]-prev);
	      prev= aligner->buffer[i];
	      if ( ++q == maxQ )
		{
		  maxQ= aligner->amodel->hmms[cs.v[++s]].Q;
		  q= 0;
		}
	    }
	  (*alignment)[i].c= cs.v[s];
	  (*alignment)[i].q= q;
	  (*alignment)[i].pos= (int) prev;
	  (*alignment)[i].length= (int) ((unsigned long) fea->cols - prev);
	}
    }
  assert ( aligner->hypmng->v[0].counter == 0 );
  
  tL_reset_eprobs ( aligner->tr.lin.eprobs, aligner->amodel, fea->cols );
  
  return 0;
  
} /* align_lin */


static int
align_gen (
	   tLAligner     *aligner,
	   const tLFea   *fea,
	   const size_t   seq,
	   tLSegment    **alignment,
	   size_t        *size,
	   tLFloat       *prob
	   )
{
  
  tLSeq cs;
  size_t nbsize, hyp, length, nsegs, i, j, s;
  const unsigned long *buffer;
  unsigned long prev, next;
  const int *dist;
  
  
  /* Prepara. */
  cs= *(tL_seqs_get_seq ( aligner->seqs, seq ));
  if ( aligner->statealign )
    {
      nbsize= cs.length + fea->cols*2;
      if ( nbsize > aligner->bsize )
	{
	  mt ( aligner->buffer=
	       mem_realloc ( unsigned long, aligner->buffer, nbsize ) );
	  aligner->bsize= nbsize;
	}
    }
  dist= aligner->tr.gen.dists[seq];
  
  /* Alinea. */
  /* LA PROBABILITAT OBLIGA A COMENTAR AÇÒ.
  if ( cs.length == 1 &&
       (!aligner->statealign || aligner->amodel->hmms[cs.v[0]].Q == 1) )
    {
      alloc_segments ( alignment, size, 1 );
      (*alignment)[0].c= cs.v[0];
      (*alignment)[0].q= 0;
      (*alignment)[0].pos= 0;
      (*alignment)[0].length= fea->cols;
      return 0;
    }
  */
  hyp= viterbi_gen ( aligner->amodel, aligner->tr.gen.tr,
		     aligner->tr.gen.eprobs, aligner->hypmng,
		     fea, cs, aligner->statealign, dist, prob );
  if ( hyp == NULLHYP ) return -1;
  else
    {
      length= tL_get_hyp ( aligner->hypmng, hyp, aligner->buffer );
      buffer= aligner->buffer;
      release_hyp ( aligner->hypmng, hyp );
      if ( !aligner->statealign )
	{
	  /* FEATURE QUE JA NO M'AGRADA */
	  /*for ( i= 1; i < length && buffer[i] == ULONG_MAX; ++i );
	    nsegs= length-i;*/
	  nsegs= length-1; i= 1;
	  alloc_segments ( alignment, size, nsegs );
	  prev= 0;
	  for ( j= 0; i < length; ++i, ++j )
	    {
	      (*alignment)[j].c= cs.v[i-1];
              (*alignment)[j].q= 0;
              (*alignment)[j].pos= (int) prev;
	      next= buffer[i];
              (*alignment)[j].length= (int) (next-prev) + 1;
              prev= next+1;
	    }
	}
      else
	{
	  nsegs= 0;
	  for ( i= 0; i < length; )
	    if ( buffer[i] == ULONG_MAX ) ++i;
	    else { ++nsegs; i+= 2; }
	  alloc_segments ( alignment, size, nsegs);
	  prev= 0;
	  for ( i= 1, s= j= 0; i < length; )
	    if ( buffer[i] == ULONG_MAX ) { ++s; ++i; }
	    else
	      {
		next= buffer[i++];
		(*alignment)[j].c= cs.v[s];
		(*alignment)[j].q= (unsigned int) buffer[i++];
		(*alignment)[j].pos= (int) prev;
		(*alignment)[j++].length= (int) (next-prev) + 1;
		prev= next+1;
	      }
	}
    }
  assert ( aligner->hypmng->v[0].counter == 0 );
  
  return 0;
  
} /* end align_gen */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_aligner_free (
		 tLAligner *aligner
		 )
{
  
  free ( aligner->buffer );
  tL_free_hypmng ( aligner->hypmng );
  if ( aligner->type == LINEAR )
    {
      free ( aligner->tr.lin.nstates );
      free_tr ( aligner->tr.lin.tr );
      tL_free_eprobs ( aligner->tr.lin.eprobs );
    }
  else /* GENERIC. */
    {
      free_dists ( aligner->tr.gen.dists, aligner->seqs );
      free_tr_gen ( aligner->tr.gen.tr );
      tL_free_eprobst ( aligner->tr.gen.eprobs );
    }
  free ( aligner );
  
} /* end tL_aligner_free */


TL_API_FUNC(tLAligner *)
tL_aligner_new (
		tLAModel      *amodel,
		const tLSeqs  *seqs,
		const tLBool   statealign,
		char         **err
		)
{
  
  tLAligner *new;
  size_t max_length, n, length, s;
  const tLSeq *cs;
  int type;
  
  
  /* Crea. */
  mt ( new= mem_alloc ( tLAligner, 1 ) );
  new->amodel= amodel;
  new->seqs= seqs;
  new->type= type= (amodel->tflags == FLINEAL) ? LINEAR : GENERIC;
  /* max_length pot reprentar el número de caràcters màxim o el número
     d'estats màxim. */
  max_length= 0;
  for ( n= 0; n < tL_seqs_get_num_seqs ( seqs ); ++n )
    {
      cs= tL_seqs_get_seq ( seqs, n );
      if ( statealign && type == LINEAR )
        {
          length= 0;
          for ( s= 0; s < cs->length; ++s )
            length+= amodel->hmms[cs->v[s]].Q;
        }
      else length= cs->length;
      if ( length > max_length ) max_length= length;
    }
  if ( type == LINEAR )
    {
      new->tr.lin.eprobs= tL_create_eprobs ( amodel, err );
      if ( new->tr.lin.eprobs == NULL ) goto error;
      new->tr.lin.tr= create_tr ();
      new->tr.lin.nstates= tL_create_nstates ( seqs->cseqs, amodel );
      new->bsize= max_length;
    }
  else /* GENERIC */
    {
      new->tr.gen.eprobs= tL_create_eprobst ( amodel, FALSE, err );
      if ( new->tr.gen.eprobs == NULL ) goto error;
      tL_amodel_alloc_all_itrans ( amodel );
      new->tr.gen.tr= create_tr_gen ( amodel, max_length );
      new->tr.gen.dists= create_dists ( amodel, seqs );
      new->bsize= statealign ? 1 : (1+max_length);
    }
  mt ( new->buffer= mem_alloc ( unsigned long, new->bsize ) );
  new->hypmng= tL_create_hypmng ();
  new->statealign= statealign;
  
  return new;
  
 error:
  free ( new );
  return NULL;
  
} /* end tL_aligner_new */


TL_API_FUNC(int)
tL_align (
	  tLAligner     *aligner,
	  const tLFea   *fea,
	  const size_t   seq,
	  tLSegment    **alignment,
	  size_t        *size,
	  tLFloat       *prob
	  )
{
  
  if ( aligner->type == LINEAR )
    return align_lin ( aligner, fea, seq, alignment, size, prob );
  else /* GENERIC */
    return align_gen ( aligner, fea, seq, alignment, size, prob );
  
} /* tL_align */
