/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  align.h - Realitza aliniament forçat.
 *
 *  Adrià Giménez Pastor, 2008-2011.
 *
 */

#ifndef __LIBTL_ALIGN_H__
#define __LIBTL_ALIGN_H__

#include <stddef.h>

#include "amodel_public.h"
#include "common_public.h"
#include "port.h"
#include "samples_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup Align Alignment of samples with symbol sequences
 */
/** \addtogroup Align
 *  @{
 */

#ifndef __TL_PRV
/** Aligner. This structure contains auxiliary data used by 'tL_align'
 *  to align a sample with a symbol sequence. The values of model
 *  parameters can be modified but not the topology.
 */
struct _tLAligner
{
  
  const tLAModel *amodel;    /**< Reference to the acoustic model used
				by the aligner. */
  const tLSeqs   *seqs;      /**< Reference to the symbol sequence
				manager used by the aligner. */
  
};
#endif
/** Aligner. This structure contains auxiliary data used by 'tL_align'
 *  to align a sample with a symbol sequence. The values of model
 *  parameters can be modified but not the topology.
 */
typedef struct _tLAligner tLAligner;

/** Frees memory. Frees the memory allocated for the aligner.
 *
 *  @param aligner The aligner
 */
TL_API_FUNC(void)
tL_aligner_free (
		 tLAligner *aligner
		 );

/** Gets a tLAligner. This function creates an aligner used to align
 *  samples with symbol sequences. This data can only be used to
 *  align samples, using the provided acoustic model, with the
 *  symbol sequences stored in seqs during the creation of the
 *  aligner. Alignment can be done at symbol level or state
 *  level. This structure stores a reference to the provided acoustic
 *  model and sequence manager, thus, the acoustic model and
 *  sequence manager must not be freed. Currently supported types
 *  are: Bernoulli, DGaussian, Mixture and TiedStates.
 *
 *  @param amodel The acoustic model.
 *  @param seqs A symbol sequence manager containing the symbol
 *  sequences which can be used to align samples.
 *  @param statealign If TRUE, alignment is performed at state level;
 *  else, at symbol level.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @return Aligner or NULL in case of error.
 */
TL_API_FUNC(tLAligner *)
tL_aligner_new (
		tLAModel      *amodel,
		const tLSeqs  *seqs,
		const tLBool   statealign,
		char         **err
		);

/** Alignment. Structure used to store the alignments produced by
 * 'tL_align'.
 */
typedef struct
{
  
  size_t       c;         /**< Symbol code. */
  unsigned int q;         /**< Number of states inside the
			     symbol. This field is only used with
			     alignments at state level. */
  int          pos;       /**< Position of the first frame. */
  int          length;    /**< Number of frames. */
  
} tLSegment;

/** Aligns a sample with a symbol sequence. This function finds the
 *  maximum probability path of a given sample, feature vector
 *  sequence, for a provided symbol sequence, and gets the
 *  alignment. The sequence is specified as integer code according to
 *  the sequence manager used during the creation of the aligner. The
 *  resulting alignment is returned into the variable alignment. The
 *  alignment is terminated with a tLSegment with length -1. If
 *  alignment points to NULL, the needed memory is allocated and the
 *  number of allocated segments is stored in the variable size. If
 *  alignment doesn't point to NULL, then alignment will be reallocated
 *  if needed, according to the value stored in size, and the variable
 *  size will be updated.
 *
 *  @param aligner The aligner.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param seq The code of the symbol sequence which will be
 *  aligned.
 *  @param alignment Pointer to the address where the alignment will be
 *  stored. It can be pointed to NULL.
 *  @param size Memory address where the number of allocated segments
 *  of alignment is stored.
 *  @param prob The probability of the sample.
 *  @return 0 if the sample has been successfully aligned, -1 otherwise.
 */
TL_API_FUNC(int)
tL_align (
	  tLAligner     *aligner,
	  const tLFea   *fea,
	  const size_t   seq,
	  tLSegment    **alignment,
	  size_t        *size,
	  tLFloat       *prob
	  );

/** @} */
/** @} */

#endif /* __LIBTL_ALIGN_H__ */
