/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  dgaussian.c - Implementació de 'dgaussian.h'.
 *
 *  Adrià Giménez Pastor, 2009,2010,2011,2012,2013.
 *
 */


#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <zlib.h>

#include "iamcounts.h"
#include "logamcounts.h"
#include "amcounts_aux.h"
#include "amodel/dgaussian.h"
#include "dgaussian.h"
#include "error.h"
#include "global.h"
#include "mem.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) DGAUSSIAN_GLOBAL(ptr)
#define CGLOBAL(ptr) ((amcount_dgaussian_global_t *) (ptr))
#define CSTATE(ptr) ((amcount_dgaussian_state_t *) (ptr))
#define STATE(ptr) DGAUSSIAN_STATE(ptr)




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
tL_amcount_dgaussian_alloc_xx (
			       amcount_dgaussian_state_t *counts,
			       const unsigned int         Q,
			       const unsigned int         D,
			       const int                  aboutxx
			       )
{
  
  tLFloat *mem, **xx;
  unsigned int q, d, D2;
  
  
  if ( aboutxx == NONE ) { counts[0].xx= NULL; return; }
  D2= (aboutxx==CMLLR)?(D+1):D;
  mt ( xx= mem_alloc ( tLFloat *, Q*(D+1) ) );
  mt ( mem= mem_alloc ( tLFloat, Q*(D+1)*D2 ) );

  for ( q= 0; q < Q; ++q, xx+= (D+1)  )
    {
      counts[q].xx= xx;
      for ( d= 0; d <= D; ++d, mem+= D2 )
	xx[d]= mem;
    }
  
} /* end tL_amcount_dgaussian_alloc_xx */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. */
static int
read_and_accum (
		tLFloat       *var,
		gzFile         from,
		tLBuffer      *buffer,
		const tLBool   binary,
		unsigned long *counter
		)
{

  tLFloat aux;
  
  
  if ( binary )
    {
      if ( gzread ( from, &aux, 1*sizeof(tLFloat) ) != 1*sizeof(tLFloat) )
	return -1;
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ) return -1;
      aux= atof ( buffer->v );
    }
  (*var)+= aux;
  ++(*counter);
  
  return 0;

} /* end read_and_accum */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. */
static int
read_and_accum_eps (
                    tLFloat       *var,
                    gzFile         from,
                    tLBuffer      *buffer,
                    const tLFloat  eps,
                    const tLBool   binary,
                    unsigned long *counter
                    )
{

  tLFloat aux;
  
  
  if ( binary )
    {
      if ( gzread ( from, &aux, 1*sizeof(tLFloat) ) != 1*sizeof(tLFloat) )
        return -1;
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ) return -1;
      aux= atof ( buffer->v );
    }
  (*var)= (*var)*(1.0-eps) + eps*aux;
  ++(*counter);
  
  return 0;

} /* end read_and_accum_eps */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. */
static int
read_count (
	    tLFloat       *var,
	    gzFile         from,
	    tLBuffer      *buffer,
	    const tLBool   binary,
	    unsigned long *counter
	    )
{

  if ( binary )
    {
      if ( gzread ( from, var, 1*sizeof(tLFloat) ) != 1*sizeof(tLFloat) )
	return -1;
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ) return -1;
      *var= atof ( buffer->v );
    }
  ++(*counter);
  
  return 0;

} /* end read_count */




/***********/
/* MÈTODES */
/***********/

static void *
create_global (
	       void      *global,
	       const int  mllrinfo
	       )
{
  
  amcount_dgaussian_global_t *new;
  
  
  mt ( new= mem_alloc ( amcount_dgaussian_global_t, 1 ) );
  new->aboutxx= mllrinfo;
  
  return new;
  
} /* end create_global */


static void *
alloc_states (
	      void               *global,
	      void               *cglobal,
	      void               *states,
	      const unsigned int  Q
	      )
{
  
  amcount_dgaussian_state_t *new;
  tLFloat *mem;
  unsigned int q, D;
  
  
  D= GLOBAL(global)->D;
  mt ( new= mem_alloc ( amcount_dgaussian_state_t, Q ) );
  mt ( mem= mem_alloc ( tLFloat, Q*(D<<1) ) );
  for ( q= 0; q < Q; ++q )
    {
      new[q].x= mem; mem+= D;
      new[q].x2= mem; mem+= D;
    }
  tL_amcount_dgaussian_alloc_xx ( new, Q, D, CGLOBAL(cglobal)->aboutxx );
  
  return (void *) new;
  
} /* end alloc_states */


static void
free_global (
             void *cglobal,
             void *global
             )
{
  free ( cglobal );
} /* end free_global */


static void
free_states (
	     void               *global,
	     void               *cglobal,
	     void               *states,
	     void               *cstates,
	     const unsigned int  Q
	     )
{
  
  if ( CGLOBAL(cglobal)->aboutxx != NONE )
    {
      free ( CSTATE(cstates)[0].xx[0]);
      free ( CSTATE(cstates)[0].xx);
    }
  free ( CSTATE(cstates)[0].x );
  free ( cstates );

} /* end free_states */


static void
init_states (
	     void               *global,
	     void               *cglobal,
	     void               *states,
	     void               *cstates,
	     const unsigned int  Q
	     )
{
  
  unsigned int q, D;
  unsigned long i, size;
  tLFloat *mem;
  int aboutxx;
  
  
  D= GLOBAL(global)->D;
  size= Q*(D<<1);
  mem= CSTATE(cstates)[0].x;
  for ( i= 0; i < size; ++i )
    mem[i]= 0.0;
  for ( q= 0; q < Q; ++q )
    CSTATE(cstates)[q].norm= 0.0;
  if ( (aboutxx= CGLOBAL(cglobal)->aboutxx) != NONE )
    {
      mem= CSTATE(cstates)[0].xx[0];
      size= Q*(D+1)*((aboutxx==CMLLR)?(D+1):D);
      for ( i= 0; i < size; ++i )
	mem[i]= 0.0;
    }
  
} /* end init_states */


static void
print_global (
              FILE         *to,
              void         *global,
              void         *cglobal,
              const tLBool  binary
              )
{
  
  if ( binary ) fwritee ( &(CGLOBAL(cglobal)->aboutxx), sizeof(int), 1, to );
  else          fprintf ( to, "%d\n", CGLOBAL(cglobal)->aboutxx);
  
} /* end print_global */


static void
init_states_wrpi (
		  void                *global,
		  void                *cglobal,
		  void                *states,
		  void                *cstates,
		  const unsigned int   Q,
		  const unsigned long  val
		  )
{
  init_states ( global, cglobal, states, cstates, Q );
} /* end init_states_wrpi */


static void
init_states_wrplog (
		    void               *global,
		    void               *cglobal,
		    void               *states,
		    void               *cstates,
		    const unsigned int  Q,
		    const tLFloat       val
		    )
{
  init_states ( global, cglobal, states, cstates, Q );
} /* end init_states_wrplog */


static void
print (
       FILE               *to,
       void               *global,
       void               *cglobal,
       void               *states,
       void               *cstates,
       const unsigned int  Q,
       const tLBool        binary
       )
{
  
  unsigned int d, d2, q, D, D2;
  amcount_dgaussian_state_t state;
  const tLFloat *xx_d;
  int aboutxx;
  
  
  D= GLOBAL(global)->D;
  aboutxx= CGLOBAL(cglobal)->aboutxx;
  if ( binary )
    for ( q= 0; q < Q; ++q )
      {
	state= CSTATE(cstates)[q];
	fwritee ( state.x, sizeof(tLFloat), D, to );
	fwritee ( state.x2, sizeof(tLFloat), D, to );
	fwritee ( &(state.norm), sizeof(state.norm), 1, to );
	if ( aboutxx != NONE )
	  fwritee ( state.xx[0], sizeof(tLFloat),
		    (D+1)*((aboutxx==CMLLR)?(D+1):D), to );
      }
  else
    for ( q= 0; q < Q; ++q )
      {
	state= CSTATE(cstates)[q];
	fprintf ( to, "%g", _DB(state.x[0]) );
	for ( d= 1; d < D; ++d ) fprintf ( to, " %g", _DB(state.x[d]) );
	putc ( '\n', to );
	fprintf ( to, "%g", _DB(state.x2[0]) );
	for ( d= 1; d < D; ++d ) fprintf ( to, " %g", _DB(state.x2[d]) );
	putc ( '\n', to );
	fprintf ( to, "%g\n", _DB(state.norm) );
	if ( aboutxx != NONE )
	  {
	    D2= (aboutxx==CMLLR)?(D+1):D;
	    for ( d= 0; d <= D; ++d )
	      {
		xx_d= state.xx[d];
		fprintf ( to, "%g", xx_d[0] );
		for ( d2= 1; d2 < D2; ++d2 )
		  fprintf ( to, "%g", xx_d[d2] );
		putc ( '\n', to );
	      }
	  }
      }
  
} /* end print */


static int
accum_global (
              gzFile         from,
              tLBuffer      *buffer,
              void          *global,
              void          *cglobal,
              const tLBool   binary,
              unsigned long *counter
              )
{
  
  int aux;
  
  
  if ( binary )
    {
      if ( gzread ( from, &aux, sizeof(int) ) != sizeof(int) ) return -1;
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ) return -1;
      aux= (atoi ( buffer->v )!=0);
    }
  assert ( aux == CGLOBAL(cglobal)->aboutxx );
  ++(counter);
  
  return 0;
  
} /* end accum_global */


static int
accum_global_eps (
                  gzFile         from,
                  tLBuffer      *buffer,
                  void          *global,
                  void          *cglobal,
                  const tLFloat  eps,
                  const tLFloat  logeps,
                  const tLFloat  logeps_1,
                  const tLBool   binary,
                  unsigned long *counter
                  )
{
  return accum_global ( from, buffer, global, cglobal, binary, counter );
} /* end accum_global_eps */


static int
accum (
       gzFile              from,
       tLBuffer           *buffer,
       void               *global,
       void               *cglobal,
       void               *states,
       void               *cstates,
       const unsigned int  Q,
       const tLBool        binary,
       unsigned long      *counter
       )
{
  
  unsigned int d, d2, q, D, D2;
  amcount_dgaussian_state_t state;
  tLFloat *xx_d;
  int aboutxx;
  
  
  D= GLOBAL(global)->D;
  for ( q= 0; q < Q; ++q )
    {
      state= CSTATE(cstates)[q];
      for ( d= 0; d < D; ++d )
	if ( read_and_accum ( &(state.x[d]), from, buffer,
			      binary, counter ) == -1 )
	  return -1;
      for ( d= 0; d < D; ++d )
	if ( read_and_accum ( &(state.x2[d]), from, buffer,
			      binary, counter ) == -1 )
	  return -1;
      if ( read_and_accum ( &(CSTATE(cstates)[q].norm), from, buffer,
			    binary, counter ) == -1 )
	return -1;
      if ( (aboutxx= CGLOBAL(cglobal)->aboutxx) != NONE )
	{
	  D2= (aboutxx==CMLLR)?(D+1):D;
	  for ( d= 0; d <= D; ++d )
	    {
	      xx_d= state.xx[d];
	      for ( d2= 0; d2 < D2; ++d2 )
		if ( read_and_accum ( &(xx_d[d2]), from, buffer,
				      binary, counter ) == -1 )
		  return -1;
	    }
	}
    }
  
  return 0;
  
} /* end accum */


static int
accum_eps (
           gzFile              from,
           tLBuffer           *buffer,
           void               *global,
           void               *cglobal,
           void               *states,
           void               *cstates,
           const unsigned int  Q,
           const tLFloat       eps,
           const tLFloat       logeps,
           const tLFloat       logeps_1,
           const tLBool        binary,
           unsigned long      *counter
           )
{
  
  unsigned int d, d2, q, D, D2;
  amcount_dgaussian_state_t state;
  tLFloat *xx_d;
  int aboutxx;
  
  
  D= GLOBAL(global)->D;
  for ( q= 0; q < Q; ++q )
    {
      state= CSTATE(cstates)[q];
      for ( d= 0; d < D; ++d )
        if ( read_and_accum_eps ( &(state.x[d]), from, buffer,
                                  eps, binary, counter ) == -1 )
          return -1;
      for ( d= 0; d < D; ++d )
        if ( read_and_accum_eps ( &(state.x2[d]), from, buffer,
                                  eps, binary, counter ) == -1 )
          return -1;
      if ( read_and_accum_eps ( &(CSTATE(cstates)[q].norm), from, buffer,
                                eps, binary, counter ) == -1 )
        return -1;
      if ( (aboutxx= CGLOBAL(cglobal)->aboutxx) != NONE )
        {
          D2= (aboutxx==CMLLR)?(D+1):D;
          for ( d= 0; d <= D; ++d )
            {
              xx_d= state.xx[d];
              for ( d2= 0; d2 < D2; ++d2 )
                if ( read_and_accum_eps ( &(xx_d[d2]), from, buffer,
                                          eps, binary, counter ) == -1 )
                  return -1;
            }
        }
    }
  
  return 0;
  
} /* end accum_eps */


static int
_read (
      gzFile              from,
      tLBuffer           *buffer,
      void               *global,
      void               *cglobal,
      void               *states,
      void               *cstates,
      const unsigned int  Q,
      const tLBool        binary,
      unsigned long      *counter
      )
{
  
  unsigned int d, d2, q, D, D2;
  amcount_dgaussian_state_t state;
  tLFloat *xx_d;
  int aboutxx;
  
  
  D= GLOBAL(global)->D;
  for ( q= 0; q < Q; ++q )
    {
      state= CSTATE(cstates)[q];
      for ( d= 0; d < D; ++d )
	if ( read_count ( &(state.x[d]), from, buffer,
			  binary, counter ) == -1 )
	  return -1;
      for ( d= 0; d < D; ++d )
	if ( read_count ( &(state.x2[d]), from, buffer,
			  binary, counter ) == -1 )
	  return -1;
      if ( read_count ( &(CSTATE(cstates)[q].norm), from, buffer,
			binary, counter ) == -1 )
	return -1;
      if ( (aboutxx= CGLOBAL(cglobal)->aboutxx) != NONE )
	{
	  D2= (aboutxx==CMLLR)?(D+1):D;
	  for ( d= 0; d <= D; ++d )
	    {
	      xx_d= state.xx[d];
	      for ( d2= 0; d2 < D2; ++d2 )
		if ( read_count ( &(xx_d[d2]), from, buffer,
				  binary, counter ) == -1 )
		  return -1;
	    }
	}
    }
  
  return 0;
  
} /* end read */


static void
update_states (
	       void               *global,
	       void               *cglobal,
	       void               *states,
	       void               *cstates,
	       const unsigned int  Q
	       )
{
  
  unsigned int q, d, D;
  dgaussian_state_t state;
  amcount_dgaussian_state_t scount;
  const tLFloat *smooth;
  
  
  D= GLOBAL(global)->D;
  smooth= GLOBAL(global)->smooth;
  for ( q= 0; q < Q; ++q )
    {
      state= STATE(states)[q];
      scount= CSTATE(cstates)[q];
      /*if ( scount.norm == 0.0 ) continue;*/
      if ( scount.norm != 0.0 )
	for ( d= 0; d < D; ++d )
	  {
	    state.mu[d]= scount.x[d]/scount.norm;
	    state.var[d]=
	      (scount.x2[d] -
	       (scount.x[d]*scount.x[d])/scount.norm)
	      /scount.norm;
	    if ( state.var[d] < smooth[d] ) /* LÍNIA PER A SMOOTH. */
	      state.var[d]= smooth[d];
	    state.ivar[d]= 1.0/state.var[d];
	}
      /* --> Don't modify parameters if when norm==0.0
      else
	for ( d= 0; d < D; ++d )
	  {
	    state.mu[d]= 0.0;
	    state.var[d]= smooth[d];
	    state.ivar[d]= 1.0/smooth[d];
	  }
      */
      tL_DGaussian_update_logc ( D, &(STATE(states)[q]) );
    }
  
} /* end update_states */


iamcount_methods_t itL_amcounts_methodsdgaussian=
  {
    create_global,
    alloc_states,
    free_global,
    free_states,
    NULL,
    init_states_wrpi,
    print_global,
    print,
    accum_global,
    accum,
    accum_global, /* Són iguals, sols comprova que el aboutxx
		     coincideix. */
    _read,
    tL_update_global_empty,
    update_states
  };

logamcount_methods_t logtL_amcounts_methodsdgaussian=
  {
    create_global,
    alloc_states,
    free_global,
    free_states,
    NULL,
    init_states_wrplog,
    print_global,
    print,
    accum_global,
    accum,
    accum_global_eps,
    accum_eps,
    accum_global, /* Són iguals, sols comprova que el aboutxx
		    coincideix. */
    _read,
    tL_update_global_empty,
    update_states
  };
