/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  amcounts_aux.c - Implementació de 'amcounts_aux.h'.
 *
 *  Adrià Giménez Pastor, 2010,2011,2012,2013.
 *
 */


#include <stddef.h>
#include <stdlib.h>
#include <zlib.h>

#include "calc.h"
#include "amcounts_aux.h"
#include "common.h"
#include "error.h"




/**********/
/* MACROS */
/**********/

#define LOG05 -0.693147




/*************/
/* CONSTANTS */
/*************/

const llcount_type llcount_type_zero= { -HUGE_VAL, -HUGE_VAL };

const rprop_aux_node_t ra_zero= { 0.0, 0.0, 0.0 };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_amcounts_inorm_pair (
			const icount_t  count,
			tLProbPair     *var
			)
{
  
  unsigned long total;
  
  
  total= count.p1 + count.p0;
  if ( total == 0 ) return; /*ret.p1= ret.p0= LOG05;*/
  else
    {
      var->p1= (tLProb) log ( (double) count.p1 / (double) total );
      var->p0= p1top0 ( var->p1 );
    }
  
} /* end tL_amcounts_inorm_pair */


void
tL_amcounts_lognorm_pair (
			  const logcount_t  count,
			  tLProbPair       *var
			  )
{
  
  tLFloat total;
  
  
  if ( count.p1 == -HUGE_VAL && count.p0 == -HUGE_VAL ) return;
  /*ret.p1= ret.p0 = LOG05;*/
  else
    {
      total= tL_radd ( count.p1, count.p0 );
      var->p1= (tLProb) (count.p1 - total);
      var->p0= (tLProb) (count.p0 - total);
    }
  
} /* end tL_amcounts_lognorm_pair */


int
tL_amcounts_inorm_probs (
			 tLProb              *probs,
			 const unsigned long *counts,
			 const unsigned int   N
			 )
{
  
    
  unsigned int n;
  unsigned long total;
  tLProb aux;
  
  
  total= 0;
  for ( n= 0; n < N; ++n ) total+= counts[n];
  if ( total == 0 ) return -1;
  /*
    {
      aux= (tLProb) log ( 1.0 / N );
      for ( n= 0; n < N; ++n )
	probs[n]= aux;
    }
  */
  else
    {
      aux= log ( total );
      for ( n= 0; n < N; ++n )
	probs[n]= (tLProb) (log ( counts[n] ) - aux);
    }
  
  return 0;
  
} /* end tL_amcounts_inorm_probs */


int
tL_amcounts_lognorm_probs (
			   tLProb             *probs,
			   const tLFloat      *counts,
			   const unsigned int  N
			   )
{
  
  unsigned int n;
  tLFloat max, total;
  
  
  max= counts[0];
  for ( n= 1; n < N; ++n )
    if ( counts[n] > max )
      max= counts[n];
  total=
    (max != -HUGE_VAL) ?
    tL_radd_vals ( counts, max, N ) :
    -HUGE_VAL;
  
  if ( total == -HUGE_VAL ) return -1;
  /*
    {
      aux= (tLProb) log ( 1.0 / N );
      for ( n= 0; n < N; ++n )
	probs[n]= aux;
    }
  */
  else
    for ( n= 0; n < N; ++n )
      probs[n]= (tLProb) (counts[n] - total);
  
  return 0;
  
} /* end tL_amcounts_lognorm_probs */


void
tL_amcounts_iprint (
		    FILE                *f,
		    const unsigned long  value,
		    const tLBool         binary
		    )
{
  
  if ( binary ) fwritee ( &value, sizeof(unsigned long), 1, f );
  else          fprintf ( f, "%lu", value );
  
} /* end tL_amcounts_iprint */


void
tL_amcounts_logprint (
		      FILE          *f,
		      const tLFloat  value,
		      const tLBool   binary
		      )
{
  
  if ( binary ) fwritee ( &value, sizeof(tLFloat), 1, f );
  else          tL_prob_print ( value, f );
  
} /* end tL_amcounts_logprint */


void
tL_amcounts_llprint (
		     FILE               *f,
		     const llcount_type  value,
		     const tLBool        binary
		     )
{
  
  if ( binary ) fwritee ( &value, sizeof(llcount_type), 1, f );
  else
    {
      tL_prob_print ( value.a, f );
      printprobsep ( f, value.b );
    }
  
} /* end tL_amcounts_llprint */


void
tL_amcounts_raprint (
		     FILE                   *f,
		     const rprop_aux_node_t  value,
		     const tLBool            binary
		     )
{
  
  if ( binary ) fwritee ( &value, sizeof(rprop_aux_node_t), 1, f );
  else          fprintf ( f, "%g %g %g",
			  _DB(value.inc), _DB(value.grad), _DB(value.l0) );
  
} /* end tL_amcounts_raprint */


int
tL_amcounts_iaccum (
		    gzFile         f,
		    tLBuffer      *buffer,
		    unsigned long *var,
		    const tLBool   binary
		    )
{
  
  unsigned long aux;
  
  
  if ( binary )
    {
      if ( gzread ( f, &aux, 1*sizeof(unsigned long) ) != 1*sizeof(unsigned long) )
	return 0;
      *var+= aux;
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      *var+= atoul ( buffer->v );
    }
  
  return 1;
  
} /* end tL_amcounts_iaccum */


int
tL_amcounts_logaccum (
		      gzFile        f,
		      tLBuffer     *buffer,
		      tLFloat      *var,
		      const tLBool  binary
		      )
{
  
  tLFloat aux;
  
  
  if ( binary )
    {
      if ( gzread ( f, &aux, 1*sizeof(tLFloat) ) != 1*sizeof(tLFloat) )
	return 0;
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      aux= tL_atop ( buffer->v );
    }
  *var= RADD ( *var, aux );
  
  return 1;
  
} /* end tL_amcounts_logaccum */


int
tL_amcounts_logaccum_eps (
                          gzFile         f,
                          tLBuffer      *buffer,
                          tLFloat       *var,
                          const tLFloat  logeps,
                          const tLFloat  logeps_1,
                          const tLBool   binary
                          )
{
  
  tLFloat aux;
  
  
  if ( binary )
    {
      if ( gzread ( f, &aux, 1*sizeof(tLFloat) ) != 1*sizeof(tLFloat) )
        return 0;
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      aux= tL_atop ( buffer->v );
    }
  (*var)+= logeps_1;
  aux+= logeps;
  *var= RADD ( *var, aux );
  
  return 1;
  
} /* end tL_amcounts_logaccum_eps */


int
tL_amcounts_llaccum (
		     gzFile        f,
		     tLBuffer     *buffer,
		     llcount_type *var,
		     const tLBool  binary
		     )
{
  
  tLFloat aux;
  llcount_type aux2;
  
  
  if ( binary )
    {
      if ( gzread ( f, &aux2, 1*sizeof(llcount_type) ) != 1*sizeof(llcount_type) ) return 0;
      var->a= RADD ( var->a, aux2.a );
      var->b= RADD ( var->b, aux2.b );
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      aux= tL_atop ( buffer->v );
      var->a= RADD ( var->a, aux );
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      aux= tL_atop ( buffer->v );
      var->b= RADD ( var->b, aux );
    }
  
  return 1;
  
} /* end tL_amcounts_llaccum */


int
tL_amcounts_iread (
		   gzFile         f,
		   tLBuffer      *buffer,
		   unsigned long *var,
		   const tLBool   binary
		   )
{
  
  if ( binary )
    {
      if ( gzread ( f, var, 1*sizeof(unsigned long) ) != 1*sizeof(unsigned long) )
	return 0;
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      *var= atoul ( buffer->v );
    }
  
  return 1;
  
} /* end tL_amcounts_iread */


int
tL_amcounts_logread (
		     gzFile        f,
		     tLBuffer     *buffer,
		     tLFloat      *var,
		     const tLBool  binary
		     )
{
  
  if ( binary )
    {
      if ( gzread ( f, var, sizeof(tLFloat) ) != sizeof(tLFloat) )
	return 0;
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      *var= tL_atop ( buffer->v );
    }
  
  return 1;
  
} /* end tL_amcounts_logread */


int
tL_amcounts_llread (
		    gzFile        f,
		    tLBuffer     *buffer,
		    llcount_type *var,
		    const tLBool  binary
		    )
{
  
  if ( binary )
    {
      if ( gzread ( f, var, 1*sizeof(llcount_type) ) != 1*sizeof(llcount_type) ) return 0;
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      var->a= tL_atop ( buffer->v );
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      var->b= tL_atop ( buffer->v );
    }
  
  return 1;
  
} /* end tL_amcounts_llread */


int
tL_amcounts_raread (
		    gzFile            f,
		    tLBuffer         *buffer,
		    rprop_aux_node_t *var,
		    const tLBool      binary
		    )
{
  
  if ( binary )
    {
      if ( gzread ( f, var, 1*sizeof(rprop_aux_node_t) ) != 1*sizeof(rprop_aux_node_t) ) return 0;
    }
  else
    {
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      var->inc= atof ( buffer->v );
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      var->grad= atof ( buffer->v );
      if ( !tL_gtoken ( f, buffer ) ) return 0;
      var->l0= atof ( buffer->v );
    }
  
  return 1;
  
} /* end tL_amcounts_raread */
