/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  amcounts_public.h - Part pública de 'amcounts.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */

#ifndef __LIBTL_AMCOUNTS_H__
#define __LIBTL_AMCOUNTS_H__

#include <stdio.h>
#include <stdlib.h>

#include "amodel_public.h"
#include "common_public.h"
#include "text_public.h"

/** \defgroup Algs Algorithms
 */
/** \addtogroup Algs
 *  @{
 */
/** \defgroup AMCounts Acoustic model counts
 */
/** \addtogroup AMCounts
 *  @{
 */

#ifndef __TL_PRV
/* TIPUS */

/** Integer counts for tLAModel. This opaque structure is used to
 *  store integer counts for a set of HMMs.
 */
typedef void tLAMiCounts;

/** Logarithmic counts for tLAModel. This opaque structure is used to
 *  store logarithmic counts for a set of HMMs.
 */
typedef void tLAMlogCounts;

/** Mixture coefficient counts. This opaque structure is used to
 *  store the counts related to the mixture coefficients.
 */
typedef void tLMixCounts;


/* FUNCIONS */

/** Accumulates the integer counts from file. This function reads the
 *  integer counts from the provided file, in text or binary format,
 *  and then it accumulates the read counts over the current counts.
 *  It is mandatory to provide the acoustic model which was used to
 *  create the counts, or a topologically equivalent one.
 *
 *  @param counts The integer counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param from File where the counts are stored.
 *  @param binary If TRUE, stored counts are assumed to be in binary
 *  format; else, in text format.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return 0 if counts have been successfully accumulated, -1
 *  otherwise.
 */
TL_API_FUNC(int)
tL_amicounts_accum (
		    const tLAMiCounts  *counts,
		    const tLAModel     *amodel,
		    gzFile              from,
		    const tLBool        binary,
		    char              **err
		    );

/** Frees memory. Frees the memory allocated for the integer
 *  counts. It is mandatory to provide the acoustic model which was
 *  used to create the counts, or a topologically equivalent one.
 *
 *  @param counts The integer counts to free.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 */
TL_API_FUNC(void)
tL_amicounts_free (
		   tLAMiCounts    *counts,
		   const tLAModel *amodel
		   );

/** Returns the counts of the mixture coefficients. This function is
 *  only applicable to HMMs of type Mixture or TiedState Mixture.
 *
 *  @param counts The integer counts.
 *  @param amodel The acoustic model.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The mixture coefficient counts, or NULL in case of error.
 */
TL_API_FUNC(tLMixCounts *)
tL_amicounts_get_mixcounts (
			    tLAMiCounts     *counts,
			    const tLAModel  *amodel,
			    char           **err
			    );

/** Initializes the integer counts to 0. Initializes the integer
 *  counts to 0. It is mandatory to provide the acoustic model
 *  which was used to create the counts, or a topologically
 *  equivalent one.
 *
 *  @param counts The integer counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 */
TL_API_FUNC(void)
tL_amicounts_init (
		   tLAMiCounts    *counts,
		   const tLAModel *amodel
		   );

/** Initializes the integer counts of an HMM to 0. Initializes
 *  the integer counts of the designated HMM to 0. It is
 *  mandatory to provide the acoustic model which was used to
 *  create the counts, or a topologically equivalent one.
 *
 *  @param counts The integer counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param c HMM to be initialized.
 */
TL_API_FUNC(void)
tL_amicounts_init_hmm (
		       tLAMiCounts        *counts,
		       const tLAModel     *amodel,
		       const unsigned int  c
		       );

/** Initializes the integer counts of an HMM. Initializes the
 *  integer counts of the designated HMM to val. It is mandatory
 *  to provide the acoustic model which was used to create the
 *  counts, or a topologically equivalent one.
 *
 *  @param counts The integer counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param c HMM to be initialized.
 *  @param val Value.
 */
TL_API_FUNC(void)
tL_amicounts_init_hmm_val (
			   tLAMiCounts        *counts,
			   const tLAModel     *amodel,
			   const unsigned int  c,
			   const unsigned long val
			   );

/** Initializes the integer counts. Initializes the integer counts to
 *  val. It is mandatory to provide the acoustic model which was
 *  used to create the counts, or a topologically equivalent one.
 *
 *  @param counts The integer counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param val Value.
 */
TL_API_FUNC(void)
tL_amicounts_init_val (
		       tLAMiCounts         *counts,
		       const tLAModel      *amodel,
		       const unsigned long  val
		       );

/** Creates a new AMiCounts. Creates a new AMiCounts to store the
 *  integer counts of the parameters of the provided HMMs. Counts are
 *  not initialized.
 *
 *  @param amodel The provided HMMs.
 *  @return The new AMiCounts.
 */
TL_API_FUNC(tLAMiCounts *)
tL_amicounts_new (
		  const tLAModel *amodel
		  );

/** Prints the integer counts. This function prints the integer counts
 *  to a file in text or binary format. Counts are printed in the
 *  same order as the parameters of amodel are printed. The binary
 *  format is not a compact nor portable representation, but a way to
 *  ensure that counts are written like they are stored in memory. It
 *  is mandatory to provide the acoustic model which was used to
 *  create the counts, or a topologically equivalent one.
 *
 *  @param counts The integer counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param to File to which the counts will be printed.
 *  @param binary If TRUE, counts are printed in binary format; else,
 *  in text format.
 */
TL_API_FUNC(void)
tL_amicounts_print (
		    const tLAMiCounts *counts,
		    const tLAModel    *amodel,
		    FILE              *to,
		    const tLBool       binary
		    );

/** Reads the integer counts from file. This function reads the
 *  integer counts from the provided file, in text or binary format,
 *  overwriting the current counts. It is mandatory to provide the
 *  acoustic model which was used to create the counts, or a
 *  topologically equivalent one.
 *
 *  @param counts The integer counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param from File where counts are stored.
 *  @param binary If TRUE, stored counts are assumed to be in binary
 *  format; else, in text format.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return 0 if counts have been successfully read, -1 otherwise.
 */
TL_API_FUNC(int)
tL_amicounts_read (
		   const tLAMiCounts  *counts,
		   const tLAModel     *amodel,
		   gzFile              from,
		   const tLBool        binary,
		   char              **err
		   );

/** Update parameters of an HMM. This function uses the current
 *  integer counts to reestimate the parameters of the given HMM. It
 *  is mandatory to provide the acoustic model which was used to
 *  create the counts, or a topologically equivalent one. This
 *  function is only supported by HMMs with linear topologies.
 *
 *  @param counts The integer counts.
 *  @param amodel The acoustic model to update.
 *  @param c HMM to be updated.
 */
TL_API_FUNC(void)
tL_amicounts_update_hmm (
			 const tLAMiCounts  *counts,
			 tLAModel           *amodel,
			 const unsigned int  c
			 );

/** Update parameters. This function uses the current integer counts
 *  to reestimate the parameters of the given acoustic model. It is
 *  mandatory to provide the acoustic model which was used to
 *  create the counts, or a topologically equivalent one. This
 *  function is only supported by HMMs with linear topologies.
 *
 *  @param counts The integer counts.
 *  @param amodel The acoustic model to update.
 */
TL_API_FUNC(void)
tL_amicounts_update_amodel (
			    const tLAMiCounts *counts,
			    tLAModel          *amodel
			    );

/** Accumulates the logarithmic counts from file. This function reads
 *  the logarithmic counts from the provided file, in text or binary
 *  format, and then accumulates the read counts over the current
 *  counts. It is mandatory to provide the acoustic model which was
 *  used to create the counts, or a topologically equivalent one.
 *
 *  @param counts The logarithmic counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param from File where counts are stored.
 *  @param binary If TRUE, stored counts are assumed to be in binary
 *  format; else, in text format.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return 0 if counts have been successfully accumulated, -1
 *  otherwise.
 */
TL_API_FUNC(int)
tL_amlogcounts_accum (
		      const tLAMlogCounts  *counts,
		      const tLAModel       *amodel,
		      gzFile                from,
		      const tLBool          binary,
		      char                **err
		      );

/** Interpolates current logarithmic counts with the logarithmic
 *  counts from file. This function reads the logarithmic counts from
 *  the provided file, in text or binary format, and then interpolates
 *  them with the current counts, \f$\mbox{Current}\cdot(1-\epsilon) +
 *  \mbox{New}\cdot\epsilon\f$. It is mandatory to provide the
 *  acoustic model which was used to create the counts, or a
 *  topologically equivalent one.
 *
 *  @param counts The logarithmic counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param from File where counts are stored.
 *  @param eps The epsilon value, must be \f$0\leq\epsilon\leq 1\f$.
 *  @param binary If TRUE, stored counts are assumed to be in binary
 *  format; else, in text format.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return 0 if counts have been successfully accumulated, -1
 *  otherwise.
 */
TL_API_FUNC(int)
tL_amlogcounts_accum_eps (
                          const tLAMlogCounts  *counts,
                          const tLAModel       *amodel,
                          gzFile                from,
                          const tLFloat         eps,
                          const tLBool          binary,
                          char                **err
                          );

/** Frees memory. Frees the memory allocated for the logarithmic
 *  counts. It is mandatory to provide the acoustic model which was
 *  used to create the counts, or a topologically equivalent one.
 *
 *  @param counts The logarithmic counts to free.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 */
TL_API_FUNC(void)
tL_amlogcounts_free (
		     tLAMlogCounts  *counts,
		     const tLAModel *amodel
		     );

/** Returns the counts of the mixture coefficients. This function is
 *  only applicable to HMMs of the type Mixture or TiedState Mixture.
 *
 *  @param counts The logarithmic counts.
 *  @param amodel The acoustic model.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The mixture coefficient counts, or NULL in case of error.
 */
TL_API_FUNC(tLMixCounts *)
tL_amlogcounts_get_mixcounts (
			      tLAMlogCounts   *counts,
			      const tLAModel  *amodel,
			      char           **err
			      );

/** Initializes the logarithmic counts to -INF. Initializes the
 *  logarithmic counts to -INF (-HUGE_VAL). It is mandatory to provide
 *  the acoustic model which was used to create the counts, or a
 *  topologically equivalent one.
 *
 *  @param counts The logarithmic counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 */
TL_API_FUNC(void)
tL_amlogcounts_init (
		     tLAMlogCounts  *counts,
		     const tLAModel *amodel
		     );

/** Initializes the logarithmic counts of an HMM to -INF. Initializes
 *  the logarithmic counts of the designated HMM to -INF. It is
 *  mandatory to provide the acoustic model which was used to create
 *  the counts, or a topologically equivalent one.
 *
 *  @param counts The logarithmic counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param c HMM to be initialized.
 */
TL_API_FUNC(void)
tL_amlogcounts_init_hmm (
			 tLAMiCounts        *counts,
			 const tLAModel     *amodel,
			 const unsigned int  c
			 );

/** Initializes the logarithmic counts of an HMM. Initializes the
 *  logarithmic counts of the designated HMM to val. It is
 *  mandatory to provide the acoustic model which was used to
 *  create the counts, or a topologically equivalent one.
 *
 *  @param counts The logarithmic counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param c HMM to be initialized.
 *  @param val Value.
 */
TL_API_FUNC(void)
tL_amlogcounts_init_hmm_val (
			     tLAMiCounts        *counts,
			     const tLAModel     *amodel,
			     const unsigned int  c,
			     const tLFloat       val
			     );

/** Initializes the logarithmic counts. Initializes the logarithmic
 *  counts to val. It is mandatory to provide the acoustic model
 *  which was used to create the counts, or a topologically equivalent
 *  one.
 *
 *  @param counts The logarithmic counts to initialize.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param val Value.
 */
TL_API_FUNC(void)
tL_amlogcounts_init_val (
			 tLAMlogCounts  *counts,
			 const tLAModel *amodel,
			 const tLFloat   val
			 );

/** Creates a new AMlogCounts. Creates a new AMlogCounts to store the
 *  logarithmic counts of the parameters of the provided acoustic model.
 *  Counts are not initialized.
 *
 *  @param amodel The provided acoustic model.
 *  @return The new AMlogCounts.
 */
TL_API_FUNC(tLAMlogCounts *)
tL_amlogcounts_new (
		    const tLAModel *amodel
		    );

/** Creates a new AMlogCounts for CMLLR. Creates a new AMlogCounts to
 *  store the logarithmic counts, as 'tL_amlogcounts_new', prepared
 *  for the CMLLR algorithm. Counts are not initialized.
 *
 *  @param amodel The provided acoustic model.
 *  @return The new AMlogCounts.
 */
TL_API_FUNC(tLAMlogCounts *)
tL_amlogcounts_new_cmllr (
			  const tLAModel *amodel
			  );

/** Creates a new AMlogCounts for MLLR. Creates a new AMlogCounts to
 *  store the logarithmic counts, as 'tL_amlogcounts_new', prepared
 *  for the MLLR algorithm. Counts are not initialized.
 *
 *  @param amodel The provided acoustic model.
 *  @return The new AMlogCounts.
 */
TL_API_FUNC(tLAMlogCounts *)
tL_amlogcounts_new_mllr (
			 const tLAModel *amodel
			 );

/** Prints the logarithmic counts. This function prints the
 *  logarithmic counts into a file in text or binary format. Counts
 *  are printed in the same order as the parameters of amodel are
 *  printed. The binary format is not a compact nor portable
 *  representation, but a way to ensure that counts are written like
 *  they are stored in memory. It is mandatory to provide the
 *  acoustic model which was used to create the counts, or
 *  a topologically equivalent one.
 *
 *  @param counts The logarithmic counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param to File to which the counts will be printed.
 *  @param binary If TRUE, counts are printed in binary format; else,
 *  in text format.
 */
TL_API_FUNC(void)
tL_amlogcounts_print (
		      const tLAMlogCounts *counts,
		      const tLAModel      *amodel,
		      FILE                *to,
		      const tLBool         binary
		      );

/** Reads the logarithmic counts from file. This function reads the
 *  logarithmic counts from the provided file, in text or binary
 *  format, overwriting the current counts. It is mandatory to provide
 *  the acoustic model which was used to create the counts, or a
 *  topologically equivalent one.
 *
 *  @param counts The logarithmic counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param from File where counts are stored.
 *  @param binary If TRUE, stored counts are assumed to be in binary
 *  format; else, in text format.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return 0 if counts have been successfully read, -1 otherwise.
 */
TL_API_FUNC(int)
tL_amlogcounts_read (
		     const tLAMlogCounts  *counts,
		     const tLAModel       *amodel,
		     gzFile                from,
		     const tLBool          binary,
		     char                **err
		     );

/** Update parameters of an HMM. This function uses the current
 *  logarithmic counts to reestimate the parameters of the given
 *  HMM. It is mandatory to provide the acoustic model which was used
 *  to create the counts, or a topologically equivalent one. This
 *  function is only supported by HMMs with linear topologies.
 *
 *  @param counts The logarithmic counts.
 *  @param amodel The acoustic model to update.
 *  @param c HMM to be updated.
 */
TL_API_FUNC(void)
tL_amlogcounts_update_hmm (
			   const tLAMlogCounts *counts,
			   tLAModel            *amodel,
			   const unsigned int   c
			   );

/** Update parameters. This function uses the current logarithmic
 *  counts to reestimate the parameters of the given acoustic model.
 *  It is mandatory to provide the acoustic model which was
 *  used to create the counts, or a topologically equivalent one.
 *  This function is only supported by HMMs with linear topologies.
 *
 *  @param counts The logarithmic counts.
 *  @param amodel The acoustic model to update.
 */
TL_API_FUNC(void)
tL_amlogcounts_update_amodel (
			      const tLAMlogCounts *counts,
			      tLAModel            *amodel
			      );

/** Frees mixture counts. This function frees the memory allocated for
 *  the mixture coefficient counts.
 *
 *  @param mcounts The mixture coefficient counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 */
TL_API_FUNC(void)
tL_mixcounts_free (
		   tLMixCounts    *mcounts,
		   const tLAModel *amodel
		   );

/** Prints mixture counts. This function prints the mixture
 *  coefficient counts in text format.
 *
 *  @param mcounts The mixture coefficient counts.
 *  @param amodel The acoustic model which was used to create the
 *  counts.
 *  @param syms The symbol dictionary.
 *  @param to File to which the counts will be printed.
 */
TL_API_FUNC(void)
tL_mixcounts_print (
		    const tLMixCounts *mcounts,
		    const tLAModel    *amodel,
		    const tLDict      *syms,
		    FILE              *to
		    );

/** Reads mixture counts from file. This function reads the mixture
 *  coefficient counts from the file provided. The labels in the file
 *  are ignored, therefore what is important is the order of the
 *  counts in the file.
 *
 *  @param amodel The acoustic model which was used to create the
 *  counts stored in the file.
 *  @param from File where counts are stored.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The mixture coefficient counts, or NULL in case of error.
 */
TL_API_FUNC(tLMixCounts *)
tL_mixcounts_read (
		   const tLAModel  *amodel,
		   gzFile           from,
		   char           **err
		   );


/** Frees occupancies. This macro frees the memory allocated for the
 *  occupancies matrix.
 *
 *  @param occs The occupancy values.
 */
#define tL_occupancies_free(occs) \
  do {				  \
    free ( (occs)[0] );		  \
    free ( (occs) );		  \
  } while(0)

/** Inits the occupancy values. Inits to zero (-HUGE_VAL) all the
 *  occupancy values.
 *
 *  @param occs The occupancy values.
 *  @param amodel The acoustic model related to the occupancy
 *  values.
 */
TL_API_FUNC(void)
tL_occupancies_init (
		     tLFloat        **occs,
		     const tLAModel  *amodel
		     );

/** Allocs memory to store the occupancies. This function allocs a
 *  matrix to store state occupancies of the specified acoustic model.
 *  The matrix is accessed as follows: m[c][q]; where 'c' is
 *  the index related to the HMM and 'q' the index related to
 *  the state. Occupancy values are stored using logarithms.
 *
 *  @param amodel The acoustic model.
 *  @return The allocated memory.
 */
TL_API_FUNC(tLFloat **)
tL_occupancies_new (
		    const tLAModel *amodel
		    );

/** Prints the occupancy values. This function prints the values of
 *  the occupancies in plain text format.
 *
 *  @param occs The occupancy values.
 *  @param to File to which the occupancy values will be printed.
 *  @param amodel The acoustic model related to the occupancy
 *  values.
 *  @param syms A dictionary with the symbols related to the HMMs.
 */
TL_API_FUNC(void)
tL_occupancies_print (
		      tLFloat        **occs,
		      FILE            *to,
		      const tLAModel  *amodel,
		      const tLDict    *syms
		      );

#endif /* __TL_PRV */
/** @} */
/** @} */

#endif /* __LIBTL_AMCOUNTS_H__ */
