/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  amodel.c - Implementació de 'amodel.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <float.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "amodel.h"
#include "amodel_itrans.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define CHANGE_PTR_OFFSET(PTR,OFF,TYPE) ((PTR)= (OFF)+((PTR)-((TYPE *)(NULL))))




/**********************/
/* TIPUS ENREGISTRATS */
/**********************/

#include "amodel/bernoulli.h"
#include "amodel/dgaussian.h"
#include "amodel/loglinear.h"
#include "amodel/hybrid.h"
#include "amodel/mixture.h"
#include "amodel/tiedstates.h"

static const amodel_type_t types[]=
  {
    { "Bernoulli", &tL_bernoulli_methods },
    { "DGaussian", &tL_dgaussian_methods },
    { "LogLinear", &tL_loglinear_methods },
    { "LogLinearB", &tL_loglinearb_methods },
    { "Hybrid", &tL_hybrid_methods },
    { "Mixture", &tL_mixture_methods },
    { "TiedStates", &tL_tiedstates_methods },
    { NULL, NULL }
  };




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
copy_trans_list_state_content (
			       tLAModelTransListState       *a,
			       const tLAModelTransListState  b,
			       tLAModelTransListEdge        *abase,
			       const tLAModelTransListEdge  *bbase
			       )
{
  
  a->fp= b.fp;
  a->begin= abase + (b.begin-bbase);
  a->end= abase + (b.end-bbase);
  
} /* end copy_trans_list_state_content */


static tLAModelTrans
copy_trans (
	    const tLAModelHMM  b,
	    tLAModel           *new
	    )
{
  
  tLAModelTrans a;
  unsigned int q;
  tLProbPair *ap;
  tLAModelTransListEdge *edges;
  ptrdiff_t NE, e;
  tLAModelTransListState *states;
  
  
  switch ( b.trans.type )
    {
      
    case TLNULLTRANS:
      a.type= TLNULLTRANS;
      break;
      
    case TLLINEAL:
      a.lineal.type= TLLINEAL;
      a.lineal.ll= b.trans.lineal.ll;
      mt ( ap= a.lineal.p= mem_alloc ( tLProbPair, b.Q ) );
      for ( q= 0; q < b.Q; ++q )
	ap[q]= b.trans.lineal.p[q];
      break;
      
    case TLLIST:
      a.list.type= TLLIST;
      NE= a.list.NE= b.trans.list.NE;
      mt ( edges= a.list.edges=
	   mem_alloc ( tLAModelTransListEdge, NE ) );
      for ( e= 0; e < NE; ++e )
	edges[e]= b.trans.list.edges[e];
      copy_trans_list_state_content ( &(a.list.init),
				      b.trans.list.init,
				      edges, b.trans.list.edges );
      mt ( states= a.list.states=
	   mem_alloc ( tLAModelTransListState, b.Q ) );
      for ( q= 0; q < b.Q; ++q )
	copy_trans_list_state_content ( &(states[q]),
					b.trans.list.states[q],
					edges, b.trans.list.edges );
      a.list.itrans= NULL;
      break;
      
    case TLPOINTER:
      a.pointer.type= TLPOINTER;
      a.pointer.c= b.trans.pointer.c;
      a.pointer.ptr= &(new->hmms[b.trans.pointer.c].trans);
      break;
      
    }
  
  return a;
  
} /* end copy_trans */


static void
alloc_hmms (
	    tLAModel     *amodel,
	    unsigned int  N
	    )
{
  
  unsigned int n;
  
  
  amodel->N= N;
  mt ( amodel->hmms= mem_alloc ( tLAModelHMM, N ) );
  for ( n= 0; n < N; ++n )
    {
      amodel->hmms[n].states= NULL;
      amodel->hmms[n].trans.type= TLNULLTRANS;
    }
  
} /* end alloc_hmms */


static void
alloc_states (
	      tLAModel           *amodel,
	      tLAModelHMM       *cchar,
	      const unsigned int  Q
	      )
{
  
  cchar->Q= Q;
  cchar->states=
    amodel->meths->alloc_states ( amodel->global, Q );
      
} /* end alloc_states */


static void
alloc_trans_lineal (
		    tLAModelHMM *cchar
		    )
{
  
  unsigned int q;
  
  
  mt ( cchar->trans.lineal.p= mem_alloc ( tLProbPair, cchar->Q ) );
  for ( q= 0; q < cchar->Q; ++q )
    cchar->trans.lineal.p[q].p1= cchar->trans.lineal.p[q].p0= 0.0;
  
} /* end alloc_trans_lineal */


static void
alloc_trans_list (
		  tLAModelHMM *cchar
		  )
{
  
  mt ( cchar->trans.list.states= mem_alloc ( tLAModelTransListState,
					     cchar->Q ) );
  cchar->trans.list.edges= NULL;
  cchar->trans.list.itrans= NULL;
  
} /* end alloc_trans_list */


static void
print_list_state (
		  FILE                            *to,
		  const unsigned int               q,
		  const tLAModelTransListState  state
		  )
{
  
  const tLAModelTransListEdge *ed;
  
  
  if ( q == 0 ) fprintf ( to, "I\n" );
  else          fprintf ( to, "%u\n", q );
  for ( ed= state.begin; ed != state.end; ++ed )
    {
      fprintf ( to, "  %u ", ed->q+1 );
      tL_prob_print ( ed->p, to );
      putc ( '\n', to );
    }
  if ( state.fp != -HUGE_VAL )
    {
      fprintf ( to, "  F " );
      tL_prob_print ( state.fp, to );
      putc ( '\n', to );
    }
  fprintf ( to, "  .\n" );
  
} /* end print_list_state */


static ptrdiff_t
print_list_state_binary (
			 FILE                            *to,
			 const tLAModelTransListState  state,
			 const tLAModelTransListEdge       *edges
			 )
{
  
  ptrdiff_t aux;
  
  
  aux= state.begin-edges;
  fwritee ( &aux, sizeof(ptrdiff_t), 1, to );
  aux= state.end-edges;
  fwritee ( &aux, sizeof(ptrdiff_t), 1, to );
  fwritee ( &state.fp, sizeof(tLProb), 1, to );
  
  return aux;
  
} /* end print_list_state_binary */


/* Torna 0 si tot ha anat bé, -1 en cas d'error. */
static int
read_list_state (
		 tLAModelTransListState  *state,
		 gzFile                   from,
		 tLBuffer                *buffer,
		 tLAModelTransListEdge  **edges,
		 unsigned long           *edges_size,
		 unsigned long           *NE,
		 tLDict                  *syms,
		 const unsigned int       n,
		 const unsigned int       Q,
		 const unsigned int       q0,
		 char                   **err
		 )
{
  
  unsigned int q, count;
  
  
  state->fp= -HUGE_VAL;
  state->begin= AMODELNULLEDGE+(*NE);
  count= 0;
  for (;;)
    { 
      ++count;
      if ( !tL_gtoken ( from, buffer ) )
	{
	  if ( q0 == 0 )
	    tL_msgerror ( err, _( "Unable to read the target state of the"
			  " %uth transition from the initial state in"
			  " the symbol '%s'" ) , count, 
			  tL_dict_get_token ( syms, n ) );
	  else
	    tL_msgerror ( err, _( "Unable to read the target state of the"
			  " %uth transition from the %uth state in"
			  " the symbol '%s'" ) , count, q0,
			  tL_dict_get_token ( syms, n ) );
	  return -1;
	}
      if ( buffer->v[0]=='.' && buffer->v[1]=='\0' ) break;
      if ( buffer->v[0]=='F' && buffer->v[1]=='\0' )
	{
	  if ( !tL_gtoken ( from, buffer ) ) goto errorp;
	  state->fp= tL_atop ( buffer->v );
	  continue;
	}
      if ( (*NE) == (*edges_size) )
	{
	  (*edges_size)*= 2;
	  mt ( *edges= mem_realloc ( tLAModelTransListEdge,
				     *edges, *edges_size ) );
	}
      q= atoui ( buffer->v );
      if ( q < 1 || q > Q )
	{
	  if ( q0 == 0 )
	    tL_msgerror ( err, _( "Invalid target state %u in the"
			  " %uth transition from the initial state in"
			  " the symbol '%s'" ) , q, count, 
			  tL_dict_get_token ( syms, n ) );
	  else
	    tL_msgerror ( err, _( "Invalid target state %u in the"
			  " %uth transition from the %uth state in"
			  " the symbol '%s'" ) , q, count, q0,
			  tL_dict_get_token ( syms, n ) );
	  return -1;
	}
      (*edges)[*NE].q= q-1;
      if ( !tL_gtoken ( from, buffer ) ) goto errorp;
      (*edges)[(*NE)++].p= tL_atop ( buffer->v );
    }
  state->end= AMODELNULLEDGE+(*NE);
  
  return 0;
  
 errorp:
  if ( q0 == 0 )
    tL_msgerror ( err, _( "Unable to read the log-probability of the"
		  " %uth transition from the initial state in"
		  " the symbol '%s'" ) , count, 
		  tL_dict_get_token ( syms, n ) );
  else
    tL_msgerror ( err, _( "Unable to read the log-probability of the"
		  " %uth transition from the %uth state in"
		  " the symbol '%s'" ) , count, q0,
		  tL_dict_get_token ( syms, n ) );
  return -1;
  
} /* end read_list_state */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. */
static int
read_list_state_binary (
			ptrdiff_t               *addr,
			gzFile                   from,
			tLAModelTransListState  *state,
			tLDict                  *syms,
			const unsigned int       n,
			const unsigned int       q,
			char                   **err
			)
{
  
  if ( gzread ( from, addr, sizeof(ptrdiff_t) ) != sizeof(ptrdiff_t) )
    {
      if ( q == 0 )
	tL_msgerror ( err, _( "Unable to read the address of the first transition"
		      " of the initial state in the symbol '%s'" ) ,
		      tL_dict_get_token ( syms, n ) );
      else
	tL_msgerror ( err, _( "Unable to read the address of the first transition"
		      " of the %uth state in the symbol '%s'" ) ,
		      q, tL_dict_get_token ( syms, n ) );
      return -1;
    }
  state->begin= ((tLAModelTransListEdge *) NULL)+*addr;
  if ( gzread ( from, addr, sizeof(ptrdiff_t) ) != sizeof(ptrdiff_t) )
    {
      if ( q == 0 )
	tL_msgerror ( err, _( "Unable to read the address of the last transition"
		      " of the initial state in the symbol '%s'" ) ,
		      tL_dict_get_token ( syms, n ) );
      else
	tL_msgerror ( err, _( "Unable to read the address of the last transition"
		      " of the %uth state in the symbol '%s'" ) ,
		      q, tL_dict_get_token ( syms, n ) );
      return -1;
    }
  state->end= ((tLAModelTransListEdge *) NULL)+*addr;
  if ( state->begin > state->end )
    {
      if ( q == 0 )
	tL_msgerror ( err, _( "Invalid transition addresses in the initial"
		      " state of the symbol '%s'" ) ,
		      tL_dict_get_token ( syms, n ) );
      else
	tL_msgerror ( err, _( "Invalid transition addresses in the %uth"
		      " state of the symbol '%s'" ) ,
		      q, tL_dict_get_token ( syms, n ) );
      return -1;
    }
  if ( gzread ( from, &(state->fp), sizeof(tLProb) ) != sizeof(tLProb) )
    {
      if ( q == 0 )
	tL_msgerror ( err, _( "Unable to read the log-probability of the"
		      " transition to the final state in the initial"
		      " state of the symbol '%s'" ) ,
		      tL_dict_get_token ( syms, n ) );
      else
	tL_msgerror ( err, _( "Unable to read the log-probability of the"
		      " transition to the final state in the %uth"
		      " state of the symbol '%s'" ) ,
		      q, tL_dict_get_token ( syms, n ) );
      return -1;
    }
  
  return 0;
  
} /* end read_list_state_binary */


static void
smooth_list_state (
		   tLAModelTransListState *state,
		   const tLFloat           f_1,
		   const tLFloat           f
		   )
{
  
  tLFloat aux2;
  unsigned long N;
  tLBool smooth_fp;
  tLAModelTransListEdge *ed;
  
  
  N= state->end-state->begin;
  if ( (smooth_fp= (state->fp!=-HUGE_VAL)) ) ++N;
  if ( N == 0 ) return;
  aux2= f * 1.0/N;
  if ( smooth_fp ) state->fp= log ( f_1*exp ( state->fp ) + aux2 );
  for ( ed= state->begin; ed != state->end; ++ed )
    ed->p= log ( f_1*exp ( ed->p ) + aux2 );
  
} /* end smooth_list_state */


void
print_amodel_text (
		   FILE           *to,
		   const tLAModel *amodel,
		   const tLDict   *syms
		   )
{
  
  unsigned int n, q;
  tLAModelHMM cchar;
  
  
  fprintf ( to, "AMODEL\n");
  fprintf ( to, "%s\n", amodel->type );
  if ( amodel->meths->print_global != NULL )
    amodel->meths->print_global ( amodel->global, to, FALSE );
  fprintf ( to, "N %u\n", amodel->N );
  
  for ( n= 0; n < amodel->N; ++n )
    {
      
      cchar= amodel->hmms[n];
      
      fprintf ( to, "'%s'\n", tL_dict_get_token ( syms, n ) );
      
      fprintf ( to, "Q %u\n", cchar.Q );
      
      switch ( cchar.trans.type )
	{
	case TLLINEAL:
	  if ( cchar.trans.lineal.ll )
	    {
	      fprintf ( to, "TransP01\n" );
	      for ( q= 0; q < cchar.Q; ++q )
		fprintf ( to, "%g %g\n",
			  _DB(cchar.trans.lineal.p[q].p0),
			  _DB(cchar.trans.lineal.p[q].p1) );
	    }
	  else
	    {
	      fprintf ( to, "Trans\n" );
	      tL_prob_print ( cchar.trans.lineal.p[0].p1, to );
	      for ( q= 1; q < cchar.Q; ++q )
		printprobsep ( to, cchar.trans.lineal.p[q].p1 );
	      fputc ( '\n', to );
	    }
	  break;
	case TLLIST:
	  fprintf ( to, "TransL\n" );
	  print_list_state ( to, 0, cchar.trans.list.init );
	  for ( q= 0; q < cchar.Q; ++q )
	    print_list_state ( to, q+1, cchar.trans.list.states[q] );
	  break;
	case TLPOINTER:
	  fprintf ( to, "TransP %s\n",
		    tL_dict_get_token ( syms, cchar.trans.pointer.c ) );
	  break;
	default: break;
	}
      
      amodel->meths->print_states ( amodel->global,
				    to, cchar.states,
				    0, cchar.Q, FALSE );
      
    }
  
} /* end print_amodel_text */


void
print_amodel_binary (
		     FILE           *to,
		     const tLAModel *amodel,
		     const tLDict   *syms
		     )
{
  
  unsigned int n, q;
  tLAModelHMM cchar;
  ptrdiff_t max, aux;
  
  
  fprintf ( to, "AMODELB\n");
  tL_print_string_binary ( to, amodel->type );
  if ( amodel->meths->print_global != NULL )
    amodel->meths->print_global ( amodel->global, to, TRUE );
  fwritee ( &(amodel->N), sizeof(amodel->N), 1, to );
  
  for ( n= 0; n < amodel->N; ++n )
    {
      
      cchar= amodel->hmms[n];
      
      tL_print_string_binary ( to, tL_dict_get_token ( syms, n ) );
      
      fwritee ( &(cchar.Q), sizeof(cchar.Q), 1, to );
      
      fwritee ( &(cchar.trans.type), sizeof(cchar.trans.type), 1, to );
      switch ( cchar.trans.type )
	{
	case TLLINEAL:
	  fwritee ( &(cchar.trans.lineal.ll),
		    sizeof(cchar.trans.lineal.ll), 1, to );
	  fwritee ( cchar.trans.lineal.p, sizeof(tLProbPair), cchar.Q, to );
	  break;
	case TLLIST:
	  max= print_list_state_binary ( to, cchar.trans.list.init,
					 cchar.trans.list.edges );
	  for ( q= 0; q < cchar.Q; ++q )
	    {
	      aux= print_list_state_binary ( to, cchar.trans.list.states[q],
					     cchar.trans.list.edges );
	      if ( aux > max ) max= aux;
	    }
	  fwritee ( cchar.trans.list.edges, sizeof(tLAModelTransListEdge),
		    max, to );
	  break;
	case TLPOINTER:
	  fwritee ( &(cchar.trans.pointer.c), sizeof(cchar.trans.pointer.c),
		    1, to );
	  break;
	default: break;
	}
      
      amodel->meths->print_states ( amodel->global,
				    to, cchar.states,
				    0, cchar.Q, TRUE );
      
    }
  
} /* end print_amodel_binary */


static tLAModel *
read_amodel_text (
		  gzFile     from,
		  tLDict    *syms,
		  tLBuffer  *buffer,
		  char     **err
		  )
{
  
  tLAModel *amodel;
  unsigned int N, Q, n, q, ne, c;
  tLAModelHMM *cchar;
  char *token;
  amodel_type_t type;
  size_t length;
  tLAModelTransListEdge *edges, *edges2;
  unsigned long edges_size, NE;
  tLAModelTransListState *tlstates;
  

  amodel= NULL;
  length= 0;
  mt ( edges= mem_alloc ( tLAModelTransListEdge, 1 ) );
  edges_size= 1;
  
  /* LLig capçalera i inicialitza. */
  if ( !tL_gtoken ( from, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the type of emission probability"
		    " functions" ) );
      goto error;
    }
  type= tL_search_amodel_type ( buffer->v );
  if ( type.name == NULL )
    {
      tL_msgerror ( err, _( "Unknown type %s" ) , buffer->v );
      goto error;
    }
  mt ( amodel= mem_alloc ( tLAModel, 1 ) );
  amodel->hmms= NULL;
  amodel->type= type.name;
  amodel->meths= type.meths;
  if ( type.meths->read_global != NULL )
    {
      amodel->global=
	type.meths->read_global ( from, buffer, FALSE, err );
      if ( amodel->global == NULL ) goto error;
    }
  else amodel->global= NULL;
  amodel->tflags= 0;
  amodel->IF_jumps= FALSE;
  
  if ( !tL_gtoken ( from, buffer ) ||
       strcmp ( "N", buffer->v ) )
    {
      tL_msgerror ( err, _( "Unable to read the token N" ) );
      goto error;
    }
  if ( !tL_gtoken ( from, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the number of character models N" ) );
      goto error;
    }
  N= atoui ( buffer->v );
  if ( N == 0 )
    {
      tL_msgerror ( err, _( "The number of character models N cannot be 0" ) );
      goto error;
    }
  
  alloc_hmms ( amodel, N );
  
  /* Llig caràcters. */
  amodel->total_Q= amodel->max_Q= 0;
  for ( n= 0; n < N; ++n )
    {
      
      cchar= &(amodel->hmms[n]);
      
      /* Caràcter. */
      if ( !tL_gtoken ( from, buffer ) ||
	   (length= strlen ( buffer->v )) < 3 ||
	   buffer->v[0] != '\'' ||
	   buffer->v[length-1] != '\'' )
	{
	  tL_msgerror ( err, _( "Unable to read the identifier of %uth symbol,"
			" or bad format ('SYM')" ) , n+1 );
	  goto error;
	}
      buffer->v[length-1]= '\0';
      token= buffer->v + 1;
      if ( !tL_dict_insert ( syms, token, NULL ) )
	{
	  tL_msgerror ( err, _( "Symbol '%s' appears repeated" ) , token );
	  goto error;
	}
      
      /* Q. */
      if ( !tL_gtoken ( from, buffer ) ||
	   strcmp ( "Q", buffer->v ) )
	{
	  tL_msgerror ( err, _( "Unable to read the token Q in the symbol '%s'" ) ,
			tL_dict_get_token ( syms, n ) );
	  goto error;
	}
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, _( "Unable to read the number of states Q"
			" in the symbol '%s'" ) , tL_dict_get_token ( syms, n ) );
	  goto error;
	}
      Q= atoui ( buffer->v );
      if ( Q == 0 )
	{
	  tL_msgerror ( err, _( "The number of states Q is 0 in the symbol '%s'" ) ,
			tL_dict_get_token ( syms, n ) );
	  goto error;
	}
      amodel->total_Q+= Q;
      if ( Q > amodel->max_Q ) amodel->max_Q= Q;
      alloc_states ( amodel, cchar, Q );
      
      /* Transicions. */
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, _( "Unable to read the token with"
			" the type of transitions in symbol '%s'" ) ,
			tL_dict_get_token ( syms, n ) );
	  goto error;
	}
      if ( !strcmp ( "TransP01", buffer->v ) )
	{
	  cchar->trans.type= TLLINEAL;
	  cchar->trans.lineal.ll= TRUE;
	  cchar->trans.lineal.p= NULL;
	  amodel->tflags|= FLINEAL;
	}
      else if ( !strcmp ( "Trans", buffer->v ) )
	{
	  cchar->trans.type= TLLINEAL;
	  cchar->trans.lineal.ll= FALSE;
	  cchar->trans.lineal.p= NULL;
	  amodel->tflags|= FLINEAL;
	}
      else if ( !strcmp ( "TransL", buffer->v ) )
	{
	  cchar->trans.type= TLLIST;
	  amodel->tflags|= FLIST;
	  cchar->trans.list.edges= NULL;
	  cchar->trans.list.states= NULL;
	  cchar->trans.list.itrans= NULL;
	}
      else if ( !strcmp ( "TransP", buffer->v ) )
	{
	  cchar->trans.type= TLPOINTER;
	  amodel->tflags|= FPOINTER;
	}
      else
	{
	  tL_msgerror ( err, _( "Unknown transition type '%s' in symbol '%s'" ) ,
			buffer->v, tL_dict_get_token ( syms, n ) );
	  goto error;
	}
      switch ( cchar->trans.type )
	{
	case TLLINEAL:
	  alloc_trans_lineal ( cchar );
	  if ( cchar->trans.lineal.ll )
	    {
	      for ( q= 0; q < Q; ++q )
		{
		  if ( !tL_gtoken ( from, buffer ) )
		    {
		      tL_msgerror ( err, _( "Unable to read TransP01[%d][0]"
				    " in symbol '%s'" ) , q,
				    tL_dict_get_token ( syms, n ) );
		      goto error;
		    }
		  cchar->trans.lineal.p[q].p0= atof ( buffer->v );
		  if ( !tL_gtoken ( from, buffer ) )
		    {
		      tL_msgerror ( err, _( "Unable to read TransP01[%d][1]"
				    " in symbol '%s'" ) , q,
				    tL_dict_get_token ( syms, n ) );
		      goto error;
		    }
		  cchar->trans.lineal.p[q].p1= atof ( buffer->v );
		}
	    }
	  else
	    {
	      for ( q= 0; q < Q; ++q )
		{
		  if ( !tL_gtoken ( from, buffer ) )
		    {
		      tL_msgerror ( err, _( "Unable to read Trans[%d] in"
				    " symbol '%s'" ) , q,
				    tL_dict_get_token ( syms, n ) );
		      goto error;
		    }
		  cchar->trans.lineal.p[q].p1= tL_atop ( buffer->v );
		  cchar->trans.lineal.p[q].p0=
		    p1top0 ( cchar->trans.lineal.p[q].p1 );
		}
	    }
	  break;
	case TLLIST:
	  alloc_trans_list ( cchar );
	  NE= 0;
	  if ( !tL_gtoken ( from, buffer ) ||
	       strcmp ( "I", buffer->v ) )
	    {
	      tL_msgerror ( err, _( "Unable to read the transition list of"
			    " the initial state I in the symbol '%s'" ) ,
			    tL_dict_get_token ( syms, n ) );
	      goto error;
	    }
	  if ( read_list_state ( &(cchar->trans.list.init), from, buffer,
				 &edges, &edges_size, &NE, syms, n, cchar->Q,
				 0, err ) == -1 ) goto error;
	  if ( cchar->trans.list.init.fp != -HUGE_VAL )
	    amodel->IF_jumps= TRUE;
	  tlstates= cchar->trans.list.states;
	  for ( q= 0; q < Q; ++q )
	    {
	      if ( !tL_gtoken ( from, buffer ) )
		{
		  tL_msgerror ( err, _( "Unable to read the transition list of"
				" the %uth state in the symbol '%s'" ) ,
				q+1, tL_dict_get_token ( syms, n ) );
		  goto error;
		}
	      if ( (atoui ( buffer->v ))-1 != q )
		{
		  tL_msgerror ( err, _( "In symbol '%s' the transition list of"
				" the %uth state has been read instead of"
				" the transition list of the %uth state" ) , 
				tL_dict_get_token ( syms, n ),
				atoui ( buffer->v ), q+1 );
		  goto error;
		}
	      if ( read_list_state ( &(tlstates[q]), from, buffer, &edges,
				     &edges_size, &NE, syms, n, cchar->Q,
				     q+1, err ) == -1 )
		goto error;
	    }
	  cchar->trans.list.NE= NE;
	  mt ( cchar->trans.list.edges= edges2=
	       mem_alloc ( tLAModelTransListEdge, NE ) );
	  for ( ne= 0; ne < NE; ++ne )
	    edges2[ne]= edges[ne];
	  cchar->trans.list.init.begin=
	    edges2+(cchar->trans.list.init.begin-AMODELNULLEDGE);
	  cchar->trans.list.init.end=
	    edges2+(cchar->trans.list.init.end-AMODELNULLEDGE);
	  for ( q= 0; q < cchar->Q; ++q )
	    {
	      tlstates[q].begin= edges2+(tlstates[q].begin-AMODELNULLEDGE);
	      tlstates[q].end= edges2+(tlstates[q].end-AMODELNULLEDGE);
	    }
	  break;
	case TLPOINTER:
	  if ( !tL_gtoken ( from, buffer ) )
	    {
	      tL_msgerror ( err, _( "Unable to read the transition list"
			    " reference in the symbol '%s'" ) ,
			    tL_dict_get_token ( syms, n ) );
	      goto error;
	    }
	  cchar->trans.pointer.c= c= tL_dict_find ( syms, buffer->v );
	  if ( c == syms->N )
	    {
	      tL_msgerror ( err, _( "In symbol '%s', the transition list"
			    " reference '%s' references to an unseen symbol" ) ,
			    tL_dict_get_token ( syms, n ), buffer->v );
	      goto error;
	    }
	  if ( cchar->Q != amodel->hmms[c].Q )
	    {
	      tL_msgerror ( err, _( "In symbol '%s', referenced transition"
			    " list has a different number of states than"
			    " the current symbol" ) ,
			    tL_dict_get_token ( syms, n ) );
	      goto error;
	    }
	  if ( amodel->hmms[c].trans.type == TLPOINTER )
	    cchar->trans.pointer.ptr= amodel->hmms[c].trans.pointer.ptr;
	  else
	    cchar->trans.pointer.ptr= &(amodel->hmms[c].trans);
	  break;
	default: break;
	}
      
      /* Estats. */
      if ( type.meths->read_states ( amodel->global, from, buffer,
				     cchar->states, 0, Q,
				     tL_dict_get_token ( syms, n ),
				     FALSE, err ) == -1 )
	goto error;
      
    }
  
  free ( edges );
  
  return amodel;
  
 error:
  
  if ( amodel != NULL ) tL_amodel_free ( amodel );
  free ( edges );
  return NULL;
  
} /* end read_amodel_text */


static tLAModel *
read_amodel_binary (
		    gzFile     from,
		    tLDict    *syms,
		    tLBuffer  *buffer,
		    char     **err
		    )
{
  
  tLAModel *amodel;
  char *token;
  amodel_type_t type;
  unsigned int N, n, Q, q, c;
  tLAModelHMM *cchar;
  size_t nmemb;
  ptrdiff_t max, aux;
  tLAModelTransListEdge *edges;
  
  
  amodel= NULL;
  
  /* Llig capçalera i inicialitza. */
  token= tL_read_string_binary ( from );
  if ( token == NULL )
    {
      tL_msgerror ( err, _( "Unable to read the type of emission probability"
                    " functions" ) );
      goto error;
    }
  type= tL_search_amodel_type ( token );
  if ( type.name == NULL )
    {
      tL_msgerror ( err, _( "Unknown type %s" ) , token );
      free ( token );
      goto error;
    }
  free ( token );
  mt ( amodel= mem_alloc ( tLAModel, 1 ) );
  amodel->hmms= NULL;
  amodel->type= type.name;
  amodel->meths= type.meths;
  if ( type.meths->read_global != NULL )
    {
      amodel->global=
        type.meths->read_global ( from, buffer, TRUE, err );
      if ( amodel->global == NULL ) goto error;
    }
  else amodel->global= NULL;
  amodel->tflags= 0;
  amodel->IF_jumps= FALSE;
  
  if ( gzread ( from, &(N), sizeof(amodel->N) ) != sizeof(amodel->N) )
    {
      tL_msgerror ( err, _( "Unable to read the number of character models N" ) );
      goto error;
    }
  if ( N == 0 )
    {
      tL_msgerror ( err, _( "The number of character models N cannot be 0" ) );
      goto error;
    }
  amodel->N= N;
  
  alloc_hmms ( amodel, N );
  
  /* Llig caràcters. */
  amodel->total_Q= amodel->max_Q= 0;
  for ( n= 0; n < N; ++n )
    {
      
      cchar= &(amodel->hmms[n]);
      
      /* Caràcter. */
      token= tL_read_string_binary ( from );
      if ( token == NULL )
	{
	  tL_msgerror ( err, _( "Unable to read the identifier of %uth symbol,"
                        " or bad format ('SYM')" ) , n+1 );
          goto error;
	}
      if ( !tL_dict_insert ( syms, token, NULL ) )
	{
	  tL_msgerror ( err, _( "Symbol '%s' appears repeated" ) , token );
	  free ( token );
          goto error;
	}
      free ( token );
      
      /* Q. */
      if ( gzread ( from, &Q, sizeof(cchar->Q) ) != sizeof(cchar->Q) )
	{
	  tL_msgerror ( err, _( "Unable to read the number of states Q"
                        " in the symbol '%s'" ) , tL_dict_get_token ( syms, n ) );
          goto error;
	}
      if ( Q == 0 )
	{
	  tL_msgerror ( err, _( "The number of states Q is 0 in the symbol '%s'" ) ,
                        tL_dict_get_token ( syms, n ) );
          goto error;
	}
      amodel->total_Q+= Q;
      if ( Q > amodel->max_Q ) amodel->max_Q= Q;
      alloc_states ( amodel, cchar, Q );
      
      /* Transicions. */
      if ( gzread ( from, &(cchar->trans.type),
		    sizeof(cchar->trans.type) ) != sizeof(cchar->trans.type) )
	{
	  tL_msgerror ( err, _( "Unable to read the type of transitions"
			" in symbol '%s'" ) , tL_dict_get_token ( syms, n ) );
          goto error;
	}
      switch ( cchar->trans.type )
	{
	case TLLINEAL:
	  amodel->tflags|= FLINEAL;
	  nmemb= gzread ( from, &(cchar->trans.lineal.ll), sizeof(tLBool) );
	  if ( nmemb != sizeof(tLBool) ) goto errorlineal;
	  mt ( cchar->trans.lineal.p= mem_alloc ( tLProbPair, Q ) );
	  nmemb= gzread ( from, cchar->trans.lineal.p, Q*sizeof(tLProbPair) );
	  if ( nmemb != Q*sizeof(tLProbPair) ) goto errorlineal;
	  break;
	case TLLIST:
	  amodel->tflags|= FLIST;
	  alloc_trans_list ( cchar );
	  if ( read_list_state_binary ( &max, from, &(cchar->trans.list.init),
					syms, n, 0, err ) == -1 )
	    goto error;
	  if ( cchar->trans.list.init.fp!=-HUGE_VAL ) amodel->IF_jumps= TRUE;
	  for ( q= 0; q < Q; ++q )
	    {
	      if ( read_list_state_binary ( &aux, from,
					    &(cchar->trans.list.states[q]),
					    syms, n, q+1, err ) == -1 )
		goto error;
	      if ( aux > max ) max= aux;
	    }
	  cchar->trans.list.NE= max;
	  mt ( edges= cchar->trans.list.edges=
	       mem_alloc ( tLAModelTransListEdge, max ) );
	  nmemb= gzread ( from, edges, max*sizeof(tLAModelTransListEdge) );
	  if ( nmemb != (size_t) (max*sizeof(tLAModelTransListEdge)) )
	    {
	      tL_msgerror ( err, _( "Unable to read the transition edges in"
			    " the symbol '%s'" ) ,
			    tL_dict_get_token ( syms, n ) );
	      goto error;
	    }
	  CHANGE_PTR_OFFSET ( cchar->trans.list.init.begin,
			      edges, tLAModelTransListEdge );
	  CHANGE_PTR_OFFSET ( cchar->trans.list.init.end,
			      edges, tLAModelTransListEdge );
	  for ( q= 0; q < Q; ++q )
	    {
	      CHANGE_PTR_OFFSET ( cchar->trans.list.states[q].begin,
				  edges, tLAModelTransListEdge );
	      CHANGE_PTR_OFFSET ( cchar->trans.list.states[q].end,
				  edges, tLAModelTransListEdge );
	    }
	  break;
	case TLPOINTER:
	  amodel->tflags|= FPOINTER;
	  if ( gzread ( from, &c, sizeof(cchar->trans.pointer.c) )
	       != sizeof(cchar->trans.pointer.c) )
	    {
	      tL_msgerror ( err, _( "Unable to read the transition list"
                            " reference in the symbol '%s'" ) ,
                            tL_dict_get_token ( syms, n ) );
              goto error;
	    }
	  cchar->trans.pointer.c= c;
	  if ( c >= n  )
	    {
	      tL_msgerror ( err, _( "In symbol '%s', the transition list"
                            " reference references %u to an unseen symbol" ) ,
                            tL_dict_get_token ( syms, n ), c );
              goto error;
	    }
	  if ( cchar->Q != amodel->hmms[c].Q )
	    {
	      tL_msgerror ( err, _( "In symbol '%s', referenced transition"
                            " list has a different number of states than"
                            " the current symbol" ) ,
                            tL_dict_get_token ( syms, n ) );
              goto error;
	    }
	  if ( amodel->hmms[c].trans.type == TLPOINTER )
	    cchar->trans.pointer.ptr= amodel->hmms[c].trans.pointer.ptr;
	  else
	    cchar->trans.pointer.ptr= &(amodel->hmms[c].trans);
	  break;
	default:
	  tL_msgerror ( err, _( "Unknown transition type %d in symbol '%s'" ) ,
                        cchar->trans.type, tL_dict_get_token ( syms, n ) );
          goto error;
	}
      
      /* Estats. */
      if ( type.meths->read_states ( amodel->global, from, buffer,
				     cchar->states, 0, Q,
				     tL_dict_get_token ( syms, n ),
				     TRUE, err ) == -1 )
	goto error;
      
    }
  
  return amodel;

 errorlineal:
  tL_msgerror ( err, _( "Unable to read the transitions"
		" of symbol '%s'" ) , tL_dict_get_token ( syms, n ) );
 error:
  if ( amodel != NULL ) tL_amodel_free ( amodel );
  return NULL;
  
} /* end read_amodel_binary */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

tLAModel *
tL_create_amodel_empty (
			const char         *type,
			const tLDict       *syms,
			const unsigned int  Q
			)
{
  
  tLAModel *new;
  amodel_type_t typ;
  tLAModelHMM *cchar;
  unsigned int q;
  size_t n;
  
  
  tLassert ( tL_create_amodel_empty, syms->N == 0 );
  typ= tL_search_amodel_type ( type );
  if ( typ.name == NULL )
    tL_error ( _( "Unexpected error: Unknown type %s" ), type );
  mt ( new= mem_alloc ( tLAModel, 1 ) );
  new->type= type;
  new->meths= typ.meths;
  new->N= syms->N;
  new->max_Q= Q;
  new->total_Q= Q*syms->N;
  new->tflags= FLINEAL;
  new->IF_jumps= FALSE;
  mt ( new->hmms= mem_alloc ( tLAModelHMM, syms->N ) );
  for ( n= 0; n < syms->N; ++n )
    {
      cchar= &(new->hmms[n]);
      cchar->Q= Q;
      cchar->trans.type= TLLINEAL;
      cchar->trans.lineal.ll= FALSE;
      mt ( cchar->trans.lineal.p= mem_alloc ( tLProbPair, Q ) );
      for ( q= 0; q < Q; ++q )
	cchar->trans.lineal.p[q].p1= cchar->trans.lineal.p[q].p0= 0.0;
    }
  
  return new;
  
} /* end tL_create_amodel_empty */


tLAModel *
tL_create_amodel_empty_vs (
			   const char         *type,
			   const tLDict       *syms,
			   const unsigned int *Qs
			   )
{
  
  tLAModel *new;
  amodel_type_t typ;
  tLAModelHMM *cchar;
  unsigned int q, Q;
  size_t n;
  
  
  typ= tL_search_amodel_type ( type );
  if ( typ.name == NULL )
    tL_error ( _( "Unexpected error: Unknown type %s" ), type );
  mt ( new= mem_alloc ( tLAModel, 1 ) );
  new->type= type;
  new->meths= typ.meths;
  new->N= syms->N;
  new->max_Q= new->total_Q= 0;
  new->tflags= FLINEAL;
  new->IF_jumps= FALSE;
  mt ( new->hmms= mem_alloc ( tLAModelHMM, syms->N ) );
  for ( n= 0; n < syms->N; ++n )
    {
      cchar= &(new->hmms[n]);
      cchar->Q= Q= Qs[n];
      if ( Q > new->max_Q ) new->max_Q= Q;
      new->total_Q+= Q;
      cchar->trans.type= TLLINEAL;
      cchar->trans.lineal.ll= FALSE;
      mt ( cchar->trans.lineal.p= mem_alloc ( tLProbPair, Q ) );
      for ( q= 0; q < Q; ++q )
	cchar->trans.lineal.p[q].p1= cchar->trans.lineal.p[q].p0= 0.0;
    }
  
  return new;
  
} /* end tL_create_amodel_empty_vs */


TL_API_FUNC(tLAModel *)
tL_amodel_copy (
		const tLAModel *amodel
		)
{
  
  tLAModel *new;
  unsigned int n;
  tLAModelHMM *a, b;
  
  
  mt ( new= mem_alloc ( tLAModel, 1 ) );
  *new= *amodel;
  new->global= new->meths->copy_global ( amodel->global );
  mt ( new->hmms= mem_alloc ( tLAModelHMM, new->N ) );
  for ( n= 0; n < new->N; ++n )
    {
      a= &(new->hmms[n]);
      b= amodel->hmms[n];
      a->Q= b.Q;
      a->states= new->meths->alloc_states ( new->global, b.Q );
      new->meths->copy_states ( new->global, b.states, a->states, 0, 0, b.Q );
      a->trans= copy_trans ( b, new );
    }
  
  return new;
  
} /* end tL_amodel_copy */


TL_API_FUNC(void)
tL_amodel_free (
		tLAModel *amodel
		)
{
  
  unsigned int n;
  
  
  if ( amodel->hmms != NULL )
    for ( n= 0; n < amodel->N; ++n )
      {
	if ( amodel->hmms[n].states != NULL )
	  amodel->meths->free_states ( amodel->global,
				       amodel->hmms[n].states,
				       amodel->hmms[n].Q );
	switch ( amodel->hmms[n].trans.type )
	  {
	  case TLLINEAL:
	    if ( amodel->hmms[n].trans.lineal.p != NULL )
	      free ( amodel->hmms[n].trans.lineal.p );
	    break;
	  case TLLIST:
	    if ( amodel->hmms[n].trans.list.edges != NULL )
	      free ( amodel->hmms[n].trans.list.edges );
	    if ( amodel->hmms[n].trans.list.states != NULL )
	      free ( amodel->hmms[n].trans.list.states );
	    if ( amodel->hmms[n].trans.list.itrans != NULL )
	      tL_amodel_free_itrans ( amodel->hmms[n].trans.list.itrans );
	    break;
	  default: break;
	  }
	}
  free ( amodel->hmms );
  if ( amodel->global != NULL )
    amodel->meths->free_global ( amodel->global );
  free ( amodel );
  
} /* end tL_amodel_free */


TL_API_FUNC(void)
tL_amodel_print (
		 const tLAModel *amodel,
		 FILE           *to,
		 const tLDict   *syms,
		 const tLBool    binary
		 )
{
  
  tLassert ( tL_amodel_print, amodel->N > syms->N );
  if ( binary ) print_amodel_binary ( to, amodel, syms );
  else          print_amodel_text ( to, amodel, syms );
  
} /* end tL_amodel_print */


TL_API_FUNC(tLAModel *)
tL_amodel_mixture_new (
		       const tLAModel *amodel
		       )
{
  
  tLAModel *new;
  amodel_type_t typ;
  mixture_global_t *global;
  tLAModelHMM *a, b;
  unsigned int n, q;
  mixture_state_t *states;
  
  
  /* Inicialització bàsica. */
  typ= tL_search_amodel_type ( "Mixture" );
  mt ( new= mem_alloc ( tLAModel, 1 ) );
  *new= *amodel;
  new->type= typ.name;
  new->meths= typ.meths;
  
  /* Global. */
  mt ( global= mem_alloc ( mixture_global_t, 1 ) );
  global->type= amodel->type;
  global->meths= amodel->meths;
  global->global= amodel->meths->copy_global ( amodel->global );
  global->max_I= 1;
  new->global= global;
  
  /* Estats. */
  mt ( new->hmms= mem_alloc ( tLAModelHMM, amodel->N ) );
  for ( n= 0; n < new->N; ++n )
    {
      a= &(new->hmms[n]);
      b= amodel->hmms[n];
      a->Q= b.Q;
      a->states= states= new->meths->alloc_states ( global, b.Q );
      for ( q= 0; q < b.Q; ++q )
	{
	  states[q].I= 1;
	  mt ( states[q].p= mem_alloc ( tLProb, 1 ) );
	  states[q].p[0]= 0.0;
	  states[q].members= amodel->meths->alloc_states ( amodel->global, 1 );
	  amodel->meths->copy_states ( amodel->global, b.states,
				       states[q].members, q, 0, 1 );
	}
      a->trans= copy_trans ( b, new );
    }
  
  return new;
  
} /* end tL_amodel_mixture_new */


TL_API_FUNC(tLAModel *)
tL_amodel_new_from_file (
			 gzFile   from,
			 tLDict  *syms,
			 char   **err
			 )
{
  
  tLAModel *amodel;
  tLBuffer *buffer;
  tLBool binary;
  
  
  tLassert ( tL_amodel_new_from_file, syms->N > 0 );
  buffer= tL_buffer_new ();
  if ( !tL_gtoken ( from, buffer ) )
    {
      TLMSG_TOKID ( err );
      goto error;
    }
  if ( strncmp ( "AMODEL", buffer->v, 6 ) &&
       strncmp ( "CMODEL", buffer->v, 6 ) )
    {
      TLMSG_UNKTOKID ( err, buffer->v );
      goto error;
    }
  binary= (buffer->v[strlen(buffer->v)-1] == 'B');
  if ( binary ) amodel= read_amodel_binary ( from, syms, buffer, err );
  else amodel= read_amodel_text ( from, syms, buffer, err );
  if ( amodel == NULL ) goto error;
  tL_buffer_free ( buffer );
  
  return amodel;
  
 error:
  
  tL_buffer_free ( buffer );
  return NULL;
  
} /* end tL_amodel_new_from_file */


amodel_type_t
tL_search_amodel_type (
		       const char *name
		       )
{
  
  int i;
  
  
  for ( i= 0; types[i].name != NULL &&
	  strcmp ( name, types[i].name); ++i );
  
  return types[i];
  
} /* end tL_search_amodel_type */


TL_API_FUNC(void)
tL_amodel_smooth (
		  tLAModel      *amodel,
		  const tLFloat  f,
		  const tLBool   smooth_trans
		  )
{
  
  unsigned int n, q;
  tLAModelHMM cchar;
  tLFloat aux, aux2;
  
  
  aux= 1.0 - f;
  aux2= f * 0.5;
  for ( n= 0; n < amodel->N; ++n )
    {
      
      cchar= amodel->hmms[n];
      
      /* Transicions. */
      if ( smooth_trans )
	{
	  switch ( cchar.trans.type )
	    {
	    case TLLINEAL:
	      for ( q= 0; q < cchar.Q; ++q )
		smooth ( cchar.trans.lineal.p[q], aux, aux2 );
	      cchar.trans.lineal.ll= FALSE;
	      break;
	    case TLLIST:
	      smooth_list_state ( &(cchar.trans.list.init), aux, f );
	      for ( q= 0; q < cchar.Q; ++q )
		smooth_list_state ( &(cchar.trans.list.states[q]), aux, f );
	      break;
	    default: break;
	    }
	}
      
      /* Estats. */
      amodel->meths->smooth_states ( amodel->global,
				     cchar.states,
				     cchar.Q, f );
      
    }
  if ( amodel->global != NULL )
    amodel->meths->smooth_global ( amodel->global, f );
  
} /* end tL_amodel_smooth */


TL_API_FUNC(const tLDict *)
tL_amodel_tiedstates_get_names (
				const tLAModel *amodel
				)
{
  
  if ( strcmp ( amodel->type, "TiedStates" ) ) return NULL;
  
  return TIEDSTATES_GLOBAL(amodel->global)->names;
  
} /* end tL_amodel_tiedstates_get_names */


TL_API_FUNC(tLAModel *)
tL_amodel_tiedstates_mixture_new (
				  const tLAModel *amodel
				  )
{
  
  tLAModel *new;
  amodel_type_t typ;
  tiedstates_global_t *global;
  mixture_global_t *mglobal;
  tLAModelHMM *a, b;
  unsigned int n;
  mixture_state_t *states;
  size_t id;
  
  
  /* Inicialització bàsica. */
  if ( strcmp ( amodel->type, "TiedStates" ) )
    tL_error ( _( "tL_amodel_tiedstates_mixture_new: amodel type is %s" ),
	       amodel->type );
  mt ( new= mem_alloc ( tLAModel, 1 ) );
  *new= *amodel;
  
  /* Global. */
  typ= tL_search_amodel_type ( "Mixture" );
  mt ( new->global= global= mem_alloc ( tiedstates_global_t, 1 ) );
  global->type= typ.name;
  global->meths= typ.meths;
  global->N= TIEDSTATES_GLOBAL(amodel->global)->N;
  global->names= tL_dict_new ();
  for ( n= 0; n < TIEDSTATES_GLOBAL(amodel->global)->names->N; ++n )
    tL_dict_insert ( global->names,
		     tL_dict_get_token
		     ( TIEDSTATES_GLOBAL(amodel->global)->names, n ), &id );
  mt ( global->global= mglobal= mem_alloc ( mixture_global_t, 1 ) );
  mglobal->type= TIEDSTATES_GLOBAL(amodel->global)->type;
  mglobal->meths= TIEDSTATES_GLOBAL(amodel->global)->meths;
  mglobal->max_I= 1;
  mglobal->global=
    mglobal->meths->copy_global ( TIEDSTATES_GLOBAL(amodel->global)->global );
  
  /* Global Estats. */
  global->states= states=
    global->meths->alloc_states ( global->global, global->N );
  for ( n= 0; n < global->N; ++n )
    {
      states[n].I= 1;
      mt ( states[n].p= mem_alloc ( tLProb, 1 ) );
      states[n].p[0]= 0.0;
      states[n].members=
	mglobal->meths->alloc_states
	( TIEDSTATES_GLOBAL(amodel->global)->global, 1 );
      mglobal->meths->copy_states
	( TIEDSTATES_GLOBAL(amodel->global)->global,
	  TIEDSTATES_GLOBAL(amodel->global)->states,
	  states[n].members, n, 0, 1 );
    }
  
  /* Estats. */
  mt ( new->hmms= mem_alloc ( tLAModelHMM, amodel->N ) );
  for ( n= 0; n < new->N; ++n )
    {
      a= &(new->hmms[n]);
      b= amodel->hmms[n];
      a->Q= b.Q;
      a->states= amodel->meths->alloc_states ( amodel->global, b.Q );
      amodel->meths->copy_states ( amodel->global, b.states,
				   a->states, 0, 0, b.Q );
      a->trans= copy_trans ( b, new );
    }
  
  return new;
  
} /* end tL_amodel_tiedstates_mixture_new */
