/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  bernoulli.c - Implementació de 'bernoulli.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2011,2012.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../../intl.h"

#include "bernoulli.h"
#include "amodel.h"
#include "common.h"
#include "error.h"
#include "global.h"
#include "mem.h"




/**********/
/* MACROS */
/**********/

#define STATE(ptr) BERNOULLI_STATE(ptr)
#define GLOBAL(ptr) BERNOULLI_GLOBAL(ptr)




/***********/
/* MÈTODES */
/***********/

static void *
alloc_states (
	      void               *global,
	      const unsigned int  Q
	      )
{
  
  bernoulli_state_t *ret;
  unsigned int D, q;
  tLProbPair *aux;
  tLProb *aux2;
  
  
  mt ( ret= mem_alloc ( bernoulli_state_t, Q ) );
  D= GLOBAL(global)->D;
  mt ( ret[0].p= mem_alloc ( tLProbPair, D*Q ) );
  for ( q= 1, aux= ret[0].p + D;
	q < Q; ++q, aux+= D )
    ret[q].p= aux;
  mt ( ret[0].l= mem_alloc ( tLProb, D*Q ) );
  for ( q= 1, aux2= ret[0].l + D;
	q < Q; ++q, aux2+= D )
    ret[q].l= aux2;
  
  return (void *) ret;
  
} /* end alloc_states */


static void *
realloc_states (
		void               *global,
		void               *states,
		const unsigned int  oldQ,
		const unsigned int  newQ
		)
{
  
  bernoulli_state_t *ret;
  unsigned int D, q;
  tLProbPair *aux;
  tLProb *aux2;
  
  
  mt ( ret= mem_realloc ( bernoulli_state_t, STATE(states), newQ ) );
  D= GLOBAL(global)->D;
  mt ( ret[0].p= mem_realloc ( tLProbPair, ret[0].p, D*newQ ) );
  for ( q= 1, aux= ret[0].p + D; q < newQ; ++q, aux+= D )
    ret[q].p= aux;
  mt ( ret[0].l= mem_realloc ( tLProb, ret[0].l, D*newQ ) );
  for ( q= 1, aux2= ret[0].l + D; q < newQ; ++q, aux2+= D )
    ret[q].l= aux2;
  
  return (void *) ret;
  
} /* end realloc_states */


static void
copy_states (
	     void               *global,
	     void               *src_states,
	     void               *dst_states,
	     const unsigned int  src_begin,
	     const unsigned int  dst_begin,
	     const unsigned int  Q
	     )
{
  
  bernoulli_state_t *a, b;
  unsigned int q, d, D;
  
  
  D= BERNOULLI_GLOBAL(global)->D;
  for ( q= 0; q < Q; ++q )
    {
      b= BERNOULLI_STATE(src_states)[src_begin+q];
      a= &(BERNOULLI_STATE(dst_states)[dst_begin+q]);
      a->logc= b.logc;
      a->zeros= b.zeros;
      for ( d= 0; d < D; ++d )
	a->p[d]= b.p[d];
      for ( d= 0; d < D; ++d )
	a->l[d]= b.l[d];
    }
  
} /* end copy_states */


static void
free_states (
	     void               *global,
	     void               *states,
	     const unsigned int  Q
	     )
{
  
  free ( STATE(states)[0].l );
  free ( STATE(states)[0].p );
  free ( states );
  
} /* end free_states */


static void
print_states (
	      void               *global,
	      FILE               *to,
	      void               *states,
	      const unsigned int  first,
	      const unsigned int  Q,
	      const tLBool        binary
	      )
{
  
  unsigned int q, D, end;
  
  
  D= GLOBAL(global)->D;
  if ( binary ) fwritee ( STATE(states)[first].p, sizeof(tLProbPair), D*Q, to );
  else
    {
      end= first+Q;
      for ( q= first; q < end; ++q )
	tL_Bernoulli_print_state ( D, to, &(STATE(states)[q]) );
    }
  
} /* end print_states */


static int
read_states (
	     void                *global,
	     gzFile               from,
	     tLBuffer            *buffer,
	     void                *states,
	     const unsigned int   first,
	     const unsigned int   Q,
	     const char          *sym,
	     const tLBool         binary,
	     char               **err
	     )
{
  
  unsigned int q, D, end;
  unsigned long size;
  
  
  end= first+Q;
  D= GLOBAL(global)->D;
  if ( binary ) 
    {
      size= D*Q;
      if ( gzread ( from, STATE(states)[first].p,
		    size*sizeof(tLProbPair) ) != size*sizeof(tLProbPair) )
	{
	  tL_msgerror ( err, _( "Unable to read the Bernoulli"
			" prototypes of symbol '%s'" ) , sym );
	  return -1;
	}
      for ( q= first; q < end; ++q )
	tL_Bernoulli_update_hiddendata ( D, &(STATE(states)[q]) );
    }
  else
    for ( q= first; q < end; ++q )
      {
	if ( tL_Bernoulli_read_state ( D, from, buffer, &(STATE(states)[q]),
				       q, sym, err ) == -1 )
	  return -1;
	tL_Bernoulli_update_hiddendata ( D, &(STATE(states)[q]) );
      }
  
  return 0;
  
} /* end read_states */


static void
smooth_states (
	       void               *global,
	       void               *states,
	       const unsigned int  Q,
	       const tLFloat       f
	       )
{
  
  unsigned int q, D;
  tLFloat aux, aux2;
  
  
  D= GLOBAL(global)->D;
  
  aux= 1.0 - f;
  aux2= f * 0.5;
  
  for ( q= 0; q < Q; ++q )
    {
      tL_Bernoulli_smooth_state ( D, &(STATE(states)[q]), aux, aux2 );
      tL_Bernoulli_update_hiddendata ( D, &(STATE(states)[q]) );
    }
  
} /* end smooth_states */


const tLAModelMethods tL_bernoulli_methods=
  {
    alloc_states,
    realloc_states,
    tL_Bernoulli_copy_global,
    copy_states,
    tL_Bernoulli_free_global,
    free_states,
    tL_get_fea_type_pbm,
    tL_Bernoulli_print_global,
    print_states,
    tL_Bernoulli_read_global_meth,
    read_states,
    tL_smooth_global_empty,
    smooth_states,
    tL_Bernoulli_test
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_Bernoulli_assign_global (
			    bernoulli_global_t       *a,
			    const bernoulli_global_t *b
			    )
{
  a->D= b->D;
} /* end tL_Bernoulli_assign_global */


void *
tL_Bernoulli_copy_global (
			  void *global
			  )
{
  
  bernoulli_global_t *new;
  
  
  mt ( new= mem_alloc ( bernoulli_global_t, 1) );
  tL_Bernoulli_assign_global ( new, BERNOULLI_GLOBAL(global) );
  
  return new;
  
} /* end tL_Bernoulli_copy_global */


void
tL_Bernoulli_free_global (
			  void *global
			  )
{
  free ( global );
} /* end tL_Bernoulli_free_global */


void
tL_Bernoulli_print_global (
			   void         *global,
			   FILE         *to,
			   const tLBool  binary
			   )
{
  
  if ( binary )
    fwritee ( &(GLOBAL(global)->D), sizeof(GLOBAL(global)->D), 1, to );
  else fprintf ( to, "D %u\n", GLOBAL(global)->D );
  
} /* end tL_Bernoulli_print_global */


void
tL_Bernoulli_print_state (
			  const unsigned int       D,
			  FILE                    *to,
			  const bernoulli_state_t *state
			  )
{
  
  unsigned int d;
  
  
  fprintf ( to, "P" );
  for ( d= 0; d < D; ++d )
    printprobsep ( to, state->p[d].p1 );
  fputc ( '\n', to );
  
} /* end tL_Bernoulli_print_state */


int
tL_Bernoulli_read_global (
			  bernoulli_global_t  *global,
			  gzFile               from,
			  tLBuffer            *buffer,
			  const tLBool         binary,
			  char               **err
			  )
{
  
  if ( binary )
    {
      if ( gzread ( from, &(global->D),
		    sizeof(global->D) ) != sizeof(global->D) )
	goto errorD;
      if ( global->D == 0 ) goto errorD0;
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ||
	   strcmp ( "D", buffer->v ) )
	{
	  MSGERR_GTOK ( err, "D" );
	  return -1;
	}
      if ( !tL_gtoken ( from, buffer ) ) goto errorD;
      global->D= atoui ( buffer->v );
      if ( global->D == 0 ) goto errorD0;
    }
  
  return 0;
 
 errorD:
  MSGERR_GDESC ( err, _( "the feature vector dimension D" ) );
  return -1;
 errorD0:
  MSGERR_GDESC0 ( err, _( "The feature vector dimension D" ) );
  return -1;
  
} /* end Bernoulli_read_global */


void *
tL_Bernoulli_read_global_meth (
			       gzFile         from,
			       tLBuffer      *buffer,
			       const tLBool   binary,
			       char         **err
			       )
{
  
  bernoulli_global_t *global;
  
  
  mt ( global= mem_alloc ( bernoulli_global_t, 1) );
  if ( tL_Bernoulli_read_global ( global, from, buffer, binary, err ) == -1 )
    {
      free ( global );
      return NULL;
    }
  
  return (void *) global;
  
} /* end tL_Bernoulli_read_global_meth */


int
tL_Bernoulli_read_state (
			 const unsigned int   D,
			 gzFile               from,
			 tLBuffer            *buffer,
			 bernoulli_state_t   *state,
			 const unsigned int   q,
			 const char          *sym,
			 char               **err
			 )
{
  
  unsigned int d;
  
  
  /* P. */
  if ( !tL_gtoken ( from, buffer ) ||
       strcmp ( "P", buffer->v ) )
    {
      MSGERR_TOK ( err, "P", q, sym );
      return -1;
    }
  for ( d= 0; d < D; ++d )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  MSGERR_VAL ( err, "P", d, q, sym );
	  return -1;
	}
      state->p[d].p1= tL_atop ( buffer->v );
      state->p[d].p0= p1top0 ( state->p[d].p1 );
    }
  
  return 0;
  
} /* end tL_Bernoulli_read_state */


void
tL_Bernoulli_smooth_state (
			   const unsigned int  D,
			   bernoulli_state_t  *state,
			   const tLFloat       _f,
			   const tLFloat       f05
			   )
{
  
  unsigned int d;
  tLProbPair *p;
  
  
  p= state->p;
  for ( d= 0; d < D; ++d )
    smooth ( p[d], _f, f05 );
  
} /* end tL_Bernoulli_smooth_states */


tLBool
tL_Bernoulli_test (
		   void         *global,
		   const tLFea  *fea,
		   char        **err
		   )
{
  
  unsigned int D;
  
  
  D= GLOBAL(global)->D;
  if ( fea->rows != D )
    {
      tL_msgerror ( err, _( "Input feature vectors are %d-dimensional,"
		    " but model is only compatible with %lu-dimensional"
		    " feature vectors" ) , fea->rows, D );
      return FALSE;
    }
  
  return TRUE;
  
} /* end tL_Bernoulli_test */


void
tL_Bernoulli_update_hiddendata (
				const unsigned int  D,
				bernoulli_state_t  *state
				)
{
  
  unsigned int d;
  bernoulli_state_t aux; 
  
  aux= *state;
  aux.logc= 0.0;
  for ( d= 0; d < D; ++d )
    {
      aux.logc+= aux.p[d].p0;
      aux.l[d]= aux.p[d].p1-aux.p[d].p0;
    }
  state->logc= aux.logc;
  state->zeros= (state->logc==-HUGE_VAL);
  
} /* end tL_Bernoulli_update_logc */
