/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  dgaussian.c - Implementació de 'dgaussian.h'.
 *
 *  Adrià Giménez Pastor, 2009,2010,2011,2012.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../../intl.h"

#include "bernoulli.h"
#include "dgaussian.h"
#include "error.h"
#include "global.h"
#include "mem.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define STATE(ptr) DGAUSSIAN_STATE(ptr)
#define GLOBAL(ptr) DGAUSSIAN_GLOBAL(ptr)

#define LOG2PI 1.83787706641




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
print_state (
	     const dgaussian_state_t  state,
	     const unsigned int       D,
	     FILE                    *to
	     )
{
  
  unsigned int d;
  
  
  fprintf ( to, "MU" );
  for ( d= 0; d < D; ++d )
    fprintf ( to, " %g", _DB(state.mu[d]) );
  fputc ( '\n', to );
  fprintf ( to, "VAR" );
  for ( d= 0; d < D; ++d )
    fprintf ( to, " %g", _DB(state.var[d]) );
  fputc ( '\n', to );
  
} /* end print_state */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. */
static int
read_state (
	    const unsigned int   D,
	    dgaussian_state_t   *state,
	    gzFile               from,
	    tLBuffer            *buffer,
	    const unsigned int   q,
	    const char          *sym,
	    char               **err
	    )
{
  
  unsigned int d;
  
  
  if ( !tL_gtoken ( from, buffer ) ||
       strcmp ( "MU", buffer->v ) )
    {
      MSGERR_TOK ( err, "MU", q, sym );
      return -1;
    }
  for ( d= 0; d < D; ++d )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  MSGERR_VAL ( err, "MU", d, q, sym );
	  return -1;
	}
      state->mu[d]= atof ( buffer->v );
    }
  
  if ( !tL_gtoken ( from, buffer ) ||
       strcmp ( "VAR", buffer->v ) )
    {
      MSGERR_TOK ( err, "VAR", q, sym );
      return -1;
    }
  for ( d= 0; d < D; ++d )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  MSGERR_VAL ( err, "VAR", d, q, sym );
	  return -1;
	}
      state->var[d]= atof ( buffer->v );
      state->ivar[d]= 1.0/state->var[d];
    }
  
  tL_DGaussian_update_logc ( D, state );
  
  return 0;
  
} /* end read_state */




/***********/
/* MÈTODES */
/***********/

static void *
alloc_states (
              void               *global,
              const unsigned int  Q
              )
{
  
  dgaussian_state_t *ret;
  unsigned int D, q;
  tLFloat *mem;
  
  
  D= GLOBAL(global)->D;
  mt ( ret= mem_alloc ( dgaussian_state_t, Q ) );
  mt ( mem= mem_alloc ( tLFloat, Q*(D*3) ) );
  for ( q= 0; q < Q; ++q )
    {
      ret[q].mu= mem; mem+= D;
      ret[q].var= mem; mem+= D;
      ret[q].ivar= mem; mem+= D;
    }
  
  return (void *) ret;
  
} /* end alloc_states */


static void *
realloc_states (
		void               *global,
		void               *states,
		const unsigned int  oldQ,
		const unsigned int  newQ
		)
{
  
  dgaussian_state_t *ret;
  unsigned int D, q;
  tLFloat *mem;
  
  
  D= GLOBAL(global)->D;
  mt ( ret= mem_realloc ( dgaussian_state_t, STATE(states), newQ ) );
  mt ( mem= mem_realloc ( tLFloat, ret[0].mu, newQ*(D*3) ) );
  for ( q= 0; q < newQ; ++q )
    {
      ret[q].mu= mem; mem+= D;
      ret[q].var= mem; mem+= D;
      ret[q].ivar= mem; mem+= D;
    }
  
  return (void *) ret;
  
} /* end realloc_states */


static void *
copy_global (
	     void *global
	     )
{
  
  dgaussian_global_t *new;
  unsigned int d;
  
  
  mt ( new= mem_alloc ( dgaussian_global_t, 1 ) );
  new->D= DGAUSSIAN_GLOBAL(global)->D;
  mt ( new->smooth= mem_alloc ( tLFloat, new->D ) );
  for ( d= 0; d < new->D; ++d )
    new->smooth[d]= DGAUSSIAN_GLOBAL(global)->smooth[d];
  
  return new;
  
} /* end copy_global */


static void
copy_states (
	     void               *global,
	     void               *src_states,
	     void               *dst_states,
	     const unsigned int  src_begin,
	     const unsigned int  dst_begin,
	     const unsigned int  Q
	     )
{
  
  dgaussian_state_t *a, b;
  unsigned int q, d, D;
  
  
  D= DGAUSSIAN_GLOBAL(global)->D;
  for ( q= 0; q < Q; ++q )
    {
      a= &(DGAUSSIAN_STATE(dst_states)[dst_begin+q]);
      b= DGAUSSIAN_STATE(src_states)[src_begin+q];
      for ( d= 0; d < D; ++d ) a->mu[d]= b.mu[d];
      for ( d= 0; d < D; ++d ) a->var[d]= b.var[d];
      for ( d= 0; d < D; ++d ) a->ivar[d]= b.ivar[d];
      a->logc= b.logc;
    }
  
} /* end copy_states */


static void
free_global (
	     void *global
	     )
{
  
  free ( GLOBAL(global)->smooth );
  free ( global );
  
} /* end free_global */


static void
free_states (
	     void               *global,
	     void               *states,
	     const unsigned int  Q
	     )
{
  
  free ( STATE(states)[0].mu );
  free ( states );
  
} /* end free_states */


static void
print_global (
	      void         *global,
	      FILE         *to,
	      const tLBool  binary
	      )
{
  
  unsigned int D, d;
  tLFloat *smooth;
  
  
  D= GLOBAL(global)->D;
  if ( binary )
    {
      fwritee ( &D, sizeof(D), 1, to );
      fwritee ( GLOBAL(global)->smooth, sizeof(tLFloat), D, to );
    }
  else
    {
      fprintf ( to, "D %u\n", D );
      smooth= GLOBAL(global)->smooth;
      fprintf ( to, "SMOOTH" );
      for ( d= 0; d < D; ++d )
	fprintf ( to, " %g", _DB(smooth[d]) );
      fputc ( '\n', to );
    }
  
} /* end print_global */


static void
print_states (
	      void               *global,
	      FILE               *to,
	      void               *states,
	      const unsigned int  first,
	      const unsigned int  Q,
	      const tLBool        binary
	      )
{
  
  unsigned int q, D, end;
  
  
  /* NOTA: Per a cada estat els tres vectors estan consecutius en
     memòria. */
  D= GLOBAL(global)->D;
  if ( binary ) fwritee ( STATE(states)[first].mu, sizeof(tLFloat), Q*D*3, to );
  else
    {
      end= first+Q;
      for ( q= first; q < end; ++q )
	print_state ( STATE(states)[q], D, to );
    }
  
} /* end print_states */


static void *
read_global (
	     gzFile         from,
	     tLBuffer      *buffer,
	     const tLBool   binary,
	     char         **err
	     )
{
  
  dgaussian_global_t *global;
  unsigned int D, d;
  
  
  mt ( global= mem_alloc ( dgaussian_global_t, 1 ) );
  global->smooth= NULL;
  
  if ( binary )
    {
      if ( gzread ( from, &(global->D), 
		    sizeof(global->D) ) != sizeof(global->D) )
        goto errorD;
      D= global->D;
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ||
	   strcmp ( "D", buffer->v ) ) goto errorD;
      if ( !tL_gtoken ( from, buffer ) )
	{
	  MSGERR_GDESC ( err, _( "the feature vector dimension D" ) );
	  goto error;
	}
      D= global->D= atoui ( buffer->v );
    }
  if ( D == 0 )
    {
      MSGERR_GDESC0 ( err, _( "The feature vector dimension D" ) );
      goto error;
    }
  
  mt ( global->smooth= mem_alloc ( tLFloat, D ) );
  if ( binary )
    {
      if ( gzread ( from, global->smooth,
		    D*sizeof(tLFloat) ) != D*sizeof(tLFloat) )
	{
	  MSGERR_GDESC ( err, _( "SMOOTH values" ) );
	  goto error;
	}
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ||
	   strcmp ( "SMOOTH", buffer->v ) )
	{
	  MSGERR_GTOK ( err, "SMOOTH" );
	  goto error;
	}
      for ( d= 0; d < D; ++d )
	{
	  if ( !tL_gtoken ( from, buffer ) )
	    {
	      MSGERR_GVAL ( err, "SMOOTH", d );
	      goto error;
	    }
	  global->smooth[d]= atof ( buffer->v );
	}
    }
  
  return (void *) global;
  
 errorD:
  MSGERR_GTOK ( err, "D" );
 error:
  if ( global->smooth != NULL ) free ( global->smooth );
  free ( global );
  return NULL;
  
} /* end read_global */


static int
read_states (
             void                *global,
             gzFile               from,
             tLBuffer            *buffer,
             void                *states,
	     const unsigned int   first,
             const unsigned int   Q,
             const char          *sym,
	     const tLBool         binary,
	     char               **err
             )
{
  
  unsigned int q, D, end;
  dgaussian_state_t *state;
  unsigned long size;
  
  
  D= GLOBAL(global)->D;
  end= first+Q;
  if ( binary )
    {
      size= Q*D*3;
      if ( gzread ( from, STATE(states)[first].mu, size*sizeof(tLFloat) ) 
          != size*sizeof(tLFloat) )
	{
	  MSGERR_DESCS ( err, _( "the contents of the states" ) , sym );
	  return -1;
	}
      for ( q= first; q < end; ++q )
	tL_DGaussian_update_logc ( D, &(STATE(states)[q]) );
    }
  else
    for ( q= first; q < end; ++q )
      {
	state= &(STATE(states)[q]);
	if ( read_state ( D, state, from, buffer, q, sym, err ) == -1 )
	  return -1;
      }
  
  return 0;
  
} /* end read_states */


static void
smooth_states (
               void               *global,
               void               *states,
               const unsigned int  Q,
               const tLFloat       f
               )
{
  /* EL SUAVITZAT ÉS OBLIGAT, FORMAT PART DE L'ACTUALIZACIÓ DE PARÀMETRES. */
  /*
  unsigned int q, D, d, count;
  dgaussian_state_t state;
  const tLFloat *smooth;
  
  
  D= GLOBAL(global)->D;
  smooth= GLOBAL(global)->smooth;
  for ( q= 0; q < Q; ++q )
    {
      state= STATE(states)[q];
      count= 0;
      for ( d= 0; d < D; ++d )
	if ( state.var[d] < smooth[d] )
	  {
	    ++count;
	    state.var[d]= smooth[d];
	    state.ivar[d]= 1.0/smooth[d];
	  }
      if ( count )
	tL_DGaussian_update_logc ( D, &(STATE(states)[q]) );
    }
  */
} /* end smooth_states */


const tLAModelMethods tL_dgaussian_methods=
  {
    alloc_states,
    realloc_states,
    copy_global,
    copy_states,
    free_global,
    free_states,
    tL_get_fea_type_real,
    print_global,
    print_states,
    read_global,
    read_states,
    tL_smooth_global_empty,
    smooth_states,
    tL_Bernoulli_test
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_DGaussian_update_logc (
			  const unsigned int  D,
			  dgaussian_state_t  *state
			  )
{
  
  unsigned int d;
  tLFloat aux;
  const tLFloat *var;
  
  
  aux= 0;
  var= state->var;
  for ( d= 0; d < D; ++d )
    aux+= log ( var[d] );
  aux+= D*LOG2PI;
  state->logc= -0.5*aux;
  
} /* end tL_DGaussian_update_logc */
