/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  loglinear.c - Implementació de 'loglinear.h'.
 *
 *  Adrià Giménez Pastor, 2012.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#include "bernoulli.h"
#include "error.h"
#include "loglinear.h"
#include "global.h"
#include "mem.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define STATE(ptr) LOGLINEAR_STATE(ptr)
#define GLOBAL(ptr) LOGLINEAR_GLOBAL(ptr)




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
print_state (
             const loglinear_state_t  state,
             const unsigned int       D,
             FILE                    *to
             )
{
  
  unsigned int d;
  
  
  fprintf ( to, "L" );
  for ( d= 0; d < D; ++d )
    fprintf ( to, " %g", _DB(state.l[d]) );
  fputc ( '\n', to );
  
} /* end print_state */


static int
read_state (
            const unsigned int   D,
            loglinear_state_t   *state,
            gzFile               from,
            tLBuffer            *buffer,
            const unsigned int   q,
            const char          *sym,
            char               **err
            )
{
  
  unsigned int d;
  
  
  if ( !tL_gtoken ( from, buffer ) ||
       strcmp ( "L", buffer->v ) )
    {
      MSGERR_TOK ( err, "L", q, sym );
      return -1;
    }
  for ( d= 0; d < D; ++d )
    {
      if ( !tL_gtoken ( from, buffer ) )
        {
          MSGERR_VAL ( err, "L", d, q, sym );
          return -1;
        }
      state->l[d]= atof ( buffer->v );
    }
  
  return 0;
  
} /* end read_state */




/***********/
/* MÈTODES */
/***********/

static void *
alloc_states (
              void               *global,
              const unsigned int  Q
              )
{
  
  loglinear_state_t *ret;
  unsigned int D, q;
  tLFloat *mem;
  
  
  D= GLOBAL(global)->D;
  mt ( ret= mem_alloc ( loglinear_state_t, Q ) );
  mt ( mem= mem_alloc ( tLFloat, Q*D ) );
  for ( q= 0; q < Q; ++q, mem+= D )
    ret[q].l= mem;
  
  return (void *) ret;
  
} /* end alloc_states */


static void *
realloc_states (
                void               *global,
                void               *states,
                const unsigned int  oldQ,
                const unsigned int  newQ
                )
{
  
  loglinear_state_t *ret;
  unsigned int D, q;
  tLFloat *mem;
  
  
  D= GLOBAL(global)->D;
  mt ( ret= mem_realloc ( loglinear_state_t, STATE(states), newQ ) );
  mt ( mem= mem_realloc ( tLFloat, ret[0].l, newQ*D ) );
  for ( q= 0; q < newQ; ++q, mem+= D )
    ret[q].l= mem;
  
  return (void *) ret;
  
} /* end realloc_states */


static void
copy_states (
             void               *global,
             void               *src_states,
	     void               *dst_states,
             const unsigned int  src_begin,
	     const unsigned int  dst_begin,
             const unsigned int  Q
             )
{
  
  loglinear_state_t *a, b;
  unsigned int q, d, D;
  
  
  D= GLOBAL(global)->D;
  for ( q= 0; q < Q; ++q )
    {
      a= &(STATE(dst_states)[dst_begin+q]);
      b= STATE(src_states)[src_begin+q];
      for ( d= 0; d < D; ++d ) a->l[d]= b.l[d];
    }
  
} /* end copy_states */


static void
free_states (
             void               *global,
             void               *states,
             const unsigned int  Q
             )
{
  
  free ( STATE(states)[0].l );
  free ( states );
  
} /* end free_states */


static void
print_states (
              void               *global,
              FILE               *to,
              void               *states,
	      const unsigned int  first,
              const unsigned int  Q,
              const tLBool        binary
              )
{
  
  unsigned int q, D, end;
  
  
  D= GLOBAL(global)->D;
  if ( binary ) fwritee ( STATE(states)[first].l, sizeof(tLFloat), Q*D, to );
  else
    {
      end= first+Q;
      for ( q= first; q < end; ++q )
	print_state ( STATE(states)[q], D, to );
    }
  
} /* end print_states */


static int
read_states (
             void                *global,
             gzFile               from,
             tLBuffer            *buffer,
             void                *states,
	     const unsigned int   first,
             const unsigned int   Q,
             const char          *sym,
             const tLBool         binary,
             char               **err
             )
{
  
  unsigned int q, D, end;
  loglinear_state_t *state;
  unsigned long size;
  
  
  end= first+Q;
  D= GLOBAL(global)->D;
  if ( binary )
    {
      size= Q*D;
      if ( gzread ( from, STATE(states)[first].l,
		    sizeof(tLFloat)*size ) != sizeof(tLFloat)*size )
        {
          MSGERR_DESCS ( err, "the contents of the states", sym );
          return -1;
        }
    }
  else
    for ( q= first; q < end; ++q )
      {
        state= &(STATE(states)[q]);
        if ( read_state ( D, state, from, buffer, q, sym, err ) == -1 )
          return -1;
      }
  
  return 0;
  
} /* end read_states */


const tLAModelMethods tL_loglinear_methods=
  {
    alloc_states,
    realloc_states,
    tL_Bernoulli_copy_global,
    copy_states,
    tL_Bernoulli_free_global,
    free_states,
    tL_get_fea_type_real,
    tL_Bernoulli_print_global,
    print_states,
    tL_Bernoulli_read_global_meth,
    read_states,
    tL_smooth_global_empty,
    tL_smooth_states_empty,
    tL_Bernoulli_test
  };


const tLAModelMethods tL_loglinearb_methods=
  {
    alloc_states,
    realloc_states,
    tL_Bernoulli_copy_global,
    copy_states,
    tL_Bernoulli_free_global,
    free_states,
    tL_get_fea_type_pbm,
    tL_Bernoulli_print_global,
    print_states,
    tL_Bernoulli_read_global_meth,
    read_states,
    tL_smooth_global_empty,
    tL_smooth_states_empty,
    tL_Bernoulli_test
  };
