/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mixture.c - Implementació de 'mixture.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "../../intl.h"

#include "amodel.h"
#include "common.h"
#include "error.h"
#include "global.h"
#include "mem.h"
#include "mixture.h"
#include "samples.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) MIXTURE_GLOBAL(ptr)
#define STATE(ptr) MIXTURE_STATE(ptr)




/***********/
/* MÈTODES */
/***********/

static void *
alloc_states (
	      void               *global,
	      const unsigned int  Q
	      )
{
  
  mixture_state_t *ret;
  unsigned int q;
  
  
  mt ( ret= mem_alloc ( mixture_state_t, Q ) );
  for ( q= 0; q < Q; ++q )
    {
      ret[q].p= NULL;
      ret[q].members= NULL;
    }
  
  return (void *) ret;
  
} /* end alloc_states */


static void *
copy_global (
	     void *global
	     )
{
  
  mixture_global_t *new;
  
  
  mt ( new= mem_alloc ( mixture_global_t, 1 ) );
  new->type= MIXTURE_GLOBAL(global)->type;
  new->meths= MIXTURE_GLOBAL(global)->meths;
  new->global= new->meths->copy_global ( MIXTURE_GLOBAL(global)->global );
  new->max_I= MIXTURE_GLOBAL(global)->max_I;
  
  return new;
  
} /* end copy_global */


static void
copy_states (
	     void               *global,
	     void               *src_states,
	     void               *dst_states,
	     const unsigned int  src_begin,
	     const unsigned int  dst_begin,
	     const unsigned int  Q
	     )
{
  
  mixture_state_t *a, b;
  unsigned int q, i;
  
  
  for ( q= 0; q < Q; ++q )
    {
      a= &(MIXTURE_STATE(dst_states)[dst_begin+q]);
      b= MIXTURE_STATE(src_states)[src_begin+q];
      if ( a->p == NULL )
	{
	  mt ( a->p= mem_alloc ( tLProb, b.I ) );
	}
      else if ( a->I != b.I )
	{
	  mt ( a->p= mem_realloc ( tLProb, a->p, b.I ) );
	}
      for ( i= 0; i < b.I; ++i ) a->p[i]= b.p[i];
      if ( MIXTURE_GLOBAL(global)->meths->realloc_states == NULL )
	{
	  MIXTURE_GLOBAL(global)->
	    meths->free_states ( MIXTURE_GLOBAL(global)->global,
				 a->members, a->I );
	  a->members= NULL;
	}
      if ( a->members == NULL )
	a->members= MIXTURE_GLOBAL(global)->
	  meths->alloc_states ( MIXTURE_GLOBAL(global)->global, b.I );
      else if ( a->I != b.I )
	a->members= MIXTURE_GLOBAL(global)->
	  meths->realloc_states ( MIXTURE_GLOBAL(global)->global,
				  a->members, a->I, b.I );
      MIXTURE_GLOBAL(global)->
	meths->copy_states ( MIXTURE_GLOBAL(global)->global,
			     b.members, a->members, 0, 0, b.I );
      a->I= b.I;
    }
  
} /* end copy_states */


static void
free_global (
	     void *global
	     )
{
  
  if ( GLOBAL(global)->global != NULL )
    GLOBAL(global)->meths->free_global ( GLOBAL(global)->global );
  free ( global );
  
} /* end free_global */


static void
free_states (
	     void               *global,
	     void               *states,
	     const unsigned int  Q
	     )
{
  
  unsigned int q;
  mixture_state_t state;
  const tLAModelMethods *smeths;
  void *sglobal;
  
  
  smeths= GLOBAL(global)->meths;
  sglobal= GLOBAL(global)->global;
  for ( q= 0; q < Q; ++q )
    {
      state= STATE(states)[q];
      if ( state.p != NULL ) free ( state.p );
      if ( state.members != NULL )
	smeths->free_states ( sglobal,
			      state.members,
			      state.I );
    }
  free ( states );
  
} /* end free_states */


static tLFeaType
get_fea_type (
	      void *global
	      )
{
  return GLOBAL(global)->meths->get_fea_type ( GLOBAL(global)->global );
} /* end get_fea_type */


static void
print_global (
	      void         *global,
	      FILE         *to,
	      const tLBool  binary
	      )
{
  
  if ( binary ) tL_print_string_binary ( to, GLOBAL(global)->type );
  else          fprintf ( to, "%s\n", GLOBAL(global)->type );
  if ( GLOBAL(global)->global != NULL )
    GLOBAL(global)->meths->print_global ( GLOBAL(global)->global,
					  to, binary );
  
} /* end print_global */


static void
print_states (
	      void               *global,
	      FILE               *to,
	      void               *states,
	      const unsigned int  first,
	      const unsigned int  Q,
	      const tLBool        binary
	      )
{
  
  unsigned int q, i, end;
  mixture_state_t state;
  
  
  end= first+Q;
  if ( binary )
    for ( q= first; q < end; ++q )
      {
	state= STATE(states)[q];
	fwritee ( &(state.I), sizeof(state.I), 1, to );
	fwritee ( state.p, sizeof(tLProb), state.I, to );
	GLOBAL(global)->meths->print_states ( GLOBAL(global)->global,
					      to, state.members,
					      0, state.I, TRUE );
      }
  else
    for ( q= first; q < end; ++q )
      {
	
	state= STATE(states)[q];
	
	fprintf ( to, "I %u\n", state.I );
	
	fprintf ( to, "PMembers" );
	for ( i= 0; i < state.I; ++i )
	  printprobsep ( to, state.p[i] );
	fputc ( '\n', to );
	
	fprintf ( to, "Members\n" );
	GLOBAL(global)->meths->print_states ( GLOBAL(global)->global,
					      to, state.members,
					      0, state.I, FALSE );
	
      }
  
} /* end print_states */


static void *
read_global (
	     gzFile         from,
	     tLBuffer      *buffer,
	     const tLBool   binary,
	     char         **err
	     )
{
  
  mixture_global_t *global;
  amodel_type_t type;
  char *token;
  
  
  mt ( global= mem_alloc ( mixture_global_t, 1 ) );
  if ( binary )
    {
      token= tL_read_string_binary ( from );
      if ( token == NULL ) goto errorgt;
      type= tL_search_amodel_type ( token );
      if ( type.name == NULL )
	{
	  MSGERR_UT ( err, token );
	  free ( token );
	  goto error;
	}
      free ( token );
    }
  else
    {
      if ( !tL_gtoken ( from, buffer ) ) goto errorgt;
      type= tL_search_amodel_type ( buffer->v );
      if ( type.name == NULL )
	{
	  MSGERR_UT ( err, buffer->v );
	  goto error;
	}
    }
  global->type= type.name;
  global->meths= type.meths;
  if ( type.meths->read_global != NULL )
    {
      global->global=
	type.meths->read_global ( from, buffer, binary, err );
      if ( global->global == NULL ) goto error;
    }
  else global->global= NULL;
  
  global->max_I= 0;
  
  return (void *) global;
  
 errorgt:
  MSGERR_GDESC ( err, _( "the emission probability functions type" ) );
 error:
  free ( global );
  return NULL;
  
} /* end read_global */


static int
read_states (
	     void                *global,
             gzFile               from,
             tLBuffer            *buffer,
             void                *states,
	     const unsigned int   first,
             const unsigned int   Q,
             const char          *sym,
	     const tLBool         binary,
	     char               **err
	     )
{
  
  unsigned int q, I, i, end;
  mixture_state_t *state;
  tLProb *p;
  const tLAModelMethods *smeths;
  void *sglobal;
  char name[100];
  
  
  end= first+Q;
  if ( binary )
    for ( q= first; q < end; ++q )
      {
	
	state= &(STATE(states)[q]);
	
	/* I. */
	if ( gzread ( from, &I, sizeof(state->I) ) != sizeof(state->I) )
	  goto errorI;
	if ( I == 0 ) goto errorI0;
	state->I= I;
	if ( I > GLOBAL(global)->max_I ) GLOBAL(global)->max_I= I;
	
	/* PMembers. */
	mt ( p= state->p= mem_alloc ( tLProb, I ) );
	if ( gzread ( from, p, I*sizeof(tLProb) ) != I*sizeof(tLProb) )
	  {
	    MSGERR_DESC ( err, _( "the mixture coefficients" ) , q, sym );
	    return -1;
	  }
	
	/* Members. */
	snprintf ( name, 100, "(%s,%u)", sym, q+1 );
	smeths= GLOBAL(global)->meths;
	sglobal= GLOBAL(global)->global;
	state->members= smeths->alloc_states ( sglobal, I );
	if ( smeths->read_states ( sglobal, from, buffer,
				   state->members, 0, I, name,
				   TRUE, err ) == -1 )
	  return -1;
	
      }
  else
    for ( q= first; q < end; ++q )
      {
	
	state= &(STATE(states)[q]);
	
	/* I. */
	if ( !tL_gtoken ( from, buffer ) ||
	     strcmp ( "I", buffer->v ) )
	  {
	    MSGERR_TOK ( err, "I", q, sym );
	    return -1;
	  }
	if ( !tL_gtoken ( from, buffer ) ) goto errorI;
	state->I= I= atoui ( buffer->v );
	if ( I == 0 ) goto errorI0;
	if ( I > GLOBAL(global)->max_I ) GLOBAL(global)->max_I= I;

	/* PMembers. */
	if ( !tL_gtoken ( from, buffer ) ||
	     strcmp ( "PMembers", buffer->v ) )
	  {
	    MSGERR_TOK ( err, "PMembers", q, sym );
	    return -1;
	  }
	mt ( p= state->p= mem_alloc ( tLProb, I ) );
	for ( i= 0; i < I; ++i )
	  {
	    if ( !tL_gtoken ( from, buffer ) )
	      {
		tL_msgerror ( err, _( "Unable to read the %uth mixture"
			      " coefficient PMembers[%u] of the %uth state"
			      " in the symbol '%s'" ) , i+1, i, q+1, sym );
		return -1;
	      }
	    p[i]= tL_atop ( buffer->v );
	  }
	
	/* Members. */
	if ( !tL_gtoken ( from, buffer ) ||
	     strcmp ( "Members", buffer->v ) )
	  {
	    MSGERR_TOK ( err, "Members", q, sym );
	    return -1;
	  }
	snprintf ( name, 100, "(%s,%u)", sym, q+1 );
	smeths= GLOBAL(global)->meths;
	sglobal= GLOBAL(global)->global;
	state->members= smeths->alloc_states ( sglobal, I );
	if ( smeths->read_states ( sglobal, from, buffer,
				   state->members, 0, I, name,
				   FALSE, err ) == -1 )
	  return -1;
	
      }
  
  return 0;
  
 errorI:
  MSGERR_DESC ( err, _( "the number of mixture components I" ) , q, sym );
  return -1;
 errorI0:
  MSGERR_DESC0 ( err, _( "the number of mixture components I" ) , q, sym );
  return -1;
  
} /* end read_states */


static void
smooth_global (
	       void          *global,
	       const tLFloat  f
	       )
{
  
  if ( GLOBAL(global)->global != NULL )
    GLOBAL(global)->meths->smooth_global ( GLOBAL(global)->global, f );
  
} /* end smooth_global */


static void
smooth_states (
	       void                *global,
	       void                *states,
	       const unsigned int   Q,
	       const tLFloat        f
	       )
{
  
  unsigned int q, i;
  mixture_state_t state;
  tLFloat aux, aux2;
  
  
  aux= log ( 1.0 - f );
  for ( q= 0; q < Q; ++q )
    {
      
      state= STATE(states)[q];
      
      aux2= f * (1.0 / state.I);
      for ( i= 0; i < state.I; ++i )
	state.p[i]=
	  log ( exp ( aux + state.p[i] ) + aux2 );
      
      GLOBAL(global)->meths->smooth_states ( GLOBAL(global)->global,
					     state.members,
					     state.I, f );
      
    }
  
} /* end smooth_states */


static tLBool
test (
      void         *global,
      const tLFea  *fea,
      char        **err
      )
{
  return GLOBAL(global)->meths->test ( GLOBAL(global)->global, fea, err );
} /* end test */


const tLAModelMethods tL_mixture_methods=
  {
    alloc_states,
    NULL,
    copy_global,
    copy_states,
    free_global,
    free_states,
    get_fea_type,
    print_global,
    print_states,
    read_global,
    read_states,
    smooth_global,
    smooth_states,
    test
  };
