/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  amodel_itrans.c - Implementació de 'amodel_itrans.h'.
 *
 *  Adrià Giménez Pastor, 2011.
 *
 */


#include <stddef.h>
#include <stdlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "error.h"
#include "amodel_itrans.h"
#include "mem.h"




/**********/
/* MACROS */
/**********/

#define stack_clear(PTR) ((PTR)->N=0)
#define stack_free(PTR) (free ( (PTR)->v ))




/*********/
/* TIPUS */
/*********/

/* Pila per a calcular les transicions inverses. */
typedef struct
{
  
  itrans_edge_t *v;
  unsigned int   N;
  unsigned int   size;
  
} stack_ti;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
stack_add (
	   stack_ti            *stack,
	   const itrans_edge_t  edge
	   )
{
  
  if ( stack->N == stack->size )
    {
      stack->size*= 2;
      mt ( stack->v= mem_realloc ( itrans_edge_t, stack->v, stack->size ) );
    }
  stack->v[stack->N++]= edge;
  
} /* end stack_add */


static void
stack_init (
	    stack_ti *stack
	    )
{
  
  mt ( stack->v= mem_alloc ( itrans_edge_t, 1 ) );
  stack->N= 0;
  stack->size= 1;
  
} /* end stack_init */


static stack_ti *
stacks_alloc (
	      const unsigned int N
	      )
{
  
  stack_ti *new;
  unsigned int n;
  
  
  mt ( new= mem_alloc ( stack_ti, N ) );
  for ( n= 0; n < N; ++n )
    stack_init ( &(new[n]) );
  
  return new;
  
} /* end stacks_alloc */


static void
stacks_free (
	     stack_ti           *stacks,
	     const unsigned int  N
	     )
{
  
  unsigned int n;
  
  
  for ( n= 0; n < N; ++n )
    stack_free ( &(stacks[n]) );
  free ( stacks );
  
} /* end stacks_free */


static itrans_edge_t *
register_edges (
		stack_ti       *stack,
		itrans_state_t *state,
		itrans_edge_t  *edges
		)
{
  
  unsigned int n;
  
  
  state->begin= edges;
  for ( n= 0; n < stack->N; ++n )
    *(edges++)= stack->v[n];
  state->end= edges;
  state->ip= NULL;
  stack_clear ( stack );
  
  return edges;
  
} /* end register_edges */


static void
make_dist (
	   tLAModelITrans *itrans,
	   unsigned int    Q,
	   int            *queue
	   )
{
  
  int *new, *begin, *end, aux;
  unsigned int q;
  itrans_state_t state;
  const itrans_edge_t *ed;
  
  
  mt ( itrans->dist= new= mem_alloc ( int, Q ) );
  for ( q= 0; q < Q; ++q ) new[q]= -1;
  itrans->init_dist= -1;
  state= itrans->final;
  end= queue; 
  for ( ed= state.begin; ed != state.end; ++ed )
    {
      new[ed->q]= 0;
      *(end++)= ed->q;
    }
  if ( state.ip != NULL && *(state.ip) != -HUGE_VAL )
    itrans->init_dist= 0;
  for ( begin= queue; begin != end; ++begin )
    {
      q= *begin;
      aux= new[q]+1;
      state= itrans->states[q];
      for ( ed= state.begin; ed != state.end; ++ed )
	if ( new[ed->q] == -1 )
	  {
	    new[ed->q]= aux;
	    *(end++)= ed->q;
	  }
      if ( state.ip != NULL && *(state.ip) != -HUGE_VAL
	   && itrans->init_dist == -1 )
	itrans->init_dist= aux;
    }
  
} /* end make_dist */


static tLAModelITrans *
itrans_new (
	    const tLAModelHMM  *cchar,
	    stack_ti           *stacks,
	    int                *queue
	    )
{
  
  tLAModelITrans *new;
  unsigned int q, NE;
  const tLAModelTransListEdge *ed;
  const tLAModelTransListState *states;
  tLAModelTransListState state;
  itrans_edge_t auxed, *nedges;
  itrans_state_t *nstates;
  
  
  /* Registra les arestes. En Q està el final. */
  states= cchar->trans.list.states;
  NE= 0;
  for ( q= 0; q < cchar->Q; ++q )
    {
      state= states[q];
      auxed.q= q;
      for ( ed= state.begin; ed != state.end; ++ed )
	{
	  auxed.p= &(ed->p);
	  stack_add ( &(stacks[ed->q]), auxed );
	  ++NE;
	}
      if ( states[q].fp != -HUGE_VAL )
	{
	  auxed.p= &(states[q].fp);
	  stack_add ( &(stacks[cchar->Q]), auxed );
	  ++NE;
	}
    }
  
  /* Crea. */
  mt ( new= mem_alloc ( tLAModelITrans, 1 ) );
  mt ( new->states= nstates= mem_alloc ( itrans_state_t, cchar->Q ) );
  mt ( new->edges= nedges= mem_alloc ( itrans_edge_t, NE ) );
  for ( q= 0; q < cchar->Q; ++q )
    nedges= register_edges ( &(stacks[q]), &(nstates[q]), nedges );
  register_edges ( &(stacks[cchar->Q]), &(new->final), nedges );
  
  /* Registra probabilitats inicials. */
  state= cchar->trans.list.init;
  for ( ed= state.begin; ed != state.end; ++ed )
    nstates[ed->q].ip= &(ed->p);
  if ( cchar->trans.list.init.fp != -HUGE_VAL )
    new->final.ip= &(cchar->trans.list.init.fp);
  
  /* Distàncies. */
  make_dist ( new, cchar->Q, queue );
  
  return new;
  
} /* end itrans_new */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_amodel_alloc_all_itrans (
			    tLAModel *amodel
			    )
{
  
  stack_ti *stacks;
  int *queue;
  unsigned int n;
  tLAModelHMM *chars;
  
  
  if ( !(amodel->tflags&TLLIST) ) return;

  stacks= stacks_alloc ( amodel->max_Q+1 );
  queue= mem_alloc ( int, amodel->max_Q );
  chars= amodel->hmms;
  for ( n= 0; n < amodel->N; ++n )
    {
      switch ( chars[n].trans.type )
	{
	case TLLIST:
	  if ( chars[n].trans.list.itrans == NULL )
	    chars[n].trans.list.itrans=
	      itrans_new ( &(chars[n]), stacks, queue );
	  break;
	case TLLINEAL:
	case TLPOINTER:
	case TLNULLTRANS:
	  break;
	}
    }
  free ( queue );
  stacks_free ( stacks, amodel->max_Q+1 );
  
} /* end tL_amodel_alloc_all_itrans */


void
tL_amodel_free_itrans (
		       tLAModelITrans *itrans
		       )
{

  free ( itrans->states );
  free ( itrans->edges );
  free ( itrans->dist );
  free ( itrans );
  
} /* end tL_amodel_free_itrans */


int
tL_init_state_dists (
		     int                *dists,
		     const tLAModelHMM *cchar,
		     int                 offset
		     )
{
  
  const tLAModelTrans *trans;
  unsigned int q;
  const tLAModelITrans *itrans;
  int i;
  
  
  i= cchar->Q;
  trans= &(cchar->trans);
  if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
  switch ( trans->type )
    {
    case TLLINEAL:
      for ( q= 0; q < cchar->Q; ++q )
	dists[--i]= offset++;
      break;
    case TLLIST:
      itrans= trans->list.itrans;
      q= cchar->Q;
      do {
	dists[--i]= offset + itrans->dist[--q];
      } while ( q != 0 );
      offset+= itrans->init_dist;
      break;
    default:
      tL_error ( _( "%s - tL_init_state_dists: This should not happen" ),
		 __FILE__ );
    }
  
  return offset;
  
} /* end tL_init_state_dists */

