/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  amodel_public.h - Part pública de 'amodel.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */

#ifndef __LIBTL_AMODEL_H__
#define __LIBTL_AMODEL_H__

#include <stddef.h>
#include <stdio.h>
#include <zlib.h>

#include "port.h"
#include "common_public.h"
#include "samples_public.h"
#include "text_public.h"

/** \defgroup Models Models
 */
/** \addtogroup Models
 *  @{
 */

/** \defgroup AModel Acoustic model (HMMs)
 */
/** \addtogroup AModel
 *  @{
 */

/* DEFINICIÓ */

/** Type of HMM transitions. */
typedef enum
  {
    TLNULLTRANS,
    TLLINEAL,    /**< Lineal topology. Only transitions to the next
		    state and current state are allowed. */
    TLLIST,      /**< Open topology. Transitions for each state are
		    stored as a list. */
    TLPOINTER    /**< Reference to previous HMM. The transition matrix
		    of the referenced HMM is used. */
  } tLAModelTransType;

typedef union _tLAModelTrans tLAModelTrans;

/** Transitions in type LIST. This structure is used to store the
 *  information of a transition to a conventional state when type LIST
 *  is used.
 */
typedef struct
{
  
  unsigned int q;    /**< Target state. */
  tLProb       p;    /**< Log-Probability. */
  
} tLAModelTransListEdge;

/** Outgoing transitions for type LIST. This structure stores the
 *  outgoing transitions of a state of an HMM using transitions of
 *  type LIST. Transitions to conventional states are stored
 *  consecutively in memory and represented as edges of a finite state
 *  automata.
 */
typedef struct
{
  
  tLAModelTransListEdge *begin;    /**< Pointer to the first
				      transition. */
  tLAModelTransListEdge *end;      /**< Pointer to the transition
				      following the last
				      transition. */
  tLProb                 fp;       /**< Probability of the transition
				      to the final state. */
  
} tLAModelTransListState;

typedef struct tLAModelITrans tLAModelITrans;

/** HMM transitions. Union used to store the transition matrix of an
 *  HMM.
 */
union _tLAModelTrans
{
  
  tLAModelTransType type;    /**< Type of transitions. */
  struct
  {
    tLAModelTransType  type;
    tLProbPair        *p;       /**< Array of transitions. For each
				   state Q, p[Q] stores the outgoing
				   transitions of Q, where P0 is the
				   transition to the current state,
				   and P1 to the next state. */
    tLBool             ll;      /* Indica si els paràmetres són
				   log-lineals. */
  }                   lineal;    /**< Data for LINEAL transitions. */
  struct
  {
    tLAModelTransType       type;
    tLAModelTransListState  init;      /**< Outgoing transitions of
					  the initial state. */
    tLAModelTransListState *states;    /**< Array of transitions. For
					  each conventional state Q,
					  states[q] stores the
					  outgoing transitions of
					  Q. */
    tLAModelTransListEdge  *edges;     /* Memòria de les
					  arestes. Primer apareixen
					  les arestes de init, després
					  les del primer estat,
					  etc. */
    unsigned long           NE;        /* Número d'arestes. */
    tLAModelITrans         *itrans;    /* Transicions inverses i
					  distància dels estast. */
  }                 list;    /**< Data for LIST transitions. */
  struct
  {
    tLAModelTransType  type;
    unsigned int       c;      /* Caràcter al qual apunta. */
    tLAModelTrans     *ptr;    /**< Pointer to the transition
				  matrix. */
  }                 pointer;    /**< Data for POINTER transitions. */
  
};

/** HMM. This structure is used to store a Hidden Markov Model
 *  (HMM). An HMM is a finite state model with emission probability
 *  functions into the states.
 */
typedef struct
{
  
  unsigned int   Q;         /**< Number of state. */
  tLAModelTrans  trans;     /**< State transition matrix. */
  void          *states;    /* |Q|. Estats. */
  
} tLAModelHMM;

typedef struct
{
  
  /* Reserva espai per a Q estats. */
  void * (*alloc_states) (void               *global,
			  const unsigned int  Q);
  
  /* Reserva més espai respetant el contingut actual. Es supossa
     newQ>=oldQ. Si hi havia més capacitat que oldQ, tota la
     informació extra es pert. Pot ser NULL. */
  void * (*realloc_states) (void               *global,
			    void               *states,
			    const unsigned int  oldQ,
			    const unsigned int  newQ);
  
  /* Copia la capçalera. */
  void * (*copy_global) (void *global);
  
  /* Copia el contingut del estat origen en el destí. */
  void   (*copy_states) (void               *global,
			 void               *src_states,
			 void               *dst_states,
			 const unsigned int  src_begin,
			 const unsigned int  dst_begin,
			 const unsigned int  Q);
  
  /* Allibera el model. */
  void   (*free_global) (void *global);

  /* Allibera Q estats. */
  void   (*free_states) (void               *global,
                         void               *states,
			 const unsigned int  Q);
  
  /* Torna el tipus de mostres que accepta. */
  tLFeaType (*get_fea_type) (void *global);
  
  /* Imprimeix capçalera pròpia del tipus. */
  void   (*print_global) (void         *global,
			  FILE         *to,
			  const tLBool  binary);
  
  /* Imprimeix Q estats començant per first. */
  void   (*print_states) (void               *global,
			  FILE               *to,
			  void               *states,
			  const unsigned int  first,
			  const unsigned int  Q,
			  const tLBool        binary);
  
  /* LLig la part de la capçalera particular
     d'aquest tipus. Es a dir els atributs globals
     a tots els estats. */
  void * (*read_global) (gzFile         from,
			 tLBuffer      *buffer,
			 const tLBool   binary,
			 char         **err);

  /* Llig les funcions de Q estats en els estats del caràcter
     especificat a partir de first, C és el nom del caracter
     actual. Torna -1 en cas d'error, 0 si tot ha anat bé. */
  int   (*read_states) (void                *global,
			gzFile               from,
			tLBuffer            *buffer,
			void                *states,
			const unsigned int   first,
			const unsigned int   Q,
			const char          *sym,
			const tLBool         binary,
			char               **err);
  
  /* Suavitza els paràmetres globals. F és el paràmetre per al
     suavitzat lineal. */
  void   (*smooth_global) (void          *global,
			   const tLFloat  f);
  
  /* Suavitza linialment Q estats amb un factor F. */
  void   (*smooth_states) (void               *global,
			   void               *states,
			   const unsigned int  Q,
			   const tLFloat       f);
  
  /* Mira si la mostra compleix unes condicions mínimes
     per a ser acceptada pel model. */
  tLBool (*test) (void         *global,
		  const tLFea  *fea,
		  char        **msg);
  
} tLAModelMethods;

/** Acoustic model. This structure is used to store a collection of
 *  HMMs of the same type and global characteristics. This
 *  structure must not be modified by hand, public fields are only
 *  provided for read access.
 */
typedef struct
{
  
  const char            *type;        /**< String identifying the type
					 of the emission probability
					 functions used in the
					 HMMs. */
  unsigned int           N;           /**< Number of HMMs. */
  tLAModelHMM           *hmms;       /**< Array with the HMMs. */
  void                  *global;      /* Atributs globals. */
  const tLAModelMethods *meths;       /* Mètodes. */
  size_t                 total_Q;     /* Número total d'estats. */
  unsigned int           max_Q;       /* Nombre màxim d'estats per
					 caràcter. */
  char                   tflags;      /* Indiquen què tipus de
					 transicions hi han en el
					 model. */
  tLBool                 IF_jumps;    /* A cert si exiteixen bots
					 I-F. */
  
} tLAModel;


/* FUNCIONS */

/** Copies an acoustic model. This function create a new copy of the
 *  provided acoustic model.
 *
 *  @param amodel The acoustic model to copy.
 *  @return The acoustic model copy.
 */
TL_API_FUNC(tLAModel *)
tL_amodel_copy (
		const tLAModel *amodel
		);

/* Es supossa que està tot el model inicialitzat. */
/** Frees memory. Frees the memory allocated for the acoustic model.
 *
 *  @param amodel The acoustic model.
 */
TL_API_FUNC(void)
tL_amodel_free (
		tLAModel *amodel
		);

/** Gets the feature type. This macro returns the type of the feature
 *  vectors supported by the given acoustic model.
 *
 *  @param AMODEL The acoustic model.
 *  @return The type of the supported feature vectors.
 */
#define tL_amodel_get_fea_type(AMODEL)			\
  (AMODEL)->meths->get_fea_type ( (AMODEL)->global )

/** Creates a new set of mixture HMMs. This function creates
 *  a new equivalent mixture version of the set of HMMs in
 *  the provided acoustic model. That is, each probability emission
 *  function T is replaced by a one component T mixture probability
 *  function.
 *
 *  @param amodel The acoustic model.
 *  @return A new equivalent mixture version of the provided HMMs.
 */
TL_API_FUNC(tLAModel *)
tL_amodel_mixture_new (
		       const tLAModel *amodel
		       );

/* Llig el model d'un fitxer, guardant en memòria els paràmetres
 * segons el tipus indicat. Se li passa un diccionari buit, que
 * plenarà amb els caràcters que llegisca.
 */
/** Reads an acoustic model from a file. This functions creates a new
 *  acoustic model from a text description or binary
 *  representation. The symbols (phonems) related to the HMMs are
 *  registered into the given dictionary, which must be initially
 *  empty.
 *
 *  @param from File where the text description or binary
 *  representation is stored.
 *  @param syms Empty dictionary in which symbols (phonems) are registered.
 *  @param err Pointer to string variable. If not NULL an error
 *  message is allocated in the variable in case of error.
 *  @return The acoustic model, or NULL in case of error.
 */
TL_API_FUNC(tLAModel *)
tL_amodel_new_from_file (
			 gzFile   from,
			 tLDict  *syms,
			 char   **err
			 );

/* Imprimeix el model en el fitxer indicat.  Es supossa que al
 * diccionari està el vocabulari de caràcters amb el que es va crear
 * el model. Si BINARY és cert, el model es guarda en un format binari
 * totalment depenent de la màquina. La ventatja és que els paràmetres
 * estan representats com en memòria.
 */
/** Prints the acoustic model. This function writes in the given
 *  file the content of the acoustic model. Each HMM is
 *  labeled according to the provided dictionary. If binary is FALSE
 *  the content is written in text format, else it is written in
 *  binary format. The binary format is not a compact nor portable
 *  representation, but a way to ensure that parameters are written as
 *  they are stored in memory.
 *
 *  @param amodel The acoustic models.
 *  @param to File where the models are written.
 *  @param syms A dictionary with the labels of the HMMs.
 *  @param binary If false the models are written in text format, else
 *  in binary format.
 */
TL_API_FUNC(void)
tL_amodel_print (
		 const tLAModel *amodel,
		 FILE           *to,
		 const tLDict   *syms,
		 const tLBool    binary
		 );

/* Suavitxat linial. */
/** Smooth parameters of acoustic model. This function is a common
 *  interface to smooth the parameters of HMMs regardless of the
 *  emission probability functions type. Diagonal Gaussian
 *  distributions are not affected by this function, instead they are
 *  smoothed using the SMOOTH values in the model description during
 *  parameter updating, the variances are floored to the SMOOTH
 *  values. Probability parameters are linearly interpolated with the
 *  uniform distribution using F.
 *
 *  @param amodel The acoustic model.
 *  @param f Weight used to perform a lineal interpolation with the
 *  uniform distribution (p*(1-f) + 0.5*f).
 *  @param smooth_trans If TRUE the transition probabilities are
 *  smoothed.
 */
TL_API_FUNC(void)
tL_amodel_smooth (
		  tLAModel      *amodel,
		  const tLFloat  f,
		  const tLBool   smooth_trans
		  );

/** Tests sample compatibility. This macro tests if the given feature
 *  vector sequence is compatible with the provided acoustic model.
 *
 *  @param AMODEL The acoustic model.
 *  @param FEA A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the AMODEL.
 *  @param ERR Pointer to string variable. If not NULL an error message
 *  is allocated in the variable in case of error.
 *  @return TRUE if the sample is compatible, FALSE otherwise.
 */
#define tL_amodel_test(AMODEL,FEA,ERR)				\
  (AMODEL)->meths->test ( (AMODEL)->global, (FEA), (ERR) )

/** Returns the names of the tied states. This function returns the
 *  dictionary containing the names of the tied states.
 *
 *  @param amodel The acoustic model.
 *  @return The dictionary containing the names, or NULL if HMM
 *  models are not of type TiedStates.
 */
TL_API_FUNC(const tLDict *)
tL_amodel_tiedstates_get_names (
				const tLAModel *amodel
				);

/** Creates a new set of tiedstates mixture HMM models. Given a
 *  TiedStates HMM model set, this function create an equivalent
 *  tiedstate version in which each probability emission T is replaced
 *  by a one component T mixture probability function.
 *
 *  @param amodel The TiedStates HMM models.
 *  @return A new equivalent mixture version of the provided
 *  TiedStates HMM models.
 */
TL_API_FUNC(tLAModel *)
tL_amodel_tiedstates_mixture_new (
				  const tLAModel *amodel
				  );

/** @} */
/** @} */

#endif /* __LIBTL_AMODEL_H__ */
