/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  arpa.c - Implementació de 'arpa.h'.
 *
 *  Adrià Giménez Pastor, 2009,2011.
 *
 */


#include <ctype.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "arpa.h"
#include "common.h"
#include "error.h"
#include "hash.h"
#include "mem.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define BARPASYM ((unsigned long) -2)
#define EARPASYM ((unsigned long) -1)

#define MEMORY_BLOCK 16777216  /* instead of duplicating memory with realloc, add a fixed size block ~ 768MB */


/*********/
/* TIPUS */
/*********/

/* Node per a la taula hash amb les històries. */
struct arpa_hist
{
  
  unsigned long next;      /* Identificador del següent node. */
  unsigned long p;         /* Història prèvia. */
  unsigned long w;         /* Última paraula de la història. */
  int           length;    /* Llargària. */
  unsigned long state;     /* Identificador de l'estat associat amb
			      aquesta història. 0 indica que no en
			      té. */
  tLProb        bowt;      /* El format permet que una història sense
			      estat associat tinga pes de Backoff
			      (Backoff 1), però en principi açò no té
			      sentit. */
  
};


/* Nodes per a la taula hash amb els estats. */
struct arpa_state
{
  
  unsigned long hist;         /* Història associada a l'estat. */
  unsigned long edges;        /* Llista d'arestes
				 eixints. Concretament és
				 l'identificador de la primera
				 aresta. */
  unsigned long null_edge;    /* Aresta nul·la. */
  tLBool        active;       /* Indica que l'estat està actiu. */
  tLProb        fp;           /* Probabilitat final. -INF és que no en
				 té. */
  unsigned long bos;          /* Estat de backoff. */
  
};


/* Nodes per a les arestes. Per als estats no nuls, l'estat 1 i totes
 * les arestes nul·les s'utilitza el camp 'edges', en els demés casos
 * el camp 'block'.
 */
struct arpa_edge
{
  
  unsigned long next;    /* Identificador de la següent aresta en la
			    llista del estat. */
  unsigned long from;    /* Estat origen. */
  unsigned long to;      /* Estat destí. Pot ser 0. */
  unsigned long w;       /* Paraula amb que transita. */
  tLProb        p;       /* Probabilitat de transició. */
  
};


/* Estructura auxiliar per a llegir un model de llenguatge en format
   ARPA. */
struct arpa
{
  
  unsigned int       N;    /* Ordre dels N-grames. */
  
  /* Pot donar-se el cas que amb una transició no nul·la es transite a
     l'estat 1, l'estat sense història. És l'únic estat que no té una
     paraula asociada. El meu LM necessita que els estats no nuls
     tinguen sempre una praula associada, la solució necessària
     consisteix en copiar l'estat 1 tantes vegades com paraules hi
     hasquen. */
  unsigned long      begin_copies1;    /* Primera copia de l'estat 1,
					  es troba al final dels
					  estats. */
  
  /* Històries. */
  unsigned long      hist_hsize;
  unsigned long     *hist_v;
  unsigned long      hist_msize;
  struct arpa_hist  *hist_mem;
  unsigned long      hist_N;
  
  /* Quan la estructura estiga totalment completada l'idea es que
     estaran ací tots els estats: primer els normals, després les
     còpies del primer i finalment els nuls. */
  /* Estats. */
  unsigned long      state_size;
  struct arpa_state *state_v;
  unsigned long      state_N;
  
  /* Arestes. */
  unsigned long      ed_size;
  struct arpa_edge  *ed_v;
  unsigned long      ed_N;
  
};




/*************/
/* CONSTANTS */
/*************/

static const unsigned long sizeul= sizeof(unsigned long)*2;
static const unsigned long sizeul2= sizeof(unsigned long)*3;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* La història nul·la se inserta al crear la taula. No s'inicialitza
   la secció de buscar estats nuls. */
static struct arpa *
create_arpa (
	     unsigned long  nhists,
	     size_t        *entries,
	     unsigned int   N
	     )
{
  
  struct arpa *new;
  unsigned long n, total_hists;
  
  
  total_hists = entries[0];
  for ( n= 0; n < N; ++n )
    total_hists += entries[n];
  
  mt ( new= mem_alloc ( struct arpa, 1 ) );
  
  /* Històries. La història nul·la és especial i sempre és la 0. A més
     serveix de sentinela. */
  new->hist_hsize= tL_hash_primesize ( nhists );
  mt ( new->hist_v= mem_alloc ( unsigned long, new->hist_hsize ) );
  for ( n= 0; n < new->hist_hsize; ++n )
    new->hist_v[n]= 0; /* La hipòtesi 0 està en totes les entrades. */
  new->hist_msize= (total_hists >> 2);  /* expected backoff ngrams */
  mt ( new->hist_mem= mem_alloc ( struct arpa_hist, (total_hists >> 2) ) );
  new->hist_N= 1;
  new->hist_mem->next= 0;
  new->hist_mem->p= 0;
  new->hist_mem->w= 0;
  new->hist_mem->length= 0;
  new->hist_mem->state= 0;
  new->hist_mem->bowt= 0.0;
  
  /* Estats. L'estat 0 és sentinela i l'estat 1 és l'estat amb
     història nul·la. */
  /* rule of thumb for state occupation expected backoff ngrams plus
     the unigrams */
  new->state_size= (total_hists >> 2) + entries[0]; 
  mt ( new->state_v= mem_alloc ( struct arpa_state,
				 (total_hists >> 2) + entries[0] ) );
  new->state_N= 2;
  new->state_v->hist= 0;
  new->state_v->edges= 0;
  new->state_v->null_edge= 0;
  new->state_v->active= FALSE;
  new->state_v->fp= -HUGE_VAL;
  new->state_v->bos= 0;
  new->state_v[1]= new->state_v[0];
  
  /* Arestes. L'aresta 0 és sentinela. */
  new->ed_size= total_hists;
  /* there are less as ending arcs are not necessary */
  mt ( new->ed_v= mem_alloc ( struct arpa_edge, total_hists ) ); 
  new->ed_N= 1;
  new->ed_v->next= 0;
  new->ed_v->from= 0;
  new->ed_v->to= 0;
  new->ed_v->w= 0;
  new->ed_v->p= -HUGE_VAL;
  
  return new;
  
} /* end create_arpa */


static void
free_arpa (
	   struct arpa *arpa
	   )
{
  
  free ( arpa->ed_v );
  
  free ( arpa->state_v );
  
  free ( arpa->hist_v );
  free ( arpa->hist_mem );
  
  free ( arpa );
  
} /* end free_arpa */


/* Se li passa el nombre de entrades per a cada ordre del N-grama. Per
   a hists se té en compter la història nul·la. */
static void
calc_arpa_sizes (
		 const size_t  *nentries,
		 size_t         N,
		 unsigned long *nhists
		 )
{
  
  size_t n;
  
  
  /* Per a les històries no interessa l'últim símbol. Com segur que es
     repeteixen moltes, rebaixe a la baixa. */
  *nhists= 1;
  for ( n= 1; n <= N; ++n )
    *nhists+= nentries[n-1]*n;
  if ( *nhists > 4 ) *nhists>>= 2;
  
} /* end calc_arpa_sizes */


/* Inserta una història a partir de la història prèvia, se li passa
   també el bowt. Si ja està torna l'identificador i sols canvia el
   bow si el que se li passa es distint de -inf. */
static unsigned long
add_hist (
	  struct arpa   *arpa,
	  unsigned long  phist,
	  unsigned long  w,
	  tLProb         bowt
	  )
{
  
  unsigned long aux[2], p, q;
  unsigned long *hist_v;
  struct arpa_hist *nodes;
  
  
  hist_v= arpa->hist_v;
  nodes= arpa->hist_mem;
  aux[0]= phist; aux[1]= w;
  p= tL_hash_jc ( (void *) &(aux[0]), sizeul, 23312819 )%arpa->hist_hsize;
  for ( q= hist_v[p];
	q != 0 && (nodes[q].p != phist || nodes[q].w != w);
	q= nodes[q].next );
  if ( q == 0 )
    {
      if ( arpa->hist_msize == arpa->hist_N )
	{
    if ( arpa->hist_msize < MEMORY_BLOCK )
      arpa->hist_msize = (arpa->hist_msize << 1);
    else
  	  arpa->hist_msize += MEMORY_BLOCK;
	  mt ( arpa->hist_mem= nodes=
	       mem_realloc ( struct arpa_hist,
			     nodes, arpa->hist_msize ) );
	}
      q= (arpa->hist_N)++;
      nodes[q].next= hist_v[p];
      nodes[q].p= phist;
      nodes[q].w= w;
      nodes[q].length= nodes[phist].length+1;
      nodes[q].state= 0;
      nodes[q].bowt= bowt;
      hist_v[p]= q;
    }
  else if ( bowt != 0.0 ) nodes[q].bowt= bowt;

  return q;

} /* end add_hist */


/* Inserta una història completa, se li passa també el bowt. Si ja
   està torna l'identificador i sols canvia el bow si el que se li
   passa es distint de 0.0. */
static unsigned long
add_full_hist (
	       struct arpa         *arpa,
	       const unsigned long *hist,
	       unsigned             length,
	       tLProb               bowt
	       )
{
  
  unsigned long id, n;
  
  
  if ( length == 0 ) return 0;
  --length;
  for ( id= 0, n= 0; n < length; ++n )
    id= add_hist ( arpa, id, hist[n], 0.0 );
  id= add_hist ( arpa, id, hist[n], bowt );
  
  return id;
  
} /* end add_full_hist */


static unsigned long
search_hist (
	     const struct arpa *arpa,
	     unsigned long      phist,
	     unsigned long      w
	     )
{
  
  unsigned long aux[2], p, q;
  const unsigned long *hist_v;
  const struct arpa_hist *nodes;
  
  
  hist_v= arpa->hist_v;
  nodes= arpa->hist_mem;
  aux[0]= phist; aux[1]= w;
  p= tL_hash_jc ( (void *) &(aux[0]), sizeul, 23312819 )%arpa->hist_hsize;
  for ( q= hist_v[p];
	q != 0 && (nodes[q].p != phist || nodes[q].w != w);
	q= nodes[q].next );
  
  return q;
  
} /* end search_hist */


/* Torna l'estat associat a una història si existeix o 0 si no
   exiteix. També torna el bowt d'eixa història. */
static unsigned long
search_state_hist (
		   const struct arpa   *arpa,
		   const unsigned long *hist,
		   int                  length,
		   tLProb              *bowt
		   )
{
  
  unsigned long ret;
  int n;
  const struct arpa_hist *nodes;
  
  
  if ( length == 0 ) return 1;
  nodes= arpa->hist_mem;
  for ( n= 0, ret= 0; n < length; ++n )
    {
      ret= search_hist ( arpa, ret, hist[n] );
      if ( ret == 0 )
	{
	  *bowt= 0.0;
	  return 0;
	}
    }
  *bowt= nodes[ret].bowt;
  
  return nodes[ret].state;
  
} /* end search_state_hist */


/* Torna l'identificador del nou estat. */
static unsigned long
add_empty_state (
		 struct arpa   *arpa,
		 unsigned long  hist
		 )
{
  
  struct arpa_state *states;
  unsigned long nstate;
  
  
  states= arpa->state_v;
  if ( arpa->state_size == arpa->state_N )
    {

      if ( arpa->state_size < MEMORY_BLOCK )
        arpa->state_size = (arpa->state_size << 1);
      else
        arpa->state_size += MEMORY_BLOCK;
    mt ( states= arpa->state_v=
       mem_realloc ( struct arpa_state,
			 arpa->state_v,
			 arpa->state_size ) );
    }
  nstate= (arpa->state_N)++;
  states[nstate].hist= hist;
  states[nstate].edges= 0;
  states[nstate].null_edge= 0;
  states[nstate].active= FALSE;
  states[nstate].fp= -HUGE_VAL;
  
  return nstate;
  
} /* end add_empty_state */


/* Inserta la hipòtesi i torna l'idenficador de l'estat associat,
   creant-lo si és necessari. Si ja estaba simplement torna
   l'identificador. */
static unsigned long
add_state_hist (
		struct arpa         *arpa,
		const unsigned long *hist,
		unsigned             length
		)
{
  
  unsigned long id, state;
  
  
  if ( length == 0 ) return 1;
  id= add_full_hist ( arpa, hist, length, 0.0 );
  if ( (state= arpa->hist_mem[id].state) == 0 )
    state= arpa->hist_mem[id].state=
      add_empty_state ( arpa, id );
  
  return state;
  
} /* end add_state_hist */


static unsigned long
insert_empty_edge (
		   struct arpa *arpa
		   )
{
  
  struct arpa_edge *edges;
  
  
  edges= arpa->ed_v;
  if ( arpa->ed_size == arpa->ed_N )
    {
      (arpa->ed_size) += MEMORY_BLOCK;
      mt ( arpa->ed_v= edges=
	   mem_realloc ( struct arpa_edge,
			 edges, arpa->ed_size ) );
    }
  return (arpa->ed_N)++;
  
} /* end insert_empty_edge */


/* FROM ha d'existir. Torna l'indentificador de la nova aresta. */
static unsigned long
insert_edge (
	     struct arpa   *arpa,
	     unsigned long  from,
	     unsigned long  to,
	     unsigned long  w,
	     tLProb         prob
	     )
{
  
  unsigned long ret;
  struct arpa_edge *edges;
  
  
  ret= insert_empty_edge ( arpa );
  edges= arpa->ed_v;
  edges[ret].to= to;
  edges[ret].from= from;
  edges[ret].w= w;
  edges[ret].p= prob;
  
  return ret;
  
} /* end insert_edge */


/* Al final dels estats copia l'estat 1 tantes vegades com
   paraules. */
static void
make_copies1 (
	      struct arpa  *arpa,
	      const tLDict *vocab
	      )
{
  
  unsigned long newsize, n, state;
  struct arpa_state *p;
  
  
  newsize= vocab->N + arpa->state_N;
  if ( newsize > arpa->state_size )
    {
      mt ( arpa->state_v=
	   mem_realloc ( struct arpa_state,
			 arpa->state_v, newsize ) );
      arpa->state_size= newsize;
    }
  arpa->begin_copies1= arpa->state_N;
  p= arpa->state_v+arpa->state_N;
  for ( n= 0, state= arpa->begin_copies1;
	n < vocab->N; ++n, ++p, ++state )
    {
      p->hist= 0;
      p->edges= 0;
      p->null_edge= insert_edge ( arpa, state, 1, 0, 0.0 );
      p->active= FALSE;
      p->fp= -HUGE_VAL;
    }
  arpa->state_N= newsize;
  
} /* end make_copies1 */


/* Llig la capçalera. L'ordre del model s'assigna a N, i torna un
   vector amb el nombre de entrades per a cada ordre. Buffer
   l'utilitza per a llegir, si tot ha anat bé en el buffer estarà
   "\1-grams:". Torna NULL en cas d'error. */
static size_t *
load_arpa_read_head (
		     gzFile         from,
		     tLBuffer      *buffer,
		     unsigned int  *N,
		     char         **err
		     )
{
  
  int ok;
  unsigned int N_, aux;
  size_t *ret, size;
  char *t1, *t2;
  
  
  /* Ignora comentaris. */
  while ( (ok= tL_gline ( from, buffer )) &&
	  strcmp ( buffer->v, "\\data\\" ) );
  if ( !ok )
    {
      tL_msgerror ( err, _( "Unable to read '\\data\\'" ) );
      return NULL;
    }
  
  /* Inicialitza. */
  N_= 0;
  mt ( ret= mem_alloc ( size_t, 1 ) );
  size= 1;
  
  /* LLig valors. */
  while ( (ok= tL_gline ( from, buffer )) &&
	  buffer->v[0] != '\\' )
    {
      if ( buffer->v[0] == '\0' ) continue;
      if ( strncmp ( buffer->v, "ngram", 5 ) ) goto errorformat;
      for ( t1= buffer->v+5; isspace ( *t1 ); ++t1 );
      if ( !*t1 ) goto errorformat;
      for ( t2= t1; *t2 && *t2 != '='; ++t2 );
      if ( !*t2 || !*(t2+1) ) goto errorformat;
      *t2= '\0'; ++t2;
      aux= atoui ( t1 );
      if ( aux != (N_+1) )
	{
	  tL_msgerror ( err, _( "read '%s' while expecting 'ngram %u=..." ) ,
			buffer->v, (N_+1) );
	  goto error;
	}
      if ( N_ == size )
	{
	  size<<= 1;
	  mt ( ret= mem_realloc ( size_t, ret, size ) );
	}
      ret[N_++]= atos ( t2 );
    }
  if ( N_ == 0 )
    {
      tL_msgerror ( err,
		    _( "Unable to read the ARPA header, or empty header" ) );
      goto error;
    }
  if ( !ok )
    {
      tL_msgerror ( err, _( "Unable to read '\\1-grams:'" ) );
      goto error;
    }
  
  if ( N_ != size )
    mt ( ret= mem_realloc ( size_t, ret, N_ ) );
  *N= N_;
  
  return ret;
  
 errorformat:
  tL_msgerror ( err, _( "Wrong format in: '%s'. Expected 'ngram"
		" <order>=<num grams>'" ) , buffer->v );
 error:
  free ( ret );
  return NULL;
  
} /* end load_arpa_read_head */


/* Fica la història completa de ID en HIST. Té que hih aure prou espai
   en HIST. Torna la llargària de la història. Es supossa que
   existeix. */
static int
get_full_hist (
	       const struct arpa *arpa,
	       unsigned long     *hist,
	       unsigned long      id
	       )
{
  
  const struct arpa_hist *nodes;
  int i, length;
  
  
  if ( id == 0 ) return 0;
  nodes= arpa->hist_mem;
  length= nodes[id].length;
  for ( i= length-1; i >= 0; --i )
    {
      hist[i]= nodes[id].w;
      id= nodes[id].p;
    }
  
  return length;
  
} /* end get_full_hist */


/* Torna -1 en cas d'error, 0 en cas contrari. */
static int
load_arpa_get_word (
		    tLDict        *vocab,
		    const tLBool   empty,
		    const char    *word,
		    size_t        *id,
		    char         **err
		    )
{
  
  if ( empty ) tL_dict_insert ( vocab, word, id );
  else
    {
      *id= tL_dict_find ( vocab, word );
      if ( *id == vocab->N )
	{
	  tL_msgerror ( err, _( "%s is out of vocabulary" ) ,	word );
	  return -1;
	}
    }
  
  return 0;
  
} /* end load_arpa_get_word */




/* Carrega un model de llenguatge en format ARPA en una estructura
   'arpa'. Torna NULL en cas d'error. */
static struct arpa *
load_arpa (
	   gzFile         from,
	   tLDict        *vocab,
	   const tLBool   empty,
	   const char    *begin_sym,
	   const char    *end_sym,
	   const char    *ninf,
	   char         **err
	   )
{
  
  tLBuffer *buffer;
  size_t *nentries, n2;
  unsigned int N, n, k;
  unsigned long nhists, *hist, word, state, ed;
  char *begin, *end;
  struct arpa *arpa;
  tLProb p, bowt;
  size_t aux;
  int last, ok;
  
  
  nentries= NULL;
  hist= NULL;
  arpa= NULL;
  
  buffer= tL_buffer_new ();
  nentries= load_arpa_read_head ( from, buffer, &N, err );
  if ( nentries == NULL ) goto error;
  mt ( hist= mem_alloc ( unsigned long, N ) );
  calc_arpa_sizes ( nentries, N, &nhists );
  arpa= create_arpa ( nhists, nentries, N );
  for ( n= 0; n < N; ++n )
    {
      for ( n2= 0; n2 < nentries[n]; ++n2 )
	{
	  
	  /* Llig una entrada. */
	  if ( !tL_gline ( from, buffer ) )
	    {
	      tL_msgerror ( err, _( "Unable to read the %luth %u-gram" ) ,
			    _PS(n2)+1, n+1 );
	      goto error;
	    }
	  if ( tL_gtokenstr ( buffer->v, &begin, &end ) ||
	       begin == NULL )
	    {
	      tL_msgerror ( err, _( "Empty line or too few fields in the"
			    " %luth %u-gram" ) , _PS(n2)+1, n+1 );
	      goto error;
	    }
	  *(end++)= '\0';
	  p= !strcmp ( begin, ninf ) ?
	    -HUGE_VAL : 
	    (tLProb) log ( pow ( 10.0, atof ( begin ) ) );
	  for ( k= 0; k < n; ++k )
	    {
	      if ( tL_gtokenstr ( end, &begin, &end ) ||
		   begin == NULL ) goto errorword;
	      *(end++)= '\0';
	      if ( begin_sym != NULL && !strcmp ( begin_sym, begin ) )
		{
		  if ( k == 0 )
		    hist[k]= BARPASYM;
		  else goto errorinitial;
		}
	      else if ( !strcmp ( end_sym, begin ) )
		{
		  tL_msgerror ( err, _( "Special final word %s appears"
				" in an invalid position in the %luth" 
				" %u-gram" ) , end_sym, _PS(n2)+1, n+1 );
		  goto error;
		}
	      else
		{
		  if ( load_arpa_get_word ( vocab, empty, begin,
					    &aux, err ) == -1 )
		    goto error;
		  hist[k]= (unsigned long) aux;
		}
	    }
	  bowt= 0.0;
	  if ( !(last= tL_gtokenstr ( end, &begin, &end )) )
	    *(end++)= '\0';
	  if ( begin == NULL ) goto errorword;
	  if ( !strcmp ( begin, end_sym ) )
	    word= EARPASYM;
	  else if ( begin_sym != NULL && !strcmp ( begin, begin_sym ) )
	    {
	      if ( n == 0 )
		word= BARPASYM;
	      else
		{
		  tL_msgerror ( err, _( "Special initial word %s appears"
				" in the last position in the %luth"
				" %u-gram" ) , begin_sym, _PS(n2)+1, n+1 );
		  goto error;
		}
	    }
	  else
	    {
	      if ( load_arpa_get_word ( vocab, empty, begin, &aux, err ) == -1 )
		goto error;
	      word= (unsigned long) aux;
	    }
	  if ( !last )
	    {
	      tL_gtokenstr ( end, &begin, &end );
	      if ( begin != NULL )
		bowt= (tLProb) log ( pow ( 10.0, atof ( begin ) ) );
	    }
	  
	  /* Inserta en l'estructura. */
	  state= add_state_hist ( arpa, &(hist[0]), n );
    if ( n+1 < N && bowt != 0.0 ) /* history of the last gram (no back-off) is not necessary  
                                      or if backoff is not present */
      add_hist ( arpa, arpa->state_v[state].hist, word, bowt );
	  if ( p != -HUGE_VAL )
	    {
	      if ( word == EARPASYM )
		arpa->state_v[state].fp= p;
	      else if ( word == BARPASYM ) goto errorinitial;
	      else
		{
		  ed= insert_edge ( arpa, state, 0, word, p );
		  arpa->ed_v[ed].next= arpa->state_v[state].edges;
		  arpa->state_v[state].edges= ed;
		}
	    }
	  
	}
      while ( (ok= tL_gline ( from, buffer )) && buffer->v[0] != '\\' );
      if ( !ok )
	{
	  tL_msgerror ( err, _( "Unable to read the next section"
			" '\\%u-grams:' or '\\end\\'" ) , n+2 );
	  goto error;
	}
    }
  
  if ( strncmp ( buffer->v, "\\end\\", 5 ) )
    {
      tL_msgerror ( err, _( "Unable to read '\\end\\'" ) );
      goto error;
    }
  
  free ( nentries );
  tL_buffer_free ( buffer );
  free ( hist );
  
  arpa->N= N;
  
  return arpa;
  
 errorinitial:
  tL_msgerror ( err, _( "Special initial word %s appears"
		" in an invalid position in the %luth"
		" %u-gram" ) , begin_sym, _PS(n2)+1, n+1 );
  goto error;
 errorword:
  tL_msgerror ( err, _( "Unable to read the %uth word, or too"
		" few fields, in the %luth %u-gram" ) ,
		k+1, _PS(n2)+1, n+1 );
 error:
  tL_buffer_free ( buffer );
  if ( nentries != NULL ) free ( nentries );
  if ( hist != NULL ) free ( hist );
  if ( arpa != NULL ) free_arpa ( arpa );
  return NULL;
  
} /* end load_arpa */


static unsigned long
search_state_to (
		 const struct arpa   *arpa,
		 const unsigned long *hist,
		 int                  length
		 )
{
  
  const unsigned long *hist2;
  unsigned long to;
  tLProb bowt;
  
  
  hist2= hist;
  while ( !(to= search_state_hist ( arpa, hist2, length, &bowt )) )
    {
      if ( bowt != 0.0 )
	tL_warning ( _( "A history without state and back-off"
		     " has been detected" ) );
      ++hist2; --length;
    }
  
  return to;
  
} /* end search_state_to */


/* Actualitza TO i inserta en la pila si no està actiu. A més
   incrementa al variable nombre d'estats actius NS. */
static void
add_stack (
	   unsigned long       to,
	   struct arpa_state  *states,
	   size_t             *NS,
	   unsigned long     **stack,
	   size_t             *N,
	   size_t             *stack_size
	   )
{
  
  ++(*NS);
  states[to].active= TRUE;
  if ( *stack_size == *N )
    {
      (*stack_size)<<= 1;
      mt ( *stack= mem_realloc ( unsigned long, *stack,
				 *stack_size ) );
    }
  (*stack)[(*N)++]= to;
  
} /* end add_stack */


static void
calc_state_to (
	       struct arpa        *arpa,
	       struct arpa_state  *states,
	       struct arpa_edge   *edges,
	       unsigned long      *hist,
	       unsigned long       ed,
	       size_t             *NS,
	       unsigned long     **stack,
	       size_t             *N,
	       size_t             *stack_size
	       )
{
  
  unsigned long id, word, to;
  int length, i;
  
  
  /* Obté l'estat destí. */
  word= edges[ed].w;
  id= states[edges[ed].from].hist;
  length= get_full_hist ( arpa, hist, id );
  if ( length == (arpa->N-1) )
    {
      for ( i= 0; i < (length-1); ++i )
	hist[i]= hist[i+1];
      hist[i]= word;
    }
  else hist[length++]= word;
  to= search_state_to ( arpa, hist, length );
  
  /* Com el meu LM necessita que cada estat tinga una
     paraula associada, tinc que replicar l'estat sense
     història tantes vegades com símbols. De totes maneres
     sols s'inserten si són necessàris. */
  if ( to == 1 )
    to= arpa->begin_copies1+word;
  
  edges[ed].to= to;
  if ( !states[to].active )
    add_stack ( to, states, NS, stack, N, stack_size );
  
} /* end calc_state_to */


/* Fixa totes les arestes i estats que falten. Torna el nombre
   d'estats i arestes actives i l'identificador de l'estat inicial. */
static void
complete_arpa (
	       struct arpa   *arpa,
	       const tLDict  *vocab,
	       const char    *begin_sym,
	       size_t        *NS,
	       size_t        *NE,
	       unsigned long *start
	       )
{
  
  unsigned long *stack, *hist, state, ed, to, bos, id;
  size_t stack_size, N;
  struct arpa_state *states;
  struct arpa_edge *edges;
  tLProb bowt;
  int length, i;
  
  
  /* Inicialitza. */
  make_copies1 ( arpa, vocab );
  mt ( hist= mem_alloc ( unsigned long, arpa->N ) );
  mt ( stack= mem_alloc ( unsigned long, 1 ) );
  stack_size= 1; N= 1;
  hist[0]= BARPASYM;
  *start= (begin_sym==NULL) ?
    1 : search_state_hist ( arpa, hist, 1, &bowt );
  if ( *start == 0 ) *start= 1;
  stack[0]= *start;
  states= arpa->state_v;
  edges= arpa->ed_v;
  
  /* Cos. */
  *NS= 1; *NE= 0;
  states[*start].active= TRUE;
  while ( N != 0 )
    {
      
      state= stack[0];
      stack[0]= stack[--N];
      
      /* Calcula l'estat de backoff. */
      id= states[state].hist;
      if ( id != 0 )
	{
	  length= get_full_hist ( arpa, hist, id );
	  --length;
	  for ( i= 0; i < length; ++i )
	    hist[i]= hist[i+1];
	  bos= search_state_to ( arpa, hist, length );
	  if ( !states[bos].active )
	    add_stack ( bos, states, NS, &stack, &N, &stack_size );
	}
      else bos= 0;
      states[state].bos= bos;
      
      /* Processa les arestes que ixen. */
      for ( ed= states[state].edges;
	    ed != 0; ed= edges[ed].next )
	{
	  ++(*NE);
	  if ( edges[ed].to == 0 )
	    calc_state_to ( arpa, states, edges, hist, ed,
			    NS, &stack, &N, &stack_size );
	}
      
      /* Aresta a estat nul. Açò és necessari per als estats copia de
	 1. */
      ed= states[state].null_edge;
      if ( ed != 0 ) /* Activem l'estat apuntat i contem l'aresta. */
	{
	  ++(*NE);
	  to= edges[ed].to;
	  if ( !states[to].active )
	    add_stack ( to, states, NS, &stack, &N, &stack_size );
	}
      
    }
  
  /* Allibera memòria. */
  free ( stack );
  free ( hist );
  
} /* end complete_arpa */


static void
create_maps (
	     const struct arpa  *arpa,
	     size_t            **map,
	     unsigned long     **imap,
	     const size_t        NS
	     )
{
  
  const struct arpa_state *astates;
  size_t s;
  unsigned long n;
  
  
  *map= mem_alloc ( size_t, arpa->state_N );
  *imap= mem_alloc ( unsigned long, NS );
  astates= arpa->state_v;
  for ( s= 0, n= 0; n < arpa->state_N; ++n )
    if ( astates[n].active )
      {
	(*map)[n]= s;
	(*imap)[s++]= n;
      }
  
} /* end create_maps */


static tLLM *
arpa2lm (
	 const struct arpa *arpa,
	 size_t             NS,
	 size_t             NE,
	 size_t             start
	 )
{
  
  size_t *map, s;
  unsigned long *imap, n, ed, bos;
  const struct arpa_state *astates;
  const struct arpa_hist *anodes;
  const struct arpa_edge *aedges;
  tLLM *lm;
  tLLMState *states, state;
  tLLMEdge *ped;

  
  /* Crea mapa d'estats. MAP arpa->lm IMAP lm->arpa. */
  create_maps ( arpa, &map, &imap, NS );
  
  /* Reserve memòria per al LM i inicialitze. */
  mt ( lm= mem_alloc ( tLLM, 1 ) );
  mt ( lm->states= mem_alloc ( tLLMState, NS ) );
  mt ( lm->edges= mem_alloc ( tLLMEdge, NE ) );
  lm->start= map[start];
  lm->NS= NS;
  lm->NE= NE;
  lm->backoff= TRUE;
  
  /* Inicialitza estats i arestes. */
  states= lm->states;
  ped= lm->edges;
  anodes= arpa->hist_mem;
  aedges= arpa->ed_v;
  astates= arpa->state_v;
  for ( s= 0; s < NS; ++s )
    {
      
      /* Paraula i inicialitza. */
      n= imap[s];
      state.word=
	(n == start || n == 1) ?
	ULONG_MAX :
	((n >= arpa->begin_copies1) ?
	 n-arpa->begin_copies1 :
	 anodes[astates[n].hist].w);
      state.fp= astates[n].fp;
      bos= astates[n].bos;
      state.bos= (bos == 0)?s:map[bos];
      state.bowt=
	(n == 1 || n >= arpa->begin_copies1) ?
	-HUGE_VAL : anodes[astates[n].hist].bowt;
      
      /* Arestes no nul·les. */
      state.edges.begin= ped;
      for ( ed= astates[n].edges;
	    ed != 0;
	    ed= aedges[ed].next )
	{
	  ped->to= map[aedges[ed].to];
	  ped->p= aedges[ed].p;
	  ++ped;
	}
      state.edges.end= ped;
      
      /* Arestes nul·les. */
      state.null_edges.begin= ped;
      if ( (ed= astates[n].null_edge) != 0 )
	{
	  ped->to= map[aedges[ed].to];
	  ped->p= aedges[ed].p;
	  ++ped;
	}
      state.null_edges.end= ped;
      
      states[s]= state;
      
    }
  
  /* Allibera recursos. */
  free ( imap );
  free ( map );
  
  return lm;
  
} /* end arpa2lm */


static void
arpa2lm_binary_file (
		     FILE              *fd,
		     const struct arpa *arpa,
		     size_t             NS,
		     size_t             NE,
		     size_t             start,
		     const tLDict      *vocab
		     )
{
  
  size_t *map, s;
  unsigned long *imap, n, ed, bos;
  const struct arpa_state *astates;
  const struct arpa_hist *anodes;
  const struct arpa_edge *aedges;
  tLLMState state;
  tLLMEdge *off, edge;
  
  
  /* Crea mapa d'estats. MAP arpa->lm IMAP lm->arpa. */
  create_maps ( arpa, &map, &imap, NS );
  
  /* Header */
  anodes= arpa->hist_mem;
  aedges= arpa->ed_v;
  astates= arpa->state_v;
  fprintf ( fd, "BLM\n" );
  fprintf ( fd, "Backoff YES\n" );
  fprintf ( fd, "NStates %lu\n", _PS(NS) );
  fprintf ( fd, "NEdges %lu\n", _PS(NE) );
  fprintf ( fd, "Start %lu\n", _PS(map[start]) );
  
  /* Lexicon */
  fprintf ( fd, "Lexicon %lu\n", _PS(vocab->N) );
  for ( s= 0; s < vocab->N; ++s )
    tL_print_string_binary ( fd, tL_dict_get_token ( vocab, s ) );
  
  /* States */
  off= 0;
  for ( s= 0; s < NS; ++s )
    {
      n= imap[s];
      state.edges.begin= off;
      for ( ed= astates[n].edges; ed != 0; ed= aedges[ed].next, ++off );
      state.edges.end= off;
      state.null_edges.begin= off;
      if ( astates[n].null_edge != 0 ) ++off;
      state.null_edges.end= off;
      state.word= 
	(n == start || n == 1) ?
        ULONG_MAX :
        ((n >= arpa->begin_copies1) ?
         n-arpa->begin_copies1 :
         anodes[astates[n].hist].w);
      state.fp= astates[n].fp;
      bos= astates[n].bos;
      state.bos= (bos == 0) ? s : map[bos];
      state.bowt=
	(n == 1 || n >= arpa->begin_copies1) ?
	-HUGE_VAL : anodes[astates[n].hist].bowt;
      fwritee ( &state, sizeof(state), 1, fd );
    }
  
  /* Edges */
  off= 0;
  fwritee ( &off, sizeof(off), 1, fd );
  for ( s= 0; s < NS; ++s )
    {
      n= imap[s];
      /* Arestes no nul·les. */
      for ( ed= astates[n].edges; ed != 0; ed= aedges[ed].next )
        {
	  edge.to= map[aedges[ed].to];
	  edge.p= aedges[ed].p;
	  edge.w= 0; /* Will be set on read */
	  fwritee ( &edge, sizeof(edge), 1, fd );
	}
      /* Arestes nul·les. */
      if ( (ed= astates[n].null_edge) != 0 )
        {
	  edge.to= map[aedges[ed].to];
	  edge.p= aedges[ed].p;
	  edge.w= 0;
	  fwritee ( &edge, sizeof(edge), 1, fd );
	}
    }
  
  /* Allibera recursos. */
  free ( imap );
  free ( map );
  
} /* end arpa2lm_binary_file */


static void
arpa2lm_text_file (
		   FILE              *fd,
		   const struct arpa *arpa,
		   size_t             NS,
		   size_t             NE,
		   size_t             start,
		   const tLDict      *vocab
		   )
{
  
  size_t *map, s;
  unsigned long *imap, n, w, ed, off, ned, bos;
  const struct arpa_state *astates;
  const struct arpa_hist *anodes;
  const struct arpa_edge *aedges;
  
  
  /* Crea mapa d'estats. MAP arpa->lm IMAP lm->arpa. */
  create_maps ( arpa, &map, &imap, NS );
  
  /* Header */
  anodes= arpa->hist_mem;
  aedges= arpa->ed_v;
  astates= arpa->state_v;
  fprintf ( fd, "LM\n" );
  fprintf ( fd, "Backoff YES\n" );
  fprintf ( fd, "NStates %lu\n", _PS(NS) );
  fprintf ( fd, "NEdges %lu\n", _PS(NE) );
  
  /* States */
  off= 0;
  fprintf ( fd, "States\n" );
  for ( s= 0; s < NS; ++s )
    {
      
      /* Word and final prob. */
      fprintf ( fd, "%lu ", _PS(s) );
      n= imap[s];
      if ( n == start ) fprintf ( fd, "START " );
      else if ( n == 1 ) fprintf ( fd, "NULL " );
      else
	{
	  w= ((n >= arpa->begin_copies1) ?
	      n-arpa->begin_copies1 :
	      anodes[astates[n].hist].w);
	  fprintf ( fd, "\"%s\" ", tL_dict_get_token ( vocab, w ) );
	}
      fprintf ( fd, "FP " );
      tL_prob_print ( astates[n].fp, fd );
      
      /* Normal edges. */
      for ( ed= astates[n].edges, ned= 0; ed != 0; ed= aedges[ed].next, ++ned );
      fprintf ( fd, " B %lu E %lu ", off, off+ned ); off+= ned;
      ned= (astates[n].null_edge != 0) ? 1 : 0;
      fprintf ( fd, "NuB %lu NuE %lu ", off, off+ned ); off+= ned;
      
      /* Backoff */
      bos= astates[n].bos;
      fprintf ( fd, "BOS %lu BOW ", _PS((bos == 0)?s:map[bos]) );
      tL_prob_print ( (n == 1 || n >= arpa->begin_copies1) ?
		      -HUGE_VAL : anodes[astates[n].hist].bowt,
		      fd );
      fputc ( '\n', fd );
      
    }
  
  /* Edges */
  fprintf ( fd, "Edges\n" );
  off= 0;
  for ( s= 0; s < NS; ++s )
    {
      n= imap[s];
      /* Arestes no nul·les. */
      for ( ed= astates[n].edges; ed != 0; ed= aedges[ed].next, ++off )
        {
	  fprintf ( fd, "%lu To %lu Prob ",
		    _PS(off), _PS(map[aedges[ed].to]) );
	  tL_prob_print ( aedges[ed].p, fd );
	  fputc ( '\n', fd );
        }
      /* Arestes nul·les. */
      if ( (ed= astates[n].null_edge) != 0 )
        {
	  fprintf ( fd, "%lu To %lu Prob ",
		    _PS(off), _PS(map[aedges[ed].to]) );
	  tL_prob_print ( aedges[ed].p, fd );
	  fputc ( '\n', fd );
	  ++off;
        }
    }
  
  /* Allibera recursos. */
  free ( imap );
  free ( map );
  
} /* end arpa2lm_text_file */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(tLLM *)
tL_lm_new_from_arpa_file (
			  gzFile         from,
			  tLDict        *vocab,
			  const tLBool   emptyvocab,
			  const char    *begin_sym,
			  const char    *end_sym,
			  const char    *ninf,
			  char         **err
			  )
{
  
  struct arpa *arpa;
  tLLM *lm;
  size_t NS, NE;
  unsigned long start;
  
  
  arpa= load_arpa ( from, vocab, emptyvocab, begin_sym, end_sym, ninf, err );
  if ( arpa == NULL ) return NULL;
  complete_arpa ( arpa, vocab, begin_sym, &NS, &NE, &start );
  lm= arpa2lm ( arpa, NS, NE, start );
  free_arpa ( arpa );
  
  return lm;
  
} /* end tL_lm_new_from_arpa_file */


TL_API_FUNC(int)
tL_arpa2tlkformat (
		   gzFile         from,
		   FILE          *to,
		   const char    *begin_sym,
		   const char    *end_sym,
		   const char    *ninf,
		   const tLBool   binary,
		   char         **err
		   )
{
  
  struct arpa *arpa;
  size_t NS, NE;
  unsigned long start;
  tLDict *vocab;
  
  
  vocab= tL_dict_new ();
  arpa= load_arpa ( from, vocab, TRUE, begin_sym, end_sym, ninf, err );
  if ( arpa == NULL ) goto error;
  complete_arpa ( arpa, vocab, begin_sym, &NS, &NE, &start );
  if ( binary ) arpa2lm_binary_file ( to, arpa, NS, NE, start, vocab );
  else          arpa2lm_text_file ( to, arpa, NS, NE, start, vocab );
  free_arpa ( arpa );
  tL_dict_free ( vocab );
  
  return 0;
  
 error:
  tL_dict_free ( vocab );
  return -1;
  
} /* end tL_arpa2tlkformat */
