/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  arpa.h - Llig un model de llenguatge en format ARPA.
 *
 *  Adrià Giménez Pastor, 2009,2011.
 *
 */

#ifndef __LIBTL_ARPA_H__
#define __LIBTL_ARPA_H__

#include <stdio.h>
#include <zlib.h>

#include "lm_public.h"
#include "port.h"
#include "text_public.h"

/** \addtogroup LM
 *  @{
 */

/* Llig un model de llenguatge en format 'ARPA' d'un fitxer. Se li ha
 * de passar un diccionari buit que omplirà amb el vocabulari del
 * model. Si el diccionari no està buit les paraules fora de
 * vocabulari es consideren un error. Es necessari que s'especifiquen
 * els símbols que indiquen principi i final. A més s'ha d'especificar
 * el token per a indicar probabilitat infinita. El símbols inicials i
 * finals poden ser el mateix, es distingiran per la posició. El
 * símbol inicial és optatiu, NULL indica que no es troba present.
 */
/** Reads a language model from an ARPA file. This function creates a
 *  new language model from a text description in ARPA format stored
 *  in the file provided. A tLDict is provided to register the words.
 *  If the vocabulary is empty, new words are registered; else, an
 *  out-of-vocabulary word is treated as an error. The tokens used to
 *  represent the special final word and -INF are required. The token
 *  for the special initial word is optional. The token used for the
 *  final word can also be used for the initial word.
 *
 *  @param from File where the text description is stored. 
 *  @param vocab The dictionary where the words are registered.
 *  @param emptyvocab Specifies whether the provided dictionary is
 *  empty (new words must be registered) or not (new words are treated
 *  as errors).
 *  @param begin_sym The token for the special initial word. Can be
 *  NULL.
 *  @param end_sym The token for the special final word.
 *  @param ninf The token representing -INF.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The language model, or NULL in case of error. 
 */
TL_API_FUNC(tLLM *)
tL_lm_new_from_arpa_file (
			  gzFile         from,
			  tLDict        *vocab,
			  const tLBool   emptyvocab,
			  const char    *begin_sym,
			  const char    *end_sym,
			  const char    *ninf,
			  char         **err
			  );

/** Reads a language model from an ARPA file and writes it in TLK
 *  format. This function has similar effects to calling first
 *  'tL_lm_new_from_arpa_file' and then 'tL_lm_print', but less
 *  memory is used.
 *
 *  @param from File where the text description is stored.
 *  @param to File to which the model will be written.
 *  @param begin_sym The token for the special initial word. Can be
 *  NULL.
 *  @param end_sym The token for the special final word.
 *  @param ninf The token representing -INF.
 *  @param binary Flag to indicate if binary output is desired.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return 0 on success, or -1 in case of error.
 */
TL_API_FUNC(int)
tL_arpa2tlkformat (
		   gzFile         from,
		   FILE          *to,
		   const char    *begin_sym,
		   const char    *end_sym,
		   const char    *ninf,
		   const tLBool   binary,
		   char         **err
		   );

/** @} */

#endif /* __LIBTL_ARPA_H__ */
