/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  auxfstats.h - Serveix per a entrenar els estats. Sustitueix a
 *  'statestr.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011.
 *
 */

#ifndef __TL_PRV_AUXFSTATS_H__
#define __TL_PRV_AUXFSTATS_H__

#include <stddef.h>
#include <stdlib.h>

#include "amodel.h"
#include "common.h"
#include "samples.h"

/* TIPUS AUXFSTATSMNG */

typedef struct auxfstats_mng auxfstats_mng_t;

/* Mètodes. */
typedef struct
{
  
  /* Prepara el gestor per a una nova mostra. Pot ser NULL. */
  void (*init) (auxfstats_mng_t *afmng,
		void            *global,
		const tLFea     *fea);
  
  /* Allibera la memòria del gestor. */
  void   (*free) (auxfstats_mng_t *afmng);
  
  /* Reserva memòria per a un vector de T auxfstats. */
  void * (*alloc_auxfstats) (auxfstats_mng_t *afmng,
			     int              T);
  
  /* Allibera la memòria de T auxfstats. */
  void   (*free_auxfstats) (auxfstats_mng_t *afmng,
			    void            *auxfstats,
			    int              T);
  
  /* Reserva més memòria per als auxfstats. Es supossa oldT <=
     newT. Modificar la grandària pot malfer el contingut. */
  void * (*resize_auxfstats) (auxfstats_mng_t *afmng,
			      void            *auxfstats,
			      int              oldT,
			      int              newT);
  
  /* Pas pseudoE del EM. Torna la probabilitat del frame t per a
     l'estat q. A més s'ha d'indicar quin vector de auxfstats ha
     d'utilitzar per a emmagatzemar els estadístics necessaris per a
     poder més tard calcular el pas M. */
  tLProb (*pseudoE_step) (auxfstats_mng_t *afmng,
			  void            *auxfstats,
			  const tLFea     *fea,
			  int              t,
			  void            *global,
			  void            *states,
			  unsigned int     q);
  
  /* Pas pseudoM. Acumula posibles comptadors globals. Pot ser
     NULL. Es crida després d'acumular-ho tot a nivell d'estat. */
  void (*pseudoM_step_global) (auxfstats_mng_t *afmng,
			       const tLFea     *fea,
			       void            *global,
			       void            *cglobal);
  
  /* Pas pseudoM. Acumula els comptes asociats a un estat dels frames
     indicats. */
  void (*pseudoM_step) (auxfstats_mng_t *afmng,
			void            *auxfstats,
			tLProb          *weights,
			int             *frames,
			int              nframes,
			const tLFea     *fea,
			void            *global,
			void            *cglobal,
			void            *states,
			void            *cstates,
			unsigned int     q);
  
} auxfstats_mng_meths_t;

#define AUXFSTATS_MNG				\
  const auxfstats_mng_meths_t *meths

/* En l'entrenament EM per a cada frame i estat hi ha que guardar
   estadístiques. Les estadístiques que depenen del tipus són
   emmagatzemats en un auxfstats. Aquesta estructura és un gestor de
   'auxfstats'. */
struct auxfstats_mng
{
  AUXFSTATS_MNG;
};

#define auxfstats_mng_init(AFMNG,GLOBAL,FEA)		\
  do {							\
    if ( (AFMNG)->meths->init != NULL )			\
      (AFMNG)->meths->init ( AFMNG, GLOBAL, FEA );	\
  } while(0)

#define auxfstats_mng_free(AFMNG)		\
  (AFMNG)->meths->free ( AFMNG )

#define auxfstats_mng_alloc_auxfstats(AFMNG,T)		\
  (AFMNG)->meths->alloc_auxfstats ( (AFMNG), (T) )

#define auxfstats_mng_free_auxfstats(AFMNG,AUXFSTATS,T)		\
  (AFMNG)->meths->free_auxfstats ( (AFMNG), (AUXFSTATS), (T) )

#define auxfstats_mng_resize_auxfstats(AFMNG,AUXFSTATS,OLDT,NEWT)	\
  (AFMNG)->meths->resize_auxfstats ( (AFMNG), (AUXFSTATS), (OLDT), (NEWT) )

#define auxfstats_mng_pseudoE_step(AFMNG,AUXFSTATS,FEA,T,GLOBAL,STATES,Q) \
  (AFMNG)->meths->pseudoE_step ( (AFMNG), (AUXFSTATS), (FEA), (T),	\
				 (GLOBAL), (STATES), (Q) )

#define auxfstats_mng_pseudoM_step_global(AFMNG,FEA,GLOBAL,CGLOBAL)	\
  do {									\
    if ( (AFMNG)->meths->pseudoM_step_global != NULL )			\
      (AFMNG)->meths->pseudoM_step_global ( (AFMNG), (FEA), (GLOBAL),	\
					    (CGLOBAL) );		\
  } while(0)

#define auxfstats_mng_pseudoM_step(AFMNG,AUXFSTATS,WEIGHTS,FRAMES,NFRAMES, \
				   FEA,GLOBAL,CGLOBAL,STATES,CSTATES,Q)	\
  (AFMNG)->meths->pseudoM_step ( (AFMNG), (AUXFSTATS), (WEIGHTS), (FRAMES), \
				 (NFRAMES), (FEA), (GLOBAL), (CGLOBAL), \
				 (STATES), (CSTATES), (Q) )

/* Constructor. */
auxfstats_mng_t *
tL_auxfstats_mng_new (
		      const char  *type,
		      void        *global,
		      char       **err
		      );

#endif /* __TL_PRV_AUXFSTATS_H__ */
