/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  bernoulli.c - Implementació de 'bernoulli.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#include "auxfstats.h"
#include "bernoulli.h"
#include "calc.h"
#include "amcounts/logbernoulli.h"
#include "amodel/bernoulli.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "samples.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) BERNOULLI_GLOBAL(ptr)
#define STATE(ptr) BERNOULLI_STATE(ptr)
#define CSTATE(ptr) ((logamcount_bernoulli_state_t *) (ptr))
#define CSTATE2(ptr) ((iamcount_bernoulli_state_t *) (ptr))
#define AFMNG(ptr) ((afmng_bernoulli_t *) (ptr))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  AUXFSTATS_MNG;
  tLFloat *calc1;    /* |T+1|. */
  tLFloat *calc0;    /* |T+1|. */
  int      T;
  
} afmng_bernoulli_t;




/***********/
/* MÈTODES */
/***********/

static void
init (
      auxfstats_mng_t *afmng,
      void            *global,
      const tLFea     *fea
      )
{
  
  int nsize;
  
  
  if ( fea->cols <= AFMNG(afmng)->T ) return;
  nsize= fea->cols+1;
  mt ( AFMNG(afmng)->calc0=
       mem_realloc ( tLFloat, AFMNG(afmng)->calc0, nsize*2 ) );
  AFMNG(afmng)->calc1= AFMNG(afmng)->calc0 + nsize;
  AFMNG(afmng)->T= fea->cols;
  
} /* end init */


static void
_free (
       auxfstats_mng_t *afmng
       )
{
  
  free ( AFMNG(afmng)->calc0 );
  free ( afmng );
  
} /* end _free */


static void *
alloc_auxfstats (
		 auxfstats_mng_t *afmng,
		 int              T
		 )
{
  return NULL;
} /* end alloc_auxfstats */


static void
free_auxfstats (
		auxfstats_mng_t *afmng,
		void            *auxfstats,
		int              T
		)
{
} /* end free_auxfstats */


static void *
resize_auxfstats (
		  auxfstats_mng_t *afmng,
		  void            *auxfstats,
		  int              oldT,
		  int              newT
		  )
{
  return auxfstats;
} /* end resize_auxfstats */


static tLProb
pseudoE_step (
	      auxfstats_mng_t *afmng,
	      void            *auxfstats,
	      const tLFea     *fea,
	      int              t,
	      void            *global,
	      void            *states,
	      unsigned int     q
	      )
{
  
  tLProb prob;
  unsigned int d, D;
  const bit *frame;
  const int *i;
  bernoulli_state_t state;
  
  
  state= STATE(states)[q];
  D= GLOBAL(global)->D;
  prob= 0.0;
  if ( state.zeros )
    {
      frame= B2cb(fea->v[t]);
      for ( d= 0; d < D; ++d )
        prob+=
          frame[d] ?
          state.p[d].p1 :
          state.p[d].p0;
    }
  else
    {
      for ( i= tL_fea_get_sparse ( fea )[t]; *i != -1; ++i )
        prob+= state.l[*i];
      prob+= state.logc;
    }
  
  return prob;
  
} /* end pseudoE_step */


static void
pseudoM_step (
	      auxfstats_mng_t *afmng,
	      void            *auxfstats,
	      tLProb          *weights,
	      int             *frames,
	      int              nframes,
	      const tLFea     *fea,
	      void            *global,
	      void            *cglobal,
	      void            *states,
	      void            *cstates,
	      unsigned int     q
	      )
{
  
  unsigned int D, d;
  logamcount_bernoulli_state_t scount;
  tLFloat *calc1, *calc0, max1, max0, aux;
  volatile tLFloat fix;
  int n, N0, N1, t;
  tLProb weight;
  tLByte **bytes;
  const bit *frame;
  
  
  D= GLOBAL(global)->D;
  scount= CSTATE(cstates)[q];
  calc0= AFMNG(afmng)->calc0;
  calc1= AFMNG(afmng)->calc1;
  bytes= fea->v;
  
  for ( d= 0; d < D; ++d )
    {
      
      max1= calc1[0]= scount.p[d].p1;
      max0= calc0[0]= scount.p[d].p0;
      N1= N0= 1;
      
      for ( n= 0; n < nframes; ++n )
	{
	  t= frames[n];
	  if ( (weight= weights[t]) != -HUGE_VAL )
	    {
	      frame= B2cb(bytes[t]);
	      if ( frame[d] )
		{
		  calc1[N1++]= aux= fix= weight;
		  if ( aux > max1 ) max1= aux;
		}
	      else
		{
		  calc0[N0++]= aux= fix= weight;
		  if ( aux > max0 ) max0= aux;
		}
	    }
	}
      
      if ( N1 > 1 )
        scount.p[d].p1= tL_radd_vals ( calc1, max1, N1 );
      
      if ( N0 > 1 )
        scount.p[d].p0= tL_radd_vals ( calc0, max0, N0 );
      
    }
  
} /* end pseudoM_step */


static const auxfstats_mng_meths_t meths=
  {
    init,
    _free,
    alloc_auxfstats,
    free_auxfstats,
    resize_auxfstats,
    pseudoE_step,
    NULL,
    pseudoM_step
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

auxfstats_mng_t *
tL_create_afmng_bernoulli (
			   void  *global,
			   char **err
			   )
{
  
  afmng_bernoulli_t *new;
  
  
  mt ( new= mem_alloc ( afmng_bernoulli_t, 1 ) );
  mt ( new->calc0= mem_alloc ( tLFloat, 4 ) );
  new->calc1= new->calc0 + 2;
  new->T= 1;
  new->meths= &meths;
  
  return (auxfstats_mng_t *) new;
  
} /* end tL_create_afmng_bernoulli */
