/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  dgaussian.c - Implementació de 'dgaussian.h'.
 *
 *  Adrià Giménez Pastor, 2009,2011,2012,2013.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#include "auxfstats.h"
#include "logamcounts.h"
#include "amcounts/dgaussian.h"
#include "amodel/dgaussian.h"
#include "common.h"
#include "dgaussian.h"
#include "error.h"
#include "mem.h"
#include "samples.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) DGAUSSIAN_GLOBAL(ptr)
#define CGLOBAL(ptr) ((amcount_dgaussian_global_t *) (ptr))
#define STATE(ptr) DGAUSSIAN_STATE(ptr)
#define CSTATE(ptr) ((amcount_dgaussian_state_t *) (ptr))
#define AFMNG(ptr) ((afmng_dgaussian_t *) (ptr))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  AUXFSTATS_MNG;
  tLFloat       *ax;
  tLFloat       *ax2;
  tLFloat      **axx;
  tLFloat       *calc;
  const float  **frames;
  int            T;
  
} afmng_dgaussian_t;




/***********/
/* MÈTODES */
/***********/

static void
init (
      auxfstats_mng_t *afmng,
      void            *global,
      const tLFea     *fea
      )
{
  
  if ( fea->cols <= AFMNG(afmng)->T ) return;
  mt ( AFMNG(afmng)->calc=
       mem_realloc ( tLFloat, AFMNG(afmng)->calc, fea->cols ) );
  mt ( AFMNG(afmng)->frames=
       mem_realloc ( const float *, AFMNG(afmng)->frames, fea->cols ) );
  AFMNG(afmng)->T= fea->cols;
  
} /* end init */


static void
_free (
       auxfstats_mng_t *afmng
       )
{
  
  if ( AFMNG(afmng)->axx != NULL )
    {
      free ( AFMNG(afmng)->axx[0] );
      free ( AFMNG(afmng)->axx );
    }
  free ( AFMNG(afmng)->ax );
  free ( AFMNG(afmng)->calc );
  free ( AFMNG(afmng)->frames );
  free ( afmng );
  
} /* end _free */


static void *
alloc_auxfstats (
                 auxfstats_mng_t *afmng,
                 int              T
                 )
{
  return NULL;
} /* end alloc_auxfstats */


static void
free_auxfstats (
                auxfstats_mng_t *afmng,
                void            *auxfstats,
                int              T
                )
{
} /* end free_auxfstats */


static void *
resize_auxfstats (
                  auxfstats_mng_t *afmng,
                  void            *auxfstats,
                  int              oldT,
                  int              newT
                  )
{
  return auxfstats;
} /* end resize_auxfstats */


static tLProb
pseudoE_step (
              auxfstats_mng_t *afmng,
              void            *auxfstats,
              const tLFea     *fea,
              int              t,
              void            *global,
              void            *states,
              unsigned int     q
              )
{
  
  tLFloat prob, aux;
  unsigned int d, D;
  const float *frame;
  dgaussian_state_t state;
  
  
  state= STATE(states)[q];
  D= GLOBAL(global)->D;
  frame= B2cR(fea->v[t]);
  prob= 0.0;
  for ( d= 0; d < D; ++d )
    {
      aux= frame[d]-state.mu[d];
      prob+= (aux*aux)*state.ivar[d];
    }
  
  return (tLProb) (-0.5*prob+state.logc);
  
} /* end pseudoE_step */


static void
pseudoM_step (
              auxfstats_mng_t *afmng,
              void            *auxfstats,
              tLProb          *weights,
              int             *frames,
              int              nframes,
              const tLFea     *fea,
              void            *global,
              void            *cglobal,
              void            *states,
              void            *cstates,
              unsigned int     q
              )
{
  
  unsigned int d, d2, D, D2;
  amcount_dgaussian_state_t scount;
  tLFloat *calc, *ax, *ax2, **axx, *axx_d, *mem, *xx_d, max,
    aux, aux2, anorm, aux3;
  tLByte **bytes;
  const float **frames_;
  const float *frame;
  int count, n, t, aboutxx;
  dgaussian_state_t state;
  
  
  D= GLOBAL(global)->D;
  aboutxx= CGLOBAL(cglobal)->aboutxx;
  scount= CSTATE(cstates)[q];
  if ( aboutxx == MLLR ) state= STATE(states)[q];
  calc= AFMNG(afmng)->calc;
  ax= AFMNG(afmng)->ax;
  ax2= AFMNG(afmng)->ax2;
  frames_= AFMNG(afmng)->frames;
  bytes= fea->v;
  
  count= 0; max= -HUGE_VAL;
  for ( n= 0; n < nframes; ++n )
    {
      t= frames[n];
      if ( (aux= weights[t]) != -HUGE_VAL )
	{
	  calc[count]= aux;
	  if ( aux > max ) max= aux;
	  frames_[count++]= B2cR(bytes[t]);
	}
    }
  
  if ( count == 0 ) return;
  for ( d= 0; d < D; ++d ) ax[d]= 0.0;
  for ( d= 0; d < D; ++d ) ax2[d]= 0.0;
  if ( aboutxx != NONE )
    {
      D2= (aboutxx==CMLLR)?(D+1):D;
      axx= AFMNG(afmng)->axx;
      if ( axx == NULL )
	{
	  mt ( AFMNG(afmng)->axx= mem_alloc ( tLFloat *, (D+1) ) );
	  mt ( AFMNG(afmng)->axx[0]= mem= mem_alloc ( tLFloat, (D+1)*D2 ) );
	  for ( d= 0; d <= D; ++d, mem+= D2 )
	    AFMNG(afmng)->axx[d]= mem;
	  axx= AFMNG(afmng)->axx;
	}
      for ( d= 0; d <= D; ++d )
	{
	  axx_d= axx[d];
	  for ( d2= 0; d2 < D2; ++d2 ) axx_d[d2]= 0.0;
	}
    }
  else axx= NULL; /*CALLA!*/
  anorm= 0;
  for ( n= 0; n < count; ++n )
    {
      frame= frames_[n];
      aux3= calc[n]-max;
      if ( aux3 < LOGEPS ) continue;
      aux= exp ( aux3 );
      for ( d= 0; d < D; ++d )
        {
          aux2= frame[d];
          ax[d]+= aux*aux2;
          ax2[d]+= aux*(aux2*aux2);
        }
      anorm+= aux;
      if ( aboutxx == CMLLR )
	{
	  axx_d= axx[0];
	  axx_d[0]+= aux;
	  for ( d2= 1; d2 <= D; ++d2 ) axx_d[d2]+= aux*frame[d2-1];
	  for ( d= 1; d <= D; ++d )
	    {
	      aux2= frame[d-1];
	      axx_d= axx[d];
	      axx_d[0]+= aux*aux2;
	      for ( d2= 1; d2 <= D; ++d2 ) axx_d[d2]+= aux*aux2*frame[d2-1];
	    }
	}
      else if ( aboutxx == MLLR )
	{
	  /* Columna offset. */
	  axx_d= axx[0];
	  for ( d2= 0; d2 < D; ++d2 )
	    axx_d[d2]+= frame[d2]*state.ivar[d2]*aux;
	  /* Columnes mitjes. */
	  for ( d= 1; d <= D; ++d )
	    {
	      axx_d= axx[d];
	      aux2= state.mu[d-1]*aux;
	      for ( d2= 0; d2 < D; ++d2 )
		axx_d[d2]+= frame[d2]*state.ivar[d2]*aux2;
	    }
	}
    }
  max= exp ( max );
  CSTATE(cstates)[q].norm+= max*anorm;
  for ( d= 0; d < D; ++d )
    {
      scount.x[d]+= max*ax[d];
      scount.x2[d]+= max*ax2[d];
    }
  if ( aboutxx != NONE )
    {
      D2= (aboutxx==CMLLR)?(D+1):D;
      for ( d= 0; d <= D; ++d )
	{
	  xx_d= scount.xx[d];
	  axx_d= axx[d];
	  for ( d2= 0; d2 < D2; ++d2 )
	    xx_d[d2]+= max*axx_d[d2];
	}
    }
  
} /* end pseudoM_step */


static const auxfstats_mng_meths_t meths=
  {
    init,
    _free,
    alloc_auxfstats,
    free_auxfstats,
    resize_auxfstats,
    pseudoE_step,
    NULL,
    pseudoM_step
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

auxfstats_mng_t *
tL_create_afmng_dgaussian (
                           void  *global,
			   char **err
                           )
{
  
  afmng_dgaussian_t *new;
  
  
  mt ( new= mem_alloc ( afmng_dgaussian_t, 1 ) );
  mt ( new->calc= mem_alloc ( tLFloat, 1 ) );
  mt ( new->frames= mem_alloc ( const float *, 1 ) );
  mt ( new->ax= mem_alloc ( tLFloat, GLOBAL(global)->D*2 ) );
  new->ax2= new->ax+GLOBAL(global)->D;
  new->axx= NULL;
  new->T= 1;
  new->meths= &meths;
  
  return (auxfstats_mng_t *) new;
  
} /* end tL_create_afmng_dgaussian */
