/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mixture.c - Implementació de 'mixture.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <stddef.h>
#include <stdlib.h>

#include "auxfstats.h"
#include "calc.h"
#include "amcounts/logmixture.h"
#include "amodel/mixture.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "samples.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) MIXTURE_GLOBAL(ptr)
#define CGLOBAL(ptr) ((logamcount_mixture_global_t *) (ptr))
#define STATE(ptr) MIXTURE_STATE(ptr)
#define CSTATE(ptr) ((logamcount_mixture_state_t *) (ptr))
#define AFMNG(ptr) ((afmng_mixture_t *) (ptr))
#define AUX(ptr) ((aux_t *) (ptr))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  AUXFSTATS_MNG;
  auxfstats_mng_t  *afmng;
  tLFloat          *calc;    /* Com a mínim max_I. */
  unsigned int      max_I;
  int               T;
  
} afmng_mixture_t;


typedef struct
{
  
  unsigned int   I;       /* Per a poder reajustar. */
  int            T;       /* Per a poder reajustar. */
  tLProb       **weights; /* Els pesos els reajuste cada vegada que es
			     calcula pseudo_E, per a aprofitar millor
			     la memòria. */
  void         **aux;     /* Açò ho deixe sempre a max_I. */
  
} aux_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
resize_auxfstats_I (
		    auxfstats_mng_t *afmng,
		    aux_t           *aux,
		    unsigned int     newI
		    )
{

  unsigned int i;
  tLProb *mem;
  
  
  if ( aux->I == newI ) return;
  mt ( aux->weights= mem_realloc ( tLProb *, aux->weights, newI ) );
  mt ( aux->weights[0]= mem=
       mem_realloc ( tLProb, aux->weights[0], newI*aux->T ) );
  for ( i= 0; i < newI; ++i, mem+= aux->T ) aux->weights[i]= mem;
  aux->I= newI;
  
} /* end resize_auxfstats_I */




/***********/
/* MÈTODES */
/***********/

static void
init (
      auxfstats_mng_t *afmng,
      void            *global,
      const tLFea     *fea
      )
{
  
  /* És important que estiga abans del if. */
  auxfstats_mng_init ( AFMNG(afmng)->afmng, GLOBAL(global)->global, fea );
  if ( fea->cols <= AFMNG(afmng)->T ) return;
  mt ( AFMNG(afmng)->calc=
       mem_realloc ( tLFloat, AFMNG(afmng)->calc, fea->cols+1 ) );
  AFMNG(afmng)->T= fea->cols;
  
} /* end init */


static void
_free (
       auxfstats_mng_t *afmng
       )
{
  
  auxfstats_mng_free ( AFMNG(afmng)->afmng );
  free ( AFMNG(afmng)->calc );
  free ( afmng );
  
} /* end _free */


static void *
alloc_auxfstats (
                 auxfstats_mng_t *afmng,
                 int              T
                 )
{
  
  aux_t *new;
  unsigned int i, max_I;
  
  
  max_I= AFMNG(afmng)->max_I;
  mt ( new= mem_alloc ( aux_t, 1 ) );
  new->I= 1;
  new->T= T;
  mt ( new->weights= mem_alloc ( tLProb *, 1 ) );
  mt ( new->weights[0]= mem_alloc ( tLProb, T ) );
  mt ( new->aux= mem_alloc ( void *, max_I ) );
  for ( i= 0; i < max_I; ++i )
    new->aux[i]= auxfstats_mng_alloc_auxfstats ( AFMNG(afmng)->afmng, T );
  
  return (void *) new;
  
} /* end alloc_auxfstats */


static void
free_auxfstats (
                auxfstats_mng_t *afmng,
                void            *auxfstats,
                int              T
                )
{
  
  unsigned int max_I, i;
  
  
  max_I= AFMNG(afmng)->max_I;
  for ( i= 0; i < max_I; ++i )
    auxfstats_mng_free_auxfstats ( AFMNG(afmng)->afmng,
				   AUX(auxfstats)->aux[i], T );
  free ( AUX(auxfstats)->aux );
  free ( AUX(auxfstats)->weights[0] );
  free ( AUX(auxfstats)->weights );
  free ( auxfstats );
  
} /* end free_auxfstats */


static void *
resize_auxfstats (
                  auxfstats_mng_t *afmng,
                  void            *auxfstats,
                  int              oldT,
                  int              newT
                  )
{

  unsigned int max_I, i, I;
  tLProb *mem;
  
  
  max_I= AFMNG(afmng)->max_I;
  I= AUX(auxfstats)->I;
  AUX(auxfstats)->T= newT;
  mt ( AUX(auxfstats)->weights[0]= mem=
       mem_realloc ( tLProb, AUX(auxfstats)->weights[0], I*newT ) );
  for ( i= 0; i < I; ++i, mem+= newT )
    AUX(auxfstats)->weights[i]= mem;
  for ( i= 0; i < max_I; ++i )
    auxfstats_mng_resize_auxfstats ( AFMNG(afmng)->afmng,
				     AUX(auxfstats)->aux[i], oldT, newT );
  
  return auxfstats;
  
} /* end resize_auxfstats */


static tLProb
pseudoE_step (
	      auxfstats_mng_t *afmng,
	      void            *auxfstats,
	      const tLFea     *fea,
	      int              t,
	      void            *global,
	      void            *states,
	      unsigned int     q
	      )
{
  
  void **auxfs;
  mixture_state_t state;
  tLFloat *calc, max, aux, ret;
  unsigned int i;
  volatile tLFloat fix;
  tLProb **weights;
  
  
  /* Calcula probabilitats. */
  auxfs= AUX(auxfstats)->aux;
  state= STATE(states)[q];
  resize_auxfstats_I ( afmng, AUX(auxfstats), state.I );
  calc= AFMNG(afmng)->calc;
  max= -HUGE_VAL;
  for ( i= 0; i < state.I; ++i )
    {
      calc[i]= aux= fix= 
	state.p[i] + auxfstats_mng_pseudoE_step ( AFMNG(afmng)->afmng,
						  auxfs[i], fea, t,
						  GLOBAL(global)->global,
						  state.members, i );
      if ( aux > max ) max= aux;
    }
  
  /* Calcula probabilitat total i renormalitza. */
  ret=
    (state.I > 1 && max != -HUGE_VAL) ?
    tL_radd_vals ( calc, max, state.I ) :
    max;
  weights= AUX(auxfstats)->weights;
  if ( ret != -HUGE_VAL )
    for ( i= 0; i < state.I; ++i )
      weights[i][t]= calc[i] - ret;
  else
    for ( i= 0; i < state.I; ++i )
      weights[i][t]= -HUGE_VAL;
  
  return (tLProb) ret;
  
} /* end pseudoE_step */


static void
pseudoM_step_global (
		     auxfstats_mng_t *afmng,
		     const tLFea     *fea,
		     void            *global,
		     void            *cglobal
		     )
{
  auxfstats_mng_pseudoM_step_global ( AFMNG(afmng)->afmng, fea,
				      GLOBAL(global)->global,
				      CGLOBAL(cglobal)->global );
} /* end pseudoM_step_global */


static void
pseudoM_step (
              auxfstats_mng_t *afmng,
              void            *auxfstats,
              tLProb          *weights,
              int             *frames,
              int              nframes,
              const tLFea     *fea,
              void            *global,
              void            *cglobal,
              void            *states,
              void            *cstates,
              unsigned int     q
              )
{
  
  unsigned int i;
  mixture_state_t state;
  logamcount_mixture_state_t scount;
  tLFloat *calc, max, aux;
  int n, N, t;
  volatile tLFloat fix;
  tLProb **weights_, *weights_i;
  void ** auxfs;
  
  
  scount= CSTATE(cstates)[q];
  state= STATE(states)[q];
  calc= AFMNG(afmng)->calc;
  weights_= AUX(auxfstats)->weights;
  auxfs= AUX(auxfstats)->aux;
  
  for ( i= 0; i < state.I; ++i )
    {
      
      weights_i= weights_[i];
      max= calc[0]= scount.p[i];
      N= 1;
      
      for ( n= 0; n < nframes; ++n )
	{
	  t= frames[n];
	  calc[N++]= aux= fix= (weights_i[t]+= weights[t]);
	  if ( aux > max ) max= aux;
	}
      
      scount.p[i]=
	(N > 1 && max != -HUGE_VAL ) ?
	tL_radd_vals ( calc, max, N ) :
	max;
      auxfstats_mng_pseudoM_step ( AFMNG(afmng)->afmng, auxfs[i], weights_i,
				   frames, nframes, fea,
				   GLOBAL(global)->global,
				   CGLOBAL(cglobal)->global,
				   state.members, scount.members, i );
      
    }
  
} /* end pseudoM_step */


static const auxfstats_mng_meths_t meths=
  {
    init,
    _free,
    alloc_auxfstats,
    free_auxfstats,
    resize_auxfstats,
    pseudoE_step,
    pseudoM_step_global,
    pseudoM_step
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

auxfstats_mng_t *
tL_create_afmng_mixture (
			 void  *global,
			 char **err
			 )
{
  
  afmng_mixture_t *new;
  
  
  mt ( new= mem_alloc ( afmng_mixture_t, 1 ) );
  new->meths= &meths;
  new->afmng= tL_auxfstats_mng_new ( GLOBAL(global)->type,
				     GLOBAL(global)->global,
				     err );
  if ( new->afmng == NULL )
    {
      free ( new );
      return NULL;
    }
  new->max_I= GLOBAL(global)->max_I;
  mt ( new->calc= mem_alloc ( tLFloat, new->max_I+1 ) );
  new->T= new->max_I;
  
  return (auxfstats_mng_t *) new;
  
} /* end tL_create_afmng_mixture */
