/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mixture.c - Implementació de 'tiedstates.h'.
 *
 *  Adrià Giménez Pastor, 2011,2013.
 *
 */


#include <assert.h>
#include <stddef.h>
#include <stdlib.h>

#include "auxfstats.h"
#include "calc.h"
#include "amcounts/logtiedstates.h"
#include "amodel/tiedstates.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "samples.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) TIEDSTATES_GLOBAL(ptr)
#define CGLOBAL(ptr) ((logamcount_tiedstates_global_t *) (ptr))
#define STATE(ptr) TIEDSTATES_STATE(ptr)
#define AFMNG(ptr) ((afmng_tiedstates_t *) (ptr))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  tLProb p;
  tLBool a;
  
} node_t;


typedef struct
{
  
  node_t *nodes;
  tLProb *weights;
  void   *auxfstats;
  int    *stack;
  int     T;
  int     N;
  
} fstats_t;


typedef struct
{
  
  AUXFSTATS_MNG;
  auxfstats_mng_t *afmng;
  fstats_t        *fstats;    /* |N| com a màxim. */
  size_t           size;      /* Capacitat de fstats, N màxim. */
  unsigned int     N;
  unsigned int    *map;       /* |N|. Torna el fstats associat a un
				 estat, N si no en té.  */
  unsigned int    *actives;   /* |N|. Pila amb els estats actius. */
  unsigned int     Na;        /* Número de fstats actius. */
  int              currentT;
  
} afmng_tiedstates_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
fstats_init (
	     fstats_t        *fstats,
	     auxfstats_mng_t *afmng
	     )
{
  
  mt ( fstats->nodes= mem_alloc ( node_t, 1 ) );
  fstats->nodes[0].a= FALSE;
  mt ( fstats->weights= mem_alloc ( tLProb, 1 ) );
  fstats->auxfstats= auxfstats_mng_alloc_auxfstats ( AFMNG(afmng)->afmng, 1 );
  mt ( fstats->stack= mem_alloc ( int, 1 ) );
  fstats->T= 1;
  fstats->N= 0;
  
} /* end fstats_init */


static void
fstats_free (
	     fstats_t        *fstats,
	     auxfstats_mng_t *afmng
	     )
{
  
  free ( fstats->nodes );
  free ( fstats->weights );
  auxfstats_mng_free_auxfstats ( AFMNG(afmng)->afmng, fstats->auxfstats,
				 fstats->T );
  free ( fstats->stack );
  
} /* end fstats_free */


static void
fstats_clear (
	      fstats_t *fstats
	      )
{
  
  int n;
  
  
  for ( n= 0; n < fstats->N; ++n )
    fstats->nodes[fstats->stack[n]].a= FALSE;
  fstats->N= 0;
  
} /* end fstats_clear */


static void
fstats_resize (
	       fstats_t        *fstats,
	       int              T,
	       auxfstats_mng_t *afmng
	       )
{
  
  int t;
  
  
  mt ( fstats->nodes= mem_realloc ( node_t, fstats->nodes, T ) );
  for ( t= fstats->T; t < T; ++t )
    fstats->nodes[t].a= FALSE;
  mt ( fstats->weights= mem_realloc ( tLProb, fstats->weights, T ) );
  fstats->auxfstats=
    auxfstats_mng_resize_auxfstats ( AFMNG(afmng)->afmng,
				     fstats->auxfstats,
				     fstats->T, T );
  mt ( fstats->stack= mem_realloc ( int, fstats->stack, T ) );
  fstats->T= T;
  
} /* end fstats_resize */




/***********/
/* MÈTODES */
/***********/

static void
init (
      auxfstats_mng_t *afmng,
      void            *global,
      const tLFea     *fea
      )
{
  
  unsigned int n;
  
  
  auxfstats_mng_init ( AFMNG(afmng)->afmng, GLOBAL(global)->global, fea );
  AFMNG(afmng)->currentT= fea->cols;
  for ( n= 0; n < AFMNG(afmng)->Na; ++n )
    AFMNG(afmng)->map[AFMNG(afmng)->actives[n]]= AFMNG(afmng)->N;
  AFMNG(afmng)->Na= 0;
  
} /* end init */


static void
_free (
       auxfstats_mng_t *afmng
       )
{
  
  unsigned int n;
  
  
  free ( AFMNG(afmng)->actives );
  free ( AFMNG(afmng)->map );
  for ( n= 0; n < AFMNG(afmng)->size; ++n )
    fstats_free ( &(AFMNG(afmng)->fstats[n]), afmng );
  free ( AFMNG(afmng)->fstats );
  auxfstats_mng_free ( AFMNG(afmng)->afmng );
  free ( afmng );
  
} /* end _free */


static void *
alloc_auxfstats (
                 auxfstats_mng_t *afmng,
                 int              T
                 )
{
  return NULL;
} /* end alloc_auxfstats */


static void
free_auxfstats (
                auxfstats_mng_t *afmng,
                void            *auxfstats,
                int              T
                )
{
} /* end free_auxfstats */


static void *
resize_auxfstats (
                  auxfstats_mng_t *afmng,
                  void            *auxfstats,
                  int              oldT,
                  int              newT
                  )
{
  return auxfstats;
} /* end resize_auxfstats */


static tLProb
pseudoE_step (
              auxfstats_mng_t *afmng,
              void            *auxfstats,
              const tLFea     *fea,
              int              t,
              void            *global,
              void            *states,
              unsigned int     q
              )
{
  
  node_t *nodes;
  fstats_t *fstats;
  unsigned int state, fstats_id, n;
  tLProb ret;
  size_t nsize;
  
  
  state= STATE(states)[q];
  
  /* Obté els fstats. */
  fstats_id= AFMNG(afmng)->map[state];
  if ( fstats_id == AFMNG(afmng)->N )
    {
      if ( AFMNG(afmng)->Na == AFMNG(afmng)->size )
	{
	  nsize= AFMNG(afmng)->size*2;
	  if ( nsize > AFMNG(afmng)->N ) nsize= AFMNG(afmng)->N;
	  mt ( AFMNG(afmng)->fstats=
	       mem_realloc ( fstats_t, AFMNG(afmng)->fstats, nsize ) );
	  for ( n= AFMNG(afmng)->size; n < nsize; ++n )
	    fstats_init ( &(AFMNG(afmng)->fstats[n]), afmng );
	  AFMNG(afmng)->size= nsize;
	}
      fstats_id= (AFMNG(afmng)->Na)++;
      AFMNG(afmng)->actives[fstats_id]= state;
      AFMNG(afmng)->map[state]= fstats_id;
      fstats= &(AFMNG(afmng)->fstats[fstats_id]);
      fstats_clear ( fstats );
    }
  else fstats= &(AFMNG(afmng)->fstats[fstats_id]);
  if ( AFMNG(afmng)->currentT > fstats->T )
    fstats_resize ( fstats, AFMNG(afmng)->currentT, afmng );
  
  /* Obté la probabilitat. */
  nodes= fstats->nodes;
  if ( nodes[t].a ) return nodes[t].p;
  nodes[t].a= TRUE;
  fstats->weights[t]= -HUGE_VAL;
  nodes[t].p= ret= auxfstats_mng_pseudoE_step ( AFMNG(afmng)->afmng,
						fstats->auxfstats,
						fea, t,
						GLOBAL(global)->global,
						GLOBAL(global)->states,
						state );
  fstats->stack[fstats->N++]= t;
  
  return ret;
  
} /* end pseudoE_step */


static void
pseudoM_step_global (
		     auxfstats_mng_t *afmng,
		     const tLFea     *fea,
		     void            *global,
		     void            *cglobal
		     )
{
  
  unsigned int n, state, fstats_id;
  fstats_t *fstats, *fstats_n;
  
  
  fstats= AFMNG(afmng)->fstats;
  for ( n= 0; n < AFMNG(afmng)->Na; ++n )
    {
      state= AFMNG(afmng)->actives[n];
      fstats_id= AFMNG(afmng)->map[state];
      AFMNG(afmng)->map[state]= AFMNG(afmng)->N;
      fstats_n= &(fstats[fstats_id]);
      if ( fstats_n->N > 0 )
	{
	  auxfstats_mng_pseudoM_step ( AFMNG(afmng)->afmng, fstats_n->auxfstats,
				       fstats_n->weights, fstats_n->stack,
				       fstats_n->N, fea,
				       GLOBAL(global)->global,
				       CGLOBAL(cglobal)->global,
				       GLOBAL(global)->states,
				       CGLOBAL(cglobal)->states, state );
	  /* fstats_clear ( fstats_n ); No és necessari, al assignar-lo
	   es fa el clear. */ 
	}
    }
  AFMNG(afmng)->Na= 0;
  auxfstats_mng_pseudoM_step_global ( AFMNG(afmng)->afmng, fea,
                                      GLOBAL(global)->global,
                                      CGLOBAL(cglobal)->global );
  
} /* end pseudoM_step_global */


static void
pseudoM_step (
              auxfstats_mng_t *afmng,
              void            *auxfstats,
              tLProb          *weights,
              int             *frames,
              int              nframes,
              const tLFea     *fea,
              void            *global,
              void            *cglobal,
              void            *states,
              void            *cstates,
              unsigned int     q
              )
{
  
  unsigned int state, fstats_id;
  tLProb *weights_, weight;
  int n, t;
  
  
  state= STATE(states)[q];
  fstats_id= AFMNG(afmng)->map[state];
  /* This is a strange case, could occur for example in this
     situation: The transcription has 3 phonems (3 states per phonem
     with lineal topology) and a SP (1 state with aIF transition), and
     the sample has 12 vectors. It can be processed but skipping the
     SP. Thus, SP will not have any statistics. */
  if ( fstats_id == AFMNG(afmng)->N ) return;
  weights_= AFMNG(afmng)->fstats[fstats_id].weights;
  for ( n= 0; n < nframes; ++n )
    {
      t= frames[n];
      if ( (weight= weights[t]) != -HUGE_VAL )
	{
	  if ( weights_[t] == -HUGE_VAL )
	    weights_[t]= weight;
	  else
	    weights_[t]= tL_radd ( weights_[t], weight );
	}
    }
  
} /* end pseudoM_step */


static const auxfstats_mng_meths_t meths=
  {
    init,
    _free,
    alloc_auxfstats,
    free_auxfstats,
    resize_auxfstats,
    pseudoE_step,
    pseudoM_step_global,
    pseudoM_step
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

auxfstats_mng_t *
tL_create_afmng_tiedstates (
			    void  *global,
			    char **err
			    )
{
  
  afmng_tiedstates_t *new;
  unsigned int n;
  
  
  mt ( new= mem_alloc ( afmng_tiedstates_t, 1 ) );
  new->meths= &meths;
  new->afmng= tL_auxfstats_mng_new ( GLOBAL(global)->type,
                                     GLOBAL(global)->global,
				     err );
  if ( new->afmng == NULL )
    {
      free ( new );
      return NULL;
    }
  new->N= GLOBAL(global)->N;
  mt ( new->fstats= mem_alloc ( fstats_t, 1 ) );
  fstats_init ( new->fstats, (auxfstats_mng_t *) new );
  new->size= 1;
  new->currentT= 1;
  mt ( new->map= mem_alloc ( unsigned int, new->N ) );
  for ( n= 0; n < new->N; ++n )
    new->map[n]= new->N;
  mt ( new->actives= mem_alloc ( unsigned int, new->N ) );
  new->Na= 0;
  
  return (auxfstats_mng_t *) new;
  
} /* end tL_create_afmng_tiedstates */
