/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mixture.c - Implementació de 'mixture.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2011.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#include "calc.h"
#include "calcelprob.h"
#include "amodel/mixture.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "mixture.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) MIXTURE_GLOBAL(ptr)
#define STATE(ptr) MIXTURE_STATE(ptr)
#define CEP(ptr) ((mixture_calcelprob_t *) (ptr))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  CALCELPROB
  calcelprob_t *cep;
  tLProb       *calc;
  unsigned int  size;
  
} mixture_calcelprob_t;




/***********/
/* MÈTODES */
/***********/

static tLProb
calc (
      calcelprob_t        *cep,
      void                *global,
      void                *states,
      const unsigned int   q,
      const tLFea         *fea,
      const int            t
      )
{
  
  unsigned int i;
  mixture_state_t state;
  void *sglobal;
  calcelprob_t *scep;
  tLProb aux, max, *calc;
  volatile tLProb fix;
  
  
  state= STATE(states)[q];
  scep= CEP(cep)->cep;
  sglobal= GLOBAL(global)->global;
  
  /* I=1. */
  if ( state.I == 1 )
    return state.p[0] +
      scep->meths->calc ( scep, sglobal, state.members, 0, fea, t );
  
  /* I>1. Suposse que normalment les probabilitats
     de les components estaràn sauvitzades. */
  if ( state.I > CEP(cep)->size )
    {
      mt ( CEP(cep)->calc= mem_realloc ( tLProb,
					 CEP(cep)->calc,
					 state.I ) );
      CEP(cep)->size= state.I;
    }
  calc= CEP(cep)->calc;
  
  max= -HUGE_VAL;
  for ( i= 0; i < state.I; ++i )
    {
      calc[i]= aux= fix=
	state.p[i] +
	scep->meths->calc ( scep, sglobal,  state.members, i, fea, t );
      if ( aux > max ) max= aux;
    }
  
  return
    (max != -HUGE_VAL) ?
    tL_radd_vals ( calc, max, state.I ) :
    -HUGE_VAL;
  
} /* end calc */


static void
_free (
       calcelprob_t *cep
       )
{
  
  calcelprob_t *scep;
  
  
  if ( CEP(cep)->calc != NULL )
    free ( CEP(cep)->calc );
  scep= CEP(cep)->cep;
  scep->meths->free ( scep );
  free ( cep );
  
} /* end _free */


static void
init (
      calcelprob_t *cep,
      void         *global,
      const tLFea  *fea,
      const int     begin,
      const int     nframes
      )
{
  
  calcelprob_t *scep;
  
  
  scep= CEP(cep)->cep;
  scep->meths->init ( scep, GLOBAL(global)->global, fea, begin, nframes );
  
} /* end init */


static tLBool
isdensity (
	   calcelprob_t *cep
	   )
{
  
  calcelprob_t *scep;
  
  
  scep= CEP(cep)->cep;
  return scep->meths->isdensity ( scep );
  
} /* end isdensity */


static const cep_methods_t methods=
  {
    calc,
    _free,
    init,
    isdensity
  };


static tLProb
calc_max (
	  calcelprob_t        *cep,
	  void                *global,
	  void                *states,
	  const unsigned int   q,
	  const tLFea         *fea,
	  const int            t
	  )
{
  
  mmprob_t aux;
  
  
  aux= tL_Mixture_calcelprob_calcmm ( cep, global, states, q, fea, t );
  
  return aux.p;
  
} /* end calc_max */


static const cep_methods_t methods_max=
  {
    calc_max,
    _free,
    init,
    isdensity
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

calcelprob_t *
tL_alloc_mixture_calcelprob (
			     void       *global,
			     const int   flags,
			     char      **err
			     )
{
  
  mixture_calcelprob_t *new;
  
  
  mt ( new= mem_alloc ( mixture_calcelprob_t, 1 ) );
  new->cep= tL_create_calcelprob_type ( GLOBAL(global)->type,
					GLOBAL(global)->global,
					flags, err );
  if ( new->cep == NULL )
    {
      free ( new );
      return NULL;
    }
  
  if ( flags&CEP_MAX )
    {
      new->meths= &methods_max;
      new->calc= NULL;
    }
  else
    {
      new->meths= &methods;
      mt ( new->calc= mem_alloc ( tLProb, 1 ) );
      new->size= 1;
    }
  
  return (calcelprob_t *) new;
  
} /* end tL_alloc_mixture_calcelprob */


mmprob_t
tL_Mixture_calcelprob_calcmm (
			      calcelprob_t        *cep,
			      void                *global,
			      void                *states,
			      const unsigned int   q,
			      const tLFea         *fea,
			      const int            t
			      )
{
  
  unsigned int i;
  mixture_state_t state;
  void *sglobal;
  calcelprob_t *scep;
  tLProb aux;
  mmprob_t ret;
  
  
  state= STATE(states)[q];
  scep= CEP(cep)->cep;
  sglobal= GLOBAL(global)->global;
  
  ret.p= -HUGE_VAL; ret.i= 0;
  for ( i= 0; i < state.I; ++i )
    {
      aux=
	state.p[i] +
	scep->meths->calc ( scep, sglobal, state.members, i, fea, t );
      if ( aux > ret.p )
	{
	  ret.p= aux;
	  ret.i= i;
	}
    }
  
  return ret;
  
} /* end tL_Mixture_calcelprob_calcmm */
