/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  classify.c - Classificador de paraules.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011.
 *
 */

#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "classify.h"
#include "amodel.h"
#include "common.h"
#include "eprobs.h"
#include "error.h"
#include "lexicon.h"
#include "mem.h"
#include "priors.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define CALC_VITERBI_ROW(PTRANS,Q)                           \
  							     \
  tr1[begin]=							\
    tr0[begin-1] + (PTRANS) +					\
    tL_eprobs_lprob ( eprobs, amodel, fea, c, (Q), begin );	\
  for ( t= begin+1; t < end; ++t )				\
    {								\
      a= tr0[t-1] + (PTRANS);					\
      b= tr1[t-1] + p[(Q)].p0;					\
      tr1[t]=							\
	MAX ( a, b ) +						\
	tL_eprobs_lprob ( eprobs, amodel, fea, c, (Q), t );	\
    }								\
  								\
  aux= tr0;							\
  tr0= tr1;							\
  tr1= aux

#define CALC_VITERBI_BODY                 \
  					  \
  for ( q= 1; q < Q; ++q )                \
    {                                     \
      ++begin; ++end;                     \
      CALC_VITERBI_ROW ( p[q-1].p1, q );  \
    }                                     \
  prob_f= p[Q-1].p1




/*********/
/* TIPUS */
/*********/

struct tr
{
  
  tLProb *tr0;
  tLProb *tr1;
  tLProb *_mem;
  int     _length;
  
};


struct _tLClassifier
{
  
  const tLAModel     *amodel;
  const tLLexicon    *lexicon;
  const tLPriors     *priors;
  struct tr          *tr;
  size_t             *nstates;
  eprobs_t           *eprobs;
  tLBool              isdensity;
  tLClassifyCallback *callback;
  void               *udata;
  
};




/************/
/* FUNCIONS */
/************/

static struct tr *
create_tr ()
{
  
  struct tr *new;
  
  
  mt ( new= mem_alloc ( struct tr, 1 ) );
  mt ( new->_mem= mem_alloc ( tLProb, 2 ) );
  new->tr0= new->_mem;
  new->tr1= new->_mem + 1;
  new->_length= 1;
  
  return new;
  
} /* end create_tr */


static void
free_tr (
	 struct tr *tr
	 )
{
  
  free ( tr->_mem );
  free ( tr );
  
} /* end free_tr */


/* Sols is és necessari. */
static void
resize_tr (
	   struct tr *tr,
	   int        length
	   )
{
  
  mt ( tr->_mem= mem_realloc ( tLProb, tr->_mem, (length<<1) ) );
  tr->tr0= tr->_mem;
  tr->tr1= tr->_mem + length;
  tr->_length= length;
  
} /* end resize_tr */


static tLProb
viterbi (
	 const tLAModel *amodel,
	 struct tr      *tr,
	 eprobs_t       *eprobs,
	 const tLFea    *fea,
	 tLSeq          cs,
	 size_t          nstates
	 )
{
  
  int diff, begin, end, t;
  tLProb *tr0, *tr1, a, b, *aux, prob_f;
  tLProbPair *p;
  tLAModelHMM *cchar;
  size_t c, s;
  unsigned int q, Q;
  
  
  /* Prepara. */
  if ( fea->cols < nstates ) return -HUGE_VAL;
  diff= fea->cols - nstates + 1;
  tr0= tr->tr0; tr1= tr->tr1;
  begin= 0; end= diff;
  
  /* s=0 */
  c= cs.v[0];
  cchar= &(amodel->hmms[c]);
  p= cchar->trans.lineal.p;
  Q= cchar->Q;
  
  /* --> q=0 */
  tr0[0]= tL_eprobs_lprob ( eprobs, amodel,
			 fea, c, 0, 0 );
  for ( t= 1; t < diff; ++t )
    tr0[t]=
      tr0[t-1] + p[0].p0 +
      tL_eprobs_lprob ( eprobs, amodel, fea, c, 0, t );
  
  /* --> q>0 */
  CALC_VITERBI_BODY;
  
  /* s>0 */
  for ( s= 1; s < cs.length; ++s )
    {
      
      c= cs.v[s];
      cchar= &(amodel->hmms[c]);
      p= cchar->trans.lineal.p;
      Q= cchar->Q;
      
      /* --> q=0 */
      ++begin; ++end;
      CALC_VITERBI_ROW ( prob_f, 0 );
      
      /* --> q>0 */
      CALC_VITERBI_BODY;
      
    }
  
  return tr0[fea->cols-1] + prob_f;
  
} /* end viterbi */


/* Torna la probabilitat de viterbi. */
static tLProb
classify_sample_iter (
		      const tLAModel     *amodel,
		      struct tr          *tr,
		      eprobs_t           *eprobs,
		      const tLFea        *fea,
		      const size_t       *nstates,
		      tLClassHyp         *hyps,
		      const unsigned int  nhyp,
		      const tLSeq       *cs,
		      const size_t        ed,
		      const size_t        class,
		      tLProb              prob,
		      unsigned int       *inserted
		      )
{
  
  tLProb aux;
  int i;
  
  
  aux= viterbi ( amodel, tr, eprobs, fea, *cs, nstates[ed] );
  prob+= aux;
  if ( prob > hyps[nhyp-1].p )
    {
      for ( i= nhyp-1; i != 0 && hyps[i-1].p < prob; --i )
	hyps[i]= hyps[i-1];
      hyps[i].p= prob;
      hyps[i].c= class;
      hyps[i].cs= cs;
      ++(*inserted);
    }
  
  return aux;
  
} /* end classify_sample_iter */


static size_t *
create_nstates_priors (
		       const tLLexicon *lexicon,
		       const tLAModel  *amodel
		       )
{
  
  size_t *ret, n, t;
  tLSeq cs;
  
  
  mt ( ret= mem_alloc ( size_t, lexicon->N+1 ) );
  for ( n= 1; n <= lexicon->N; ++n )
    {
      ret[n]= 0;
      cs= *(lexicon->entries[n].cs);
      for ( t= 0; t < cs.length; ++t )
	ret[n]+= amodel->hmms[cs.v[t]].Q;
    }
  
  return ret;
  
} /* end create_nstates_priors */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_classifier_free (
		    tLClassifier *cls
		    )
{
  
  free ( cls->nstates );
  free_tr ( cls->tr );
  tL_free_eprobs ( cls->eprobs );
  free ( cls );
  
} /* end aux_classifier_free */


TL_API_FUNC(tLClassifier *)
tL_classifier_new (
		   const tLAModel      *amodel,
		   const tLLexicon     *lexicon,
		   const tLPriors      *priors,
		   tLClassifyCallback  *callback,
		   void                *udata,
		   char               **err
		   )
{
  
  tLClassifier *new;
  
  
  if ( amodel->tflags != FLINEAL )
    {
      tL_msgerror ( err, _( "tL_classifier_new is only implemented"
                    " for character models with linear topologies" ) );
      return NULL;
    }
  mt ( new= mem_alloc ( tLClassifier, 1 ) );
  new->amodel= amodel;
  new->lexicon= lexicon;
  new->priors= priors;
  new->eprobs= tL_create_eprobs ( amodel, err );
  if ( new->eprobs == NULL )
    {
      free ( new );
      return NULL;
    }
  new->tr= create_tr ();
  new->isdensity= isdensity_calcelprob ( new->eprobs->cep );
  new->nstates= create_nstates_priors ( lexicon, amodel );
  new->callback= callback;
  new->udata= udata;
  
  return new;
  
} /* end tL_classifier_new */


TL_API_FUNC(unsigned int)
tL_classify (
	     tLClassifier       *cls,
	     const tLFea        *fea,
	     tLClassHyp         *hyps,
	     const unsigned int  N
	     )
{
  
  unsigned int i, ret;
  size_t *init, n, ed;
  tLLexiconEntry *entries;
  tLProb prob, prob2, aux2;
  
  
  if ( fea->cols > cls->tr->_length ) resize_tr ( cls->tr, fea->cols );
  if ( fea->cols > cls->eprobs->length )
    tL_resize_eprobs ( cls->eprobs, cls->amodel, fea->cols );
  for ( i= 0; i < N; ++i )
    {
      hyps[i].p= -HUGE_VAL;
      hyps[i].c= 0;
      hyps[i].cs= NULL;
    }
  
  /* Classifica. */
  ret= 0;
  init_eprobs ( cls->eprobs, cls->amodel, fea );
  init= cls->lexicon->init;
  entries= cls->lexicon->entries;
  for ( n= 0; n < cls->lexicon->words->N; ++n )
    if ( (prob= cls->priors[n]) != -HUGE_VAL &&
	 (cls->isdensity || prob > hyps[N-1].p) )
      for ( ed= init[n]; ed != NULL_ENTRY; ed= entries[ed].next )
	if ( (prob2= prob+entries[ed].p) != -HUGE_VAL &&
	     (cls->isdensity || prob2 > hyps[N-1].p) )
	  {
	    aux2= classify_sample_iter ( cls->amodel, cls->tr, cls->eprobs, fea,
					 cls->nstates, hyps, N, entries[ed].cs,
					 ed, n, prob2, &ret );
	    if ( cls->callback != NULL )
	      cls->callback ( cls->udata, cls->priors[n], entries[ed].p,
			      aux2, n, entries[ed].cs );
	  }
  
  tL_reset_eprobs ( cls->eprobs, cls->amodel, fea->cols );
  
  return ret>N?N:ret;
  
} /* end tL_classify */
