/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  classify.h - Classificador de paraules.
 *
 *  Adrià Giménez Pastor, 2008-2011.
 *
 */

#ifndef __LIBTL_CLASSIFY_H__
#define __LIBTL_CLASSIFY_H__

#include <stddef.h>

#include "common_public.h"
#include "amodel_public.h"
#include "lexicon_public.h"
#include "priors.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup Classify Classification of samples
 */
/** \addtogroup Classify
 *  @{
 */

#ifndef __TL_PRV
/** Classifier. This structure contains auxiliary data used by
 *  'tL_classify' to classify a given sample. The values of model
 *  parameters can be modified, but not the topology.
 */
struct _tLClassifier
{
  
  const tLAModel  *amodel;     /**< Reference to the acoustic model
				  used by the classifier. */
  const tLLexicon *lexicon;    /**< Reference to the lexicon used by
				  the classifier. */
  const tLPriors  *priors;     /**< Reference to the prior
				  probabilities used by the
				  classifier. */
  
};
#endif

/** Classifier. This structure contains auxiliary data used by
 *  'tL_classify' to classify a given sample. The values of model
 *  parameters can be modified, but not the topology.
 */
typedef struct _tLClassifier tLClassifier;

/** Frees memory. Frees the memory allocated for the classifier.
 *
 *  @param cls The classifier.
 */
TL_API_FUNC(void)
tL_classifier_free (
		    tLClassifier *cls
		    );

/** Callback function for tL_classify. This callback is used to
 *  capture probability \f$p(c)p(e\mid c)p(x\mid e)\f$ for every
 *  symbol sequence.
 */
typedef void
(tLClassifyCallback) (
		      void         *udata,           /**< User
							data. */
		      tLProb        prob_class,      /**< \f$p(c)\f$. */
		      tLProb        prob_lexicon,    /**< \f$p(e\mid c)\f$. */
		      tLProb        prob_sample,     /**< \f$p(x\mid e)\f$. */
		      size_t        class,           /**< The class
							identifier. */
		      const tLSeq *cs               /**< The
							symbol
							sequence. */
		      );

/** Gets a Classifier. This function creates a classifier used to
 *  classify samples. Currently supported types are: Bernoulli,
 *  DGaussian, Mixture and TiedStates.
 *
 *  @param amodel The acoustic model used to classify.
 *  @param lexicon The lexicon used to classify.
 *  @param priors The prior probabilities used to classify.
 *  @param callback A callback function used to capture the calculated
 *  probabilities during the classification process. It can be NULL.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @param udata User data passed to the callback function.
 *  @return Classifier, or NULL in case of error.
 */
TL_API_FUNC(tLClassifier *)
tL_classifier_new (
		   const tLAModel      *amodel,
		   const tLLexicon     *lexicon,
		   const tLPriors      *priors,
		   tLClassifyCallback  *callback,
		   void                *udata,
		   char               **err
		   );

/** Classification hypothesis. This structure is used to store a
 *  classification hypothesis.
 */
typedef struct
{
  
  tLProb        p;     /**< Probability. */
  size_t        c;     /**< Class identifier. */
  const tLSeq *cs;    /**< Pointer to the symbol sequence. */
  
} tLClassHyp;

/** Classifies a sample. This function classifies the given sample,
 *  feature vector sequence, according to the next expression:
 *  \f$(c,e)^*=\arg\max_{(c,e)} p(c)p(e\mid c)p(x\mid e)\, \f$ where
 *  \f$x\f$ is the given sample and \f$(c,e)\f$ is a valid pair of
 *  symbol sequence and class. The probability \f$p(x\mid e)\f$ is
 *  approximated using the Viterbi algorithm. This function can be
 *  used to obtain the N most probable \f$(c,e)\f$ pairs or
 *  hypotheses. The variable hyps must contain enough memory to
 *  allocate N hypotheses.
 *
 *  @param cls The classifier.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param hyps An array with enough memory to allocate the required
 *  hypotheses.
 *  @param N Number of required hypotheses.
 *  @return The number of hypotheses returned.
 */
TL_API_FUNC(unsigned int)
tL_classify (
	     tLClassifier       *cls,
	     const tLFea        *fea,
	     tLClassHyp         *hyps,
	     const unsigned int  N
	     );

/** @} */
/** @} */

#endif /* __LIBTL_CLASSIFY_H__ */
