/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  cmllr.c - Implementació de 'cmllr.h'.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */


#include <assert.h>
#include <cblas.h>
#include <clapack.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "logamcounts.h"
#include "amcounts/dgaussian.h"
#include "amcounts/logmixture.h"
#include "amcounts/logtiedstates.h"
#include "cmllr.h"
#include "cmllr_alg.h"
#include "cmllr_alg_full.h"
#include "amodel/dgaussian.h"
#include "amodel/mixture.h"
#include "amodel/tiedstates.h"
#include "common.h"
#include "error.h"
#include "mem.h"




/**********/
/* MACROS */
/**********/

#define DGAUSSIAN_COUNT(ptr) ((amcount_dgaussian_state_t *) (ptr))
#define MIXTURE_COUNT(ptr) ((logamcount_mixture_state_t *) (ptr))




/*********/
/* TIPUS */
/*********/

/* Transformació W. */
typedef struct
{
  
  double **W;
  double **A;
  double  *b;
  
} W_t;


/* Per a saber cada transformació a que estats està associat. */
typedef struct
{
  
  unsigned int c;   /* Si és tied serà 0. */
  unsigned int q;
  
} map_node_t;


/* Els estats de cada transformació i els compters i paràmetres
   necessaris pera a adaptar. */
typedef struct
{
  
  map_node_t               *nodes;
  unsigned int              N;
  cmllr_alg_paramscounts_t  pcs;
  
} map_t;


struct _tLCMLLR
{
  
  const tLAModel         *amodel;
  enum { TIED, NORMAL }   type;
  unsigned int            D;
  const tLFloat          *smooth;
  unsigned int            N;         /* Número de transformacions. */
  map_t                  *map;       /* Els estats relacionats amb
					cada transformació. */
  unsigned int          **imap;      /* |C|*|Q|. Mapa invers. En cas
                                         de Tied c és 0. ARA MATEIXA
                                         ÉS INÚTIL!!!!  */
  W_t                    *Ws;        /* |N+1|. L'últim es gasta com
                                         auxiliar.  */
  int                     mlevel;    /* Número de nivells Mixture. */
  cmllr_alg_data_t        data;
  cmllr_alg_full_data_t   data_full;
  double                 *aux_A;     /* D*D. Té diverses utilitats. */
  int                    *ipiv;      /* |D| */
  double                 *in,*out;
  
};




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* Els nodes tenen que estar inicialitzats. */
static void
alloc_pcs (
	   map_t          *map,
	   const tLAModel *amodel,
	   const int       type,
	   const int       mlevel
	   )
{
  
  unsigned int n, I, i;
  map_node_t node;
  mixture_state_t *state;
  tiedstates_global_t *global;
  
  
  /* Conta estats. */
  map->pcs.I= 0;
  if ( type == NORMAL )
    for ( n= 0; n < map->N; ++n )
      {
	node= map->nodes[n];
	if ( mlevel == 1 )
	  {
	    state= &(MIXTURE_STATE(amodel->hmms[node.c].states)[node.q]);
	    map->pcs.I+= state->I;
	  }
	else ++(map->pcs.I);
      }
  else /* TIED */
    {
      global= TIEDSTATES_GLOBAL(amodel->global);
      for ( n= 0; n < map->N; ++n )
	{
	  node= map->nodes[n];
	  if ( mlevel == 1 )
	    {
	      state= &(MIXTURE_STATE(global->states)[node.q]);
	      map->pcs.I+= state->I;
	    }
	  else ++(map->pcs.I);
	}
    }
  
  /* Reserva memòria. */
  mt ( map->pcs.params= mem_alloc ( dgaussian_state_t *, map->pcs.I ) );
  mt ( map->pcs.counts= mem_alloc ( amcount_dgaussian_state_t *, map->pcs.I ) );
  
  /* Enllaça paràmetres */
  I= 0;
  if ( type == NORMAL )
    for ( n= 0; n < map->N; ++n )
      {
	node= map->nodes[n];
	if ( mlevel == 1 )
	  {
	    state= &(MIXTURE_STATE(amodel->hmms[node.c].states)[node.q]);
	    for ( i= 0; i < state->I; ++i, ++I )
	      map->pcs.params[I]= &(DGAUSSIAN_STATE(state->members)[i]);
	  }
	else map->pcs.params[I++]=
	       &(DGAUSSIAN_STATE(amodel->hmms[node.c].states)[node.q]);
      }
  else /* TIED */
    {
      global= TIEDSTATES_GLOBAL(amodel->global);
      for ( n= 0; n < map->N; ++n )
	{
	  node= map->nodes[n];
	  if ( mlevel == 1 )
	    {
	      state= &(MIXTURE_STATE(global->states)[node.q]);
	      for ( i= 0; i < state->I; ++i, ++I )
		map->pcs.params[I]= &(DGAUSSIAN_STATE(state->members)[i]);
	    }
	  else map->pcs.params[I++]=
		 &(DGAUSSIAN_STATE(global->states)[node.q]);
	}
    }
  
} /* end alloc_pcs */


/* Els nodes tenen que estar inicialitzats. */
static void
pcs_set_counts (
		map_t               *map,
		const tLAModel      *amodel,
		const tLAMlogCounts *counts,
		const int            type,
		const int            mlevel
		)
{
  
  unsigned int n, I, i;
  map_node_t node;
  mixture_state_t *state;
  logamcount_mixture_state_t *cstate;
  tiedstates_global_t *global;
  logamcount_tiedstates_global_t *cglobal;
  
  
  /* Enllaça compters. */
  I= 0;
  if ( type == NORMAL )
    for ( n= 0; n < map->N; ++n )
      {
	node= map->nodes[n];
	if ( mlevel == 1 )
	  {
	    state= &(MIXTURE_STATE(amodel->hmms[node.c].states)[node.q]);
	    cstate= &(MIXTURE_COUNT(counts->hmms[node.c].states)[node.q]);
	    for ( i= 0; i < state->I; ++i, ++I )
	      map->pcs.counts[I]= &(DGAUSSIAN_COUNT(cstate->members)[i]);
	  }
	else map->pcs.counts[I++]=
	       &(DGAUSSIAN_COUNT(counts->hmms[node.c].states)[node.q]);
      }
  else /* TIED */
    {
      global= TIEDSTATES_GLOBAL(amodel->global);
      cglobal= (logamcount_tiedstates_global_t *) counts->global;
      for ( n= 0; n < map->N; ++n )
	{
	  node= map->nodes[n];
	  if ( mlevel == 1 )
	    {
	      state= &(MIXTURE_STATE(global->states)[node.q]);
	      cstate= &(MIXTURE_COUNT(cglobal->states)[node.q]);
	      for ( i= 0; i < state->I; ++i, ++I )
		map->pcs.counts[I]= &(DGAUSSIAN_COUNT(cstate->members)[i]);
	    }
	  else map->pcs.counts[I++]=
		 &(DGAUSSIAN_COUNT(cglobal->states)[node.q]);
	}
    }
  
} /* end pcs_set_counts */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. */
static int
check_type_and_set_values (
			   tLCMLLR   *cmllr,
			   tLAModel  *amodel,
			   char     **err
			   )
{
  
  void *global;
  const char *type;
  unsigned int *mem, n, c, q;
  
  
  if ( !strcmp ( "TiedStates", amodel->type ) )
    {
      global= TIEDSTATES_GLOBAL(amodel->global)->global;
      type= TIEDSTATES_GLOBAL(amodel->global)->type;
      cmllr->type= TIED;
      mt ( cmllr->imap= mem_alloc ( unsigned int *, 1 ) );
      mt ( cmllr->imap[0]= mem_alloc ( unsigned int,
				       TIEDSTATES_GLOBAL(amodel->global)->N ) );
      /* HARDCODED MAPPING !!!!! *********************************************/
      mt ( cmllr->map= mem_alloc ( map_t, 1 ) );
      mt ( cmllr->map[0].nodes=
	   mem_alloc ( map_node_t, TIEDSTATES_GLOBAL(amodel->global)->N ) );
      for ( n= 0; n < TIEDSTATES_GLOBAL(amodel->global)->N; ++n )
	{
	  cmllr->map[0].nodes[n].c= 0;
	  cmllr->map[0].nodes[n].q= n;
	}
      cmllr->map[0].N= TIEDSTATES_GLOBAL(amodel->global)->N;
      cmllr->N= 1;
      for ( n= 0; n < TIEDSTATES_GLOBAL(amodel->global)->N; ++n )
	cmllr->imap[0][n]= 0;
      /***********************************************************************/
    }
  else
    {
      global= amodel->global;
      type= amodel->type;
      cmllr->type= NORMAL;
      mt ( cmllr->imap= mem_alloc ( unsigned int *, amodel->N ) );
      mt ( mem= mem_alloc ( unsigned int, amodel->total_Q ) );
      for ( n= 0; n < amodel->N; ++n )
	{
	  cmllr->imap[n]= mem;
	  mem+= amodel->hmms[n].Q;
	}
      /* HARDCODED MAPPING !!!!! *********************************************/
      mt ( cmllr->map= mem_alloc ( map_t, 1 ) );
      mt ( cmllr->map[0].nodes=
	   mem_alloc ( map_node_t, TIEDSTATES_GLOBAL(amodel->global)->N ) );
      for ( n= c= 0; c < amodel->N; ++c )
	for ( q= 0; q < amodel->hmms[c].Q; ++q, ++n )
	  {
	    cmllr->map[0].nodes[n].c= c;
	    cmllr->map[0].nodes[n].q= q;
	  }
      cmllr->map[0].N= amodel->total_Q;
      cmllr->N= 1;
      for ( n= 0; n < amodel->total_Q; ++n ) cmllr->imap[0][n]= 0;
      /***********************************************************************/
    }
  
  cmllr->mlevel= 0;
  while ( !strcmp ( "Mixture", type ) )
    {
      type= MIXTURE_GLOBAL(global)->type;
      global= MIXTURE_GLOBAL(global)->global;
      ++(cmllr->mlevel);
    }
  /* Açò de moment no es possible per que a l'algorisme per culpa de
     'tL_cmllr_update', però si fora possible es podria canviar. */
  if ( cmllr->mlevel > 1 )
    {
      tL_msgerror ( err, _( "nested Mixtures are not supported" ) );
      return -1;
    }
  if ( strcmp ( "DGaussian", type ) )
    {
      if ( cmllr->type == TIED && !strcmp ( "TiedStates", type ) )
	tL_msgerror ( err, _( "nested TiedStates are not supported" ) );
      else tL_msgerror ( err, _( "type %s is not supported" ) , type );
      return -1;
    }
  cmllr->D= DGAUSSIAN_GLOBAL(global)->D;
  cmllr->smooth= DGAUSSIAN_GLOBAL(global)->smooth;
  
  /* Reserva pcs. */
  for ( n= 0; n < cmllr->N; ++n )
    alloc_pcs ( &(cmllr->map[n]), amodel, cmllr->type, cmllr->mlevel );
  
  return 0;
  
} /* end check_type_and_set_values */


static void
free_Ws (
	 W_t            *Ws,
	 const unsigned  N
	 )
{
  
  unsigned int n;
  
  
  for ( n= 0; n <= N; ++n )
    {
      free ( Ws[n].W[0] );
      free ( Ws[n].W );
      free ( Ws[n].A );
    }
  free ( Ws );
  
} /* end free_Ws */


static W_t *
alloc_Ws (
	  const tLAModel     *amodel,
	  const int           type,
	  const unsigned int  D,
	  const unsigned int  N
	  )
{
  
  W_t *new;
  unsigned int n, d;
  double *aux;
  
  
  mt ( new= mem_alloc ( W_t, N+1 ) );
  for ( n= 0; n <= N; ++n )
    {
      mt ( new[n].W= mem_alloc ( double *, D+1 ) );
      mt ( new[n].A= mem_alloc ( double *, D ) );
      mt ( aux= mem_alloc ( double, (D+1)*D ) );
      new[n].W[0]= new[n].b= aux; aux+= D;
      for ( d= 0; d < D; ++d, aux+= D )
	new[n].W[d+1]= new[n].A[d]= aux;
    }
  
  return new;
  
} /* end alloc_Ws */


static void
print_W (
	 const W_t           W,
	 const unsigned int  D,
	 FILE               *to
	 )
{
  
  unsigned int d, d2;
  const double *A_d;
  
  
  for ( d= 0; d < D; ++d )
    {
      A_d= W.A[d];
      fprintf ( to, "%.15f", A_d[0] );
      for ( d2= 1; d2 < D; ++d2 ) fprintf ( to, " %.15f", A_d[d2] );
      putc ( '\n', to );
    }
  fprintf ( to, "%.15f", W.b[0] );
  for ( d2= 1; d2 < D; ++d2 ) fprintf ( to, " %.15f", W.b[d2] );
  putc ( '\n', to );
  
} /* end print_W */


static void
init_W (
	W_t                *W,
	const unsigned int  D
	)
{
  
  unsigned int d, d2;
  double *A_d;
  
  
  for ( d= 0; d < D; ++d )
    {
      A_d= W->A[d];
      for ( d2= 0; d2 < D; ++d2 )
	A_d[d2]= (d==d2)?1.0:0.0;
    }
  for ( d= 0; d < D; ++d ) W->b[d]= 0.0;
  
} /* end init_W */







/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_cmllr_adapt_fea (
		    tLCMLLR         *cmllr,
		    const tLFea     *fea_in,
		    tLFea           *fea_out
		    )
{
  
  int t;
  W_t *W;
  unsigned int D, d;
  float **frames_in, **frames_out, *frames_in_t, *frames_out_t;
  double *in, *out;
  
  
  assert ( fea_in->rows == cmllr->D );
  assert ( cmllr->N == 1 );
  frames_in= tL_fea_get_real_data ( fea_in );
  tL_fea_resize ( fea_out, fea_in->rows, fea_in->cols );
  frames_out= tL_fea_get_real_data ( fea_out );
  in= cmllr->in; out= cmllr->out;
  D= cmllr->D;
  /* Açò és prou inútil ara mateix, però si en el futur hi haguera que
     tornar a múltiples traswnforamcions sería útil. */
  /*
  if ( cmllr->type == TIED )
    {
      state= TIEDSTATES_STATE(cmllr->amodel->hmms[seg->c].states)[seg->q];
      W= &(cmllr->Ws[cmllr->imap[0][state]]);
    }
  else W= &(cmllr->Ws[cmllr->imap[seg->c][seg->q]]);
  */
  W= &(cmllr->Ws[0]);
  for ( t= 0; t < fea_in->cols; ++t )
    {
      frames_in_t= frames_in[t];
      frames_out_t= frames_out[t];
      for ( d= 0; d < D; ++d ) in[d]= (double) frames_in_t[d];
      cblas_dcopy ( D, W->b, 1, out, 1 );
      cblas_dgemv ( CblasColMajor, CblasNoTrans, D, D,
		    1.0, W->A[0], D, in, 1, 1.0, out, 1 );
      for ( d= 0; d < D; ++d ) frames_out_t[d]= (float) out[d];
    }
  
} /* end tL_cmllr_adapt_fea */


TL_API_FUNC(void)
tL_cmllr_free (
	       tLCMLLR *cmllr
	       )
{

  unsigned int n;
  
  
  free ( cmllr->aux_A );
  free ( cmllr->in );
  free ( cmllr->out );
  free ( cmllr->ipiv );
  if ( cmllr->data.newmu != NULL ) tL_cmllr_alg_data_free ( &(cmllr->data) );
  if ( cmllr->data_full.G != NULL )
    tL_cmllr_alg_full_data_free ( &(cmllr->data_full) );
  for ( n= 0; n < cmllr->N; ++n )
    {
      free ( cmllr->map[n].nodes );
      free ( cmllr->map[n].pcs.params );
      free ( cmllr->map[n].pcs.counts );
    }
  free ( cmllr->map );
  free ( cmllr->imap[0] );
  free ( cmllr->imap );
  free_Ws ( cmllr->Ws, cmllr->N );
  free ( cmllr );
  
} /* end tL_cmllr_free */


TL_API_FUNC(tLFloat)
tL_cmllr_get_lv_constant (
			  const tLCMLLR       *cmllr,
			  const tLAMlogCounts *counts
			  )
{
  
  tLFloat beta;
  double logA, *ptr;
  unsigned int d, i, D;
  int info, counter;
  const cmllr_alg_paramscounts_t *pcs;
  
  
  /* Càlcul log(|A|). */
  D= cmllr->D;
  assert ( cmllr->N == 1 );
  memcpy ( cmllr->aux_A, cmllr->Ws[0].A[0], D*D*sizeof(double) );
  info= clapack_dgetrf ( CblasColMajor, D, D, cmllr->aux_A, D, cmllr->ipiv );
  if ( info != 0.0 )
    {
      fprintf ( stderr, _( "Warning: unable to calculated |A|"
		" in tL_mllr_get_lv_constant" ) );
      return 0.0;
    }
  logA= 1.0;
  for ( d= 0, ptr= cmllr->aux_A; d < D; ++d, ptr+= (D+1) )
    logA*= *ptr;
  for ( d= counter= 0; d < D; ++d )
    if ( cmllr->ipiv[d]!=d ) ++counter;
  
  logA= (counter%2)?-logA:logA;
  logA= log ( logA*logA );
  
  /* Calcula constant. */
  pcs_set_counts ( &(cmllr->map[0]), cmllr->amodel,
		   counts, cmllr->type, cmllr->mlevel );
  pcs= &(cmllr->map[0].pcs);
  beta= 0;
  for ( i= 0; i < pcs->I; ++i ) beta+= pcs->counts[i]->norm;
  
  return 0.5*logA*beta;
  
} /* end tL_cmllr_get_lv_constant */


TL_API_FUNC(void)
tL_cmllr_init (
	       tLCMLLR *cmllr
	       )
{
  
  unsigned int n;
  W_t *ms;
  
  
  ms= cmllr->Ws;
  for ( n= 0; n <= cmllr->N; ++n )
    init_W ( &(ms[n]), cmllr->D );
  
} /* end tL_cmllr_init */


TL_API_FUNC(tLCMLLR *)
tL_cmllr_new (
	      tLAModel  *amodel,
	      char     **err
	      )
{
  
  tLCMLLR *new;
  
  
  mt ( new= mem_alloc ( tLCMLLR, 1 ) );
  new->amodel= amodel;
  if ( check_type_and_set_values ( new, amodel, err ) != 0 )
    {
      free ( new );
      return NULL;
    }
  new->Ws= alloc_Ws ( amodel, new->type, new->D, new->N );
  new->data.newmu= NULL;
  new->data_full.G= NULL;
  tL_cmllr_init ( new );
  mt ( new->aux_A= mem_alloc ( double, new->D*new->D ) );
  mt ( new->ipiv= mem_alloc ( int, new->D ) );
  mt ( new->in= mem_alloc ( double, new->D ) );
  mt ( new->out= mem_alloc ( double, new->D ) );
  
  return new;
  
} /* end tL_cmllr_new */


TL_API_FUNC(void)
tL_cmllr_print_matrix (
		       const tLCMLLR *cmllr,
		       FILE          *to
		       )
{
  
  unsigned int n;
  
  
  for ( n= 0; n < cmllr->N; ++n )
    {
      fprintf ( to, "%lu\n", _PS(n) );
      print_W ( cmllr->Ws[n], cmllr->D, to );
    }
  
} /* end tL_cmllr_print_matrix */


TL_API_FUNC(int)
tL_cmllr_read_matrix (
		      const tLCMLLR  *cmllr,
		      gzFile          from,
		      char          **err
		      )
{
  
  tLBuffer *buffer;
  unsigned int n, d, d2;
  
  
  buffer= tL_buffer_new ();
  for ( n= 0; n < cmllr->N; ++n )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, _( "Unable to read the label of the"
			" %uth CMLLR transformation" ) , n+1);
	  goto error;
	}
      for ( d= 0; d < cmllr->D; ++d )
	for ( d2= 0; d2 < cmllr->D; ++d2 )
	  {
	    if ( !tL_gtoken ( from, buffer ) )
	      {
		tL_msgerror ( err, _( "Unable to read A[%u][%u][%u]" ) , n, d, d2 );
		goto error;
	      }
	    cmllr->Ws[n].A[d][d2]= atof ( buffer->v );
	  }
      for ( d= 0; d < cmllr->D; ++d )
	{
	  if ( !tL_gtoken ( from, buffer ) )
	    {
	      tL_msgerror ( err, _( "Unable to read b[%u][%u][%u]" ) , n, d );
	      goto error;
	    }
	  cmllr->Ws[n].b[d]= atof ( buffer->v );
	}
    }
  tL_buffer_free ( buffer );
  
  return 0;
  
 error:
  tL_buffer_free ( buffer );
  return -1;
  
} /* end tL_cmllr_read_matrix */


TL_API_FUNC(void)
tL_cmllr_update_matrix (
			tLCMLLR             *cmllr,
			const tLAMlogCounts *counts,
			const unsigned int   niters
			)
{
  
  unsigned int n, D;
  void *global;
  char *err;
  int ret;
  W_t *aux_W;
  
  
  D= cmllr->D;
  if ( niters == 0 )
    {
      if ( cmllr->data.newmu == NULL )
	tL_cmllr_alg_data_alloc ( &(cmllr->data), D );
    }
  else
    {
      if ( cmllr->data_full.G == NULL )
	tL_cmllr_alg_full_data_alloc ( &(cmllr->data_full), D );
    }
  global= (cmllr->type==NORMAL) ?
    cmllr->amodel->global :
    TIEDSTATES_GLOBAL(cmllr->amodel->global)->global;
  if ( cmllr->mlevel == 1 ) global= MIXTURE_GLOBAL(global)->global;
  aux_W= &(cmllr->Ws[cmllr->N]);
  for ( n= 0; n < cmllr->N; ++n )
    {
      pcs_set_counts ( &(cmllr->map[n]), cmllr->amodel,
		       counts, cmllr->type, cmllr->mlevel );
      memcpy ( cmllr->aux_A, cmllr->Ws[n].A[0], sizeof(double)*D*D);
      if ( niters == 0 )
	ret= tL_cmllr_alg ( &(cmllr->map[n].pcs), DGAUSSIAN_GLOBAL(global),
			      aux_W->W, &(cmllr->data), &err );
	  
      else
	{
	  init_W ( aux_W, D );
	  ret= tL_cmllr_alg_full ( &(cmllr->map[n].pcs),
				   DGAUSSIAN_GLOBAL(global),
				   niters, aux_W->W,
				   &(cmllr->data_full), &err );
	  
	}
      if ( ret == -1 )
	{
	  fprintf ( stderr, _( "Warning: an error has occurred while"
		    " calculating the %uth CMLLR transformation matrix: %s\n" ) ,
		    n+1, err );
	  free ( err );
	}
      cblas_dgemm ( CblasColMajor, CblasNoTrans, CblasNoTrans, D, D, D,
		    1.0, aux_W->A[0], D, cmllr->aux_A, D,
		    0.0, cmllr->Ws[n].A[0], D );
      cblas_dgemv ( CblasColMajor, CblasNoTrans, D, D,
		    1.0, aux_W->A[0], D, cmllr->Ws[n].b, 1,
		    1.0, aux_W->b, 1 );
      memcpy ( cmllr->Ws[n].b, aux_W->b, sizeof(double)*D );
    }
  
} /* end tL_cmllr_update_matrix */
