/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  cmllr.h - CMLLR.
 *
 *  Adrià Giménez Pastor, 2012.
 *
 */

#ifndef __LIBTL_CMLLR_H__
#define __LIBTL_CMLLR_H__

#include <stdio.h>
#include <stddef.h>
#include <zlib.h>

#include "align.h"
#include "amcounts_public.h"
#include "amodel_public.h"
#include "samples_public.h"
#include "text_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup CMLLR CMLLR adaptation
 */
/** \addtogroup CMLLR
 *  @{
 */

#ifndef __TL_PRV
/** Used in CMLLR functions. This structure contains auxiliary data
 *  used to adapt models using the CMLLR technique.
 */
struct _tLCMLLR
{
  
  const tLAModel *amodel;    /**< Reference to the acoustic model
				used by the CMLLR adapter. */
  
};
#endif

/** Used in CMLLR functions. This structure contains auxiliary data
 *  used to adapt models using the CMLLR technique.
 */
typedef struct _tLCMLLR tLCMLLR;

/** Adapt a feature sequence. This function applies the CMLLR
 *  transformation matrix to the feature sequence stored in the
 *  provided feature sequence manager. The feature sequence is
 *  assumed to be real and to have the same dimension as the
 *  acoustic model related to the CMLLR structure.
 *
 *  @param cmllr The CMLLR structure.
 *  @param fea_in A feature sequence manager containing the real
 *  feature sequence to adapt.
 *  @param fea_out The feature sequence manager to which the
 *  transformed sequence will be stored.
 */
TL_API_FUNC(void)
tL_cmllr_adapt_fea (
		    tLCMLLR     *cmllr,
		    const tLFea *fea_in,
		    tLFea       *fea_out
		    );

/** Frees memory. Frees the memory allocated for the CMLLR structure.
 *
 *  @param cmllr The CMLLR structure.
 */
TL_API_FUNC(void)
tL_cmllr_free (
	       tLCMLLR *cmllr
	       );

/** Gets the additional \f$\frac{1}{2}\sum_m\sum_t\log{(|A|)}\f$
 *  term. This function calculates the additional term
 *  \f$\frac{1}{2}\sum_m\sum_t\log{(|A|)}\f$, which must be added to
 *  log-likelihood.
 *
 *  @param cmllr The CMLLR structure.
 *  @param counts Logarithmic counts related to the acoustic model
 *  of the CMLLR structure.
 *  @return The additional term.
 */
TL_API_FUNC(tLFloat)
tL_cmllr_get_lv_constant (
			  const tLCMLLR       *cmllr,
			  const tLAMlogCounts *counts
			  );

/** Initializes the CMLLR transformation matrix. This function
 *  initializes the CMLLR transformation matrix. The A matrix is
 *  initialized as the identity matrix, and the b vector is
 *  initialized with zeroes.
 *
 *  @param cmllr The CMLLR structure.
 */
TL_API_FUNC(void)
tL_cmllr_init (
	       tLCMLLR *cmllr
	       );

/** Get a tLCMLLR. This function creates a CMLLR structure used to
 *  adapt models and transform features using the CMLLR
 *  technique. Only DGaussian based types are supported. This function
 *  calls 'tL_cmllr_init'.
 *
 *  @param amodel The acoustic model.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return CMLLR structure, or NULL in case of error.
 */
TL_API_FUNC(tLCMLLR *)
tL_cmllr_new (
	      tLAModel  *amodel,
	      char     **err
	      );

/** Prints the transformation matrix. This function prints the
 *  CMLLR transformation matrix to the specified file.
 *
 *  @param cmllr The CMLLR structure.
 *  @param to File to which the matrix will be written.
 */
TL_API_FUNC(void)
tL_cmllr_print_matrix (
		       const tLCMLLR *cmllr,
		       FILE          *to
		       );

/** Reads the transformation matrix. This function reads the
 *  CMLLR transformation matrix from the specified file.
 *
 *  @param cmllr The CMLLR structure.
 *  @param from File where the matrix is stored.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return -1 if error, 0 if success.
 */
TL_API_FUNC(int)
tL_cmllr_read_matrix (
		      const tLCMLLR  *cmllr,
		      gzFile          from,
		      char          **err
		      );

/** Update the transformation matrix. This function recalculates the
 *  CMLLR transformation matrix, using the previous values and the
 *  provided logarithmic counts. When niters is 0, diagonal CMLLR
 *  matrix is calculated. When full matrix is calculated, counts
 *  must be created using 'tL_cmlogcounts_new_cmllr'.
 *
 *  @param cmllr The CMLLR structure.
 *  @param counts Logarithmic counts related to the acoustic model
 *  of the CMLLR structure. The counts must be created using
 *  'tL_cmlogcounts_new_cmllr' if niters>0.
 *  @param niters The number of iterations for the CMLLR, or 0 to
 *  indicate that diagonal CMLLR matrix must be calculated.
 *  algorithm. Must be greater than 0.
 */
TL_API_FUNC(void)
tL_cmllr_update_matrix (
			tLCMLLR             *cmllr,
			const tLAMlogCounts *counts,
			const unsigned int   niters
			);

/** @} */
/** @} */

#endif /* __LIBTL_CMLLR_H__ */
