/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  cmllr_alg.c - Algorisme CMLLR.
 *
 *  Miguel Ángel del Agua Teba.
 */

#include <assert.h>
#include <stddef.h>
#include <stdlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV

#include "amcounts/dgaussian.h"
#include "amodel/dgaussian.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "cmllr_alg.h"
#include <math.h>
#include <cblas.h>
#include <clapack.h>

#define ABS(a) (((a) < 0) ? -(a) : (a))


void tL_cmllr_alg_data_alloc(
			     cmllr_alg_data_t         *mem_vars,
			     uint                      D
			     )
{
  
  mt ( mem_vars->newmu= mem_alloc ( double, 1 ) );
  mt ( mem_vars->newvar= mem_alloc ( double, 1 ) );
  mem_vars->size= 1;
  
}

void
tL_cmllr_alg_data_free (
			cmllr_alg_data_t *mem_vars
			)
{
  
  free ( mem_vars->newmu );
  free ( mem_vars->newvar );
  
}

/* ****************************************************************************/

static double
getA(
     amcount_dgaussian_state_t **dgcounts,  /* Cuentas de la
					       gaussiana */
     const uint                  I
     )
{
  uint i;
  double result;
  
  result= 0.0;
  for ( i= 0; i < I; ++i ) result+=dgcounts[i]->norm; 
  
  return result;
}

static double
getB(
     amcount_dgaussian_state_t **dgcounts,  /* Cuentas de la
					       gaussiana */
     dgaussian_state_t         **dgstate,   /* Componentes de mix de
					       Gaussianas del
					       estado */
     const uint                   I,
     const uint                   d
     )
{
  uint i;
  double result;

  result= 0.0;
  for ( i= 0; i < I; ++i ) result+=dgcounts[i]->norm*dgstate[i]->ivar[d]; 

  return result;
}

static double
getC(
     amcount_dgaussian_state_t **dgcounts,  /* Cuentas de la
					       gaussiana */
     dgaussian_state_t          **dgstate,   /* Componentes de mix de
						Gaussianas del
						estado */
     const uint                   I,
     const uint                   d
     )
{
  uint i;
  double result;

  result= 0.0;
  for ( i= 0; i < I; ++i )
    result+=dgcounts[i]->norm * dgstate[i]->ivar[d] * dgstate[i]->mu[d]; 

  return result;
}

static void
getMu(
      amcount_dgaussian_state_t   **dgcounts,  /* Cuentas de la
						  gaussiana */
      uint                          I,
      uint                          d,
      double                      *mu
      )
{
  uint i;

  for ( i= 0; i < I; ++i ) 
    if ( dgcounts[i]->norm == 0 )
      mu[i]= 1;
    else
      mu[i]=dgcounts[i]->x[d]/dgcounts[i]->norm;

}

static void
getVar(
       amcount_dgaussian_state_t **dgcounts,  /* Cuentas de la
						 gaussiana */
       uint                        I,
       uint                        d,
       const double              *smooth,
       double                    *var
       )
{
  uint i;
  
  for ( i= 0; i < I; ++i ) 
    if ( dgcounts[i]->norm == 0 )
      var[i]= smooth[d];
    else
      {
	var[i]= ( dgcounts[i]->x2[d] - 
		  ( dgcounts[i]->x[d] * dgcounts[i]->x[d] )
		  / dgcounts[i]->norm ) / dgcounts[i]->norm;
	if(var[i] < smooth[d])
	  var[i]= smooth[d];
      }
}

static double
getD(
     amcount_dgaussian_state_t **dgcounts,  /* Cuentas de la
					       gaussiana */
     dgaussian_state_t         **dgstate,   /* Componentes de mix de
					       Gaussianas del
					       estado */
     double                    *newmu,
     const uint                  I,
     const uint                  d
      )
{
  uint i;
  double result; 

  result= 0.0;
  for ( i= 0; i < I; ++i )
    result+=dgcounts[i]->norm * dgstate[i]->ivar[d]
      * dgstate[i]->mu[d] * newmu[i]; 

  return result;
}

static double
getE(
     amcount_dgaussian_state_t **dgcounts,  /* Cuentas de la
					       gaussiana */
     dgaussian_state_t         **dgstate,   /* Componentes de mix de
					       Gaussianas del
					       estado */
     double                    *newmu,
     const uint                  I,
     const uint                  d
     )
{
  uint i;
  double result; 

  result= 0.0;
  for ( i= 0; i < I; ++i )
    result+=dgcounts[i]->norm * dgstate[i]->ivar[d] * newmu[i]; 
  
  return result;
}

static double
getF(
     amcount_dgaussian_state_t **dgcounts,  /* Cuentas de la gaussiana */
     dgaussian_state_t         **dgstate,   /* Componentes de mix de
					       Gaussianas del
					       estado */
     double                    *newmu,
     double                    *newvar,
     const uint                  I,
     const uint                  d
     )
{
  uint i;
  double result; 

  result= 0.0;
  for ( i= 0; i < I; ++i )
    result+= (newmu[i] * newmu[i] + newvar[i])
      * dgstate[i]->ivar[d] * dgcounts[i]->norm; 
  
  return result;
}

int
tL_cmllr_alg (
	      const cmllr_alg_paramscounts_t   *pcs,
	      const dgaussian_global_t         *global,
	      double                          **W,
	      cmllr_alg_data_t                 *mem_vars,
	      char                            **err
	      )
{

  double *newmu;
  double *newvar;
  double res1, a ,b;
  /*double res2;*/
  uint D,d,i;
  double opA,opB,opC,opD,opE,opF;
  double sdA,sdB,sdC;
  const double *smooth;


  D= global->D; /* Dimension (39) */
  smooth= global->smooth;
  
  if ( mem_vars->size < pcs->I )
    {
      mt ( mem_vars->newmu= mem_realloc ( double, mem_vars->newmu, pcs->I ) );
      mt ( mem_vars->newvar=
	   mem_realloc ( double, mem_vars->newvar, pcs->I ) );
      mem_vars->size= pcs->I;
    }
  newmu= mem_vars->newmu;
  newvar= mem_vars->newvar;
  
  for( d= 0; d < D; ++d )
  {
    for( i= 0; i < pcs->I; ++i )
    {
      newmu[i]= 0.0;
      newvar[i]= 0.0;
    }

    getMu( pcs->counts, pcs->I, d, newmu);
    getVar( pcs->counts, pcs->I, d, smooth, newvar);

    opA= getA( pcs->counts, pcs->I );
    if ( opA == 0.0 )
      {
	tL_msgerror ( err, _( "all the counts are 0" ) );
	return -1;
      }
    opB= getB( pcs->counts, pcs->params, pcs->I, d );
    opC= getC( pcs->counts, pcs->params, pcs->I, d );
    opD= getD( pcs->counts, pcs->params, newmu, pcs->I, d );
    opE= getE( pcs->counts, pcs->params, newmu, pcs->I, d );
    opF= getF( pcs->counts, pcs->params, newmu, newvar, pcs->I, d );

    sdA= opA * opB;
    sdB= -opC * opE + opB * opD;
    sdC= opE * opE - opB * opF;

    res1= (-sdB + sqrt(sdB*sdB - 4*sdA*sdC))/( 2 * sdA );
    /*res2= (-sdB - sqrt(sdB*sdB - 4*sdA*sdC))/( 2 * sdA );*/
    
    b= (opE - res1*opC) / opB;
    a= res1;
    
    W[0][d]= -(b/a);
    W[d+1][d]= 1.0/a;
    
  }
  
  return 0;
  
}
