/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  create_amodels.c - Implementació de 'create_amodels.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#include "amodel.h"
#include "amodel/bernoulli.h"
#include "amodel/dgaussian.h"
#include "create_amodels.h"
#include "error.h"
#include "mem.h"
#include "text.h"




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
amodel_bernoulli_init (
		       tLAModel           *amodel,
		       const unsigned int  D
		       )
{
  
  unsigned int *aux;
  size_t n;
  tLAModelHMM *cchar;
  bernoulli_state_t state;
  unsigned int q, d;
  
  
  mt ( aux= mem_alloc ( unsigned int, 1 ) );
  *aux= D;
  amodel->global= (void *) aux;
  
  for ( n= 0; n < amodel->N; ++n )
    {
      cchar= &(amodel->hmms[n]);
      cchar->states= amodel->meths->alloc_states ( amodel->global, cchar->Q );
      for ( q= 0; q < cchar->Q; ++q )
	{
	  state= BERNOULLI_STATE(cchar->states)[q];
	  for ( d= 0; d < D; ++d )
	    state.p[d].p1= state.p[d].p0= 0.0;
	}
    }
  
} /* end amodel_bernoulli_init */


static void
amodel_dgaussian_init_state (
			     dgaussian_state_t *state,
			     unsigned int       D
			     )
{
  
  unsigned int d;
  
  
  for ( d= 0; d < D; ++d )
    state->mu[d]= 0.0;
  for ( d= 0; d < D; ++d )
    state->var[d]= 0.0;
  state->logc= -HUGE_VAL;
  
} /* end amodel_dgaussian_init_state */


static void
amodel_dgaussian_init (
		       tLAModel           *amodel,
		       const unsigned int  D
		       )
{
  
  dgaussian_global_t *global;
  tLFloat *smooth;
  size_t n;
  tLAModelHMM *cchar;
  unsigned int q, d;
  
  
  mt ( global= mem_alloc ( dgaussian_global_t, 1 ) );
  global->D= D;
  mt ( global->smooth= smooth= mem_alloc ( tLFloat, D ) );
  for ( d= 0; d < D; ++d ) smooth[d]= 0.0;
  amodel->global= global;
  
  for ( n= 0; n < amodel->N; ++n )
    {
      cchar= &(amodel->hmms[n]);
      cchar->states= amodel->meths->alloc_states ( amodel->global, cchar->Q );
      for ( q= 0; q < cchar->Q; ++q )
	amodel_dgaussian_init_state ( &(DGAUSSIAN_STATE(cchar->states)[q]), D );
    }
  
} /* end amodel_dgaussian_init */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(tLAModel *)
tL_amodel_bernoulli_new (
			 const tLDict       *syms,
			 const unsigned int  Q,
			 const unsigned int  D
			 )
{
  
  tLAModel *new;
  
  
  new= tL_create_amodel_empty ( "Bernoulli", syms, Q );
  amodel_bernoulli_init ( new, D );
  
  return new;
  
} /* end tL_amodel_bernoulli_new */


TL_API_FUNC(tLAModel *)
tL_amodel_bernoulli_new_vs (
			    const tLDict       *syms,
			    const unsigned int *Qs,
			    const unsigned int  D
			    )
{
  
  tLAModel *new;
  
  
  new= tL_create_amodel_empty_vs ( "Bernoulli", syms, Qs );
  amodel_bernoulli_init ( new, D );
  
  return new;
  
} /* end tL_amodel_bernoulli_new_vs */


TL_API_FUNC(tLAModel *)
tL_amodel_dgaussian_new (
			 const tLDict       *syms,
			 const unsigned int  Q,
			 const unsigned int  D
			 )
{
  
  tLAModel *new;
  
  
  new= tL_create_amodel_empty ( "DGaussian", syms, Q );
  amodel_dgaussian_init ( new, D );
  
  return new;
  
} /* end tL_amodel_dgaussian_new */


TL_API_FUNC(tLAModel *)
tL_amodel_dgaussian_new_vs (
			    const tLDict       *syms,
			    const unsigned int *Qs,
			    const unsigned int  D
			    )
{
  
  tLAModel *new;
  
  
  new= tL_create_amodel_empty_vs ( "DGaussian", syms, Qs );
  amodel_dgaussian_init ( new, D );
  
  return new;
  
} /* end tL_amodel_dgaussian_new_vs */


TL_API_FUNC(tLFloat *)
tL_amodel_dgaussian_get_smooth (
				tLAModel *amodel
				)
{
  return DGAUSSIAN_GLOBAL(amodel->global)->smooth;
} /* end tL_amodel_dgaussian_get_smooth */
