/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  extract.c - Implementació de 'extract.h'.
 *
 *  Adrià Giménez Pastor, Juan Daniel Valor Miró.
 */


#include <assert.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "error.h"
#include "extract.h"
#include "fft.h"
#include "mem.h"




/**********/
/* MACROS */
/**********/

#define PI 3.14159265358979323846




/*********************/
/* PRIVATE FUNCTIONS */
/*********************/

static void
free_frame (
	    tLFeatureExtractor *fext
	    )
{
  if ( fext->frame[0] != NULL ) free ( fext->frame[0] );
} /* end free_frame */


static void
prepare_frame (
	       tLFeatureExtractor *fext,
	       const int           wsize,
               const int           ffsize
	       )
{
  
  if ( wsize == fext->frame_wsize ) return;
  free_frame ( fext );
  fext->frame_wsize= wsize;
  fext->frame[0]= mem_alloc ( double, ffsize*2 );
  fext->frame[1]= fext->frame[0]+ffsize;
  
} /* end prepare_frame */


static double
get_frame (
	   const double       *input,
	   const int           begin,
	   const int           end,
	   const int           length,
	   tLFeatureExtractor *fext
	   )
{
  
  int i, j;
  double sum, val, prev, factor, e;
  
  
  sum= e= 0.0;
  for ( i= begin, j= 0; i < end; ++i, ++j )
    {
      fext->frame[0][j]= val= input[i];
      sum+= val; e+= val*val;
    }
  for ( ; j < length; ++j ) fext->frame[0][j]= 0; /* In any case calc
						  something. */
  
  if ( fext->zeromean ) /* ZEROMEAN */
    {
      sum/= length;
      for ( j= 0; j < length; ++j ) {
        fext->frame[0][j]-= sum;
      }
    }
  
  if ( fext->preemcoef != 0.0 ) /* PRE-EMPHASIS */
    {
      prev= fext->frame[0][0];
      for ( j= 1; j < length; ++j )
	{
	  val= fext->frame[0][j];
	  fext->frame[0][j]= val - fext->preemcoef*prev;
	  prev= val;
	}
        fext->frame[0][0] *= 1.0 - fext->preemcoef;
    }
  
  if ( fext->usehamming && length > 1 ) /* Hamming window */
    {
      factor= (2*PI)/(length-1);
      for ( j= 0; j < length; ++j )
	fext->frame[0][j]= (0.54 - 0.46*cos ( factor*j )) * fext->frame[0][j];
    }
  
  return e;
  
} /* end get_frame */

#if 0
static void
free_fftw (
	   tLFeatureExtractor *fext
	   )
{
} /* end free_fftw */


static void
prepare_fftw (
	      tLFeatureExtractor *fext,
	      int                 N
	      )
{
} /* end prepare_fftw */
#endif

/* Stores the spectrum in frame. */
static void
calc_fft (
	  tLFeatureExtractor *fext,
	  const int           ffsize
	  )
{

  int i;


  for ( i= fext->frame_wsize; i < ffsize; ++i ) fext->frame[0][i]= 0.0;
  for ( i= 0; i < ffsize; ++i ) fext->frame[1][i]= 0.0;
  tL_FFTStockham ( fext->frame[0], fext->frame[1], ffsize );
  
} /* end calc_fft */


static void
free_ftb (
	  tLFeatureExtractor *fext
	  )
{
  
  if ( fext->ftb_weights != NULL ) free ( fext->ftb_weights );
  if ( fext->ftb_out != NULL ) free ( fext->ftb_out );
  
} /* end free_ftb */


static double
ftb_freq2mel (
	      const int val
	      )
{
  return 1127 * log ( 1.0 + ((float) val)/700.0 );
} /* end ftb_freq2mel */


static double
ftb_ind2mel (
	     const int    ind,
	     const double windur
	     )
{
  return 1127 * log ( 1.0 + ind/(windur*700.0) );
} /* end ftb_ind2mel */


/* Returns an array of |numchans+1| with the central mel value for
   each channel. The 1th channel is in position 1. Position 0 is the
   mel value related with the low frequency. The center of the last
   channel is related with the high frequency. */
static double *
ftb_calc_mel_chans (
		    tLFeatureExtractor *fext,
		    const int           ifreq,
                    double             *lomels
		    )
{
  
  double *chans, ratio, himels;
  int i, lofreq, hifreq, ifreq2;
  
  
  /* Low and high frequencies. */
  lofreq= (fext->ftb_lofreq<0) ? 0 : fext->ftb_lofreq;
  ifreq2= ifreq/2;
  hifreq= (fext->ftb_hifreq>ifreq2) ? ifreq2 : fext->ftb_hifreq;
  assert ( lofreq < hifreq );
  *lomels= ftb_freq2mel ( lofreq );
  himels= ftb_freq2mel ( hifreq );
  
  /* Calc central mel values for channels. */
  chans= mem_alloc ( double, fext->numchans+1 );
  ratio= (himels-*lomels);
  for ( i= 0; i <= fext->numchans; ++i )
    chans[i]= *lomels + ((i+1)/((double)fext->numchans+1))*ratio;
  
  return chans;
  
} /* end ftb_calc_mel_chans */


static void
ftb_create_weights (
		    tLFeatureExtractor *fext,
		    const double       *chans,
		    const int           Ny2,
                    double              lomels
		    )
{
  
  int i, pos;
  double mel, windur;
  struct tLFextFtbWeight__ *weights;
 
  fext->ftb_weights= weights= mem_alloc ( struct tLFextFtbWeight__, Ny2 );
  pos= 0;
  windur= (Ny2*2)/(double)fext->ftb_infreq;

  for ( i= 0; i < Ny2; ++i )
    {
      mel= ftb_ind2mel ( i, windur );
      for ( ; pos <= fext->numchans && mel > chans[pos]; ++pos );
      if ( pos > fext->numchans ) break;
      if ( mel < lomels ) { weights[i].pos= -1; continue; }
      weights[i].pos= pos;
      if ( pos == 0 ) weights[i].w= (mel - lomels) / (chans[pos] - lomels);
      else weights[i].w= (mel - chans[pos-1]) / (chans[pos] - chans[pos-1]);
    }
  fext->ftb_end= i;
  for ( ; i < Ny2; ++i ) weights[i].pos= -1;
  for ( i= 0; i < Ny2 && weights[i].pos == -1; ++i );
  fext->ftb_begin= i+1;

} /* end ftb_create_weights */


static void
prepare_ftb (
	     tLFeatureExtractor *fext,
	     const int           infreq,
	     const int           ffsize /* ffsize is power of 2. */
	     )
{
  
  double *mel_chans, lomels;
  int Ny2;
  
  if ( fext->ftb_numchans == fext->numchans && 
       fext->ftb_infreq == infreq &&
       fext->ftb_lofreq == fext->lofreq &&
       fext->ftb_hifreq == fext->hifreq &&
       fext->ftb_wsize == ffsize )
    return;
  
  /* Free and update values. */
  free_ftb ( fext );
  fext->ftb_infreq= infreq;
  fext->ftb_wsize= ffsize;
  fext->ftb_lofreq= fext->lofreq;
  fext->ftb_hifreq= fext->hifreq;
  
  /* Prepare for filterbank. */
  fext->ftb_out= mem_alloc ( double, fext->numchans );
  Ny2= ffsize/2;
  mel_chans= ftb_calc_mel_chans ( fext, infreq, &lomels );
  ftb_create_weights ( fext, mel_chans, Ny2, lomels );
  free ( mel_chans );
  
} /* end prepare_ftb */


/* Using the output of FFT as input calculates the mel filterbanks in
   ftb_out. */
static void
calc_ftb (
	  tLFeatureExtractor *fext,
	  const int           Ny2
	  )
{
  
  int n;
  double amp, a, b, aux;
  struct tLFextFtbWeight__ weight;
  
  
  for ( n= 0; n < fext->numchans; ++n ) fext->ftb_out[n]= 0.0;
  for ( n= fext->ftb_begin; n < fext->ftb_end; ++n )
    {
      weight= fext->ftb_weights[n];
      if ( weight.pos == -1 ) continue;
      a= fext->frame[0][n];
      b= fext->frame[1][n];
      amp= a*a + b*b;
      
      if ( !(fext->usepower) ) amp= sqrt ( amp );

      aux= amp*weight.w;
      if(weight.pos<fext->numchans) fext->ftb_out[weight.pos]+= aux;
      if(weight.pos>0) fext->ftb_out[weight.pos-1]+= amp - aux;
    }
} /* end calc_ftb */


static void
calc_logftb (
	     tLFeatureExtractor *fext
	     )
{
  
  int n;
  double val;
  
  
  for ( n= 0; n < fext->numchans; ++n )
    {
      val= fext->ftb_out[n];
      fext->ftb_out[n]= (val<1.0) ? 0.0 : log ( val );
    }
  
} /* end calc_logftb */


static void
free_dct (
	   tLFeatureExtractor *fext
	   )
{
  if ( fext->dct_out != NULL ) free ( fext->dct_out );
} /* end free_dct */


static void
prepare_dct (
	     tLFeatureExtractor *fext
	     )
{
  
  if ( fext->dct_numceps == fext->numceps &&
       fext->dct_numchans == fext->numchans ) return;
  free_dct ( fext );
  fext->dct_numceps= fext->numceps;
  fext->dct_numchans= fext->numchans;
  fext->dct_out= mem_alloc ( double, fext->numceps );
  fext->dct_sqrt_2_N= sqrt ( 2.0 / fext->numchans );
  fext->dct_pi_N= PI / fext->numchans;
  
} /* end prepare_dct */


/* Calculates Discrete Cosine Transform. ftb_out -> dct_out */
static void
calc_dct (
	  tLFeatureExtractor *fext
	  )
{
  
  int i, j;
  double val, aux, aux2;
  
  
  for ( i= 0; i < fext->numceps; ++i )
    {
      val= 0.0; aux= fext->dct_pi_N*(i+1);
      for ( j= 0; j < fext->numchans; ++j )
	val+= fext->ftb_out[j] * cos ( aux*(j+1-0.5) );
      fext->dct_out[i]= val * fext->dct_sqrt_2_N;
    }
  if ( fext->ceplifter > 0 )
    {
      aux= ((double) fext->ceplifter) / 2.0;
      aux2= PI / (double) fext->ceplifter;
      for ( i= 0; i < fext->numceps; ++i )
	fext->dct_out[i]= fext->dct_out[i] * (1 + aux*sin ( aux2*(i+1) ));
    }
      
} /* end calc_dct */


static void
calc_energy (
	     float        **frames,
	     const int      length,
	     const int      pos,
	     const double   maxenergy
	     )
{
  
  int t;
  double floor, min, val;
  
  
  if ( maxenergy == -HUGE_VAL )
    {
      for ( t= 0; t < length; ++t )
	frames[t][pos]= 1.0;
      return;
    }
  floor= 5*log(10); /* 50dB = 10*log10(W1/W0) --> 5*log(10) = log(W1/W0) */
  min= maxenergy - floor; /* Minum value is max value minus 50dB. */
  for ( t= 0; t < length; ++t )
    {
      val= frames[t][pos];
      if ( val < min ) val= min;
      /* Normalize between 1-floor and 1.0. 0.1 is used as a scale factor. */
      frames[t][pos]= 1.0 + (val-maxenergy)*0.1;
    }
  
} /* end norm_energy */


static void
calc_derivatives (
		  float     **frames,
		  const int   length,
		  const int   d0_src,   /* First source dimension. */
		  const int   d0_dst,   /* First destination
					   dimension. */
		  const int   D,        /* original dimension. */
		  const int   wsize
		  )
{
  
  int t, aux, i, l, r, d, d_src;
  double norm, val;
  
  
  aux= 1; for ( i= 2; i <= wsize; ++i ) aux+= i*i;
  norm= 1.0/(2.0 * (double) aux);

  for ( t= 0; t < length; ++t )
    for ( d= 0; d < D; ++d )
      {
	val= 0.0;
	d_src= d0_src+d;
	for ( i= 1; i <= wsize; ++i )
	  {
	    l= t-i; if ( l < 0 ) l= 0;
	    r= t+i; if ( r >= length ) r= length-1;
	    val+= i * ( frames[r][d_src] - frames[l][d_src] );
	  }
	frames[t][d0_dst+d]= val*norm;
      }
  
} /* end calc_derivatives */

static void
norm_mean_z (
	     float     **frames,
	     const int   length,
	     const int   D        /* original dimension. */
	     )
{
  
  int t, d;
  double sum;
  
  for ( d= 0; d < D; ++d )
    {
      sum= 0;
      for ( t= 0; t < length; ++t ) sum+= frames[t][d];
      sum/= length;
      for ( t= 0; t < length; ++t ) frames[t][d]-= sum;
    }
} /* end norm_mean_z */

static double
cep0func (
          tLFeatureExtractor  *fext
         )
{
  
  int n;
  double sum;
  
  sum= 0;
  for ( n= 0; n < fext->numchans; ++n ) sum+= fext->ftb_out[n];

  return sum*sqrt(2.0/(float)fext->numchans);
} /* end cep0func */

/********************/
/* PUBLIC FUNCTIONS */
/********************/

TL_API_FUNC(void)
tL_extract_features (
		     tLFeatureExtractor  *fext,
		     const double        *input,
		     int                  length,
		     int                  infreq,
		     tLFea               *output
		     )
{
  
  double spf, p, energy, maxenergy;
  int begin, end, wsize, Ny2, d, nframes, t, D, ED, RD, ffsize, nframes2;
  float **frames, *frame;
  
  assert ( length > 0 );
  assert ( infreq > 0 );
  assert ( fext->outfreq > 0 );
  assert ( fext->wduration > 0 );
  assert ( fext->numchans > 0 );
  assert ( fext->preemcoef >= 0 && fext->preemcoef < 1.0 );
  
  /* Samples per frame and nframes max. */
  spf= infreq / (double) fext->outfreq;
  nframes= (int) ((length / spf)+0.5);
  
  /* Window size. */
  wsize= (int) (infreq*fext->wduration+0.5);
  if ( wsize > length ) wsize= length;
  if ( wsize < 4 ) wsize= 4;
  else if ( wsize&0x1 ) --wsize;
  ffsize= 2;   
  while (wsize>ffsize) ffsize*= 2;
  Ny2 = wsize/2;
  
  /* Prepare structures. */
  prepare_frame ( fext, wsize, ffsize );
  /*prepare_fftw ( fext, ffsize );*/
  prepare_ftb ( fext, infreq, ffsize );
  switch ( fext->ftype )
    {
    case TL_MELSPEC: 
    case TL_FBANK: D= fext->numchans; break;
    case TL_MFCC:
      assert ( fext->numceps > 0 );
      prepare_dct ( fext );
      D= fext->numceps;
      break;
    default: D= 0; tL_error ( _( "tL_extract_features: unknown ftype" ));
    }
  RD= ED= D;
  if ( (fext->ftype == TL_MFCC) &&
       (fext->energy || fext->cep0) ) { ++RD; ++ED; }
  if ( fext->deltawindow > 0 )
    RD*= (fext->accwindow > 0) ? 3 : 2;
  tL_fea_resize ( output, RD, nframes );
  
  /* Calc frames. */
  nframes2= 0;
  maxenergy= -HUGE_VAL;
  frames= tL_fea_get_real_data ( output );
  for ( t= 0, p= 0; p < length; ++t, p+= spf )
    {
      
      frame= frames[t];
      
      /* Calculate begin and end. If it is not possible to fit the
	 window in the sample, then begin must be 0. */
      begin= (int) (p+0.5);
      end= begin+wsize;
      if ( end > length && nframes2 > 0) break;
      nframes2+= 1;
      
      /* Common steps. */
      energy= get_frame ( input, begin, end, wsize, fext );
      if ( fext->energy )
	{
	  frame[D]= energy= log ( energy );
	  if ( energy > maxenergy ) maxenergy= energy;
	}

      calc_fft ( fext, ffsize );
      calc_ftb ( fext, Ny2 );
      
      /* Generate output. */
      if ( fext->ftype == TL_MELSPEC ) /* MELSPEC */
	{
	  for ( d= 0; d < D; ++d ) frame[d]= (float) fext->ftb_out[d];
	  continue;
	}
      calc_logftb ( fext );
      if ( fext->ftype == TL_FBANK ) /* FBANK */
	{
	  for ( d= 0; d < D; ++d ) frame[d]= (float) fext->ftb_out[d];
	  continue;
	}
      if ( fext->cep0 && !fext->energy )
        {
          frame[D]= energy= cep0func( fext );
          if ( energy > maxenergy ) maxenergy= energy;
        }
      calc_dct ( fext ); /* MFCC */
      for ( d= 0; d < D; ++d ) frame[d]= (float) fext->dct_out[d];
      
    }
  
  nframes= nframes2;
  output->cols= nframes;
  
  /* Normalise energy and derivatives. */
  if ( fext->meannorm ) norm_mean_z ( frames, nframes, ED );
  if ( fext->energy ) calc_energy ( frames, nframes, D, maxenergy );
  if ( fext->deltawindow > 0 )
    {
      calc_derivatives ( frames, nframes, 0, ED, ED, fext->deltawindow );
      if ( fext->accwindow > 0 )
	calc_derivatives ( frames, nframes, ED, 2*ED, ED, fext->accwindow );
    }
  
} /* end tL_extract_features */


TL_API_FUNC(void)
tL_feature_extractor_free (
			   tLFeatureExtractor *fext
			   )
{
  
  free_dct ( fext );
  /*free_fftw ( fext );*/
  free_ftb ( fext );
  free_frame ( fext );
  free ( fext );
  
} /* end tL_feature_extractor_free */


TL_API_FUNC(tLFeatureExtractor *)
tL_feature_extractor_new (void)
{
  
  tLFeatureExtractor *new;
  
  
  new= mem_alloc ( tLFeatureExtractor, 1 );
  
  /* Default values. */
  new->ftype= TL_MFCC;
  new->outfreq= 100;
  new->wduration= 0.025;
  new->numchans= 26;
  new->lofreq= 0;
  new->hifreq= INT_MAX;
  new->numceps= 12;
  new->zeromean= TRUE;
  new->preemcoef= 0.97;
  new->usehamming= TRUE;
  new->usepower= TRUE;
  new->ceplifter= 22;
  new->energy= TRUE;
  new->deltawindow= 2;
  new->accwindow= 2;
  new->cep0= TRUE;
  new->meannorm= TRUE;
  
  /* Frame. */
  new->frame_wsize= -1;
  new->frame[0]= NULL;
  
  /* Filterbank. */
  new->ftb_numchans= -1;
  new->ftb_weights= NULL;
  new->ftb_out= NULL;
  
  /* DCT. */
  new->dct_out= NULL;
  new->dct_numceps= -1;
  
  return new;
  
} /* end tL_feature_extractor_new */
