/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  extract.h - Extracció de característiques.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */

#ifndef __LIBTL_EXTRACT_H__
#define __LIBTL_EXTRACT_H__

#include <limits.h>
#include <stddef.h>
#include <stdio.h>

#include "port.h"
#include "common_public.h"
#include "samples_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup Extract Feature extraction
 */
/** \addtogroup Extract
 *  @{
 */

/* For each position of the window (Ny2) stores index in 'mel_chans'
   of the left bank filter and its weight. */
struct tLFextFtbWeight__
{
  
  int    pos;
  double w;
  
};

typedef enum
  {
    TL_MELSPEC,
    TL_FBANK,
    TL_MFCC
  } tLTypeFeatures;

/** Feature extractor. This structure is used for feature
 *  extraction. Public fields could be modified at any time.
 */
typedef struct
{
  
  tLTypeFeatures ftype;          /**< Type of the features. Default
				    TL_MFCC. */
  int            outfreq;        /**< Output frequency. Default
				    100Hz. */
  double         wduration;      /**< Duration for each extraction
				    window in seconds. Default 0.025
				    (25ms). */
  int            numchans;       /**< Number of channels used for the
				    Mel filter bank. Default 26. */
  int            lofreq;         /**< Low frequency cut-off for
				    filterbank analysis. Default 0. */
  int            hifreq;         /**< High frequency cut-off for
				    filterbank analysis. Default
				    INT_MAX. */
  int            numceps;        /**< Number of cepstral features
				    (only for MFCC). Default 12. */
  tLBool         zeromean;       /**< Removes the DC mean from the
				    source. Default TRUE. */
  double         preemcoef;      /**< Coefficient used to
				    pre-emphasize the signal
				    ([0,1[). Default 0.97. */
  tLBool         usehamming;     /**< Apply Hamming windowing. Default
				    TRUE. */
  tLBool         usepower;       /**< Uses the power in filterbank
				    analysis. Default TRUE. */
  int            ceplifter;      /**< Value used to lifter the
				    cepstra. Default 22. */
  tLBool         energy;         /**< Calculates the energy and
				    normalizes it. Default TRUE. */
  int            deltawindow;    /**< Width of the window used to
				    calculate delta coefficients. A 0
				    value disables the calculation of
				    delta coefficients. Default value
				    2. */
  tLBool         cep0;           /**< Add Cepstral 0. Default value TRUE. */
  tLBool         meannorm;       /**< Normalize the vectors substracting
                                    the mean. Default value TRUE. */
  int            accwindow;      /**< Width of the window used to
				    calculate acceleration
				    coefficients. A 0 value disables
				    the calculation of acceleration
				    coeeficients. Default value 2. */
  
  
  /* Private. frame. */
  int     frame_wsize;
  double *frame[2]; /* The second one is for imaginary values. */
  
  /* Private. FFT. */
  
  /* Private. Mel filters bank. */
  int                       ftb_numchans;
  int                       ftb_infreq;
  int                       ftb_wsize;
  int                       ftb_lofreq;
  int                       ftb_hifreq;
  int                       ftb_begin;
  int                       ftb_end;
  struct tLFextFtbWeight__ *ftb_weights;
  double                   *ftb_out;
  
  /* Private. DCT. */
  double *dct_out;
  int     dct_numceps;
  int     dct_numchans;
  double  dct_sqrt_2_N;
  double  dct_pi_N;
  
} tLFeatureExtractor;

/** Extracts features. This functions transforms an input sound wave,
 *  stored as a sequence of doubles, into a sequence of feature
 *  vectors. Feature extraction parameters are stored in the feature
 *  extractor.
 *
 *  @param fext The feature extractor.
 *  @param input Input sound wave.
 *  @param length Length of the input.
 *  @param infreq Input frequency in Hz.
 *  @param fea Feature vector managers to which the output will be
 *  stored.
 */
TL_API_FUNC(void)
tL_extract_features (
		     tLFeatureExtractor *fext,
		     const double       *input,
		     int                 length,
		     int                 infreq,
		     tLFea              *output
		     );

/** Frees memory. Frees the memory allocated for the feature extractor.
 *
 *  @param fext The feature extractor.
 */
TL_API_FUNC(void)
tL_feature_extractor_free (
			   tLFeatureExtractor *fext
			   );

/** Gets a tLFeatureExtractor. This function creates a feature
 *  extractor. Parameters are initialized to their default values.
 *
 *  @return The new feature extractor.
 */
TL_API_FUNC(tLFeatureExtractor *)
tL_feature_extractor_new (void);

/** @} */
/** @} */

#endif /* __LIBTL_EXTRACT_H__ */
