/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  init_amodel.h - Inicialitza el model de caràcters.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2013.
 */

#ifndef __LIBTL_INIT_AMODEL_H__
#define __LIBTL_INIT_AMODEL_H__

#include "amodel_public.h"
#include "amcounts_public.h"
#include "port.h"
#include "samples_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup Init Initialization of an acoustic model using linear segmentation
 */
/** \addtogroup Init
 *  @{
 */

#ifndef __TL_PRV
/** Used in 'tL_init'. This structure contains auxiliary data used by
 *  'tL_init' to initialize an acoustic model from samples and their
 *  transcriptions using linear segmentation.
 */
struct _tLInitializer
{
  
  const tLAModel *amodel;    /**< Reference to the acoustic model
				used by the model initializer. */
  const tLSeqs   *seqs;      /**< Reference to the symbol sequence
                                manager used by the initializer. */
  
};
#endif

/** Used in 'tL_init'. This structure contains auxiliary data used by
 *  'tL_init' to initialize an acoustic model from samples and their
 *  transcriptions by linear segmentation.
 */
typedef struct _tLInitializer tLInitializer;

/** Frees memory. Frees the memory allocated for the initializer.
 *
 *  @param ini The initializer.
 */
TL_API_FUNC(void)
tL_initializer_free (
		     tLInitializer *ini
		     );

/* Torna NULL en cas d'error. */
/** Gets a tLInitializer. This function creates a new initializer
 *  used to initialize an acoustic model from samples and their
 *  transcriptions using linear segmentation. Currently supported types
 *  are: see tL_hmm_initializer_new.
 *
 *  @param amodel The acoustic model.
 *  @param seqs A symbol sequence manager containing the
 *  transcription of the samples.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @return Initializer, or NULL in case of error.
 */
TL_API_FUNC(tLInitializer *)
tL_initializer_new (
		    const tLAModel  *amodel,
		    const tLSeqs    *seqs,
		    char           **err
		    );

/** Initializes an acoustic model from a sample and its transcription. This
 *  function linearly segments the input frames, according to the
 *  number of symbols of the transcription. According to this
 *  segmentation, acoustic model counts are increased using
 *  tL_init_hmm. The transcription is specified as an integer code
 *  according to the sequence manager used during the creation of the
 *  auxiliary data.
 *
 *  @param ini Initializer.
 *  @param counts Integer counts related to amodel.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param seq The code of the symbol sequence related to fea.
 *  @return 0 if the sample has been successfully used to increase the
 *  acoustic model counts or the minimum number of required frames.
 */
TL_API_FUNC(int)
tL_init (
	 tLInitializer *ini,
	 tLAMiCounts   *counts,
	 const tLFea   *fea,
	 const size_t   seq
	 );

/** @} */
/** @} */

#endif /* __LIBTL_INIT_AMODEL_H__ */
