/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  init_hmm.h - Inicialitza els comptadors associats a un caràcter
 *               mitjançant segmentació uniforme.
 *
 *  Adrià Giménez Pastor, 2008,2009,2011,2013.
 *
 */

#ifndef __LIBTL_INIT_HMM_H__
#define __LIBTL_INIT_HMM_H__

#include "amodel_public.h"
#include "amcounts_public.h"
#include "port.h"
#include "samples_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup InitHMM Initialization from transcriptions of HMMs \
 using linear segmentation
 */
/** \addtogroup InitHMM
 *  @{
 */

/** HMM initializer. This structure contains auxiliary
 *  data used by 'tL_init_hmm' to initialize an HMM using
 *  linear segmentation.
 */
typedef struct _tLHMMInitializer tLHMMInitializer;

typedef struct
{
  
  void (*init) (struct _tLHMMInitializer *init,
		void                     *global,
		void                     *states,
		void                     *cstates,
		const unsigned int        q,
		const tLFea              *fea,
		const int                 t);
  void (*free) (struct _tLHMMInitializer *init);
  
} tLInitHMMAuxMeths;

#define TL_INIT_HMM_AUX			\
  const tLInitHMMAuxMeths *meths

struct _tLHMMInitializer
{
  const tLAModel *amodel;    /**< Reference to the acoustic model
				used by the HMM
				initializer. */
  TL_INIT_HMM_AUX;
};

/** Frees memory. Frees the memory allocated for the HMM
 *  initializer.
 *
 *  @param INIT The HMM initializer.
 */
#define tL_hmm_initializer_free(INIT)		\
  (INIT)->meths->free ( INIT )

/* Torna NULL en cas d'error. */
/** Gets a HMM initializer. This function creates a new
 *  HMM initializer used to initialize HMMs using
 *  linear segmentation. Currently supported types are: Bernoulli and
 *  DGaussian.
 *
 *  @param amodel The acoustic model.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @return HMM initializer, or NULL in case of error.
 */
TL_API_FUNC(tLHMMInitializer *)
tL_hmm_initializer_new (
			const tLAModel  *amodel,
			char           **err
			);

/** Initializes an HMM using linear segmentation of a sample segment.
 *  This function linearly segments the input frames, which
 *  are a segment of the given feature vector sequence, according to
 *  the number of states of the specified HMM. According
 *  to this segmentation, input frames are assigned to states, and
 *  then, state counts are consequently increased.
 *
 *  @param init HMM initializer.
 *  @param counts Integer counts related to amodel used by the
 *  HMM initializer.
 *  @param c The index of the HMM to be initialized.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param begin The first input frame.
 *  @param nframes The number of input frames.
 */
TL_API_FUNC(void)
tL_init_hmm_seg (
		 tLHMMInitializer *init,
		 tLAMiCounts      *counts,
		 const int         c,
		 const tLFea      *fea,
		 const int         begin,
		 const int         nframes
		 );

/** Initializes an HMM using linear segmentation of
 *  sample. This macro is equivalent to tL_init_hmm_seg but using all
 *  frames contained in fea.
 *
 *  @param INIT The HMM initializer.
 *  @param COUNTS Integer counts related to AMODEL used by INIT.
 *  @param C The index of the HMM to be initialized.
 *  @param FEA A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the AMODEL.
 */
#define tL_init_hmm(INIT,COUNTS,C,FEA)					\
  tL_init_hmm_seg ( INIT, COUNTS, C, FEA, 0, (FEA)->cols )

/** @} */
/** @} */

#endif /* __LIBTL_INIT_HMM_H__ */
