/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  itertrain.c - Implementació de 'itertrain.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#define __TL_PRV
#include "abmat.h"
#include "auxfstats.h"
#include "calc.h"
#include "amodel_itrans.h"
#include "common.h"
#include "error.h"
#include "linabmat.h"
#include "logamcounts.h"
#include "itertrain.h"
#include "mem.h"
#include "smodels.h"




/**********/
/* MACROS */
/**********/

#define CALC_PROB(FSC,AFMNG,FEA,T,GLOBAL,STATES,Q)			\
  ((FSC)[(Q)].v[(T)].a ? (FSC)[(Q)].v[(T)].p :				\
   fstats_add ( &((FSC)[(Q)]), (AFMNG), (FEA), (T), (GLOBAL), (STATES), (Q) ))

#define MBASEFB(PTR) ((mbasefb_t *) (PTR))
#define MLINFB(PTR) ((mlinfb_t *) (PTR))
#define MFB(PTR) ((mfb_t *) (PTR))




/*********/
/* TIPUS */
/*********/

/* FSTATS *********************************************************************/
/* NOTA: p sempre estarà a -HUGE_VAL quan no està actiu. */
typedef struct
{
  
  tLProb p;    /* Probabilitat. */
  tLBool a;    /* Actiu. */
  
} fstats_node_t;

/* Estadístiques dels frames per a un estat. */
typedef struct
{
  
  fstats_node_t *v;          /* |T|. Nodes. */
  tLProb        *weights;    /* |T|. Pesos. */
  void          *aux;        /* |T|. auxfstats. */
  int           *stack;      /* |T|. Instants actius. */
  int            N;          /* Número de nodes actius. */
  int            T;          /* Capacitat màxima. */
  
} fstats_t;

/* MAUX ***********************************************************************/
/* Wrapper per al forward-backward lineal. */
#define MBASEFB_CLASS				\
  /* S'han d'assignar al crear-lo. */		\
  fstats_t          **fstats;			\
  auxfstats_mng_t    *afmng;			\
  void               *global;			\
  /* S'han d'assignar abans de alpha. */	\
  const smodel_t     *smodel;			\
  /* S'han d'assignar abans de beta. */		\
  tLProb              prob;			\
  /* Variables. */				\
  fstats_t           *fstats_c;			\
  logamcount_hmm_t  *ccount

typedef struct
{
  MBASEFB_CLASS;
} mbasefb_t;

typedef struct
{
  MBASEFB_CLASS;
} mlinfb_t;

typedef struct
{
  MBASEFB_CLASS;
  /* S'han d'assignar al crear-lo. */
  tLAMlogCounts *counts;
} mfb_t;


/* Memòria auxiliar per a càlculs. */
typedef struct maux
{
  
  unsigned int      max_C;
  unsigned int      max_Q;
  auxfstats_mng_t  *afmng;
  fstats_t        **fstats;    /* |C|*|Q|. */
  enum {
    LINEAR,
    GENERIC
  }                 type;
  union
  {
    struct
    {
      abmat_t *m;
      mfb_t   *fb;
    }           gen;
    struct
    {
      linabmat_t *m;
      mlinfb_t   *fb;
    }           lin;
  }                 abmat;
  int               max_width;
  tLFloat          *occs_calc;    /* |T|. NULL si no es gasta. */
  
} maux_t;


struct _tLTrainer
{
  
  const tLAModel  *amodel;
  const tLSeqs    *seqs;
  tLAMlogCounts   *counts;
  tLFloat        **occs;
  smodels_t        smodels;
  maux_t          *maux;
  int              type;
  
};




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* FSTATS *********************************************************************/
static void
fstats_clear (
	      fstats_t *fstats
	      )
{
  
  int n, t;
  
  
  for ( n= 0; n < fstats->N; ++n )
    {
      t= fstats->stack[n];
      fstats->v[t].p= -HUGE_VAL;
      fstats->v[t].a= FALSE;
    }
  fstats->N= 0;
  
} /* end fstats_clear */


/* També neteja. */
static void
fstats_accum (
	      fstats_t        *fstats,
	      auxfstats_mng_t *afmng,
	      const tLFea     *fea,
	      void            *global,
	      void            *cglobal,
	      void            *states,
	      void            *cstates,
	      tLFloat         *occs,
	      tLFloat         *occs_calc,
	      unsigned int     q,
	      tLProb           prob
	      )
{
  
  int n, t;
  tLFloat max, aux;
  
  
  for ( n= 0; n < fstats->N; ++n )
    {
      t= fstats->stack[n];
      fstats->weights[t]-= prob;
      fstats->v[t].p= -HUGE_VAL;
      fstats->v[t].a= FALSE;
    }
  if ( occs != NULL )
    {
      max= -HUGE_VAL;
      for ( n= 0; n < fstats->N; ++n )
	{
	  t= fstats->stack[n];
	  aux= occs_calc[n]= fstats->weights[t];
	  if ( aux > max ) max= aux;
	}
      aux= (max!=-HUGE_VAL && fstats->N>1) ?
	tL_radd_vals ( occs_calc, max, fstats->N ) : max;
      occs[q]= RADD ( occs[q], aux );
    }
  auxfstats_mng_pseudoM_step ( afmng, fstats->aux, fstats->weights,
			       fstats->stack, fstats->N, fea,
			       global, cglobal, states, cstates, q );
  fstats->N= 0;
  
} /* end fstats_accum */


/* Es supossa que no està. */
static tLProb
fstats_add (
	    fstats_t        *fstats,
	    auxfstats_mng_t *afmng,
	    const tLFea     *fea,
	    int              t,
	    void            *global,
	    void            *states,
	    unsigned int     q
	    )
{
  
  tLProb ret;
  
  
  fstats->weights[t]= -HUGE_VAL;
  fstats->v[t].a= TRUE;
  fstats->v[t].p= ret= 
    auxfstats_mng_pseudoE_step ( afmng, fstats->aux, fea, t,
				 global, states, q );
  fstats->stack[fstats->N++]= t;
  
  return ret;
  
} /* end fstats_add */


static void
fstats_free (
	     fstats_t         fstats,
	     auxfstats_mng_t *afmng
	     )
{
  
  free ( fstats.v );
  free ( fstats.weights );
  free ( fstats.stack );
  auxfstats_mng_free_auxfstats ( afmng, fstats.aux, fstats.T );
  
} /* end fstats_free */


static void
fstats_init (
	     fstats_t        *fstats,
	     auxfstats_mng_t *afmng
	     )
{
  
  mt ( fstats->v= mem_alloc ( fstats_node_t, 1 ) );
  fstats->v[0].p= -HUGE_VAL;
  fstats->v[0].a= FALSE;
  mt ( fstats->weights= mem_alloc ( tLProb, 1 ) );
  fstats->aux= auxfstats_mng_alloc_auxfstats ( afmng, 1 );
  mt ( fstats->stack= mem_alloc ( int, 1 ) );
  fstats->N= 0;
  fstats->T= 1;
  
} /* end fstats_init */


static void
fstats_require_size (
		     fstats_t        *fstats,
		     auxfstats_mng_t *afmng,
		     int              size
		     )
{
  
  int t;
  
  
  if ( size <= fstats->T ) return;
  mt ( fstats->v= mem_realloc ( fstats_node_t, fstats->v, size ) );
  mt ( fstats->weights= mem_realloc ( tLProb, fstats->weights, size ) );
  fstats->aux= auxfstats_mng_resize_auxfstats ( afmng, fstats->aux,
						fstats->T, size );
  mt ( fstats->stack= mem_realloc ( int, fstats->stack, size ) );
  for ( t= fstats->T; t < size; ++t )
    {
      fstats->v[t].p= -HUGE_VAL;
      fstats->v[t].a= 0;
    }
  fstats->T= size;
  
} /* end fstats_require_resize */

/* FORWARD-BACKWARD ***********************************************************/
static void
accum_weight (
	      fstats_t      *fstats,
	      const int      t,
	      const tLFloat  weight
	      )
{
  
  if ( weight != -HUGE_VAL )
    {
      if ( fstats->weights[t] == -HUGE_VAL )
	fstats->weights[t]= weight;
      else
	fstats->weights[t]= tL_radd ( fstats->weights[t], weight );
    }
  
} /* end accum_weight */


static void
accum_trans (
             const tLFloat       count0,
             const tLFloat       count1,
             const tLProb        prob,
             const unsigned int  q,
             logcount_t         *p
             )
{
  
  tLFloat aux;
  
  
  aux= count0 - prob;
  if ( aux != -HUGE_VAL ) p[q].p0= tL_radd ( p[q].p0, aux );
  aux= count1 - prob;
  if ( aux != -HUGE_VAL ) p[q].p1= tL_radd ( p[q].p1, aux );
  
} /* end accum_trans */


static size_t
mbasefb_get_length (
		    void *udata
		    )
{
  return MBASEFB(udata)->smodel->length;
} /* end mbasefb_get_length */


static const tLAModelHMM *
mbasefb_a_set_sym (
		   const size_t  sym,
		   void         *udata
		   )
{
  
  unsigned int c;
  const smodel_t *smodel;
  
  
  smodel= MBASEFB(udata)->smodel;
  c= smodel->seq[sym];
  MBASEFB(udata)->fstats_c= MBASEFB(udata)->fstats[c];
  
  return smodel->hmms[c];
  
} /* end mbasefb_a_set_sym */


static tLFloat
mbasefb_a_get_prob (
		    const tLFea        *fea,
		    const int           t,
		    void               *states,
		    const unsigned int  q,
		    void               *udata
		    )
{
  
  fstats_t *fstats_c;
  
  
  fstats_c= MBASEFB(udata)->fstats_c;
  
  return fstats_c[q].v[t].a ? fstats_c[q].v[t].p :
    fstats_add ( &(fstats_c[q]), MBASEFB(udata)->afmng,
		 fea, t, MBASEFB(udata)->global, states, q  );
  
} /* end mbasefb_a_get_prob */


static const tLAModelHMM *
mbasefb_b_set_sym (
		   const size_t  sym,
		   void         *udata
		   )
{
  
  unsigned int c;
  const smodel_t *smodel;
  
  
  smodel= MBASEFB(udata)->smodel;
  c= smodel->seq[sym];
  MBASEFB(udata)->fstats_c= MBASEFB(udata)->fstats[c];
  MBASEFB(udata)->ccount= smodel->counts[c];
  
  return smodel->hmms[c];
  
} /* end mbasefb_b_set_sym */


static void *
mlinfb_b_get_fstats (
		     const unsigned int  q,
		     void               *udata
		     )
{
  return &(MLINFB(udata)->fstats_c[q]);
} /* end mlinfb_b_get_fstats */


static tLFloat
mlinfb_b_get_prob (
		   void      *fstats,
		   const int  t
		   )
{
  return ((fstats_t *) fstats)->v[t].p;
} /* end mlinfb_b_get_prob */


static void
mlinfb_b_accum_weight (
		       void          *fstats,
		       const int      t,
		       const tLFloat  weight,
		       void          *udata
		       )
{
  accum_weight ( (fstats_t *) fstats, t, weight );
} /* end mlinfb_b_accum_weight */


static void
mlinfb_b_accum_trans (
		      const tLFloat  count0,
		      const tLFloat  count1,
		      unsigned int   q,
		      void          *udata
		      )
{
  accum_trans ( count0, count1, MLINFB(udata)->prob,
		q, MLINFB(udata)->ccount->trans.lineal.p );
} /* end mlinfb_b_accum_trans */


static void
mfb_b_reset_counts (
		    const unsigned int  c,
		    void               *udata
		    )
{
  MFB(udata)->ccount= &(MFB(udata)->counts->hmms[c]);
} /* end mfb_b_reset_counts */


static tLFloat
mfb_b_get_prob (
		const unsigned int  q,
		const int           t,
		void               *udata
		)
{
  return MFB(udata)->fstats_c[q].v[t].p;
} /* end mfb_b_get_prob */


static void
mfb_b_accum_weight (
		    const unsigned int  q,
		    const int           t,
		    const tLFloat       weight,
		    void               *udata
		    )
{
  accum_weight ( &( MFB(udata)->fstats_c[q]), t, weight );
} /* end mfb_b_accum_weight */


static void
mfb_b_accum_trans_p0 (
		      const tLFloat       weight,
		      const unsigned int  q,
		      void               *udata
		      )
{
  
  logcount_t *counts_p;
  
  
  counts_p= MFB(udata)->ccount->trans.lineal.p;
  counts_p[q].p0= tL_radd ( counts_p[q].p0, weight-MFB(udata)->prob );
  
} /* end mfb_b_accum_trans_p0 */


static void
mfb_b_accum_trans_p1 (
		      const tLFloat       weight,
		      const unsigned int  q,
		      void               *udata
		      )
{
  
  logcount_t *counts_p;
  
  
  counts_p= MFB(udata)->ccount->trans.lineal.p;
  counts_p[q].p1= tL_radd ( counts_p[q].p1, weight-MFB(udata)->prob );
  
} /* end mfb_b_accum_trans_p1 */


static void
mfb_b_accum_trans_fp (
		      const tLFloat       weight,
		      const unsigned int  q,
		      void               *udata
		      )
{
  
  tLFloat *ccounts_fp;
  
  
  ccounts_fp= MFB(udata)->ccount->trans.list.fp;
  ccounts_fp[q]= tL_radd ( ccounts_fp[q], weight-MFB(udata)->prob );
  
} /* end mfb_b_accum_trans_fp */


static void
mfb_b_accum_trans_edge (
			const tLFloat  weight,
			ptrdiff_t      pos,
			void          *udata
			)
{
  
  tLFloat *ced;
  
  
  ced= MFB(udata)->ccount->trans.list.edges;
  ced[pos]= tL_radd ( ced[pos],  weight-MFB(udata)->prob );
  
} /* end mfb_b_accum_trans_edge */


static const linfb_meths_t mlinfb_meths=
  {
    mbasefb_get_length,
    mbasefb_a_set_sym,
    mbasefb_a_get_prob,
    mbasefb_b_set_sym,
    mlinfb_b_get_fstats,
    mlinfb_b_get_prob,
    mlinfb_b_accum_weight,
    mlinfb_b_accum_trans
  };

static const fb_meths_t mfb_meths=
  {
    mbasefb_get_length,
    mbasefb_a_set_sym,
    mbasefb_a_get_prob,
    mbasefb_b_set_sym,
    mfb_b_reset_counts,
    mfb_b_get_prob,
    mfb_b_accum_weight,
    mfb_b_accum_trans_p0,
    mfb_b_accum_trans_p1,
    mfb_b_accum_trans_fp,
    mfb_b_accum_trans_edge
  };


/* MAUX ***********************************************************************/
static void
maux_free (
	   maux_t *maux
	   )
{
  
  unsigned int c, q;
  
  
  if ( maux->occs_calc != NULL ) free ( maux->occs_calc );
  if ( maux->type == LINEAR )
    {
      free ( maux->abmat.lin.fb );
      tL_linabmat_free ( maux->abmat.lin.m );
    }
  else /* GENERIC */
    {
      free ( maux->abmat.gen.fb );
      tL_abmat_free ( maux->abmat.gen.m );
    }
  for ( c= 0; c < maux->max_C; ++c )
    for ( q= 0; q < maux->max_Q; ++q )
      fstats_free ( maux->fstats[c][q], maux->afmng );
  free ( maux->fstats[0] );
  free ( maux->fstats );
  auxfstats_mng_free ( maux->afmng );
  free ( maux );
  
} /* end maux_free */


static maux_t *
maux_new (
	  const tLAModel  *amodel,
	  tLAMlogCounts   *counts,
	  smodels_t        smodels,
	  int              type,
	  const tLBool     occs,
	  char           **err
	  )
{
  
  maux_t *new;
  fstats_t **fstats, *fstats_mem;
  unsigned int c, q;
  mlinfb_t *linfb;
  mfb_t *fb;
  
  
  mt ( new= mem_alloc ( maux_t, 1 ) );
  
  new->max_C= smodels->max_C;
  new->max_Q= smodels->max_Q;
  new->type= type;
  new->max_width= 1;
  
  /* afmng. */
  new->afmng= tL_auxfstats_mng_new ( amodel->type, amodel->global, err );
  if ( new->afmng == NULL )
    {
      free ( new );
      return NULL;
    }
  
  /* fstats. */
  mt ( new->fstats= fstats= mem_alloc ( fstats_t *, smodels->max_C ) );
  mt ( fstats[0]= fstats_mem=
       mem_alloc ( fstats_t, smodels->max_C * smodels->max_Q ) );
  for ( c= 0; c < smodels->max_C; ++c, fstats_mem+= smodels->max_Q )
    {
      fstats[c]= fstats_mem;
      for ( q= 0; q < smodels->max_Q; ++q )
	fstats_init ( &(fstats_mem[q]), new->afmng );
    }
  
  /* abmat. */
  if ( type == LINEAR )
    {
      new->abmat.lin.m= tL_linabmat_new ( amodel, smodels->max_length );
      mt ( new->abmat.lin.fb= linfb= mem_alloc ( mlinfb_t, 1 ) );
      linfb->fstats= new->fstats;
      linfb->afmng= new->afmng;
      linfb->global= amodel->global;
    }
  else /* GENERIC */
    {
      new->abmat.gen.m= tL_abmat_new ( amodel, smodels->max_length );
      mt ( new->abmat.gen.fb= fb= mem_alloc ( mfb_t, 1 ) );
      fb->fstats= new->fstats;
      fb->afmng= new->afmng;
      fb->global= amodel->global;
      fb->counts= counts;
    }
  
  /* occs_calc. */
  if ( occs )
    {
      mt ( new->occs_calc= mem_alloc ( tLFloat, new->max_width ) );
    }
  else new->occs_calc= NULL;
  
  return new;
  
} /* end maux_new */


/* FORWARD-BACKWARD ***********************************************************/
static int
itertrain_sample_init (
		       void         *global,
		       smodel_t     *smodel,
		       maux_t       *maux,
		       const tLFea  *fea,
		       unsigned int  max_length,
		       unsigned int  max_Q
		       )
{
  
  int diff;
  unsigned int c, q;
  fstats_t **fstats, *fstats_c;
  const tLAModelHMM *cchar;
  
  
  /* afmng. */
  auxfstats_mng_init ( maux->afmng, global, fea );
  
  /* fstats. */
  fstats= maux->fstats;
  for ( c= 0; c < smodel->C; ++c )
    {
      fstats_c= fstats[c];
      cchar= smodel->hmms[c];
      for ( q= 0; q < cchar->Q; ++q )
	fstats_require_size ( &(fstats_c[q]), maux->afmng, fea->cols );
    }
  
  /* 'm' */
  if ( maux->type == LINEAR )
    {
      if ( fea->cols > maux->abmat.lin.m->max_width )
	tL_linabmat_resize ( maux->abmat.lin.m, max_length, max_Q, fea->cols );
      maux->abmat.lin.fb->smodel= smodel;
    }
  else /* GENERIC */
    {
      if ( fea->cols > maux->abmat.gen.m->max_width )
	tL_abmat_resize ( maux->abmat.gen.m, max_length, max_Q, fea->cols );
      maux->abmat.gen.fb->smodel= smodel;
    }
  
  /* occs_calc. */
  if ( maux->occs_calc != NULL && fea->cols > maux->max_width )
    {
      mt ( maux->occs_calc=
	   mem_realloc ( tLFloat, maux->occs_calc, fea->cols ) );
    }
  
  /* 'max_width'. */
  maux->max_width= fea->cols;
  
  /* Calcula diff. */
  diff= fea->cols - smodel->nstates + 1;
  
  return diff;
  
} /* end intertrain_sample_init */


static void
clear_fstats (
	      smodel_t  *smodel,
	      fstats_t **fstats
	      )
{
  
  unsigned int c, q;
  const tLAModelHMM *cchar;
  fstats_t *fstats_c;
  
  
  for ( c= 0; c < smodel->C; ++c )
    {
      cchar= smodel->hmms[c];
      fstats_c= fstats[c];
      for ( q= 0; q < cchar->Q; ++q )
	fstats_clear ( &(fstats_c[q]) );
    }
  
} /* end clear_fstats */


static void
accum_fstats (
	      maux_t         *maux,
	      smodel_t       *smodel,
	      const tLFea    *fea,
	      const tLAModel *amodel,
	      tLAMlogCounts  *counts,
	      const tLProb    prob
	      )
{
  
  void *global, *cglobal;
  unsigned int c, q;
  const tLAModelHMM *cchar;
  logamcount_hmm_t *ccount;
  fstats_t **fstats, *fstats_c;
  auxfstats_mng_t *afmng;
  tLFloat *occs, *occs_calc;
  
  
  global= amodel->global;
  cglobal= counts->global;
  afmng= maux->afmng;
  fstats= maux->fstats;
  occs_calc= maux->occs_calc;
  for ( c= 0; c < smodel->C; ++c )
    {
      cchar= smodel->hmms[c];
      ccount= smodel->counts[c];
      occs= (smodel->occs==NULL) ? NULL : smodel->occs[c];
      fstats_c= fstats[c];
      for ( q= 0; q < cchar->Q; ++q )
	fstats_accum ( &(fstats_c[q]), afmng, fea, global, cglobal,
		       cchar->states, ccount->states, occs, occs_calc,
		       q, prob );
    }
  auxfstats_mng_pseudoM_step_global ( afmng, fea, global, cglobal );
  
} /* end accum_fstats */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_trainer_free (
		 tLTrainer *trainer
		 )
{
  
  maux_free ( trainer->maux );
  tL_free_smodels ( trainer->smodels );
  free ( trainer );
  
} /* end tL_trainer_free */


TL_API_FUNC(tLTrainer *)
tL_trainer_new (
		tLAModel       *amodel,
		tLAMlogCounts  *counts,
		const tLSeqs   *seqs,
		tLFloat       **occs,
		char          **err
		)
{
  
  tLTrainer *new;
  int type;

  
  if ( amodel->tflags == FLINEAL ) type= LINEAR;
  else
    {
      type= GENERIC;
      tL_amodel_alloc_all_itrans ( amodel );
    }
  mt ( new= mem_alloc ( tLTrainer, 1 ) );
  new->amodel= amodel;
  new->seqs= seqs;
  new->counts= counts;
  new->occs= occs;
  new->smodels= tL_create_smodels ( amodel, counts, NULL, occs, seqs );
  new->maux= maux_new ( amodel, counts, new->smodels, type, occs!=NULL, err );
  if ( new->maux == NULL )
    {
      tL_free_smodels ( new->smodels );
      free ( new );
      return NULL;
    }
  new->type= type;
  
  return new;
  
} /* end tL_trainer_new */


TL_API_FUNC(tLProb)
tL_train (
	  tLTrainer    *trainer,
          const tLFea  *fea,
          const size_t  seq
          )
{
  
  maux_t *maux;
  smodel_t *smodel;
  int diff;
  tLProb prob;
  
  
  /* Prepara. */ 
  maux= trainer->maux;
  smodel= &(trainer->smodels->models[seq]);
  diff= itertrain_sample_init ( trainer->amodel->global,
				smodel, maux, fea,
				trainer->smodels->max_length,
				trainer->smodels->max_Q );
  
  /* Forward-Backward. */
  if ( trainer->type == LINEAR )
    {
      if ( diff <= 0 ) return -HUGE_VAL;
      maux->abmat.lin.fb->prob= prob=
	tL_calc_alpha_lin ( maux->abmat.lin.m, &mlinfb_meths,
			    fea, diff, 0.0, maux->abmat.lin.fb );
      if ( prob == -HUGE_VAL ) goto clear;
      tL_calc_beta_lin ( maux->abmat.lin.m, &mlinfb_meths,
			 fea->cols, diff, maux->abmat.lin.fb );
    }
  else /* GENERIC */
    {
      maux->abmat.gen.fb->prob= prob=
	tL_calc_alpha ( maux->abmat.gen.m, &mfb_meths, fea,
			tL_smodel_get_dist ( smodel ),
			maux->abmat.gen.fb );
      if ( prob == -HUGE_VAL ) goto clear;
      tL_calc_beta ( maux->abmat.gen.m, &mfb_meths, fea->cols,
		     maux->abmat.gen.fb );
      
    }
  
  /* Acumula comptadors. */
  accum_fstats ( maux, smodel, fea, trainer->amodel, trainer->counts, prob );
  
  return prob;
  
 clear:
  clear_fstats ( smodel, maux->fstats );
  return -HUGE_VAL;
  
} /* end tL_train */
