/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  itertrain.h - Per a fer una iteració de EM.
 *
 *  Adrià Giménez Pastor, 2008,2009,2011,2013.
 *
 */

#ifndef __LIBTL_ITERTRAIN_H__
#define __LIBTL_ITERTRAIN_H__

#include <stddef.h>

#include "amcounts_public.h"
#include "amodel_public.h"
#include "port.h"
#include "samples_public.h"
#include "text_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup EM EM training from transcriptions of acoustic models
 */
/** \addtogroup EM
 *  @{
 */

#ifndef __TL_PRV
/** Used in 'tL_train'. This structure contains auxiliary data used by
 *  'tL_train' to train an acoustic model from transcriptions.
 */
struct _tLTrainer
{
  
  const tLAModel  *amodel;    /**< Reference to the acoustic model
                                 used by the model trainer. */
  const tLSeqs    *seqs;      /**< Reference to the symbol sequence
                                 manager used by the trainer. */
  tLAMlogCounts   *counts;    /**< Reference to the acoustic model
	 			counts used. */
  tLFloat        **occs;      /**< Reference to the matrix to store
				 the occupancy values of the
				 states. Could be NULL. */
  
};
#endif

/** Used in 'tL_train'. This structure contains auxiliary data used by
 *  'tL_train' to train an acoustic model from transcriptions.
 */
typedef struct _tLTrainer tLTrainer;

/** Frees memory. Frees the memory allocated for the trainer.
 *
 *  @param trainer The trainer.
 */
TL_API_FUNC(void)
tL_trainer_free (
		 tLTrainer *trainer
		 );

/** Gets a tLTrainer. This function creates a trainer used to train
 *  an acoustic model from transcriptions using the EM algorithm. This
 *  data can only be used to train the symbol sequences stored in
 *  seqs during the creation of the auxiliary data. Currently
 *  supported types are: Bernoulli, DGaussian and Mixture.
 *
 *  @param amodel The acoustic model.
 *  @param counts The logarithmic counts which will be increased.
 *  @param seqs A symbol sequence manager containing the symbol
 *  sequences which can be used to train samples.
 *  @param occs If not NULL, it must be a pointer to a matrix in which
 *  the state occupancy values will be accumulated.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @return Trainer, or NULL in case of error.
 */
TL_API_FUNC(tLTrainer *)
tL_trainer_new (
		tLAModel       *amodel,
		tLAMlogCounts  *counts,
		const tLSeqs   *seqs,
		tLFloat       **occs,
		char          **err
		);

/** Updates an acoustic model's counts according to the EM algorithm. This
 *  function increases the acoustic model's counts, which are stored in
 *  the provided trainer, from one sample (feature vector sequence)
 *  and its transcription, using the EM algorithm in order to obtain
 *  the logarithmic counts. The sequence is specified as an integer
 *  code according to the sequence manager used during the creation of
 *  the trainer.
 *
 *  @param trainer The trainer.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param seq The code of the transcription of the sample.
 *  @return The sample probability for that transcription; -HUGE_VAL
 *  in case the sample cannot be used to update the counts.
 */
TL_API_FUNC(tLProb)
tL_train (
	  tLTrainer    *trainer,
          const tLFea  *fea,
          const size_t  seq
          );

/** @} */
/** @} */

#endif /* __LIBTL_ITERTRAIN_H__ */
