/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  itertrain_viterbi.c - Implementació de 'itertrain_viterbi.h'.
 *
 *  Adrià Giménez Pastor, 2010,2011,2013.
 *
 */

#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "common.h"
#include "error.h"
#include "iamcounts.h"
#include "itertrain_viterbi.h"
#include "mem.h"
#include "smodels.h"
#include "statestr.h"




/*********/
/* TIPUS */
/*********/

/* Node del trellis. */
typedef struct
{
  
  tLProb p;
  tLBool jump;    /* A cert si és l'anterior estat no és ell. */
  
} mauxv_mnode_t;

/* Memòria auxiliar per a càlculs. */
typedef struct mauxv
{
  
  tLBool        ***actives;
  tLProb        ***probs;
  int            **ends;
  int              max_width;
  statestrv_t     *str;
  void           **trs;
  mauxv_mnode_t  **m;
  
} mauxv_t;

struct _tLVTrainer
{
  
  const tLAModel *amodel;
  const tLSeqs   *seqs;
  tLAMiCounts    *counts;
  smodels_t       smodels;
  mauxv_t        *maux;
  
};



/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* Es suposa que és necessari. */
static void
resize_maux_width (
                   mauxv_t      *maux,
                   unsigned int  max_C,
                   unsigned int  max_Q,
		   size_t        max_length,
                   int           width
                   )
{
  
  unsigned int c, q, i;
  tLBool *aux_active;
  tLProb *aux_prob;
  size_t old_size, new_size;
  int t;
  mauxv_mnode_t *aux_m;
  
  
  /* 'actives' i 'probs'. */
  old_size= max_C * max_Q * maux->max_width;
  new_size= max_C * max_Q * width;
  mt ( maux->probs[0][0]=
       mem_realloc ( tLProb, maux->probs[0][0], new_size ) );
  mt ( maux->actives[0][0]=
       mem_realloc ( tLBool, maux->actives[0][0], new_size ) );
  
  aux_active= maux->actives[0][0];
  for ( i= old_size; i < new_size; ++i )
    aux_active[i]= FALSE;
  
  for ( c= 0, aux_prob= maux->probs[0][0],
	  aux_active= maux->actives[0][0];
	c < max_C; ++c )
    for ( q= 0; q < max_Q;
	  ++q, aux_prob+= width, aux_active+= width )
      {
	maux->probs[c][q]= aux_prob;
	maux->actives[c][q]= aux_active;
      }
  
  /* 'm'. Reserve un de més per al marge. */
  new_size= (max_length * max_Q)+1;
  mt ( maux->m= mem_realloc ( mauxv_mnode_t *, maux->m, width ) );
  mt ( maux->m[0]= mem_realloc ( mauxv_mnode_t,
				 maux->m[0], new_size * width ) );
  for ( t= 0, aux_m= maux->m[0]; t < width; ++t, aux_m+= new_size )
    maux->m[t]= aux_m;
  
  /* 'max_width'. */
  maux->max_width= width;
  
} /* end resize_maux_width */


static void
viterbi_iter (
	      mauxv_mnode_t    *mt_1,
	      mauxv_mnode_t    *mt,
	      const smodel_t   *smodel,
	      tLProb         ***probs,
	      const long        begin,
	      const long        end,
	      const int         t
	      )
{
  
  tLProb aux, prev, aux2, prob;
  long remain;
  unsigned int c, nstates, Q;
  const tLAModelHMM *cchar;
  tLProb **probs_c;
  const tLProbPair *pc;
  smodel_state_t id;
  
  
  /* Inicialitza. */
  remain= end-begin;
  mt_1+= begin; mt+= begin;
  if ( begin == 0 ) prev= -HUGE_VAL;
  else
    {
      prev= (mt_1-1)->p;
      id= smodel->smap[begin-1];
      prev+= smodel->hmms[smodel->seq[id.s]]->trans.lineal.p[id.q].p1;
    }
  id= smodel->smap[begin];
  
  /* Cos. */
  while ( remain )
    {
      c= smodel->seq[id.s];
      cchar= smodel->hmms[c];
      probs_c= probs[c];
      pc= cchar->trans.lineal.p;
      Q= cchar->Q; nstates= Q-id.q;
      if ( nstates > remain ) { Q= remain+id.q; remain= 0; }
      else remain-= nstates;
      for ( ; id.q < Q; ++id.q, ++mt_1, ++mt )
	{
	  aux= prev;
	  prev= mt_1->p;
	  aux2= prev+pc[id.q].p0;
	  prob= probs_c[id.q][t];
	  if ( aux > aux2 ) { mt->p= aux + prob; mt->jump= TRUE; }
	  else              { mt->p= aux2 + prob; mt->jump= FALSE; }
	  prev+= pc[id.q].p1;
	}
      ++id.s; id.q= 0;
    }
  mt->p= -HUGE_VAL;
  
} /* end viterbi_iter */


static tLProb
viterbi (
	 mauxv_mnode_t   **m,
	 const smodel_t   *smodel,
	 const int         T,
	 tLProb         ***probs
	 )
{
  
  long Q, diff, begin, end;
  int t;
  smodel_state_t id;
  
  
  /* Inicialitza. */
  Q= smodel->nstates;
  diff= Q-T;
  m[0][0].p= probs[smodel->seq[0]][0][0];
  m[0][1].p= -HUGE_VAL;
  
  /* Cos. */
  for ( t= 1; t < T; ++t )
    {
      begin= diff+t; begin= MAX ( 0, begin );
      end= t+1; end= MIN ( Q, end );
      viterbi_iter ( m[t-1], m[t], smodel, probs, begin, end, t );
    }
  
  return m[T-1][Q-1].p +
    smodel->hmms[smodel->seq[id.s]]->trans.lineal.p[id.q].p1;
  
} /* end viterbi */


static void
accum (
       statestrv_t      *str,
       void             *global,
       void            **trs,
       mauxv_mnode_t   **m,
       const smodel_t   *smodel,
       const int         T
       )
{
  
  unsigned long q;
  int t;
  smodel_state_t id;
  const smodel_state_t *smap;
  const unsigned int *seq;
  tLAModelHMM **chars;
  const tLAModelHMM *cchar;
  iamcount_hmm_t **icounts, *ccount;
  unsigned int c;
  void *trs_c;
  
  
  smap= smodel->smap;
  q= smodel->nstates-1;
  id= smap[q];
  seq= smodel->seq;
  chars= smodel->hmms; icounts= smodel->icounts;
  c= seq[id.s];
  cchar= chars[c]; ccount= icounts[c];
  trs_c= trs[c];
  for ( t= T-1; t > 0; --t )
    {
      statestrv_accum ( str, global, t, id.q, cchar->states,
			ccount->states, trs_c );
      if ( m[t][q].jump )
	{
	  id= smap[--q];
	  c= seq[id.s];
	  cchar= chars[c];
	  ccount= icounts[c];
	  trs_c= trs[c];
	}
    }
  statestrv_accum ( str, global, t, id.q, cchar->states,
		    ccount->states, trs_c );
  
} /* end accum */


static int
itertrain_sample_init (
                       smodel_t     *smodel,
                       mauxv_t      *maux,
		       unsigned int  max_C,
		       unsigned int  max_Q,
		       size_t        max_length,
		       const tLFea  *fea,
                       void         *global
                       )
{
  
  size_t s, begin, begin2, end, t;
  unsigned int c, q;
  int diff;
  tLBool *aux;
  tLAModelHMM **chars;
  
  
  /* Prepara memòria auxiliar. */
  if ( smodel->smap == NULL ) tL_smodel_create_smap ( smodel );
  chars= smodel->hmms;
  statestrv_get_ready ( maux->str, fea, global );
  if ( fea->cols > maux->max_width )
    resize_maux_width ( maux, max_C, max_Q,
                        max_length, fea->cols );
  
  /* Inicialitza els vectors actius. */
  if ( fea->cols < smodel->nstates ) return -1;
  for ( c= 0; c < smodel->C; ++c )
    for ( q= 0; q < chars[c]->Q; ++q )
      {
        aux= maux->actives[c][q];
        for ( t= 0; t < fea->cols; ++t )
          aux[t]= FALSE;
      }
  diff= fea->cols - smodel->nstates + 1;
  for ( s= 0, begin= 0, end= diff;
        s < smodel->length; ++s )
    {
      c= smodel->seq[s];
      for ( q= 0; q < chars[c]->Q; ++q, ++begin, ++end )
        {
          aux= maux->actives[c][q];
          begin2= MAX ( begin, maux->ends[c][q] );
          for ( t= begin2; t < end; ++t )
            aux[t]= TRUE;
          maux->ends[c][q]= end;
        }
    }
  for ( c= 0; c < smodel->C; ++c )
    for ( q= 0; q < chars[c]->Q; ++q )
      maux->ends[c][q]= 0;
  
  return diff;
  
} /* end itertrain_sample_init */


static mauxv_t *
create_mauxv (
	      const tLAModel *amodel,
	      smodels_t       smodels
	      )
{
  
  mauxv_t *new;
  unsigned int c, w, i;
  tLBool **aux_active, *aux_active2;
  tLProb **aux_prob, *aux_prob2;
  int *aux_ends;
  void *aux_trs;
  size_t trs_size;
  strv_factory_t *strf;
  
  
  mt ( new= mem_alloc ( mauxv_t, 1 ) );
  
  /* 'str'. */
  strf= tL_create_strv_factory ( amodel->type, amodel->global );
  for ( c= 0; c < amodel->N; ++c )
    if ( smodels->active[c] )
      strv_factory_add ( strf,
			 amodel->hmms[c].states,
			 amodel->hmms[c].Q );
  new->str= tL_strv_factory_make ( strf,
				amodel->global,
				smodels->max_C *
				smodels->max_Q );
  free_strv_factory ( strf );
  
  /* 'trs'. */
  trs_size= statestrv_sizeof ( new->str, smodels->max_Q );
  mt ( new->trs= mem_alloc ( void *, smodels->max_C ) );
  aux_trs= new->str->statetrs;
  for ( c= 0; c < smodels->max_C; ++c )
    {
      new->trs[c]= aux_trs;
      aux_trs= (void *) (((char *) aux_trs) + trs_size);
    }
  
  /* 'probs', 'actives' i 'max_width'. */
  mt ( new->probs= mem_alloc ( tLProb **,
                               smodels->max_C ) );
  mt ( new->actives= mem_alloc ( tLBool **,
                                 smodels->max_C ) );
  mt ( new->probs[0]=
       mem_alloc ( tLProb *, smodels->max_C * smodels->max_Q ) );
  mt ( new->actives[0]=
       mem_alloc ( tLBool *, smodels->max_C * smodels->max_Q ) );
  for ( c= 1, aux_prob= new->probs[0] + smodels->max_Q,
	  aux_active= new->actives[0] + smodels->max_Q;
        c < smodels->max_C; ++c, aux_prob+= smodels->max_Q,
	  aux_active+= smodels->max_Q )
    {
      new->probs[c]= aux_prob;
      new->actives[c]= aux_active;
    }
  mt ( new->probs[0][0]=
       mem_alloc ( tLProb, smodels->max_C * smodels->max_Q ) );
  mt ( new->actives[0][0]=
       mem_alloc ( tLBool, smodels->max_C * smodels->max_Q ) );
  aux_active2= new->actives[0][0];
  for ( c= 0, i= 0; c < smodels->max_C; ++c )
    for ( w= 0; w < smodels->max_Q; ++w, ++i )
      aux_active2[i]= FALSE;
  for ( c= 0, aux_prob2= new->probs[0][0],
	  aux_active2= new->actives[0][0];
        c < smodels->max_C; ++c )
    for ( w= 0; w < smodels->max_Q;
          ++w, ++aux_prob2, ++aux_active2 )
      {
	new->probs[c][w]= aux_prob2;
	new->actives[c][w]= aux_active2;
      }
  new->max_width= 1;
  
  /* 'ends'. */
  mt ( new->ends= mem_alloc ( int *, smodels->max_C ) );
  mt ( new->ends[0]= mem_calloc ( int, smodels->max_Q *
                                  smodels->max_C ) );
  for ( c= 1, aux_ends= new->ends[0] + smodels->max_Q;
        c < smodels->max_C; ++c, aux_ends+= smodels->max_Q  )
    new->ends[c]= aux_ends;
  
  /* 'm'. Reserve un de més per al marge de baix. */
  mt ( new->m= mem_alloc ( mauxv_mnode_t *, 1 ) );
  mt ( new->m[0]= mem_alloc ( mauxv_mnode_t, (smodels->max_length
					      * smodels->max_Q)+1 ) );
  
  return new;
  
} /* end create_mauxv */


static void
free_mauxv (
	    mauxv_t *maux
	    )
{
  
  free ( maux->m[0] );
  free ( maux->m );
  free_statestrv ( maux->str );
  free ( maux->ends[0] );
  free ( maux->ends );
  free ( maux->actives[0][0] );
  free ( maux->actives[0] );
  free ( maux->actives );
  free ( maux->probs[0][0] );
  free ( maux->probs[0] );
  free ( maux->probs );
  free ( maux->trs );
  free ( maux );
  
} /* end free_mauxv */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_vtrainer_free (
		  tLVTrainer *vtrainer
		  )
{
  
  free_mauxv ( vtrainer->maux );
  tL_free_smodels ( vtrainer->smodels );
  free ( vtrainer );
  
} /* end tL_vtrainer_free */


TL_API_FUNC(tLVTrainer *)
tL_vtrainer_new (
		 const tLAModel  *amodel,
		 tLAMiCounts     *counts,
		 const tLSeqs    *seqs,
		 char           **err
		 )
{
  
  tLVTrainer *new;
  
  
  if ( amodel->tflags != FLINEAL )
    {
      tL_msgerror ( err, _( "tL_vtrainer_new is only implemented"
                    " for character models with lineal topologies" ) );
      return NULL;
    }
  mt ( new= mem_alloc ( tLVTrainer, 1 ) );
  new->amodel= amodel;
  new->seqs= seqs;
  new->counts= counts;
  new->smodels= tL_create_smodels ( amodel, NULL, counts, NULL, seqs );
  new->maux= create_mauxv ( amodel, new->smodels );
  
  return new;
  
} /* end tL_vtrainer_new */


TL_API_FUNC(tLProb)
tL_vtrain (
	   tLVTrainer   *vtrainer,
	   const tLFea  *fea,
	   const size_t  seq
	   )
{
  
  int diff;
  unsigned int c, q;
  tLBool **aux_actives;
  const tLAModelHMM *cchar;
  tLProb **aux_probs, prob;
  void *trs_c;
  mauxv_t *maux;
  smodel_t *smodel;
  
  
  /* Prepara. */
  maux= vtrainer->maux;
  smodel= &(vtrainer->smodels->models[seq]);
  diff= itertrain_sample_init ( smodel, maux, vtrainer->smodels->max_C,
				vtrainer->smodels->max_Q,
				vtrainer->smodels->max_length,
				fea, vtrainer->amodel->global );
  if ( diff == -1 ) return -HUGE_VAL;
  
  /* Probabilitats d'emissió. */
  for ( c= 0; c < smodel->C; ++c )
    {
      
      aux_actives= maux->actives[c];
      cchar= smodel->hmms[c];
      aux_probs= maux->probs[c];
      trs_c= maux->trs[c];
      
      for ( q= 0; q < cchar->Q; ++q )
	statestrv_get_probs ( maux->str, vtrainer->amodel->global, fea,
			      aux_actives[q], cchar->states,
			      trs_c, q, aux_probs[q] );
      
    }
  
  /* Viterbi. */
  prob= viterbi ( maux->m, smodel, fea->cols, maux->probs );
  if ( prob == -HUGE_VAL ) return prob;
  
  /* Actualitza comptadors estats. */
  accum ( maux->str, vtrainer->amodel->global, maux->trs,
	  maux->m, smodel, fea->cols );
  
  return prob;
  
} /* end tL_vtrain */
