/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  itertrain_viterbi.h - Per a fer una iteració d'entrenament amb
 *                        viterbi.
 *
 *  Adrià Giménez Pastor, 2010,2011,2013.
 *
 */

#ifndef __LIBTL_ITERTRAIN_VITERBI_H__
#define __LIBTL_ITERTRAIN_VITERBI_H__

#include <stddef.h>

#include "amcounts_public.h"
#include "amodel_public.h"
#include "port.h"
#include "samples_public.h"
#include "text_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup ViterbiTr Viterbi training from transcriptions of acoustic models
 */
/** \addtogroup ViterbiTr
 *  @{
 */

#ifndef __TL_PRV
/** Used in 'tL_vtrain'. This structure contains auxiliary data used
 *  by 'tL_vtrain' to train an acoustic model from transcriptions using
 *  the Viterbi algorithm.
 */
struct _tLVTrainer
{
  
  const tLAModel *amodel;    /**< Reference to the acoustic model
                                used by the Viterbi trainer. */
  const tLSeqs   *seqs;      /**< Reference to the symbol sequence
                                manager used by the Viterbi
                                trainer. */
  tLAMiCounts    *counts;    /**< Reference to the acoustic model
                                counts used. */
  
};
#endif

/** Used in 'tL_vtrain'. This structure contains auxiliary data used
 *  by 'tL_vtrain' to train an acoustic model from transcriptions using
 *  the Viterbi algorithm.
 */
typedef struct _tLVTrainer tLVTrainer;

/** Frees memory. Frees the memory allocated for the auxiliary data.
 *
 *  @param vtrainer The Viterbi trainer.
 */
TL_API_FUNC(void)
tL_vtrainer_free (
		  tLVTrainer *vtrainer
		  );

/** Gets a tLVTrainer. This function creates auxiliary data used to
 *  train an acoustic model from transcriptions using the Viterbi
 *  algorithm. This data can only be used to train the symbol
 *  sequences stored in seqs during the creation of the
 *  trainer. Currently supported types are: Bernoulli, DGaussian and
 *  Mixture.
 *
 *  @param amodel The acoustic model.
 *  @param counts The integer counts which will be increased.
 *  @param seqs A symbol sequence manager containing the symbol
 *  sequences which can be used to train samples.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @return Viterbi trainer, or NULL in case of error.
 */
TL_API_FUNC(tLVTrainer *)
tL_vtrainer_new (
		 const tLAModel  *amodel,
		 tLAMiCounts     *counts,
		 const tLSeqs    *seqs,
		 char           **err
		 );

/** Updates acoustic model counts according to the Viterbi
 *  algorithm. This function increases the acoustic model's counts,
 *  which are stored in the trainer, from one sample (feature vector
 *  sequence) and its transcription, using the Viterbi algorithm in
 *  order to obtain the integer counts. The sequence is specified as
 *  an integer code according to the sequence manager used during the
 *  creation of the viterbi trainer.
 *
 *  @param vtrainer The viterbi trainer.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param seq The code of the transcription of the sample.
 *  @return The sample probability for that transcription; -HUGE_VAL
 *  in case the sample cannot be used to update the counts.
 */
TL_API_FUNC(tLProb)
tL_vtrain (
	   tLVTrainer   *vtrainer,
	   const tLFea  *fea,
	   const size_t  seq
	   );

/** @} */
/** @} */

#endif /* __LIBTL_ITERTRAIN_VITERBI_H__ */
