/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lexicon.c - Implementació de 'lexicon.h'.
 *
 *  Adrià Giménez Pastor, 2008,2011,2012.
 *
 */


#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "error.h"
#include "hash.h"
#include "lexicon.h"
#include "mem.h"




/**********/
/* MACROS */
/**********/

#define HSIZE 32363U




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  size_t wc;
  size_t sym;
  size_t next;
  
} cwhash_node_t;


typedef struct
{
  
  size_t        *table;
  cwhash_node_t *nodes;    /* 0 és NULL. */
  size_t         N;
  size_t         size;
  
} cwhash_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* Torna -1 en cas d'error. */
static int
register_sym (
	      const char    *token,
	      const size_t   n,
	      size_t        *sym,
	      tLDict        *syms,
	      const tLBool   emptysyms,
	      const tLBool   isentry,
	      char         **err
	      )
{
  
  if ( !emptysyms )
    {
      *sym= tL_dict_find ( syms, token );
      if ( *sym == syms->N )
	{
	  if ( isentry )
	    tL_msgerror ( err, _( "In the %uth entry, the character '%s'"
			  " is out of vocabulary" ) , _PS(n), token );
	  else if ( n != 0 )
	    tL_msgerror ( err, _( "In the %uth crossword transition, the"
			  " character '%s' is out of vocabulary" ) ,
			  _PS(n), token );
	  else
	    tL_msgerror ( err, _( "The character '%s' is out of vocabulary" ) ,
			  token );
	  return -1;
	}
    }
  else tL_dict_insert ( syms, token, sym );
  
  return 0;
  
} /* end register_sym */


static cwhash_t *
create_cwhash (void)
{
  
  cwhash_t *new;
  size_t n;
  
  
  mt ( new= mem_alloc ( cwhash_t, 1 ) );
  mt ( new->table= mem_alloc ( size_t, HSIZE ) );
  for ( n= 0; n < HSIZE; ++n )
    new->table[n]= 0;
  mt ( new->nodes= mem_alloc ( cwhash_node_t, 2 ) );
  new->N= 1;
  new->size= 2;
  
  return new;
  
} /* end create_cwhash */


static void
free_cwhash (
	     cwhash_t *cwh
	     )
{
  
  free ( cwh->nodes );
  free ( cwh->table );
  free ( cwh );
  
} /* end free_cwhash */


static size_t
cwhash_find (
	     cwhash_t *cwh,
	     size_t    wc,
	     size_t    sym
	     )
{
  
  static unsigned int buffer[2];
  static const unsigned long size= sizeof(unsigned int)*2;
  unsigned long key;
  size_t pos;
  
  
  /* Calcula clau. */
  buffer[0]= (unsigned int) wc;
  buffer[1]= (unsigned int) sym;
  key= tL_hash_gnu_cxx ( buffer, size )%HSIZE;
  
  /* Busca. */
  for ( pos= cwh->table[key];
	pos != 0 && (cwh->nodes[pos].wc!=wc || cwh->nodes[pos].sym!=sym);
	pos= cwh->nodes[pos].next );
  return pos;
  
} /* end cwhash_find */


static size_t
cwhash_get_pos (
		cwhash_t *cwh,
		size_t    wc,
		size_t    sym
		)
{
  
  static unsigned int buffer[2];
  static const unsigned long size= sizeof(unsigned int)*2;
  unsigned long key;
  size_t pos;
  
  
  /* Calcula clau. */
  buffer[0]= (unsigned int) wc;
  buffer[1]= (unsigned int) sym;
  key= tL_hash_gnu_cxx ( buffer, size )%HSIZE;
  
  /* Busca. */
  for ( pos= cwh->table[key];
	pos != 0 && (cwh->nodes[pos].wc!=wc || cwh->nodes[pos].sym!=sym);
	pos= cwh->nodes[pos].next );
  if ( pos != 0 ) return pos;
  
  /* Inserta. */
  if ( cwh->N == cwh->size )
    {
      cwh->size*= 2;
      mt ( cwh->nodes= mem_realloc ( cwhash_node_t, cwh->nodes, cwh->size ) );
    }
  pos= cwh->N++;
  cwh->nodes[pos].wc= wc;
  cwh->nodes[pos].sym= sym;
  cwh->nodes[pos].next= cwh->table[key];
  cwh->table[key]= pos;
  
  return pos;
  
} /* end cwhash_get_pos */


/* Torna 0 si tot ha anat bé. En pos guarda la posició del node. */
static int
add_cwnode (
	    tLLexicon     *lexicon,
	    tLDict        *syms,
	    const tLBool   emptysyms,
	    cwhash_t      *cwh,
	    const char    *wc,
	    const char    *sym,
	    const size_t   n,
	    size_t        *pos,
	    char         **err
	    )
{
  
  size_t id, idsym;
  
  
  /* Codifica tokens. */
  if ( tL_dict_insert ( lexicon->wildcards, wc, &id ) )
    {
      if ( lexicon->_size_wcinits == lexicon->wildcards->N )
	{
	  lexicon->_size_wcinits*= 2;
	  mt ( lexicon->wcinits= mem_realloc ( size_t, lexicon->wcinits,
					       lexicon->_size_wcinits ) );
	}
      lexicon->wcinits[id]= 0;
    }
  if ( register_sym ( sym, n, &idsym, syms, emptysyms, FALSE, err ) )
    return -1;
  
  /* Obté la posició del node i comprova si és nou. */
  *pos= cwhash_get_pos ( cwh, id, idsym	);
  if ( *pos != (lexicon->NCW+1) ) return 0;
  
  /* Crea el node. */
  if ( *pos == lexicon->_size_cwnodes )
    {
      lexicon->_size_cwnodes*= 2;
      mt ( lexicon->cwnodes= mem_realloc ( tLLexiconCWNode, lexicon->cwnodes,
					   lexicon->_size_cwnodes ) );
    }
  lexicon->cwnodes[*pos].wildcard= id;
  lexicon->cwnodes[*pos].sym= idsym;
  lexicon->cwnodes[*pos].next= lexicon->wcinits[id];
  lexicon->wcinits[id]= *pos;
  ++(lexicon->NCW);
  
  return 0;
  
} /* end add_cwnode */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. El lèxic ha de ser
 * de crosswords. Crea també els nodes necessaris, registrant els
 * símbols i wildcards que calga. No comprova la repetició de
 * transicions.
 */
static int
add_cwtrans (
	     tLLexicon     *lexicon,
	     tLDict        *syms,
	     cwhash_t      *cwh,
	     const size_t   from,
	     const char    *dst_sym,
	     const char    *dst_wc,
	     const size_t   n,
	     char         **err
	     )
{
  
  size_t to, pos, wc, sym;
  
  
  
  wc= tL_dict_find ( lexicon->wildcards, dst_wc );
  if ( wc == lexicon->wildcards->N ) goto error;
  sym= tL_dict_find ( syms, dst_sym );
  if ( sym == syms->N ) goto error;
  to= cwhash_find ( cwh, wc, sym );
  if ( to == 0 ) goto error;
  pos= ++(lexicon->NCWT);
  if ( pos == lexicon->_size_cwtrans )
    {
      lexicon->_size_cwtrans*= 2;
      mt ( lexicon->cwtrans= mem_realloc ( tLLexiconCWTrans, lexicon->cwtrans,
					   lexicon->_size_cwtrans ) );
    }
  lexicon->cwtrans[pos].to= to;
  lexicon->cwtrans[pos].next= lexicon->cwitrans[from];
  lexicon->cwitrans[from]= pos;
  
  return 0;

 error:
  tL_msgerror ( err, _( "out of vocabulary crossword node (%s,%s)"
		" in the %uth crossword transition" ) ,
		dst_wc, dst_sym, _PS(n) );
  return -1;
  
} /* end add_cwtrans */


static tLLexicon *
read_cwlexicon_text (
		     gzFile         from,
		     tLDict        *syms,
		     const tLBool   emptysyms,
		     tLBuffer      *buffer,
		     char         **err
		     )
{
  
  tLLexicon *new;
  char *begin, *end, *word, *wc, *src_sym, *dst_sym, *dst_wc, *aux;
  tLSeq cs;
  size_t cs_size, n, sym;
  tLProb prob;
  int stop, stop2, readtoken, counter;
  cwhash_t *cwh;
  

  stop2= -1; /* CALLA!!! */
  new= NULL;
  cs.v= NULL;
  cwh= NULL;
  
  /* Capçalera i inicialitza. */
  new= tL_lexicon_new ( TL_LEXICON_CW );
  cs_size= 1;
  mt ( cs.v= mem_alloc ( size_t, 1 ) );
  cwh= create_cwhash ();
  
  /* WILDCARDS. */
  if ( !tL_gline ( from, buffer ) ) goto error_wildcards_tok;
  /* No me molesta que hi hasquen més tokens en la línia. */
  tL_gtokenstr ( buffer->v, &begin, &end );
  if ( begin == NULL ) goto error_wildcards_tok;
  *end= '\0';
  if ( strcmp ( begin, "WILDCARDS" ) ) TLMSG_UNKTOKID ( err, begin );
  n= 1;
  while ( (readtoken= tL_gline ( from, buffer )) )
    {
      
      /* Wildcard origen. Siga wildcard o CWSTART necessita almenys
	 dos tokens.*/
      if ( tL_gtokenstr ( buffer->v, &wc, &end ) )
	goto errorwildcards;
      *end= '\0';
      if ( !strcmp ( wc, "CWSTART" ) ) break;
      
      /* Llig símbols. */
      counter= 0;
      stop= FALSE;
      while ( !stop )
	{
	  stop= tL_gtokenstr ( end+1, &src_sym, &end );
	  if ( src_sym == NULL ) break;
	  *end= '\0';
	  if ( add_cwnode ( new, syms, emptysyms, cwh,
			    wc, src_sym, 0, &sym, err ) )
	    goto error;
	  ++counter;
	}
      if ( counter == 0 ) goto errorwildcards;
      ++n;
      
    }
  if ( !readtoken ) goto error_cwstart_tok;
  new->NS= new->_size_syms= syms->N;
  mt ( new->stopcw= mem_realloc ( tLBool, new->stopcw, syms->N ) );
  mt ( new->cwitrans= mem_realloc ( size_t, new->cwitrans, syms->N ) );
  for ( n= 0; n < syms->N; ++n )
    {
      new->stopcw[n]= FALSE;
      new->cwitrans[n]= 0;
    }
  
  /* Llig CWSTART. Ací estic segur de que encara no hem alcançat el
     final del buffer. */
  tL_gtokenstr ( end+1, &src_sym, &end );
  if ( src_sym == NULL ) goto errorcwstart;
  *end= '\0';
  sym= tL_dict_find ( syms, src_sym );
  if ( sym == syms->N ) goto error_src_sym_not_in_wc;
  new->initcw= sym;
  
  /* Llig CWSTOP. */
  if ( !tL_gline ( from, buffer ) ) goto error_cwstop_tok;
  tL_gtokenstr ( buffer->v, &begin, &end );
  if ( begin == NULL ) goto error_cwstop_tok;
  *end= '\0';
  if ( strcmp ( begin, "CWSTOP" ) ) TLMSG_UNKTOKID ( err, begin );
  counter= 0;
  stop= FALSE;
  while ( !stop )
    {
      stop= tL_gtokenstr ( end+1, &src_sym, &end );
      if ( src_sym == NULL ) break;
      *end= '\0';
      sym= tL_dict_find ( syms, src_sym );
      if ( sym == syms->N ) goto error_src_sym_not_in_wc;
      new->stopcw[sym]= TRUE;
      ++counter;
    }
  if ( counter == 0 ) goto errorcwstop;
  
  /* Llig les transicions de crosswords. */
  if ( !tL_gline ( from, buffer ) ) goto error_cwtrans_tok;
  /* No me molesta que hi hasquen més tokens en la línia. */
  tL_gtokenstr ( buffer->v, &begin, &end );
  if ( begin == NULL ) goto error_cwtrans_tok;
  *end= '\0';
  if ( strcmp ( begin, "CWTRANS" ) ) TLMSG_UNKTOKID ( err, begin );
  n= 1;
  while ( tL_gline ( from, buffer ) )
    {
      
      /* Símbol origen. */
      if ( tL_gtokenstr ( buffer->v, &src_sym, &end ) )
	{
	  if ( wc != NULL )
	    {
	      *end= '\0';
	      if ( !strcmp ( "ENTRIES", src_sym ) )
		break;
	    }
	  goto errorcwtrans;
	}
      *end= '\0';
      sym= tL_dict_find ( syms, src_sym );
      if ( sym == syms->N ) goto error_src_sym_not_in_wc;
      
      /* Símbol destí. */
      if ( tL_gtokenstr ( end+1, &dst_sym, &end ) )
	{
	  if ( dst_sym == NULL && !strcmp ( "ENTRIES", src_sym ) )
	    break;
	  goto errorcwtrans;
	}
      *end= '\0';
      
      /* wildcard destí. */
      stop= tL_gtokenstr ( end+1, &dst_wc, &end );
      if ( dst_wc == NULL ) goto errorcwtrans;
      *end= '\0';
      
      /* Comprova que no hi han més. */
      if ( !stop )
	{
	  tL_gtokenstr ( end+1, &begin, &end );
	  if ( begin != NULL ) goto errorcwtrans;
	}
      
      /* Inserta aresta. */
      if ( add_cwtrans ( new, syms, cwh, sym, dst_sym, dst_wc, n, err ) )
	goto error;
      
      ++n;
      
    }
  
  /* Llig entrades. */
  n= 1;
  while ( tL_gline ( from, buffer ) )
    {
      
      /* Paraula. */
      if ( tL_gtokenstr ( buffer->v, &begin, &end ) ) goto errorline;
      *end= '\0';
      word= begin;
      
      /* Probabilitat. */
      if ( tL_gtokenstr ( end+1, &begin, &end ) ) goto errorline;
      *end= '\0';
      prob= tL_atop ( begin );
      
      /* Seqüència de símbols. */
      cs.length= 0;
      stop= tL_gtokenstr ( end+1, &begin, &end );
      while ( begin != NULL )
	{
	  aux= begin;
	  *end= '\0';
	  if ( !stop ) stop2= tL_gtokenstr ( end+1, &begin, &end );
	  if ( cs.length == 0 || stop || begin == NULL )
	    {
	      sym= tL_dict_find ( new->wildcards, aux );
	      if ( sym == new->wildcards->N )
		{
		  tL_msgerror ( err, _( "Wildcard '%s' in the %uth entry is"
				" not related to any crossword transition" ) ,
				aux, _PS(n));
		  goto error;
		}
	    }
	  else
	    {
	      if ( register_sym ( aux, n, &sym, syms,
				  emptysyms, TRUE, err ) == -1 )
		goto error;
	    }
	  if ( cs.length == cs_size )
	    {
	      cs_size<<= 1;
	      mt ( cs.v= mem_realloc ( size_t, cs.v, cs_size ) );
	    }
	  cs.v[cs.length++]= sym;
	  if ( stop ) break;
	  stop= stop2;
	}
      if ( cs.length == 0 ) goto errorline;
      tL_lexicon_add ( new, word, prob, &cs );
      
      ++n;
      
    }
  
  /* Allibera i ajusta memòria. */
  free_cwhash ( cwh );
  free ( cs.v );
  tL_lexicon_adjust ( new );
  
  return new;

 error_src_sym_not_in_wc:
  tL_msgerror ( err, _( "symbol %s does not belong to a wildcard" ) , src_sym );
  goto error;
 error_wildcards_tok:
  TLMSG_GTOK ( err, "WILDCARDS" );
  goto error;
 error_cwstart_tok:
  TLMSG_GTOK ( err, "CWSTART" );
  goto error;
 error_cwstop_tok:
  TLMSG_GTOK ( err, "CWSTOP" );
  goto error;
 errorcwstart:
  tL_msgerror ( err, _( "Two fields were expected in CWSTART" ) );
  goto error;
 errorcwstop:
  tL_msgerror ( err, _( "At least two fields were expected in CWSTOP" ) );
  goto error;
 error_cwtrans_tok:
  TLMSG_GTOK ( err, "CWTRANS" );
  goto error;
 errorcwtrans:
  tL_msgerror ( err, _( "Three fields were expected"
                " in the %uth crossword transition" ) , _PS(n) );
  goto error;
 errorwildcards:
  tL_msgerror ( err, _( "At least two fields were expected"
                " in the %uth wildcard" ) , _PS(n) );
  goto error;
 errorline:
  tL_msgerror ( err, _( "At least three fields were expected"
                " in the %uth entry" ) , _PS(n) );
 error:
  if ( cwh != NULL ) free_cwhash ( cwh );
  if ( cs.v != NULL ) free ( cs.v );
  if ( new != NULL ) tL_lexicon_free ( new );
  return NULL;
  
} /* end read_cwlexicon_text */


static tLLexicon *
read_lexicon_text (
		   gzFile         from,
		   tLDict        *syms,
		   const tLBool   emptysyms,
		   tLBuffer      *buffer,
		   char         **err
		   )
{
  
  tLLexicon *new;
  char *begin, *end, *word;
  tLSeq cs;
  size_t cs_size, n, sym;
  tLProb prob;
  int stop;
  
  
  new= NULL;
  cs.v= NULL;
  
  /* Capçalera i inicialitza. */
  new= tL_lexicon_new ( TL_LEXICON_CONV );
  cs_size= 1;
  mt ( cs.v= mem_alloc ( size_t, 1 ) );
  
  /* Llig entrades. */
  n= 1;
  while ( tL_gline ( from, buffer ) )
    {
      
      /* Paraula. */
      if ( tL_gtokenstr ( buffer->v, &begin, &end ) ) goto errorline;
      *end= '\0';
      word= begin;
      
      /* Probabilitat. */
      if ( tL_gtokenstr ( end+1, &begin, &end ) ) goto errorline;
      *end= '\0';
      prob= tL_atop ( begin );
      
      /* Seqüència de símbols. */
      cs.length= 0;
      stop= tL_gtokenstr ( end+1, &begin, &end );
      while ( begin != NULL )
	{
	  *end= '\0';
	  if ( register_sym ( begin, n, &sym, syms,
			      emptysyms, TRUE, err ) == -1 )
	    goto error;
	  if ( cs.length == cs_size )
	    {
	      cs_size<<= 1;
	      mt ( cs.v= mem_realloc ( size_t, cs.v, cs_size ) );
	    }
	  cs.v[cs.length++]= sym;
	  if ( stop ) break;
	  stop= tL_gtokenstr ( end+1, &begin, &end );
	}
      if ( cs.length == 0 ) goto errorline;
      tL_lexicon_add ( new, word, prob, &cs );
      
      ++n;
      
    }
  
  /* Allibera i ajusta memòria. */
  free ( cs.v );
  tL_lexicon_adjust ( new );
  
  return new;

errorline:
  tL_msgerror ( err, _( "At least three fields were expected"
                " in the %uth entry" ) , _PS(n) );
 error:
  if ( cs.v != NULL ) free ( cs.v );
  if ( new != NULL ) tL_lexicon_free ( new );
  return NULL;
  
} /* end read_lexicon_text */


static void
cwlexicon_print (
		 const tLLexicon *lexicon,
		 FILE            *to,
		 const tLDict    *syms
		 )
{
  
  size_t n, ed, *init, t, p, q;
  const tLLexiconEntry *entries;
  tLLexiconCWNode node;
  
  
  fprintf ( to, "CWLEXICON\n" );
  fprintf ( to, "WILDCARDS\n" );
  for ( n= 0; n < lexicon->wildcards->N; ++n )
    {
      fprintf ( to, "%s", tL_dict_get_token ( lexicon->wildcards, n ) );
      for ( p= lexicon->wcinits[n]; p != 0; p= lexicon->cwnodes[p].next )
	fprintf ( to, " %s", tL_dict_get_token ( syms,
						 lexicon->cwnodes[p].sym ) );
      putc ( '\n', to );
    }
  fprintf ( to, "CWSTART %s\n", tL_dict_get_token ( syms, lexicon->initcw ) );
  fprintf ( to, "CWSTOP" );
  for ( n= 0; n < lexicon->NS; ++n )
    if ( lexicon->stopcw[n] )
      fprintf ( to, " %s", tL_dict_get_token ( syms, n ) );
  putc ( '\n', to );
  fprintf ( to, "CWTRANS\n" );
  for ( n= 0; n < lexicon->NS; ++n )
    {
      for ( q= lexicon->cwitrans[n]; q != 0; q= lexicon->cwtrans[q].next )
	{
	  node= lexicon->cwnodes[lexicon->cwtrans[q].to];
	  fprintf ( to, "%s %s %s\n",
		    tL_dict_get_token ( syms, n ),
		    tL_dict_get_token ( syms, node.sym ),
		    tL_dict_get_token ( lexicon->wildcards,
					node.wildcard ) );
	    }
	}

  fprintf ( to, "ENTRIES\n" );
  init= lexicon->init;
  entries= lexicon->entries;
  for ( n= 0; n < lexicon->words->N; ++n )
    for ( ed= init[n]; ed != NULL_ENTRY;
	  ed= entries[ed].next )
      {
	fprintf ( to, "%s",
		  tL_dict_get_token ( lexicon->words, n ) );
	printprobsep ( to, entries[ed].p );
	fprintf ( to, " %s", tL_dict_get_token ( lexicon->wildcards,
						 entries[ed].cs->v[0] ) );
	for ( t= 1; t < entries[ed].cs->length-1; ++t )
	  fprintf ( to, " %s", tL_dict_get_token ( syms,
						   entries[ed].cs->v[t] ) );
	if ( entries[ed].cs->length > 1 )
	  fprintf ( to, " %s", tL_dict_get_token ( lexicon->wildcards,
						   entries[ed].cs->v[t] ) );
	fputc ( '\n', to );
      }
  
} /* end cwlexicon_print */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(tLLexicon *)
tL_lexicon_new (
		const tLLexiconType type
		)
{
  
  tLLexicon *new;
  
  
  mt ( new= mem_alloc ( tLLexicon, 1 ) );
  new->type= type;
  new->words= tL_dict_new ();
  new->N= 0;
  mt ( new->init= mem_alloc ( size_t, 1 ) );
  mt ( new->entries= mem_alloc ( tLLexiconEntry, 2 ) );
  new->_size_init= 1;
  new->_size_entries= 2;
  if ( type == TL_LEXICON_CW )
    {
      new->wildcards= tL_dict_new ();
      mt ( new->wcinits= mem_alloc ( size_t, 1 ) );
      mt ( new->cwnodes= mem_alloc ( tLLexiconCWNode, 2 ) );
      memset ( new->cwnodes, 0, sizeof(tLLexiconCWNode) );
      new->NCW= 0;
      mt ( new->stopcw= mem_alloc ( tLBool, 1 ) );
      mt ( new->cwitrans= mem_alloc ( size_t, 1 ) );
      new->NS= 0;
      mt ( new->cwtrans= mem_alloc ( tLLexiconCWTrans, 2 ) );
      memset ( new->cwtrans, 0, sizeof(tLLexiconCWTrans) );
      new->NCWT= 0;
      new->_size_wcinits= 1;
      new->_size_cwnodes= 2;
      new->_size_cwtrans= 2;
      new->_size_syms= 1;
    }
  
  return new;
  
} /* end tL_lexicon_new */


TL_API_FUNC(void)
tL_lexicon_free (
		 tLLexicon *lexicon
		 )
{
  
  size_t n;
  
  
  if ( lexicon->type == TL_LEXICON_CW )
    {
      free ( lexicon->stopcw );
      free ( lexicon->cwitrans );
      free ( lexicon->wcinits );
      free ( lexicon->cwtrans );
      free ( lexicon->cwnodes );
      tL_dict_free ( lexicon->wildcards );
    }
  for ( n= 1; n <= lexicon->N; ++n )
    tL_seq_free ( lexicon->entries[n].cs );
  free ( lexicon->entries );
  free ( lexicon->init );
  tL_dict_free ( lexicon->words );
  free ( lexicon );
  
} /* end tL_lexicon_free */


int 
tL_lexicon_add (
		tLLexicon    *lexicon,
		const char   *word,
		const tLProb  p,
		const tLSeq *cs
		)
{
  
  int ret;
  size_t id, N;
  tLLexiconEntry *entries;
  
  
  if ( (ret= tL_dict_insert ( lexicon->words, word, &id )) )
    {
      if ( lexicon->words->N > lexicon->_size_init )
	{
	  (lexicon->_size_init)<<= 1;
	  mt ( lexicon->init= mem_realloc ( size_t,
					    lexicon->init,
					    lexicon->_size_init ) );
	}
      lexicon->init[id]= NULL_ENTRY;
    }
  
  N= ++(lexicon->N);
  if ( lexicon->N == lexicon->_size_entries )
    {
      (lexicon->_size_entries)<<= 1;
      mt ( lexicon->entries= mem_realloc ( tLLexiconEntry,
					   lexicon->entries,
					   lexicon->_size_entries ) );
    }
  
  entries= lexicon->entries;
  entries[N].cs= cscopy ( cs );
  entries[N].p= p;
  entries[N].w= id;
  entries[N].next= lexicon->init[id];
  lexicon->init[id]= N;
  
  return ret;
  
} /* end tL_lexicon_add */


int
tL_lexicon_add_cwtrans (
			tLLexicon     *lexicon,
			tLDict        *syms,
			const tLBool   emptysyms,
			const char    *src_wc,
			const char    *src_sym,
			const char    *dst_sym,
			const char    *dst_wc,
			char         **err
			)
{
  
  assert ( lexicon->type == TL_LEXICON_CW );
  
  
  return 0;
  
} /* end tL_lexicon_add_cwtrans */


void
tL_lexicon_adjust (
		   tLLexicon *lexicon
		   )
{
  
  if ( lexicon->N == 0 ) return;
  if ( (lexicon->N+1) < lexicon->_size_entries )
    {
      mt ( lexicon->entries= mem_realloc ( tLLexiconEntry,
					   lexicon->entries,
					   lexicon->N+1 ) );
      lexicon->_size_entries= lexicon->N+1;
    }
  
  if ( lexicon->words->N < lexicon->_size_init )
    {
      mt ( lexicon->init= mem_realloc ( size_t,
					lexicon->init,
					lexicon->words->N ) );
      lexicon->_size_init= lexicon->words->N;
    }
  
  if ( lexicon->type == TL_LEXICON_CW )
    {
      if ( lexicon->wildcards->N < lexicon->_size_wcinits )
	{
	  mt ( lexicon->wcinits= mem_realloc ( size_t, lexicon->wcinits,
					       lexicon->wildcards->N ) );
	  lexicon->_size_wcinits= lexicon->wildcards->N;
	}
      if ( (lexicon->NCW+1) < lexicon->_size_cwnodes )
	{
	  mt ( lexicon->cwnodes= mem_realloc ( tLLexiconCWNode, 
					       lexicon->cwnodes,
					       lexicon->NCW+1 ) );
	  lexicon->_size_cwnodes= lexicon->NCW+1;
	}
      if ( (lexicon->NCWT+1) < lexicon->_size_cwtrans )
	{
	  mt ( lexicon->cwtrans= mem_realloc ( tLLexiconCWTrans, 
					       lexicon->cwtrans,
					       lexicon->NCWT+1 ) );
	  lexicon->_size_cwtrans= lexicon->NCWT+1;
	}
      if ( lexicon->NS < lexicon->_size_syms && lexicon->NS > 0 )
	{
	  lexicon->_size_syms= lexicon->NS;
	  mt ( lexicon->stopcw= mem_realloc ( tLBool, lexicon->stopcw,
					      lexicon->NS ) );
	  mt ( lexicon->cwitrans= mem_realloc ( size_t, lexicon->cwitrans,
						lexicon->NS ) );
	}
    }
  
} /* end tL_lexicon_adjust */


TL_API_FUNC(void)
tL_lexicon_print (
		  const tLLexicon *lexicon,
		  FILE            *to,
		  const tLDict    *syms
		  )
{
  
  size_t n, ed, *init;
  const tLLexiconEntry *entries;
  
  
  if ( lexicon->type == TL_LEXICON_CW )
    {
      cwlexicon_print ( lexicon, to, syms );
      return;
    }
  
  fprintf ( to, "LEXICON\n" );
  init= lexicon->init;
  entries= lexicon->entries;
  for ( n= 0; n < lexicon->words->N; ++n )
    for ( ed= init[n]; ed != NULL_ENTRY;
	  ed= entries[ed].next )
      {
	fprintf ( to, "%s",
		  tL_dict_get_token ( lexicon->words, n ) );
	printprobsep ( to, entries[ed].p );
	fputc ( ' ', to );
	tL_seq_print ( entries[ed].cs, to, syms, ' ' );
	fputc ( '\n', to );
      }
  
} /* end tL_lexicon_print */


TL_API_FUNC(tLLexicon *)
tL_lexicon_new_from_file (
			  gzFile         from,
			  tLDict        *syms,
			  const tLBool   emptysyms,
			  char         **err
			  )
{
  
  tLBuffer *buffer;
  tLLexicon *lexicon;
  int cw;
  
  
  buffer= tL_buffer_new ();
  if ( !tL_gtoken ( from, buffer ) )
    {
      TLMSG_TOKID ( err );
      goto error;
    }
  if ( (cw= strcmp ( "LEXICON", buffer->v )) &&
       strcmp ( "CWLEXICON", buffer->v ) )
    {
      TLMSG_UNKTOKID ( err, buffer->v );
      goto error; 
    }
  lexicon= cw ?
    read_cwlexicon_text ( from, syms, emptysyms, buffer, err ) :
    read_lexicon_text ( from, syms, emptysyms, buffer, err );
  
  tL_buffer_free ( buffer );
  
  return lexicon;
  
 error:
  tL_buffer_free ( buffer );
  return NULL;
  
} /* end tL_lexicon_new_from_file */
