/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lexicon_public.h - Part pública de 'lexicon.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */

#ifndef __LIBTL_LEXICON_H__
#define __LIBTL_LEXICON_H__

#include "common_public.h"
#include "port.h"
#include "text_public.h"

/** \addtogroup Models
 *  @{
 */

/** \defgroup Lexicon Lexicon model
 */
/** \addtogroup Lexicon
 *  @{
 */

/* TIPUS */

typedef struct _tLLexiconEntry tLLexiconEntry;
typedef struct _tLLexiconCWNode tLLexiconCWNode;
typedef struct _tLLexiconCWTrans tLLexiconCWTrans;

/** Type of Lexicon. */
typedef enum
  {
    TL_LEXICON_CONV,    /**< Conventional lexicon. */
    TL_LEXICON_CW       /**< Crossword lexicon. */
  } tLLexiconType;

/* Diccionari de paraules. Per a cada
 * paraula indica les distintes formes
 * de escriurela i la probabilitat de
 * cadascuna de les formes.
 */
/** Lexicon model. This structure stores how words are represented as
 *  symbol sequences. Each word can be written in several ways,
 *  hence, for each way its log-probability is also stored. For
 *  crossword lexicons, the first and last symbol of each sequence
 *  must be wildcards. Wildcards are special symbols, which are
 *  stored in the lexicon, and represent sets of
 *  symbols. Crossword lexicons also store the possible transitions
 *  between wildcards.
 */
typedef struct
{
  
  tLLexiconType     type;       /**< Type of the lexicon. */
  tLDict           *words;      /**< Dictionary with the words of the
				   lexicon. */
  size_t            N;          /**< Number of current entries in the
				   lexicon. */
  size_t           *init;       /* |Words|. Posició de la primera
				   entrada per a cada paraula del
				   vocabulari. */
  tLLexiconEntry   *entries;    /* Entrades del diccionari. */
  size_t            _size_init;
  size_t            _size_entries;
  
  /* CW. */
  tLDict           *wildcards;
  size_t           *wcinits;     /* Primer node per a cada wildcard. */
  tLLexiconCWNode  *cwnodes;     /* Nodes CW. */
  size_t            NCW;         /* Número de nodes. EL 0 és NULL. */
  size_t            initcw;      /* El reconeixement començarà
				    supossant que l'últim símbol
				    reconegut ha sigut este. */
  size_t            NS;          /* Número de símbols utilitzats per a
				    stopcw i cwitrans. Pot ser que
				    siguent tots els del vocabulari o
				    no, però és segur que tots els
				    símbols que forment part d'algun
				    wildcard estan ací. */
  tLBool           *stopcw;      /* |NS|. Símbols que poden acabar una
				    frase. */
  size_t           *cwitrans;    /* |NS|. La primera transició de
                                     cada símbol. */
  tLLexiconCWTrans *cwtrans;     /* Transicions CW. El 0 és NULL. */
  size_t            NCWT;        /* Número de transicions. */
  size_t            _size_wcinits;
  size_t            _size_cwnodes;
  size_t            _size_cwtrans;
  size_t            _size_syms;
  
} tLLexicon;

/* FUNCIONS */

/** Frees memory. Frees the memory allocated for the lexicon model.
 *
 *  @param lexicon The lexicon model.
 */
TL_API_FUNC(void)
tL_lexicon_free (
		 tLLexicon *lexicon
		 );

/* Crea una diccionari de paraules buit. */
/** Creates a lexicon. This function creates a new empty lexicon
 *  model.
 *
 *  @param type The type of the new lexicon.
 *  @return The lexicon model.
 */
TL_API_FUNC(tLLexicon *)
tL_lexicon_new (
		const tLLexiconType type
		);

/* Llig un diccionari de paraules. Si un
 * símbol està fora de vocabulari genera
 * un tL_error.
 */
/** Reads a lexicon model from a file. This function creates a new
 *  lexicon model from a text description. The provided dictionary is
 *  the dictionary of symbols.
 *
 *  @param from File where the text description is stored. 
 *  @param syms Dictionary with the symbols.
 *  @param emptysyms Specifies whether the provided dictionary is
 *  empty (new symbols must be registered) or not (new symbols are
 *  treated as errors).
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The lexicon model, or NULL in case of error.
 */
TL_API_FUNC(tLLexicon *)
tL_lexicon_new_from_file (
			  gzFile         from,
			  tLDict        *syms,
			  const tLBool   emptysyms,
			  char         **err
			  );

/* Imprimeix el diccionari de paraules. */
/** Prints the lexicon model. This function writes to the given file
 *  the content of the lexicon model, using the provided symbol
 *  dictionary. The provided symbol dictionary is assumed to
 *  contain all required symbols, otherwise an unexpected error
 *  could happen.
 *
 *  @param lexicon The lexicon model.
 *  @param to File to which the model will be written.
 *  @param syms Dictionary containing the symbols.
 */
TL_API_FUNC(void)
tL_lexicon_print (
		  const tLLexicon *lexicon,
		  FILE            *to,
		  const tLDict    *syms
		  );

/** @} */
/** @} */

#endif /* __LIBTL_LEXICON_H__ */
