/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lexicon_recaux.c - Implementació de 'lexicon_recaux.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */

#include <assert.h>
#include <stddef.h>
#include <stdlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "error.h"
#include "hash.h"
#include "lexicon_recaux.h"
#include "mem.h"




/**********/
/* MACROS */
/**********/

#define HSIZE 32363U




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  unsigned long hist;    /* Identificador del node previ. */
  size_t        sym;     /* Símbol. */
  unsigned long next;    /* Posició (Identificador) del següent
			    node. */
  
} lph_node_t;


/* Taula hash auxiliar utilitzada per a crear un lexprefix_t a partir
   d'un lèxicon. Torna un idenfitificador de node a partir de un
   identificador de node i un identificador de símbol. Comença a
   contar sempre per l'u. */
typedef struct
{
  
  unsigned long  NId;             /* Següent identificador, comença
				     amb 1. */
  unsigned long  table[HSIZE];    /* Taula. */
  lph_node_t    *mem;             /* Memòria. La posició 0 no es
				     gasta. */
  unsigned long  size;            /* Capacitat de mem. */
  
} lphash_t;


/* Estructura auxiliar utilitzada en la creació d'un
   'lexpref_t'. Representa una aresta. */
struct lp_auxed
{
  
  unsigned long to;
  unsigned long next;
  
};


/* Estructura auxiliar utilitzada en la creació d'un
   'lexpref_t'. Representa una aresta final. */
struct lp_auxfed
{
  
  lexpref_fedge_t fed;
  unsigned long   next;
  
};


/* Estructura auxiliar utilitzada en la creació d'un
   'lexpref_t'. Representa les arestes inicials d'un estat en les
   estructures auxiliars. */
struct lp_auxst
{
  
  unsigned long init_ed;
  unsigned long init_fed;
  
};




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* Calcula per a un wildcard el número de transicions entrants. */
static int *
calc_npredwc (
	      const tLLexicon *lexicon
	      )
{
  
  int *new;
  size_t n;
  
  
  mt ( new= mem_calloc ( int, lexicon->wildcards->N ) );
  for ( n= 1; n <= lexicon->NCWT; ++n )
    ++new[lexicon->cwnodes[lexicon->cwtrans[n].to].wildcard];
  
  return new;
  
} /* end calc_npredwc */


static lphash_t *
create_lphash ()
{

  lphash_t *new;
  int i;
  
  
  mt ( new= mem_alloc ( lphash_t, 1 ) );
  mt ( new->mem= mem_alloc ( lph_node_t, 1 ) );
  new->size= 1;
  new->NId= 1;
  for ( i= 0; i < HSIZE; ++i ) new->table[i]= 0;
  
  return new;
  
} /* end create_lphash */


static void
free_lphash (
	     lphash_t *lphash
	     )
{
  
  free ( lphash->mem );
  free ( lphash );
  
} /* end free_lphash */


static unsigned long
lphash_hash (
	     const unsigned long  hist,
	     const size_t         sym
	     )
{
  
  static unsigned int buffer[2];
  static const unsigned long size= sizeof(unsigned int)*2;
  
  
  buffer[0]= (unsigned int) hist;
  buffer[1]= (unsigned int) sym;
  
  return tL_hash_gnu_cxx ( buffer, size );
  
} /* end lphash_hash */


/* Torna l'identificador associat a un
   identificador(història)+símbol. */
static unsigned long
lphash_get_id (
	       lphash_t            *lphash,
	       const unsigned long  hist,
	       const size_t         sym
	       )
{
  
  lph_node_t *mem;
  unsigned int pos;
  unsigned long id;
  
  
  /* Cerca. */
  pos= (unsigned int) (lphash_hash ( hist, sym )%HSIZE);
  mem= lphash->mem;
  for ( id= lphash->table[pos];
	id != 0 && (mem[id].hist != hist || mem[id].sym != sym);
	id= mem[id].next );
  if ( id != 0 ) return id;
  
  /* Inserta. */
  if ( lphash->size == lphash->NId )
    {
      lphash->size*= 2;
      mt ( mem= lphash->mem= mem_realloc ( lph_node_t, mem, lphash->size ) );
    }
  id= lphash->NId++;
  mem[id].hist= hist;
  mem[id].sym= sym;
  mem[id].next= lphash->table[pos];
  lphash->table[pos]= id;
  
  return id;
  
} /* end lphash_get_id */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

lexmap_t *
tL_create_lexmap (
		  const tLLexicon *lexicon,
		  const tLAModel  *amodel
		  )
{
  
  lexmap_t *new;
  lexmap_node_t *map;
  lexmap_inode_t *aux, **imap;
  unsigned long N, state;
  size_t n, e, *init, *nentries, t, ne;
  const tLLexiconEntry *entries;
  const tLSeq *cs;
  const tLAModelHMM *chars;
  tLBool *mpred;
  
  
  /* Prepara. */
  chars= amodel->hmms;
  init= lexicon->init;
  entries= lexicon->entries;
  mt ( new= mem_alloc ( lexmap_t, 1 ) );
  
  /* Nombre d'estats globals. */
  N= 0;
  for ( n= 1; n <= lexicon->N; ++n )
    N+= entries[n].cs->length;
  new->N= N;
  mt ( new->mpred= mpred= mem_alloc ( tLBool, new->N ) );
  
  /* Reserva memòria. */
  mt ( map= new->map= mem_alloc ( lexmap_node_t, N ) );
  mt ( imap= new->imap= mem_alloc ( lexmap_inode_t *, lexicon->words->N ) );
  mt ( new->imap[0]= mem_alloc ( lexmap_inode_t, lexicon->N ) );
  mt ( nentries= new->nentries= mem_alloc ( size_t, lexicon->words->N ) );
  
  /* Plena. */
  for ( n= 0, state= 0, aux= new->imap[0];
	n < lexicon->words->N; ++n )
    {
      imap[n]= aux;
      for ( e= init[n], ne= 0; e != NULL_ENTRY;
	    e= entries[e].next, ++aux, ++ne )
	{
	  aux->state= state;
	  aux->p= entries[e].p;
	  cs= entries[e].cs;
	  for ( t= 0; t < cs->length; ++t, ++state )
	    {
	      map[state].cseq= cs;
	      map[state].state= t;
	      map[state].hmm= &(chars[cs->v[t]]);
	      map[state].word= n;
	      map[state].e= e;
	      mpred[state]= FALSE;
	    }
	}
      nentries[n]= ne;
    }
  
  return new;
  
} /* end tL_create_lexmap */


void
tL_free_lexmap (
		lexmap_t *lexmap
		)
{

  free ( lexmap->mpred );
  free ( lexmap->nentries );
  free ( lexmap->imap[0] );
  free ( lexmap->imap );
  free ( lexmap->map );
  free ( lexmap );
  
} /* end tL_free_lexmap */


lexpref_t *
tL_create_lexpref (
		   const tLLexicon *lexicon,
		   const tLAModel  *amodel,
		   const size_t     sil
		   )
{
  
  unsigned long NE, NFE, size_eds, size_feds, size, N, Nh, NhS, n, state,
    prev, *edges, ed, sizeh, *cwnodemap, off, end;
  ptrdiff_t NECW;
  struct lp_auxst *auxstates;
  struct lp_auxed *auxedges;
  struct lp_auxfed *auxfedges;
  lexpref_t *new;
  lexpref_state_t *states, *statesh;
  lphash_t *lphash;
  size_t e, t, sym, p, cwsym, ncwnodes, wcsym;
  const size_t *init;
  const tLLexiconEntry *entries;
  const tLSeq *cs;
  const tLAModelHMM *chars;
  lexpref_fedge_t *fedges;
  tLBool iscw, iswildcard, *mpred, next_is_mpred, issil;
  const tLLexiconCWNode *cwnodes;
  tLLexiconCWNode cwnode;
  lexpref_state_edges_t *steds;
  int *npredwc;
  unsigned long *psils, NS;
  
  
  iscw= (lexicon->type==TL_LEXICON_CW);
  issil= (sil<amodel->N);
  
  /* 'auxedges' es adecuat per a afegir arestes. La idea és al final
     processar auxedges per a crear les arestes de lexpref. auxstates
     és per a cada estat la primera aresta en auxedges. */
  /* Reserva memòria i inicialitza. L'aresta 0 de auxstates es
     nul·la. NE és el número d'arestes contant la 0. */
  mt ( states= mem_alloc ( lexpref_state_t, 1 ) );
  mt ( mpred= mem_alloc ( tLBool, 1 ) );
  mt ( auxstates= mem_alloc ( struct lp_auxst, 1 ) );
  if ( issil )
    {
      /* En psils guarde de moment +1 per a saber que el 0 és que no té. */
      mt ( psils= mem_alloc ( unsigned long, 1 ) );
      NS= 0;
    }
  else psils= NULL; /* CALLA. */
  size= 1; N= 1;
  auxstates[0].init_ed= 0;
  auxstates[0].init_fed= 0;
  mt ( auxedges= mem_alloc ( struct lp_auxed, 1 ) );
  size_eds= 1; NE= 1;
  mt ( auxfedges= mem_alloc ( struct lp_auxfed, 1 ) );
  size_feds= 1; NFE= 1;

  /* statesh guarda provisionalment els nodes ocults. De moment el
     primer està guardat en 1, el 0 es gasta per a indicar node
     nul. Nh inclou al 0. */
  /* Cada cwnode sols pot tindre un estat ocult com a principi de
     frase. Per tant cwnodemap indica eixe estat ocult per a cada
     cwnode, 0 si no apareix mai al principi. El punter que gaurda
     ésel offset de l'estat ocult +1.  */
  /* Reserva memòria per a les estructures de crosswords. */
  NhS= 0;
  if ( iscw )
    {
      mt ( statesh= mem_alloc ( lexpref_state_t, 2 ) );
      sizeh= 2; Nh= 1;
      mt ( cwnodemap= mem_alloc ( unsigned long, lexicon->NCW+1 ) );
      for ( n= 1; n <= lexicon->NCW; ++n ) cwnodemap[n]= 0;
      npredwc= calc_npredwc ( lexicon );
    }
  else { Nh= 0; statesh= NULL; cwnodemap= NULL; npredwc= NULL; } /* CALLA!! */
  
  /* Inserta els estats i les arestes en 'auxedges'. */
  lphash= create_lphash ();
  init= lexicon->init;
  entries= lexicon->entries;
  chars= amodel->hmms;
  cwnodes= iscw?lexicon->cwnodes:NULL;
  states[0].type= NORMAL;
  mpred[0]= FALSE;
  if ( issil ) psils[0]= 0;
  wcsym= 0; /* CALLA */
  for ( n= 0; n < lexicon->words->N; ++n )
    for ( e= init[n]; e != NULL_ENTRY; e= entries[e].next )
      {
	cs= entries[e].cs;
	prev= 0;
	next_is_mpred= FALSE;
	for ( t= 0; t < cs->length; ++t )
	  {
	    iswildcard= (iscw && (t==0 || t==(cs->length-1)));
	    sym= cs->v[t];
	    if ( iswildcard )
	      {
		wcsym= sym;
		sym+= amodel->N; /* Separa els símbols
				    dels wildcards. */
		if ( t == 0 && npredwc[wcsym] > 1 ) next_is_mpred= TRUE;
	      }
	    state= lphash_get_id ( lphash, prev, sym );
	    if ( state == N ) /* NOTA!!!: Està fet de manera que açò
				 indica estat nou. */
	      {
		
		/* Inserta estat. */
		if ( N == size )
		  {
		    size*= 2;
		    mt ( states=
			 mem_realloc ( lexpref_state_t, states, size ) );
		    mt ( auxstates=
			 mem_realloc ( struct lp_auxst, auxstates, size ) );
		    mt ( mpred= mem_realloc ( tLBool, mpred, size ) );
		    if ( issil )
		      mt ( psils= mem_realloc ( unsigned long, psils, size ) );
		  }
		++N;
		if ( issil ) psils[state]= 0;
		auxstates[state].init_ed= 0;
		auxstates[state].init_fed= 0;
		mpred[state]= next_is_mpred;
		if ( !iswildcard )
		  {
		    states[state].type= NORMAL;
		    states[state].n.hmm= &(chars[sym]);
		    states[state].n.sym= sym;
		  }
		else
		  {
		    /* Provisionalment guarde els offsets +1 dels
		       estats ocults com a punters. */
		    states[state].type= WILDCARD;
		    states[state].w.begin= Nh;
		    for ( p= lexicon->wcinits[wcsym], ncwnodes= 0;
			  p != 0;
			  p= cwnodes[p].next, ++ncwnodes )
		      {
			cwnode= cwnodes[p];
			if ( Nh == sizeh )
			  {
			    sizeh*= 2;
			    mt ( statesh= mem_realloc ( lexpref_state_t,
							statesh, sizeh ) );
			  }
			if ( t == 0 )
			  {
			    assert ( cwnodemap[p] == 0 );
			    cwnodemap[p]= Nh;
			  }
			statesh[Nh].type= HIDDEN;
			statesh[Nh].h.hmm= &(chars[cwnode.sym]);
			statesh[Nh].h.sym= cwsym= cwnode.sym;
			statesh[Nh++].h.wcstate= state;
		      }
		    states[state].w.end= Nh;
		    next_is_mpred= (ncwnodes>1);
		    /* Is a word-end wildcard. */
		    if ( t==(cs->length-1) && issil ) NhS+= ncwnodes;
		  }
		
		/* Inserta aresta. */
		if ( NE == size_eds )
		  {
		    size_eds*= 2;
		    mt ( auxedges=
			 mem_realloc ( struct lp_auxed, auxedges, size_eds ) );
		  }
		auxedges[NE].to= state;
		auxedges[NE].next= auxstates[prev].init_ed;
		auxstates[prev].init_ed= NE++;
		
	      }
	    prev= state;
	  }
	if ( NFE == size_feds )
	  {
	    size_feds*= 2;
	    mt ( auxfedges=
		 mem_realloc ( struct lp_auxfed, auxfedges, size_feds ) );
	  }
	auxfedges[NFE].fed.word= n;
	auxfedges[NFE].fed.e= e;
	auxfedges[NFE].fed.p= entries[e].p;
	auxfedges[NFE].next= auxstates[prev].init_fed;
	auxstates[prev].init_fed= NFE++;
	if ( issil && psils[prev] == 0 )
	  psils[prev]= ++NS; /* Guarde momentaneament +1, aixina el 0
				és que no en té. */
      }
  free_lphash ( lphash );
  
  /* Finalitza de construir i allibera memòria innecessària. */
  /* Estats. */
  mt ( new= mem_alloc ( lexpref_t, 1 ) );
  new->iscw= iscw;
  new->N= N;
  new->NS= N+NS+1; /* +1 per el silenci inicial. */
  if ( new->NS != size )
    mt ( mpred= mem_realloc ( tLBool, mpred, new->NS ) );
  new->mpred= mpred;
  if ( !iscw )
    {
      new->cwlist= NULL;
      new->cwedges= NULL;
      new->NR= new->NS;
      if ( new->NS != size )
	mt ( states= mem_realloc ( lexpref_state_t, states, new->NS ) );
    }
  else
    {
      
      free ( npredwc );
      assert ( Nh > 1 );
      new->NR= N+NS+1+(Nh-1)+NhS;
      
      /* Copia els estats i reajusta memòria. */
      mt ( states= mem_realloc ( lexpref_state_t, states, new->NR ) );
      for ( n= 1, off= new->NS; n < Nh; ++n, ++off )
	states[off]= statesh[n];
      free ( statesh );
      
      /* Transicions crossword. */
      mt ( new->cwlist= mem_alloc ( lexpref_cwlist_t, lexicon->NS ) );
      mt ( new->cwedges= edges= mem_alloc ( unsigned long, lexicon->NCWT ) );
      for ( n= 0; n < lexicon->NS; ++n )
	{
	  new->cwlist[n].begin= edges;
	  for ( p= lexicon->cwitrans[n];
		p != 0;
		p= lexicon->cwtrans[p].next )
	    {
	      state= cwnodemap[lexicon->cwtrans[p].to];
	      if ( state != 0 )
		*(edges++)= new->NS+state-1;
	    }
	  new->cwlist[n].end= edges;
	}
      NECW= edges-new->cwedges;
      if ( NECW < lexicon->NCWT )
	{
	  edges= new->cwedges;
	  mt ( new->cwedges= mem_realloc ( unsigned long,
					   new->cwedges, NECW ) );
	  if ( edges != new->cwedges )
	    for ( n= 1; n <= lexicon->NCW; ++n )
	      {
		new->cwlist[n].begin= new->cwedges+(new->cwlist[n].begin-edges);
		new->cwlist[n].end= new->cwedges+(new->cwlist[n].end-edges);
	      }
	}
      free ( cwnodemap );
      
      /* is final i inicial. */
      mt ( new->isfinal= mem_alloc ( tLBool, lexicon->NS ) );
      memcpy ( new->isfinal, lexicon->stopcw, lexicon->NS*sizeof(tLBool) );
      new->initcw= lexicon->initcw;
      
    }
  new->states= states;
  mt ( new->state_eds= mem_alloc ( lexpref_state_edges_t, new->NS ) );
  if ( NE == 1 ) tL_error ( _( "Lexicon is empty" ) );    
  
  /* Arestes. */
  mt ( new->edges= edges= mem_alloc ( unsigned long, NE-1 ) );
  new->NE= NE-1;
  mt ( new->fedges= fedges= mem_alloc ( lexpref_fedge_t, NFE-1 ) );
  for ( n= 0; n < N; ++n )
    {
      if ( states[n].type == WILDCARD )
	{
	  states[n].w.begin= (new->NS+states[n].w.begin)-1;
	  states[n].w.end= (new->NS+states[n].w.end)-1;
	}
      steds= &(new->state_eds[n]);
      steds->begin= edges;
      for ( ed= auxstates[n].init_ed; ed != 0; ed= auxedges[ed].next )
	*(edges++)= auxedges[ed].to;
      steds->end= edges;
      steds->fbegin= fedges;
      for ( ed= auxstates[n].init_fed; ed != 0; ed= auxfedges[ed].next )
	*(fedges++)= auxfedges[ed].fed;
      steds->fend= fedges;
    }
  free ( auxfedges );
  free ( auxedges );
  free ( auxstates );
  
  /* Silencis. */
  if ( issil )
    {
      if ( N < size )
	mt ( psils= mem_realloc ( unsigned long, psils, N ) );
      mt ( new->_sil2vis= mem_alloc ( unsigned long, NS+1 ) );
      new->sil2vis= new->_sil2vis-N;
      new->psils= psils;
      for ( n= 0; n < N; ++n )
	if ( (state= psils[n]) != 0 )
	  {
	    state+= N-1;
	    psils[n]= state;
	    new->sil2vis[state]= n;
	    states[state].type= SIL;
	    states[state].n.hmm= &(chars[sil]);
	    states[state].n.sym= sil;
	    steds= &(new->state_eds[state]);
	    steds->begin= steds->end= NULL;
	    steds->fbegin= new->state_eds[n].fbegin;
	    steds->fend= new->state_eds[n].fend;
	    new->mpred[state]= (states[n].type==WILDCARD);
	  }
      state= new->isil= N+NS;
      states[state].type= SIL;
      states[state].n.hmm= &(chars[sil]);
      states[state].n.sym= sil;
      steds= &(new->state_eds[state]);
      steds->begin= steds->end= NULL;
      steds->fbegin= steds->fend= NULL;
      new->mpred[state]= FALSE;
      new->sil2vis[state]= 0;
    }
  else new->psils= new->_sil2vis= NULL;
  
  /* Silencies xwords. */
  if ( issil && iscw )
    {
      new->_hinfo= mem_alloc ( lexpref_hinfo_t, Nh+NhS-1 );
      new->hinfo= new->_hinfo-new->NS;
      end= new->NS+Nh-1;
      for ( n= new->NS, off= end; n < end; ++n )
	{
	  steds= &(new->state_eds[states[n].h.wcstate]);
	  new->hinfo[n].issil= FALSE;
	  if ( steds->fbegin == steds->fend ) /* Is a node of a
						 word-start wildcard ? */
	    {
	      new->hinfo[n].node= 0;
	      continue;
	    }
	  new->hinfo[n].node= off;
	  new->hinfo[off].issil= TRUE;
	  new->hinfo[off].node= n;
	  states[off].h= states[n].h;
	  states[off].h.hmm= &(chars[sil]);
	  states[off].h.sym= sil;
	  ++off;
	}
    }
  else { new->_hinfo= new->hinfo= NULL; }
  
  return new;
  
} /* end tL_create_lexpref */


void
tL_free_lexpref (
		 lexpref_t *lexpref
		 )
{

  if ( lexpref->_hinfo != NULL ) free ( lexpref->_hinfo );
  if ( lexpref->_sil2vis != NULL ) free ( lexpref->_sil2vis );
  if ( lexpref->psils != NULL ) free ( lexpref->psils );
  if ( lexpref->iscw )
    {
      free ( lexpref->isfinal );
      free ( lexpref->cwedges );
      free ( lexpref->cwlist );
    }
  free ( lexpref->mpred );
  free ( lexpref->fedges );
  free ( lexpref->edges );
  free ( lexpref->state_eds );
  free ( lexpref->states );
  free ( lexpref );
  
} /* end tL_free_lexpref */


#if 0
void
lexpref_print_dot (
		   const lexpref_t *lexpref,
		   const tLDict    *syms,
		   const tLDict    *words,
		   FILE            *to
		   )
{
  
  unsigned long n;
  const unsigned long *i;
  
  
  fprintf ( to, "digraph g {\n" );
  fprintf ( to, "n0[label=\"START\"];\n" );
  for ( n= 1; n < lexpref->N; ++n )
    if ( lexpref->states[n].p != -HUGE_VAL )
      fprintf ( to, "n%lu[label=\"<f0> %s|<f1> %s\"];\n",
		n, tL_dict_get_token ( syms, lexpref->states[n].sym ),
		tL_dict_get_token ( words, lexpref->states[n].word ) );
    else
      fprintf ( to, "n%lu[label=\"%s\"];\n",
		n, tL_dict_get_token ( syms, lexpref->states[n].sym) );
  for ( n= 0; n < lexpref->N; ++n )
    {
      for ( i= lexpref->states[n].begin; i != lexpref->states[n].end; ++i )
	fprintf ( to, "\"n%lu\" -> \"n%lu\";\n", n, *i );
    }
  fprintf ( to, "}\n" );
  
} /* end lexpref_print_dot */
#endif
