/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lexicon_recaux.h - Diverses estructures utilitzades durant el
 *                     reconeixement creades a partir del lèxic.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */

#ifndef __TL_PRV_LEXICON_RECAUX_H__
#define __TL_PRV_LEXICON_RECAUX_H__

#include <stddef.h>
#include <stdio.h>

#include "amodel.h"
#include "common.h"
#include "lexicon.h"

/* LEXMAP */
/* Mapeja un estat global del lèxic amb l'entrada del lèxic, el número
 * d'estat intern i el model de Markov associat.
 */
typedef struct
{
  
  const tLSeq       *cseq;     /* Seqüència de caràcters. */
  size_t             state;    /* Número d'estat dins de
				   l'entrada. */
  const tLAModelHMM *hmm;      /* Model de Markov. */
  unsigned long      word;     /* Paraula del lèxic associada. */
  size_t             e;        /* Entrada del lèxic. */
  
} lexmap_node_t;


typedef struct
{
  
  unsigned long state;
  tLProb        p;
  
} lexmap_inode_t;


/* Mapeja estats globals del lèxic a estats locals del lèxic. Amb
 * aquesta estructura podem ignorar el lèxic. NOTA: els identificadors
 * globals d'una mateixa entrada apareixen consecutius en el mateix
 * ordre que en la seqüència.
 */
typedef struct
{
  
  unsigned long    N;           /* Nombre d'estats globals. */
  lexmap_node_t   *map;         /* Donat un identificador global
				   d'estat del lèxic torna, l'entrada
				   associada, el número d'estat intern
				   en l'entrada i el model de
				   caràcters associat. */
  lexmap_inode_t **imap;        /* |WORDS|*|ENTRADES|. Donat una
				   paraula i una entrada d'eixa
				   paraula al lèxic, torna
				   l'identificador global del primer
				   estat d'eixa entrada i la
				   probabilitat de transitar. */
  size_t          *nentries;    /* Número d'entrades en cada paraula. */
  tLBool          *mpred;       /* |N|. Indica si un estat té
                                    multiples predecessors. */
  
} lexmap_t;

lexmap_t *
tL_create_lexmap (
		  const tLLexicon *lexicon,
		  const tLAModel  *amodel
		  );

void
tL_free_lexmap (
		lexmap_t *lexmap
		);


/* LEXPREF */

typedef struct
{
  
  unsigned long word;     /* Si és un estat final, paraula
			     associada. */
  tLProb        p;        /* Probabilitat de l'entrada. */
  size_t        e;        /* Número d'entrada. */
  
} lexpref_fedge_t;

/* Transicions d'un estat. */
typedef struct
{
  
  const unsigned long   *begin;     /* Primera aresta. */
  const unsigned long   *end;       /* Aresta que segueix a l'última
				       aresta. */
  const lexpref_fedge_t *fbegin;    /* Primera aresta final. */
  const lexpref_fedge_t *fend;      /* Aresta que segueix a l'última
				       aresta final. */
  
} lexpref_state_edges_t;

/* Els nodes SIL són nodes NORMALS no visibles que tenen un tractament
   especial. */
enum { NORMAL, WILDCARD, HIDDEN, SIL };

typedef union lexpref_state lexpref_state_t;

union lexpref_state
{
  
  int type;
  struct
  {
    int                          type;
    const tLAModelHMM           *hmm;       /* Model de Markov. */
    size_t                       sym;       /* Símbol. */
  }   nh;    /* Part comú de Norma i Hidden. */
  struct
  {
    int                          type;
    const tLAModelHMM           *hmm;       /* Model de Markov. */
    size_t                       sym;       /* Símbol. */
  }   n;    /* Estat Normal. */
  struct
  {
    int           type;
    unsigned long begin;     /* Punter al primer node ocult. */
    unsigned long end;       /* Punter al node ocult
					       que segueix a l'últim
					       node ocult. */
  }   w;    /* Estat Wildcard. */
  struct
  {
    int                          type;
    const tLAModelHMM           *hmm;        /* Model de Markov. */
    size_t                       sym;        /* Símbol. */
    unsigned long                wcstate;    /* Estat wildcard al que
						pertany. */
    
  }   h;    /* Estat ocult. */
  
};

typedef struct
{
  
  const unsigned long *begin;    /* Primera aresta. */
  const unsigned long *end;      /* Aresta que segueix a l'última
				    aresta. */
  
} lexpref_cwlist_t;

/* Extra information for hidden nodes. Connect standard hidden nodes
   with special silence hidden nodes. */
typedef struct
{
  
  tLBool        issil;
  unsigned long node;    /* issil=TRUE: Related standard hidden node.
			    issil=FALSE: Related silence hidden node or
			                 0 if theare is no a related node.
			 */
  
} lexpref_hinfo_t;

/* Lèxic representat com un acceptor de prefixes, amb informació dels
   models de Markov embeguts en cada estat i de les paraules
   associades als estats finals de paraula. L'estat 0 és sempre
   l'estat inicial i representa la cadena buida. Quan es gasta
   crosswords, exceptuant el silenci inicial, la resta es tracte com
   estats ocults associats a un wildcard. */
typedef struct
{

  tLBool                 iscw;         /* Indica si és un lèxic de
					  crosswords. */
  unsigned long          N;            /* Número d'estats. */
  unsigned long          NS;           /* Número d'estats visibles més
					  estats de silenci. */
  unsigned long          NR;           /* Número d'estats reals: N + NS + els
					  ocults. */
  unsigned long          NE;           /* Número d'arestes de l'arbre. */
  lexpref_state_t       *states;       /* Estats. */
  lexpref_state_edges_t *state_eds;    /* |NS|. Arestes per als estats
					  no ocults i de silenci. */
  unsigned long         *edges;        /* Cada aresta representa
					  l'identificador de l'estat
					  destí. */
  lexpref_fedge_t       *fedges;       /* Cada aresta representa una
					  paraula i una probabilitat
					  de transitar. */
  tLBool                *mpred;        /* |NS|. Indica si un estat té
					  multiples
					  predecessors. Special hidden
					  silece nodes always have
					  only one predecessor. */
  unsigned long         *psils;        /* |N|. Quan hi ha silenci,
                                           punter al node de silenci
                                           associat. */
  unsigned long         *sil2vis;      /* |NS|. Però el primer element
                                           és N. Inversa de psils. */
  unsigned long         *_sil2vis;
  unsigned long          isil;         /* Identificador del silenci
					  inicial. */
  
  /* CW. */
  lexpref_hinfo_t       *hinfo;     /* |NR|. First element is NS. Only
                                        available in
                                        crosswords. Connecte standard
                                        hidden nodes with special
                                        silence hidden nodes. */
  lexpref_hinfo_t       *_hinfo;
  size_t                 initcw;    /* Símbol inicial quan s'utilitza
				       crosswords. */
  tLBool                *isfinal;   /* Per a cada símbol (realment és
				       per a cada NS del lexicon) diu
				       si és o no final. */
  lexpref_cwlist_t      *cwlist;    /* Per a cada símbol (de NS), té
				       la llista de nodes ocults de
				       l'arbre de prefixe que es poden
				       alcançar al començar
				       paraula. */
  unsigned long         *cwedges;   /* Arestes de les llistes. */
  
  
} lexpref_t;

/* Si sil >= amodel->N no hi ha símbol de silenci. */
lexpref_t *
tL_create_lexpref (
		   const tLLexicon *lexicon,
		   const tLAModel  *amodel,
		   const size_t     sil
		   );
void
tL_free_lexpref (
		 lexpref_t *lexpref
		 );

/*
void
lexpref_print_dot (
		   const lexpref_t *lexpref,
		   const tLDict    *syms,
		   const tLDict    *words,
		   FILE            *to
		   );
*/


#endif /* __TL_PRV_LEXICON_RECAUX_H__ */
