/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  linabmat.c - Implementació de 'linabmat.h'.
 *
 *  Adrià Giménez Pastor, 2010,2011,2012.
 *
 */


#include <stddef.h>
#include <stdlib.h>

#include "error.h"
#include "linabmat.h"
#include "mem.h"




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
calc_alpha_lin_row (
		    const linfb_meths_t *fb_meths,
                    const tLAModelHMM  *cchar,
                    const unsigned int   q,
                    tLFloat             *asq1,
                    tLFloat             *asq0,
                    const int            begin,
                    const int            end,
                    const tLFloat        prev_trans,
                    const tLFea         *fea,
		    void                *udata
                    )
{
  
  int t;
  tLFloat a, b, trans;
  
  
  asq1[begin]= asq0[begin-1] + prev_trans +
    fb_meths->a_get_prob ( fea, begin, cchar->states, q, udata );
  trans= cchar->trans.lineal.p[q].p0;
  for ( t= begin + 1; t < end; ++t )
    {
      a= asq0[t-1] + prev_trans;
      b= asq1[t-1] + trans;
      asq1[t]= RADD ( a, b ) +
	fb_meths->a_get_prob ( fea, t, cchar->states, q, udata );
    }
  
} /* end calc_alpha_lin_row */


static tLFloat *
calc_alpha_lin_body (
		     const linfb_meths_t  *fb_meths,
                     const tLAModelHMM   *cchar,
                     tLFloat              *asq1,
                     tLFloat             **as,
                     int                  *begin,
                     int                  *end,
                     const tLFea          *fea,
		     void                 *udata
                     )
{
  
  unsigned int q;
  tLFloat *asq0;
  
  
  for ( q= 1; q < cchar->Q; ++q )
    {
      
      asq0= asq1; asq1= as[q];
      ++(*begin); ++(*end);
      
      calc_alpha_lin_row ( fb_meths, cchar, q, asq1, asq0, *begin, *end,
			   cchar->trans.lineal.p[q-1].p1, fea, udata );
      
    }
  
  return asq1;
  
} /* end calc_alpha_lin_body */


static void
calc_beta_lin_row (
		   const linfb_meths_t *fb_meths,
                   tLFloat             *b,
                   tLFloat             *asq,
                   const int            begin,
                   const int            end,
                   const tLFloat        trans_p0,
                   const tLFloat        trans_p1,
		   void                *fstats_cq0,
                   void                *fstats_cq1,
                   tLFloat             *calc0,
                   tLFloat             *calc1,
                   const unsigned int   q,
		   void                *udata
                   )
{
  
  int t, N0, N1;
  tLFloat a, a2, b0, b1, max0, max1, aux, alpha, count0, count1;
  
  
  b1= b[begin];
  b[begin]= b0= trans_p1 +
    fb_meths->b_get_prob ( fstats_cq1, begin+1 ) + b[begin+1];
  calc1[0]= max1= aux= asq[begin] + b0;
  fb_meths->b_accum_weight ( fstats_cq0, begin, aux, udata );
  max0= -HUGE_VAL; N0= 0; N1= 1;
  for ( t= begin - 1; t > end; --t )
    {
      a=  trans_p0 + fb_meths->b_get_prob ( fstats_cq0, t+1 ) + b0;
      a2= trans_p1 + fb_meths->b_get_prob ( fstats_cq1, t+1 ) + b1;
      b1= b[t];
      b[t]= b0= RADD ( a, a2 );
      alpha= asq[t];
      fb_meths->b_accum_weight ( fstats_cq0, t, alpha + b0, udata );
      calc0[N0++]= aux= alpha + a;  if ( aux > max0 ) max0= aux;
      calc1[N1++]= aux= alpha + a2; if ( aux > max1 ) max1= aux;
    }
  count0= (N0>1&&max0!=-HUGE_VAL) ? tL_radd_vals ( calc0, max0, N0 ) : max0;
  count1= (N1>1&&max1!=-HUGE_VAL) ? tL_radd_vals ( calc1, max1, N1 ) : max1;
  fb_meths->b_accum_trans ( count0, count1, q, udata );
  
} /* end calc_beta_lin_row */


static void
calc_beta_lin_body (
		    const linfb_meths_t  *fb_meths,
                    const tLAModelHMM   *cchar,
                    tLFloat              *b,
                    tLFloat             **as,
                    int                  *begin,
                    int                  *end,
                    tLFloat              *calc0,
                    tLFloat              *calc1,
		    void                 *udata
                    )
{
  
  unsigned int q;
  void *fstats_cq1, *fstats_cq0;
  tLFloat *asq;
  
  
  q= cchar->Q-1;
  fstats_cq0= fb_meths->b_get_fstats ( q, udata );
  while ( q > 0 )
    {
      
      asq= as[--q];
      --(*begin); --(*end);
      fstats_cq1= fstats_cq0;
      fstats_cq0= fb_meths->b_get_fstats ( q, udata );
      
      calc_beta_lin_row ( fb_meths, b, asq, *begin, *end,
                          cchar->trans.lineal.p[q].p0,
                          cchar->trans.lineal.p[q].p1,
                          fstats_cq0, fstats_cq1,
			  calc0, calc1, q, udata );
      
    }
  
} /* end calc_beta_lin_body */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_linabmat_free (
		  linabmat_t *m
		  )
{
  
  free ( m->calc1 );
  free ( m->b );
  free ( m->a[0][0] );
  free ( m->a[0] );
  free ( m->a );
  free ( m );
  
} /* end tL_linabmat_free */


linabmat_t *
tL_linabmat_new (
		 const tLAModel *amodel,
		 unsigned int    max_length
		 )
{
  
  linabmat_t *new;
  unsigned int c, max_Q;
  tLFloat **aux;
  
  
  max_Q= amodel->max_Q;
  mt ( new= mem_alloc ( linabmat_t, 1 ) );
  mt ( new->a= mem_alloc ( tLFloat **, max_length ) );
  mt ( new->a[0]= mem_alloc ( tLFloat *, max_length * max_Q ) );
  for ( c= 1, aux= new->a[0] + max_Q; c < max_length; ++c, aux+= max_Q )
    new->a[c]= aux;
  mt ( new->a[0][0]= mem_alloc ( tLFloat, max_length * max_Q ) );
  mt ( new->b= mem_alloc ( tLFloat, 1 ) );
  new->max_width= 1;
  mt ( new->calc1= mem_alloc ( tLFloat, 2 ) );
  new->calc0= new->calc1 + 1;
  
  return new;
  
} /* end tL_linabmat_new */


void
tL_linabmat_resize (
		    linabmat_t   *m,
		    unsigned int  max_length,
		    unsigned int  max_Q,
		    int           width
		    )
{
  
  unsigned int q, s;
  tLFloat ***a, **a_s, *aux;
  
  
  mt ( m->a[0][0]= aux=
       mem_realloc ( tLFloat, m->a[0][0], max_length * max_Q * width ) );
  a= m->a;
  for ( s= 0; s < max_length; ++s )
    {
      a_s= a[s];
      for ( q= 0; q < max_Q; ++q, aux+= width )
        a_s[q]= aux;
    }
  mt ( m->b= mem_realloc ( tLFloat, m->b, width ) );
  mt ( m->calc1= mem_realloc ( tLFloat, m->calc1, width*2 ) );
  m->calc0= m->calc1 + width;
  m->max_width= width;
  
} /* end tL_linabmat_resize */


tLFloat
tL_calc_alpha_lin (
		   linabmat_t          *m,
		   const linfb_meths_t *fb_meths,
		   const tLFea         *fea,
		   int                  diff,
		   const tLFloat        prior,
		   void                *udata
		   )
{
  
  tLFloat ***a, **as, *asq1, *asq0, prob_f, auxp;
  const tLAModelHMM *cchar;
  int begin, end, t;
  size_t s, length;
  
  
  /* Prepara. */
  a= m->a;
  length= fb_meths->get_length ( udata );
  
  /* s=0. */
  cchar= fb_meths->a_set_sym ( 0, udata );
  as= a[0];
  
  /* --> q=0. */
  begin= 0; end= diff;
  asq1= as[0];
  asq1[0]= prior + fb_meths->a_get_prob ( fea, 0, cchar->states, 0, udata );
  auxp= cchar->trans.lineal.p[0].p0;
  for ( t= 1; t < diff; ++t )
    asq1[t]= asq1[t-1] + auxp +
      fb_meths->a_get_prob ( fea, t, cchar->states, 0, udata );
  
  /* --> q>0. */
  asq1= calc_alpha_lin_body ( fb_meths, cchar, asq1, as,
			      &begin, &end, fea, udata );
  prob_f= cchar->trans.lineal.p[cchar->Q-1].p1;
  
  /* s>0. */
  for ( s= 1; s < length; ++s )
    {
      
      cchar= fb_meths->a_set_sym ( s, udata );
      as= a[s];
      
      /* --> q=0. */
      ++begin; ++end;
      asq0= asq1; asq1= as[0];
      calc_alpha_lin_row ( fb_meths, cchar, 0, asq1, asq0,
			   begin, end, prob_f, fea, udata );
      
      /* --> q>0. */
      asq1= calc_alpha_lin_body ( fb_meths, cchar, asq1, as,
				  &begin, &end, fea, udata );
      prob_f= cchar->trans.lineal.p[cchar->Q-1].p1;
      
    }
  
  return asq1[end-1] + prob_f;
  
} /* end tL_calc_alpha_lin */


void
tL_calc_beta_lin (
		  linabmat_t          *m,
		  const linfb_meths_t *fb_meths,
		  const int            T,
		  const int            diff,
		  void                *udata
		  )
{
  
  const tLAModelHMM *cchar;
  tLFloat *b, b0, auxp, ***a, **as, *asq, count0, count1, max, aux;
  size_t s, length;
  int begin, end, t, N;
  void *fstats_cq;
  tLFloat *calc0, *calc1;
  
  
  /* Prepara. */
  a= m->a;
  b= m->b;
  calc0= m->calc0;
  calc1= m->calc1;
  length= fb_meths->get_length ( udata );
  
  /* s=Length-1. */
  s= length-1;
  cchar= fb_meths->b_set_sym ( s, udata );
  as= a[s];
  
  /* --> q=Q-1. */
  asq= as[cchar->Q-1];
  fstats_cq= fb_meths->b_get_fstats ( cchar->Q-1, udata );
  begin= T-1; end= begin - diff;
  b[begin]= b0= cchar->trans.lineal.p[cchar->Q-1].p1;
  count1= asq[begin] + b0;
  fb_meths->b_accum_weight ( fstats_cq, begin, count1, udata );
  max= -HUGE_VAL; N= 0;
  auxp= cchar->trans.lineal.p[cchar->Q-1].p0;
  for ( t= begin-1; t > end; --t )
    {
      b[t]= b0= auxp + fb_meths->b_get_prob ( fstats_cq, t+1 ) + b[t+1];
      calc0[N++]= aux= asq[t] + b0;
      fb_meths->b_accum_weight ( fstats_cq, t, aux, udata );
      if ( aux > max ) max= aux;
    }
  count0= (N>1&&max!=-HUGE_VAL) ? tL_radd_vals ( calc0, max, N ) : max;
  fb_meths->b_accum_trans ( count0, count1, cchar->Q-1, udata );
  
  /* --> q<(Q-1). */
  calc_beta_lin_body ( fb_meths, cchar, b, as, &begin,
		       &end, calc0, calc1, udata );
  
  /* s<(Length-1). */
  while ( s > 0 )
    {
      
      fstats_cq= fb_meths->b_get_fstats ( 0, udata );
      cchar= fb_meths->b_set_sym ( --s, udata );
      as= a[s];
      
      /* --> q=Q-1. */
      --begin; --end;
      calc_beta_lin_row ( fb_meths, b, as[cchar->Q-1], begin, end,
                          cchar->trans.lineal.p[cchar->Q-1].p0,
                          cchar->trans.lineal.p[cchar->Q-1].p1,
			  fb_meths->b_get_fstats ( cchar->Q-1, udata ),
                          fstats_cq, calc0, calc1, cchar->Q-1, udata );
      
      /* --> q<(Q-1). */
      calc_beta_lin_body ( fb_meths, cchar, b, as, &begin,
			   &end, calc0, calc1, udata );
      
    }
  
} /* end tL_calc_beta_lin */
