/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lm.c - Implementació de 'lm.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2011,2012.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "common.h"
#include "error.h"
#include "lm.h"
#include "mem.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define INOB(I) ((iter_nob_t *) (I))
#define IB(I) ((iterb_t *) (I))

#define ED(PTR) ((const tLLMEdge *) (PTR))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  LM_ITERATOR_CONTENT
  tLLMIter i;
  tLBool    set;
  
} iter_nob_t;


typedef struct
{
  
  size_t          state;
  const tLLMEdge *next;
  unsigned long   limitw;
  tLProb          abowt;
  
} iterb_node_t;


typedef struct
{
  
  LM_ITERATOR_CONTENT
  tLLMEdge      *buffer;      /* Buffer on escriure les arestes*/
  iterb_node_t  *stack;       /* Pila. */
  size_t         size;        /* Grandària de la pila. */
  size_t         N;           /* Número d'elements en la pila. */
  unsigned long  nextw;       /* Següent paraula a buscat. */
  
} iterb_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* Torna -1 en cas d'error, 0 en cas contrari. */
static int
next_token (
	    char                 *str,
	    char                **begin,
	    char                **end,
	    const char           *token,
	    const unsigned long   line,
	    char                **err
	    )
{
  
  if ( tL_gtokenstr ( str, begin, end ) ||
       begin == NULL )
    {
      tL_msgerror ( err, _( "In line %lu, unable to read the token %s,"
		    " or too few fields" ) , line, token );
      return -1;
    }
  *((*end)++)= '\0';
  if ( strcmp ( *begin, token ) )
    {
      tL_msgerror ( err, _( "In line %lu, the token %s was read instead of the"
		    " token %s" ) , line, *begin, token );
      return -1;
    }
  
  return 0;
  
} /* end next_token */


static void
iter_nob_free (
	       tLLMIterator i
	       )
{
  free ( i );
} /* end iter_nob_free */


static void
iter_nob_set (
	      tLLMIterator i,
	      size_t        state
	      )
{
  
  INOB(i)->i= INOB(i)->lm->states[state].edges;
  INOB(i)->set= TRUE;
  
} /* end iter_nob_set */


static tLBool
iter_nob_next (
	       tLLMIterator     i,
	       const tLLMEdge **begin,
	       const tLLMEdge **end
	       )
{
  
  if ( INOB(i)->set )
    {
      *begin= INOB(i)->i.begin;
      *end= INOB(i)->i.end;
      INOB(i)->set= FALSE;
      return TRUE;
    }
  
  return FALSE;
  
} /* end iter_nob_next */


static const lm_iterator_meths_t iter_nob_meths=
  {
    iter_nob_free,
    iter_nob_set,
    iter_nob_next
  };


static tLLMIterator
make_iter_nob (
	       const tLLM *lm
	       )
{
  
  iter_nob_t *new;
  
  
  mt ( new= mem_alloc ( iter_nob_t, 1 ) );
  new->lm= lm;
  new->meths= &iter_nob_meths;
  new->set= FALSE;
  
  return (tLLMIterator) new;
  
} /* end make_iter_nob */


/* Torna, del conjunt d'arestes apuntades per l'iterador, l'aresta amb
   la paraula més xicoteta que és major o igual que W. Torna
   l'indicador de END si totes les paraules són inferiors. */
static const tLLMEdge *
search_edge (
	     tLLMIter      i,
	     unsigned long w
	     )
{
  
  size_t N, l, r, m, diff;
  const tLLMEdge *v;
  unsigned long wm;
  
  
  v= i.begin;
  N= (size_t) (i.end - i.begin);
  if ( N == 0 || v[N-1].w < w ) return i.end;
  if ( N == 1 || w <= v->w ) return v;
  l= 0; r= N;
  for (;;)
    {
      diff= r-l;
      if ( diff == 1 ) return &(v[l]);
      if ( (v[r-1].w-v[l].w)+1 == diff )
	return &(v[l+(w-v[l].w)]);
      m= l+(diff>>1); wm= v[m].w;
      if ( wm == w ) return &(v[m]);
      else if ( wm > w )
	{
	  if ( l == m || v[m-1].w < w ) return &(v[m]);
	  r= m;
	}
      else
	{
	  l= m+1;
	  if ( w <= v[l].w ) return &(v[l]);
	}
    }
  
} /* end search_edge */


static void
iterb_free (
	    tLLMIterator i
	    )
{
  
  free ( IB(i)->buffer );
  free ( IB(i)->stack );
  free ( i );
  
} /* end iterb_free */


static void
iterb_set (
	   tLLMIterator i,
	   size_t        state
	   )
{
  
  IB(i)->stack->state= state;
  IB(i)->stack->next= IB(i)->lm->states[state].edges.begin;
  IB(i)->stack->limitw= IB(i)->lm->NW;
  IB(i)->stack->abowt= 0.0;
  IB(i)->N= 1;
  IB(i)->nextw= 0;
  
} /* end iterb_set */


static void
iterb_next_add_stack (
		      iterb_t      *j,
		      iterb_node_t  node
		      )
{
  
  if ( j->N == j->size )
    {
      (j->size)<<= 1;
      mt ( j->stack= mem_realloc ( iterb_node_t,
				   j->stack, j->size ) );
    }
  j->stack[(j->N)++]= node;
  
} /* end iterb_next_add_stack */


static tLBool
iterb_next (
	    tLLMIterator i,
	    const tLLMEdge **begin,
	    const tLLMEdge **end
	    )
{
  
  iterb_t j;
  const tLLMState *states;
  tLProb bowt;
  iterb_node_t node;
  const tLLMEdge *aux;
  tLLMEdge *p;
  tLBool cond;
  
  
  j= *(IB(i));
  if ( j.N == 0 ) return FALSE;
  node= j.stack[--j.N];
  states= j.lm->states;
  aux= states[node.state].edges.end;
  
  /* Mentres no siga el que busque intente baixar tot el que puga. */
  while ( ((cond= (node.next ==  aux || node.next->w >= node.limitw)) ||
	   node.next->w > j.nextw) &&
	  (bowt= states[node.state].bowt) != -HUGE_VAL )
    {
      
      /* Insert recovery point if we didn't reach the end edge of this
	 level */
      if ( !cond )
	{
	  iterb_next_add_stack ( &j, node );
	  node.limitw= node.next->w;
	}
      
      /* Update bow and search for the begining. */
      node.abowt+= bowt;
      node.state= states[node.state].bos;
      node.next= search_edge ( states[node.state].edges, j.nextw );
      aux= states[node.state].edges.end;
      
    }
  
  /* Si no he trobat res baix m'en vaig dalt a la primera que tinga
     arestes encara que no siga la pròxima que busque. Com no he
     insertat els nodes que no tenen arestes d'interes, el que busque
     és el següent de la pila. */
  if ( cond )
    {
      if ( j.N != 0 )
	{
	  node= j.stack[--j.N];
	  aux= states[node.state].edges.end;
	}
      else { *(IB(i))= j; return FALSE; }
    }
  j.nextw= node.next->w; /* Si no he trobat l'aresta concreta, me vote
			    arestes. El motiu pot ser perque no puga
			    baixar per backoff o per que no he trobat
			    res de trellat baixant. */
  
  /* Crea les arestes. */
  for ( *begin= p= j.buffer;
	node.next != aux &&
	  node.next->w == j.nextw &&
	  j.nextw < node.limitw;
	++p, ++node.next, ++j.nextw )
    {
      *p= *node.next;
      p->p+= node.abowt;
    }
  if ( j.nextw < node.limitw )
    iterb_next_add_stack ( &j, node );
  *(IB(i))= j;
  *end= p;
  
  return TRUE;
  
} /* end iterb_next */


static const lm_iterator_meths_t iterb_meths=
  {
    iterb_free,
    iterb_set,
    iterb_next
  };


static tLLMIterator
make_iterb (
	    const tLLM *lm
	    )
{
  
  iterb_t *new;
  
  
  new= mem_alloc ( iterb_t, 1 );
  new->lm= lm;
  new->meths= &iterb_meths;
  new->buffer= mem_alloc ( tLLMEdge, lm->NW );
  new->stack= mem_alloc ( iterb_node_t, 1 );
  new->size= 1;
  new->N= 0;
  
  return (tLLMIterator) new;
  
} /* end make_iterb */


static int
cmpedge (
	 const void *a,
	 const void *b
	 )
{
  
  if ( ED(a)->w < ED(b)->w ) return -1;
  return (ED(a)->w != ED(b)->w);
  
} /* end cmpedge */


static void
tL_lm_print_text (
		  const tLLM   *lm,
		  FILE         *to,
		  const tLDict *words
		  )
{
  
  size_t n;
  
  
  fprintf ( to, "LM\n" );
  fprintf ( to, "Backoff %s\n", lm->backoff?"YES":"NO"  );
  fprintf ( to, "NStates %lu\n", _PS(lm->NS) );
  fprintf ( to, "NEdges %lu\n", _PS(lm->NE) );
  fprintf ( to, "States\n" );
  for ( n= 0; n < lm->NS; ++n )
    {
      fprintf ( to, "%lu ", _PS(n) );
      if ( n == lm->start )
	fprintf ( to, "START " );
      else if ( lm->states[n].word == ULONG_MAX )
	fprintf ( to, "NULL " );
      else
	fprintf ( to, "\"%s\" ",
		  tL_dict_get_token ( words,
				lm->states[n].word ) );
      fprintf ( to, "FP " );
      tL_prob_print ( lm->states[n].fp, to );
      fprintf ( to, " B %lu E %lu NuB %lu NuE %lu",
		_PS(lm->states[n].edges.begin - lm->edges),
		_PS(lm->states[n].edges.end - lm->edges),
		_PS(lm->states[n].null_edges.begin - lm->edges),
		_PS(lm->states[n].null_edges.end - lm->edges) );
      if ( lm->backoff )
	{
	  fprintf ( to, " BOS %lu BOW ", _PS(lm->states[n].bos) );
	  tL_prob_print ( lm->states[n].bowt, to );
	}
      fputc ( '\n', to );
    }
  fprintf ( to, "Edges\n" );
  for ( n= 0; n < lm->NE; ++n )
    {
      fprintf ( to, "%lu To %lu Prob ",
		_PS(n),	_PS(lm->edges[n].to) );
      tL_prob_print ( lm->edges[n].p, to );
      fputc ( '\n', to );
    }
  
} /* end tL_lm_print_text */


static void
tL_lm_print_binary (
		    const tLLM   *lm,
		    FILE         *to,
		    const tLDict *words
		    )
{
  
  size_t n;
  
  
  fprintf ( to, "BLM\n" );
  fprintf ( to, "Backoff %s\n", lm->backoff?"YES":"NO"  );
  fprintf ( to, "NStates %lu\n", _PS(lm->NS) );
  fprintf ( to, "NEdges %lu\n", _PS(lm->NE) );
  fprintf ( to, "Start %lu\n", _PS(lm->start) ); 

  fprintf ( to, "Lexicon %lu\n", _PS(words->N) );
  for ( n= 0; n < words->N ; ++n )
  {
    tL_print_string_binary ( to, tL_dict_get_token ( words, n ) );
  }

  fwritee ( lm->states, sizeof(tLLMState), lm->NS, to ); 

  /* offset of states */
  fwritee ( &lm->edges , sizeof(tLLMEdge *), 1, to);
  /* the edges */
  fwritee ( lm->edges, sizeof(tLLMEdge), lm->NE, to );
  
} /* end tL_lm_print_binary */


TL_API_FUNC(void)
tL_lm_print (
	     const tLLM   *lm,
	     FILE         *to,
	     const tLDict *words,
	     const tLBool  binary
	     )
{
  if ( binary )
  {
    tL_lm_print_binary( lm, to, words );
  }
  else
  {
    tL_lm_print_text( lm, to, words );
  }

} /* end tL_lm_print_text */


static tLLM *
tL_lm_new_from_file_text (
			  gzFile         from,
			  tLLexicon     *lexicon,
			  const tLBool   emptylex,
			  const tLDict  *syms,
			  char         **err
			  )
{
  
  tLLM *lm;
  tLBuffer *buffer;
  size_t cs_size, NS, NE, s, ind, length, e;
  tLSeq cs;
  tLLMState *states, state;
  char *begin, *end;
  tLBool start;
  tLLMEdge *ped, *aux;
  unsigned long line;
  int ret;  
  
  
  /* Inicialitza. */
  line= 2;
  buffer= tL_buffer_new ();
  mt ( lm= mem_alloc ( tLLM, 1 ) );
  lm->_gsf= 1.0; lm->_wip= 0.0;
  lm->states= NULL;
  lm->edges= NULL;
  if ( emptylex )
    {
      cs_size= 1;
      cs.length= 0;
      mt ( cs.v= mem_alloc ( size_t, 1 ) );
      cs.v[0]= 0;
    }
  else cs.v= NULL;
  
  /* Capçalera. */
  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "Backoff ", 8 ) ||
       buffer->v[8] == '\0' )
    {
      tL_msgerror ( err, _( "In line %lu, unable to read 'Backoff <val>'" ) ,
		    line );
      goto error;
    } ++line;
  if ( !strcmp ( buffer->v+8, "NO" ) ) lm->backoff= FALSE;
  else if ( !strcmp ( buffer->v+8, "YES" ) ) lm->backoff= TRUE;
  else
    {
      tL_msgerror ( err, _( "Unknown Backoff value '%s', valid values"
		    " are YES or NO" ) , buffer->v+8 );
      goto error;
    }
  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "NStates ", 8 ) ||
       buffer->v[8] == '\0' )
    {
      tL_msgerror ( err, _( "In line %lu, unable to read 'NStates <val>'" ) ,
		    line );
      goto error;
    } ++line;
  NS= lm->NS= atos ( buffer->v+8 );
  if ( NS == 0 )
    {
      tL_msgerror ( err, _( "NStates cannot be 0" ) );
      goto error;
    }
  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "NEdges ", 7 ) ||
       buffer->v[7] == '\0' )
    {
      tL_msgerror ( err, _( "In line %lu, unable to read the token"
		    " 'NEdges <val>'" ) , line );
      goto error;
    } ++line;
  NE= lm->NE= atos ( buffer->v+7 );
  if ( NE == 0 )
    {
      tL_msgerror ( err, _( "NEdges cannot be 0" ) );
      goto error;
    }
  mt ( states= lm->states= mem_alloc ( tLLMState, NS ) );
  mt ( ped= lm->edges= mem_alloc ( tLLMEdge, NE ) );

  /* Estats. */
  start= FALSE;
  if ( !tL_gline ( from, buffer ) ||
       strcmp ( buffer->v, "States" ) )
    {
      tL_msgerror ( err, _( "In line %lu, unable to read 'States'" ) , line );
      goto error;
    } ++line;
  for ( s= 0; s < NS; ++s )
    {
      
      if ( !tL_gline ( from, buffer ) )
	{
	  tL_msgerror ( err, _( "Description of the %luth state was expected"
				" to be in line %lu, but line %lu doesn't"
				" exist" ) ,
			_PS(s), line, line );
	  goto error;
	}
      
      /* Índex. */
      if ( tL_gtokenstr ( buffer->v, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the state index,"
				" or too few fields" ) ,
			line );
	  goto error;
	}
      *(end++)= '\0';
      if ( (ind= atos ( buffer->v )) != s )
	{
	  tL_msgerror ( err, _( "The index of state %lu was expected to be in"
				" line %lu, instead the index of state"
				" %lu was read" ) ,
			_PS(s), line, _PS(ind) );
	  goto error;
	}
      
      /* Paraula. */
      if ( tL_gtokenstr ( end, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the word,"
				" or too few fields" ) , line );
	  goto error;
	}
      *(end++)= '\0';
      if ( !strcmp ( begin, "START" ) )
	{
	  start= TRUE;
	  state.word= ULONG_MAX;
	  lm->start= s;
	}
      else if ( !strcmp ( begin, "NULL" ) )
	state.word= ULONG_MAX;
      else
	{
	  if ( (length= strlen ( begin )) < 3 ||
	       begin[0] != '\"' ||
	       begin[length-1] != '\"' )
	    {
	      tL_msgerror ( err, _( "In line %lu, the word was expected"
				    " to be in format \"word\"" ) , line );
	      goto error;
	    }
	  (begin++)[length-1]= '\0';
	  state.word=
	    (unsigned long) tL_dict_find ( lexicon->words, begin );
	  if ( state.word == lexicon->words->N )
            {
              if ( emptylex )
                {
		  if ( syms != NULL )
		    {
		      if ( tL_utf8tocseq ( begin, syms, &cs,
					   &cs_size, err ) == -1 )
			goto error;
		    }
                  tL_lexicon_add ( lexicon, begin, 0.0, &cs );
                }
              else
		{
		  tL_msgerror ( err,
				_( "In line %lu, '%s' is out of vocabulary" ),
				line, begin );
		  goto error;
		}
            }
	}
      
      /* Probabilitat final. */
      if ( next_token ( end, &begin, &end, "FP", line, err ) == -1 )
	goto error;
      if ( tL_gtokenstr ( end, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the final"
				" log-probability, or too few fields" ),
			line );
	  goto error;
	}
      *(end++)= '\0';
      state.fp= tL_atop ( begin );
      
      /* Arestes. Begin. */
      if ( next_token ( end, &begin, &end, "B", line, err ) == -1 )
	goto error;
      if ( tL_gtokenstr ( end, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the identifier"
				" of the beginning edge B, or too few fields" ),
			line );
	  goto error;
	}
      *(end++)= '\0';
      state.edges.begin= lm->edges + atos ( begin );
      
      /* Arestes. End. */
      if ( next_token ( end, &begin, &end, "E", line, err ) == -1 )
	goto error;
      if ( tL_gtokenstr ( end, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the identifier"
				" of the ending edge E, or too few fields" ),
			line );
	  goto error;
	}
      *(end++)= '\0';
      state.edges.end= lm->edges + atos ( begin );
      
      /* Arestes. Begin null. */
      if ( next_token ( end, &begin, &end, "NuB", line, err ) == -1 )
	goto error;
      if ( tL_gtokenstr ( end, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the identifier"
				" of the beginning null edge NuB, or too few"
				" fields" ) ,
			line );
	  goto error;
	}
      *(end++)= '\0';
      state.null_edges.begin= lm->edges + atos ( begin );
      
      
      /* Arestes. End null. */
      if ( next_token ( end, &begin, &end, "NuE", line, err ) == -1 )
	goto error;
      ret= tL_gtokenstr ( end, &begin, &end );
      if ( begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the identifier"
				" of the ending null-edge NuE" ) , line );
	  goto error;
	}
      if ( lm->backoff ) *(end++)= '\0';
      state.null_edges.end= lm->edges + atos ( begin );
      
      if ( lm->backoff )
	{
	  
	  if ( ret )
	    {
	      tL_msgerror ( err, _( "Too few fields in line %lu" ) , line );
	      goto error;
	    }
	  
	  /* Estat de backoff. */
	  if ( next_token ( end, &begin, &end, "BOS", line, err ) == -1 )
	    goto error;
	  if ( tL_gtokenstr ( end, &begin, &end ) ||
	       begin == NULL )
	    {
	      tL_msgerror ( err, _( "In line %lu, unable to read the back-off"
				    " state BOS, or too few fields" ) , line );
	      goto error;
	    }
	  *(end++)= '\0';
	  state.bos= atos ( begin );
	  if ( state.bos >= lm->NS )
	    {
	      tL_msgerror ( err, _( "In line %lu, back-off state %lu"
				    " doesn't exist" ) , line, state.bos );
	      goto error;
	    }
	  
	  /* Pes de backoff. */
	  if ( next_token ( end, &begin, &end, "BOW", line, err ) == -1 )
	    goto error;
	  tL_gtokenstr ( end, &begin, &end );
	  if ( begin == NULL )
	    {
	      tL_msgerror ( err, _( "In line %lu, unable to read the"
				    " back-off weight BOW" ) , line );
	      goto error;
	    }
	  state.bowt= tL_atop ( begin );
	  
	}
      else { state.bos= 0; state.bowt= -HUGE_VAL; }
      
      if ( !tL_gtokenstr ( end, &begin, &end ) ||
	   begin != NULL )
	{
	  tL_msgerror ( err, _( "Too many fields in line %lu" ) , line );
	  goto error;
	}
      
      aux= lm->edges+lm->NE;
      if ( state.edges.begin > aux ||
	   state.edges.begin > state.edges.end ||
	   state.edges.end > aux ||
	   state.null_edges.begin > aux ||
	   state.null_edges.begin > state.edges.end ||
	   state.null_edges.end > aux )
	{
	  tL_msgerror ( err, _( "In line %lu, edge indexes are out of"
				" range" ) , line );
	  goto error;
	}
      lm->states[s]= state;
      
      ++line;
      
    }
  lm->NW= lexicon->words->N;
  if ( !start )
    {
      tL_msgerror ( err, _( "There is not a initial state START" ) );
      goto error;
    }
  
  /* Arestes. */
  if ( !tL_gline ( from, buffer ) ||
       strcmp ( buffer->v, "Edges" ) )
    {
      tL_msgerror ( err, _( "In line %lu, unable to read 'Edges'" ) , line );
      goto error;
    } ++line;
  for ( e= 0; e < lm->NE; ++e )
    {
      
      if ( !tL_gline ( from, buffer ) )
	{
	  tL_msgerror ( err, _( "Description of the %luth edge was expected"
				" to be in line %lu, but line %lu doesn't"
				" exist" ) ,
                        _PS(e), line, line );
	  goto error;
	}
      
      /* Índex. */
      if ( tL_gtokenstr ( buffer->v, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err, _( "In line %lu, unable to read the edge index,"
				" or too few fields" ) , line );
	  goto error;
	}
      *(end++)= '\0';
      if ( (ind= atos ( buffer->v )) != e )
	{
	  tL_msgerror ( err,
			_( "The index of edge %lu was expected to be in"
			   " line %lu, instead the index of edge %lu"
			   " was read" ),
                        _PS(e), line, _PS(ind) );
	  goto error;
	}
      
      /* Estat destí. */
      if ( next_token ( end, &begin, &end, "To", line, err ) == -1 )
	goto error;
      if ( tL_gtokenstr ( end, &begin, &end ) ||
	   begin == NULL )
	{
	  tL_msgerror ( err,
			_( "In line %lu, unable to read the"
			   " target state To, or too few fields" ),
			line );
	  goto error;
	}
      *(end++)= '\0';
      ped->to= atos ( begin );
      if ( ped->to >= lm->NS )
	{
	  tL_msgerror ( err, _( "In line %lu, the target state %lu doesn't"
				" exist" ) , line, ped->to );
	  goto error;
	}
      ped->w= states[ped->to].word;
      
      /* Probabilitat. */
      if ( next_token ( end, &begin, &end, "Prob", line, err ) == -1 )
	goto error;
      tL_gtokenstr ( end, &begin, &end );
      if ( begin == NULL )
	{
	  tL_msgerror ( err,
			_( "In line %lu, unable to read the log-probability"
			   " Prob of the %luth edge" ), line, _PS(e) );
	  goto error;
	}
      ped->p= tL_atop ( begin );
      
      if ( !tL_gtokenstr ( end, &begin, &end ) ||
	   begin != NULL )
	{
	  tL_msgerror ( err, _( "Too many fields in line %lu" ) , line );
	  goto error;
	}
      
      ++ped;
      ++line;
      
    }
  
  /* Allibera memòria. */
  if ( emptylex )
    {
      tL_lexicon_adjust ( lexicon );
      free ( cs.v );
    }
  tL_buffer_free ( buffer );
  
  /* Ordena arestes. */
  if ( lm->backoff )
    for ( s= 0; s < NS; ++s )
      {
	state= states[s];
	qsort ( state.edges.begin,
		(size_t) (state.edges.end-state.edges.begin),
		sizeof(tLLMEdge), cmpedge );
      }
  
  return lm;
  
 error:
  if ( cs.v != NULL ) free ( cs.v );
  tL_buffer_free ( buffer );
  tL_lm_free ( lm );
  return NULL;
  
} /* end tL_lm_new_from_file_text */


static tLLM *
tL_lm_new_from_file_binary (
			    gzFile         from,
			    tLLexicon     *lexicon,
			    const tLBool   emptylex,
			    const tLDict  *syms,
			    char         **err
			    )
{
  
  tLLM *lm;
  tLBuffer *buffer;
  char *token;
  size_t cs_size, NS, NE, NWords, s, e;
  tLSeq cs;
  tLLMState *states, state;
  tLLMEdge *ped, *edges_offset;
  unsigned long *lex_mapping;
  int toread;  
  
  
  /* Inicialitza. */
  buffer= tL_buffer_new ();
  mt ( lm= mem_alloc ( tLLM, 1 ) );
  lm->_gsf= 1.0; lm->_wip= 0.0;
  lm->states= NULL;
  lm->edges= NULL;
  lex_mapping= NULL;
  if ( emptylex )
    {
      cs_size= 1;
      cs.length= 0;
      mt ( cs.v= mem_alloc ( size_t, 1 ) );
      cs.v[0]= 0;
    }
  else cs.v= NULL;
  
  /* Capçalera. */
  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "Backoff ", 8 ) ||
       buffer->v[8] == '\0' )
    {
      tL_msgerror ( err, _( "Unable to read 'Backoff <val>'" ) );
      goto error;
    }
  if ( !strcmp ( buffer->v+8, "NO" ) ) lm->backoff= FALSE;
  else if ( !strcmp ( buffer->v+8, "YES" ) ) lm->backoff= TRUE;
  else
    {
      tL_msgerror ( err, _( "Unknown Backoff value '%s', valid values"
		    " are YES or NO" ) , buffer->v+8 );
      goto error;
    }
  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "NStates ", 8 ) ||
       buffer->v[8] == '\0' )
    {
      tL_msgerror ( err, _( "Unable to read 'NStates <val>'" ) );
      goto error;
    }
  NS= lm->NS= atos ( buffer->v+8 );
  if ( NS == 0 )
    {
      tL_msgerror ( err, _( "NStates cannot be 0" ) );
      goto error;
    }
  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "NEdges ", 7 ) ||
       buffer->v[7] == '\0' )
    {
      tL_msgerror ( err, _( "Unable to read the token 'NEdges <val>'" ) );
      goto error;
    }
  NE= lm->NE= atos ( buffer->v+7 );
  if ( NE == 0 )
    {
      tL_msgerror ( err, _( "NEdges cannot be 0" ) );
      goto error;
    }
  mt ( states= lm->states= mem_alloc ( tLLMState, NS ) );
  mt ( ped= lm->edges= mem_alloc ( tLLMEdge, NE ) );

  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "Start ", 6 ) ||
       buffer->v[6] == '\0' )
    {
      tL_msgerror ( err, _( "Unable to read the token 'Start <val>'" ) );
      goto error;
    }
  lm->start= atos ( buffer->v+6 );
  if ( lm->start < 0 || lm->start >= NE )
    {
      tL_msgerror ( err,
		    _( "Start cannot be less than 0 or greater than NEdges" ) );
      goto error;
    } 

  /* Lexicon Start */

  mt ( lex_mapping= mem_alloc ( unsigned long, NS ) );

  if ( !tL_gline ( from, buffer ) ||
       strncmp ( buffer->v, "Lexicon ", 8 ) ||
       buffer->v[8] == '\0' )
    {
      tL_msgerror ( err, _( "Unable to read 'Lexicon <val>'" ) );
      goto error;
    }
  NWords= atos ( buffer->v+8 );
  if ( NS == 0 )
    {
      tL_msgerror ( err, _( "Lexicon cannot be 0" ) );
      goto error;
    }

  for ( e= 0; e < NWords; ++e )
  {
    token = tL_read_string_binary ( from );
    if ( token == NULL )
    {
      tL_msgerror ( err, _( "Unable to read the %uth word of the lexicon,"
                          " or bad format ('SYM')" ) , e+1 );
            goto error;
    }
    state.word=
      (unsigned long) tL_dict_find ( lexicon->words, token );
    if ( state.word == lexicon->words->N )
      {
	if ( emptylex )
	  {
	    if ( syms != NULL )
	      {
		if ( tL_utf8tocseq ( token, syms, &cs,
				     &cs_size, err ) == -1 )
		  goto error;
	      }
	    tL_lexicon_add ( lexicon, token, 0.0, &cs );
	  }
	else
	  {
	    tL_msgerror ( err,
			  _( "'%s' is out of vocabulary" ), token );
	    goto error;
	  }
      }
    /* store mapping to remap the following states to be read */
    lex_mapping[e] = state.word;
    
    free ( token );
  }
  lm->NW= lexicon->words->N;

  /* Start states */

  e = NS;
  states = lm->states;
  while ( e > 0)
  {
    toread = e > ( INT_MAX / sizeof(tLLMState) ) ? 
              ( INT_MAX / sizeof(tLLMState) ) : (int) e;

    if ( gzread ( from, states,
		  sizeof(tLLMState)*toread ) != sizeof(tLLMState)*toread )
      {
	tL_msgerror ( err, _( "Unable to read all the states" ) );
	goto error;
      }
    e -= toread;
    ped += toread;
  }
  states = lm->states;

  /* Start Edges */

  if ( gzread ( from, &edges_offset,
		sizeof(tLLMEdge *) )  != sizeof(tLLMEdge *) )
    {
      tL_msgerror ( err, _( "Unable to read the pointer offset of edges" ) );
      goto error;
    }
  
  e = NE;
  ped = lm->edges;
  while ( e > 0)
  {
    toread = e > ( INT_MAX / sizeof(tLLMEdge) ) ? 
              ( INT_MAX / sizeof(tLLMEdge) ) : (int) e;

    if ( gzread ( from, ped,
		  sizeof(tLLMEdge)*toread) != sizeof(tLLMEdge)*toread )
      {
	tL_msgerror ( err, _( "Unable to read all the edges" ) );
	goto error;
      }
    e -= toread;
    ped += toread;
  }
  
  ped = lm->edges;
  
  /* remapping */
  for ( e= 0; e < NS; ++e )
    {
    /* adjusting offsets and lexicon correspondency */
    if ( states[e].word != ULONG_MAX )
      states[e].word = lex_mapping[states[e].word];
    
    states[e].edges.begin = lm->edges + ( states[e].edges.begin - edges_offset );
    states[e].edges.end = lm->edges + ( states[e].edges.end - edges_offset ); 
    states[e].null_edges.begin = lm->edges + ( states[e].null_edges.begin - edges_offset );
    states[e].null_edges.end = lm->edges + ( states[e].null_edges.end - edges_offset );
  }

  /* THIS IS MANDATORY !!! */
  for ( e= 0; e < NE; ++e )
  {
    ped[e].w = states[ped[e].to].word;
  }

  /* Allibera memòria. */
  if ( emptylex )
    {
      tL_lexicon_adjust ( lexicon );
      free ( cs.v );
    }
  tL_buffer_free ( buffer );
  free ( lex_mapping );
  
  /* Ordena arestes. */
  if ( lm->backoff )
    for ( s= 0; s < NS; ++s )
      {
	state= states[s];
	qsort ( state.edges.begin,
		(size_t) (state.edges.end-state.edges.begin),
		sizeof(tLLMEdge), cmpedge );
      }
  
  return lm;

 error:
  if ( cs.v != NULL ) free ( cs.v );
  tL_buffer_free ( buffer );
  tL_lm_free ( lm );
  if ( lex_mapping != NULL ) free ( lex_mapping );
  return NULL;
  
} /* end tL_lm_new_from_file_binary */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_lm_free (
	    tLLM *lm
	    )
{
  
  if ( lm->states != NULL ) free ( lm->states );
  if ( lm->edges != NULL ) free ( lm->edges );
  free ( lm );
  
} /* end tL_lm_free */


void
tL_lm_bo_get_trans (
		    const tLLM          *lm,
		    const unsigned long  from,
		    const unsigned long  word,
		    unsigned long       *to,
		    tLProb              *prob
		    )
{
  
  tLLMState state;
  const tLLMEdge *ed;
  tLProb bowt;
  unsigned long aux;
  
  
  bowt= 0.0; aux= from;
  do {
    state= lm->states[aux];
    ed= search_edge ( state.edges, word );
    if ( ed != state.edges.end && ed->w == word )
      {
	*prob= bowt + ed->p;
	state= lm->states[ed->to];
	if ( state.null_edges.begin != state.null_edges.end )
	  {
	    *to= state.null_edges.begin->to;
	    (*prob)+= state.null_edges.begin->p;
	  }
	else *to= ed->to;
	return;
      }
    aux= state.bos;
    bowt+= state.bowt;
  } while ( bowt != -HUGE_VAL ); 
  *prob= -HUGE_VAL;
  
} /* end tL_lm_bo_get_trans */


tLProb
tL_lm_get_fp (
	      const tLLM    *lm,
	      unsigned long  state
	      )
{
  
  const tLLMState *states;
  tLProb ret, bowt, abowt;
  
  
  states= lm->states;
  abowt= 0.0;
  for ( ret= states[state].fp;
	ret == -HUGE_VAL && (bowt= states[state].bowt) != -HUGE_VAL;
	state= states[state].bos, ret= states[state].fp )
    abowt+= bowt;
  
  return ret+abowt;
  
} /* end tL_lm_get_fp */


tLLMIterator
tL_lm_get_iterator (
		    const tLLM *lm
		    )
{
  return lm->backoff ?
    make_iterb ( lm ) :
    make_iter_nob ( lm );
} /* end tL_lm_get_iterator */


size_t
tL_lm_nnstates (
		const tLLM *lm
		)
{
  
  size_t n, ret;
  tLLMState *states;
  
  
  ret= 0;
  states= lm->states;
  for ( n= 0; n < lm->NS; ++n )
    if ( states[n].word == ULONG_MAX )
      ++ret;
  
  return ret;
  
} /* end tL_lm_nnstates */


TL_API_FUNC(void)
tL_lm_set_gsf (
	       tLLM          *lm,
	       const tLFloat  gsf
	       )
{
  
  size_t n;
  
  
  for ( n= 0; n < lm->NS; ++n )
    {
      lm->states[n].fp*= gsf;
      if ( lm->backoff )
	lm->states[n].bowt*= gsf;
    }
  for ( n= 0; n < lm->NE; ++n )
    lm->edges[n].p*= gsf;
  lm->_gsf*= gsf;
  lm->_wip*= gsf;
  
} /* end tL_lm_set_gsf */


TL_API_FUNC(void)
tL_lm_set_wip (
	       tLLM         *lm,
	       const tLProb  wip
	       )
{
  
  /* NOTA: Es suma a totes les transicions que ixen d'un estat no
     nul. */
  
  size_t n;
  tLLMEdge *ed;
  
  
  for ( n= 0; n < lm->NS; ++n )
    {
      lm->states[n].fp+= wip;
      for ( ed= lm->states[n].edges.begin;
	    ed != lm->states[n].edges.end;
	    ++ed )
	ed->p+= wip;
    }
  lm->_wip+= wip;
  
} /* end tL_lm_set_wip */


TL_API_FUNC(tLLM *)
tL_lm_new_from_file (
		     gzFile         from,
		     tLLexicon     *lexicon,
		     const tLBool   emptylex,
		     const tLDict  *syms,
		     char         **err
		     )
{
  
  tLLM *lm;
  tLBuffer *buffer;
  
  /* Inicialitza. */
  buffer= tL_buffer_new ();
  
  /* Capçalera. */
  if ( !tL_gline ( from, buffer ) ||
       ( strcmp ( buffer->v, "LM" ) && 
       strcmp ( buffer->v, "BLM" ) ) )
    {
      tL_msgerror ( err, _( "In line %lu, unable to read 'LM' or 'BLM'" ) , 1 );
      goto error;
    } 
  if (  strcmp ( buffer->v, "LM" ) == 0 )
    lm= tL_lm_new_from_file_text ( from, lexicon, emptylex, syms, err );
  else 
    lm= tL_lm_new_from_file_binary ( from, lexicon, emptylex, syms, err );

  tL_buffer_free ( buffer );
  return lm;

 error:
  tL_buffer_free ( buffer );
  return NULL;
  
} /* end tL_lm_new_from_file */

