/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lm_public.h - Part pública de 'lm.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012.
 *
 */

#ifndef __LIBTL_LM_H__
#define __LIBTL_LM_H__

#include <stddef.h>
#include <stdio.h>

#include "common_public.h"
#include "lexicon_public.h"
#include "port.h"
#include "text_public.h"

/** \addtogroup Models
 *  @{
 */

/** \defgroup LM Language model
 */
/** \addtogroup LM
 *  @{
 */

/* DEFINICIÓ. */

typedef struct _tLLMIterator *tLLMIterator;

typedef struct
{
  
  size_t        to;    /* Estat destí. */
  tLProb        p;     /* Probabilitat de transició. */
  unsigned long w;     /* Paraula amb que transitat. Aquesta
			  informació és redundant però ho fique per
			  qüestions d'eficiència durant la
			  búsqueda. */
  
} tLLMEdge;

typedef struct
{
  
  tLLMEdge *begin;
  tLLMEdge *end;
  
} tLLMIter;

typedef struct
{
  
  tLLMIter      edges;         /* Arestes a estats que tenen associat
				   una paraula. */
  tLLMIter      null_edges;    /* Arestes a estats que no tenen
				   associada ninguna paraula. */
  unsigned long  word;          /* Paraula associada a l'estat.  Un
				   valor de ULONG_MAX indica estat
				   sense paraula associada. */
  tLProb         fp;            /* Probabilitat de transitar a l'estat
				   final. */
  size_t         bos;           /* Estat on baixar en cas de
				   backoff. */
  tLProb         bowt;          /* Pes de backoff. */
  
} tLLMState;

/** Language model. This structure is used to store ARPA language
 *  models and Wordnet models, as finite-state automata.
 */
typedef struct
{
  
  tLBool     backoff;    /**< If TRUE, then the stored model has
			    back-off (ARPA). */
  size_t     NS;         /**< The number of states. */
  size_t     NE;         /**< The number of edges. */
  size_t     start;      /* Estat inicial. */
  tLLMState *states;     /* Estats. */
  tLLMEdge  *edges;      /* Emmagatzema les arestes. */
  size_t     NW;         /* Número de paraules del vocabulari. */
  tLFloat    _gsf;       /* Grammar scale factor acumulat. */
  tLProb     _wip;       /* WIP acumulat. */
  
} tLLM;


/* FUNCIONS */

/** Frees memory. Frees the memory allocated for the language model.
 *
 *  @param lm The language model.
 */
TL_API_FUNC(void)
tL_lm_free (
	    tLLM *lm
	    );

/* Imprimeix el model de llenguatge en el fitxer indicat. Es supossa
 * que és el vocabulari per al que es va crear el model de llenguatge
 */
/** Prints the language model. This function writes to the given file
 *  the content of the language model, using the provided word
 *  dictionary. The provided word dictionary is assumed to contain
 *  all needed words, otherwise an unexpected error could
 *  happen. The content is not written using the ARPA format or
 *  Wordnet format; instead, a private text representation is used.
 *
 *  @param lm The language model.
 *  @param to File to which the model will be written.
 *  @param words Dictionary containing the words.
 *  @param binary Flag to indicate if binary output is desired.
 */
TL_API_FUNC(void)
tL_lm_print (
	     const tLLM   *lm,
	     FILE         *to,
	     const tLDict *words,
	     const tLBool  binary
	     );

/* Fixa el 'Grammar Scale Factor'. */
/** Sets the grammar scale factor. This function modifies the given
 *  language model by applying the given grammar scale factor. Note
 *  that if a previous grammar scale factor (gsf0) was applied, the
 *  resulting model will have a grammar scale factor of gsf0*gsf.
 *
 *  @param lm The language model to modify.
 *  @param gsf Grammar scale factor.
 */
TL_API_FUNC(void)
tL_lm_set_gsf (
	       tLLM          *lm,
	       const tLFloat  gsf
	       );

/* Fixa el 'Word Insertion Penalty'. És una log-probabilitat. */
/** Sets the word insertion penalty. This function modifies the given
 *  language model by applying the word insertion penalty. Note that
 *  if a previous word insertion penalty (wip0) was applied, the
 *  resulting model will have a word insertion penalty of wip0+wip.
 *
 *  @param lm The language model to modify.
 *  @param wip Word insertion penalty.
 */
TL_API_FUNC(void)
tL_lm_set_wip (
	       tLLM         *lm,
	       const tLProb  wip
	       );

/* Llig un model de llenguatge d'un fitxer.  Se li ha de passar un
 * diccionari de paraules. Si està buit ha de omplir-lo suposant que
 * cada paraula és una seqüència de caràcters en UTF-8. Si no ho està
 * generarà un tL_error en cas de paraula fora de diccionari. Si SYMS és
 * NULL s'ignoren els símbols, es a dir el lèxic està en un estat
 * inestable (no es pot imprimir, les entrades no tenen sentit, etc.),
 * en cas contrari els símbols que no apareguen al vocabulari de
 * caràcters produeixen un tL_error.
 *
 * NOTA: D'un estat no poden eixir dos arestes que transiten a estats
 * amb la mateixa paraula associada.
 */
/** Reads a language model from a file. This function creates a new
 *  language model from a text description stored in the file
 *  provided. The language model is assumed to be in the same format
 *  as the one generated by tL_lm_print. A tLLexicon is provided to
 *  register the words. If the lexicon is empty, new words are
 *  registered in the lexicon; else, an out-of-vocabulary word is
 *  treated as an error. If a dictionary is provided, new registered
 *  words are considered to be coded in UTF-8 and split into
 *  characters, an out-of-dictionary character being an error.
 *  Otherwise, new words are registered whit an empty sequence of
 *  symbols. Hence, in this case the resulting tLLexicon is a badly
 *  formed tLLexicon; however, it can be used to print the language
 *  model.
 *
 *  @param from File where the text description is stored. 
 *  @param lexicon The lexicon where words are registered.
 *  @param emptylex Specifies if the provided lexicon is empty (new
 *  words must be registered) or not (new words are treated as
 *  errors).
 *  @param syms A character dictionary used to register new words. If
 *  not NULL, out-of-vocabulary characters are treated as errors; if
 *  NULL, characters are ignored.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The language model, or NULL in case of error.
 */
TL_API_FUNC(tLLM *)
tL_lm_new_from_file (
		     gzFile         from,
		     tLLexicon     *lexicon,
		     const tLBool   emptylex,
		     const tLDict  *syms,
		     char         **err
		     );

/** @} */
/** @} */

#endif /* __LIBTL_LM_H__ */
