/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lookahead.c - Implementació de 'lokahead.h'.
 *
 *  Adrià Giménez Pastor, 2012.
 *
 */


#include <assert.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#include "common.h"
#include "error.h"
#include "lookahead.h"
#include "mem.h"




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  unsigned long next;
  unsigned long to;
  
} toaux_edge_t;


typedef struct
{
  
  unsigned long edges;
  unsigned int  counter;
  
} toaux_state_t;


/* Esctructura auxiliar per a otindre l'ordre topològic de l'arbre. */
typedef struct
{
  
  toaux_state_t *states;
  toaux_edge_t  *edges;
  unsigned long *stack;
  unsigned long  N;
  
} toaux_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static toaux_t *
create_toaux (
	      const lexpref_t *lexpref
	      )
{
  
  toaux_t *new;
  unsigned long n, NE;
  const unsigned long *ed;
  lexpref_state_edges_t steds;
  
  
  mt ( new= mem_alloc ( toaux_t, 1 ) );
  mt ( new->states= mem_alloc ( toaux_state_t, lexpref->N ) );
  for ( n= 0; n < lexpref->N; ++n )
    new->states[n].edges= 0;
  mt ( new->edges= mem_alloc ( toaux_edge_t, lexpref->NE+1 ) );
  mt ( new->stack= mem_alloc ( unsigned long, lexpref->N ) );
  new->N= 0;
  NE= 1;
  for ( n= 0; n < lexpref->N; ++n )
    {
      steds= lexpref->state_eds[n];
      new->states[n].counter= (unsigned int) (steds.end-steds.begin);
      if ( new->states[n].counter == 0 )
	new->stack[new->N++]= n;
      for ( ed= steds.begin; ed != steds.end; ++ed )
	{
	  new->edges[NE].to= n;
	  new->edges[NE].next= new->states[*ed].edges;
	  new->states[*ed].edges= NE++;
	}
    }
  
  return new;
  
} /* end create_toaux */


static void
free_toaux (
	    toaux_t *toaux
	    )
{
  
  free ( toaux->states );
  free ( toaux->stack );
  free ( toaux->edges );
  free ( toaux );
  
} /* end free_toaux */


static unsigned long *
create_toporder (
		 const lexpref_t *lexpref
		 )
{
  
  unsigned long *new;
  toaux_t *aux;
  unsigned long s, N, ed;
  
  
  mt ( new= mem_alloc ( unsigned long, lexpref->N ) );
  aux= create_toaux ( lexpref );
  N= 0;
  while ( aux->N != 0 )
    {
      s= aux->stack[--(aux->N)];
      new[N++]= s;
      for ( ed= aux->states[s].edges; ed != 0; ed= aux->edges[ed].next )
	if ( --(aux->states[aux->edges[ed].to].counter) == 0 )
	  aux->stack[aux->N++]= aux->edges[ed].to;
    }
  free_toaux ( aux );
  
  return new;
  
} /* end create_toporder */


/* Torna la grandària de les taules de lookahead. */
static unsigned long
fill_maps (
	   unsigned long       *map_leaves,
	   unsigned long       *map_states,
	   const unsigned long *toporder,
	   const lexpref_t     *lexpref
	   )
{
  
  unsigned int counter, aux;
  unsigned long n, s, N;
  const lexpref_fedge_t *ed;
  lexpref_state_edges_t steds;
  
  
  N= 0;
  for ( n= 0; n < lexpref->N; ++n )
    {
      s= toporder[n];
      steds= lexpref->state_eds[s];
      counter= 0;
      for ( ed= steds.fbegin; ed != steds.fend; ++ed )
	{
	  map_leaves[ed->e]= N++;
	  ++counter;
	}
      aux= (unsigned int) (steds.end-steds.begin);
      if ( (counter+aux) > 1 )
	map_states[s]= N++;
      else if ( aux )
	map_states[s]= map_states[*(steds.begin)];
      else
	map_states[s]= N-1; /* Últim assignat de les fulles. */
    }
  
  return N;
  
} /* end fill_maps */


static void
free_cache (
	    lookahead_cache_t *cache
	    )
{
  
  lookahead_node_t *nodes, *p, *q;
  
  
  nodes= &(cache->nodes);
  p= nodes->next;
  while ( p != nodes )
    {
      q= p;
      p= p->next;
      free ( q->table.to );
      free ( q->table.probs );
      free ( q );
    }
  free ( cache->states );
  free ( cache );
  
} /* end free_cache */


static lookahead_cache_t *
cache_new (
	   const tLLM   *lm,
	   const size_t  capacity
	   )
{
  
  lookahead_cache_t *new;
  unsigned long n;
  
  
  mt ( new= mem_alloc ( lookahead_cache_t, 1 ) );
  mt ( new->states= mem_alloc ( lookahead_node_t *, lm->NS ) );
  for ( n= 0; n < lm->NS; ++n ) new->states[n]= NULL;
  new->nodes.next= new->nodes.prev= &(new->nodes);
  new->N= 0;
  new->capacity= capacity;
  
  return new;
  
} /* end cache_new */


static void
calc_table (
	    const lookahead_t       *lookahead,
	    const lexpref_t         *lexpref,
	    const tLLM              *lm,
	    tLLMIterator             i,
	    const unsigned long      lms,
	    const lookahead_table_t  table
	    )
{
  
  unsigned long n, s;
  tLFloat max, aux;
  const lexpref_fedge_t *fed;
  const tLLMEdge *begin, *end, *edi;
  tLProb *auxlm;
  const unsigned long *map_leaves, *map_states, *ed;
  tLLMState lmstate;
  lexpref_state_edges_t steds;
  
  
  auxlm= lookahead->auxlm;
  map_leaves= lookahead->map_leaves;
  map_states= lookahead->map_states;
  
  /* Obté probabilitats de transició. */
  lm_iterator_set ( i, lms );
  while ( lm_iterator_next ( i, &begin, &end ) )
    for ( edi= begin; edi != end; ++edi, ++n )
      {
	lmstate= lm->states[edi->to];
	/* Supossem que és LM creat a partir d'un model ARPA (Està
	   suavitzat) i que les paraules que no apareixen ja s'han
	   inicialitzat amb -HUGE_VAL al crear auxlm. */
	if ( lmstate.null_edges.begin != lmstate.null_edges.end )
          {
	    table.to[edi->w]= lmstate.null_edges.begin->to;
	    auxlm[edi->w]= edi->p + lmstate.null_edges.begin->p;
	  }
	else
	  {
	    table.to[edi->w]= edi->to;
	    auxlm[edi->w]= edi->p;
	  }
      }
  
  /* Ompli la taula. */
  for ( n= 0; n < lexpref->N; ++n )
    {
      s= lookahead->toporder[n];
      steds= lexpref->state_eds[s];
      max= -HUGE_VAL;
      for ( fed= steds.fbegin; fed != steds.fend; ++fed )
	{
	  table.probs[map_leaves[fed->e]]= aux= fed->p + auxlm[fed->word];
	  if ( aux > max ) max= aux;
	}
      for ( ed= steds.begin; ed != steds.end; ++ed )
	{
	  aux= table.probs[map_states[*ed]];
	  if ( aux > max ) max= aux;
	}
      table.probs[map_states[s]]= max;
    }
  
} /* end calc_table */


static void
unlink_node (
	     lookahead_node_t *node
	     )
{
  
  node->prev->next= node->next;
  node->next->prev= node->prev;
  
} /* end unlink_node */


static unsigned long *
make_map_hists (
		const tLLM         *lm,
		const unsigned int  depth
		)
{
  
  unsigned long *new, *stack, n, state, prev_state;
  unsigned int *depths, current_depth;
  int ssize, sN;
  
  
  assert ( lm->backoff );
  assert ( depth >= 1 );
  
  /* Allocates memory. */
  new= mem_alloc ( unsigned long, lm->NS );
  depths= mem_alloc ( unsigned int, lm->NS );
  stack= mem_alloc ( unsigned long, 1 );
  ssize= 1;
  
  /* Fill values. */
  prev_state= 0; /* For compiler warnings. */
  for ( n= 0; n < lm->NS; ++n ) depths[n]= 0;
  for ( n= 0; n < lm->NS; ++n )
    if ( depths[n] == 0 )
      {
	
	/* Get all back-off states until the first seen state. */
	stack[0]= state= n; sN= 1;
	current_depth= 0;
	while ( lm->states[state].bowt != -HUGE_VAL )
	  {
	    state= lm->states[state].bos;
	    if ( (current_depth= depths[state]) != 0 )
	      {
		prev_state= new[state];
		break;
	      }
	    if ( ssize == sN )
	      {
		ssize*= 2;
		stack= mem_realloc ( unsigned long, stack, ssize );
	      }
	    stack[sN++]= state;
	  }
	
	/* Update depths and new. */
	for ( --sN; sN >= 0; --sN )
	  {
	    state= stack[sN];
	    ++current_depth;
	    depths[state]= current_depth;
	    if ( current_depth <= depth )
	      {
		new[state]= state;
		prev_state= state;
	      }
	    else new[state]= prev_state;
	  }
	
      }
  
  /* Frees memory. */
  free ( stack );
  free ( depths );
  
  return new;
  
} /* end make_map_hists */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_lookahead_change_capacity (
			      lookahead_t  *lookahead,
			      const size_t  ncapacity
			      )
{
  
  lookahead_cache_t *cache;
  lookahead_node_t *nodes, *prev, *p, *q;
  size_t counter;
  
  
  cache= lookahead->cache;
  if ( ncapacity >= cache->N )
    {
      cache->capacity= ncapacity;
      return;
    }
  nodes= &(cache->nodes);
  for ( prev= nodes, p= nodes->next, counter= 0;
	p != nodes && counter < ncapacity;
	prev= p, p= p->next, ++counter );
  while ( p != nodes )
    {
      q= p;
      p= p->next;
      cache->states[q->state]= NULL;
      free ( q->table.to );
      free ( q->table.probs );
      free ( q );
    }
  prev->next= nodes;
  nodes->prev= prev;
  cache->N= cache->capacity= ncapacity;
  
} /* end tL_lookahead_change_capacity */


void
tL_lookahead_free (
		   lookahead_t *lookahead
		   )
{
  
  if ( lookahead->map_hists != NULL )
    free ( lookahead->map_hists );
  free_cache ( lookahead->cache );
  free ( lookahead->auxlm );
  free ( lookahead->map_leaves );
  free ( lookahead->map_states );
  free ( lookahead->toporder );
  free ( lookahead );
  
} /* end tL_lookahead_free */


lookahead_table_t
tL_lookahead_get_table (
			lookahead_t         *lookahead,
			const lexpref_t     *lexpref,
			const tLLM          *lm,
			tLLMIterator         i,
			const unsigned long  lms,
			unsigned long       *fails
			)
{
  
  lookahead_cache_t *cache;
  lookahead_node_t *ret;
  tLBool insert;
  unsigned long state;
  
  
  cache= lookahead->cache;
  
  /* Selecciona node. */
  state= lookahead->map_hists==NULL ? lms : lookahead->map_hists[lms];
  ret= cache->states[state];
  if ( ret != NULL )
    {
      if ( ret->prev == &(cache->nodes) ) insert= FALSE;
      else { unlink_node ( ret ); insert= TRUE; }
    }
  else
    {
      if ( cache->N == cache->capacity )
	{
	  ret= cache->nodes.prev;
	  cache->states[ret->state]= NULL;
	  unlink_node ( ret );
	}
      else
	{
	  mt ( ret= mem_alloc ( lookahead_node_t, 1 ) );
	  mt ( ret->table.to= mem_alloc ( unsigned long, lm->NW ) );
	  mt ( ret->table.probs= mem_alloc ( tLFloat, lookahead->N ) );
	  (cache->N)++;
	}
      calc_table ( lookahead, lexpref, lm, i, state, ret->table );
      cache->states[state]= ret;
      ret->state= state;
      insert= TRUE;
      ++(*fails);
    }
  
  /* Inserta */
  if ( insert )
    {
      ret->next= cache->nodes.next;
      cache->nodes.next->prev= ret;
      ret->prev= &(cache->nodes);
      cache->nodes.next= ret;
    }
  
  return ret->table;
  
} /* end tL_lookahead_get_table */


lookahead_t *
tL_lookahead_new (
		  const lexpref_t    *lexpref,
		  const tLLexicon    *lexicon,
		  const tLLM         *lm,
		  const size_t        capacity,
		  const unsigned int  depth
		  )
{
  
  lookahead_t *new;
  unsigned long n;
  
  
  mt ( new= mem_alloc ( lookahead_t, 1 ) );
  new->toporder= create_toporder ( lexpref );
  mt ( new->map_leaves= mem_alloc ( unsigned long, lexicon->N+1 ) );
  mt ( new->map_states= mem_alloc ( unsigned long, lexpref->N ) );
  new->N= fill_maps ( new->map_leaves, new->map_states,
		      new->toporder, lexpref );
  mt ( new->auxlm= mem_alloc ( tLProb, lexicon->words->N ) );
  /* Açò fa que totes les paraules del lèxic que no apareguen en el LM
     tinguen sempre probabilitat -INF. Vore 'calc_table'. */
  for ( n= 0; n < lexicon->words->N; ++n ) new->auxlm[n]= -HUGE_VAL;
  new->cache= cache_new ( lm, capacity );
  new->map_hists= depth==0 ? NULL : make_map_hists ( lm, depth );
  
  return new;
  
} /* end tL_lookahed_new */
