/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mixtcompmult.c - Implementació de 'mixtcompmult.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012.
 *
 */


#include <assert.h>
#include <float.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "check_is_mixture.h"
#include "amcounts_aux.h"
#include "amodel.h"
#include "amodel/bernoulli.h"
#include "amodel/dgaussian.h"
#include "amodel/mixture.h"
#include "amodel/tiedstates.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "mixtcompmult.h"




/*********/
/* TIPUS */
/*********/

/* Obté dos versions perturbades de a, una la guarda en a i l'altra en
   b. */
typedef void (mu_t) (void               *global,
		     void               *states,
		     const unsigned int  a,
		     const unsigned int  b
		     );

typedef struct
{
  
  const char *name;
  mu_t       *mu;
  
} mu_type_t;


typedef struct
{
  
  unsigned int *v;
  unsigned int  N;
  unsigned int  size;
  
} heap_t;


struct _tLMixtCompMult
{
  
  tLAModel    *amodel;
  tLMixCounts *mcounts;
  tLFloat      min_occ;
  mu_t        *mu;
  heap_t      *heap;
  int          type;
  
};




/***************/
/* FUNCIONS MU */
/***************/
/* NOTA: Si en un futur açò augmenta ja ho trauré d'ací. */

static void
mu_bernoulli (
	      void               *global,
	      void               *states,
	      const unsigned int  a,
	      const unsigned int  b
	      )
{
  
  bernoulli_state_t *sa, *sb;
  unsigned int D, d;
  tLFloat aux, aux2, diff;
  
  
  D= BERNOULLI_GLOBAL(global)->D;
  sa= &(BERNOULLI_STATE(states)[a]);
  sb= &(BERNOULLI_STATE(states)[b]);
  for ( d= 0; d < D; ++d )
    {
      aux= exp ( sa->p[d].p1 );
      aux2= aux - 0.05;
      aux+= 0.05;
      if ( aux > 1.0 )
	{
	  diff= aux - 1.0;
	  aux= 1.0;
	  aux2-= diff;
	}
      else if ( aux2 < 0.0 )
	{
	  diff= -aux2;
	  aux+= diff;
	  aux2= 0.0;
	}
      sa->p[d].p1= log ( aux );
      sa->p[d].p0= log ( 1.0 -aux );
      sb->p[d].p1= log ( aux2 );
      sb->p[d].p0= log ( 1.0 - aux2 );
    }
  tL_Bernoulli_update_hiddendata ( D, sa );
  tL_Bernoulli_update_hiddendata ( D, sb );
  
} /* end mu_bernoulli */


static void
mu_dgaussian (
	      void               *global,
	      void               *states,
	      const unsigned int  a,
	      const unsigned int  b
	      )
{
  
  dgaussian_state_t *sa, *sb;
  unsigned int D, d;
  tLFloat aux, aux2;
  
  
  D= DGAUSSIAN_GLOBAL(global)->D;
  sa= &(DGAUSSIAN_STATE(states)[a]);
  sb= &(DGAUSSIAN_STATE(states)[b]);
  for ( d= 0; d < D; ++d )
    {
      sb->ivar[d]= sa->ivar[d];
      aux= sb->var[d]= sa->var[d];
      aux= sqrt ( aux )*0.2;
      aux2= sa->mu[d];
      sa->mu[d]-= aux;
      sb->mu[d]= aux2 + aux;
    }
  tL_DGaussian_update_logc ( D, sa );
  tL_DGaussian_update_logc ( D, sb );
  
} /* end mu_dgaussian */


static const mu_type_t types[]=
  {
    { "Bernoulli", mu_bernoulli },
    { "DGaussian", mu_dgaussian },
    { NULL, NULL }
  };




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static heap_t *
make_heap (void)
{
  
  heap_t *new;
  
  
  mt ( new= mem_alloc ( heap_t, 1 ) );
  mt ( new->v= mem_alloc ( unsigned int, 1 ) );
  new->N= 0;
  new->size= 1;
  
  return new;
  
} /* end make_heap */


static void
free_heap (
	   heap_t *heap
	   )
{
  
  free ( heap->v );
  free ( heap );
  
} /* end free_heap */


static void
prepare_heap (
	      heap_t             *heap,
	      const unsigned int  I
	      )
{
  
  heap->N= 0;
  if ( I+1 <= heap->size ) return;
  mt ( heap->v= mem_realloc ( unsigned int, heap->v, I+1 ) );
  heap->size= I+1;
  
} /* end prepare_heap */


static void
heap_push (
	   heap_t             *heap,
	   const unsigned int  i,
	   const tLProb       *probs
	   )
{
  
  unsigned int p, p2, *v;
  
  
  v= heap->v;
  for ( p= ++heap->N;
	p != 1 && probs[i] > probs[v[(p2= p>>1)]];
	p= p2 )
    v[p]= v[p2];
  v[p]= i;
  
} /* end heap_push */


static unsigned int
heap_pop (
	  heap_t       *heap,
	  const tLProb *probs
	  )
{
  
  unsigned int ret, N, p, p2, q, *v;
  
  
  v= heap->v;
  ret= v[1];
  N= heap->N--;
  for ( p= 1; (p2= (p<<1)) < N; p= p2 )
    {
      if ( (q=(p2|1)) < N && probs[v[q]] > probs[v[p2]] ) p2= q;
      if ( probs[v[N]] < probs[v[p2]] ) v[p]= v[p2];
      else break;
    }
  v[p]= v[N];
  
  return ret;
  
} /* end heap_pop */


/* Torna NULL si no el troba. */
static mu_t *
get_mu (
	const char *name
	)
{
  
  const mu_type_t *p;
  
  
  for ( p= &(types[0]); p->name != NULL; ++p )
    if ( !strcmp ( p->name, name ) )
      return p->mu;
  
  return NULL;
  
} /* end get_mu */


static void
split_components (
		  tLMixtCompMult    *mcmu,
		  mixture_state_t   *state,
		  tLFloat          **ptr_counts,
		  mixture_global_t  *global
		  )
{
  
  unsigned int i, N;
  tLProb p;
  tLFloat log_2, *counts, occ;
  
  
  /* Prepara. */
  mt ( state->p= mem_realloc ( tLProb, state->p, state->I*2 ) );
  state->members=
    global->meths->realloc_states ( global->global, state->members,
                                    state->I, state->I*2 );
  counts= NULL;
  if ( ptr_counts != NULL )
    mt ( *ptr_counts= counts=
	 mem_realloc ( tLFloat, *ptr_counts, state->I*2 ) );
  
  /* Crea noves components. */
  log_2= log ( 2 );
  N= state->I;
  for ( i= 0; i < state->I; ++i )
    {
      if ( counts != NULL )
	{
	  occ= counts[i]/2.0;
	  if ( occ < mcmu->min_occ ) continue;
	  counts[N]= counts[i]= occ;
	}
      p= state->p[i] - log_2;
      state->p[i]= p;
      state->p[N]= p;
      mcmu->mu ( global->global, state->members, i, N );
      ++N;
    }
  if ( N != state->I*2 ) /* Sols en cas de counts!=NULL */
    {
      mt ( state->p= mem_realloc ( tLProb, state->p, N ) );
      /* Encara que había reservat més capacitat, sols m'interen els
	 primers state->I estats. */
      state->members=
	global->meths->realloc_states ( global->global, state->members,	N, N );
       mt ( *ptr_counts= mem_realloc ( tLFloat, *ptr_counts, N ) );
    }
  state->I= N;
  if ( state->I > global->max_I ) global->max_I= state->I;
  
} /* end split_components */


static void
mu_components (
	       tLMixtCompMult    *mcmu,
	       mixture_state_t   *state,
	       tLFloat          **ptr_counts,
	       mixture_global_t  *global,
	       unsigned int       newI
	       )
{
  
  unsigned int i, aux;
  tLProb p;
  tLFloat log_2, *counts, occ;
  
  
  if ( state->I >= newI ) return;
  
  /* Prepara. */
  mt ( state->p= mem_realloc ( tLProb, state->p, newI ) );
  state->members=
    global->meths->realloc_states ( global->global, state->members,
                                    state->I, newI );
  prepare_heap ( mcmu->heap, newI );
  counts= NULL;
  if ( ptr_counts != NULL )
    {
      mt ( *ptr_counts= counts= mem_realloc ( tLFloat, *ptr_counts, newI ) );
      for ( i= 0; i < state->I; ++i )
	heap_push ( mcmu->heap, i, counts );
    }
  else
    for ( i= 0; i < state->I; ++i )
      heap_push ( mcmu->heap, i, state->p );
  
  /* Crea noves components. */
  log_2= log ( 2 );
  for ( i= state->I; i < newI; ++i )
    {
      if ( counts != NULL )
	{
	  aux= heap_pop ( mcmu->heap, counts );
	  occ= counts[aux]/2.0;
	  if ( occ < mcmu->min_occ ) break;
	  counts[aux]= counts[i]= occ;
	}
      else aux= heap_pop ( mcmu->heap, state->p );
      p= state->p[aux] - log_2;
      state->p[aux]= p;
      state->p[i]= p;
      mcmu->mu ( global->global, state->members, aux, i );
      if ( counts != NULL )
	{
	  heap_push ( mcmu->heap, aux, counts );
	  heap_push ( mcmu->heap, i, counts );
	}
      else
	{
	  heap_push ( mcmu->heap, aux, state->p );
	  heap_push ( mcmu->heap, i, state->p );
	}
    }
  state->I= i;
  if ( state->I > global->max_I ) global->max_I= state->I;
  if ( state->I != newI ) /* Sols en cas de counts!=NULL */
    {
      mt ( state->p= mem_realloc ( tLProb, state->p, state->I ) );
      /* Encara que había reservat més capacitat, sols m'interen els
	 primers state->I estats. */
      state->members=
	global->meths->realloc_states ( global->global, state->members,
					state->I, state->I );
       mt ( *ptr_counts= mem_realloc ( tLFloat, *ptr_counts, state->I ) );
    }
  
} /* end mu_components */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_mixtcompmult_free (
		      tLMixtCompMult *mcmu
		      )
{
  
  free_heap ( mcmu->heap );
  free ( mcmu );
  
} /* end tL_mixtcompmult_free */


TL_API_FUNC(tLMixtCompMult *)
tL_mixtcompmult_new (
		     tLAModel     *amodel,
		     tLMixCounts  *mcounts,
		     char        **err
		     )
{
  
  tLMixtCompMult *new;
  mixture_global_t *global;
  mu_t *mu;
  int type;
  
  
  /* Comprovació. */
  if ( !tL_check_is_mixture ( amodel, &type, err ) ) return NULL;
  if ( mcounts != NULL ) assert ( type == mcounts->type );
  if ( type == MIXTURE ) global= MIXTURE_GLOBAL(amodel->global);
  else global= MIXTURE_GLOBAL(TIEDSTATES_GLOBAL(amodel->global)->global);
  mu= get_mu ( global->type );
  if ( mu == NULL || global->meths->realloc_states == NULL )
    {
      tL_msgerror ( err, _( "%s mixtures are not supported" ) , global->type );
      return NULL;
    }
  
  /* Crea. */
  mt ( new= mem_alloc ( tLMixtCompMult, 1 ) );
  new->amodel= amodel;
  new->mcounts= mcounts;
  new->min_occ= 10.0;
  new->mu= mu;
  new->heap= make_heap ();
  new->type= type;
  
  return new;
  
} /* end tL_mixtcompmult_new */


TL_API_FUNC(void)
tL_mu_mixture_components (
			  tLMixtCompMult *mcmu,
			  unsigned int    c,
			  unsigned int    q,
			  unsigned int    newI
			  )
{
  
  tLAModelHMM cchar;
  mixture_state_t *state;
  mixture_global_t *global;
  tLFloat **counts_cq;
  
  
  assert ( mcmu->type == MIXTURE );
  assert ( c < mcmu->amodel->N );
  cchar= mcmu->amodel->hmms[c];
  assert ( q < cchar.Q );
  state= &(MIXTURE_STATE(cchar.states)[q]);
  global= MIXTURE_GLOBAL(mcmu->amodel->global);
  counts_cq= (mcmu->mcounts!=NULL) ? &(mcmu->mcounts->counts[c][q]) : NULL;
  if ( newI == 0 ) split_components ( mcmu, state, counts_cq, global );
  else             mu_components ( mcmu, state, counts_cq, global, newI );
  
} /* end tL_mu_mixture_components */


TL_API_FUNC(int)
tL_mu_tmixture_components (
			   tLMixtCompMult  *mcmu,
			   unsigned int     newI,
			   const char      *name,
			   char           **err
			   )
{
  
  tiedstates_global_t *global0;
  mixture_state_t *states;
  mixture_global_t *global;
  unsigned int n;
  tLFloat **counts_cq;
  
  
  assert ( mcmu->type == TIEDSTATES );
  global0= TIEDSTATES_GLOBAL(mcmu->amodel->global);
  states= MIXTURE_STATE(global0->states);
  global= MIXTURE_GLOBAL(global0->global);
  if ( name == NULL )
    {
      for ( n= 0; n < global0->N; ++n )
	{
	  counts_cq= (mcmu->mcounts!=NULL) ?
	    &(mcmu->mcounts->counts[0][n]) : NULL;
	  if ( newI == 0 )
	    split_components ( mcmu, &(states[n]), counts_cq, global );
	  else
	    mu_components ( mcmu, &(states[n]), counts_cq, global, newI );
	}
    }
  else
    {
      n= tL_dict_find ( global0->names, name );
      if ( n == global0->names->N )
	{
	  tL_msgerror ( err, _( "Cannot find the tied state '%s'"
			" in the acoustic model" ) , name );
	  return -1;
	}
      counts_cq= (mcmu->mcounts!=NULL) ?
	&(mcmu->mcounts->counts[0][n]) : NULL;
      if ( newI == 0 )
	split_components ( mcmu, &(states[n]), counts_cq, global );
      else
	mu_components ( mcmu, &(states[n]), counts_cq, global, newI );
    }
  
  return 0;
  
} /* end tL_mu_tmixture_components */


TL_API_FUNC(void)
tL_mu_tmixture_components_state (
				 tLMixtCompMult *mcmu,
				 const char     *state,
				 unsigned int    newI
				 )
{
  
  tiedstates_global_t *global0;
  mixture_state_t *states;
  mixture_global_t *global;
  unsigned int id;
  tLFloat **counts_cq;
  
  
  assert ( mcmu->type == TIEDSTATES );
  global0= TIEDSTATES_GLOBAL(mcmu->amodel->global);
  states= MIXTURE_STATE(global0->states);
  global= MIXTURE_GLOBAL(global0->global);
  id= tL_dict_find ( global0->names, state );
  if ( id == global0->names->N )
    tL_warning ( _( "There is no state with the name '%s'" ) , state );
  else
    {
      counts_cq= (mcmu->mcounts!=NULL) ? &(mcmu->mcounts->counts[0][id]) : NULL;
      mu_components ( mcmu, &(states[id]), counts_cq, global, newI );
    }
  
} /* end tL_mu_tmixture_components_state */
