/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mixtcompmult.h - Multiplicador de components de mixtura.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */

#ifndef __LIBTL_MIXTCOMPMULT_H__
#define __LIBTL_MIXTCOMPMULT_H__

#include <stddef.h>
#include <stdlib.h>

#include "amcounts_public.h"
#include "amodel_public.h"
#include "port.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup MU Multiplication of mixture components
 */
/** \addtogroup MU
 *  @{
 */

#ifndef __TL_PRV
/** Mixture component multiplier. This structure is used to multiply
 *  the mixture components of a given acoustic model.
 */
struct _tLMixtCompMult
{
  
  tLAModel    *amodel;     /**< Reference to the acoustic model to be
			      modified by the multiplier. */
  tLMixCounts *mcounts;    /**< Reference to the mixture coefficient
			      counts to be used by the multiplier. */
  tLFloat      min_occ;    /**< Minimum occupancy for each mixture
			      component. This value can be modified at
			      any time. Default value 10. */
  
};
#endif

/** Mixture component multiplier. This structure is used to multiply
 *  the mixture components of an acoustic model.
 */
typedef struct _tLMixtCompMult tLMixtCompMult;

/** Frees memory. Frees the memory allocated for the mixture component
 *  multiplier.
 *
 *  @param mcmu The mixture component multiplier.
 */
TL_API_FUNC(void)
tL_mixtcompmult_free (
		      tLMixtCompMult *mcmu
		      );

/** Gets a tLMixtCompMult. This function creates a mixture component
 *  multiplier for a given acoustic model. Currently supported types
 *  are: Bernoulli mixtures and DGaussian mixtures.
 *
 *  @param amodel The acoustic model to modify.
 *  @param mcounts The mixture coefficient counts of the model, or
 *  NULL. If the counts are provided, then the 'min_occ' parameter is
 *  used as stop criterion, guarantying a minimum occupancy in each
 *  mixture component.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @return The mixture component multiplier, or NULL in case of error.
 */
TL_API_FUNC(tLMixtCompMult *)
tL_mixtcompmult_new (
		     tLAModel     *amodel,
		     tLMixCounts  *mcounts,
		     char        **err
		     );

/** Multiplies mixture components. This function multiplies the
 *  mixture components of the specified state. New components are
 *  created until reaching the desired number, or minimum occupancy
 *  value is reached (if mixture coefficient counts are provided). If
 *  the desired number is less than the current number of components,
 *  then nothing happens. Each new component is created by removing
 *  the component with the highest mixture coefficient (or count) and
 *  adding two new components which are perturbations of the removed
 *  component with half of the mixture coefficient probability. If
 *  mixture counts are provided, then they are also updated. Bernoulli
 *  mixture components are perturbed as
 *  \f$\vec{p}\pm\vec{0.05}\f$. DGaussian mixture components are
 *  perturbed as \f$\vec{\mu}\pm 0.2*\sqrt{\sigma^2}\f$. If newI is 0,
 *  then mixture coefficients are ignored and all mixture
 *  components of the specified state are split (if the minimum
 *  occupancy allows for it).
 *
 *  @param mcmu The mixture component multiplier.
 *  @param c HMM to be modified.
 *  @param q State to be modified.
 *  @param newI The desired number of mixture components.
 */
TL_API_FUNC(void)
tL_mu_mixture_components (
			  tLMixtCompMult *mcmu,
			  unsigned int    c,
			  unsigned int    q,
			  unsigned int    newI
			  );

/** Multiplies the mixture components of a TiedState model. This
 *  function multiplies all the mixture components of the related
 *  TiedStates-Mixture model. If the name of a tied state is specified
 *  then only the mixture components of that state are
 *  multiplied. Components are multiplied as in
 *  tL_mu_mixture_components.
 *
 *  @param mcmu The mixture component multiplier.
 *  @param newI The desired number of mixture components in each
 *  state.
 *  @param name The name of the tied state for which mixture
 *  components are to be multiplied, or NULL to multiply all mixture
 *  components.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return -1 in case of error, 0 otherwise.
 */
TL_API_FUNC(int)
tL_mu_tmixture_components (
			   tLMixtCompMult  *mcmu,
			   unsigned int     newI,
			   const char      *name,
			   char           **err
			   );

/** Multiplies the mixture components of a tied state. This function
 *  multiplies the mixture components of the specified tied
 *  state. Components are multiplied as in tL_mu_mixture_components.
 *
 *  @param mcmu The mixture component multiplier.
 *  @param state Name of the tied state.
 *  @param newI The desired number of mixture components.
 */
TL_API_FUNC(void)
tL_mu_tmixture_components_state (
				 tLMixtCompMult *mcmu,
				 const char     *state,
				 unsigned int    newI
				 );

/** @} */
/** @} */

#endif /* __LIBTL_MIXTCOMPMULT_H__ */
