/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mllr.c - Implementació de 'mllr.h'.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */


#include <assert.h>
#include <cblas.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "logamcounts.h"
#include "amcounts/dgaussian.h"
#include "amcounts/logmixture.h"
#include "amcounts/logtiedstates.h"
#include "cmllr_alg.h"
#include "amodel/dgaussian.h"
#include "amodel/mixture.h"
#include "amodel/tiedstates.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "mllr.h"
#include "mllr_alg.h"




/**********/
/* MACROS */
/**********/

#define DGAUSSIAN_COUNT(ptr) ((amcount_dgaussian_state_t *) (ptr))
#define MIXTURE_COUNT(ptr) ((logamcount_mixture_state_t *) (ptr))



/*********/
/* TIPUS */
/*********/

/* Transformació W. */
typedef struct
{
  
  double **W;
  double **A;
  double  *b;
  
} W_t;


/* Per a saber cada transformació a que estats està associat. */
typedef struct
{
  
  unsigned int c;   /* Si és tied serà 0. */
  unsigned int q;
  
} map_node_t;


/* Els estats de cada transformació i els compters i paràmetres
   necessaris pera a adaptar. */
typedef struct
{
  
  map_node_t               *nodes;
  unsigned int              N;
  cmllr_alg_paramscounts_t  pcs;
  
} map_t;


struct _tLMLLR
{
  
  const tLAModel         *amodel;
  enum { TIED, NORMAL }   type;
  unsigned int            D;
  const tLFloat          *smooth;
  unsigned int            N;         /* Número de transformacions. */
  map_t                  *map;       /* Els estats relacionats amb
                                        cada transformació. */
  unsigned int          **imap;      /* |C|*|Q|. Mapa invers. En cas
                                         de Tied c és 0.  */
  W_t                    *Ws;        /* |N+1|. L'últim es gasta com
					auxiliar. */
  int                     mlevel;    /* Número de nivells Mixture. */
  mllr_alg_data_t         data;
  double                 *aux_v;
  
};




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* Els nodes tenen que estar inicialitzats. */
static void
alloc_pcs (
           map_t          *map,
           const tLAModel *amodel,
           const int       type,
           const int       mlevel
           )
{
  
  unsigned int n;
  map_node_t node;
  mixture_state_t *state;
  tiedstates_global_t *global;
  
  
  /* Conta estats. */
  map->pcs.I= 0;
  if ( type == NORMAL )
    for ( n= 0; n < map->N; ++n )
      {
        node= map->nodes[n];
        if ( mlevel == 1 )
          {
            state= &(MIXTURE_STATE(amodel->hmms[node.c].states)[node.q]);
            map->pcs.I+= state->I;
          }
        else ++(map->pcs.I);
      }
  else /* TIED */
    {
      global= TIEDSTATES_GLOBAL(amodel->global);
      for ( n= 0; n < map->N; ++n )
        {
          node= map->nodes[n];
          if ( mlevel == 1 )
            {
              state= &(MIXTURE_STATE(global->states)[node.q]);
              map->pcs.I+= state->I;
            }
          else ++(map->pcs.I);
        }
    }
  
  /* Reserva memòria. */
  mt ( map->pcs.params= mem_alloc ( dgaussian_state_t *, map->pcs.I ) );
  mt ( map->pcs.counts= mem_alloc ( amcount_dgaussian_state_t *, map->pcs.I ) );
  
} /* end alloc_pcs */


/* Els nodes tenen que estar inicialitzats. */
static void
pcs_set_counts (
                map_t               *map,
                const tLAModel      *amodel,
		const tLAModel      *amodel2,
                const tLAMlogCounts *counts,
                const int            type,
                const int            mlevel
                )
{
  
  unsigned int n, I, i;
  map_node_t node;
  mixture_state_t *state, *state2;
  logamcount_mixture_state_t *cstate;
  tiedstates_global_t *global, *global2;
  logamcount_tiedstates_global_t *cglobal;
  
  
  /* Enllaça compters. */
  I= 0;
  if ( type == NORMAL )
    for ( n= 0; n < map->N; ++n )
      {
        node= map->nodes[n];
        if ( mlevel == 1 )
          {
            state= &(MIXTURE_STATE(amodel->hmms[node.c].states)[node.q]);
	    state2= &(MIXTURE_STATE(amodel2->hmms[node.c].states)[node.q]);
            cstate= &(MIXTURE_COUNT(counts->hmms[node.c].states)[node.q]);
            for ( i= 0; i < state->I; ++i, ++I )
	      {
		map->pcs.params[I]= &(DGAUSSIAN_STATE(state2->members)[i]);
		map->pcs.counts[I]= &(DGAUSSIAN_COUNT(cstate->members)[i]);
	      }
          }
        else
	  {
	    map->pcs.params[I]=
	      &(DGAUSSIAN_STATE(amodel2->hmms[node.c].states)[node.q]);
	    map->pcs.counts[I]=
	      &(DGAUSSIAN_COUNT(counts->hmms[node.c].states)[node.q]);
	    ++I;
	  }
      }
  else /* TIED */
    {
      global= TIEDSTATES_GLOBAL(amodel->global);
      global2= TIEDSTATES_GLOBAL(amodel2->global);
      cglobal= (logamcount_tiedstates_global_t *) counts->global;
      for ( n= 0; n < map->N; ++n )
        {
          node= map->nodes[n];
          if ( mlevel == 1 )
            {
              state= &(MIXTURE_STATE(global->states)[node.q]);
	      state2= &(MIXTURE_STATE(global2->states)[node.q]);
              cstate= &(MIXTURE_COUNT(cglobal->states)[node.q]);
              for ( i= 0; i < state->I; ++i, ++I )
		{
		  map->pcs.params[I]= &(DGAUSSIAN_STATE(state2->members)[i]);
		  map->pcs.counts[I]= &(DGAUSSIAN_COUNT(cstate->members)[i]);
		}
            }
          else
	    {
	      map->pcs.params[I]= &(DGAUSSIAN_STATE(global2->states)[node.q]);
	      map->pcs.counts[I]= &(DGAUSSIAN_COUNT(cglobal->states)[node.q]);
	      ++I;
	    }
        }
    }
  
} /* end pcs_set_counts */


static int
fill_imap_and_alloc_map_normal (
				tLMLLR              *mllr,
				tLAModel            *amodel,
				const tLMLLRrc      *rcs,
				const unsigned int   length_rcs,
				char               **err
				)
{
  
  unsigned int n, N, *aux, c, q;
  tLMLLRrc rc;
  map_node_t *nodes, node;
  
  
  /* Fill imap. */
  N= 1;
  for ( n= 0; n < length_rcs; ++n )
    {
      rc= rcs[n];
      if ( rc.c >= amodel->N )
	{
	  tL_msgerror ( err, _( "symbol %u is out of range" ) , rc.c );
	  return -1;
	}
      if ( rc.q >= amodel->hmms[rc.c].Q )
	{
	  tL_msgerror ( err, _( "in symbol %u, state %u is out of range" ) , rc.q );
	  return -1;
	}
      if ( (rc.rc+1) > N ) N= rc.rc+1;
      mllr->imap[rc.c][rc.q]= rc.rc;
    }
  
  /* Crea map. */
  mt ( aux= mem_calloc ( unsigned int, N ) );
  for ( c= 0; c < amodel->N; ++c )
    for ( q= 0; q < amodel->hmms[c].Q; ++q )
      ++aux[mllr->imap[c][q]];
  mt ( mllr->map= mem_alloc ( map_t, N ) );
  mt ( nodes= mem_alloc ( map_node_t, amodel->total_Q ) );
  for ( n= 0; n < N; nodes+= aux[n], ++n )
    {
      mllr->map[n].N= 0;
      mllr->map[n].nodes= nodes;
    }
  for ( c= 0; c < amodel->N; ++c )
    for ( q= 0; q < amodel->hmms[c].Q; ++q )
      {
	node.c= c;
	node.q= q;
	n= mllr->imap[c][q];
	mllr->map[n].nodes[mllr->map[n].N++]= node;
      }
  free ( aux );
  
  mllr->N= N;
  
  return 0;
  
} /* end fill_imap_and_alloc_map_normal */


static int
fill_imap_and_alloc_map_tiedstates (
				    tLMLLR              *mllr,
				    tLAModel            *amodel,
				    const tLMLLRrc      *rcs,
				    const unsigned int   length_rcs,
				    char               **err
				    )
{
  
  unsigned int n, N, *aux, Q, q;
  tLMLLRrc rc;
  map_node_t *nodes, node;
  
  
  /* Fill imap. */
  Q= TIEDSTATES_GLOBAL(amodel->global)->N;
  N= 1;
  for ( n= 0; n < length_rcs; ++n )
    {
      rc= rcs[n];
      if ( rc.q >= Q )
	{
	  tL_msgerror ( err, _( "state %u is out of range" ) , rc.q );
	  return -1;
	}
      if ( (rc.rc+1) > N ) N= rc.rc+1;
      mllr->imap[0][rc.q]= rc.rc;
    }
  
  /* Crea map. */
  mt ( aux= mem_calloc ( unsigned int, N ) );
  for ( q= 0; q < Q; ++q )
    ++aux[mllr->imap[0][q]];
  mt ( mllr->map= mem_alloc ( map_t, N ) );
  mt ( nodes= mem_alloc ( map_node_t, Q ) );
  for ( n= 0; n < N; nodes+= aux[n], ++n )
    {
      mllr->map[n].N= 0;
      mllr->map[n].nodes= nodes;
    }
  for ( q= 0; q < Q; ++q )
      {
	node.c= 0;
	node.q= q;
	n= mllr->imap[0][q];
	mllr->map[n].nodes[mllr->map[n].N++]= node;
      }
  free ( aux );
  
  mllr->N= N;
  
  return 0;
  
} /* end fill_imap_and_alloc_map_tiedstates */


/* Torna 0 si tot ha anat bé, -1 en cas contrari. */
static int
check_type_and_set_values (
                           tLMLLR              *mllr,
                           tLAModel            *amodel,
			   const tLMLLRrc      *rcs,
			   const unsigned int   length_rcs,
                           char               **err
                           )
{
  
  void *global;
  const char *type;
  unsigned int *mem, n;
  
  
  mllr->map= NULL;
  if ( !strcmp ( "TiedStates", amodel->type ) )
    {
      global= TIEDSTATES_GLOBAL(amodel->global)->global;
      type= TIEDSTATES_GLOBAL(amodel->global)->type;
      mllr->type= TIED;
      mt ( mllr->imap= mem_alloc ( unsigned int *, 1 ) );
      mt ( mllr->imap[0]= mem_calloc ( unsigned int,
				       TIEDSTATES_GLOBAL(amodel->global)->N ) );
      if ( fill_imap_and_alloc_map_tiedstates ( mllr, amodel, rcs,
						(rcs==NULL)?0:length_rcs,
						err ) == -1 )
	goto error;
    }
  else
    {
      global= amodel->global;
      type= amodel->type;
      mllr->type= NORMAL;
      mt ( mllr->imap= mem_alloc ( unsigned int *, amodel->N ) );
      mt ( mem= mem_calloc ( unsigned int, amodel->total_Q ) );
      for ( n= 0; n < amodel->N; ++n )
	{
          mllr->imap[n]= mem;
          mem+= amodel->hmms[n].Q;
        }
      if ( fill_imap_and_alloc_map_normal ( mllr, amodel, rcs,
					    (rcs==NULL)?0:length_rcs,
					    err ) == -1 )
	goto error;
    }
  
  mllr->mlevel= 0;
  while ( !strcmp ( "Mixture", type ) )
    {
      type= MIXTURE_GLOBAL(global)->type;
      global= MIXTURE_GLOBAL(global)->global;
      ++(mllr->mlevel);
    }
  /* Açò de moment no es possible per que a l'algorisme per culpa de
     'tL_mllr_update', però si fora possible es podria canviar. */
  if ( mllr->mlevel > 1 )
    {
      tL_msgerror ( err, _( "nested Mixtures are not supported" ) );
      goto error;
    }
  if ( strcmp ( "DGaussian", type ) )
    {
      if ( mllr->type == TIED && !strcmp ( "TiedStates", type ) )
        tL_msgerror ( err, _( "nested TiedStates are not supported" ) );
      else tL_msgerror ( err, _( "type %s is not supported" ) , type );
      goto error;
    }
  mllr->D= DGAUSSIAN_GLOBAL(global)->D;
  mllr->smooth= DGAUSSIAN_GLOBAL(global)->smooth;
  
  /* Reserva pcs. */
  for ( n= 0; n < mllr->N; ++n )
    if ( mllr->map[n].N != 0 )
      alloc_pcs ( &(mllr->map[n]), amodel, mllr->type, mllr->mlevel );
  
  return 0;
  
 error:
  free ( mllr->imap[0] );
  free ( mllr->imap );
  if ( mllr->map != NULL )
    {
      free ( mllr->map[0].nodes );
      free ( mllr->map );
    }
  return -1;
  
} /* end check_type_and_set_values */


static void
free_Ws (
         W_t            *Ws,
         const unsigned  N
         )
{
  
  unsigned int n;
  
  
  for ( n= 0; n < N; ++n )
    {
      free ( Ws[n].W[0] );
      free ( Ws[n].W );
      free ( Ws[n].A );
    }
  free ( Ws );
  
} /* end free_Ws */


static W_t *
alloc_Ws (
          const tLAModel     *amodel,
          const int           type,
          const unsigned int  D,
          const unsigned int  N
          )
{
  
  W_t *new;
  unsigned int n, d;
  double *aux;
  
  
  mt ( new= mem_alloc ( W_t, N ) );
  for ( n= 0; n < N; ++n )
    {
      mt ( new[n].W= mem_alloc ( double *, D+1 ) );
      mt ( new[n].A= mem_alloc ( double *, D ) );
      mt ( aux= mem_alloc ( double, (D+1)*D ) );
      new[n].W[0]= new[n].b= aux; aux+= D;
      for ( d= 0; d < D; ++d, aux+= D )
        new[n].W[d+1]= new[n].A[d]= aux;
    }
  
  return new;
  
} /* end alloc_Ws */


static void
adapt_dgaussian (
		 double             *aux_v,
                 const W_t           W,
                 dgaussian_state_t  *params,
                 const unsigned int  D,
                 const tLFloat      *smooth
                 )
{
  
  /* NOTA: A està organitzat en columnes. */
  
  /* MU */
  cblas_dcopy ( D, params->mu, 1, aux_v, 1 );
  cblas_dcopy ( D, W.b, 1, params->mu, 1 );
  cblas_dgemv ( CblasColMajor, CblasNoTrans, D, D, 1.0,
                W.A[0], D, aux_v, 1, 1.0, params->mu, 1 );
  
} /* adapt_dgaussian */


static void
adapt_state (
	     double             *aux_v,
             void               *states,
             const unsigned int  q,
             const W_t           W,
             const unsigned int  D,
             const tLFloat      *smooth,
             const int           level
             )
{
  
  mixture_state_t state;
  unsigned int i;
  
  
  if ( level == 0 )
    adapt_dgaussian ( aux_v, W, &(DGAUSSIAN_STATE(states)[q]), D, smooth );
  else
    {
      state= MIXTURE_STATE(states)[q];
      for ( i= 0; i < state.I; ++i )
        adapt_state ( aux_v, state.members, i, W, D, smooth, level-1 );
    }
  
} /* end adapt_state */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_mllr_free (
	      tLMLLR *mllr
	      )
{

  unsigned int n;
  
  
  tL_mllr_alg_data_free ( &(mllr->data) );
  for ( n= 0; n < mllr->N; ++n )
    if ( mllr->map[n].N != 0 )
      {
	free ( mllr->map[n].pcs.params );
	free ( mllr->map[n].pcs.counts );
      }
  free ( mllr->map[0].nodes );
  free ( mllr->aux_v );
  free ( mllr->map );
  free ( mllr->imap[0] );
  free ( mllr->imap );
  free_Ws ( mllr->Ws, mllr->N );
  free ( mllr );
  
} /* end tL_mllr_free */


TL_API_FUNC(tLAModel *)
tL_mllr_get_adapted_model (
			   tLMLLR   *mllr,
			   tLAModel *amodel
			   )
{
  
  unsigned int n, q;
  tLAModelHMM cchar;
  tiedstates_global_t *global;
  
  
  if ( amodel == NULL ) amodel= tL_amodel_copy ( mllr->amodel );
  if ( mllr->type == NORMAL )
    for ( n= 0; n < mllr->amodel->N; ++n )
      {
        cchar= amodel->hmms[n];
        for ( q= 0; q < cchar.Q; ++q )
          adapt_state ( mllr->aux_v, cchar.states, q,
                        mllr->Ws[mllr->imap[n][q]],
                        mllr->D, mllr->smooth, mllr->mlevel );
      }
  else /* TIED */
    {
      global= TIEDSTATES_GLOBAL(amodel->global);
      for ( n= 0; n < global->N; ++n )
        adapt_state ( mllr->aux_v, global->states, n,
                      mllr->Ws[mllr->imap[0][n]],
                      mllr->D, mllr->smooth, mllr->mlevel );
    }
  
  return amodel;
  
} /* end tL_mllr_get_adapted_model */


TL_API_FUNC(void)
tL_mllr_init (
	      tLMLLR *mllr
	      )
{
  
  unsigned int n, d, d2;
  W_t *ms;
  double **A, *A_d, *b;
  
  
  ms= mllr->Ws;
  for ( n= 0; n < mllr->N; ++n )
    {
      A= ms[n].A;
      for ( d= 0; d < mllr->D; ++d )
        {
          A_d= A[d];
          for ( d2= 0; d2 < mllr->D; ++d2 )
            A_d[d2]= (d==d2)?1.0:0.0;
        }
      b= ms[n].b;
      for ( d= 0; d < mllr->D; ++d ) b[d]= 0.0;
    }
  
} /* end tL_mllr_init */


TL_API_FUNC(tLMLLR *)
tL_mllr_new (
	     tLAModel            *amodel,
	     const tLMLLRrc      *rcs,
	     const unsigned int   length_rcs,
	     char               **err
	     )
{
  
  tLMLLR *new;
  
  
  mt ( new= mem_alloc ( tLMLLR, 1 ) );
  new->amodel= amodel;
  if ( check_type_and_set_values ( new, amodel, rcs, length_rcs, err ) != 0 )
    {
      free ( new );
      return NULL;
    }
  new->Ws= alloc_Ws ( amodel, new->type, new->D, new->N );
  new->data.G= NULL;
  tL_mllr_init ( new );
  mt ( new->aux_v= mem_alloc ( double, new->D ) );
  
  return new;
  
} /* end tL_mllr_new */


TL_API_FUNC(void)
tL_mllr_update_matrices (
			 tLMLLR              *mllr,
			 const tLAMlogCounts *counts,
			 const tLAModel      *amodel
			 )
{
  
  unsigned int n;
  void *global;
  char *err;
  int ret;
  
  
  if ( mllr->data.G == NULL )
    tL_mllr_alg_data_alloc ( &(mllr->data), mllr->D );
  global= (mllr->type==NORMAL) ?
    mllr->amodel->global :
    TIEDSTATES_GLOBAL(mllr->amodel->global)->global;
  if ( mllr->mlevel == 1 ) global= MIXTURE_GLOBAL(global)->global;
  for ( n= 0; n < mllr->N; ++n )
    if ( mllr->map[n].N != 0 )
      {
	pcs_set_counts ( &(mllr->map[n]), mllr->amodel, amodel,
			 counts, mllr->type, mllr->mlevel );
	ret= tL_mllr_alg ( &(mllr->map[n].pcs), DGAUSSIAN_GLOBAL(global),
			   mllr->Ws[n].W, &(mllr->data), &err );
	if ( ret == -1 )
	  {
	    fprintf ( stderr, _( "Warning: an error has occurred while"
		      " calculating the %uth MLLR transformation matrix: %s\n" ) ,
		      n+1, err );
	    free ( err );
	  }
      }
  
} /* end tL_mllr_update_matrices */
