/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mllr.h - MLLR.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */

#ifndef __LIBTL_MLLR_H__
#define __LIBTL_MLLR_H__

#include <stdio.h>
#include <stddef.h>

#include "amcounts_public.h"
#include "amodel_public.h"
#include "samples_public.h"
#include "text_public.h"

/** \addtogroup Algs Algorithms
 *  @{
 */
/** \defgroup MLLR MLLR adaptation
 */
/** \addtogroup MLLR
 *  @{
 */

#ifndef __TL_PRV
/** Used in MLLR functions. This structure contains auxiliary data
 *  used to adapt models using the MLLR technique.
 */
struct _tLMLLR
{
  
  const tLAModel *amodel;    /**< Reference to the acoustic model used
                                by the MLLR adapter. */
  
};
#endif

/** Used in MLLR functions. This structure contains auxiliary data
 *  used to adapt models using the MLLR technique.
 */
typedef struct _tLMLLR tLMLLR;

/** Used to assign states to regression classes.
 */
typedef struct
{
  
  unsigned int c;
  unsigned int q;
  unsigned int rc;
  
} tLMLLRrc;

/** Frees memory. Frees the memory allocated for the MLLR structure.
 *
 *  @param mllr The MLLR structure.
 */
TL_API_FUNC(void)
tL_mllr_free (
	      tLMLLR *mllr
	      );

/** Adapts the model. This function adapts the acoustic model passed
 *  using the current MLLR transformation matrices. The acoustic model
 *  must have the same structure as the original acoustic model
 *  which was used to make the MLLR structure.
 *
 *  @param mllr The MLLR structure.
 *  @param amodel The acoustic model to adapt. If NULL, then
 *  the original acoustic is copied and adapted.
 *  @return The adapted acoustic model.
 */
TL_API_FUNC(tLAModel *)
tL_mllr_get_adapted_model (
			   tLMLLR   *mllr,
			   tLAModel *amodel
			   );

/** Initializes the MLLR transformation matrices. This function
 *  initializes the MLLR transformation matrices. The A matrix is
 *  initialized to the identity matrix, and the b vector is
 *  initialized with zeroes.
 *
 *  @param mllr The MLLR structure.
 */
TL_API_FUNC(void)
tL_mllr_init (
	      tLMLLR *mllr
	      );

/** Get a tLMLLR. This function creates a MLLR structure used to
 *  adapt models and transform features using the MLLR technique. Only
 *  DGaussian based types are supported. This function calls
 *  'tL_mllr_init'.
 *
 *  @param amodel The acoustic model.
 *  @param rcs An array used to assign states to regression
 *  classes. By default, all classes are assigned to the regression
 *  class 0. If NULL, then all states are assigned to the regression
 *  class 0.
 *  @param length_rcs The size of rcs.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return MLLR structure, or NULL in case of error.
 */
TL_API_FUNC(tLMLLR *)
tL_mllr_new (
	     tLAModel            *amodel,
	     const tLMLLRrc      *rcs,
	     const unsigned int   length_rcs,
	     char               **err
	     );

/** Update the transformation matrices. This function recalculates the
 *  MLLR transformation matrices using the provided logarithmic counts
 *  and the parameters of the acoustic model used to estimate the
 *  counts. Counts must be created using 'tL_amlogcounts_new_mllr',
 *  and the acoustic model must have the same structure as the
 *  structure of the acoustic model used to create the MLLR
 *  structure.
 *
 *  @param mllr The MLLR structure.
 *  @param counts Logarithmic counts related to the acoustic model
 *  of the CMLLR structure. The counts must be created using
 *  'tL_amlogcounts_new_mllr'.
 *  @param amodel The acoustic model parameters used to estimate the
 *  counts.
 */
TL_API_FUNC(void)
tL_mllr_update_matrices (
			 tLMLLR              *mllr,
			 const tLAMlogCounts *counts,
			 const tLAModel      *amodel
			 );

/** @} */
/** @} */

#endif /* __LIBTL_MLLR_H__ */
