/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mllr_alg.c - Algorisme MLLR.
 *
 *  Miguel Ángel del Agua Teba.
 */

#include <assert.h>
#include <stddef.h>
#include <stdlib.h>

#define __TL_PRV

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "amcounts/dgaussian.h"
#include "amodel/dgaussian.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "mllr_alg.h"
#include <math.h>
#include <cblas.h>
#include <clapack.h>

#define ABS(a) (((a) < 0) ? -(a) : (a))


void tL_mllr_alg_data_alloc(
				  mllr_alg_data_t *mem_vars,
				  uint                   D
				  )
{
  double *mem;
  uint d;

  /* Reserva de Memoria para G */
  mt(mem_vars->G= mem_alloc(double *, (D+1))); 
  mt(mem= mem_alloc(double, (D+1)*(D+1)));
  for ( d= 0; d < D+1; ++d, mem+= (D+1))mem_vars->G[d]= mem;
  /******************************/

  /* Reserva de Memoria para Z */
  mt(mem_vars->Z= mem_alloc(double *, D+1)); 
  mt(mem= mem_alloc(double, D*(D+1)));
  for ( d= 0; d <= D; ++d, mem+= D)mem_vars->Z[d]= mem;
  /******************************/

  /* Reserva de memoria para el vector ipiv */
  mt(mem_vars->ipiv= mem_alloc(int, D+1));
  /******************************/
  
  mt ( mem_vars->z= mem_alloc ( double, D+1 ) );
  
}

void
tL_mllr_alg_data_free (
			     mllr_alg_data_t *mem_vars
			     )
{
  free ( mem_vars->z );
  free(mem_vars->G[0]); free(mem_vars->G);
  free(mem_vars->Z[0]); free(mem_vars->Z);
  free(mem_vars->ipiv);
}


/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
getEps  (
        dgaussian_state_t       **dgstate,
        const uint                      D,
        const uint                      i,
        double                       *eps
        )
{
  uint d;
  eps[0]= 1;
  for ( d= 1; d < D+1; ++d ) eps[d]= dgstate[i]->mu[d-1];
}

static void
getZ    (
        amcount_dgaussian_state_t  **dgcounts, 
        const uint                      D,    /* Max dimension */
        const double                    I,    /* Number of components */
        double                        **Z     /* Returned matrix */
        )
{
  uint i,j;

  
  for ( j= 0; j < D+1; ++j )
    for ( i= 0; i < D; ++i )
      Z[j][i]= 0;
  
  for ( i= 0; i < I; ++i )
    cblas_daxpy ( (D+1) * D, 1, dgcounts[i]->xx[0], 1, Z[0], 1 );
  
}

static void
getD    (
        dgaussian_state_t       **dgstate,
        const uint                      D,
        const uint                      i,
        double                  **Dmatrix
        )
{
  double *eps;
  mt(eps= mem_alloc(double, (D+1)));
  
  getEps( dgstate, D, i, eps );
  /* A (M*K) B(K*N) = (M*N) */
  cblas_dgemm(CblasColMajor, CblasNoTrans, CblasTrans, /* m */ D+1, /* n */ D+1, /* k */ 1, 1, eps, D+1, eps, D+1, 0, Dmatrix[0], D+1); 
  
  free(eps); 
}

static void
getG    (
        dgaussian_state_t           **dgstate,
        amcount_dgaussian_state_t  **dgcounts,
        uint                                d, /* Fila solicitada */
        uint                                D,
        const double                        I,
        double                            **G
        )
{
  uint i,j;
  double consti, **Dmatrix, *mem;

  mt(Dmatrix= mem_alloc(double*, (D+1)));
  mt(mem= mem_alloc(double, (D+1)*(D+1)));
  for ( i= 0; i < D+1; ++i, mem+=D+1 ) Dmatrix[i]=mem;

  for ( i= 0; i < D+1; ++i )
    for ( j= 0; j < D+1; ++j )
      G[j][i]= 0;
  
  for ( i= 0; i < I; ++i )
  {
    consti= dgstate[i]->ivar[d] * dgcounts[i]->norm;
    getD( dgstate, D, i, Dmatrix );
    cblas_daxpy ( (D+1) * (D+1), consti, Dmatrix[0], 1, G[0], 1 );
  }

  free(Dmatrix[0]); free(Dmatrix);
}


/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

int
tL_mllr_alg (
	     const cmllr_alg_paramscounts_t        *pcs,
	     const dgaussian_global_t           *global,
	     double                                 **W,
	     mllr_alg_data_t                  *mem_vars,
	     char                                 **err
	     )
{

  uint j,d,D;
  double **Z;
  double **G, *z;
  int *ipiv;
  int result;

  Z= mem_vars->Z;
  G= mem_vars->G;
  ipiv= mem_vars->ipiv;
  z= mem_vars->z;

  assert ( pcs->counts[0]->xx != NULL );
  D= global->D; /* Dimension (39) */

  /* Obtengo Z */
  getZ( pcs->counts, D, pcs->I, Z ); 

  for ( d= 0; d < D; ++d )
  {
    getG( pcs->params, pcs->counts, d, D, pcs->I, G );
    
    /* Calculo wiT */
    result= clapack_dgetrf(CblasColMajor, D+1, D+1, G[0], D+1, ipiv); 
    if ( result != 0 ) { tL_msgerror ( err, _( "error in LU factorization" ) ); return -1; }

    for ( j= 0; j < D+1; ++j )
      z[j]= Z[j][d];
    
    /* NRHS= Number of columns of B . W[D] will store the result (wi)*/
    result= clapack_dgetrs( CblasColMajor, CblasNoTrans, D+1, 1, G[0], D+1, ipiv, z, D+1 );
    if ( result != 0 ){ tL_msgerror ( err, _( "error in wi estimation" ) ); return -1; }
    for ( j= 0; j < D+1; ++j ) W[j][d]= z[j];
  }
  
  return 0;
  
} /* end tL_mllr_alg */
