/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mustd.h - Mean and standard deviation estimation.
 *
 *  Adrià Giménez Pastor, 2013.
 *
 */


#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "common.h"
#include "error.h"
#include "mem.h"
#include "mustd.h"
#include "text.h"




/********************/
/* PUBLIC FUNCTIONS */
/********************/

TL_API_FUNC(void)
tL_mustd_accum (
		tLMuStd     *mustd,
		const tLFea *fea
		)
{
  
  float **frames, aux;
  const float *frame;
  int t, d;
  
  
  frames= tL_fea_get_real_data ( fea );
  for ( t= 0; t < fea->cols; ++t )
    {
      frame= frames[t];
      for ( d= 0; d < fea->rows; ++d )
        {
          mustd->mu[d]+= aux= frame[d];
          mustd->std[d]+= aux*aux;
        }
    }
  mustd->N+= fea->cols;
  
} /* tL_mustd_accum */


TL_API_FUNC(void)
tL_mustd_calc (
	       tLMuStd *mustd
	       )
{
  
  unsigned int d;
  double mu;
  
  
  for ( d= 0; d < mustd->D; ++d )
    {
      mustd->mu[d]= mu= mustd->mu[d]/(double) mustd->N;
      mustd->std[d]= sqrt ( (mustd->std[d] - mu*mu)/(double) mustd->N );
    }
  
} /* end tL_mustd_calc */


TL_API_FUNC(void)
tL_mustd_free (
	       tLMuStd *mustd
	       )
{
  
  free ( mustd->mu );
  free ( mustd );
  
} /* end tL_mustd_free */


TL_API_FUNC(void)
tL_mustd_init (
	       tLMuStd *mustd
	       )
{
  
  unsigned int d;
  
  
  for ( d= 0; d < mustd->D; ++d )
    {
      mustd->mu[d]= 0.0;
      mustd->std[d]= 0.0;
    }
  mustd->N= 0;
  
} /* end tL_mustd_init */


TL_API_FUNC(tLMuStd *)
tL_mustd_new (
	      const unsigned int D
	      )
{
  
  tLMuStd *new;
  
  
  new= mem_alloc ( tLMuStd, 1 );
  new->D= D;
  new->N= 0;
  new->mu= mem_calloc ( double, D*2 );
  new->std= new->mu+D;
  
  return new;
  
} /* end tL_mustd_new */


TL_API_FUNC(tLMuStd *)
tL_mustd_new_from_file (
			gzFile   from,
			char   **err
			)
{
  
  tLBuffer *buffer;
  tLMuStd *new;
  unsigned int D, d;
  
  
  /* Prepare */
  new= NULL;
  buffer= tL_buffer_new ();
  
  /* File id. */
  if ( !tL_gtoken ( from, buffer ) )
    {
      TLMSG_TOKID ( err );
      goto error;
    }
  if ( strcmp ( "MUSTD", buffer->v ) )
    {
      TLMSG_UNKTOKID ( err, buffer->v );
      goto error; 
    }
  
  /* Dimension and allocate. */
  if ( !tL_gtoken ( from, buffer ) || strcmp ( "D", buffer->v ) )
    {
      tL_msgerror ( err, _( "Unable to read the token D" ) );
      goto error;
    }
  if ( !tL_gtoken ( from, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the feature vector dimension" ) );
      goto error;
    }
  D= atoui ( buffer->v );
  if ( D == 0 )
    {
      tL_msgerror ( err, _( "The feature vector dimension D cannot be 0" ) );
      goto error;
    }
  new= tL_mustd_new ( D );
  
  /* MU. */
  if ( !tL_gtoken ( from, buffer ) || strcmp ( "MU", buffer->v ) )
    {
      tL_msgerror ( err, _( "Unable to read the token MU" ) );
      goto error;
    }
  for ( d= 0; d < D; ++d )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, _("Unable to read MU[%u]"), d );
	  goto error;
	}
      new->mu[d]= atof ( buffer->v );
    }
  
  /* STD. */
  if ( !tL_gtoken ( from, buffer ) || strcmp ( "STD", buffer->v ) )
    {
      tL_msgerror ( err, _( "Unable to read the token STD" ) );
      goto error;
    }
  for ( d= 0; d < D; ++d )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, _("Unable to read STD[%u]"), d );
	  goto error;
	}
      new->std[d]= atof ( buffer->v );
    }
  
  /* N. */
  if ( !tL_gtoken ( from, buffer ) || strcmp ( "N", buffer->v ) )
    {
      tL_msgerror ( err, _( "Unable to read the token N" ) );
      goto error;
    }
  if ( !tL_gtoken ( from, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the number"
			    " of accumulated samples" ) );
      goto error;
    }
  new->N= atoui ( buffer->v );
  
  
  return new;
  
 error:
  if ( new != NULL ) tL_mustd_free ( new );
  tL_buffer_free ( buffer );
  return NULL;
  
} /* end tL_mustd_new_from_file */


TL_API_FUNC(void)
tL_mustd_print (
		const tLMuStd *mustd,
		FILE          *to
		)
{
  
  unsigned int d;
  
  
  fprintf ( to, "MUSTD\n" );
  fprintf ( to, "D %u\n", mustd->D );
  fprintf ( to, "MU" );
  for ( d= 0; d < mustd->D; ++d )
    fprintf ( to, " %g", mustd->mu[d] );
  fputc ( '\n', to );
  fprintf ( to, "STD" );
  for ( d= 0; d < mustd->D; ++d )
    fprintf ( to, " %g", mustd->std[d] );
  fputc ( '\n', to );
  fprintf ( to, "N %u\n", mustd->N );
  
} /* end tL_mustd_print */
