/*
 *  Copyright 2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mustd.h - Mean and standard deviation estimation.
 *
 *  Adrià Giménez Pastor, 2013.
 *
 */

#ifndef __LIBTL_MUSTD_H__
#define __LIBTL_MUSTD_H__

#include "port.h"
#include "common_public.h"
#include "samples_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \addtogroup Extract
 *  @{
 */

/** Used to calculate and store the mean and the standard deviation of
 *  real feature vectors.
 */
typedef struct
{
  
  unsigned int  D;      /**< Dimension of the feature vectors. */
  double       *mu;     /**< Mean values. */
  double       *std;    /**< Standard Deviation values. */
  unsigned int  N;      /**< Total number of frames. */
  
} tLMuStd;

/** Accumulates statistics. This function is used to accumulate
 *  statistics, from the provided feature vectors, for the calculation
 *  of the mean and the standard deviation. The mu and std fields are
 *  used to store these statistics provisionally. The feature vectors
 *  must be of dimension D and type TL_FEA_REAL.
 *
 *  @param mustd The tLMuStd structure.
 *  @param fea The feature vectors.
 */
TL_API_FUNC(void)
tL_mustd_accum (
		tLMuStd     *mustd,
		const tLFea *fea
		);

/** Calculates the mean and the standard deviations. This function
 *  calculates the mean and the standard deviation using the
 *  statistics in the structure.
 *
 *  @param mustd The tLMuStd structure.
 */
TL_API_FUNC(void)
tL_mustd_calc (
	       tLMuStd *mustd
	       );

/** Frees memory. Frees the memory allocated for a tLMuStd.
 *
 *  @param mustd The tLMuStd structure.
 */
TL_API_FUNC(void)
tL_mustd_free (
	       tLMuStd *mustd
	       );

/** Initializes a tLMuStd structure. This function initializes the
 *  mean and standard deviation values to zero.
 *
 *  @param mustd The tLMuStd structure.
 */
TL_API_FUNC(void)
tL_mustd_init (
	       tLMuStd *mustd
	       );

/** Allocs a new tLMuStd structure. This function allocates memory for
 *  new tLMuStd values, which will be used to estimate the mean and
 *  the standard deviation of real feature vectors of the given
 *  dimension. The new structure is always intialized using
 *  tL_mustd_init.
 *
 *  @param D The dimension of the real fature vectors.
 *  @return The new tLMuStd structure.
 */
TL_API_FUNC(tLMuStd *)
tL_mustd_new (
	      const unsigned int D
	      );

/** Reads the mean and standard deviation. This function allocates
 *  memory for new tLMuStd values, and initializes them using the
 *  values from the provided text description.
 *
 *  @param from File where the text description is stored.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The new tLMuStd structure, or NULL in case of error.
 */
TL_API_FUNC(tLMuStd *)
tL_mustd_new_from_file (
			gzFile   from,
			char   **err
			);

/** Prints the mean and standard deviation. This function writes to
 *  the given file the current value for mean and standard deviation.
 *
 *  @param lexicon The tLMuStd structure.
 *  @param to File to which the mean and standard deviation will be
 *  written.
 */
TL_API_FUNC(void)
tL_mustd_print (
		const tLMuStd *mustd,
		FILE          *to
		);

/** @} */
/** @} */

#endif /* __LIBTL_MUSTD_H__ */
