/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  priors.c - Implementació de 'priors.h'.
 *
 *  Adrià Giménez Pastor, 2008,2011,2012,2013.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "common.h"
#include "error.h"
#include "lexicon.h"
#include "mem.h"
#include "priors.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define EMSG_UREADN "Unable to read the number of entries N"
#define EMSG_N0 "N can not be 0"
#define EMSG_UREADENTRY(N) "Unable to read the %luth entry", _PS(N+1)
#define EMSG_NDIFFERS "N differs from the number of words in the lexicon"




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

/* Torna -1 en cas d'error. */
static int
register_label (
		const char    *token,
		tLLexicon     *lexicon,
		const tLBool   emptylex,
		tLDict        *syms,
		tLSeq        *cs,
		size_t        *cs_size,
		size_t         n,
		size_t        *id,
		char         **err
		)
{
  
  if ( emptylex )
    {
      if ( syms != NULL )
	{
	  if ( tL_utf8tocseq ( token, syms, cs, cs_size, err ) == -1 )
	    return -1;
	}
      if ( !tL_lexicon_add ( lexicon, token, 0.0, cs ) )
	{
	  tL_msgerror ( err, _( "The label %s appears repeated in the"
			" %luth entry" ) , token, _PS(n+1) );
	  return -1;
	}
      *id= n;
    }
  else
    {
      *id= tL_dict_find ( lexicon->words, token );
      if ( *id == lexicon->words->N )
	{
	  tL_msgerror ( err, _( "Label %s in the %luth entry is out"
			" of vocabulary" ) , token, _PS(n+1) );
	  return -1;
	}
    }
  
  return 0;
  
} /* end register_label */


static tLPriors *
read_priors_text (
		  gzFile         from,
		  tLLexicon     *lexicon,
		  const tLBool   emptylex,
		  tLDict        *syms,
		  tLBuffer      *buffer,
		  char         **err
		  )
{
  
  tLPriors *new;
  size_t N, n, cs_size, id;
  tLSeq cs;
  
  
  new= NULL;
  cs.v= NULL;
  
  /* Capçalera i inicialitza. */
  if ( !tL_gtoken ( from, buffer ) ||
       strcmp ( "N", buffer->v ) )
    {
      tL_msgerror ( err, _( "Unable to read the token N" ) );
      goto error;
    }
  if ( !tL_gtoken ( from, buffer ) )
    {
      tL_msgerror ( err, EMSG_UREADN );
      goto error;
    }
  N= atos ( buffer->v );
  if ( N == 0 )
    {
      tL_msgerror ( err, EMSG_N0 );
      goto error;
    }
  if ( !emptylex && lexicon->words->N != N )
    {
      tL_msgerror ( err, EMSG_NDIFFERS );
      goto error;
    }
  
  new= tL_priors_new ( N );
  
  /* Entrades. */
  if ( emptylex )
    {
      cs_size= 1;
      mt ( cs.v= mem_alloc ( size_t, 1 ) );
      if ( syms == NULL ) cs.length= 0;
    }
  for ( n= 0; n < N; ++n )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, EMSG_UREADENTRY(n) );
	  goto error;
	}
      if ( register_label ( buffer->v, lexicon, emptylex, syms,
			    &cs, &cs_size, n, &id, err ) == -1 )
	goto error;
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, _( "Unable to read the log-probability of"
			" the %luth entry" ) , _PS(n+1) );
	  goto error;
	}
      new[id]= tL_atop ( buffer->v );
    }
  if ( emptylex )
    {
      tL_lexicon_adjust ( lexicon );
      free ( cs.v );
    }
  
  return new;
  
 error:
  if ( new != NULL ) tL_priors_free ( new );
  if ( cs.v != NULL ) free ( cs.v );
  return NULL;
  
} /* end read_priors_text */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(tLPriors *)
tL_priors_new (
	       const size_t N
	       )
{
  
  tLPriors *new;
  size_t n;
  
  
  mt ( new= mem_alloc ( tLProb, N ) );
  for ( n= 0; n < N; ++n )
    new[n]= 0.0;
  
  return new;
  
} /* end tL_priors_new */


TL_API_FUNC(void)
tL_priors_print (
		 const tLPriors *priors,
		 FILE           *to,
		 const tLDict   *words
		 )
{
  
  size_t n;
  
  
  fprintf ( to, "PRIORS\n" );
  fprintf ( to, "N %lu\n", _PS(words->N) );
  for ( n= 0; n < words->N; ++n )
    {
      fprintf ( to, "%s ", tL_dict_get_token ( words, n) );
      tL_prob_print ( priors[n], to );
      putc ( '\n', to );
    }
  
} /* end tL_priors_print */


TL_API_FUNC(tLPriors *)
tL_priors_new_from_file (
			 gzFile         from,
			 tLLexicon     *lexicon,
			 const tLBool   emptylex,
			 tLDict        *syms,
			 char         **err
			 )
{
  
  tLPriors *new;
  tLBuffer *buffer;
  
  
  buffer= tL_buffer_new ();
  if ( !tL_gtoken ( from, buffer ) )
    {
      TLMSG_TOKID ( err );
      goto error;
    }
  if ( strcmp ( "PRIORS", buffer->v ) )
    {
      TLMSG_UNKTOKID ( err, buffer->v );
      goto error; 
    }
  new= read_priors_text ( from, lexicon, emptylex, syms, buffer, err );
  tL_buffer_free ( buffer );
  
  return new;
  
 error:
  tL_buffer_free ( buffer );
  return NULL;
  
} /* end tL_priors_new_from_file */


TL_API_FUNC(void)
tL_priors_set_gsf (
		   tLPriors      *priors,
		   const size_t   N,
		   const tLFloat  gsf
		   )
{
  
  size_t n;
  
  
  for ( n= 0; n < N; ++n )
    priors[n]*= gsf;
  
} /* end tL_priors_set_gsf */
