/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  priors.h - Definició del model de paraules.
 *
 *  Adrià Giménez Pastor, 2008,2011,2012,2013.
 *
 */

#ifndef __LIBTL_PRIORS_H__
#define __LIBTL_PRIORS_H__


#include <stddef.h>
#include <stdio.h>
#include <zlib.h>

#include "common_public.h"
#include "lexicon_public.h"
#include "port.h"
#include "text_public.h"

/** \addtogroup Models
 *  @{
 */

/** \defgroup Priors Prior probabilities
 */
/** \addtogroup Priors
 *  @{
 */

/* DEFINICIÓ */

/** Table of Prior probabilities. This structure stores a table of
 *  log-probabilities, one entry for each prior probability.
 */
typedef tLProb tLPriors;


/* FUNCIONS */

/** Creates a table of prior probabilities. This functions creates a
 *  new table of prior probabilities of the given size, initialized to
 *  0.
 *
 *  @param N Size of the new table.
 *  @return The new prior probabilities table.
 */
TL_API_FUNC(tLPriors *)
tL_priors_new (
	       const size_t N
	       );

/** Frees memory. Frees the memory allocated for the table of prior
 *  probabilities.
 *
 *  @param PRIORS The table of priors probabilities.
 */
#define tL_priors_free(PRIORS) (free ( PRIORS ))

/* Imprimeix el model de paraules de de probabilitats a priori. Es
 * supossa que els símbols i el diccionari de paraules són els
 * mateixos que amb els que s'ha creat.
 */
/** Prints the table of priors probabilities. This functions writes to
 *  the given file the content of the table of prior probabilities in
 *  text format. Each component is linked to a word or class label,
 *  which is obtained from the provided tLDict. The size of the table
 *  is the number of words in the given tLDict.
 *
 *  @param priors The table of prior probabilities.
 *  @param to File to which the table is written.
 *  @param words A dictionary with the words or class labels. The
 *  number of words must be equal to the size of the table.
 */
TL_API_FUNC(void)
tL_priors_print (
		 const tLPriors *priors,
		 FILE           *to,
		 const tLDict   *words
		 );

/* Llig un model de paraules de probabilitats a priori d'un fitxer. Se
 * li ha de passar un diccionari de paraules. Si se li passa buit el
 * ompli supossant que cada classe és una paraula en UTF-8. Si no ho
 * està generarà un tL_error en cas de paraula fora de diccionari.
 * Els símbols que no apareguen al vocabulari de caràcters produeixen
 * tL_error. syms pot ser NULL, en eixe cas al crear el lèxic es
 * claven seqüències de llargària 0.
 */
/** Reads a table of prior probabilities. This function creates a new
 *  table of prior probabilities from a text description stored in the
 *  given file. A tLLexicon is provided to register the words or class
 *  labels. If the lexicon is empty, new labels are registered into
 *  the lexicon; else, an out-of-vocabulary label is treated as an
 *  error. If a tLDict is provided, new registered labels are
 *  considered to be coded in UTF-8 and split into characters, an
 *  out-of-dictionary character being an error. Otherwise, new labels
 *  are registered whit an empty sequence of symbols. Hence, in this
 *  case the resulting tLLexicon is a badly formed tLLexicon; however,
 *  it can be used to print the table of prior probabilities.
 *
 *  @param from	File where the text description is stored.
 *  @param lexicon The lexicon where labels are registered.
 *  @param emptylex Specifies if the provided lexicon is empty (new
 *  labels must be registered) or not (new labels are treated as
 *  errors).
 *  @param syms A character dictionary used to register new labels. If
 *  not NULL, out-of-vocabulary characters are treated as errors; if
 *  NULL, characters are ignored.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The table of prior probabilities, or NULL in case of
 *  error.
 */
TL_API_FUNC(tLPriors *)
tL_priors_new_from_file (
			 gzFile         from,
			 tLLexicon     *lexicon,
			 const tLBool   emptylex,
			 tLDict        *syms,
			 char         **err
			 );

/* Fixa el 'Grammar Scale Factor'. */
/** Sets the grammar scale factor. This function modifies the given
 *  table of prior probabilities by applying the given grammar scale
 *  factor. Note that if a previous grammar scale factor (gsf0) was
 *  applied, the resulting table will have a grammar scale factor of
 *  gsf0*gsf.
 *
 *  @param priors The table of prior probabilities.
 *  @param N The size of the table.
 *  @param gsf Grammar scale factor.
 */
TL_API_FUNC(void)
tL_priors_set_gsf (
		   tLPriors      *priors,
		   const size_t   N,
		   const tLFloat  gsf
		   );

/** @} */
/** @} */

#endif /* __LIBTL_PRIORS_H__ */
