/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  recognise.c - Reconeixedor de text (o el que es vullga
 *                interpretar).
 *
 *  Adrià Giménez Pastor, 2009,2010,2011,2012,2013.
 *
 *  NOTA: Les funcions que sols s'utilitzen en el reconeixement amb
 *  hmm-in-edges desactivat, he ignorat HMMP i assumit que mai s'usarà
 *  el hypmng en mode wordgraph.
 *
 */


#include <assert.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "calcelprob.h"
#include "amodel.h"
#include "common.h"
#include "eprobst.h"
#include "error.h"
#include "hash.h"
#include "hypmng.h"
#include "lexicon.h"
#include "lexicon_recaux.h"
#include "lookahead.h"
#include "lm.h"
#include "mem.h"
#include "recognise.h"
#include "samples.h"
#include "text.h"
#include "tr_lm.h"
#include "wghypmng.h"




/**********/
/* MACROS */
/**********/

#define take_uhyp(HYPMNG,H)						\
  do {									\
    if ( (HYPMNG).type == HYP ) take_hyp ( (HYPMNG).hyp.mng, H );	\
    else                        take_wghyp ( (HYPMNG).wghyp.mng, H );	\
  } while(0)

#define release_uhyp(HYPMNG,H)						\
  do {									\
    if ( (HYPMNG).type == HYP ) release_hyp ( (HYPMNG).hyp.mng, H );	\
    else                        release_wghyp ( (HYPMNG).wghyp.mng, H ); \
  } while(0)




/*********/
/* TIPUS */
/*********/

/* Gestor d'hipòtesis. */
typedef enum { HYP, WGHYP } hypmng_type_t;
typedef union
{
  hypmng_type_t type;
  struct
  {
    hypmng_type_t  type;
    hypmng_t      *mng;
  }             hyp;
  struct
  {
    hypmng_type_t  type;
    wghypmng_t    *mng;
  }             wghyp;
} uhypmng_t;

/* Representació auxiliar del lèxic. */
typedef enum { MAP, PREFIX } lexaux_type_t;
typedef union
{
  lexaux_type_t type;
  struct
  {
    lexaux_type_t  type;
    lexmap_t      *lex;
  }             map;
  struct
  {
    lexaux_type_t  type;
    lexpref_t     *lex;
    lookahead_t   *la;
  }             prefix;
} lexaux_t;

/* Alguns paràmetres interns que van pegant moltes voltes i que no
   tenen molt de pes. Igual acabe per ficant ací tot. */
typedef struct
{
  
  /* Tipus d'aliniament. */
  tLRecAlignType align;
  
  /* Beam acustic */
  tLProb         max;
  tLProb         thr;
  tLFloat        beam;
  
  /* Beam lm */
  tLProb         lm_max;
  tLProb         lm_thr;
  tLFloat        lm_beam;
  
  /* Fallades de look-ahead. */
  unsigned long  la_fails;
  
  /* Sobre el silenci. */
  size_t         sil;
  tLProb         lpnosil;
  tLBool         wsil;
  
} args_t;


/* AUXILIAR *******************************************************************/
struct _tLRecogniser
{
  
  const tLAModel  *amodel;
  const tLLexicon *lexicon;
  const tLLM      *lm; 
  unsigned long    nmaxstates;
  tLFloat          beam;
  tLFloat          lm_beam;
  tLRecAlignType   align;
  uhypmng_t        hypmng;
  tr_t            *tr;
  letr_t          *letr;
  letr_t          *cwlmtr; /* Te dos camps, però el les fa referència
			      al crossword node. */
  lmtr_t          *lmtr;
  lmtr_t          *lmtrn;
  tLLMIterator     iter;
  eprobst_t       *eprobs;
  lexaux_t         lex;
  unsigned long   *buffer;
  size_t           bsize;
  unsigned int     sil;
  tLProb           lpnosil;
  
};




/************/
/* FUNCIONS */
/************/

/* VITERBI ********************************************************************/
static void
viterbi_iter_lm_insert (
			const tLLMEdge *edi,
			tLProb           p,
			size_t           hyp,
			lmtrnodes_t     *nodes,
			size_t          *actives,
			hypmng_t        *hypmng
			)
{
  
  tLLMEdge ed;
  tLProb aux;
  size_t pnode;
  lmtrnode_t new_node;
  
  
  ed= *edi;
  aux= p+ed.p;
  if ( (pnode= actives[ed.to]) == 0 )
    {
      new_node.id= ed.to;
      new_node.h= hyp;
      new_node.p= aux;
      take_hyp ( hypmng, hyp );
      tL_lmtrnodes_add ( nodes, new_node, actives );
    }
  else if ( aux > nodes->v[pnode].p )
    {
      release_hyp ( hypmng, nodes->v[pnode].h );
      nodes->v[pnode].h= hyp;
      nodes->v[pnode].p= aux;
      take_hyp ( hypmng, hyp );
    }
  
} /* end viterbi_iter_lm_insert */


static void
viterbi_lm_null (
		 lmtr_t       *tr,
		 lmtr_t       *trn,
		 tLLMIterator  i,
		 hypmng_t     *hypmng,
		 const tLLM   *lm
		 )
{
  
  lmtrnodes_t *nnodes0, *nnodes1, *nodes1, *switch_ns;
  size_t *nactives0, *nactives1, *actives1, *switch_as;
  lmtrnode_t *v, node;
  size_t N, n;
  const tLLMEdge *begin, *end, *edi;
  tLLMIter j;
  const tLLMState *states;
  
  
  nodes1= tr->nodes1; actives1= tr->actives1;
  states= lm->states;
  while ( !LMTRNODES_EMPTY ( trn->nodes0 ) )
    {
      nnodes0= trn->nodes0; nnodes1= trn->nodes1;
      nactives0= trn->actives0; nactives1= trn->actives1;
      v= nnodes0->v; N= nnodes0->N;
      for ( n= 1; n <= N; ++n )
	{
	  node= v[n];
	  nactives0[node.id]= 0;
	  lm_iterator_set ( i, node.id );
	  while ( lm_iterator_next ( i, &begin, &end ) )
	    for ( edi= begin; edi != end; ++edi )
	      viterbi_iter_lm_insert ( edi, node.p, node.h,
				       nodes1, actives1, hypmng );
	  j= states[node.id].null_edges;
	  for ( edi= j.begin; edi != j.end; ++edi )
	    viterbi_iter_lm_insert ( edi, node.p, node.h,
				     nnodes1, nactives1, hypmng );
	  release_hyp ( hypmng, node.h );
	}
      nnodes0->N= 0;
      switch_ns= trn->nodes0;
      trn->nodes0= trn->nodes1;
      trn->nodes1= switch_ns;
      switch_as= trn->actives0;
      trn->actives0= trn->actives1;
      trn->actives1= switch_as;
    }
  
} /* end viterbi_lm_null */


static void
viterbi_iter_lm (
		 lmtr_t       *tr,
		 lmtr_t       *trn,
		 tLLMIterator  i,
		 hypmng_t     *hypmng,
		 const tLLM   *lm
		 )
{
  
  lmtrnodes_t *nodes0, *nodes1, *nnodes0;
  size_t *actives0, *actives1, N, n, *nactives0;
  lmtrnode_t *v, node;
  const tLLMEdge *begin, *end, *edi;
  tLLMIter j;
  const tLLMState *states;
  
  
  nodes0= tr->nodes0; nodes1= tr->nodes1; nnodes0= trn->nodes0;
  actives0= tr->actives0; actives1= tr->actives1; nactives0= trn->actives0;
  v= nodes0->v; N= nodes0->N;
  states= lm->states;
  for ( n= 1; n <= N; ++n )
    {
      node= v[n];
      actives0[node.id]= 0;
      lm_iterator_set ( i, node.id );
      while ( lm_iterator_next ( i, &begin, &end ) )
	for ( edi= begin; edi != end; ++edi )
	  viterbi_iter_lm_insert ( edi, node.p, node.h,
				   nodes1, actives1, hypmng );
      j= states[node.id].null_edges;
      for ( edi= j.begin; edi != j.end; ++edi )
	viterbi_iter_lm_insert ( edi, node.p, node.h,
				 nnodes0, nactives0, hypmng );
      release_hyp ( hypmng, node.h );
    }
  nodes0->N= 0;
  if ( !LMTRNODES_EMPTY ( nnodes0 ) )
    viterbi_lm_null ( tr, trn, i, hypmng, lm );
    
} /* end viterbi_iter_lm */


static void
viterbi_iter_le_insert_node (
			     letra_t     *actives,
			     letrnodes_t *nodes,
			     uhypmng_t    hypmng,
			     letrid_t     id,
			     tLProb       p,
			     size_t       h,
			     tLProb       hmmp
			     )
{
  
  unsigned long trapos, pnode;
  letrnode_t new_node;
  
  
  trapos= tL_letra_active ( actives, id );
  if ( (pnode= actives->nodes[trapos].pos) == 0 )
    {
      new_node.id= id;
      new_node.p= p;
      new_node.h= h; take_uhyp ( hypmng, h );
      new_node.hmmp= hmmp;
      tL_letrnodes_add ( nodes, new_node, trapos, actives );
    }
  else if ( p > nodes->v[pnode].p )
    {
      nodes->v[pnode].p= p;
      nodes->v[pnode].hmmp= hmmp;
      release_uhyp ( hypmng, nodes->v[pnode].h );
      nodes->v[pnode].h= h; take_uhyp ( hypmng, h );
    }
  
} /* end viterbi_iter_le_insert_node */


static void
viterbi_lm2le (
	       letr_t         *tr,
	       lmtr_t         *lmtr,
	       const lexmap_t *lexmap,
	       const tLLM     *lm,
	       uhypmng_t       hypmng,
	       const int       t,
	       const args_t   *args
	       )
{
 
  lmtrnode_t *lmv, lmnode;
  size_t lmN, n, *lmactives1, i;
  const size_t *nentries;
  const lexmap_inode_t *imap_w;
  lexmap_inode_t **imap, lexinode;
  const tLLMState *states;
  letrnode_t node;
  letrnodes_t *nodes1;
  letra_t *actives;
  unsigned long word;
  
  
  lmv= lmtr->nodes1->v;
  lmN= lmtr->nodes1->N;
  lmactives1= lmtr->actives1;
  nentries= lexmap->nentries;
  imap= lexmap->imap;
  states= lm->states;
  nodes1= tr->nodes1; actives= tr->actives;
  for ( n= 1; n <= lmN; ++n )
    {
      
      lmnode= lmv[n];
      lmactives1[lmnode.id]= 0;
      node.id.lms= lmnode.id;
      node.h= lmnode.h;
      word= states[lmnode.id].word;
      imap_w= imap[word];
	  
      /* El HMMP sols s'utilitza en el càlcul del wordgraph, per tant
	 és irrelevant. */
      lexinode= imap_w[0];
      node.id.les= lexinode.state;
      node.p= lmnode.p + lexinode.p;
      tL_letrnodes_add ( nodes1, node, 0, actives );
      for ( i= 1; i < nentries[word]; ++i )
	{
	  lexinode= imap_w[i];
	  node.id.les= lexinode.state;
	  node.p= lmnode.p + lexinode.p;
	  take_hyp ( hypmng.hyp.mng, node.h );
	  tL_letrnodes_add ( nodes1, node, 0, actives );
	}
      
    }
  lmtr->nodes1->N= 0;
      
} /* end viterbi_lm2le */


/* LENODE té l'estat del model de llenguatge que s'està insertant, a
   més de la probabilitat i la hipòtesis. lenode.les no es gasta.  */
static void
viterbi_iter_le_insert_lm (
			   hypmng_t         *hypmng,
			   lmtrnodes_t      *lmnodes,
			   size_t           *actives,
			   const letrnode_t  lenode,
			   const size_t      e,
			   const int         t,
			   const tLBool      align,
			   args_t           *args
			   )
{
  
  size_t pnode, h;
  lmtrnode_t lmnode;
  
  
  if ( lenode.p < args->lm_thr ) return;
  if ( (pnode= actives[lenode.id.lms]) == 0 )
    {
      if ( lenode.p > args->lm_max )
	{ args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }
      lmnode.id= lenode.id.lms;
      if ( align )
	{
	  h= tL_hypmng_add ( hypmng, lenode.h, t );
	  lmnode.h= tL_hypmng_add ( hypmng, h, e );
	  release_hyp ( hypmng, h );
	}
      else lmnode.h= tL_hypmng_add ( hypmng, lenode.h, e );
      lmnode.p= lenode.p;
      tL_lmtrnodes_add ( lmnodes, lmnode, actives );
    }
  else if ( lenode.p > lmnodes->v[pnode].p )
    {
      if ( lenode.p > args->lm_max )
	{ args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }
      release_hyp ( hypmng, lmnodes->v[pnode].h );
      if ( align )
	{
	  h= tL_hypmng_add ( hypmng, lenode.h, t );
	  lmnodes->v[pnode].h= tL_hypmng_add ( hypmng, h, e );
	  release_hyp ( hypmng, h );
	}
      else lmnodes->v[pnode].h=	tL_hypmng_add ( hypmng, lenode.h, e );
      lmnodes->v[pnode].p= lenode.p;
    }
  
} /* end viterbi_iter_le_insert_lm */


static void
viterbi_iter_le (
		 letr_t         *tr,
		 lmtr_t         *lmtr,
		 lmtr_t         *lmtrn,
		 tLLMIterator    iter,
		 uhypmng_t       hypmng,
		 const lexmap_t *lexmap,
		 const tLLM     *lm,
		 int             t,
		 args_t         *args
		 )
{
  
  letrnodes_t *nodes0, *nodes1;
  letra_t *actives;
  letrnode_t *v, node;
  size_t n, N, *lmactives0, h;
  lexmap_node_t lexmnode;
  const lexmap_node_t *map;
  lmtrnodes_t *lmnodes0;
  tLBool isfinal;
  
  
  nodes0= tr->nodes0; nodes1= tr->nodes1; actives= tr->actives;
  lmnodes0= lmtr->nodes0; lmactives0= lmtr->actives0;
  v= nodes0->v; N= nodes0->N;
  map= lexmap->map;
  /* Pot haver-se gastat en el nivell inferior. */
  tL_trac_getready ( TRAC(actives) );
  /* NOTA: el letrnodes comença en el 1. */
  for ( n= 1; n <= N; ++n )
    {
      /* NOTA: com el léxicon és lineal les transicions a aquest
	 nivell sols necessiten controlar si ja està actiu o no, en el
	 cas que hi hasquen transicions de l'estat inicial a l'estat
	 final. */
      /* NOTA: Recordar que els identificadors globals dels estats
	 d'una seqüència són consecutius, en el mateix ordre que
	 apareixen en l'entrada. */
      node= v[n];
      lexmnode= map[node.id.les];
      if ( ++lexmnode.state < lexmnode.cseq->length )
	{
	  if ( args->align == TL_REC_ALIGN_SYMS )
	    {
	      h= node.h;
	      node.h= tL_hypmng_add ( hypmng.hyp.mng, h, t );
	      release_hyp ( hypmng.hyp.mng, h );
	    }
	  ++node.id.les;
	  isfinal= FALSE;
	  tL_letrnodes_add ( nodes1, node, 0, actives );
	}
      else isfinal= TRUE;
      
      /* Transició a l'estat final, es a dir, bot al model de
	 llenguatge. 'node.p' ha de contindre la probabilitat del
	 bot. */
      if ( isfinal )
	{
	  /* NOTA!!!: En cas de bot IF estem guardant l'identificador
	     'les' del node anterior. No obstant, açò no és un
	     problema doncs de l'identificador l'única cosa que volem
	     és saber la paraula del lèxic a què pertany. */
	  viterbi_iter_le_insert_lm ( hypmng.hyp.mng, lmnodes0, lmactives0,
				      node, lexmnode.e, t,
				      args->align!=TL_REC_ALIGN_NONE,
				      args );
	  release_hyp ( hypmng.hyp.mng, node.h );
	}
    }
  nodes0->N= 0;
  
  /* Inserta les arestes que venen de dalt. */
  if ( !LMTRNODES_EMPTY ( lmnodes0 ) )
    {
      viterbi_iter_lm ( lmtr, lmtrn, iter, hypmng.hyp.mng, lm );
      viterbi_lm2le ( tr, lmtr, lexmap, lm, hypmng, t, args );
    }
  
  /* Neteja. */
  tL_trac_getready ( TRAC(actives) );
  
} /* end viterbi_iter_le */


static void
viterbi_iter_le_insert_lm_wg (
			      wghypmng_t          *hypmng,
			      lmtrnodes_t         *lmnodes,
			      size_t              *actives,
			      const letrnode_t     lenode,
			      const size_t         e,
			      const int            t,
			      const tLProb         lmp,
			      const unsigned long  w,
			      args_t              *args
			      )
{
  
  /* NOTA: Al crear el WG no pode en aquest punt, però continue
     recalculant el threshold per a poder més avant. */
  
  size_t pnode;
  lmtrnode_t lmnode;

  
  if ( lenode.p > args->lm_max )
    { args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }

  if ( (pnode= actives[lenode.id.lms]) == 0 )
    {
      /*
      if ( lenode.p < args->lm_thr ) return;
      if ( lenode.p > args->lm_max )
        { args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }
      */
      lmnode.id= lenode.id.lms;
      lmnode.h= tL_wghypmng_add ( hypmng, lenode.h, lenode.id.lms,
				  t, lenode.p, lenode.hmmp, lmp, w, e );
      lmnode.p= lenode.p;
      tL_lmtrnodes_add ( lmnodes, lmnode, actives );
    }
  else
    {
      tL_wghypmng_insert ( hypmng, lmnodes->v[pnode].h, lenode.h,
			   lenode.p, lenode.hmmp, lmp, w, e );
      if ( lenode.p > lmnodes->v[pnode].p )
	{
	  /*
	  if ( lenode.p > args->lm_max )
	    { args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }
	  */
	  lmnodes->v[pnode].p= lenode.p;
	}
    }
  
} /* end viterbi_iter_le_insert_lm_wg */


static size_t
viterbi_iter_le_insert_cwlm_hypmng (
				    hypmng_t         *hypmng,
				    const size_t      hyp,
				    const int         t,
				    unsigned long     cwsym,
				    const size_t      e,
				    const tLBool      align,
				    args_t           *args
				    )
{
  
  size_t h2, h, ret;
  tLBool eraseh2;
  
  
  if ( args->wsil )
    {
      if ( align )
	{
	  h2= tL_hypmng_add ( hypmng, hyp, t );
	  eraseh2= TRUE;
	}
      else
	{
	  h2= hyp;
	  eraseh2= FALSE;
	}
    }
  else
    {
      eraseh2= TRUE;
      if ( align )
	{
	  h= tL_hypmng_add ( hypmng, hyp, t );
	  h2= tL_hypmng_add ( hypmng, h, cwsym );
	  release_hyp ( hypmng, h );
	}
      else h2= tL_hypmng_add ( hypmng, hyp, cwsym );
    }
  ret= tL_hypmng_add ( hypmng, h2, e );
  if ( eraseh2 ) release_hyp ( hypmng, h2 );
  
  return ret;
      
} /* end viterbi_iter_le_insert_cwlm_hypmng */


/* LENODE té l'últim símbol reconegut i l'estat del model de
   llenguatge que s'està insertant, a més de la probabilitat i la
   hipòtesis. */
static void
viterbi_iter_le_insert_cwlm (
			     hypmng_t         *hypmng,
			     letrnodes_t      *cwlmnodes,
			     letra_t          *actives,
			     const letrnode_t  lenode,
			     const size_t      e,
			     const int         t,
			     const tLBool      align,
			     args_t           *args
			     )
{
  
  letrnode_t lmnode;
  unsigned long trapos, pnode;
  
  
  if ( lenode.p < args->lm_thr ) return;
  trapos= tL_letra_active ( actives, lenode.id );
  if ( (pnode= actives->nodes[trapos].pos) == 0 )
    {
      if ( lenode.p > args->lm_max )
	{ args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }
      lmnode.id= lenode.id;
      lmnode.h= viterbi_iter_le_insert_cwlm_hypmng ( hypmng, lenode.h, t,
						     lenode.id.les, e,
						     align, args );
      lmnode.p= lenode.p;
      lmnode.hmmp= 0.0;
      tL_letrnodes_add ( cwlmnodes, lmnode, trapos, actives );
    }
  else if ( lenode.p > cwlmnodes->v[pnode].p )
    {
      if ( lenode.p > args->lm_max )
	{ args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }
      release_hyp ( hypmng, cwlmnodes->v[pnode].h );
      cwlmnodes->v[pnode].h=
	viterbi_iter_le_insert_cwlm_hypmng ( hypmng, lenode.h, t,
					     lenode.id.les, e,
					     align, args );
      cwlmnodes->v[pnode].p= lenode.p;
    }
  
} /* end viterbi_iter_le_insert_cwlm */


static void
viterbi_iter_le_insert_cwlm_wg (
				wghypmng_t          *hypmng,
				letrnodes_t         *cwlmnodes,
				letra_t             *actives,
				const letrnode_t     lenode,
				const size_t         e,
				const int            t,
				const tLProb         lmp,
				const unsigned long  w,
				args_t              *args
				)
{
  
  /* NOTA: Al crear el WG no pode en aquest punt, però continue
     recalculant el threshold per a poder més avant. */
  
  unsigned long trapos, pnode;
  letrnode_t lmnode;
  size_t h;
  
  
  if ( lenode.p > args->lm_max )
    { args->lm_max= lenode.p; args->lm_thr= lenode.p-args->lm_beam; }

  trapos= tL_letra_active ( actives, lenode.id );
  if ( (pnode= actives->nodes[trapos].pos) == 0 )
    {
      lmnode.id= lenode.id;
      if ( args->wsil )
	lmnode.h= tL_wghypmng_add ( hypmng, lenode.h, lenode.id.lms,
				    t, lenode.p, lenode.hmmp, lmp, w, e );
      else
	{
	  h= tL_wghypmng_add_sym ( hypmng, lenode.h, lenode.id.les );
	  lmnode.h= tL_wghypmng_add ( hypmng, h, lenode.id.lms,
				      t, lenode.p, lenode.hmmp, lmp, w, e );
	  release_wghyp ( hypmng, h );
	}
      lmnode.p= lenode.p;
      lmnode.hmmp= 0.0;
      tL_letrnodes_add ( cwlmnodes, lmnode, trapos, actives );
    }
  else
    {
      if ( args->wsil )
	tL_wghypmng_insert ( hypmng, cwlmnodes->v[pnode].h, lenode.h,
			     lenode.p, lenode.hmmp, lmp, w, e );
      else
	{
	  h= tL_wghypmng_add_sym ( hypmng, lenode.h, lenode.id.les );
	  tL_wghypmng_insert ( hypmng, cwlmnodes->v[pnode].h, h,
			       lenode.p, lenode.hmmp, lmp, w, e );
	  release_wghyp ( hypmng, h );
	}
      if ( lenode.p > cwlmnodes->v[pnode].p )
	cwlmnodes->v[pnode].p= lenode.p;
    }
  
} /* end viterbi_iter_le_insert_cwlm_wg */


static tLFloat
calc_sil_laprob (
		 const lexpref_t         *lexprefix,
		 const lookahead_t       *lookahead,
		 const lookahead_table_t *latable,
		 unsigned long            les
		 )
{
  
  lexpref_state_edges_t steds;
  tLFloat max, aux;
  const lexpref_fedge_t *i;
  
  
  steds= lexprefix->state_eds[les];
  max= -HUGE_VAL;
  for ( i= steds.fbegin; i != steds.fend; ++i )
    {
      aux= latable->probs[lookahead->map_leaves[i->e]];
      if ( aux > max ) max= aux;
    }
  
  return max;
  
} /* end calc_sils_laprob */


/* NOTA: Si s'està gastant lookahead, la probabilitat de l'estat
   actual ja està restada. */
/* En la transició final, no s'inserta el silenci. */
static void
viterbi_iter_le_prefix_final_trans (
				    const tLLM                  *lm,
				    lmtr_t                      *lmtr,
				    letr_t                      *cwlmtr,
				    uhypmng_t                    hypmng,
				    const lexpref_t             *lexprefix,
				    letrnodes_t                 *nodes1,
				    letra_t                     *actives,
				    letrnode_t                   node,
				    const int                    t,
				    const tLRecAlignType         alignt,
				    lookahead_t                 *lookahead,
				    const lookahead_table_t     *latable,
				    args_t                      *args,
				    const tLBool                 isfinaltrans
				    )
{
  
  tLProb prob, aux, hmmprob;
  lexpref_state_t state;
  lexpref_state_edges_t steds;
  const lexpref_fedge_t *i;
  unsigned long from, psil, cwsym, vi_les;
  tLBool eraseh, align, ishidden;
  letrnode_t sil_node;
  size_t h2;
  lexpref_hinfo_t info;
  const lexpref_hinfo_t *hinfo;
  
  
  
  align= (alignt!=TL_REC_ALIGN_NONE);
  eraseh= FALSE;
  state= lexprefix->states[node.id.les];
  hinfo= lexprefix->hinfo;
  if ( state.type == HIDDEN )
    {
      info= hinfo[node.id.les];
      vi_les= state.h.wcstate;
      cwsym= (args->wsil && info.issil) ?
	lexprefix->states[info.node].h.sym : state.h.sym;
      /* CORRECTE!! Si hi han silencis tampoc podria transitar al
	 silenci perquè no és un símbol final. */
      if ( isfinaltrans && !(lexprefix->isfinal[cwsym]) ) return;
    }
  else
    {
      vi_les= node.id.les;
      /* Per si estem reconeguent amb CW i és silenci. */
      cwsym= state.n.sym;
    }
  steds= lexprefix->state_eds[vi_les];
  
  /* Silence. psil is the silence node related to the current node. It
     is also indicating that current node is at the end of a
     word. There are two type of silence nodes: conventional silence
     nodes, and hidden silence nodes (xwords).  */
  ishidden= FALSE;
  if ( args->wsil )
    {
      if ( state.type == SIL ) psil= 0;
      else if ( state.type == HIDDEN && !lexprefix->isfinal[cwsym] )
	{
	  ishidden= TRUE;
	  psil= hinfo[node.id.les].issil ? 0 : hinfo[node.id.les].node;
	}
      else psil= lexprefix->psils[vi_les];
    }
  else psil= 0;
  if ( psil )
    {
      /* Nova hipòtesis. El que acaba insertarà altra volta t i per
	 tant tindrà sil length=0. El sil quan acabe ficarà la t
	 corresponent. */
      if ( alignt == TL_REC_ALIGN_SYMS )
	{
	  if ( lexprefix->iscw )
	    {
	      if ( hypmng.type == HYP )
		{
		  h2= tL_hypmng_add ( hypmng.hyp.mng, node.h, t );
		  node.h= tL_hypmng_add ( hypmng.hyp.mng, h2, cwsym );
		  
		}
	      else
		{
		  h2= tL_wghypmng_add_seg ( hypmng.wghyp.mng, node.h, t );
		  node.h= tL_wghypmng_add_sym ( hypmng.wghyp.mng, h2, cwsym );
		}
	      release_uhyp ( hypmng, h2 );
	    }
	  else
	    {
	      if ( hypmng.type == HYP )
		node.h= tL_hypmng_add ( hypmng.hyp.mng, node.h, t );
	      else
		node.h= tL_wghypmng_add_seg ( hypmng.wghyp.mng, node.h, t );
	    }
	  eraseh= TRUE;
	}
      else if ( lexprefix->iscw )
	{
	  if ( hypmng.type == HYP )
	    node.h= tL_hypmng_add ( hypmng.hyp.mng, node.h, cwsym );
	  else
	    node.h= tL_wghypmng_add_sym ( hypmng.wghyp.mng, node.h, cwsym );
	  eraseh= TRUE;
	}
      /* Inserta el silenci. */
      if ( !isfinaltrans )
	{
	  sil_node= node;
	  sil_node.id.les= psil;
	  if ( lookahead != NULL ) /* Probabilitat del node associat. */
	    sil_node.p+= ishidden ?
	      latable->probs[lookahead->map_states[vi_les]] :
	      calc_sil_laprob ( lexprefix, lookahead, latable, vi_les );
	  /* Special hidden silece nodes always have only one predecessor.*/
	  if ( !ishidden && lexprefix->mpred[psil] )
	    viterbi_iter_le_insert_node ( actives, nodes1, hypmng,
					  sil_node.id, sil_node.p,
					  sil_node.h, sil_node.hmmp );
	  else
	    {
	      take_uhyp ( hypmng, sil_node.h );
	      tL_letrnodes_add ( nodes1, sil_node, 0, actives );
	    }
	}
      /* Actualitza el node. */
      node.p+= args->lpnosil;
      node.hmmp+= args->lpnosil;
    }
  /* Preliminars si és CW. */
  if ( lexprefix->iscw )
    {
      node.id.les= cwsym;      
      if ( !isfinaltrans &&  /* Comprovació de node útil. */
	   lexprefix->cwlist[cwsym].begin==lexprefix->cwlist[cwsym].end )
	{
	  if ( eraseh ) release_uhyp ( hypmng, node.h );
	  return;
	}
    }
  /* Transicions finals. */
  from= node.id.lms; prob= node.p; hmmprob= node.hmmp;
  for ( i= steds.fbegin; i != steds.fend; ++i )
    {
      if ( lookahead != NULL &&
	   tL_lookahead_check_state_has_table ( lookahead, from ) )
	{
	  node.id.lms= latable->to[i->word];
	  aux= latable->probs[lookahead->map_leaves[i->e]];
	  if ( aux == -HUGE_VAL ) continue;
	  node.p= prob + aux;
	  aux-= i->p;
	}
      else
	{
	  tL_lm_bo_get_trans ( lm, from, i->word, &node.id.lms, &aux );
	  if ( aux == -HUGE_VAL ) continue;
	  node.p= prob + aux + i->p;
	}
      if ( lexprefix->iscw )
	{
	  assert ( state.type == HIDDEN || state.type == SIL );
	  if ( hypmng.type == HYP )
	    viterbi_iter_le_insert_cwlm ( hypmng.hyp.mng, cwlmtr->nodes0,
					  cwlmtr->actives, node, i->e,
					  t, align, args );
	  else
	    {
	      node.hmmp= hmmprob + i->p;
	      viterbi_iter_le_insert_cwlm_wg ( hypmng.wghyp.mng, cwlmtr->nodes0,
					       cwlmtr->actives, node, i->e, t,
					       aux, i->word, args );
	    }
	}
      else
	{
	  if ( hypmng.type == HYP )
	    viterbi_iter_le_insert_lm ( hypmng.hyp.mng, lmtr->nodes0,
					lmtr->actives0,
					node, i->e, t, align, args );
	  else
	    {
	      node.hmmp= hmmprob + i->p;
	      viterbi_iter_le_insert_lm_wg ( hypmng.wghyp.mng, lmtr->nodes0,
					     lmtr->actives0, node, i->e, t,
					     aux, i->word, args );
	    }
	}
    }
  /* Esborra hipòtesis si és el cas. */
  if ( eraseh ) release_uhyp ( hypmng, node.h );
  
} /* end viterbi_iter_le_prefix_final_trans */


static void
viterbi_iter_le_insert_wildcardnode (
				     const lexpref_state_t  state,
				     letra_t               *actives,
				     letrnodes_t           *nodes,
				     uhypmng_t              hypmng,
				     letrid_t               id,
				     tLProb                 prob,
				     size_t                 h,
				     tLProb                 hmmp,
				     const tLBool           ismpred
				     )
{
  
  unsigned long p;
  letrnode_t node;
  
  
  assert ( state.type == WILDCARD );
  if ( ismpred )
    for ( p= state.w.begin; p != state.w.end; ++p )
      {
	id.les= p;
	viterbi_iter_le_insert_node ( actives, nodes, hypmng,
				      id, prob, h, hmmp );
      }
  else
    {
      node.id= id;
      node.p= prob;
      node.hmmp= hmmp;
      node.h= h;
      for ( p= state.w.begin; p != state.w.end; ++p )
	{
	  node.id.les= p;
	  tL_letrnodes_add ( nodes, node, 0, actives );
	  take_uhyp ( hypmng, h );
	}
    }
  
} /* end viterbi_iter_le_insert_wildcardnode */


/* node porta ja fixada la probabilitat (p i hmmp) i l'hipòtesis i
   l'estat del LM. */
/* NOTA!!: Quan es gasta per a fer les transicions de isil, podrà
   botar-se un fonema i anar al següent, però mai podrà acabar la
   paraula. Açò considere que té sentit, perquè una paraula amb un
   únic fonema te que reconeixer almenys eixe fonema. */
static void
viterbi_lm2le_prefix_insert_nodes (
				   letrnode_t           node,
				   letr_t              *tr,
				   const lexpref_t     *lexpref,
				   lookahead_table_t   *latable,
				   lookahead_t         *lookahead,
				   uhypmng_t            hypmng,
				   const int            t,
				   args_t              *args,
				   const unsigned long  isil_hist
				   )
{
  
  lexpref_state_edges_t isteds;
  tLProb cprob, prob;
  const unsigned long *i;
  
  
  isteds= lexpref->state_eds[0];
  cprob= node.p;
  for ( i= isteds.begin; i != isteds.end; ++i )
    {
      node.id.les= *i;
      if ( lookahead != NULL )
	{
	  prob=
	    cprob +
	    latable->probs[lookahead->map_states[node.id.les]];
	}
      else prob= cprob;
      if ( lexpref->mpred[node.id.les] || node.id.lms == isil_hist )
	viterbi_iter_le_insert_node ( tr->actives, tr->nodes1, hypmng, node.id,
				      prob, node.h, node.hmmp );
      else
	{
	  take_uhyp ( hypmng, node.h );
	  node.p= prob;
	  tL_letrnodes_add ( tr->nodes1, node, 0, tr->actives );
	  node.p= cprob;
	}
    }
  
} /* end viterbi_lm2le_prefix_insert_nodes */


static void
viterbi_lm2le_prefix (
		      letr_t              *tr,
		      lmtr_t              *lmtr,
		      const lexpref_t     *lexpref,
		      lookahead_t         *lookahead,
		      const tLLM          *lm,
		      tLLMIterator         iter,
		      uhypmng_t            hypmng,
		      const int            t,
		      args_t              *args,
		      const tLProb         init_hmmp,
		      const unsigned long  isil_hist
		      )
{
 
  lmtrnode_t *lmv, lmnode;
  size_t lmN, n, *lmactives;
  letrnode_t node;
  lookahead_table_t latable;
  
  
  latable.probs=NULL;latable.to=NULL;node.id.les= 0;/*CALLA*/
  
  /* NOTA: Si estem generant wordgraphs, les hipòtesis ja s'han
     registrat quan s'estava transitant a l'actual estat. */
  lmv= lmtr->nodes0->v;
  lmN= lmtr->nodes0->N;
  lmactives= lmtr->actives0;
  for ( n= 1; n <= lmN; ++n )
    {
      lmnode= lmv[n];
      lmactives[lmnode.id]= 0;
      if ( lmnode.p < args->thr || lmnode.p < args->lm_thr )
	/* AÇÒ ÉS IMPORTANT PER AL L-A. */
	{
	  release_uhyp ( hypmng, lmnode.h );
	  continue;
	}
      node.id.lms= lmnode.id;
      node.h= lmnode.h;
      node.p= lmnode.p;
      node.hmmp= init_hmmp;
      if ( lookahead != NULL )
	latable= tL_lookahead_get_table ( lookahead, lexpref,
					  lm, iter, lmnode.id,
					  &(args->la_fails) );
      viterbi_lm2le_prefix_insert_nodes ( node, tr, lexpref, &latable,
					  lookahead, hypmng, t, args,
					  isil_hist );
      release_uhyp ( hypmng, node.h );
    }
  lmtr->nodes0->N= 0;
  
} /* end viterbi_lm2le_prefix */


/* El node inclou tot la informació (prob,hyp,etc.) i el wildcard d'on
   anem a transitar. */
/* NOTA!!: Quan es gasta per a fer les transicions de isil, podrà
   botar-se un fonema i anar al següent, però mai podrà acabar la
   paraula. Açò considere que té sentit, perquè una paraula amb un
   únic fonema te que reconeixer almenys eixe fonema. */
static void
viterbi_lm2le_cwprefix_insert_nodes (
				     letrnode_t           node,
				     letr_t              *tr,
				     const lexpref_t     *lexpref,
				     uhypmng_t            hypmng,
				     lookahead_table_t   *latable,
				     lookahead_t         *lookahead,
				     args_t              *args,
				     const int            t,
				     const unsigned long  isil_hist
				     )
{

  lexpref_cwlist_t cwedges;
  lexpref_state_t state;
  tLProb prob, cprob;
  const unsigned long *i;
  size_t h;
  
  
  cwedges= lexpref->cwlist[node.id.les];
  h= node.h;
  cprob= node.p;
  for ( i= cwedges.begin; i != cwedges.end; ++i )
    {
      node.id.les= *i;
      state= lexpref->states[node.id.les];
      if ( hypmng.type == HYP )
	node.h= tL_hypmng_add ( hypmng.hyp.mng, h, state.h.sym );
      else
	node.h= tL_wghypmng_add_sym ( hypmng.wghyp.mng, h, state.h.sym );
      if ( lookahead != NULL )
	prob= node.p + latable->probs[lookahead->map_states[state.h.wcstate]];
      else prob= node.p;
      if ( lexpref->mpred[state.h.wcstate] || node.id.lms == isil_hist )
	viterbi_iter_le_insert_node ( tr->actives, tr->nodes1, hypmng, node.id,
				      prob, node.h, node.hmmp );
      else
	{
	  take_uhyp ( hypmng, node.h );
	  node.p= prob;
	  tL_letrnodes_add ( tr->nodes1, node, 0, tr->actives );
	  node.p= cprob;
	}
      release_uhyp ( hypmng, node.h );
    }
  
} /* end viterbi_lm2le_cwprefix_insert_nodes */


static void
viterbi_lm2le_cwprefix (
			letr_t              *tr,
			letr_t              *cwlmtr,
			const lexpref_t     *lexpref,
			lookahead_t         *lookahead,
			const tLLM          *lm,
			tLLMIterator         iter,
			uhypmng_t            hypmng,
			const int            t,
			args_t              *args,
			const unsigned long  isil_hist
			)
{
 
  size_t lmN, n;
  letrnode_t node, *lmv, lmnode;
  letra_t *lmactives;
  lookahead_table_t latable;
  
  
  latable.probs=NULL;latable.to=NULL;/*CALLA*/
  
  /* NOTA: Si estem generant wordgraphs, les hipòtesis ja s'han
     registrat quan s'estava transitant a l'actual estat. */
  lmv= cwlmtr->nodes0->v;
  lmN= cwlmtr->nodes0->N;
  lmactives= cwlmtr->actives;
  for ( n= 1; n <= lmN; ++n )
    {
      
      lmnode= lmv[n];
      if ( lmnode.p < args->thr || lmnode.p < args->lm_thr )
	/* AÇÒ ÉS IMPORTANT PER AL L-A. */
	{
	  release_uhyp ( hypmng, lmnode.h );
	  continue;
	}
      node.id.lms= lmnode.id.lms;
      node.h= lmnode.h;
      node.p= lmnode.p;
      node.hmmp= 0; /* Comença una nova aresta entre estats del LM i
		       per tant reseteje la probabilitat. */
      
      if ( lookahead != NULL )
	latable= tL_lookahead_get_table ( lookahead, lexpref,
					  lm, iter, lmnode.id.lms,
					  &(args->la_fails) );
      
      node.id.les= lmnode.id.les; /* Wilcard origen. Necessari per a
				     insert_nodes.*/
      viterbi_lm2le_cwprefix_insert_nodes ( node, tr, lexpref, hypmng,
					    &latable, lookahead,
					    args, t, isil_hist );
      release_uhyp ( hypmng, node.h );
      
    }
  cwlmtr->nodes0->N= 0;
  
  tL_trac_getready ( TRAC(lmactives) );
  
} /* end viterbi_lm2le_cwprefix */


static void
viterbi_insert_isil (
		     letr_t          *tr,
		     const lexpref_t *lexpref,
		     lookahead_t     *lookahead,
		     const tLLM      *lm,
		     tLLMIterator     iter,
		     uhypmng_t        hypmng,
		     args_t          *args
		     )
{
  
  letrnode_t node;
  lookahead_table_t latable;
  
  
  node.id.lms= lm->start;
  node.id.les= lexpref->isil;
  node.p= 0.0;
  node.hmmp= 0.0;
  node.h= (hypmng.type==HYP) ? NULLHYP : WGNULLHYP;
  if ( lookahead != NULL )
    {
      latable= tL_lookahead_get_table ( lookahead, lexpref,
					lm, iter, node.id.lms,
					&(args->la_fails) );
      node.p+= latable.probs[lookahead->map_states[0]];
    }
  take_uhyp ( hypmng, node.h );
  tL_letrnodes_add ( tr->nodes1, node, 0, tr->actives );
  
} /* end viterbi_insert_isil */


static void
viterbi_iter_le_prefix_isil (
			     letr_t           *tr,
			     tLLMIterator      iter,
			     uhypmng_t         hypmng,
			     const lexpref_t  *lexprefix,
			     lookahead_t      *lookahead,
			     const tLLM       *lm,
			     int               t,
			     letrnode_t        node,
			     args_t           *args
			     )
{
  
  lookahead_table_t latable;
  
  
  if ( args->align == TL_REC_ALIGN_SYMS )
    {
      if ( hypmng.type == HYP )
	node.h= tL_hypmng_add ( hypmng.hyp.mng, node.h, t );
      else
	node.h= tL_wghypmng_add_seg ( hypmng.wghyp.mng, node.h, t );
    }
  if ( lookahead != NULL )
    {
      latable= tL_lookahead_get_table ( lookahead, lexprefix,
					lm, iter, node.id.lms,
					&(args->la_fails) );
      node.p-= latable.probs[lookahead->map_states[0]];
    }
  if ( lexprefix->iscw )
    {
      node.id.les= lexprefix->initcw;
      viterbi_lm2le_cwprefix_insert_nodes ( node, tr, lexprefix, hypmng,
					    &latable, lookahead, args,
					    t, node.id.lms );
    }
  else viterbi_lm2le_prefix_insert_nodes ( node, tr, lexprefix, &latable,
					   lookahead, hypmng, t, args,
					   node.id.lms );
  if ( args->align == TL_REC_ALIGN_SYMS ) release_uhyp ( hypmng, node.h );
  
} /* end viterbi_iter_le_prefix_isil */


static void
viterbi_iter_le_prefix (
			letr_t          *tr,
			lmtr_t          *lmtr,
			letr_t          *cwlmtr,
			tLLMIterator     iter,
			uhypmng_t        hypmng,
			const lexpref_t *lexprefix,
			lookahead_t     *lookahead,
			const tLLM      *lm,
			int              t,
			args_t          *args
			)
{
  
  letrnodes_t *nodes0, *nodes1;
  letra_t *actives;
  letrnode_t *v, node;
  size_t n, N, h;
  lexpref_state_edges_t steds;
  lexpref_state_t state, state2;
  const lexpref_state_edges_t *state_edges;
  const lexpref_state_t *states;
  const unsigned long *i;
  lookahead_table_t latable;
  tLFloat cprob, prob, aux;
  const tLBool *mpred;
  tLBool iscw;
  unsigned long vis_les, isil_hist;
  
  
  iscw= lexprefix->iscw;
  nodes0= tr->nodes0; nodes1= tr->nodes1; actives= tr->actives;
  v= nodes0->v; N= nodes0->N;
  state_edges= lexprefix->state_eds;
  states= lexprefix->states;
  mpred= lexprefix->mpred;
  isil_hist= ULONG_MAX;
  /* Pot haver-se gastat en el nivell anterior. */
  tL_trac_getready ( TRAC(actives) );
  /* NOTA: el letrnodes comença en el 1. */
  for ( n= 1; n <= N; ++n )
    {
      /* NOTA: com el léxicon és un aceptor de prefixes les
	 transicions a aquest nivell sols necessiten controlar si ja
	 està actiu o no l'estat destí, en el cas que hi hasquen
	 transicions de l'estat inicial a l'estat final. */
      node= v[n];
      if ( args->wsil && node.id.les == lexprefix->isil )
	{
	  viterbi_iter_le_prefix_isil ( tr, iter, hypmng, lexprefix,
					lookahead, lm, t, node, args );
	  isil_hist= node.id.lms;
	  release_uhyp ( hypmng, node.h );
	  continue;
	}
      state= states[node.id.les];
      vis_les= (state.type==HIDDEN)?state.h.wcstate:node.id.les;
      steds= state_edges[vis_les];
      if ( lookahead != NULL )
	{
	  latable= tL_lookahead_get_table ( lookahead, lexprefix,
					    lm, iter, node.id.lms,
					    &(args->la_fails) );
	  aux=
	    (state.type!=SIL) ?
	    latable.probs[lookahead->map_states[vis_les]] :
	    calc_sil_laprob ( lexprefix, lookahead, &latable,
			      lexprefix->sil2vis[vis_les] );
	  v[n].p= node.p= cprob= node.p - aux;
	}
      else cprob= node.p;
      h= node.h;
      if ( args->align == TL_REC_ALIGN_SYMS )
	{
	  if ( hypmng.type == HYP )
	    node.h= tL_hypmng_add ( hypmng.hyp.mng, node.h, t );
	  else
	    node.h= tL_wghypmng_add_seg ( hypmng.wghyp.mng, node.h, t );
	}
      for ( i= steds.begin; i != steds.end; ++i )
	{
	  node.id.les= *i;
	  if ( lookahead != NULL )
            prob=
              cprob +
              latable.probs[lookahead->map_states[node.id.les]];
          else prob= cprob;
          state2= states[node.id.les];
          if ( state2.type == WILDCARD )
            viterbi_iter_le_insert_wildcardnode ( state2, actives, nodes1,
                                                  hypmng, node.id, prob,
                                                  node.h, node.hmmp,
                                                  mpred[node.id.les] );
          else if ( mpred[node.id.les] )
            viterbi_iter_le_insert_node ( actives, nodes1, hypmng,
                                          node.id, prob, node.h,
                                          node.hmmp );
          else
            {
              take_uhyp ( hypmng, node.h );
              node.p= prob;
              tL_letrnodes_add ( nodes1, node, 0, actives );
              node.p= cprob;
            }
	}
      if ( args->align == TL_REC_ALIGN_SYMS ) release_uhyp ( hypmng, node.h );
      viterbi_iter_le_prefix_final_trans ( lm, lmtr, cwlmtr,
					   hypmng, lexprefix, nodes1, actives,
					   v[n], t, args->align, lookahead,
					   &latable, args, FALSE );
      release_uhyp ( hypmng, h );
    }
  nodes0->N= 0;
  
  /* Inserta les arestes que venen de dalt. */
  if ( iscw )
    {
      if ( !LETRNODES_EMPTY ( cwlmtr->nodes0 ) )
	viterbi_lm2le_cwprefix ( tr, cwlmtr, lexprefix, lookahead,
				 lm, iter, hypmng, t, args, isil_hist );
    }
  else
    {
      if ( !LMTRNODES_EMPTY ( lmtr->nodes0 ) )
	viterbi_lm2le_prefix ( tr, lmtr, lexprefix, lookahead,
			       lm, iter, hypmng, t, args, 0.0, isil_hist );
    }
  
  tL_trac_getready ( TRAC(actives) );
  
} /* end viterbi_iter_le_prefix */


static void
viterbi_iter_insert_node (
			  unsigned long  nmaxstates,
			  tra_t         *actives,
			  trnodes_t     *nodes1,
			  uhypmng_t      hypmng,
			  trid_t         id,
			  tLProb         p,
			  size_t         h,
			  tLProb         hmmp,
			  args_t        *args
			  )
{
  
  trnode_t new_node, aux;
  unsigned long pnode;
  tLBool full;

  
  if ( p < args->thr ) return;
  if ( (full= (TRNODES_NSTATES ( nodes1 ) == nmaxstates)) &&
       p <= TRNODES_MINP ( nodes1 ) ) return;
  
  new_node.trapos= tL_tra_active ( actives, id );
  if ( (pnode= actives->nodes[new_node.trapos].pos) == 0 )
    {
      if ( p > args->max ) { args->max= p; args->thr= p-args->beam; }
      new_node.p= p;
      new_node.id= id;
      new_node.h= h; take_uhyp ( hypmng, h );
      new_node.hmmp= hmmp;
      if ( full )
	{
	  aux= tL_trnodes_pop_add ( nodes1, new_node, actives );
	  release_uhyp ( hypmng, aux.h );
	}
      else tL_trnodes_add ( nodes1, new_node, actives );
    }
  else if ( p > nodes1->v[pnode].p )
    {
      if ( p > args->max ) { args->max= p; args->thr= p-args->beam; }
      nodes1->v[pnode].p= p;
      release_uhyp ( hypmng, nodes1->v[pnode].h );
      nodes1->v[pnode].h= h; take_uhyp ( hypmng, h );
      nodes1->v[pnode].hmmp= hmmp;
      tL_trnodes_update ( nodes1, pnode, actives );
    }
  
} /* end viterbi_iter_insert_node */


static void
viterbi_le2_ (
	      unsigned long  nmaxstates,
	      tr_t          *tr,
	      letr_t        *letr,
	      uhypmng_t      hypmng,
	      lexaux_t       lex,
	      args_t        *args
	      )
{
  
  letrnodes_t *lenodes1;
  size_t N;
  letrnode_t *v, lenode;
  size_t i;
  trid_t id;
  trnodes_t *nodes1;
  tra_t *actives;
  const tLAModelTrans *trans;
  tLAModelTransListState ptrl;
  const tLAModelTransListEdge *ed;
  
  
  /* NOTA: el letrnodes comença en el 1. */
  lenodes1= letr->nodes1;
  N= lenodes1->N; v= lenodes1->v;
  nodes1= tr->nodes1; actives= tr->actives;
  for ( i= 1; i <= N; ++i )
    {
      lenode= v[i];
      id.lms= lenode.id.lms;
      id.les= lenode.id.les;
      trans= lex.type==MAP ?
	&(lex.map.lex->map[id.les].hmm->trans) :
	&(lex.prefix.lex->states[id.les].nh.hmm->trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
	{
	case TLLINEAL:
	  id.q= 0;
	  viterbi_iter_insert_node ( nmaxstates, actives, nodes1, hypmng,
				     id, lenode.p, lenode.h, lenode.hmmp,
				     args );
	  break;
	case TLLIST:
	  ptrl= trans->list.init;
	  for ( ed= ptrl.begin; ed != ptrl.end; ++ed )
	    {
	      id.q= ed->q;
	      viterbi_iter_insert_node ( nmaxstates, actives, nodes1, hypmng,
					 id, lenode.p+ed->p, lenode.h,
					 lenode.hmmp+ed->p, args );
	    }
	  break;
	default: break;
	}
      release_uhyp ( hypmng, lenode.h );
    }
  lenodes1->N= 0;
  
} /* end viterbi_le2_ */


/* Esta funció es crida supossant que hi han nodes actius. */
static void
viterbi_iter (
	      tLRecogniser   *aux,
	      const tLAModel *amodel,
	      const tLLM     *lm,
	      const tLFea    *fea,
	      const int       t,
	      args_t         *args
	      )
{
  
  tr_t *tr;
  letr_t *letr;
  unsigned long nmaxstates;
  eprobst_t *eprobs;
  uhypmng_t hypmng;
  lexaux_t lex;
  trnodes_t *nodes0, *nodes1;
  letrnodes_t *lenodes0;
  letrid_t leid;
  letrnode_t lenode;
  tra_t *actives;
  letra_t *leactives;
  trnode_t node, *vnodes0;
  unsigned long n, N;
  tLProbPair ptr;
  tLProb p, old_max, hmmp, auxp, old_thr;
  tLBool lineal, isfinal;
  const tLAModelTrans *trans;
  tLAModelTransListState ptrl;
  const tLAModelTransListEdge *ed;
  size_t sym;
  const tLAModelHMM *hmm;
  
  
  /* Obté Camps. */
  tr= aux->tr;
  letr= aux->letr;
  nmaxstates= aux->nmaxstates;
  eprobs= aux->eprobs;
  hypmng= aux->hypmng;
  lex= aux->lex;
  
  /* Inicialització. */
  nodes0= tr->nodes0; nodes1= tr->nodes1; actives= tr->actives;
  vnodes0= nodes0->v;
  lenodes0= letr->nodes0; leactives= letr->actives;
  old_max= args->max; old_thr= args->thr;
  args->max= -HUGE_VAL; args->thr= -HUGE_VAL;
  args->lm_max= -HUGE_VAL; args->lm_thr= -HUGE_VAL;
  N= TRNODES_NSTATES ( nodes0 );
  /*if(bframe+10==frame){while(trnodes_pop(nodes0,&node))printf("%g (%ld,%ld,%ld)\n",node.p,node.id.lms,node.id.les,node.id.q);exit(1);}*/
  for ( n= N; n != 0; --n ) /* NOTA: estaria be testejar quin ordre és
			       el més òptim. */
    {
      
      /* Prepara. */
      node= vnodes0[n];
      if ( node.p < old_thr )
	{
	  release_uhyp ( hypmng, node.h );
	  continue;
	}
      if ( lex.type == MAP )
	{
	  hmm= lex.map.lex->map[node.id.les].hmm;
	  sym= lex.map.lex->map[node.id.les].cseq->
	    v[lex.map.lex->map[node.id.les].state];
	}
      else
	{
	  hmm= lex.prefix.lex->states[node.id.les].nh.hmm;
	  sym= lex.prefix.lex->states[node.id.les].nh.sym;
	}
      node.p-= old_max;
      auxp= tL_eprobst_lprob ( eprobs, amodel, fea, sym, node.id.q, t );
      node.p+= auxp;
      node.hmmp+= auxp;
      if ( node.p == -HUGE_VAL )
	{
	  release_uhyp ( hypmng, node.h );
	  continue;
	}
      
      /* Transicions no final. */
      lineal= FALSE;
      trans= &(hmm->trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
	{
	  
	case TLLINEAL:
	  lineal= TRUE;
	  ptr= trans->lineal.p[node.id.q];
	  if ( ptr.p0 != -HUGE_VAL )
	    {
	      p= node.p+ptr.p0;
	      hmmp= node.hmmp+ptr.p0;
	      viterbi_iter_insert_node ( nmaxstates, actives, nodes1, hypmng,
					 node.id, p, node.h, hmmp, args );
	    }
	  p= node.p+ptr.p1;
	  hmmp= node.hmmp+ptr.p1;
	  if ( ++node.id.q < hmm->Q )
	    {
	      isfinal= FALSE;
	      viterbi_iter_insert_node ( nmaxstates, actives, nodes1, hypmng,
					 node.id, p, node.h, hmmp, args );
	    }
	  else isfinal= TRUE;
	  break;
	  
	case TLLIST:
	  ptrl= trans->list.states[node.id.q];
	  for ( ed= ptrl.begin; ed != ptrl.end; ++ed )
	    {
	      p= node.p+ed->p;
	      hmmp= node.hmmp+ed->p;
	      node.id.q= ed->q;
	      viterbi_iter_insert_node ( nmaxstates, actives, nodes1, hypmng,
					 node.id, p, node.h, hmmp, args );
	    }
	  if ( ptrl.fp != -HUGE_VAL )
	    {
	      p= node.p+ptrl.fp;
	      hmmp= node.hmmp+ptrl.fp;
	      isfinal= TRUE;
	    }
	  else isfinal= FALSE;
	  break;
	  
	default: p= hmmp= -HUGE_VAL; isfinal= FALSE; break; /* CALLA!!!. */
	}
      
      /* Transicions finals. Si es produeix la variable 'p' ('hmmp') ha de
	 contindre la probabilitat. */
      if ( isfinal )
	{
	  if ( lineal )
	    {
	      /* Com el HMM és lineal, el node (lms,les) sols es pot
		 alcançar una vegada i per tant no cal registar-lo en
		 la taula hash. */
	      lenode.id.lms= node.id.lms;
	      lenode.id.les= node.id.les;
	      lenode.p= p;
	      lenode.hmmp= hmmp;
	      lenode.h= node.h;
	      tL_letrnodes_add ( lenodes0, lenode, 0, leactives );
	      take_uhyp ( hypmng, node.h );
	    }
	  else
	    {
	      leid.lms= node.id.lms;
	      leid.les= node.id.les;
	      viterbi_iter_le_insert_node ( leactives, lenodes0, hypmng,
					    leid, p, node.h, hmmp );
	    }
	}
      
      release_uhyp ( hypmng, node.h );
      
    }
  
  /* Inserta les arestes que venen de dalt.  */
  if ( !LETRNODES_EMPTY ( lenodes0 ) )
    {
      if ( lex.type == MAP )
	viterbi_iter_le ( letr, aux->lmtr, aux->lmtrn, aux->iter,
			  hypmng, lex.map.lex, lm, t, args );
      else
	viterbi_iter_le_prefix ( letr, aux->lmtr, aux->cwlmtr, aux->iter,
				 hypmng, lex.prefix.lex, lex.prefix.la,
				 lm, t, args );
      viterbi_le2_ ( nmaxstates, tr, letr, hypmng, lex, args );
    }
  
  /* Prepara següent iteració. */
  tL_reset_eprobst ( eprobs );
  tL_tr_getready ( tr );
  
} /* end viterbi_iter */


static void
viterbi_lm_null_final (
		       lmtr_t     *trn,
		       hypmng_t   *hypmng,
		       const tLLM *lm,
		       tLProb     *fprob,
		       size_t     *fhist
		       )
{
  
  lmtrnodes_t *nnodes0, *nnodes1, *switch_ns;
  size_t *nactives0, *nactives1, *switch_as;
  lmtrnode_t *v, node;
  size_t N, n;
  const tLLMEdge *edi;
  tLLMIter j;
  const tLLMState *states;
  tLProb aux;
  
  
  states= lm->states;
  while ( !LMTRNODES_EMPTY ( trn->nodes0 ) )
    {
      nnodes0= trn->nodes0; nnodes1= trn->nodes1;
      nactives0= trn->actives0; nactives1= trn->actives1;
      v= nnodes0->v; N= nnodes0->N;
      for ( n= 1; n <= N; ++n )
	{
	  node= v[n];
	  nactives0[node.id]= 0;
	  aux= node.p + tL_lm_get_fp ( lm, node.id );
	  if ( aux > *fprob )
	    {
	      *fprob= aux;
	      release_hyp ( hypmng, *fhist );
	      *fhist= node.h;
	      take_hyp ( hypmng, node.h );
	    }
	  j= states[node.id].null_edges;
	  for ( edi= j.begin; edi != j.end; ++edi )
	    viterbi_iter_lm_insert ( edi, node.p, node.h,
				     nnodes1, nactives1, hypmng );
	  release_hyp ( hypmng, node.h );
	}
      nnodes0->N= 0;
      switch_ns= trn->nodes0;
      trn->nodes0= trn->nodes1;
      trn->nodes1= switch_ns;
      switch_as= trn->actives0;
      trn->actives0= trn->actives1;
      trn->actives1= switch_as;
    }
  
} /* end viterbi_lm_null_final */


static void
viterbi_iter_final_lm (
		       lmtr_t     *tr,
		       lmtr_t     *trn,
		       hypmng_t   *hypmng,
		       const tLLM *lm,
		       tLProb     *fprob,
		       size_t     *fhist
		       )
{
  
  lmtrnodes_t *nodes0, *nnodes0;
  size_t *actives0, N, n, *nactives0;
  lmtrnode_t *v, node;
  const tLLMEdge *edi;
  tLLMIter j;
  const tLLMState *states;
  tLProb aux;
  
  
  nodes0= tr->nodes0; actives0= tr->actives0;
  nnodes0= trn->nodes0; nactives0= trn->actives0;
  v= nodes0->v; N= nodes0->N;
  states= lm->states;
  for ( n= 1; n <= N; ++n )
    {
      node= v[n];
      actives0[node.id]= 0;
      aux= node.p + tL_lm_get_fp ( lm, node.id );
      if ( aux > *fprob )
	{
	  *fprob= aux;
	  release_hyp ( hypmng, *fhist );
	  *fhist= node.h;
	  take_hyp ( hypmng, node.h );
	}
      j= states[node.id].null_edges;
      for ( edi= j.begin; edi != j.end; ++edi )
	viterbi_iter_lm_insert ( edi, node.p, node.h,
				 nnodes0, nactives0, hypmng );
      release_hyp ( hypmng, node.h );
    }
  nodes0->N= 0;
  if ( !LMTRNODES_EMPTY ( nnodes0 ) )
    viterbi_lm_null_final ( trn, hypmng, lm, fprob, fhist );
  
} /* end viterbi_iter_final_lm */


static void
viterbi_iter_final_le (
		       letr_t         *tr,
		       lmtr_t         *lmtr,
		       lmtr_t         *lmtrn,
		       hypmng_t       *hypmng,
		       const lexmap_t *lexmap,
		       const tLLM     *lm,
		       int             t,
		       tLBool          align,
		       tLProb         *fprob,
		       size_t         *fhist
		       )
{
  
  letrnodes_t *nodes0;
  letra_t *actives;
  letrnode_t *v, node;
  size_t n, N, *lmactives0, pnode, h;
  lexmap_node_t lexmnode;
  const lexmap_node_t *map;
  lmtrnodes_t *lmnodes0;
  lmtrnode_t lmnode;
  tLBool isfinal;
  const tLAModelTrans *trans;
  
  
  nodes0= tr->nodes0; actives= tr->actives;
  lmnodes0= lmtr->nodes0; lmactives0= lmtr->actives0;
  v= nodes0->v; N= nodes0->N;
  map= lexmap->map;
  /* Pot ser s'haja gasta en l'anterior nivell. */
  tL_trac_getready ( TRAC(actives) );
  for ( n= 1; n <= N; ++n )
    {
      node= v[n];
      lexmnode= map[node.id.les];
      if ( ++lexmnode.state == lexmnode.cseq->length ) isfinal= TRUE;
      else
	{
	  lexmnode= map[++node.id.les];
	  if ( ++lexmnode.state == lexmnode.cseq->length )
	    {
	      trans= &(lexmnode.hmm->trans);
	      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
	      if (trans->type == TLLIST &&
		  (node.p+= trans->list.init.fp) != -HUGE_VAL )
		isfinal= TRUE;
	      else isfinal= FALSE;
	    }
	  else isfinal= FALSE;
	}
      /* En cas de bot la probabilitat del bot ha d'estar en
	 'node.p'. */
      if ( isfinal )
	{
	  /* NOTA: Com pasa en 'viterbi_iter_le', en aquest punt
	     l'unica informació d'interés que aporta 'node.id.les' és
	     la paraula a la qual pertany. */
	  if ( (pnode= lmactives0[node.id.lms]) == 0 )
	    {
	      lmnode.id= node.id.lms;
	      if ( align )
		{
		  h= tL_hypmng_add ( hypmng, node.h, t );
		  lmnode.h= tL_hypmng_add ( hypmng, h, lexmnode.e );
		  release_hyp ( hypmng, h );
		}
	      else lmnode.h= tL_hypmng_add ( hypmng, node.h, lexmnode.e );
	      lmnode.p= node.p;
	      tL_lmtrnodes_add ( lmnodes0, lmnode, lmactives0 );
	    }
	  else if ( node.p > lmnodes0->v[pnode].p )
	    {
	      release_hyp ( hypmng, lmnodes0->v[pnode].h );
	      if ( align )
		{
		  h= tL_hypmng_add ( hypmng, node.h, t );
		  lmnodes0->v[pnode].h= tL_hypmng_add ( hypmng, h, lexmnode.e );
		  release_hyp ( hypmng, h );
		}
	      else lmnodes0->v[pnode].h=
		     tL_hypmng_add ( hypmng, node.h, lexmnode.e );
	      lmnodes0->v[pnode].p= node.p;
	    }
	}
      release_hyp ( hypmng, node.h );
    }
  nodes0->N= 0;
  
  /* Inserta les arestes que venen de dalt. */
  if ( !LMTRNODES_EMPTY ( lmnodes0 ) )
    viterbi_iter_final_lm ( lmtr, lmtrn, hypmng, lm, fprob, fhist );
  
} /* end viterbi_iter_final_le */


static void
viterbi_iter_final_lm_prefix_wg (
				 lmtr_t     *tr,
				 wghypmng_t *hypmng,
				 const tLLM *lm,
				 const int   t,
				 tLProb     *fprob,
				 size_t     *fhyp
				 )
{
  
  lmtrnodes_t *nodes0;
  size_t *actives0, N, n;
  lmtrnode_t *v, node;
  const tLLMState *states;
  tLProb aux, lmp;
  
  
  nodes0= tr->nodes0; actives0= tr->actives0;
  v= nodes0->v; N= nodes0->N;
  states= lm->states;
  for ( n= 1; n <= N; ++n )
    {
      node= v[n];
      actives0[node.id]= 0;
      lmp= tL_lm_get_fp ( lm, node.id );
      if ( lmp != -HUGE_VAL )
	{
	   aux= node.p + lmp;
	   if ( aux > *fprob ) *fprob= aux;
	   if ( *fhyp == WGNULLHYP )
            {
              release_wghyp ( hypmng, WGNULLHYP );
	      /* L'important és ficar ULONG_MAX a l'estat, NULL_ENTRY
		 a l'entrada i talla de vocabulari a la paraula. */
	      *fhyp= tL_wghypmng_add ( hypmng, node.h, ULONG_MAX,
				       t, aux, 0.0, lmp, 
				       lm->NW, NULL_ENTRY );
            }
	   else tL_wghypmng_insert ( hypmng, *fhyp, node.h, aux, 0.0,
				     lmp, lm->NW, NULL_ENTRY );
	}
      assert ( states[node.id].null_edges.begin ==
	       states[node.id].null_edges.end );
      release_wghyp ( hypmng, node.h );
    }
  nodes0->N= 0;
  
} /* end viterbi_iter_final_lm_prefix_wg */


static void
viterbi_iter_final_lm_cwprefix (
				letr_t     *tr,
				uhypmng_t   hypmng,
				const tLLM *lm,
				const int   t,
				tLProb     *fprob,
				size_t     *fhyp
				)
{
  
  letrnodes_t *nodes0;
  letra_t *actives;
  size_t N, n;
  letrnode_t *v, node;
  const tLLMState *states;
  tLProb aux, lmp;
  
  
  nodes0= tr->nodes0; actives= tr->actives;
  v= nodes0->v; N= nodes0->N;
  states= lm->states;
  for ( n= 1; n <= N; ++n )
    {
      node= v[n];
      lmp= tL_lm_get_fp ( lm, node.id.lms );
      if ( lmp != -HUGE_VAL )
	{
	  aux= node.p + lmp;
	  if ( aux > *fprob )
	    {
	      *fprob= aux;
	      if ( hypmng.type == HYP )
		{
		  release_hyp ( hypmng.hyp.mng, *fhyp );
		  *fhyp= node.h;
		  take_hyp ( hypmng.hyp.mng, node.h );
		}
	    }
	  if ( hypmng.type == WGHYP )
	    {
	      if ( *fhyp == WGNULLHYP )
		{
		  release_wghyp ( hypmng.wghyp.mng, WGNULLHYP );
		  /* L'important és ficar ULONG_MAX a l'estat, NULL_ENTRY
		     a l'entrada i talla de vocabulari a la paraula. */
		  *fhyp= tL_wghypmng_add ( hypmng.wghyp.mng, node.h,
					   ULONG_MAX, t, aux, 0.0, lmp, 
					   lm->NW, NULL_ENTRY );
		}
	      else tL_wghypmng_insert ( hypmng.wghyp.mng, *fhyp, node.h,
					aux, 0.0, lmp, lm->NW, NULL_ENTRY );
	    }
	}
      assert ( states[node.id.lms].null_edges.begin ==
	       states[node.id.lms].null_edges.end );
      release_uhyp ( hypmng, node.h );
    }
  nodes0->N= 0;
  tL_trac_getready ( TRAC(actives) );
  
} /* end viterbi_iter_final_lm_cwprefix */


static void
viterbi_iter_final_le_prefix (
			      letr_t          *tr,
			      lmtr_t          *lmtr,
			      letr_t          *cwlmtr,
			      lmtr_t          *lmtrn,
			      uhypmng_t        hypmng,
			      const lexpref_t *lexprefix,
			      lookahead_t     *lookahead,
			      const tLLM      *lm,
			      tLLMIterator     iter,
			      int              t,
			      tLProb          *fprob,
			      size_t          *fhist,
			      args_t          *args
			      )
{
  
  letrnodes_t *nodes0;
  letra_t *actives;
  letrnode_t *v, node;
  size_t n, N, h;
  lexpref_state_t state;
  const lexpref_state_t *states;
  lookahead_table_t latable;
  unsigned long vis_les;
  tLFloat aux;
  
  
  nodes0= tr->nodes0; actives= tr->actives;
  v= nodes0->v; N= nodes0->N;
  states= lexprefix->states;
  /* Pot ser s'haja gastat. */
  tL_trac_getready ( TRAC(actives) );
  /* NOTA: el letrnodes comença en el 1. */
  for ( n= 1; n <= N; ++n )
    {
      /* NOTA: com el léxicon és un aceptor de prefixes les
	 transicions a aquest nivell sols necessiten controlar si ja
	 està actiu o no l'estat destí, en el cas que hi hasquen
	 transicions de l'estat inicial a l'estat final. */
      node= v[n];
      h= node.h;
      state= states[node.id.les];
      vis_les= (state.type==HIDDEN)?state.h.wcstate:node.id.les;
      if ( lookahead != NULL )
        {
          latable= tL_lookahead_get_table ( lookahead, lexprefix,
                                            lm, iter, node.id.lms,
					    &(args->la_fails) );
	  aux=
	    (state.type!=SIL) ?
	    latable.probs[lookahead->map_states[vis_les]] :
	    calc_sil_laprob ( lexprefix, lookahead, &latable,
			      lexprefix->sil2vis[vis_les] );
	  v[n].p= node.p= node.p - aux;
        }
      viterbi_iter_le_prefix_final_trans ( lm, lmtr, cwlmtr,
					   hypmng, lexprefix, NULL, NULL,
					   v[n], t, args->align, lookahead,
					   &latable, args, TRUE );
      release_uhyp ( hypmng, h );
    }
  nodes0->N= 0;
  
  /* Inserta les arestes que venen de dalt. */
  if ( lexprefix->iscw )
    {
      if ( !LETRNODES_EMPTY ( cwlmtr->nodes0 ) )
	viterbi_iter_final_lm_cwprefix ( cwlmtr, hypmng, lm, t, fprob, fhist );
    }
  else
    {
      if ( !LMTRNODES_EMPTY ( lmtr->nodes0 ) )
	{
	  if ( hypmng.type == HYP )
	    viterbi_iter_final_lm ( lmtr, lmtrn, hypmng.hyp.mng,
				    lm, fprob, fhist );
	  else
	    viterbi_iter_final_lm_prefix_wg ( lmtr, hypmng.wghyp.mng, lm, t,
					      fprob, fhist );
	}
    }
  
} /* end viterbi_iter_final_le_prefix */


/* Es supossa que té estats actius. */
static void
viterbi_iter_final (
		    tLRecogniser   *aux,
		    const tLAModel *amodel,
		    const tLLM     *lm,
		    const tLFea    *fea,
		    const int       t,
		    tLProb         *fprob,
		    size_t         *fhyp,
		    args_t         *args
		    )
{
  
  tr_t *tr;
  letr_t *letr;
  eprobst_t *eprobs;
  uhypmng_t hypmng;
  lexaux_t lex;
  trnodes_t *nodes0;
  letrnodes_t *lenodes0;
  letrnode_t lenode;
  letra_t *leactives;
  trnode_t node, *vnodes0;
  unsigned long n, N;
  tLProbPair ptr;
  tLProb old_max, auxp, old_thr;
  tLBool lineal, isfinal;
  const tLAModelTrans *trans;
  tLAModelTransListState ptrl;
  size_t sym;
  const tLAModelHMM *hmm;
  
  
  /* Obté paràmtres. */
  tr= aux->tr;
  letr= aux->letr;
  eprobs= aux->eprobs;
  hypmng= aux->hypmng;
  lex= aux->lex;
  
  /* Inicialització del màxim. */
  args->lm_max= args->lm_thr= -HUGE_VAL;
  
  /* Inicialització. */
  nodes0= tr->nodes0;
  vnodes0= nodes0->v;
  lenodes0= letr->nodes0; leactives= letr->actives;
  old_max= args->max; old_thr= args->thr; args->max= -HUGE_VAL;
  N= TRNODES_NSTATES ( nodes0 );
  for ( n= N; n != 0; --n )
    {
      
      /* Prepara. */
      node= vnodes0[n];
      if ( node.p < old_thr )
	{
	  release_uhyp ( hypmng, node.h );
	  continue;
	}
      if ( lex.type == MAP )
        {
          hmm= lex.map.lex->map[node.id.les].hmm;
          sym= lex.map.lex->map[node.id.les].cseq->
            v[lex.map.lex->map[node.id.les].state];
        }
      else
        {
          hmm= lex.prefix.lex->states[node.id.les].nh.hmm;
          sym= lex.prefix.lex->states[node.id.les].nh.sym;
        }
      node.p-= old_max;
      auxp= tL_eprobst_lprob ( eprobs, amodel, fea, sym, node.id.q, t );
      node.p+= auxp;
      node.hmmp+= auxp;
      if ( node.p == -HUGE_VAL )
	{
	  release_uhyp ( hypmng, node.h );
	  continue;
	}
      
      /* Determina si hi ha transició. */
      lineal= FALSE;
      trans= &(hmm->trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
        {
          
        case TLLINEAL:
	  lineal= TRUE;
          ptr= trans->lineal.p[node.id.q];
          if ( ++node.id.q == hmm->Q )
            {
	      lenode.p= node.p+ptr.p1;
	      lenode.hmmp= node.hmmp+ptr.p1;
              isfinal= TRUE;
            }
          else isfinal= FALSE;
          break;
	  
	case TLLIST:
          ptrl= trans->list.states[node.id.q];
          if ( ptrl.fp != -HUGE_VAL )
            {
              lenode.p= node.p+ptrl.fp;
	      lenode.hmmp= node.hmmp+ptrl.fp;
              isfinal= TRUE;
            }
          else isfinal= FALSE;
          break;
          
        default:
	  isfinal= FALSE; /* CALLA!!!! */
	  tL_error ( _( "Unknown type of transition" ) );;
        }
      
      /* Transició cap a dalt. Si hi ha transició la probabilitat ha
	 d'estar en 'lenode.p'. */
      if ( isfinal )
	{
	  lenode.id.lms= node.id.lms;
	  lenode.id.les= node.id.les;
	  if ( lineal )
	    {
	      lenode.h= node.h;
	      tL_letrnodes_add ( lenodes0, lenode, 0, leactives );
	    }
	  else
	    {
	      viterbi_iter_le_insert_node ( leactives, lenodes0, hypmng,
					    lenode.id, lenode.p, node.h,
					    lenode.hmmp );
	      release_uhyp ( hypmng, node.h );
	    }
	}
      else release_uhyp ( hypmng, node.h );
      
    }
  
  /* Inserta les arestes que venen de dalt.  */
  if ( !LETRNODES_EMPTY ( lenodes0 ) )
    {
      if ( lex.type == MAP )
	viterbi_iter_final_le ( letr, aux->lmtr, aux->lmtrn, hypmng.hyp.mng,
				lex.map.lex, lm, t, aux->align, fprob, fhyp );
      else
	viterbi_iter_final_le_prefix ( letr, aux->lmtr, aux->cwlmtr, aux->lmtrn,
				       hypmng, lex.prefix.lex, lex.prefix.la,
				       lm, aux->iter, t, fprob, fhyp, args );
    }
  
  /* Prepara següent iteració. */
  tL_reset_eprobst ( eprobs );
  tL_tr_getready ( tr );
  
} /* end viterbi_iter_final */


static void
viterbi_init (
	      tLRecogniser *aux,
	      const tLLM   *lm,
	      args_t       *args
	      )
{
  
  letr_t *letr;
  lmtr_t *lmtr, *lmtrn;
  tLLMIterator iter;
  uhypmng_t hypmng;
  const tLLMEdge *begin, *end, *edi;
  tLLMIter j;
  lmtrnodes_t *lmnodes1, *lmnnodes0;
  size_t *lmactives1, *lmnactives0;
  tLFloat tmp_beam, tmp_lm_beam;
  
  
  /* Obté camps. */
  letr= aux->letr;
  lmtr= aux->lmtr;
  lmtrn= aux->lmtrn;
  iter= aux->iter;
  hypmng= aux->hypmng;
  assert ( hypmng.type == HYP );
  
  /* Inicialització del màxim. */
  args->max= args->thr= -HUGE_VAL;
  args->lm_max= args->lm_thr= -HUGE_VAL;
  tmp_lm_beam= args->lm_beam; args->lm_beam= HUGE_VAL; /* Recupera al final. */
  
  /* Model de llenguatge. */
  lmnodes1= lmtr->nodes1; lmactives1= lmtr->actives1;
  lm_iterator_set ( iter, lm->start );
  while ( lm_iterator_next ( iter, &begin, &end ) )
    for ( edi= begin; edi != end; ++edi )
      viterbi_iter_lm_insert ( edi, 0.0, NULLHYP,
			       lmnodes1, lmactives1, hypmng.hyp.mng );
  lmnnodes0= lmtrn->nodes0; lmnactives0= lmtrn->actives0;
  j= lm->states[lm->start].null_edges;
  for ( edi= j.begin; edi != j.end; ++edi )
    viterbi_iter_lm_insert ( edi, 0.0, NULLHYP,
			     lmnnodes0, lmnactives0, hypmng.hyp.mng );
  if ( !LMTRNODES_EMPTY ( lmnnodes0 ) )
    viterbi_lm_null ( lmtr, lmtrn, iter, hypmng.hyp.mng, lm );
  
  /* Model lèxic. */
  if ( !LMTRNODES_EMPTY ( lmnodes1 ) )
    {
      
      viterbi_lm2le ( letr, lmtr, aux->lex.map.lex, lm, hypmng, -1, args );
      tL_trac_getready ( TRAC(letr->actives) );
      
      /* Models de Markov. En la inicialització clavem tots els estats
	 que pogam. */
      if ( !LETRNODES_EMPTY ( letr->nodes1 ) )
	{
	  tmp_beam= args->beam; args->beam= HUGE_VAL;
	  viterbi_le2_ ( ULONG_MAX, aux->tr, letr, hypmng, aux->lex, args );
	  args->beam= tmp_beam;
	  tL_tr_getready ( aux->tr );
	}
      
    }
  
  args->lm_beam= tmp_lm_beam;
  
} /* end viterbi_init */


static void
viterbi_init_prefix (
		     tLRecogniser *aux,
		     const tLLM   *lm,
		     args_t       *args
		     )
{
  
  lmtrnode_t new_node;
  letrnode_t new_nodecw;
  letr_t *letr, *cwlmtr;
  lmtr_t *lmtr;
  uhypmng_t hypmng;
  tLFloat tmp_beam;
  tLBool eraseh;
  size_t h;
  
  
  /* Obté camps. */
  letr= aux->letr;
  lmtr= aux->lmtr;
  cwlmtr= aux->cwlmtr;
  hypmng= aux->hypmng;
  
  /* Inicialització del màxim. */
  args->max= args->thr= -HUGE_VAL;
  args->lm_max= args->lm_thr= -HUGE_VAL;
  
  /* Model de llenguatge, Inserta estat inicial. */
  if ( hypmng.type == HYP )
    {
      take_hyp ( hypmng.hyp.mng, NULLHYP );
      new_nodecw.h= new_node.h= NULLHYP;
    }
  else
    {
      take_wghyp ( hypmng.wghyp.mng, WGNULLHYP );
      new_nodecw.h= new_node.h= WGNULLHYP;
    }
  if ( args->wsil && args->align==TL_REC_ALIGN_SYMS )
    {
      h= new_node.h;
      if ( hypmng.type == HYP )
	new_node.h= tL_hypmng_add ( hypmng.hyp.mng, new_node.h, -1 );
      else
	new_node.h= tL_wghypmng_add_seg ( hypmng.wghyp.mng, new_node.h, -1 );
      new_nodecw.h= new_node.h;
      eraseh= TRUE;
    }
  else eraseh= FALSE;
  if ( aux->lex.prefix.lex->iscw )
    {
      
      new_nodecw.id.lms= lm->start;
      new_nodecw.id.les= aux->lex.prefix.lex->initcw;
      if ( args->wsil )
	{
	  new_nodecw.p= args->lpnosil;
	  new_nodecw.hmmp= args->lpnosil;
	}
      else
	{
	  new_nodecw.p= 0.0;
	  new_nodecw.hmmp= 0.0;
	}
      tL_letrnodes_add ( cwlmtr->nodes0, new_nodecw, 0, cwlmtr->actives );
      
      /* Model lèxic. */
      viterbi_lm2le_cwprefix ( letr, cwlmtr, aux->lex.prefix.lex,
			       aux->lex.prefix.la, lm, aux->iter,
			       hypmng, -1, args, ULONG_MAX );
      
    }
  else
    {
      
      new_node.id= lm->start;
      new_node.p= args->wsil ? args->lpnosil : 0.0;
      tL_lmtrnodes_add ( lmtr->nodes0, new_node, lmtr->actives0 );
      
      /* Model lèxic. */
      viterbi_lm2le_prefix ( letr, lmtr, aux->lex.prefix.lex,
			     aux->lex.prefix.la, lm, aux->iter,
			     hypmng, -1, args, new_node.p, ULONG_MAX );
      
    }
  if ( args->wsil )
    {
      viterbi_insert_isil ( letr, aux->lex.prefix.lex, aux->lex.prefix.la,
			    lm, aux->iter, hypmng, args );
      if ( eraseh ) release_uhyp ( hypmng, h );
    }
  tL_trac_getready ( TRAC(letr->actives) );
  
  /* Models de Markov. En la inicialització clavem tots els estats
     que pogam. */
  if ( !LETRNODES_EMPTY ( letr->nodes1 ) )
    {
      tmp_beam= args->beam; args->beam= HUGE_VAL;
      viterbi_le2_ ( ULONG_MAX, aux->tr, letr, hypmng, aux->lex, args );
      args->beam= tmp_beam;
      tL_tr_getready ( aux->tr );
    }
  
} /* end viterbi_init_prefix */


static void
viterbi (
	 tLRecogniser *aux,
	 const tLFea  *fea,
	 tLProb       *fprob,
	 size_t       *fhyp,
	 tLRecStats   *stats
	 )
{
  
  int t;
  tLProb fprob2;
  tr_t *tr;
  const tLAModel *amodel;
  const tLLM *lm;
  args_t args;
  
  
  /* Inicialitza. */
  tr= aux->tr; amodel= aux->amodel; lm= aux->lm;
  args.align= aux->align;
  args.beam= aux->beam;
  args.lm_beam= aux->lm_beam;
  args.lpnosil= aux->lpnosil;
  args.wsil= (aux->sil<amodel->N);
  args.sil= aux->sil;
  if ( aux->hypmng.type == HYP )
    { *fhyp= NULLHYP; take_hyp ( aux->hypmng.hyp.mng, NULLHYP ); }
  else
    { *fhyp= WGNULLHYP; take_wghyp ( aux->hypmng.wghyp.mng, WGNULLHYP ); }
  fprob2= -HUGE_VAL; *fprob= 0.0;
  init_eprobst ( aux->eprobs, amodel, fea );
  if ( stats != NULL )
    args.la_fails= stats->min_astates=
      stats->max_astates= stats->total_astates= 0;
  
  /* Cos. */
  if ( aux->lex.type == MAP ) viterbi_init ( aux, lm, &args );
  else                        viterbi_init_prefix ( aux, lm, &args );
  (*fprob)+= args.max;
  if ( stats != NULL )
    stats->min_astates= stats->max_astates= stats->total_astates= tr->nodes0->N;
  if ( TRNODES_EMPTY ( tr->nodes0 ) ) goto exit;
  for ( t= 0; t < fea->cols-1; ++t )
    {
      viterbi_iter ( aux, amodel, lm, fea, t, &args );
      (*fprob)+= args.max;
      if ( stats != NULL )
	{
	  if ( tr->nodes0->N > stats->max_astates )
	    stats->max_astates= tr->nodes0->N;
	  if ( tr->nodes0->N < stats->min_astates )
	    stats->min_astates= tr->nodes0->N;
	  stats->total_astates+= tr->nodes0->N;
	}
      if ( TRNODES_EMPTY ( tr->nodes0 ) ) goto exit;
    }
  args.lm_beam= HUGE_VAL; /* no poda en l'última iteració. */
  viterbi_iter_final ( aux, amodel, lm, fea, t, &fprob2, fhyp, &args );
  (*fprob)+= fprob2;
  
 exit:
  stats->lookahead_fails= args.la_fails;
  
  return;
  
} /* end viterbi */


/* N no inclou l'elemnt buit. Sempre fica el sentinela. */
static void
alloc_words (
	     tLRecWord    **rec,
	     size_t        *size,
	     const size_t   N
	     )
{
  
  size_t aux;
  
  
  aux= N+1;
  if ( *rec == NULL )
    {
      *size= aux;
      mt( *rec= mem_alloc ( tLRecWord, aux ) );
    }
  else if ( aux > *size )
    {
      *size= aux;
      mt ( *rec= mem_realloc ( tLRecWord, *rec, aux ) );
    }
  (*rec)[N].cs= NULL;
  
} /* end alloc_words */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_recogniser_free (
		    tLRecogniser *aux
		    )
{
  
  free ( aux->buffer );
  if ( aux->lex.type == MAP ) tL_free_lexmap ( aux->lex.map.lex );
  else
    {
      if ( aux->lex.prefix.la != NULL )
	tL_lookahead_free ( aux->lex.prefix.la );
      tL_free_lexpref ( aux->lex.prefix.lex );
    }
  tL_free_eprobst ( aux->eprobs );
  free_lm_iterator ( aux->iter );
  if ( aux->lmtrn != NULL ) tL_free_lmtr ( aux->lmtrn );
  if ( aux->lmtr != NULL ) tL_free_lmtr ( aux->lmtr );
  if ( aux->cwlmtr != NULL ) tL_free_letr ( aux->cwlmtr );
  tL_free_letr ( aux->letr );
  tL_free_tr ( aux->tr );
  if ( aux->hypmng.type == HYP ) tL_free_hypmng ( aux->hypmng.hyp.mng );
  else                           tL_free_wghypmng ( aux->hypmng.wghyp.mng );
  free ( aux );
  
} /* end tL_recogniser_free */


TL_API_FUNC(tLRecogniser *)
tL_recogniser_new (
		   const tLAModel      *amodel,
		   const tLLexicon     *lexicon,
		   const tLLM          *lm,
		   const tLBool         viterbi_states,
		   const unsigned int   wordgraph_density,
		   const unsigned int   lookahead_capacity,
		   const unsigned int   lookahead_depth,
		   const unsigned int   sil,
		   char               **err
		   )
{
  
  tLRecogniser *new;
  unsigned long init_capacity;
  const tLAModelTrans *trans;
  
  
  mt ( new= mem_alloc ( tLRecogniser, 1 ) );
  new->amodel= amodel;
  new->lexicon= lexicon;
  new->lm= lm;
  new->nmaxstates= ULONG_MAX;
  new->beam= HUGE_VAL;
  new->align= TL_REC_ALIGN_NONE;
  new->sil= sil;
  if ( sil < amodel->N ) /* Takes a_IF as lpnosil */
    {
      trans= &(amodel->hmms[sil].trans);
      if ( trans->type == TLPOINTER ) trans= trans->pointer.ptr;
      switch ( trans->type )
	{
	case TLLIST:
	  new->lpnosil=
	    (trans->list.init.fp!=-HUGE_VAL) ? trans->list.init.fp : 0.0;
	  break;
	default: new->lpnosil= 0.0;
	}
    }
  else new->lpnosil= 0.0;
  new->eprobs= tL_create_eprobst ( amodel, viterbi_states, err );
  if ( new->eprobs == NULL ) goto error;
  if ( lm->backoff )
    {
      /*
      if ( !lm->backoff )
	{
	  tL_msgerror ( err, "HMM in edges is only allowed using"
			" backoff language models" );
	  tL_free_eprobst ( new->eprobs );
	  goto error;
	}
      */
      new->lex.type= PREFIX;
      new->lex.prefix.lex= tL_create_lexpref ( lexicon, amodel, sil );
      new->lex.prefix.la=
	(lookahead_capacity==0) ? NULL :
	tL_lookahead_new ( new->lex.prefix.lex, lexicon,
			   lm, lookahead_capacity, lookahead_depth );
      /*
	lexpref_print_dot ( lex.prefix.lex, syms, lexicon->words, stdout );
	exit ( 0 );
      */
    }
  else
    {
      if ( sil < amodel->N )
	{
	  tL_msgerror ( err, _( "Automatic silence insertion is only"
			" available using back-off LMs" ) );
	  tL_free_eprobst ( new->eprobs );
	  goto error;
	}
      if ( wordgraph_density )
	{
	  tL_msgerror ( err, _( "Word graphs can only be generated using"
			" back-off LMs" ) );
	  tL_free_eprobst ( new->eprobs );
	  goto error;
	}
      if ( lookahead_capacity )
	{
	  tL_msgerror ( err, _( "Lookahead only works using back-off LMs" ) );
	  tL_free_eprobst ( new->eprobs );
	  goto error;
	}
      if ( lexicon->type == TL_LEXICON_CW )
	{
	  tL_msgerror ( err, _( "Crossword lexicon are only allowed using"
			" back-off LMs" ) );
	  tL_free_eprobst ( new->eprobs );
	  goto error;
	}
      new->lex.type= MAP;
      new->lex.map.lex= tL_create_lexmap ( lexicon, amodel );
    }
  if ( wordgraph_density )
    {
      new->hypmng.type= WGHYP;
      new->hypmng.wghyp.mng= tL_create_wghypmng ( wordgraph_density );
    }
  else
    {
      new->hypmng.type= HYP;
      new->hypmng.hyp.mng= tL_create_hypmng ();
    }
  init_capacity= (lexicon->words->N)<<1;
  new->tr= tL_make_tr ( init_capacity==0?1:init_capacity );
  new->letr= tL_make_letr ( 1 );
  if ( lexicon->type == TL_LEXICON_CONV )
    {
      new->cwlmtr= NULL;
      new->lmtr= tL_make_lmtr ( lm );
      new->lmtrn= tL_make_lmtr ( lm );
    }
  else
    {
      init_capacity= (lexicon->words->N)<<3;
      new->cwlmtr= tL_make_letr ( init_capacity==0?1:init_capacity );
      new->lmtr= NULL;
      new->lmtrn= NULL;
    }
  new->iter= tL_lm_get_iterator ( lm );
  mt ( new->buffer= mem_alloc ( unsigned long, 1 ) );
  new->bsize= 1;
  
  return new;
  
 error:
  free ( new );
  return NULL;
  
} /* end tL_recogniser_new */


TL_API_FUNC(int)
tL_recognise (
	      tLRecogniser  *reco,
	      const tLFea   *fea,
	      tLRecWord    **rec,
	      size_t        *size,
	      tLWordGraph  **wg,
	      tLRecStats    *stats
	      )
{
  
  tLProb fprob;
  size_t fhyp, length, n, aux_size, i, nwords, nsyms, j, cwsym, desp, stop;
  unsigned long w;
  int end, prev;
  const tLLexicon *lexicon;
  const tLSeq *cs;
  tLBool iscw, wsil;
  
  
  /* Prepara. */
  lexicon= reco->lexicon;
  iscw= lexicon->type==TL_LEXICON_CW;
  tLassert ( tL_recognise, reco->nmaxstates==0 );
  aux_size= fea->cols;
  if ( reco->align == TL_REC_ALIGN_WORDS )
    {
      if ( iscw ) aux_size*= 4;
      else        aux_size*= 2;
    }
  else if ( reco->align == TL_REC_ALIGN_SYMS )
    {
      /* 1W per Frame --> SYM-BOT-FRAME-BOT-SYM+WORD*/
      /* ATENCIÓ!!!: Si wsil aleshores SYM i el penúltim s'intercanvien. */
      if ( iscw ) aux_size= aux_size*6+1;
      /* 1W per Frame --> BOT-FRAME-BOT+WORD*/
      else        aux_size*= 4;
    }
  else if ( iscw ) aux_size*= 3;
  if ( aux_size > reco->bsize )
    {
      reco->bsize= aux_size;
      mt ( reco->buffer= mem_realloc ( unsigned long,
				       reco->buffer, reco->bsize ) );
    }
  
  /* Reconeix. */
  viterbi ( reco, fea, &fprob, &fhyp, stats );
  if ( (reco->hypmng.type == HYP && fhyp == NULLHYP) ||
       (reco->hypmng.type == WGHYP && fhyp == WGNULLHYP) ) fprob= -HUGE_VAL;
  else
    {
      wsil= (reco->sil<reco->amodel->N);
      if ( reco->hypmng.type == HYP )
	length= tL_get_hyp ( reco->hypmng.hyp.mng, fhyp, reco->buffer );
      else
	length= tL_get_wghyp ( reco->hypmng.wghyp.mng, fhyp,
			       reco->buffer,
			       reco->align!=TL_REC_ALIGN_NONE, iscw, wsil );
      if ( reco->align == TL_REC_ALIGN_SYMS )
	{
	  n= length; nwords= 0;
	  stop= wsil ? 1 : 0;
	  do {
	    --n;
	    nsyms= lexicon->entries[reco->buffer[n]].cs->length;
	    if ( wsil ) ++nsyms;
	    nwords+= nsyms;
	    n-= nsyms;
	    if ( iscw ) n-= 2;
	  } while ( n!=wsil );
	  if ( wsil ) ++nwords;
	  alloc_words ( rec, size, nwords );
	  n= length; i= nwords; end= -10000; cwsym= 0; /*CALLAT*/
	  do {
	    --n;
	    w= lexicon->entries[reco->buffer[n]].w;
	    cs= lexicon->entries[reco->buffer[n]].cs;
	    if ( iscw )
	      {
		if ( wsil )
		  {
		    cwsym= (size_t) reco->buffer[n-2];
		    reco->buffer[n-2]= reco->buffer[n-1];
		    --n;
		  }
		else cwsym= (size_t) reco->buffer[--n];
	      }
	    j= cs->length;
	    if ( wsil ) ++j;
	    do {
	      --n; --j;
	      if ( i != nwords )
		{
		  /* Quan primer prev==-1 -> pos=0 length=ok 
		     Quan segon... prev==-1 -> pos=0 length=0 */
		  prev= (int) reco->buffer[n];
		  (*rec)[i].pos= prev+1;
		  (*rec)[i].length= end-prev;
		  end= prev;
		}
	      else end= (int) reco->buffer[n];
	      --i;
	      (*rec)[i].w= w;
	      (*rec)[i].cs= cs;
	      (*rec)[i].sym= (j==cs->length)?reco->sil:cs->v[j];
	      (*rec)[i].sympos= j;
	    } while ( j!=0 );
	    if ( iscw )
	      {
		(*rec)[i].sym= (size_t) reco->buffer[--n];
		(*rec)[i+cs->length-1].sym= cwsym;
	      }
	  } while ( n!=stop );
	  if ( wsil )
	    {
	      prev= (int) reco->buffer[0];
	      (*rec)[1].pos= prev+1;
	      (*rec)[1].length= end-prev;
	      end= prev;
	      (*rec)[0].w= w;
	      (*rec)[0].cs= cs;
	      (*rec)[0].sym= reco->sil;
	      (*rec)[0].sympos= -1;
	    }
	  (*rec)[0].pos= 0;
	  (*rec)[0].length= end+1;
	}
      else if ( reco->align == TL_REC_ALIGN_WORDS )
	{
	  desp= iscw ? 4 : 2;
	  alloc_words ( rec, size, (length/desp) );
	  prev= 0;
          for ( n= desp-1, i= 0; n < length; n+= desp, ++i )
            {
	      (*rec)[i].w= lexicon->entries[reco->buffer[n]].w;
	      (*rec)[i].cs= cs= lexicon->entries[reco->buffer[n]].cs;
	      (*rec)[i].pos= prev;
	      if ( iscw )
		{
		  end= (int) reco->buffer[n-2];
		  (*rec)[i].sympos= (size_t) reco->buffer[n-1];
		  (*rec)[i].sym= (size_t) reco->buffer[n-3];
		}
	      else
		{	      
		  end= (int) reco->buffer[n-1];
		  (*rec)[i].sympos= cs->v[cs->length-1];
		  (*rec)[i].sym= cs->v[0];
		}
	      (*rec)[i].length= end-prev+1;
              prev= end+1;
            }
	}
      else /* TL_REC_ALIGN_NONE */
	{
	  desp= iscw ? 3 : 1;
	  alloc_words ( rec, size, (length/desp) );
	  for ( n= desp-1, i= 0; n < length; n+= desp, i++ )
            {
	      (*rec)[i].w= lexicon->entries[reco->buffer[n]].w;
	      (*rec)[i].cs= cs= lexicon->entries[reco->buffer[n]].cs;
	      (*rec)[i].pos= 0;
	      (*rec)[i].length= 0;
	      if ( iscw )
		{
		  (*rec)[i].sympos= (size_t) reco->buffer[n-1];
		  (*rec)[i].sym= (size_t) reco->buffer[n-2];
		}
	      else
		{	      
		  (*rec)[i].sympos= cs->v[cs->length-1];
		  (*rec)[i].sym= cs->v[0];
		}
            }
	}
    }
  if ( reco->hypmng.type == HYP )
    {
      release_hyp ( reco->hypmng.hyp.mng, fhyp );
      assert ( reco->hypmng.hyp.mng->v[0].counter == 0 );
    }
  else
    {
      if ( fhyp != WGNULLHYP )
	tL_wghypmng_get_wordgraph ( reco->hypmng.wghyp.mng,
				    fhyp, reco->lm, wg,
				    (reco->align==TL_REC_ALIGN_SYMS)?
				    lexicon:NULL,
				    (reco->sil<reco->amodel->N),
				    reco->sil );
      release_wghyp ( reco->hypmng.wghyp.mng, fhyp );
      assert ( reco->hypmng.wghyp.mng->v[0].counter == 0 );
    }
  assert ( TRNODES_EMPTY ( reco->tr->nodes0 ) );
  
  if ( stats != NULL ) stats->prob= fprob;
  
  return (fprob!=-HUGE_VAL) ? 0 : -1;
  
} /* end tL_recognise */
