/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  recognize.h - Reconeixedor de text (o el que es vullga
 *                interpretar).
 *
 *  Adrià Giménez Pastor, 2009-2013.
 *
 */

#ifndef __LIBTL_RECOGNISE_H__
#define __LIBTL_RECOGNISE_H__

#include <float.h>
#include <limits.h>
#include <stddef.h>
#include <stdlib.h>

#include "amodel_public.h"
#include "common_public.h"
#include "lexicon_public.h"
#include "lm_public.h"
#include "port.h"
#include "samples_public.h"
#include "wordgraph_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup Reco Recognition of samples
 */
/** \addtogroup Reco
 *  @{
 */

/** Recognition alignment type. Type of alignment using during
 *  recognition.
 */
typedef enum {
  TL_REC_ALIGN_NONE,     /**< Recognize without aligning. */
  TL_REC_ALIGN_WORDS,    /**< Align words. */
  TL_REC_ALIGN_SYMS      /**< Align words and symbols. */
} tLRecAlignType;

#ifndef __TL_PRV
/** Recognizer. This structure is used to recognize samples. Public
 *  fields can be modified at any time, except during the execution of
 *  'tL_recognise'. The value of model parameters can be modified, but
 *  not the topology.
 */
struct _tLRecogniser
{
  
  const tLAModel  *amodel;        /**< Reference to the acoustic
				     model used by the recognizer. */
  const tLLexicon *lexicon;       /**< Reference to the lexicon model
				     used by the recognizer. */
  const tLLM      *lm;            /**< Reference to the language model
				     used by the recognizer. */
  unsigned long    nmaxstates;    /**< The maximum number of active
				     states in the trellis. Must be
				     greater than 0. By default,
				     ULONG_MAX. */
  tLFloat          beam;          /**< All states whose probability is
				     lower than LPMAX-beam are purged
				     from the trellis, where LPMAX is
				     the log-probability of current
				     most probable active state. By
				     default, HUGE_VAL. */
  tLFloat          lm_beam;       /**< All LM states whose probability
				     is lower than LPMAX-beam are not
				     inserted into the trellis, where
				     LPMAX is the log-probability of
				     current most probable active
				     state. By default, HUGE_VAL. */
  tLRecAlignType   align;         /**< Specifies the alignment used
				     during recognition. By default,
				     TL_REC_ALIGN_NONE. */
  
};
#endif

/** Recognizer. This structure is used to recognize samples. Public
 *  fields can be modified at any time, except during the execution of
 *  'tL_recognise'. The value of model parameters can be modified, but
 *  not the topology.
 */
typedef struct _tLRecogniser tLRecogniser;

/** Frees memory. Frees the memory allocated for the recognizer.
 *
 *  @param rec The recognizer.
 */
TL_API_FUNC(void)
tL_recogniser_free (
		    tLRecogniser *rec
		    );

/** Gets a tLRecogniser. This function creates a recognizer used
 *  to recognize samples. By default, recognition is carried out by
 *  embedding the HMMs into the states of the language model; however,
 *  when the language model is a back-off model, HMMs can be
 *  optionally embedded into the edges and the lexicon internally
 *  represented as a prefix tree, which can entail a faster
 *  recognition process. By default, the Viterbi approach, that is,
 *  the replacement of operator \f$\sum\f$ by operator \f$\max\f$, is
 *  only applied at state transition level. If desired, the Viterbi
 *  approach can be also applied to the emission probability
 *  functions. Currently supported types are: Bernoulli, DGaussian,
 *  Mixture and TiedStates.
 *
 *  @param amodel The acoustic model.
 *  @param lexicon The lexicon model.
 *  @param lm The language model.
 *  @param viterbi_states If TRUE, the viterbi approach is also
 *  applied inside the emission probability functions.
 *  @param wordgraph_density A positive number indicates that a
 *  word graph with that number of incoming hypothesis in each state
 *  will be generated. A zero value indicates that no word graph will
 *  be generated. Currently, word graphs can only be generated using
 *  back-off LMs.
 *  @param lookahead_capacity Indicates the capacity of the lookahead
 *  cache. A zero value means lookahead disabled. Currently,
 *  lookahead only works with back-off LMs.
 *  @param lookahead_depth Indicates the depth of the n-grams used in
 *  look-ahead, i.e., 1 refers to unigrams, 2 for bigrams, etc. 0 to
 *  the order of the language model.
 *  @param sil If it is a valid symbol (sil<amodel->N), it is the
 *  symbol used as silence, which would be inserted between words.
 *  Otherwise, any symbol is used as silence.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return Auxiliary data, or NULL in case of error.
 */
TL_API_FUNC(tLRecogniser *)
tL_recogniser_new (
		   const tLAModel      *amodel,
		   const tLLexicon     *lexicon,
		   const tLLM          *lm,
		   const tLBool         viterbi_states,
		   const unsigned int   wordgraph_density,
		   const unsigned int   lookahead_capacity,
		   const unsigned int   lookahead_depth,
		   const unsigned int   sil,
		   char               **err
		   );

/** Recognized word. Structure used to store a word recognized by
 *  'tL_recognise'.
 */
typedef struct
{
  
  unsigned long  w;         /**< Word identifier. */
  const tLSeq  *cs;        /**< Symbol sequences. It is worth
			       noting that when a crossword lexicon
			       is used, the first and last symbols are
			       wildcards. */
  int            pos;       /**< Position of the first frame. In case
			       of alignment. */
  int            length;    /**< Number of frames in the word or the
			       current symbol. In case of
			       alignment. */
  size_t         sym;       /**< In case of symbol alignment, this is
			       the current symbol. Usually this value
			       is the same as the one contained in
			       cs, however it could change when a
			       crossword lexicon is used. In case of
			       word alignment, this is the first
			       symbol (the real one, not the wildcard)
			       of cs. */
  size_t         sympos;    /**< In case of symbol alignment, this is
			       the position of the current symbol
			       inside the word. In case of word
			       alignment, this is the last symbol (the
			       real one, not the wildcard) of cs. If
			       silences are inserted between words, in
			       symbol alignment each word ends with a
			       silence whose sympos is cs->length+1,
			       and the silence at the beginning of the
			       sentence has sympos= (size_t) -1. */
  
} tLRecWord;

/** Recognition statistics. Structure used to store statistics from
 *  'tL_recognise'.
 */
typedef struct
{
  
  tLProb        prob;               /**< The log-probability of the
				       path. */
  unsigned long total_astates;      /**< The total number of active
				       states in the trellis. */
  unsigned long max_astates;        /**< Maximum number of
				       simultaneously active
				       states. */
  unsigned long min_astates;        /**< Minimum number of
				       simultaneously active
				       states. */
  unsigned long lookahead_fails;    /**< Lookahead cache fails */
  
} tLRecStats;

/** Recognizes a sample. This function finds the maximum probability
 *  path of a given sample, feature vector sequence, for a provided
 *  recognizer. The resulting recognition is returned into the
 *  variable rec. The recognition is terminated with a tLRecoWord with
 *  a NULL symbol sequence. If rec points to NULL, the needed memory
 *  is allocated and the number of allocated words is stored in the
 *  variable size. If rec doesn't point to NULL, then rec will be
 *  reallocated if needed, according to the value stored in size, and
 *  the variable size will be updated.
 *
 *  @param recogniser The recognizer.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param rec Pointer to the address where the recognition will be
 *  stored. It can point to NULL.
 *  @param size Memory address where the number of allocated words
 *  of the recognition is stored.
 *  @param wg Pointer to the address of a word graph structure where
 *  the generated word graph will be stored. If the address is NULL, a
 *  new word graph is allocated. The value of this variable is ignored
 *  if no word graph is generated during the recognition process.
 *  @param stats Pointer to a RecStats structure, used to store
 *  statistics from the recognition process. It can be NULL.
 *  @return 0 if the sample has been recognized, -1 otherwise.
 */
TL_API_FUNC(int)
tL_recognise (
	      tLRecogniser  *recogniser,
	      const tLFea   *fea,
	      tLRecWord    **rec,
	      size_t        *size,
	      tLWordGraph  **wg,
	      tLRecStats    *stats
	      );

/** @} */
/** @} */

#endif /* __LIBTL_RECOGNISE_H__ */
