/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  samples.c - Implementació de 'samples.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2011,2012.
 *
 */


#include <ctype.h>
#include <limits.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "error.h"
#include "mem.h"
#include "samples.h"
#include "text.h"




/**********/
/* MACROS */
/**********/

#define c2i(c) ((int) ((unsigned char) (c)))
#define i2c(c) ((char) ((unsigned char) (c)))
#define isnl(c) (c2i('\n') == c || c2i('\r') == c)

#define MSGEMPTY(ERR,NAME)				\
  tL_msgerror ( ERR, "File '%s' is empty", (NAME) )




/*********/
/* TIPUS */
/*********/

enum { ERR, P1, P4 };




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static tLFeaSparse *
alloc_fea_sparse ()
{
  
  tLFeaSparse *new;
  
  
  mt ( new= mem_alloc ( tLFeaSparse, 1 ) );
  mt ( new->v= mem_alloc ( int *, 1 ) );
  mt ( new->v[0]= mem_alloc ( int, 1 ) );
  new->sv= new->s= 1;
  new->updated= FALSE;
  
  return new;
  
} /* end alloc_fea_sparse */


static void
free_fea_sparse (
		 tLFeaSparse *fea
		 )
{
  
  free ( fea->v[0] );
  free ( fea->v );
  free ( fea );
  
} /* ebd free_fea_sparse */


static void
resize_fea_sparse (
		   tLFeaSparse *fea,
		   const int    rows,
		   const int    cols
		   )
{
  
  int *mem, c;
  size_t nsize;
  
  
  nsize= ((size_t)(rows+1))*((size_t)cols);
  if ( nsize > fea->s )
    {
      mt ( fea->v[0]= mem_realloc ( int, fea->v[0], nsize ) );
      fea->s= nsize;
    }
  if ( cols > fea->sv )
    {
      mt ( fea->v= mem_realloc ( int *, fea->v, cols ) );
      fea->sv= cols;
    }
  for ( c= 0, mem= fea->v[0]; c < cols; ++c, mem+= rows+1 )
    fea->v[c]= mem;
  
} /* end resize_fea_sparse */


static int
gpbmtoken (
	   gzFile    fd,
	   tLBuffer *buffer
	   )
{
  
  int c;
  size_t pos;
  
  
  /* Busca el primer caracter del següent token. */
  for (;;)
    {
      while ( (c= gzgetc ( fd )) != EOF && isspace ( c ) );
      if ( c == EOF ) return 0;
      else if ( c ==  c2i('#') )
	{
	  while ( (c= gzgetc ( fd )) != EOF && !isnl ( c ) );
	  if ( c == EOF ) return 0;
	  else break;
	}
      else break;
    }
  
  /* Llig el token. */
  pos= 1;
  buffer->v[0]= i2c(c);
  while ( (c= gzgetc ( fd )) != EOF &&
	  !isspace ( c ) && c != c2i('#') )
    {
      if ( pos >= buffer->s )
	tL_resize_buffer ( buffer );
      buffer->v[pos++]= i2c(c);
    }
  if ( pos >= buffer->s )
    tL_resize_buffer ( buffer );
  buffer->v[pos]= '\0';
  
  /* Ignora el comentari si és el cas. */
  if ( c == c2i('#') )
    while ( (c= gzgetc ( fd )) != EOF && !isnl ( c ) );
  
  return 1;
  
} /* end gpbmtoken */


/* Torna 'P1' o 'P4'. Deixa el descriptor de
 * fitxer preparat per a llegir el raster.
 */
static int
read_pbm_header (
		 gzFile     fd,
		 tLBuffer  *buffer,
		 int       *width,
		 int       *height,
		 char     **err
		 )
{
  
  char aux[3];
  int ret;
  
  
  if ( gzread ( fd, &(aux[0]), 3*sizeof(char) ) < 3*sizeof(char) ||
       aux[0] != 'P' || !isspace ( aux[2] ) ||
       (aux[1] != '1' && aux[1] != '4') )
    {
      tL_msgerror ( err, _("The file is not a PBM file" ) );
      return ERR;
    }
  ret= (aux[1] == '1') ? P1 : P4;
  
  if ( !gpbmtoken ( fd, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the width" ) );
      return ERR;
    }
  *width= atoi ( buffer->v );
  if ( *width == 0 )
    {
      tL_msgerror ( err, _( "The width is 0" ) );
      return ERR;
    }
  
  if ( !gpbmtoken ( fd, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the height" ) );
      return ERR;
    }
  *height= atoi ( buffer->v );
  if ( *height == 0 )
    {
      tL_msgerror ( err, _( "The height is 0" ) );
      return ERR;
    }
  
  return ret;
  
} /* read_pbm_header */


/* Torna 0 si tot ha anat bé, -1 en cas d'error. */
static int
read_pbm_raster_P4 (
		    gzFile  fd,
		    tLFea  *fea,
		    int     width,
		    int     height,
		    char  **err

		    )
{
  
  int nbytes, x, y, i, j;
  bit b, mask;
  
  
  nbytes= (width>>3) + ((width & 0x7) ? 1 : 0);
  if ( fea->_aux == NULL )
    {
      mt ( fea->_aux= mem_alloc ( tLByte, nbytes ) );
      fea->_auxs= nbytes;
    }
  else if ( fea->_auxs < nbytes )
    {
      mt ( fea->_aux= mem_realloc ( tLByte, fea->_aux, nbytes ) );
      fea->_auxs= nbytes;
    }
  
  for ( y= height-1; y >= 0; --y )
    {
      
      if ( gzread ( fd, fea->_aux, nbytes*sizeof(bit) ) < nbytes*sizeof(bit) )
	{
	  tL_msgerror ( err, _("Unable to read the %dth row of raster" ) ,
			height-y );
	  return -1;
	}
      
      for ( x= 0, j= 0; x < (nbytes-1); ++x )
	{
	  b= (bit) (fea->_aux[x]);
	  mask= 0x80;
	  for ( i= 0; i < 8; ++i, ++j )
	    {
	      fea->v[j][y]= (tLByte) ((b&mask) ? 1 : 0);
	      mask>>= 1;
	    }
	}
      b= (bit) (fea->_aux[x]);
      mask= 0x80;
      for ( i= 0; j < width; ++i, ++j )
	{
	  fea->v[j][y]= (tLByte) ((b&mask) ? 1 : 0);
	  mask>>= 1;
	}
      
    }
  
  return 0;
  
} /* end read_pbm_raster_P4 */


/* Torna 0 si tot ha anat bé, -1 en cas d'error. */
static int
read_pbm_raster_P1 (
		    gzFile  fd,
		    tLFea  *fea,
		    int     width,
		    int     height,
		    char  **err
		    )
{
  
  int x, y, c;
  
  
  for ( y= height-1; y >= 0; --y )
    for ( x= 0; x < width; ++x )
      {
	
	while ( (c= gzgetc ( fd )) != EOF && isspace ( c ) );
	if ( c == EOF )
	  {
	    tL_msgerror ( err, _( "Unable to read the bit [row:%d,col:%d]" ),
			  height-y, x+1 );
	    return -1;
	  }
	
	if ( c == c2i('1') )
	  fea->v[x][y]= 1;
	else if ( c == c2i('0') )
	  fea->v[x][y]= 0;
	else
	  {
	    tL_msgerror ( err, _( "Invalid character in the bit"
			  " [row:%d,col:%d]" ) , height-y, x+1 );
	    return -1;
	  }
	
      }
  
  return 0;
  
} /* end read_pbm_raster_P1 */


static void
resize_fea (
	    tLFea *fea,
	    int    rows,
	    int    cols
	    )
{
  
  size_t nsize;
  int x, desp;
  tLByte *p;
  
  
  nsize= ((size_t)rows)*((size_t)cols)*fea->_sizee;
  if ( nsize > fea->_s )
    {
      fea->_s= nsize;
      mt ( fea->v[0]=
	   mem_realloc ( tLByte, fea->v[0], nsize ) );
    }
  if ( cols > fea->_sv )
    {
      fea->_sv= cols;
      mt ( fea->v= mem_realloc ( tLByte *, fea->v, cols ) );
    }
  desp= ((size_t)rows)*fea->_sizee;
  for ( x= 1, p= fea->v[0] + desp;
	x < cols; ++x, p+= desp )
    fea->v[x]= p;
  
} /* end resize_fea */


static int
load_pbm (
	  tLFea  *fea,
	  gzFile  f,
	  char  **err
	  )
{
  
  int format, cols, rows, ret;
  
  
  format= read_pbm_header ( f, fea->_buffer, &cols, &rows, err );
  if ( format == ERR ) return -1;
  
  /* Reserva memòria. */
  fea->_sparse->updated= FALSE;
  resize_fea ( fea, rows, cols );
  
  /* Llig imatge. */
  if ( format == P1 )
    ret= read_pbm_raster_P1 ( f, fea, cols, rows, err );
  else
    ret= read_pbm_raster_P4 ( f, fea, cols, rows, err );
  fea->rows= rows;
  fea->cols= cols;
  
  return ret;
  
} /* end load_pbm */


static void
print_pbm (
	   tLFea *fea,
	   FILE  *to
	   )
{
  
  int x, y;
  bit *p;
  
  
  for ( x= 0; x < fea->cols; ++x )
    {
      p= B2b ( fea->v[x] );
      for ( y= 0; y < fea->rows; ++y )
	fprintf ( to, "%d", p[y] );
      fputc ( '\n', to );
    }
  
} /* end print_pbm */


static int
load_real_fea_bin (
		   tLFea  *fea,
		   gzFile  f,
		   char  **err
		   )
{
  
  int aux[2], rows, cols;
  int nfloats;
  size_t nelements, pos;
  
  
  /* Llig capçalera. */
  if ( gzread ( f, aux, 2*sizeof(int) ) != 2*sizeof(int) )
    {
      tL_msgerror ( err, _( "Unable to read the number of rows and columns" ) );
      return -1;
    }
  rows= aux[0];
  cols= aux[1];
  if ( rows < 1 )
    {
      tL_msgerror ( err, _( "The number of rows must be positive" ) );
      return -1;
    }
  if ( cols < 1 )
    {
      tL_msgerror ( err, _( "The number of columns must be positive" ) );
      return -1;
    }
  
  /* Reserva memòria i llig contingut. */
  resize_fea ( fea, rows, cols );
  fea->rows= rows;
  fea->cols= cols;
  nelements= ((size_t)rows)*((size_t)cols);

  pos = 0;
  while ( nelements > 0 )
    { 
      /* gzread can only read a maximum of INT_MAX bytes */
      nfloats = nelements > ( INT_MAX / sizeof(float) ) ?
	( INT_MAX / sizeof(float) ) : (int) nelements;
      
      if ( gzread ( f, &(fea->v[0][pos]),
		    nfloats*sizeof(float) ) != nfloats*sizeof(float) )
	{
	  tL_msgerror ( err, _( "Unable to read the data" ) );
	  return -1;
	}
      nelements -= nfloats;
      pos += nfloats*sizeof(float);
    }
  
  return 0;
  
} /* end load_real_fea_bin */


static int
load_real_fea_text (
		    tLFea  *fea,
		    gzFile  f,
		    char  **err
		    )
{
  
  tLBuffer *buffer;
  int rows, cols, r, c;
  float *p;
  
  
  buffer= tL_buffer_new ();
  
  /* Llig capçalera. */
  if ( !tL_gtoken ( f, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the feature vector dimension" ) );
      goto error;
    }
  rows= atoi ( buffer->v );
  if ( rows == 0 )
    {
      tL_msgerror ( err, _( "The feature vector dimension is 0" ) );
      goto error;
    }
  if ( !tL_gtoken ( f, buffer ) )
    {
      tL_msgerror ( err, _( "Unable to read the number of feature vectors" ) );
      goto error;
    }
  cols= atoi ( buffer->v );
  if ( cols == 0 )
    {
      tL_msgerror ( err, _( "The number of feature vectors is 0" ) );
      goto error;
    }
  
  /* Reserva memòria i llig contingut. */
  resize_fea ( fea, rows, cols );
  fea->rows= rows;
  fea->cols= cols;
  for ( c= 0; c < cols; ++c )
    {
      p= B2R ( fea->v[c] );
      for ( r= 0; r < rows; ++r )
	{
	  if ( !tL_gtoken ( f, buffer ) )
	    {
	      tL_msgerror ( err, _( "Unable to read the %dth element"
			    " of the %dth feature vector" ) , r+1, c+1 );
	      goto error;
	    }
	  p[r]= atof ( buffer->v );
	}
    }
  
  tL_buffer_free ( buffer );
  
  return 0;
  
 error:
  tL_buffer_free ( buffer );
  return -1;
  
} /* end load_real_fea_text */


static int
load_real_fea (
	       tLFea  *fea,
	       gzFile  f,
	       char  **err
	       )
{
  
  char fmt[8];
  
  
  /* Llig capçalera. */
  if ( gzread ( f, fmt, 8 ) != 8 )
    {
      tL_msgerror ( err, _( "Unable to read the file identifier" ) );
      return -1;
    }
  if ( !strncmp ( fmt, "AKREALBF" , 8 ) )
    return load_real_fea_bin ( fea, f, err );
  else if ( !strncmp ( fmt,  "AKREALTF" , 8 ) )
    return load_real_fea_text ( fea, f, err );
  else
    {
      tL_msgerror ( err, _( "Invalid format" ) );
      return -1;
    }
  
  return 0; /* CALLAT!!!. */
  
} /* end load_real_fea */


static void
print_real_fea (
		tLFea *fea,
		FILE  *to
		)
{
  
  int x, y;
  const float *p;
  
  
  for ( x= 0; x < fea->cols; ++x )
    {
      fprintf ( to, "%d:", x );
      p= B2cR ( fea->v[x] );
      for ( y= 0; y < fea->rows; ++y )
	fprintf ( to, "  %0.4g", p[y] );
      fputc ( '\n', to );
    }
  
} /* end print_real_fea */

static int
write_pbm (
	   const tLFea   *fea,
	   FILE          *to,
	   const tLBool   binary,
	   char         **err
	   )
{
  
  char byte;
  int r, c, x;
  bit **p;
  
  
  p= (bit **) fea->v;
  if ( binary )
    {
      fprintf ( to, "P4\n" );
      fprintf ( to, "%d %d\n", fea->cols, fea->rows );
      byte= 0;
      for ( r= fea->rows-1; r >= 0; --r )
	{
	  for ( c= x= 0; c < fea->cols; ++c )
	    {
	      byte= (byte<<1)|p[c][r];
	      if ( ++x == 8 )
		{
		  if ( fwrite ( &byte, 1, 1, to ) != 1 )
		    goto error;
		  x= 0;
		}
	    }
	  if ( x != 0 )
	    {
	      byte<<= 8-x;
	      if ( fwrite ( &byte, 1, 1, to ) != 1 )
		goto error;
	    }
	}
    }
  else
    {
      fprintf ( to, "P1\n" );
      fprintf ( to, "%d %d\n", fea->cols, fea->rows );
      for ( x= 0, r= fea->rows-1; r >= 0; --r )
	{
	  for ( c= 0; c < fea->cols; ++c )
	    {
	      fprintf ( to, "%d ", p[c][r] );
	      if ( ++x == 35 )
		{
		  putc ( '\n', to );
		  x= 0;
		}
	    }
	}
      if ( x != 0 ) putc ( '\n', to );
    }
  
  return 0;
  
 error:
  tL_msgerror ( err, _( "Unable to write the data" ) );
  return -1;
  
} /* end write_pbm */


static int
write_real_fea (
		const tLFea   *fea,
		FILE          *to,
		const tLBool   binary,
		char         **err
		)
{
  
  int c, r, aux[2];
  size_t nelements;
  const float *p;
  
  
  if ( binary )
    {
      if ( fwrite ( "AKREALBF", 1, 8, to ) != 8 ) goto error;
      aux[0]= fea->rows; aux[1]= fea->cols;
      if ( fwrite ( aux, sizeof(int), 2, to ) != 2 ) goto error;
      nelements= ((size_t)fea->rows)*((size_t)fea->cols);
      if ( fwrite ( fea->v[0], sizeof(float), nelements, to ) != nelements )
        goto error;
    }
  else
    {
      fprintf ( to, "AKREALTF\n" );
      fprintf ( to, "%d %d\n", fea->rows, fea->cols );
      for ( c= 0; c < fea->cols; ++c )
        {
          p= (const float *) ( fea->v[c] );
          fprintf ( to, "%g", p[0] );
          for ( r= 1; r < fea->rows; ++r )
            fprintf ( to, " %g", p[r] );
          putc ( '\n', to );
        }
    }
  
  return 0;
  
 error:
  tL_msgerror ( err, _( "Unable to write the data" ) );
  return -1;
  
} /* end write_real_fea */




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(tLFea *)
tL_fea_new (
	    const tLFeaType type
	    )
{
  
  tLFea *new;
  
  
  mt ( new= mem_alloc ( tLFea, 1 ) );
  mt ( new->v= mem_alloc ( tLByte *, 1 ) );
  mt ( new->v[0]= mem_alloc ( tLByte, 1 ) );
  new->_sv= 1;
  new->_s= 1;
  new->_buffer= tL_buffer_new ();
  new->_aux= NULL;
  new->_sparse= NULL;
  
  switch ( type )
    {
    case TL_FEA_BINARY:
      new->_load= load_pbm;
      new->_print= print_pbm;
      new->_write= write_pbm;
      new->_sizee= sizeof(bit);
      new->_sparse= alloc_fea_sparse ();
      break;
    case TL_FEA_REAL:
      new->_load= load_real_fea;
      new->_print= print_real_fea;
      new->_write= write_real_fea;
      new->_sizee= sizeof(float);
      break;
    default:
      tL_error ( _( "Unknown sample type: %d\n" ) , type );
    }
  
  return new;
  
} /* end tL_fea_new */


TL_API_FUNC(void)
tL_fea_free (
	     tLFea *fea
	     )
{
  
  if ( fea->_sparse != NULL )
    free_fea_sparse ( fea->_sparse );
  if ( fea->_aux != NULL )
    free ( fea->_aux );
  tL_buffer_free ( fea->_buffer );
  free ( fea->v[0] );
  free ( fea->v );
  free ( fea );
  
} /* end tL_fea_free */


TL_API_FUNC(char **)
tL_fea_get_binary_data (
			tLFea *fea
			)
{
  
  fea->_sparse->updated= FALSE;
  
  return (char **) fea->v;
  
} /* end tL_fea_get_binary_data */


TL_API_FUNC(int)
tL_fea_load (
	     tLFea       *fea,
	     const char  *file_name,
	     char       **err
	     )
{
  
  gzFile f;
  int ret;
  
  
  f= gzopen ( file_name, "r" );
  if ( f == NULL )
    {
      tL_msgerror ( err, _( "File '%s' cannot be opened" ) , file_name );
      return -1;
    }
  ret= tL_fea_load_fd ( fea, f, err );
  gzclose ( f );
  
  return ret;
  
} /* end tL_fea_load */


TL_API_FUNC(void)
tL_fea_resize (
	       tLFea     *fea,
	       const int  dim,
	       const int  nvecs
	       )
{
  
  tLassert ( tL_fea_resize, dim == 0 || nvecs == 0 );
  resize_fea ( fea, dim, nvecs );
  fea->rows= dim;
  fea->cols= nvecs;
  if ( fea->_sparse != NULL )
    fea->_sparse->updated= FALSE;
  
} /* end tL_fea_resize */


TL_API_FUNC(int)
tL_fea_write (
	      const tLFea   *fea,
	      const char    *file_name,
	      const tLBool   binary,
	      char         **err
	      )
{
  
  FILE *f;
  int ret;
  
  
  f= fopen ( file_name, binary?"wb":"w" );
  if ( f == NULL )
    {
      tL_msgerror ( err, _( "File '%s' cannot be opened for writing" ) ,
		    file_name );
      return -1;
    }
  ret= tL_fea_write_fd ( fea, f, binary, err );
  fclose ( f );
  
  return ret;
  
} /* end tL_fea_write */


TL_API_FUNC(void)
tL_seqs_adjust (
		tLSeqs *seqs
		)
{
  
  if ( seqs->size == 0 || seqs->size == seqs->N || seqs->N == 0 ) return;
  mt ( seqs->seqs= mem_realloc ( size_t, seqs->seqs, seqs->N ) );
  seqs->size= 0;
  
} /* end tL_seqs_adjust */


TL_API_FUNC(size_t)
tL_seqs_append (
		tLSeqs       *seqs,
		const tLSeq *cseq
		)
{
  
  size_t aux;
  
  
  if ( seqs->size == 0 )
    {
      for ( aux= 1; aux <= seqs->N; aux<<= 1 );
      seqs->size= aux;
      mt ( seqs->seqs= mem_realloc ( size_t, seqs->seqs, aux ) );
    }
  else if ( seqs->N == seqs->size )
    {
      seqs->size*= 2;
      mt ( seqs->seqs= mem_realloc ( size_t, seqs->seqs, seqs->size ) );
    }
  tL_dict_insert ( seqs->cseqs, cseq, &aux );
  seqs->seqs[seqs->N++]= aux;
  
  return aux;
  
} /* end tL_seqs_append */


TL_API_FUNC(void)
tL_seqs_free (
	      tLSeqs *seqs
	      )
{
  
  tL_dict_free ( seqs->cseqs );
  free ( seqs->seqs );
  free ( seqs );
  
} /* tL_seqs_free */


TL_API_FUNC(tLSeqs *)
tL_seqs_load (
	      const tLFileList  *filelist,
	      const tLDict      *syms,
	      char             **err,
              tLTar             *tar
	      )
{
  
  tLSeqs *new;
  size_t n, cs_size, sym;
  gzFile f;
  tLBuffer *buffer;
  tLSeq cs;
  long file_size=0, file_read=0, read=0;
  
  
  /* Inicialitza. */
  mt ( new= mem_alloc ( tLSeqs, 1 ) );
  mt ( new->seqs= mem_alloc ( size_t, filelist->N ) );
  new->cseqs= tL_dict_cseqs_new ();
  new->N= filelist->N;
  new->size= 0;
  
  /* Carrega. */
  buffer= tL_buffer_new ();
  mt ( cs.v= mem_alloc ( size_t, 1 ) );
  cs_size= 1;
  for ( n= 0; n < filelist->N; ++n )
    {

      if ( tar == NULL ) f= gzopen ( filelist->files[n], "r" );
      else f= tL_tar_get_file ( tar, filelist->files[n], &file_size, err );
      if ( f == NULL )
	{
	  TLMSG_ORFILE ( err, filelist->files[n] );
	  goto error;
	}
      
      cs.length= 0;

      if ( tar == NULL )
        {
          while ( tL_gtoken ( f, buffer ) )
            {
              if ( cs.length == cs_size )
                {
                  cs_size<<= 1;
                  mt ( cs.v= mem_realloc ( size_t, cs.v, cs_size ) );
                }
              sym= tL_dict_find ( syms, buffer->v );
              if ( sym == syms->N )
                {
                  tL_msgerror ( err, _( "Out of vocabulary token '%s'" ) , buffer->v );
                  goto error;
                }
              cs.v[cs.length++]= sym;
            }
        }
      else /* reading tar, stop reading after file_size chars */
        {
          file_read = read = 0;
          while ( ( read = tL_gntoken ( f, buffer, file_size - file_read ) ) )
            {
              file_read += read;
              if ( cs.length == cs_size )
                {
                  cs_size<<= 1;
                  mt ( cs.v= mem_realloc ( size_t, cs.v, cs_size ) );
                }
              sym= tL_dict_find ( syms, buffer->v );
              if ( sym == syms->N )
                {
                  tL_msgerror ( err, _( "Out of vocabulary token '%s'" ) , buffer->v );
                  goto error;
                }
              cs.v[cs.length++]= sym;
            }
        }
      if ( cs.length == 0 )
	{
	  MSGEMPTY ( err, filelist->files[n] );
	  goto error;
	}
      
      tL_dict_insert ( new->cseqs, &cs, &(new->seqs[n]) );
      
      if ( tar == NULL ) { gzclose ( f ); f= NULL; }
      
    }
  free ( cs.v );
  tL_buffer_free ( buffer );
  
  return new;
  
 error:
  if ( f != NULL ) gzclose ( f );
  free ( cs.v );
  tL_buffer_free ( buffer );
  tL_seqs_free ( new );
  return NULL;
  
} /* end tL_seqs_load */


TL_API_FUNC(tLSeqs *)
tL_seqs_load_words (
		    const tLFileList  *filelist,
		    const tLDict      *syms,
		    char             **err,
                    tLTar             *tar
		    )
{
  
  tLSeqs *new;
  size_t n, cs_size;
  gzFile f;
  tLBuffer *buffer;
  tLSeq cs;
  long file_size=0, read=0;
  
  
  /* Inicialitza. */
  mt ( new= mem_alloc ( tLSeqs, 1 ) );
  mt ( new->seqs= mem_alloc ( size_t, filelist->N ) );
  new->cseqs= tL_dict_cseqs_new ();
  new->N= filelist->N;
  new->size= 0;
  
  /* Carrega. */
  buffer= tL_buffer_new ();
  mt ( cs.v= mem_alloc ( size_t, 1 ) );
  cs_size= 1;
  for ( n= 0; n < filelist->N; ++n )
    {
      if ( tar == NULL ) f= gzopen ( filelist->files[n], "r" );
      else  f= tL_tar_get_file ( tar, filelist->files[n], &file_size, err );
      if ( f == NULL )
	{
          TLMSG_ORFILE ( err, filelist->files[n] );
          goto error;
        }
      
      if ( tar == NULL ) read = tL_gtoken ( f, buffer );
      else read = tL_gntoken ( f, buffer, file_size );
      if ( read == 0 )
        {
          MSGEMPTY ( err, filelist->files[n] );
          goto error;
            }
      if ( tL_utf8tocseq ( buffer->v, syms, &cs, &cs_size, err ) == -1 )
	goto error;
      tL_dict_insert ( new->cseqs, &cs, &(new->seqs[n]) );
      
      gzclose ( f );
      
    }
  free ( cs.v );
  tL_buffer_free ( buffer );
  
  return new;
  
 error:
  if ( f != NULL ) gzclose ( f );
  free ( cs.v );
  tL_buffer_free ( buffer );
  tL_seqs_free ( new );
  return NULL;
  
} /* end tL_seqs_load_words */


TL_API_FUNC(tLSeqs *)
tL_seqs_new (void)
{
  
  tLSeqs *new;
  
  
  mt ( new= mem_alloc ( tLSeqs, 1 ) );
  mt ( new->seqs= mem_alloc ( size_t, 1 ) );
  new->cseqs= tL_dict_cseqs_new ();
  new->N= 0;
  new->size= 1;
  
  return new;
  
} /* end tL_seqs_new */


int **
tL_fea_get_sparse (
		   const tLFea *fea
		   )
{
  
  tLFeaSparse *sp;
  bit **a, *a_c;
  int **b, *b_c;
  int r, c;
  
  
  sp= fea->_sparse;
  if ( sp->updated ) return sp->v;
  resize_fea_sparse ( sp, fea->rows, fea->cols );
  b= sp->v;
  a= (bit **) (fea->v);
  for ( c= 0; c < fea->cols; ++c )
    {
      a_c= a[c];
      b_c= b[c];
      for ( r= 0; r < fea->rows; ++r )
	if ( a_c[r] ) *(b_c++)= r;
      *b_c= -1;
    }
  sp->updated= TRUE;
  
  return b;
  
} /* end tL_fea_get_sparse */
