/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  samples_public.h - Part pública de 'samples.h'.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 */

#ifndef __LIBTL_SAMPLES_H__
#define __LIBTL_SAMPLES_H__

#include <stddef.h>
#include <stdio.h>
#include <zlib.h>

#include "common_public.h"
#include "filelist.h"
#include "port.h"
#include "text_public.h"

/** \addtogroup Basic
 *  @{
 */

/* TAR */

/** Tar files. This structure is used to access tar files. **/
typedef struct
{
  
  gzFile  f;
  tLDict *files;
  long   *offsets;
  long   *sizes;
  
} tLTar;

/** Frees memory. Frees the memory allocated for an tLTar.
 *
 *  @param tar The tLTar structure.
 */
TL_API_FUNC(void)
tL_tar_free (
	     tLTar *tar
	     );

/** Returns a file descriptor to desired file. This function returns
 *  the file descriptor of the tar file, seeked to the beginning of
 *  the desired file. Therefore, the file descriptor must not be
 *  closed.
 *
 *  @param tar The tLTar structure.
 *  @param name The name of the desired file.
 *  @param size Pointer to a long variable. If not NULL the size of
 *  the file in bytes is returned.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The file descriptor to the desired file, or NULL in case
 *  of error. This descriptor must not be closed.
 */
TL_API_FUNC(gzFile)
tL_tar_get_file (
		 tLTar       *tar,
		 const char  *name,
                 long        *size,
		 char       **err
		 );

/** Creates a new tLTar. This functions creates a
 *  new tLTar for a given file.
 *
 *  @param file_name The name of the file.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The new feature vector manager.
 */
TL_API_FUNC(tLTar *)
tL_tar_new (
	    const char  *file_name,
	    char       **err
	    );


/* FEATURES */

/* Tipus de mostres soportats. */
/** Type of features. */
typedef enum
  {
    TL_FEA_BINARY,    /**< Binary feature. */
    TL_FEA_REAL       /**< Real feature. */
  } tLFeaType;

typedef unsigned char tLByte;

typedef struct
{
  
  int    **v;
  tLBool   updated;
  int      sv;
  size_t   s;
  
} tLFeaSparse;

/** Feature vectors. This structure is used to manage feature
 *  vectors. Once it is created, it can be used to load sequences of
 *  feature vectors. Currently, two types of features are supported:
 *  binary and real. When created, the type of feature must be
 *  specified.
 */
typedef struct fea
{
  
  tLByte      **v;     /* |COLS|*|ROWS|. */
  int           rows;
  int           cols;
  int           _sv;
  size_t        _s;
  tLBuffer     *_buffer;
  tLByte       *_aux;
  size_t        _auxs;
  size_t        _sizee;
  tLFeaSparse  *_sparse;
  int         (*_load) (struct fea  *fea,
			gzFile       f,
			char       **err);
  void        (*_print) (struct fea *fea,
			 FILE       *to);
  int        (*_write) (const struct fea  *fea,
			FILE              *to,
			const tLBool       binary,
			char             **err);
  
} tLFea;

/** Frees memory. Frees the memory allocated for the feature vector
 *  manager.
 *
 *  @param fea The feature vector manager.
 */
TL_API_FUNC(void)
tL_fea_free (
	     tLFea *fea
	     );

/** Gets a pointer to real data. This macro gets a pointer to the
 *  feature vector manager data. The data can be directly modified by
 *  accessing as a C matrix of floats [N][D], where N is the feature
 *  vector index, and D is the selected dimension. This macro is only
 *  valid for TL_FEA_REAL feature vector managers.
 *
 *  @param FEA The feature vector manager.
 *  @return A pointer (float **) to the real data.
 */
#define tL_fea_get_real_data(FEA) ((float **) (FEA)->v)

/** Gets a pointer to binary data. This function gets a pointer to the
 *  feature vector manager data. The data can be directly modified by
 *  accessing as a C matrix of chars [N][D], where N is the feature
 *  vector index, and D is the selected dimension. This macro is only
 *  valid for TL_FEA_BINARY feature vector managers. The only allowed
 *  values are 0 and 1, other values could produce undesirable
 *  behaviour. Once the feature vector manager has been used, this
 *  function must be called again if we want to modify the values
 *  again.
 *
 *  @param fea The feature vector manager.
 *  @return A pointer to the binary data.
 */
TL_API_FUNC(char **)
tL_fea_get_binary_data (
			tLFea *fea
			);

/** Loads a feature vector sequence. This function does the same as
 *  tL_fea_load_fd. The difference is that in this function the name
 *  of the file is provided instead of the file descriptor.
 *
 *  @param fea The feature vector manager.
 *  @param file_name The name of the file where a feature vector
 *  sequence is stored.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return -1 in case of error, 0 otherwise.
 */
TL_API_FUNC(int)
tL_fea_load (
	     tLFea       *fea,
	     const char  *file_name,
	     char       **err
	     );

/* Carrega la mostra en FEA. */
/** Loads a feature vector sequence. This macro loads into the feature
 *  vector manager a feature vector sequence stored in the given
 *  file. The feature vector sequence must be of the same type as
 *  the feature vector manager. Binary feature vector sequences are
 *  expected to be in PBM image, where each column is a binary feature
 *  vector.
 *
 *  @param FEA The feature vector manager.
 *  @param FD The file where a feature vector sequence is stored.
 *  @param ERR Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return -1 in case of error, 0 otherwise.
 */
#define tL_fea_load_fd(FEA,FD,ERR) (FEA)->_load ( (FEA), (FD), (ERR) )

/** Creates a new feature vector manager. This function creates a
 *  new manager for feature vectors of the provided type.
 *
 *  @param type The type of the feature vectors.
 *  @return The new feature vector manager.
 */
TL_API_FUNC(tLFea *)
tL_fea_new (
	    const tLFeaType type
	    );

/** Prints the content. This macro prints the content in text
 *  representation. The feature vector manager must be loaded,
 *  otherwise an unexpected error can happen.
 *
 *  @param FEA The feature vector manager.
 *  @param TO File to which the content is written.
 */
#define tL_fea_print(FEA,TO) (FEA)->_print ( (FEA), (TO) )

/** Resizes memory. This function resizes the memory used by the
 *  manager to store the sample data. If the current capacity is
 *  enough, this function does nothing.
 *
 *  @param fea The feature vector manager.
 *  @param dim The required feature vector dimension.
 *  @param nvecs The required number of feature vectors.
 */
TL_API_FUNC(void)
tL_fea_resize (
	       tLFea     *fea,
	       const int  dim,
	       const int  nvecs
	       );

/** Writes the content. This macro writes the content to a file. The
 *  feature vector manager must be loaded, otherwise an unexpected
 *  error can happen.
 *
 *  @param fea The feature vector manager.
 *  @param file_name Name of the file to which content must be
 *  written.
 *  @param binary Specifies whether the content must be written using
 *  binary or text representation.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return -1 in case of error, 0 otherwise.
 */
TL_API_FUNC(int)
tL_fea_write (
	      const tLFea   *fea,
	      const char    *file_name,
	      const tLBool   binary,
	      char         **err
	      );

/** Writes the content. This macro writes the content to a file
 *  descriptor. The feature vector manager must be loaded, otherwise
 *  an unexpected error can happen.
 *
 *  @param FEA The feature vector manager.
 *  @param TO File to which the content is written.
 *  @param BINARY Specifies whether the content must be written using
 *  binary or text representation.
 *  @param ERR Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return -1 in case of error, 0 otherwise.
 */
#define tL_fea_write_fd(FEA,TO,BINARY,ERR)		\
  (FEA)->_write ( (FEA), (TO), (BINARY), (ERR) )

/* SEQÜÈNCIES */

/** Symbol sequence manager. This structure is used to manage sets
 *  of symbol sequences. Symbol sequences are encoded into
 *  integers, thus, repeated symbol sequences are encoded using the
 *  same number.
 */
typedef struct
{
  
  size_t  N;       /**< Number of symbol sequences, including the
		      repeated ones. */
  size_t *seqs;    /**< Encoded symbol sequences. */
  tLDict *cseqs;
  size_t  size;    /* Capacitat. 0 vol dir que la capacitat és N. */
  
} tLSeqs;

/** Adjusts memory. This function adjusts the memory related to the
 *  field 'seqs'.
 *
 *  @param seqs The symbol sequence manager.
 */
TL_API_FUNC(void)
tL_seqs_adjust (
		tLSeqs *seqs
		);

/** Appends a new symbol sequence. This function appends a new
 *  symbol sequence to the symbol sequence manager (the
 *  symbol sequence may be repeated) and returns the identifier of
 *  the inserted symbol sequence.
 *
 *  @param seqs The symbol sequence manager.
 *  @param seq The new symbol sequence.
 *  @return The identifier of the inserted symbol sequence.
 */
TL_API_FUNC(size_t)
tL_seqs_append (
		tLSeqs      *seqs,
		const tLSeq *seq
		);

/** Frees memory. Frees the memory allocated for the symbol
 *  sequence manager.
 *
 *  @param seqs The symbol sequence manager.
 */
TL_API_FUNC(void)
tL_seqs_free (
	      tLSeqs *seqs
	      );

/** Gets the number of different symbol sequences. This macro
 *  returns the number of different symbol sequences encoded in the
 *  symbol sequence manager. Symbol sequences are always encoded
 *  as consecutive numbers starting from 0.
 *
 *  @param SEQS The symbol sequence manager.
 *  @return The number of different symbol sequences in the
 *  manager.
 */
#define tL_seqs_get_num_seqs(SEQS) ((SEQS)->cseqs->N)

/** Gets a symbol sequence. This macro returns, from the provided
 *  symbol sequence manager, a reference to the symbol sequence
 *  encoded as ID.
 *
 *  @param SEQS The symbol sequence manager.
 *  @param ID An integer encoding a symbol sequence.
 *  @return A reference to the required symbol sequence.
 */
#define tL_seqs_get_seq(SEQS,ID)		\
  ((const tLSeq *) (SEQS)->cseqs->data[(ID)])

/* Llig les seqüències de la llista de mostres, fent ús del
 * vocabulari. En cas de que hi hasca un caràcter fora de vocabulari
 * genera un tL_error.
 */
/** Creates a new symbol sequence manager from files. This function
 *  creates a new sequence manager and loads all the symbol
 *  sequences from the provided tLFileList. Each file is assumed to
 *  contain a symbol sequence encoded as a sequence of tokens
 *  separated by blank characters. Sequences are loaded respecting the
 *  original order in the tLFileList. The provided symbol
 *  dictionary is used to encode the symbols.
 *
 *  @param filelist List of the files containing the symbol
 *  sequences.
 *  @param syms The dictionary with the symbols.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @param tar information of a tar file from which samples will be read
 *  @return The new symbol sequence manager, or NULL in case of
 *  error.
 */
TL_API_FUNC(tLSeqs *)
tL_seqs_load (
	      const tLFileList  *filelist,
	      const tLDict      *syms,
	      char             **err,
              tLTar             *tar
	      );

/* Llig les paraules de la llista de mostres, i ompli el diccionari de
 * seqüències codificades, fent ús del vocabulari. En cas de que hi
 * hasca un caràcter fora de vocabulari genera un tL_error. Es suposa
 * que cada paraula està escrita en UTF-8 i cada caràcter és un
 * símbol.
 */
/** Creates a new symbol sequence manager from files containing
 *  words. This function works as tL_seqs_load, but the symbol
 *  sequences are assumed to be words encoded using UTF-8.
 *
 *  @param filelist List of the files containing the symbol
 *  sequences.
 *  @param syms The dictionary with the symbols (UTF-8 characters).
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @param tar information of a tar file from which samples will be read
 *  @return The new symbol sequence manager, or NULL in case of
 *  error.
 */
TL_API_FUNC(tLSeqs *)
tL_seqs_load_words (
		    const tLFileList  *filelist,
		    const tLDict      *syms,
		    char             **err,
                    tLTar             *tar
		    );

/** Creates a new symbol sequence manager. This function creates a
 *  new empty symbol sequence manager.
 *
 *  @return The new symbol sequence manager.
 */
TL_API_FUNC(tLSeqs *)
tL_seqs_new (void);

/** @} */

#endif /* __LIBTL_SAMPLES_H__ */
