/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  statestr.h - Serveix per a entrenar els estats.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011.
 *
 */

#ifndef __TL_PRV_STATESTR_H__
#define __TL_PRV_STATESTR_H__

#include <stddef.h>
#include <stdlib.h>

#include "amodel.h"
#include "common.h"
#include "samples.h"


/* TIPUS STATESTR */

typedef struct statestr statestr_t;

typedef struct
{
  
  /* Allibera la memòria. Pero el punter. */
  void (*free) (statestr_t *statestr);
  
  /* Prepara l'entrenador d'estats per a la mostra indicada. */
  void (*get_ready) (statestr_t  *statestr,
		     const tLFea *fea,
		     void        *global);
  
  /* Realitza un pseudo pas E en la funció de probabilitat d'un
   * estat. Es suposa que 'maux' té la memòria necessària.  Els
   * 'frames' no actius tornen probabilitat 0.
   */
  void (*train_state_pseudoE) (void *,
                               const tLFea *,
                               tLBool *,
                               void *,
                               void *,
                               unsigned int,
                               void *,
                               tLProb *);
  
  /* Aquesta funció es crida abans de començar el pas pseudo E. Es
   * supossa que 'maux' té la memòria necessària.
   */
  void (*train_global_pseudoE) (void *,
				const tLFea *,
				void *);
  
  /* Realitza un pseudo M en la funció de probabilitat d'un estat. Es
   * suposa que els vectors de càlcul de 'maux' tenen almenys espai
   * per a N+1 valors.
   */
  void (*train_state_pseudoM) (void *,
                               int,
                               tLProb *,
                               void *,
                               void *,
                               void *,
                               unsigned int,
                               void *);
  
  /* Aquesta funció es crida després de realitzar el pas pseudo M en
   * cada estat. Es suposa que els vectors de càlcul de 'maux' tenen
   * almenys espai per a N+1 valors.
   */
  void (*train_global_pseudoM) (void *,
				void *,
				int,
				void *);
  
  /* Es crida abans de començar el pas M. */
  void (*train_init_pseudoM) (void *,
			      int,
			      void *);
  
} strm_t;

/* 'statestr' és una estructura per a entrenar els estats d'un model
 * de caràcters. Cada 'statestr' te un número concret de 'statetr's i
 * un estat auxiliar comú a tots ells. Cada 'statetr' serveix per a
 * entrenar un estat, poguent ser utilizat per a entrenar qualsevol
 * estat per al que es va crear 'statestr'.
 */
struct statestr
{
  
  size_t        N;    /* Nombre de 'statetr's. */
  void         *statetrs;
  void         *maux;
  const strm_t *meths;
  size_t        _statetr_size;
  
};

#define free_statestr(STATESTR)                 \
    do                                          \
      {                                         \
        (STATESTR)->meths->free ( (STATESTR) ); \
        free ( (STATESTR) );                    \
      } while(0)

#define statestr_get_ready(STATESTR,FEA,GLOBAL)		       \
  (STATESTR)->meths->get_ready ( (STATESTR), (FEA), (GLOBAL) )

#define statestr_sizeof(STATESTR,Q)   \
    ((STATESTR)->_statetr_size * (Q))

#define statestr_pseudoE(STATESTR,GLOBAL,FEA,ACTIVES,STATES,STATETRS,Q,PROBS) \
    (STATESTR)->meths->train_state_pseudoE ( (GLOBAL), (FEA), (ACTIVES),      \
                                             (STATES), (STATETRS), (Q),       \
                                             (STATESTR)->maux, (PROBS) )

#define statestr_pseudoE_global(STATESTR,GLOBAL,FEA)             \
    (STATESTR)->meths->train_global_pseudoE ( (GLOBAL), (FEA),   \
                                              (STATESTR)->maux )

#define statestr_pseudoM(STATESTR,GLOBAL,N,WEIGHTS,STATES,CSTATES,STATETRS,Q) \
    (STATESTR)->meths->train_state_pseudoM ( (GLOBAL), (N), (WEIGHTS),        \
                                             (STATES), (CSTATES), (STATETRS), \
                                             (Q), (STATESTR)->maux )

#define statestr_pseudoM_global(STATESTR,GLOBAL,CGLOBAL,N)	        \
    (STATESTR)->meths->train_global_pseudoM ( (GLOBAL), (CGLOBAL), (N),	\
                                              (STATESTR)->maux )

#define statestr_pseudoM_init(STATESTR,GLOBAL,N)               \
    (STATESTR)->meths->train_init_pseudoM ( (GLOBAL), (N),     \
                                            (STATESTR)->maux )


/* TIPUS STATESTR_FACTORY */

typedef struct str_factory str_factory_t;

typedef struct
{
  
  /* Allibera memòria. */
  void (*free) (str_factory_t *strf);
  
  /* Inidica que també ha de tractar amb els
   * següents estats.
   */
  void (*add) (str_factory_t      *strf,
	       void               *states,
	       const unsigned int  Q);
  
  /* Inicialitza 'statestr' amb l'sepai
   * indicat.
   */
  void (*init_statestr) (str_factory_t *strf,
			 statestr_t    *str,
			 void          *global,
			 const size_t   N);
  
} strfm_t;

/* 'str_factory' s'utilizat per a crear
 *  'statestr'. L'ús típic és crearlo indicar
 *  mitjançant el mètode 'add' (repetidament
 *  si és necessari) amb quins estats ha de tractar,
 *  i finalment amb 'make' es crea un 'statestr'
 *  amb l'espai indicat.
 */
#define STR_FACTORY       \
    const strfm_t *meths;

struct str_factory
{
  
  STR_FACTORY
  
};

str_factory_t *
tL_create_str_factory (
		       const char *type,
		       void       *global
		       );

#define free_str_factory(STR_FACTORY)            \
    (STR_FACTORY)->meths->free ( (STR_FACTORY) )

#define str_factory_add(STR_FACTORY,STATES,Q)  \
    (STR_FACTORY)->meths->add ( (STR_FACTORY), \
                                (STATES), (Q) )

statestr_t *
tL_str_factory_make (
		     str_factory_t *strf,
		     void          *global,
		     const size_t   N
		     );


/* TIPUS STATESTRV */

typedef struct statestrv statestrv_t;

typedef struct
{
  
  /* Allibera la memòria. Pero el punter. */
  void (*free) (statestrv_t *statestrv);
  
  /* Prepara l'entrenador d'estats per a la mostra indicada. */
  void (*get_ready) (statestrv_t *statestrv,
		     const tLFea *fea,
		     void        *global);
  
  /* Torna la probabilitat de l'estat per a cada frame actiu,
   * enmagatzemant la informació que crega necessària per a més tart
   * acumular els comptadors si és requerit. Es suposa que 'maux' té
   * la memòria necessària. Els 'frames' no actius tornen probabilitat
   * 0.
   */
  void (*train_state_get_probs) (void         *global,
				 const tLFea  *fea,
				 tLBool       *actives,
				 void         *states,
				 void         *statetrs,
				 unsigned int  Q,
				 void         *maux,
				 tLProb       *probs);
  
  /* Acumula el frame 't' en l'estat 'q'. Es supossa que 'maux' té la
     memòria necessària. */
  void (*train_state_accum) (void         *global,
			     const int     t,
			     unsigned int  q,
			     void         *states,
			     void         *cstates,
			     void         *statetrs,
			     void         *maux);
  
} strvm_t;

/* 'statestrv' és una estructura per a entrenar els estats d'un model
 * de caràcters a la viterbi. Cada 'statestr' te un número concret de
 * 'statetr's i un estat auxiliar comú a tots ells. Cada 'statetr'
 * serveix per a entrenar un estat, poguent ser utilizat per a
 * entrenar qualsevol estat per al que es va crear 'statestr'.
 */
struct statestrv
{
  
  size_t         N;    /* Nombre de 'statetrv's. */
  void          *statetrs;
  void          *maux;
  const strvm_t *meths;
  size_t         _statetr_size;
  
};

#define free_statestrv(STATESTRV)                 \
  do						  \
    {						  \
      (STATESTRV)->meths->free ( (STATESTRV) );	  \
      free ( (STATESTRV) );			  \
    } while(0)

#define statestrv_get_ready(STATESTRV,FEA,GLOBAL)		\
  (STATESTRV)->meths->get_ready ( (STATESTRV), (FEA), (GLOBAL) )

#define statestrv_sizeof(STATESTRV,Q)   \
    ((STATESTRV)->_statetr_size * (Q))

#define statestrv_get_probs(STATESTRV,GLOBAL,FEA,ACTIVES,		  \
			    STATES,STATETRS,Q,PROBS)			  \
  (STATESTRV)->meths->train_state_get_probs ( (GLOBAL), (FEA), (ACTIVES), \
					      (STATES), (STATETRS), (Q),  \
					      (STATESTRV)->maux, (PROBS) )

#define statestrv_accum(STATESTRV,GLOBAL,T,Q,STATES,CSTATES,STATETRS)	\
  (STATESTRV)->meths->train_state_accum ( (GLOBAL), (T), (Q), (STATES), \
					  (CSTATES), (STATETRS),	\
					  (STATESTRV)->maux )


/* TIPUS STATESTRV_FACTORY */

typedef struct strv_factory strv_factory_t;

typedef struct
{
  
  /* Allibera memòria. */
  void (*free) (strv_factory_t *strvf);
  
  /* Inidica que també ha de tractar amb els
   * següents estats.
   */
  void (*add) (strv_factory_t     *strvf,
	       void               *states,
	       const unsigned int  Q);
  
  /* Inicialitza 'statestrv' amb l'sepai
   * indicat.
   */
  void (*init_statestrv) (strv_factory_t *strvf,
			  statestrv_t    *str,
			  void           *global,
			  const size_t    N);
  
} strvfm_t;

/* 'strv_factory' s'utilizat per a crear
 *  'statestrv'. L'ús típic és crearlo indicar
 *  mitjançant el mètode 'add' (repetidament
 *  si és necessari) amb quins estats ha de tractar,
 *  i finalment amb 'make' es crea un 'statestr'
 *  amb l'espai indicat.
 */
#define STRV_FACTORY       \
    const strvfm_t *meths;

struct strv_factory
{
  
  STRV_FACTORY
  
};

strv_factory_t *
tL_create_strv_factory (
			const char *type,
			void       *global
			);

#define free_strv_factory(STRV_FACTORY)            \
    (STRV_FACTORY)->meths->free ( (STRV_FACTORY) )

#define strv_factory_add(STRV_FACTORY,STATES,Q)  \
    (STRV_FACTORY)->meths->add ( (STRV_FACTORY), \
				 (STATES), (Q) )

statestrv_t *
tL_strv_factory_make (
		      strv_factory_t *strf,
		      void           *global,
		      const size_t    N
		      );


#endif /* __TL_PRV_STATESTR_H__ */
