/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  bernoulli.c - Implementació de 'bernoulli.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#include "bernoulli.h"
#include "calc.h"
#include "amcounts/ibernoulli.h"
#include "amcounts/logbernoulli.h"
#include "amodel/bernoulli.h"
#include "common.h"
#include "empty.h"
#include "error.h"
#include "mem.h"
#include "samples.h"
#include "statestr.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) BERNOULLI_GLOBAL(ptr)
#define STATE(ptr) BERNOULLI_STATE(ptr)
#define CSTATE(ptr) ((logamcount_bernoulli_state_t *) (ptr))
#define CSTATE2(ptr) ((iamcount_bernoulli_state_t *) (ptr))
#define MAUX(ptr) ((Bernoulli_maux_t *) (ptr))
#define STRF(ptr) ((str_factory_t *) (ptr))
#define TRS(ptr) ((statetr_t *) (ptr))
#define STRVF(ptr) ((strv_factory_t *) (ptr))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  int   size;      /* Espai disponible. */
  bit **frames;    /* Els frames actius. */
  
} statetr_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static tLFloat
calc_prob (
	   unsigned int       D,
	   bernoulli_state_t  state,
	   const tLFea       *fea,
	   const int          t
	   )
{
  
  tLFloat prob;
  unsigned int d;
  const bit *frame;
  const int *i;
  
  
  prob= 0.0;
  if ( state.zeros )
    {
      frame= B2cb(fea->v[t]);
      for ( d= 0; d < D; ++d )
	prob+=
	  frame[d] ?
	  state.p[d].p1 :
	  state.p[d].p0;
    }
  else
    {
      for ( i= tL_fea_get_sparse ( fea )[t]; *i != -1; ++i )
	prob+= state.l[*i];
      prob+= state.logc;
    }
  
  return prob;
  
} /* end calc_prob */


static void
resize_maux_calc (
		  Bernoulli_maux_t *maux,
		  unsigned int      size
		  )
{

  size_t size_aux;


  size_aux= size; size_aux<<= 1;
  mt ( maux->calc1=
       mem_realloc ( tLFloat, maux->calc1, size_aux ) );
  maux->calc0= maux->calc1 + size;

  maux->calc_size= size;

} /* end resize_maux_calc */




/**********************/
/* STATESTR (MÈTODES) */
/**********************/

static void
free_strvm (
	    statestrv_t *statestrv
	    )
{
  
  size_t n;
  
  
  for ( n= 0; n < statestrv->N; ++n )
    free ( TRS(statestrv->statetrs)[n].frames );
  free ( statestrv->statetrs );
  
} /* end free_strvm */


static void
free_strm (
           statestr_t *statestr
           )
{
  
  size_t n;
  
  
  for ( n= 0; n < statestr->N; ++n )
    free ( TRS(statestr->statetrs)[n].frames );
  free ( statestr->statetrs );
  tL_Bernoulli_maux_free ( MAUX(statestr->maux) );
  
} /* end free_strm */


static void
train_state_pseudoE (
		     void         *global,
                     const tLFea  *fea,
                     tLBool       *actives,
                     void         *states,
                     void         *statetrs,
                     unsigned int  q,
                     void         *maux,
                     tLProb       *probs
                     )
{
  
  bernoulli_state_t state;
  statetr_t *statetr;
  int n;
  unsigned int D;
  tLFloat prob;
  
  
  state= STATE(states)[q];
  statetr= &(TRS(statetrs)[q]);
  D= GLOBAL(global)->D;
  
  if ( fea->cols > statetr->size )
    {
      mt ( statetr->frames=
	   mem_realloc ( bit *,
			 statetr->frames,
			 fea->cols ) );
      statetr->size= fea->cols;
    }
  
  for ( n= 0; n < fea->cols; ++n )
    if ( actives[n] )
      {
	prob= calc_prob ( D, state, fea, n );
	if ( prob == -HUGE_VAL )
	  {
	    probs[n]= -HUGE_VAL;
	    statetr->frames[n]= NULL;
	  }
	else
	  {
	    probs[n]= prob;
	    statetr->frames[n]= B2b(fea->v[n]);
	  }
      }
    else
      {
	probs[n]= -HUGE_VAL;
	statetr->frames[n]= NULL;
      }
  
} /* end train_state_pseudoE */


static void
train_state_pseudoM (
		     void         *global,
                     int           N,
                     tLProb       *weights,
                     void         *states,
                     void         *cstates,
                     void         *statetrs,
                     unsigned int  q,
                     void         *maux
                     )
{
  
  logamcount_bernoulli_state_t scount;
  statetr_t statetr;
  int n, N1, N0;
  unsigned int d, D;
  tLFloat max1, max0, aux;
  tLFloat *calc1, *calc0;
  bit *frame;
  volatile tLFloat fix;
  
  
  D= GLOBAL(global)->D;
  scount= CSTATE(cstates)[q];
  statetr= TRS(statetrs)[q];
  
  calc0= MAUX(maux)->calc0; calc1= MAUX(maux)->calc1;
  
  for ( d= 0; d < D; ++d )
    {
      
      max1= calc1[0]= scount.p[d].p1;
      max0= calc0[0]= scount.p[d].p0;
      N1= N0= 1;
      
      for ( n= 0; n < N; ++n )
	if ( (frame= statetr.frames[n]) != NULL &&
	     weights[n] != -HUGE_VAL )
	  {
	    if ( frame[d] )
	      {
		calc1[N1++]= aux= fix= weights[n];
		if ( aux > max1 ) max1= aux;
	      }
	    else
	      {
		calc0[N0++]= aux= fix= weights[n];
		if ( aux > max0 ) max0= aux;
	      }
	  }
      
      if ( N1 > 1 )
	scount.p[d].p1= tL_radd_vals ( calc1, max1, N1 );
      
      if ( N0 > 1 )
	scount.p[d].p0= tL_radd_vals ( calc0, max0, N0 );
      
    }
  
} /* end train_state_pseudoM */


static void
train_state_accum (
		   void         *global,
		   const int     t,
		   unsigned int  q,
		   void         *states,
		   void         *cstates,
		   void         *statetrs,
		   void         *maux
		   )
{
  
  iamcount_bernoulli_state_t scount;
  const bit *frame;
  unsigned int d, D;
  
  
  D= GLOBAL(global)->D;
  scount= CSTATE2(cstates)[q];
  frame= TRS(statetrs)[q].frames[t];
  for ( d= 0; d < D; ++d )
    if ( frame[d] ) ++scount.p[d].p1;
    else            ++scount.p[d].p0;
  
} /* end train_state_accum */


static const strm_t strm=
  {
    free_strm,
    tL_Bernoulli_get_ready,
    train_state_pseudoE,
    tL_empty_statestr_train_global_pseudoE,
    train_state_pseudoM,
    tL_empty_statestr_train_global_pseudoM,
    tL_empty_statestr_train_init_pseudoM
  };


static const strvm_t strvm=
  {
    free_strvm,
    tL_empty_statestrv_get_ready,
    train_state_pseudoE,
    train_state_accum
  };




/*************************/
/* STR_FACTORY (MÈTODES) */
/*************************/

static void
init_statestr (
               str_factory_t *strf,
               statestr_t    *str,
	       void          *global,
               const size_t   N
               )
{
  
  size_t n;
  statetr_t *aux;
  
  
  str->N= N;
  mt ( aux= mem_alloc ( statetr_t, N ) );
  for ( n= 0; n < N; ++n )
    {
      aux[n].size= 1;
      mt ( aux[n].frames= mem_alloc ( bit *, 1 ) );
    }
  str->statetrs= (void *) aux;
  str->maux= (void *) tL_Bernoulli_maux_create ( 1 );
  str->meths= &strm;
  str->_statetr_size= sizeof(statetr_t);
  
} /* end init_statestr */


static void
init_statestrv (
		strv_factory_t *strf,
		statestrv_t    *str,
		void           *global,
		const size_t    N
		)
{
  
  size_t n;
  statetr_t *aux;
  
  
  str->N= N;
  mt ( aux= mem_alloc ( statetr_t, N ) );
  for ( n= 0; n < N; ++n )
    {
      aux[n].size= 1;
      mt ( aux[n].frames= mem_alloc ( bit *, 1 ) );
    }
  str->statetrs= (void *) aux;
  str->maux= NULL;
  str->meths= &strvm;
  str->_statetr_size= sizeof(statetr_t);
  
} /* end init_statestrv */


static const strfm_t strfm=
  {
    tL_empty_free_strf,
    tL_empty_strf_add,
    init_statestr
  };


static const strvfm_t strvfm=
  {
    tL_empty_free_strvf,
    tL_empty_strvf_add,
    init_statestrv
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

void
tL_Bernoulli_get_ready (
			statestr_t  *statestr,
			const tLFea *fea,
			void        *global
			)
{
  
  if ( fea->cols >= MAUX(statestr->maux)->calc_size )
    resize_maux_calc ( MAUX(statestr->maux),
		       fea->cols+1 );
  
} /* end tL_Bernoulli_get_ready */


Bernoulli_maux_t *
tL_Bernoulli_maux_create (
			  const unsigned int size
			  )
{
  
  Bernoulli_maux_t *new;
  
  
  mt ( new= mem_alloc ( Bernoulli_maux_t, 1 ) );
  mt ( new->calc1= mem_alloc ( tLFloat, size<<1 ) );
  new->calc0= new->calc1 + size;
  new->calc_size= size;
  
  return new;
  
} /* end tL_Bernoulli_maux_create */


void
tL_Bernoulli_maux_free (
			Bernoulli_maux_t *maux
			)
{

  free ( maux->calc1 );
  free ( maux );

} /* end tL_Bernoulli_maux_free */


str_factory_t *
tL_create_strf_bernoulli (
			  void *global
			  )
{

  str_factory_t *new;


  mt ( new= mem_alloc ( str_factory_t, 1 ) );
  new->meths= &strfm;
  
  return new;

} /* end tL_create_strf_bernoulli */


strv_factory_t *
tL_create_strvf_bernoulli (
			   void *global
			   )
{

  strv_factory_t *new;
  
  
  mt ( new= mem_alloc ( strv_factory_t, 1 ) );
  new->meths= &strvfm;
  
  return new;
  
} /* end tL_create_strvf_bernoulli */
