/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  dgaussian.c - Implementació de 'dgaussian.h'.
 *
 *  Adrià Giménez Pastor, 2009,2011,2012,2013.
 *
 */


#include <math.h>
#include <stddef.h>
#include <stdlib.h>

#include "amcounts/dgaussian.h"
#include "amodel/dgaussian.h"
#include "common.h"
#include "dgaussian.h"
#include "empty.h"
#include "error.h"
#include "mem.h"
#include "samples.h"
#include "statestr.h"




/**********/
/* MACROS */
/**********/

#define GLOBAL(ptr) DGAUSSIAN_GLOBAL(ptr)
#define STATE(ptr) DGAUSSIAN_STATE(ptr)
#define CSTATE(ptr) ((amcount_dgaussian_state_t *) (ptr))
#define MAUX(ptr) ((maux_t *) (ptr))
#define STRF(ptr) ((str_factory_t *) (ptr))
#define TRS(ptr) ((statetr_t *) (ptr))




/*********/
/* TIPUS */
/*********/

typedef struct
{
  
  tLFloat       *calc;
  tLFloat       *ax;
  tLFloat       *ax2;
  float        **frames;
  unsigned int   calc_size;
  
} maux_t;


typedef struct
{
  
  int     size;      /* Grandària. */
  float **frames;    /* Els frames actius. */
  
} statetr_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static maux_t *
maux_create (
	     unsigned int size,
	     unsigned int D
	     )
{
  
  maux_t *new;
  
  
  mt ( new= mem_alloc ( maux_t, 1 ) );
  mt ( new->calc= mem_alloc ( tLFloat, size ) );
  mt ( new->ax= mem_alloc ( tLFloat, D<<1 ) );
  new->ax2= new->ax+D;
  mt ( new->frames= mem_alloc ( float *, size ) );
  new->calc_size= size;
  
  return new;
  
} /* end maux_create */


static void
maux_free (
	   maux_t *maux
	   )
{
  
  free ( maux->calc );
  free ( maux->ax );
  free ( maux->frames );
  free ( maux );
  
} /* end maux_free */


static void
resize_maux_calc (
		  maux_t       *maux,
		  unsigned int  size
		  )
{
  
  mt ( maux->calc= mem_realloc ( tLFloat, maux->calc, size ) );
  mt ( maux->frames= mem_realloc ( float *, maux->frames, size ) );
  maux->calc_size= size;
  
} /* end resize_maux_calc */


static tLFloat
calc_prob (
	   unsigned int       D,
	   dgaussian_state_t  state,
	   const float       *frame
	   )
{
  
  tLFloat prob, aux;
  unsigned int d;
  
  
  prob= 0.0;
  for ( d= 0; d < D; ++d )
    {
      aux= frame[d]-state.mu[d];
      prob+= (aux*aux)*state.ivar[d];
    }
  
  return -0.5*prob+state.logc;
  
} /* end calc_prob */




/**********************/
/* STATESTR (MÈTODES) */
/**********************/

static void
free_strm (
	   statestr_t *statestr
	   )
{
  
  size_t n;
  
  
  for ( n= 0; n < statestr->N; ++n )
    free ( TRS(statestr->statetrs)[n].frames );
  free ( statestr->statetrs );
  maux_free ( MAUX(statestr->maux) );
  
} /* end free_strm */


static void
get_ready (
	   statestr_t  *statestr,
	   const tLFea *fea,
	   void        *global
	   )
{
  
  if ( fea->cols >= MAUX(statestr->maux)->calc_size )
    resize_maux_calc ( MAUX(statestr->maux),
		       fea->cols );
  
} /* end get_ready */


static void
train_state_pseudoE (
		     void         *global,
                     const tLFea  *fea,
                     tLBool       *actives,
                     void         *states,
                     void         *statetrs,
                     unsigned int  q,
                     void         *maux,
                     tLProb       *probs
		     )
{
  
  dgaussian_state_t state;
  statetr_t *statetr;
  int n;
  unsigned int D;
  tLFloat prob;
  
  
  state= STATE(states)[q];
  statetr= &(TRS(statetrs)[q]);
  D= GLOBAL(global)->D;
  
  if ( fea->cols > statetr->size )
    {
      mt ( statetr->frames=
	   mem_realloc ( float *,
			 statetr->frames,
			 fea->cols ) );
      statetr->size= fea->cols;
    }
  
  for ( n= 0; n < fea->cols; ++n )
    if ( actives[n] )
      {
	prob= calc_prob ( D, state, B2cR(fea->v[n]) );
	if ( prob == -HUGE_VAL )
	  {
	    probs[n]= -HUGE_VAL;
	    statetr->frames[n]= NULL;
	  }
	else
	  {
	    probs[n]= prob;
	    statetr->frames[n]= B2R(fea->v[n]);
	  }
      }
    else
      {
	probs[n]= -HUGE_VAL;
	statetr->frames[n]= NULL;
      }
  
} /* end train_state_pseudoE */


static void
train_state_pseudoM (
                     void         *global,
                     int           N,
                     tLProb       *weights,
                     void         *states,
                     void         *cstates,
                     void         *statetrs,
                     unsigned int  q,
                     void         *maux
                     )
{
  
  amcount_dgaussian_state_t scount;
  statetr_t statetr;
  unsigned int d, D;
  tLFloat *calc, *ax, *ax2, aux, aux2, max, anorm;
  float **frames, *frame;
  int count, n;
  
  
  D= GLOBAL(global)->D;
  scount= CSTATE(cstates)[q];
  statetr= TRS(statetrs)[q];
  
  calc= MAUX(maux)->calc;
  ax= MAUX(maux)->ax;
  ax2= MAUX(maux)->ax2;
  frames= MAUX(maux)->frames;
  
  count= 0; max= -HUGE_VAL;
  for ( n= 0; n < N; ++n )
    if ( (frame= statetr.frames[n]) != NULL &&
	 weights[n] != -HUGE_VAL )
      {
	aux= calc[count]= weights[n];
	if ( aux > max ) max= aux;
	frames[count++]= frame;
      }
  
  if ( count == 0 ) return;
  for ( d= 0; d < D; ++d ) ax[d]= 0.0;
  for ( d= 0; d < D; ++d ) ax2[d]= 0.0;
  anorm= 0;
  for ( n= 0; n < count; ++n )
    {
      frame= frames[n];
      aux= exp ( calc[n]-max );
      for ( d= 0; d < D; ++d )
	{
	  aux2= frame[d];
	  ax[d]+= aux*aux2;
	  ax2[d]+= aux*(aux2*aux2);
	}
      anorm+= aux;
    }
  max= exp ( max );
  CSTATE(cstates)[q].norm+= max*anorm;
  for ( d= 0; d < D; ++d )
    {
      scount.x[d]+= max*ax[d];
      scount.x2[d]+= max*ax2[d];
    }
  
} /* end train_state_pseudoM */


static const strm_t strm=
  {
    free_strm,
    get_ready,
    train_state_pseudoE,
    tL_empty_statestr_train_global_pseudoE,
    train_state_pseudoM,
    tL_empty_statestr_train_global_pseudoM,
    tL_empty_statestr_train_init_pseudoM
  };




/*************************/
/* STR_FACTORY (MÈTODES) */
/*************************/

static void
init_statestr (
	       str_factory_t *strf,
	       statestr_t    *str,
	       void          *global,
	       const size_t   N
	       )
{
  
  size_t n;
  statetr_t *aux;
  
  
  str->N= N;
  mt ( aux= mem_alloc ( statetr_t, N ) );
  for ( n= 0; n < N; ++n )
    {
      aux[n].size= 1;
      mt ( aux[n].frames= mem_alloc ( float *, 1 ) );
    }
  str->statetrs= (void *) aux;
  str->maux= (void *) maux_create ( 1, GLOBAL(global)->D );
  str->meths= &strm;
  str->_statetr_size= sizeof(statetr_t);
  
} /* end init_statestr */


static const strfm_t strfm=
  {
    tL_empty_free_strf,
    tL_empty_strf_add,
    init_statestr
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

str_factory_t *
tL_create_strf_dgaussian (
			  void *global
			  )
{
  
  str_factory_t *new;
  
  
  mt ( new= mem_alloc ( str_factory_t, 1 ) );
  new->meths= &strfm;
  
  return new;
  
} /* end tL_create_strf_dgaussian */
