/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mixture.c - Implementació de 'mixture.h'.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <stddef.h>
#include <stdlib.h>

#include "calc.h"
#include "amcounts/imixture.h"
#include "amcounts/logmixture.h"
#include "amodel/mixture.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "samples.h"
#include "statestr.h"




/**********/
/* MACROS */
/**********/

#define SINFO(ptr) ((stateinfo_t *) (ptr))
#define GLOBAL(ptr) MIXTURE_GLOBAL(ptr)
#define CGLOBAL(ptr) ((logamcount_mixture_global_t *) (ptr))
#define STATE(ptr) MIXTURE_STATE(ptr)
#define CSTATE(ptr) ((logamcount_mixture_state_t *) (ptr))
#define CSTATE2(ptr) ((iamcount_mixture_state_t *) (ptr))
#define MAUX(ptr) ((maux_t *) (ptr))
#define MAUXV(ptr) ((mauxv_t *) (ptr))
#define STRF(ptr) ((mixture_strf_t *) (ptr))
#define STRVF(ptr) ((mixture_strvf_t *) (ptr))
#define TRS(ptr) ((statetr_t *) (ptr))
#define TRVS(ptr) ((statetrv_t *) (ptr))

#define TRSF_INITIALIZED(TRSF) ((TRSF).z != NULL)




/*********/
/* TIPUS */
/*********/

/* Memòria auxiliar per als càlculs. */
typedef struct
{
  
  statestr_t   *str;
  unsigned int  I;
  tLFloat      *calc;
  unsigned int  size;
  
} maux_t;


/* Memòria auxiliar per als càlculs. */
typedef struct
{
  
  statestrv_t   *str;
  unsigned int   I;
  tLProb        *aux_probs;
  int           *actives;
  int            size;
  
} mauxv_t;


typedef struct
{
  
  int       size;        /* Grandària. */
  tLFloat **z;           /* |I|*?. variables indicadores. */
  void     *statetrs;    /* |I|. Els 'statetr' de les
			   components. */
  
} statetr_t;


typedef struct
{
  
  int           size;        /* Grandària. */
  unsigned int *i;           /* ?. Component més probable en cada
				instant. */
  void         *statetrs;    /* |I|. Els 'statetr' de les
                                 components. */
  
} statetrv_t;


typedef struct
{
  
  STR_FACTORY
  str_factory_t *strf;
  unsigned int   max_I;
  
} mixture_strf_t;


typedef struct
{
  
  STRV_FACTORY
  strv_factory_t *strvf;
  unsigned int    max_I;
  
} mixture_strvf_t;




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static void
accum_z (
	 int            N,
	 unsigned int   I,
	 tLFloat      **z,
	 tLFloat       *counts,
	 tLFloat       *calc
	 )
{
  
  int N2, n;
  unsigned int i;
  tLFloat max, aux, *zi;
  
  
  for ( i= 0; i < I; ++i )
    {
      
      zi= z[i]; N2= 1;
      calc[0]= max= counts[i];
      
      for ( n= 0; n < N; ++n )
	if ( (aux= zi[n]) != -HUGE_VAL )
	  {
	    calc[N2++]= aux;
	    if ( aux > max ) max= aux;
	  }
      
      if ( N2 > 1 )
	counts[i]= tL_radd_vals ( calc, max, N2 );
      
    }
  
} /* end accum_z */


static maux_t *
create_maux (
	     str_factory_t          *strf,
	     const mixture_global_t *global,
	     unsigned int            max_I,
	     size_t                  N
	     )
{
  
  maux_t *new;
  size_t tsize;
  
  
  mt ( new= mem_alloc ( maux_t, 1 ) );
  new->I= max_I;
  tsize= N * max_I;
  new->str= tL_str_factory_make ( strf, global->global, tsize );
  mt ( new->calc= mem_alloc ( tLFloat, max_I ) );
  new->size= max_I;
  
  return new;
  
} /* end create_maux */


static mauxv_t *
create_mauxv (
	      strv_factory_t         *strvf,
	      const mixture_global_t *global,
	      unsigned int            max_I,
	      size_t                  N
	      )
{
  
  mauxv_t *new;
  size_t tsize;
  
  
  mt ( new= mem_alloc ( mauxv_t, 1 ) );
  new->I= max_I;
  tsize= N * max_I;
  new->str= tL_strv_factory_make ( strvf, global->global, tsize );
  mt ( new->aux_probs= mem_alloc ( tLProb, 1 ) );
  mt ( new->actives= mem_alloc ( int, 1 ) );
  new->size= 1;
  
  return new;
  
} /* end create_mauxv */


static void
free_maux (
	   maux_t *maux
	   )
{
  
  free ( maux->calc );
  free_statestr ( maux->str );
  free ( maux );
  
} /* free_maux */


static void
free_mauxv (
	    mauxv_t *maux
	    )
{
  
  free ( maux->actives );
  free ( maux->aux_probs );
  free_statestrv ( maux->str );
  free ( maux );
  
} /* end free_mauxv */


static void
free_statetr (
	      statetr_t statetr
	      )
{
  
  free ( statetr.z[0] );
  free ( statetr.z );
  
} /* end free_statetr */


static void
init_statetr (
	      statetr_t    *statetr,
	      unsigned int  max_I,
	      void         *statetrs
	      )
{
  
  unsigned int i;
  tLFloat *aux;
  
  
  statetr->size= 1;
  mt ( statetr->z=
       mem_alloc ( tLFloat *, max_I ) );
  mt ( statetr->z[0]=
       mem_alloc ( tLFloat, max_I ) );
  for ( i= 1, aux= statetr->z[0] + 1;
	i < max_I; ++i, ++aux )
    statetr->z[i]= aux;
  statetr->statetrs= statetrs;
  
} /* end init_statetr */


static void
init_statetrv (
	       statetrv_t   *statetrv,
	       void         *statetrs
	       )
{
  
  statetrv->size= 1;
  mt ( statetrv->i= mem_alloc ( unsigned int, 1 ) );
  statetrv->statetrs= statetrs;
  
} /* end init_statetrv */


static void
resize_maux_calc (
                  maux_t       *maux,
                  unsigned int  size
                  )
{

  mt ( maux->calc=
       mem_realloc ( tLFloat, maux->calc, size ) );
  maux->size= size;

} /* end resize_maux_calc */


static void
resize_mauxv_aux_probs (
			mauxv_t *maux,
			int      size
			)
{
  
  mt ( maux->aux_probs=
       mem_realloc ( tLProb, maux->aux_probs, size ) );
  mt ( maux->actives=
       mem_realloc ( int, maux->actives, size ) );
  maux->size= size;
  
} /* end resize_mauxv_aux_probs */


static void
resize_statetr (
		statetr_t    *statetr,
		unsigned int  newsize,
		unsigned int  max_I
		)
{
  
  tLFloat *aux;
  unsigned int i;
  
  
  mt ( statetr->z[0]=
       mem_realloc ( tLFloat, statetr->z[0],
		     max_I * newsize ) );
  for ( i= 1, aux= statetr->z[0] + newsize;
	i < max_I; ++i, aux+= newsize )
    statetr->z[i]= aux;
  statetr->size= newsize;
  
} /* end resize_statetr */


static void
resize_statetrv (
		 statetrv_t   *statetrv,
		 unsigned int  newsize,
		 unsigned int  max_I
		 )
{
  
  mt ( statetrv->i=
       mem_realloc ( unsigned int, statetrv->i,
		     max_I * newsize ) );
  statetrv->size= newsize;
  
} /* end resize_statetrv */




/**********************/
/* STATESTR (MÈTODES) */
/**********************/

static void
free_strm (
	   statestr_t *statestr
	   )
{
  
  size_t n;
  
  
  for ( n= 0; n < statestr->N; ++n )
    free_statetr ( TRS(statestr->statetrs)[n] );
  free ( statestr->statetrs );
  free_maux ( MAUX(statestr->maux) );
  
} /* end free_strm */


static void
free_strvm (
	    statestrv_t *statestrv
	    )
{
  
  size_t n;
  
  
  for ( n= 0; n < statestrv->N; ++n )
    free ( TRVS(statestrv->statetrs)[n].i );
  free ( statestrv->statetrs );
  free_mauxv ( MAUXV(statestrv->maux) );
  
} /* end free_strvm */


static void
get_ready (
           statestr_t  *statestr,
           const tLFea *fea,
	   void        *global
           )
{
  
  maux_t *maux;
  
  
  maux= MAUX(statestr->maux);
  if ( fea->cols >= maux->size )
    resize_maux_calc ( maux, fea->cols+1 );
  statestr_get_ready ( maux->str, fea,
		       GLOBAL(global)->global );
  
} /* end get_ready */


static void
get_readyv (
	    statestrv_t *statestrv,
	    const tLFea *fea,
	    void        *global
	    )
{
  
  mauxv_t *maux;
  
  
  maux= MAUXV(statestrv->maux);
  if ( fea->cols > maux->size )
    resize_mauxv_aux_probs ( maux, fea->cols );
  statestrv_get_ready ( maux->str, fea,
			GLOBAL(global)->global );
  
} /* end get_readyv */


/* NOTA: En aquest cas no és de molta utilitat
 * enrecordar-se de quins estats estaven actius,
 * simplement, aprofitant que les componenents
 * tornen probabilitat 0, les probabilitats 0 es
 * votaran en el pas M.
 */
static void
train_state_pseudoE (
                     void         *global,
                     const tLFea  *fea,
                     tLBool       *actives,
                     void         *states,
                     void         *statetrs,
                     unsigned int  q,
                     void         *maux,
                     tLProb       *probs
                     )
{
  
  mixture_state_t state;
  statetr_t statetr;
  unsigned int i;
  int n;
  tLFloat *calc, max, aux;
  volatile tLFloat fix;
  
  
  calc= MAUX(maux)->calc;
  state= STATE(states)[q];
  statetr= TRS(statetrs)[q];
  
  if ( fea->cols > statetr.size )
    {
      resize_statetr ( TRS(statetrs)+q,
		       (unsigned int) (fea->cols),
		       MAUX(maux)->I );
      statetr= TRS(statetrs)[q];
    }
  
  for ( i= 0; i < state.I; ++i )
    statestr_pseudoE ( MAUX(maux)->str,
		       GLOBAL(global)->global,
		       fea, actives,
		       state.members,
		       statetr.statetrs, i,
		       (tLProb *) (statetr.z[i]) );
  
  for ( n= 0; n < fea->cols; ++n )
    if ( actives[n] )
      {
	
	max= -HUGE_VAL;
	for ( i= 0; i < state.I; ++i )
	  {
	    calc[i]= aux= fix=
	      state.p[i] + statetr.z[i][n];
	    if ( aux > max ) max= aux;
	    statetr.z[i][n]= aux;
	  }
	
	probs[n]= aux=
	  (state.I > 1 && max != -HUGE_VAL) ?
	  tL_radd_vals ( calc, max, state.I ) :
	  max;
	
	/* NOTA: no actualitze
	   les z perquè si es 0 totes tenen q ser 0.
	   Com normalment estaran suavitzades les probabilitats
	   no és una gran perdua d'eficiència. */
	  if ( aux != -HUGE_VAL )
	    for ( i= 0; i < state.I; ++i )
	      statetr.z[i][n]-= aux;
	  
      }
    else probs[n]= -HUGE_VAL;
  
} /* train_state_pseudoE */


static void
train_state_get_probs (
		       void         *global,
		       const tLFea  *fea,
		       tLBool       *actives,
		       void         *states,
		       void         *statetrs,
		       unsigned int  q,
		       void         *maux,
		       tLProb       *probs
		       )
{
  
  mixture_state_t state;
  statetrv_t statetr;
  unsigned int i;
  int n, *aux_actives, N, n2;
  tLProb *aux_probs, prob, pi;
  
  
  state= STATE(states)[q];
  statetr= TRVS(statetrs)[q];
  
  if ( fea->cols > statetr.size )
    {
      resize_statetrv ( TRVS(statetrs)+q,
			(unsigned int) (fea->cols),
			MAUXV(maux)->I );
      statetr= TRVS(statetrs)[q];
    }
  aux_probs= MAUXV(maux)->aux_probs;
  aux_actives= MAUXV(maux)->actives;

  N= 0;
  for ( n= 0; n < fea->cols; ++n )
    {
      probs[n]= -HUGE_VAL;
      if ( actives[n] ) aux_actives[N++]= n;
    }
  for ( i= 0; i < state.I; ++i )
    {
      statestrv_get_probs ( MAUXV(maux)->str,
			    GLOBAL(global)->global,
			    fea, actives,
			    state.members,
			    statetr.statetrs, i,
			    aux_probs );
      pi= state.p[i];
      for ( n2= 0; n2 < N; ++n2 )
	{
	  n= aux_actives[n2];
	  prob= aux_probs[n] + pi;
	  if ( prob > probs[n] )
	    {
	      probs[n]= prob;
	      statetr.i[n]= i;
	    }
	}
    }
  
} /* train_state_get_probs */


static void
train_global_pseudoE (
		      void        *global,
		      const tLFea *fea,
		      void        *maux
		      )
{
  
  statestr_pseudoE_global ( MAUX(maux)->str,
			    GLOBAL(global)->global,
			    fea );
  
} /* end train_global_pseudoE */


static void
train_state_pseudoM (
                     void         *global,
                     int           N,
                     tLProb       *weights,
                     void         *states,
                     void         *cstates,
                     void         *statetrs,
                     unsigned int  q,
                     void         *maux
                     )
{
  
  int n;
  unsigned int i;
  mixture_state_t state;
  logamcount_mixture_state_t scount;
  statetr_t statetr;
  tLFloat *zi;
  
  
  state= STATE(states)[q];
  scount= CSTATE(cstates)[q];
  statetr= TRS(statetrs)[q];
  
  for ( i= 0; i < state.I; ++i )
    {
      
      zi= statetr.z[i];
      for ( n= 0; n < N; ++n )
	zi[n]+= weights[n];
      
      statestr_pseudoM ( MAUX(maux)->str,
			 GLOBAL(global)->global,
			 N, zi, state.members,
			 scount.members,
			 statetr.statetrs, i );
      
    }
  
  accum_z ( N, state.I, statetr.z,
	    scount.p, MAUX(maux)->calc );
  
} /* end train_state_pseudoM */


static void
train_state_accum (
		   void         *global,
		   const int     t,
		   unsigned int  q,
		   void         *states,
		   void         *cstates,
		   void         *statetrs,
		   void         *maux
		   )
{
  
  mixture_state_t state;
  iamcount_mixture_state_t scount;
  statetrv_t statetr;
  unsigned int i;
  
  
  state= STATE(states)[q];
  scount= CSTATE2(cstates)[q];
  statetr= TRVS(statetrs)[q];
  
  i= statetr.i[t];
  ++scount.p[i];
  statestrv_accum ( MAUXV(maux)->str,
		    GLOBAL(global)->global,
		    t, i, state.members,
		    scount.members,
		    statetr.statetrs );
  
} /* end train_state_accum */


static void
train_global_pseudoM (
		      void *global,
		      void *cglobal,
		      int   N,
		      void *maux
		      )
{
  
  statestr_pseudoM_global ( MAUX(maux)->str,
			    GLOBAL(global)->global,
			    CGLOBAL(cglobal)->global,
			    N );
  
} /* end train_global_pseudoM */


static void
train_init_pseudoM (
		    void *global,
		    int   N,
		    void *maux
		    )
{
  
  statestr_pseudoM_init ( MAUX(maux)->str,
			  GLOBAL(global)->global,
			  N );
  
} /* end train_init_pseudoM */


static const strm_t strm=
  {
    free_strm,
    get_ready,
    train_state_pseudoE,
    train_global_pseudoE,
    train_state_pseudoM,
    train_global_pseudoM,
    train_init_pseudoM
  };


static const strvm_t strvm=
  {
    free_strvm,
    get_readyv,
    train_state_get_probs,
    train_state_accum
  };




/*************************/
/* STR_FACTORY (MÈTODES) */
/*************************/

static void
free_strf (
           str_factory_t *strf
           )
{
  
  free_str_factory ( STRF(strf)->strf );
  free ( strf );
  
} /* end free_strf */


static void
add (
     str_factory_t      *strf,
     void               *states,
     const unsigned int  Q
     )
{
  
  unsigned int q, i;
  mixture_state_t state;
  
  
  for ( q= 0; q < Q; ++q )
    {
      state= STATE(states)[q];
      if ( state.I > STRF(strf)->max_I )
	STRF(strf)->max_I= state.I;
      for ( i= 0; i < state.I; ++i )
	str_factory_add ( STRF(strf)->strf,
			  state.members,
			  state.I );
    }
  
} /* end add */


static void
init_statestr (
               str_factory_t *strf,
               statestr_t    *str,
	       void          *global,
               const size_t   N
               )
{
  
  size_t n, inc;
  maux_t *maux;
  statetr_t *aux;
  unsigned int max_I;
  void *aux2;
  
  
  max_I= STRF(strf)->max_I;
  maux= create_maux ( STRF(strf)->strf,
		      GLOBAL(global),
		      max_I, N );
  str->N= N;
  mt ( aux= mem_alloc ( statetr_t, N ) );
  inc= statestr_sizeof ( maux->str, max_I );
  aux2= maux->str->statetrs;
  for ( n= 0; n < N; ++n )
    {
      init_statetr ( &(aux[n]), max_I, aux2 );
      aux2= (void *) (((char *) aux2) + inc);
    }
  str->statetrs= (void *) aux;
  str->maux= (void *) maux;
  str->meths= &strm;
  str->_statetr_size= sizeof(statetr_t);
  
} /* end init_statestr */


static void
free_strvf (
	    strv_factory_t *strvf
	    )
{
  
  free_strv_factory ( STRVF(strvf)->strvf );
  free ( strvf );
  
} /* end free_strvf */


static void
addv (
      strv_factory_t     *strvf,
      void               *states,
      const unsigned int  Q
      )
{
  
  unsigned int q, i;
  mixture_state_t state;
  
  
  for ( q= 0; q < Q; ++q )
    {
      state= STATE(states)[q];
      if ( state.I > STRVF(strvf)->max_I )
	STRF(strvf)->max_I= state.I;
      for ( i= 0; i < state.I; ++i )
	strv_factory_add ( STRVF(strvf)->strvf,
			   state.members,
			   state.I );
    }
  
} /* end addv */


static void
init_statestrv (
		strv_factory_t *strvf,
		statestrv_t    *str,
		void           *global,
		const size_t    N
		)
{
  
  size_t n, inc;
  mauxv_t *maux;
  statetrv_t *aux;
  unsigned int max_I;
  void *aux2;
  
  
  max_I= STRVF(strvf)->max_I;
  maux= create_mauxv ( STRVF(strvf)->strvf,
		       GLOBAL(global),
		       max_I, N );
  str->N= N;
  mt ( aux= mem_alloc ( statetrv_t, N ) );
  inc= statestrv_sizeof ( maux->str, max_I );
  aux2= maux->str->statetrs;
  for ( n= 0; n < N; ++n )
    {
      init_statetrv ( &(aux[n]), aux2 );
      aux2= (void *) (((char *) aux2) + inc);
    }
  str->statetrs= (void *) aux;
  str->maux= (void *) maux;
  str->meths= &strvm;
  str->_statetr_size= sizeof(statetrv_t);
  
} /* end init_statestrv */


static const strfm_t strfm=
  {
    free_strf,
    add,
    init_statestr
  };


static const strvfm_t strvfm=
  {
    free_strvf,
    addv,
    init_statestrv
  };




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

str_factory_t *
tL_create_strf_mixture (
			void *global
			)
{
  
  mixture_strf_t *new;
  
  
  mt ( new= mem_alloc ( mixture_strf_t, 1 ) );
  new->meths= &strfm;
  new->strf= tL_create_str_factory ( GLOBAL(global)->type,
				     GLOBAL(global)->global );
  new->max_I= 1;
  
  return (str_factory_t *) new;
  
} /* end tL_create_strf_mixture */


strv_factory_t *
tL_create_strvf_mixture (
			 void *global
			 )
{
  
  mixture_strvf_t *new;
  
  
  mt ( new= mem_alloc ( mixture_strvf_t, 1 ) );
  new->meths= &strvfm;
  new->strvf= tL_create_strv_factory ( GLOBAL(global)->type,
				       GLOBAL(global)->global );
  new->max_I= 1;
  
  return (strv_factory_t *) new;
  
} /* end tL_create_strvf_mixture */
