/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  tar.c - Implementació de tLTar.
 *
 *  Adrià Giménez Pastor, 2012.
 *
 */


#include <errno.h>
#include <limits.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "error.h"
#include "mem.h"
#include "samples.h"




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_tar_free (
	     tLTar *tar
	     )
{
  
  tL_dict_free ( tar->files );
  free ( tar->offsets );
  free ( tar->sizes );
  gzclose ( tar->f );
  free ( tar );
  
} /* end tL_tar_free */


TL_API_FUNC(gzFile)
tL_tar_get_file (
		 tLTar       *tar,
		 const char  *name,
                 long        *size,
		 char       **err
		 )
{
  
  size_t pos;
  
  
  pos= tL_dict_find ( tar->files, name );
  if ( tar->files->N == pos )
    {
      tL_msgerror ( err, _( "'%s' is not in the tar file" ) , name );
      return NULL;
    }
  else
    {
      if ( gzseek ( tar->f, tar->offsets[pos], SEEK_SET ) == -1 )
	{
	  tL_msgerror ( err, _( "unexpected error reading: %s" ) ,
			strerror ( errno ) );
	  return NULL;
	}
      if ( size != NULL ) *size = tar->sizes[pos];
      return tar->f;
    }
  
} /* end tL_tar_get_file */


TL_API_FUNC(tLTar *)
tL_tar_new (
	    const char  *file_name,
	    char       **err
	    )
{
  
  tLTar *new;
  char buffer[512], name[257], *end, type, *longname, *aux;
  const char *error_msg;
  long size, nblocks, offset;
  int i;
  size_t id, offsize, lnamesize, nsize;
  tLBool jump, uselongname;
  
  
  mt ( new= mem_alloc ( tLTar, 1 ) );
  new->f= gzopen ( file_name, "rb" );
  if ( new->f == NULL )
    {
      TLMSG_ORFILE ( err, file_name );
      free ( new );
      return NULL;
    }
  new->files= tL_dict_new ();
  mt ( new->offsets= mem_alloc ( long, 1 ) );
  mt ( new->sizes= mem_alloc ( long, 1 ) );
  offsize= 1;
  lnamesize= 512;
  mt ( longname= mem_alloc ( char, 512 ) );
  
  /* Processa el TAR. */
  uselongname= FALSE;
  offset= 0;
  for (;;)
    {
      
      jump= TRUE;
      if ( gzread ( new->f, buffer, 512*1 ) != 512 )
	{
	  tL_msgerror ( err, _( "unable to read the next block from '%s'" ) ,
			file_name );
	  goto error;
	}
      offset+= 512;
      
      /* Comprova el bloc. */
      if ( strncmp ( "ustar ", &(buffer[257]), 6 ) )
	{
	  for ( i= 257; i < 263; ++i )
	    if ( buffer[i] != 0 ) goto formaterror;
	  break;
	}
      
      /* Tipus i grandària. */
      type= buffer[156];
      buffer[135]= '\0';
      size= strtol ( &(buffer[124]), &end, 8 );
      if ( size == LONG_MIN || size == LONG_MAX
	   || buffer[124]=='\0' || *end!='\0')
	goto formaterror;
      nblocks= size/512;
      if ( size%512 ) ++nblocks;
      
      if ( type == 'L' )
	{
	  nsize= 512*nblocks;
	  if ( nsize > lnamesize )
	    {
	      lnamesize= nsize;
	      mt ( longname= mem_realloc ( char, longname, nsize ) );
	    }
	  if ( gzread ( new->f, longname, nsize*1 ) != nsize*1 )
	    {
	      tL_msgerror ( err,
			    _( "unable to read blocks for a longname from '%s'" ) ,
			    file_name );
	      goto error;
	    }
	  jump= FALSE;
	  offset+= nsize;
	  longname[nsize-1]= '\0';
	  uselongname= TRUE;
	}
      else if ( size > 0 && (type == '0' || type == '\0') )
	{

	  /* Obté el nom. */
	  if ( uselongname )
	    {
	      aux= longname;
	      uselongname= FALSE;
	    }
	  else
	    {
	      buffer[100]= '\0';
	      if ( buffer[345] == '\0' ) strcpy ( name, buffer );
	      else
		{
		  buffer[499]= '\0';
		  strcpy ( name, &(buffer[345]) );
		  strcat ( name, "/" );
		  strcat ( name, buffer );
		}
	      aux= name;
	    }
	  
	  /* Inserta fitxer. */
	  if ( tL_dict_insert ( new->files, aux, &id ) )
	    {
	      if ( offsize < new->files->N )
		{
		  offsize*= 2;
		  mt ( new->offsets=
		       mem_realloc ( long, new->offsets, offsize ) );
		  mt ( new->sizes=
		       mem_realloc ( long, new->sizes, offsize ) );
		}
	      new->offsets[id]= offset;
              new->sizes[id]= size;
	    }
	  else tL_warning ( _( "file '%s' is repeated in '%s'" ) ,
			    aux, file_name );
	}
      else uselongname= FALSE;
      
      /* Bota al següent bloc de dades. */
      if ( jump )
	{
	  offset+= nblocks*512;
	  if ( gzseek ( new->f, offset, SEEK_SET ) == -1 )
	    {
        error_msg = gzerror ( new->f, &errno ) ;
        if ( errno == Z_ERRNO )
          tL_msgerror ( err, _( "unexpected error reading '%s': %s" ) ,
            file_name, strerror(errno) );
        else
          tL_msgerror ( err, _( "unexpected error reading '%s': %s" ) ,
            file_name, error_msg );
	      goto error;
	    }
	}
      
    }
  free ( longname );
  
  return new;

 formaterror:
  tL_msgerror ( err, _( "'%s' is not a valid tar file" ) , file_name );
 error:
  tL_tar_free ( new );
  free ( longname );
  return NULL;
  
} /* end tL_tar_new */
