/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  text_public.h - Part pública de 'text.h'.
 *
 *  Adrià Giménez Pastor, 2008,2010,2011,2012,2013.
 *
 */

#ifndef __LIBTL_TEXT_H__
#define __LIBTL_TEXT_H__

#include <stddef.h>
#include <stdio.h>
#include <zlib.h>

#include "port.h"

/** \addtogroup Basic
 *  @{
 */

/* TIPUS BUFFER */

/** Text buffer. This structure is used to process text.
 */
typedef struct
{
  
  char   *v;    /**< Content of the buffer. */
  size_t  s;
  
} tLBuffer;

/* Crea un buffer. */
/** Creates a new buffer.
 *
 *  @return The new buffer.
 */
TL_API_FUNC(tLBuffer *)
tL_buffer_new ();

/* Allibera un buffer. */
/** Frees memory. Frees the memory allocated for the buffer.
 *
 *  @param buffer The buffer to be freed.
 */
TL_API_FUNC(void)
tL_buffer_free (
		tLBuffer *buffer
		);

/* Llig la següent línia en el buffer. Torna 0 si no s'ha pogut
 * llegir res.
 */
/** Reads line. This function reads the next line from a file into the
 *  provided text buffer. Line separators are ignored.
 *
 *  @param f Input file descriptor.
 *  @param buffer Text buffer to which the line is read.
 *  @return 0 if no line has been read.
 */
TL_API_FUNC(int)
tL_gline (
	  gzFile    f,
	  tLBuffer *buffer
	  );

/* Llig el següent token en el buffer.  Torna 0 si no s'ha pogut
 * llegir res. S'assegura que sols llegirà un caràcter de més.
 */
/** Reads token. This function reads the next token from a file into
 *  the provided text buffer. Blank separators are ignored.
 *
 *  @param f Input file descriptor.
 *  @param buffer Text buffer to which the token is read.
 *  @return 0 if no token has been read.
 */
TL_API_FUNC(int)
tL_gtoken (
	   gzFile    f,
	   tLBuffer *buffer
	   );

/** Reads the next token from a file descriptor until the end of file or
 *  until n bytes are read
 *
 *  @param f Input file descriptor.
 *  @param buffer Text buffer to which the token is read.
 *  @param n number of bytes to be read
 *  @return 0 if no token has been read, else the number of read characters
 *  is returned.
 */
TL_API_FUNC(int)
tL_gntoken (
	   gzFile    f,
	   tLBuffer *buffer,
           size_t   n
	   );

/* Apunta 'begin' al principi del primer token de 'str' i 'end' la
 * posició del següent caràcter. 'begin' apunta a NULL si no hi ha
 * res. Torna 1 si s'aplega al final de la cadena 0 en cas contrari.
 */
/** Reads token from string. This function reads the next token from a
 *  given string.
 *
 *  @param str Input string.
 *  @param begin A pointer to string, where the address of the first
 *  character of the token is stored, or NULL if there is no token
 *  in str.
 *  @param end A pointer to string, where, if it exists, the address
 *  of the next character of the token is stored.
 *  @return 1 if the input string has been fully processed, 0
 *  otherwise.
 */
TL_API_FUNC(int)
tL_gtokenstr (
	      char  *str,
	      char **begin,
	      char **end
	      );


/* TIPUS DICCIONARI */

#define TLDICTS 32363U      /* hash table size: 101U 1001U 32363U */

struct tLDict_h
{
  
  void            *n;
  size_t           i;
  struct tLDict_h *p;
  
};

/* Aquest tipus serveix tant per a tokens com per a seqüències
   codificades. En la part pública sols es va a presentar com a
   diccionari de tokens. */
/** Dictionary of tokens. Dictionaries are used to map tokens to
 *  numbers.
 */
typedef struct
{
  
  size_t            N;    /**< Current number of inserted tokens. */
  void            **data;
  struct tLDict_h  *_table[TLDICTS];
  size_t            _data_size;
  void            (*free_h) (void *);
  int             (*eq) (const void *,const void *);
  void *          (*copy) (const void *);
  unsigned int    (*hash) (const void *);
  
} tLDict;

/* Crea un diccionari de 'tokens'. */
/** Creates a new dictionary.
 *
 *  @return The new dictionary.
 */
TL_API_FUNC(tLDict *)
tL_dict_new ();

/* Afegeix un valor. Torna 0 si ja estaba o 1 si s'ha pogut
 * insertar. Si id != NULL fica l'identificador. El primer element
 * insertat serà 0, i els següents s'incrementaran seqüencialment.
 * Serveix també per a cseqs.
 */
/** Inserts a token. Tries to insert a token into the dictionary. If
 *  the token already exists, then it is not inserted. In any case,
 *  the number mapped to the token is stored in id.
 *  
 *  The first token inserted into the dictionary is mapped to 0, the
 *  second token is mapped to 1, and so on.
 *
 *  @param dict The dictionary.
 *  @param token A C string (char *) containing the token.
 *  @param id A pointer to a variable of type 'size_t'. The mapped
 *  number is stored in this variable.
 *  @return Returns 1 if the token has been inserted, 0 if the token is
 *  already in the dictionary.
 */
TL_API_FUNC(int)
tL_dict_insert (
		tLDict     *dict,
		const void *token,
		size_t     *id
		);

/* Torna el 'token' associat a l'identificador. */
/** Gets the corresponding token. Returns the mapped token to a given
 *  number. The number is assumed to be mapped to an existing token.
 *
 *  @param DICT The dictionary.
 *  @param ID The number of the requested token.
 *  @return The token.
 */
#define tL_dict_get_token(DICT,ID)		\
  ((const char *) (DICT)->data[(ID)])

/* Torna el índex associat a un valor
 * o N si no està.
 */
/** Searches a token. Returns the number mapped to the token, or
 *  dict->N if the token is not inserted in the dictionary.
 *
 *  @param dict The dictionary.
 *  @param token A C string (char *) containing the token.
 *  @return The number mapped to the token, or dict->N if the token
 *  is not found.
 */
TL_API_FUNC(size_t)
tL_dict_find (
	      const tLDict *dict,
	      const void   *token
	      );

/* Allibera un diccionari. */
/** Frees memory. Frees the memory allocated for the dictionary.
 *
 *  @param dict The dictionary.
 */
TL_API_FUNC(void)
tL_dict_free (
	      tLDict *dict
	      );

/* TIPUS SEQÜÈNCIA CODIFICADA. */

/** Symbol sequence. This structure is used to store symbol
 *  sequences which have been encoded using a tLDict.
 */
typedef struct
{
  
  size_t  length;    /**< Number of symbols. */
  size_t *v;         /**< Symbols. */
  
} tLSeq;

/** Create a new symbol sequence from another one. This function
 *  creates a new symbol sequence with the same length and content
 *  of the provided symbol sequence.
 *
 *  @param seq The symbol sequence to be copied.
 *  @return The new symbol sequence.
 */
TL_API_FUNC(tLSeq *)
tL_seq_copy (
	      const tLSeq *seq
	      );

/** Frees memory. Frees the memory allocated for the symbol
 *  sequence.
 */
TL_API_FUNC(void)
tL_seq_free (
	      tLSeq *seq
	      );

/** Creates a new empty symbol sequence. This function creates a new
 *  symbol sequence of the specified length, and sets all symbol
 *  identifiers to 0.
 *
 *  @param length The length of the new symbol sequence. Must be
 *  greater than 0.
 *  @return The new symbol sequence.
 */
TL_API_FUNC(tLSeq *)
tL_seq_new (
	     const size_t length
	     );

/** Creates a new symbol sequence from a word. This function splits
 *  the provided word into UTF-8 characters, and using the provided
 *  symbol dictionary, creates a new symbol sequence containing the
 *  encoded word.
 *
 *  @param word A string containing a UTF-8 word.
 *  @param syms A dictionary used to convert UTF-8 characters (symbols) to
 *  numbers.
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return The new symbol sequence, or NULL in case of error.
 */
TL_API_FUNC(tLSeq *)
tL_seq_new_from_word (
		       const char    *word,
		       const tLDict  *syms,
		       char         **err
		       );

/* Imprimeix una seqüència codificada.  El separador '\0' indica que
   no hi ha separador. Es supossa que syms conté tots els símbols de
   la seqüència. */
/** Prints a symbol sequence. This function writes to the given
 *  file the symbol sequence. The symbol sequence is decoded
 *  using the given symbol dictionary. If sep is set to '\\0', the
 *  sequence is printed without spaces between symbols.
 *
 *  @param seq The symbol sequence.
 *  @param to File to which the sequence is written.
 *  @param syms Dictionary with the symbols.
 *  @param sep Character used as separator. '\\0' means no separator.
 */
TL_API_FUNC(void)
tL_seq_print (
	       const tLSeq  *seq,
	       FILE         *to,
	       const tLDict *syms,
	       const char    sep
	       );

/** Prints a symbol sequence with wildcards. This function writes
 *  to the given file the symbol sequence, but replacing the first
 *  and last symbols, which are assumed to be wildcards, with the
 *  given symbols. The symbol sequence is decoded using the
 *  given symbol dictionary. If sep is set to '\\0', the sequence is
 *  printed without spaces between symbols.
 *
 *  @param seq The symbol sequence.
 *  @param to File to which the sequence is written.
 *  @param syms Dictionary with the symbols.
 *  @param sep Character used as separator. '\\0' means no separator.
 *  @param first First symbol.
 *  @param last Last symbol.
 */
TL_API_FUNC(void)
tL_seq_print_wildcards (
			 const tLSeq  *seq,
			 FILE         *to,
			 const tLDict *syms,
			 const char    sep,
			 const size_t  first,
			 const size_t  last
			 );

/** @} */

#endif /* __LIBTL_TEXT_H__ */
