/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  train_viterbi_hmm.c - Implementació de 'train_viterbi_hmm.h'.
 *
 *  Adrià Gimenez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#define __TL_PRV
#include "calcelprob.h"
#include "amodel.h"
#include "common.h"
#include "error.h"
#include "iamcounts.h"
#include "init_hmm.h"
#include "init_proto.h"
#include "mem.h"
#include "samples.h"
#include "train_viterbi_hmm.h"




/**********/
/* MACROS */
/**********/

/* Càlcul de la probabilitat en 'viterbi'. */
#define VLPROB(Q,T)							\
   calc_elprob ( cep, amodel, cchar.states, (Q), fea, (T)+frame_begin )




/*********/
/* TIPUS */
/*********/

struct tr_node
{
  
  tLProb       p;
  unsigned int from;
  
};


struct tr
{
  
  struct tr_node **v;
  int              length;
  
};


struct _tLHMMVTrainer
{
  
  const tLAModel   *amodel;
  struct tr        *tr;
  tLHMMInitializer *init_hmm;
  calcelprob_t     *cep;
  
};




/*********************/
/* FUNCIONS PRIVADES */
/*********************/

static struct tr *
create_tr (
           unsigned int max_Q
           )
{
  
  struct tr *new;
  struct tr_node *aux;
  unsigned int q;
  
  
  mt ( new= mem_alloc ( struct tr, 1 ) );
  mt ( new->v= mem_alloc ( struct tr_node *, max_Q ) );
  mt ( new->v[0]= mem_alloc ( struct tr_node, max_Q ) );
  for ( q= 1, aux= new->v[0]+1; q < max_Q; ++q, ++aux )
    new->v[q]= aux;
  new->length= 1;
  
  return new;
  
} /* end create_tr */


static void
free_tr (
         struct tr *tr
         )
{
  
  free ( tr->v[0] );
  free ( tr->v );
  free ( tr );
  
} /* end free_tr */


static void
init_tr (
	 struct tr    *tr,
	 unsigned int  max_Q,
	 int           length
	 )
{
  
  size_t nsize;
  unsigned int q;
  struct tr_node *aux;
  
  
  if ( length > tr->length )
    {
      tr->length= length;
      nsize= length * max_Q;
      mt ( tr->v[0]= mem_realloc ( struct tr_node,
				   tr->v[0], nsize ) );
    }
  
  for ( q= 1, aux= tr->v[0]+length;
	q < max_Q; ++q, aux+= length )
    tr->v[q]= aux;
  
} /* end init_tr */


static tLFloat
viterbi (
	 const tLAModel   *amodel,
	 tLAModelHMM       cchar,
	 iamcount_hmm_t    ccount,
	 const tLFea      *fea,
	 const int         frame_begin,
	 const int         length,
	 struct tr        *tr,
	 calcelprob_t     *cep,
	 tLHMMInitializer *init_hmm
	 )
{
  
  int diff, begin, end, t;
  tLProbPair *p;
  struct tr_node *tr0, *tr1;
  unsigned int q, to, from;
  tLProb a, b, max;
  
  
  /* Prepara. */
  diff= length - cchar.Q + 1;
  begin= 0; end= diff;
  p= cchar.trans.lineal.p;
  
  /* q=0 */
  tr1= tr->v[0];
  tr1[0].p= VLPROB ( 0, 0 );
  tr1[0].from= cchar.Q;
  for ( t= 1; t < diff; ++t )
    {
      tr1[t].p= tr1[t-1].p + p[0].p0 + VLPROB ( 0, t );
      tr1[t].from= 0;
    }
  
  /* q>0 */
  for ( q= 1; q < cchar.Q; ++q )
    {
      
      ++begin; ++end;
      tr0= tr1;
      tr1= tr->v[q];
      
      tr1[begin].p=
	tr0[begin-1].p + p[q-1].p1 + VLPROB ( q, begin );
      tr1[begin].from= q-1;
      for ( t= begin+1; t < end; ++t )
	{
	  a= tr0[t-1].p + p[q-1].p1;
	  b= tr1[t-1].p + p[q].p0;
	  if ( a > b )
	    {
	      tr1[t].p= a;
	      tr1[t].from= q-1;
	    }
	  else
	    {
	      tr1[t].p= b;
	      tr1[t].from= q;
	    }
	  tr1[t].p+= VLPROB ( q, t );
	}
      
    }
  
  /* Iteració final. */
  max= tr->v[cchar.Q-1][length-1].p;
  from= cchar.Q-1;
  
  /* Acumula comptadors. */
  if ( max != -HUGE_VAL )
    {
      ++(ccount.trans.lineal.p[from].p1);
      to= from;
      for ( t= length-1; t != 0; --t ) /* La transició inicial la ignore. */
	{
	  init_proto ( init_hmm, cchar.states, ccount.states,
		       to, fea, frame_begin+t );
	  from= tr->v[to][t].from;
	  if ( to == from )
	    ++(ccount.trans.lineal.p[from].p0);
	  else
	    ++(ccount.trans.lineal.p[from].p1);
	  to= from;
	}
      init_proto ( init_hmm, cchar.states, ccount.states,
		   to, fea, frame_begin+t );
    }
  
  return max;
  
} /* end viterbi */




/*********************/
/* FUNCIOS PÚBLIQUES */
/*********************/

TL_API_FUNC(void)
tL_hmm_vtrainer_free (
		      tLHMMVTrainer *cvtrainer
		      )
{
  
  free_calcelprob ( cvtrainer->cep );
  free_tr ( cvtrainer->tr );
  tL_hmm_initializer_free ( cvtrainer->init_hmm );
  free ( cvtrainer );
  
} /* end tL_hmm_vtrainer_free */


TL_API_FUNC(tLHMMVTrainer *)
tL_hmm_vtrainer_new (
		     const tLAModel  *amodel,
		     char           **err
		     )
{
  
  tLHMMVTrainer *new;
  
  
  if ( amodel->tflags != FLINEAL )
    {
      tL_msgerror ( err, _( "tL_hmm_vtrainer_new is only implemented"
                    " for acoustic models with linear topologies" ) );
      return NULL;
    }
  mt ( new= mem_alloc ( tLHMMVTrainer, 1 ) );
  new->amodel= amodel;
  new->init_hmm= tL_hmm_initializer_new ( amodel, err );
  if ( new->init_hmm == NULL ) goto error;
  new->cep= create_calcelprob ( amodel, 0, err );
  if ( new->cep == NULL )
    {
      tL_hmm_initializer_free ( new->init_hmm );
      goto error;
    }
  new->tr= create_tr ( amodel->max_Q );
  
  return new;
  
 error:
   free ( new );
   return NULL;
   
} /* end tL_hmm_vtrainer_new */


TL_API_FUNC(tLProb)
tL_train_viterbi_hmm_seg (
			  tLHMMVTrainer *cvtrainer,
			  tLAMiCounts   *counts,
			  const int      c,
			  const tLFea   *fea,
			  const int      begin,
			  const int      nframes
			  )
{
  
  tLAModelHMM cchar;
  iamcount_hmm_t ccount;
  tLFloat ret;
  
  
  cchar= cvtrainer->amodel->hmms[c];
  if ( nframes < cchar.Q ) return -HUGE_VAL;
  ccount= counts->hmms[c];
  init_tr ( cvtrainer->tr, cvtrainer->amodel->max_Q, nframes );
  init_calcelprob ( cvtrainer->cep, cvtrainer->amodel, fea, begin, nframes );
  ret= viterbi ( cvtrainer->amodel, cchar, ccount, fea, begin, nframes,
		 cvtrainer->tr, cvtrainer->cep, cvtrainer->init_hmm );
  
  return (tLProb) ret;
  
} /* end tL_train_viterbi_hmm_seg */
