/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  train_viterbi_hmm.h - Entrenament a la viterbi d'un
 *                        caràcter. Algorisme de l'antic
 *                        'hinit'. Inicialització a la HTK (HInit).
 *
 *  Adrià Giménez Pastor, 2008-2011,2013
 *
 */

#ifndef __LIBTL_TRAIN_VITERBI_HMM_H__
#define __LIBTL_TRAIN_VITERBI_HMM_H__

#include "amodel_public.h"
#include "amcounts_public.h"
#include "common_public.h"
#include "port.h"
#include "samples_public.h"

/** \addtogroup Algs
 *  @{
 */
/** \defgroup TrainVitHMM Viterbi training of HMMs
 */
/** \addtogroup TrainVitHMM
 *  @{
 */

#ifndef __TL_PRV
/** Used in 'tL_train_viterbi_hmm'. This structure contains auxiliary
 *  data used by 'tL_train_viterbi_hmm' to train an HMM
 *  using the Viterbi algorithm.
 */
struct _tLHMMVTrainer
{
  
  const tLAModel *amodel;    /**< Reference to the acoustic model
                                used by the model trainer. */
};
#endif

/** Used in 'tL_train_viterbi_hmm'. This structure contains auxiliary
 *  data used by 'tL_train_viterbi_hmm' to train an HMM
 *  using the Viterbi algorithm.
 */
typedef struct _tLHMMVTrainer tLHMMVTrainer;

/** Frees memory. Frees the memory allocated for the HMM Viterbi
 * trainer.
 *
 *  @param hvtrainer The HMM Viterbi trainer.
 */
TL_API_FUNC(void)
tL_hmm_vtrainer_free (
		      tLHMMVTrainer *hvtrainer
		      );

/** Gets a tLHMMVTrainer. This function creates an HMM Viterbi
 *  trainer used to train HMMs with the Viterbi algorithm.
 *  Currently supported types are: See tLHMMInitializer.
 *
 *  @param amodel The acoustic model.
 *  @param err Pointer to string variable. If not NULL, an error message
 *  is allocated in the variable in case of error.
 *  @return The HMM Viterbi trainer, or NULL in case of error.
 */
TL_API_FUNC(tLHMMVTrainer *)
tL_hmm_vtrainer_new (
		     const tLAModel  *amodel,
		     char           **err
		     );

/** Trains an HMM from a sample segment using the
 *  Viterbi algorithm. This function finds the maximum probability
 *  path for the specified HMM of input frames, which are a segment
 *  of the given feature vector sequence, and for each transition used
 *  it increases the related count by one. Furthermore, the maximum
 *  probability path is used to assign the input frames to states,
 *  then, states counts are increased as in tL_init_hmm.
 *
 *  @param hvtrainer The HMM Viterbi trainer.
 *  @param counts Integer counts related to amodel.
 *  @param c The index of the HMM to be trained.
 *  @param fea A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the amodel.
 *  @param begin The first input frame.
 *  @param nframes The number of input frames.
 *  @return The log-probability of the maximum probability path.
 */
TL_API_FUNC(tLProb)
tL_train_viterbi_hmm_seg (
			  tLHMMVTrainer *hvtrainer,
			  tLAMiCounts   *counts,
			  const int      c,
			  const tLFea   *fea,
			  const int      begin,
			  const int      nframes
			  );

/** Trains an HMM from a sample using the Viterbi
 *  algorithm. This macro is equivalent to tL_train_viterbi_hmm_seg
 *  but using all frames contained in fea.
 *
 *  @param HVTRAINER The HMM Viterbi trainer.
 *  @param COUNTS Integer counts related to AMODEL.
 *  @param C The index of the HMM to be trained.
 *  @param FEA A feature vector manager, which must contain a
 *  sequence of feature vectors of an adequate type for the AMODEL.
 *  @return The log-probability of the maximum probability path.
 */
#define tL_train_viterbi_hmm(HVTRAINER,COUNTS,C,FEA)			\
  tL_train_viterbi_hmm_seg ( HVTRAINER, COUNTS, C, FEA, 0, (FEA)->cols )

/** @} */
/** @} */

#endif /* __LIBTL_TRAIN_VITERBI_HMM_H__ */
