/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  wghypmanager.h - Un gestor d'hipòtesis senzill per a generar un
 *                   'word graph'.
 *
 *  Adrià Giménez Pastor, 2010,2011,2012.
 *
 */

#ifndef __TL_PRV_WGHYPMANAGER_H__
#define __TL_PRV_WGHYPMANAGER_H__

#include <stddef.h>
#include <stdio.h>

#include "common.h"
#include "lexicon.h"
#include "lm.h"
#include "text.h"
#include "wordgraph.h"


/* DEFINICIÓ */

#define WGNULLHYP 0

struct wghhnode
{
  
  tLProb        p;           /* Probabilitat dela hipòtesis. */
  tLProb        hmmp;        /* Probabilitat dels models de Markov+lèxic. */
  tLProb        lmp;         /* Probabilitat de l'aresta del model de
				llenguatge amb la que s'ha feta la
				transició. */
  size_t        e;           /* Entrada del lèxic emprada. NULL_ENTRY
				quan la paraula és nul·la o símbol
				especial. */
  unsigned long w;           /* Paraula. */
  size_t        ptr;         /* Punter a la hipòtesis pare. */
  
};

struct wghypmng_node
{
  
  unsigned long    lms;
  int              t;          /* Estos dos camps denoten
				  l'identificador d'estat, que vol dir
				  que en l'instant 't' ha acabat en
				  l'estat 'lms' l'estat associat. */
  size_t           counter;    /* Comptador associat. */
  unsigned int     N;          /* Número d'hipòtesis pare que conté,
				  el màxim ve donat per la K del
				  gestor. 0 vol dir que ptr apunta a
				  una hipòtesis pare, > 0 que apunta a
				  un minheap on estan. */
  size_t           ptr;        /* Punter al pare o al minheap dels
				  pares. */
  size_t           id;         /* S'empra per a recuperar el
				  wordgraph, durant la generació no té
				  sentit. */
  char             flag;       /* Per a saber durant la generació del
				  graf si ja l'he visitat o que. */
  
};

struct wghypmng_stack
{
  
  size_t *v;
  size_t  s;
  size_t  N;
  
};

typedef struct
{
  
  unsigned int            K;
  struct wghypmng_node   *v;
  struct wghhnode       **hnodes;
  size_t                  s,shnodes;
  struct wghypmng_stack   stack;
  size_t                  empty,emptyhnodes;
  
} wghypmng_t;


/* FUNCIONS */

/* Es supossa una K>0. */
wghypmng_t *
tL_create_wghypmng (
		    const unsigned int K
		    );

void
tL_free_wghypmng (
		  wghypmng_t *wghypmng
		  );

/* Elimina una hipòtesi, independentment de si està referenciada o no.
 */
void
tL_free_wghyp (
	       wghypmng_t *wghypmng,
	       size_t      hyp
	       );

/* Torna el reconeixement com si haguerem gastat hypmng. */
size_t
tL_get_wghyp (
	      wghypmng_t        *wghypmng,
	      size_t             hyp,
	      unsigned long int *buffer,
	      const tLBool       align,
	      const tLBool       iscw,
	      const tLBool       wsil
	      );

/* Si no té pare el pare deu de ser WGNULLHYP. */
size_t
tL_wghypmng_add (
		 wghypmng_t    *wghypmng,
		 size_t         parent,
		 unsigned long  lms,        /* Estat model
					       llenguatge. ULONG_MAX
					       indica estat final. */
		 int            t,          /* Instant. */
		 tLProb         p,          /* Prob. camí. */
		 tLProb         hmmp,       /* Prob. HMM+LEXIC. */
		 tLProb         lmp,        /* Prob. LM. */
		 unsigned long  w,          /* Paraula. ULONG_MAX
					       paraula final. Talla
					       vocabulari paraula
					       buida. */
		 size_t         e           /* Entrada del lèxic. */
		 );

/* Si no té pare el pare deu de ser WGNULLHYP. Es gasta sols per a la
   segmentació a nivell de símbol. NOTA: t=-1234 és un valor reservat
   que no s'ha de gastar en un segment. */
size_t
tL_wghypmng_add_seg (
		     wghypmng_t    *wghypmng,
		     size_t         parent,
		     int            t          /* Instant. */
		     );

/* Si no té pare el pare deu de ser WGNULLHYP. Es gasta sols per a
   insertar els símbols de final i principi de paraula quan s'està
   fent crosswords. */
size_t
tL_wghypmng_add_sym (
		     wghypmng_t *wghypmng,
		     size_t      parent,
		     size_t      sym
		     );

/* Inserta un altre camí en la hipòtesis. */
void
tL_wghypmng_insert (
		    wghypmng_t    *wghypmng,
		    size_t         hyp,
		    size_t         parent,
		    tLProb         p,          /* Prob. camí. */
		    tLProb         hmmp,       /* Prob. HMM+LEXIC. */
		    tLProb         lmp,        /* Prob. LM. */
		    unsigned long  w,          /* Paraula. ULONG_MAX
						  paraula final. Talla
						  vocabulari paraula
						  buida. */
		    size_t         e           /* Entrada del lèxic. */
		    );

/* Ompli el wordgraph. Si apunta a NULL el crea. El lexicon sols es
   passa per a generar segments. */
void
tL_wghypmng_get_wordgraph (
			   wghypmng_t       *wghypmng,
			   size_t            hyp,
			   const tLLM       *lm,
			   tLWordGraph     **wg,
			   const tLLexicon  *lexicon, /* Pot ser NULL*/
			   const tLBool      wsil,
			   const size_t      sil
			   );

#define release_wghyp(WGHYPMNG,HYP)                               \
  do {								  \
     if ( --((WGHYPMNG)->v[(HYP)].counter) == 0 && (HYP) != 0 )	  \
       tL_free_wghyp ( (WGHYPMNG), (HYP) );                          \
   } while(0)

#define take_wghyp(WGHYPMNG,HYP) ++((WGHYPMNG)->v[(HYP)].counter)


#endif /* __TL_PRV_WGHYPMANAGER_H__ */
