/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  wordgraph.c - Implementació de 'wordgrap.h'
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */


#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "common.h"
#include "error.h"
#include "mem.h"
#include "text.h"
#include "wordgraph.h"




/**********/
/* MACROS */
/**********/

#define READ_TOKEN(TOK)							\
  do {									\
    if ( !tL_gtoken ( from, buffer ) || strcmp ( TOK, buffer->v ) )	\
      {									\
	TLMSG_GTOK ( err, TOK );					\
	goto error;							\
      }									\
  } while(0)
#define READ_VAL(DESC)					\
  do {							\
    if ( !tL_gtoken ( from, buffer ) )			\
      {							\
	TLMSG_GDESC ( err, _( "the number of states NS" ) );	\
	goto error;					\
      }							\
  } while(0)

#define READ_ED_TOKEN(ED,TOK)						\
  do {									\
    if ( !tL_gtoken ( from, buffer ) || strcmp ( TOK, buffer->v ) )	\
      {									\
	tL_msgerror ( err, _( "In %luth edge: unabe to read the token %s" ) ,	\
		      _PS(ED)+1, TOK );					\
	goto error;							\
      }									\
  } while(0)
#define READ_ED_VAL(ED,DESC)						\
  do {									\
    if ( !tL_gtoken ( from, buffer ) )					\
      {									\
	tL_msgerror ( err, _( "In %luth edge: unable to read %s" ) ,		\
		      _PS(ED)+1, DESC );				\
	goto error;							\
      }									\
  } while(0)
#define READ_SEG_VAL(ED,SEG,DESC)					\
  do {									\
    if ( !tL_gtoken ( from, buffer ) )					\
      {									\
	tL_msgerror ( err, _( "In %luth segment of %luth edge:"		\
		      " unable to read %s" ) ,				\
		      _PS(SEG)+1, _PS(ED)+1, DESC );			\
	goto error;							\
      }									\
  } while(0)




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(void)
tL_wordgraph_free (
		   tLWordGraph *wg
		   )
{
  
  free ( wg->_nodes );
  if ( wg->_maxNSG != 0 )
    free ( wg->_segs );
  free ( wg->edges );
  free ( wg->states );
  free ( wg );
  
} /* end tL_wordgraph_free */


TL_API_FUNC(int)
tL_wordgraph_load (
		   tLWordGraph  **wg,
		   gzFile         from,
		   const tLDict  *words,
		   const tLDict  *syms,
		   const char    *end_sym,
		   char         **err
		   )
{
  
  tLBuffer *buffer;
  tLFloat gsf, wip;
  size_t NS, NE, NSG, start, end, n, src, s, total_segs;
  tLWordGraph *ret;
  tLWGEdge edge;
  tLWGList *nodes;
  tLWGSegment *segs, seg;
  
  
  buffer= tL_buffer_new ();
  
  /* Capçalera. */
  if ( !tL_gtoken ( from, buffer ) )
    {
      TLMSG_TOKID ( err );
      goto error;
    }
  if ( strcmp ( "WORDGRAPH", buffer->v ) )
    {
      TLMSG_UNKTOKID ( err, buffer->v );
      goto error; 
    }
  READ_TOKEN ( "GSF" );
  READ_VAL ( _( "the grammar scale factor GSF" ) );
  gsf= atof ( buffer->v );
  READ_TOKEN ( "WIP" );
  READ_VAL ( _( "the word insertion penalty WIP" ) );
  wip= atof ( buffer->v );
  READ_TOKEN ( "NS" );
  READ_VAL ( _( "the number of states NS" ) );
  NS= atoui ( buffer->v );
  READ_TOKEN ( "NE" );
  READ_VAL ( _( "the number of edges NE" ) );
  NE= atoui ( buffer->v );
  READ_TOKEN ( "NSG" );
  READ_VAL ( _( "the number of segments NSG" ) );
  NSG= atoui ( buffer->v );
  READ_TOKEN ( "START" );
  READ_VAL ( _( "the initial state START" ) );
  start= atoui ( buffer->v );
  if ( start >= NS )
    {
      tL_msgerror ( err, _( "Initial state is out of range" ) );
      goto error;
    }
  READ_TOKEN ( "END" );
  READ_VAL ( _( "the final state END" ) );
  end= atoui ( buffer->v );
  if ( end >= NS )
    {
      tL_msgerror ( err, _( "Final state is out of range" ) );
      goto error;
    }
  
  /* Ajusta memòria. */
  if ( *wg == NULL ) *wg= tL_wordgraph_alloc ( NS, NE, NSG );
  else tL_wordgraph_realloc ( *wg, NS, NE, NSG );
  ret= *wg;
  ret->NS= NS; ret->NE= NE; ret->NSG= NSG;
  ret->gsf= gsf; ret->wip= wip;
  ret->start= &(ret->states[start]);
  ret->end= &(ret->states[end]);
  nodes= ret->_nodes;
  segs= ret->_segs;
  
  /* Llig estats. */
  READ_TOKEN ( "STATES" );
  for ( n= 0; n < NS; ++n )
    {
      if ( !tL_gtoken ( from, buffer ) )
	{
	  tL_msgerror ( err, _( "Unable to read the time instant of %luth state" ) ,
			_PS(n)+1 );
	  goto error;
	}
      ret->states[n].t= atoi ( buffer->v );
      ret->states[n].oedges= NULL;
      ret->states[n].iedges= NULL;
    }
  
  /* Llig arestes. */
  total_segs= 0;
  READ_TOKEN ( "EDGES" );
  for ( n= 0; n < NE; ++n )
    {
      
      /* Origen. */
      READ_ED_TOKEN ( n, "S" );
      READ_ED_VAL ( n, _( "the source state S" ) );
      src= atoui ( buffer->v );
      if ( src >= NS )
	{
	  tL_msgerror ( err, _( "In %luth edge: source state is out of range" ) ,
			_PS(n)+1 );
	  goto error;
	}
      edge.from= &(ret->states[src]);
      nodes->edge= &(ret->edges[n]);
      nodes->next= ret->states[src].oedges;
      ret->states[src].oedges= nodes++;
      
      /* Destí. */
      READ_ED_TOKEN ( n, "E" );
      READ_ED_VAL ( n, _( "the end state E" ) );
      end= atoui ( buffer->v );
      if ( end >= NS )
	{
	  tL_msgerror ( err, _( "In %luth edge: end state is out of range" ) ,
			_PS(n)+1 );
	  goto error;
	}
      edge.to= &(ret->states[end]);
      nodes->edge= &(ret->edges[n]);
      nodes->next= ret->states[end].iedges;
      ret->states[end].iedges= nodes++;
      
      /* Paraula. */
      READ_ED_VAL ( n, _( "the word" ) );
      if ( !strcmp ( "!NULL", buffer->v ) ) edge.w= ULONG_MAX;
      else if ( !strcmp ( end_sym==NULL?"</s>":end_sym, buffer->v ) )
	edge.w= words->N;
      else if ( (edge.w= tL_dict_find ( words, buffer->v )) == words->N )
	{
	  tL_msgerror ( err, _( "In %luth edge: word '%s' is"
			" out of vocabulary" ) , _PS(n)+1, buffer->v );
	  goto error;
	}
      
      /* Acoustic probability. */
      READ_ED_TOKEN ( n, "A" );
      READ_ED_VAL ( n, _( "the acoustic probability A" ) );
      edge.a= atof ( buffer->v );
      
      /* LM probability. */
      READ_ED_TOKEN ( n, "L" );
      READ_ED_VAL ( n, _( "the LM probability L" ) );
      edge.l= atof ( buffer->v );
      
      /* NSegs probability. */
      READ_ED_TOKEN ( n, "N" );
      READ_ED_VAL ( n, _( "the number of segments" ) );
      edge.nsegs= atoui ( buffer->v );
      if ( (total_segs+= edge.nsegs) > NSG )
	{
	  tL_msgerror ( err, _( "In %luth edge: total number"
			" of segments overflow" ) , _PS(n)+1 );
	  goto error;
	}
      
      /* Segments. */
      edge.segs= segs;
      if ( edge.nsegs )
	{
	  for ( s= 0; s < edge.nsegs; ++s )
	    {
	      
	      /* Symbol. */
	      READ_SEG_VAL ( n, s, _( "the symbol" ) );
	      if ( (seg.sym= tL_dict_find ( syms, buffer->v )) == syms->N )
		{
		  tL_msgerror ( err, _( "In %luth segment of %luth edge:"
				" symbol '%s' is out of vocabulary" ) ,
				_PS(s)+1, _PS(n)+1, buffer->v );
		  goto error;
		}
	      
	      /* Position. */
	      READ_SEG_VAL ( n, s, _( "the initial position" ) );
	      seg.pos= atoi ( buffer->v );
	      
	      /* Number of frames. */
	      READ_SEG_VAL ( n, s, _( "the number of frames" ) );
	      seg.nframes= atoi ( buffer->v );
	      
	      segs[s]= seg;
	      
	    }
	  segs+= edge.nsegs;
	}
      
      ret->edges[n]= edge;
      
    }
  
  tL_buffer_free ( buffer );
  
  return 0;
  
 error:
  tL_buffer_free ( buffer );
  return -1;
  
} /* end tL_wordgraph_load */


TL_API_FUNC(void)
tL_wordgraph_print (
		    const tLWordGraph *wg,
		    FILE              *to,
		    const tLDict      *words,
		    const tLDict      *syms,
		    const char        *end_sym
		    )
{
  
  size_t n,s;
  tLWGEdge ed;
  tLWGSegment seg;
  
  
  fprintf ( to, "WORDGRAPH\n" );
  fprintf ( to, "GSF %g WIP %g\n", _DB(wg->gsf), _DB(wg->wip) );
  fprintf ( to, "NS %lu NE %lu NSG %lu\n",
	    _PS(wg->NS), _PS(wg->NE), _PS(wg->NSG) );
  fprintf ( to, "START %lu END %lu\n",
	    _PS(wg->start-wg->states),
	    _PS(wg->end-wg->states) );
  fprintf ( to, "STATES\n" );
  for ( n= 0; n < wg->NS; ++n  )
    fprintf ( to, "%d\n", wg->states[n].t );
  fprintf ( to, "EDGES\n" );
  for ( n= 0; n < wg->NE; ++n )
    {
      ed= wg->edges[n];
      fprintf ( to, "S %lu E %lu %s A %.10f L %.10f N %lu",
		_PS(ed.from-wg->states), _PS(ed.to-wg->states),
		(ed.w==ULONG_MAX)?"!NULL":
		(ed.w==words->N?
		 (end_sym==NULL?"</s>":end_sym):
		 tL_dict_get_token ( words, ed.w )),
		_DB(ed.a), _DB(ed.l), _PS(ed.nsegs) );
      for ( s= 0; s < ed.nsegs; ++s )
	{
	  seg= ed.segs[s];
	  fprintf ( to, " %s %d %d",
		    tL_dict_get_token ( syms, seg.sym ),
		    seg.pos, seg.nframes );
	}
      fputc ( '\n', to );
    }
  
} /* end tL_wordgraph_print */


TL_API_FUNC(void)
tL_wordgraph_print_htk (
			const tLWordGraph *wg,
			FILE              *to,
			const tLDict      *words,
			const tLDict      *syms,
			const char        *feaname,
			const char        *start_sym,
			const char        *end_sym
			)
{
  
  size_t n,s;
  tLWGEdge ed;
  tLWGSegment seg;
  
  
  fprintf ( to, "VERSION=1.0\n" );
  fprintf ( to, "UTTERANCE=%s\n", feaname );
  fprintf ( to, "lmscale=%g  wdpenalty=%g\n", _DB(wg->gsf), _DB(wg->wip) );
  fprintf ( to, "acscale=1.00\n" );
  fprintf ( to, "N=%lu L=%lu\n", _PS(wg->NS)+1, _PS(wg->NE)+1 );
  fprintf ( to, "I=0  t=0\n" );
  for ( n= 0; n < wg->NS; ++n  )
    fprintf ( to, "I=%lu  t=%d\n", _PS(n)+1, wg->states[n].t );
  fprintf ( to, "J=0 S=0 E=%lu W=%s a=0.0 l=0.0\n",
	    _PS(wg->start-wg->states)+1,
	    (start_sym==NULL)?"<s>":start_sym  );
  for ( n= 0; n < wg->NE; ++n )
    {
      ed= wg->edges[n];
      fprintf ( to, "J=%lu S=%lu E=%lu W=%s a=%.10f l=%.10f", _PS(n)+1,
		_PS(ed.from-wg->states)+1, _PS(ed.to-wg->states)+1,
		(ed.w==ULONG_MAX)?"!NULL":
		(ed.w==words->N?
		 (end_sym==NULL?"</s>":end_sym):
		 tL_dict_get_token ( words, ed.w )),
		_DB(ed.a), _DB(ed.l) );
      if ( ed.nsegs != 0 ) fprintf ( to, " :" );
      for ( s= 0; s < ed.nsegs; ++s )
	{
	  seg= ed.segs[s];
	  fprintf ( to, "%s,%d,%d:",
		    tL_dict_get_token ( syms, seg.sym ),
		    seg.pos, seg.pos+seg.nframes-1 );
	}
      fputc ( '\n', to );
    }
  
} /* end tL_wordgraph_print_htk */


tLWGEdge *
tL_wordgraph_add_edge (
		       tLWordGraph         *wg,
		       const size_t         from,
		       const size_t         to,
		       const unsigned long  word
		       )
{
  
  size_t ed;
  tLWGEdge *edge;
  tLWGList *node;
  
  
  edge= &(wg->edges[ed= (wg->NE)++]);
  edge->w= word;
  edge->a= edge->l= 0.0;
  edge->from= &(wg->states[from]);
  edge->to= &(wg->states[to]);
  edge->segs= NULL;
  edge->nsegs= 0;
  
  ed*= 2;
  node= &(wg->_nodes[ed]);
  node->next= edge->from->oedges;
  edge->from->oedges= node;
  node->edge= edge;
  
  node= &(wg->_nodes[ed+1]);
  node->next= edge->to->iedges;
  edge->to->iedges= node;
  node->edge= edge;
  
  return edge;
  
} /* end tL_wordgraph_add_edge */


tLWordGraph *
tL_wordgraph_alloc (
		    const size_t NS,
		    const size_t NE,
		    const size_t NSG
		    )
{
  
  tLWordGraph *new;
  
  
  mt ( new= mem_alloc ( tLWordGraph, 1 ) );
  new->_maxNS= NS;
  new->_maxNE= NE;
  new->_maxNSG= NSG;
  mt ( new->states= mem_alloc ( tLWGState, NS ) );
  mt ( new->edges= mem_alloc ( tLWGEdge, NE ) );
  mt ( new->_nodes= mem_alloc ( tLWGList, 2*NE ) );
  if ( NSG > 0 )
    mt ( new->_segs= mem_alloc ( tLWGSegment, NSG ) );
  
  return new;
  
} /* end tL_wordgraph_alloc */


void
tL_wordgraph_realloc (
		      tLWordGraph  *wg,
		      const size_t  NS,
		      const size_t  NE,
		      const size_t  NSG
		      )
{
  
  if ( NS > wg->_maxNS )
    {
      mt ( wg->states= mem_realloc ( tLWGState, wg->states, NS ) );
      wg->_maxNS= NS;
    }
  if ( NE > wg->_maxNE )
    {
      mt ( wg->edges= mem_realloc ( tLWGEdge, wg->edges, NE ) );
      mt ( wg->_nodes= mem_realloc ( tLWGList, wg->_nodes, 2*NE ) );
      wg->_maxNE= NE;
    }
  if ( NSG > wg->_maxNSG )
    {
      if ( wg->_maxNSG == 0 )
	{
	  mt ( wg->_segs= mem_alloc ( tLWGSegment, NSG ) );
	}
      else
	{
	  mt ( wg->_segs=
	       mem_realloc ( tLWGSegment, wg->_segs, NSG ) );
	}
      wg->_maxNSG= NSG;
    }
  
} /* end tL_wordgraph_realloc */
