/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  wordgraph_public.h - Representa un grag de paraules.
 *
 *  Adrià Giménez Pastor, 2011,2012.
 *
 */

#ifndef __LIBTL_WORDGRAPH_H__
#define __LIBTL_WORDGRAPH_H__

#include <stddef.h>
#include <stdio.h>
#include <zlib.h>

#include "common_public.h"
#include "port.h"
#include "text_public.h"

/** \addtogroup Basic
 *  @{
 */

typedef struct tLWGState tLWGState;
typedef struct tLWGSegment tLWGSegment;
typedef struct tLWGEdge tLWGEdge;
typedef struct tLWGList tLWGList;

/** Word graph state. This is the structure used to represent the
 *  states in a word graph.
 */
struct tLWGState
{
  
  int       t;         /**< Time instant. */
  tLWGList *oedges;    /**< Outgoing edges. */
  tLWGList *iedges;    /**< Incoming edges. */
  
};

/** Word graph segment. This structure is used to represent the start
 *  and end position of a symbol.
 */
struct tLWGSegment
{
  
  unsigned int sym;        /**< Symbol. */
  int          pos;        /**< Position of the first frame. */
  int          nframes;    /**< Number of frames in the segment. */
  
};

/** Word graph edge. This is the structure used to represent the edges
 *  in a word graph.
 */
struct tLWGEdge
{
  
  unsigned long  w;        /**< Encoded word of the transition. A
			      value equal to the size of the
			      dictionary of words used to generate the
			      word graph is interpreted as the
			      special final word. ULONG_MAX denotes
			      the special lambda (empty) word. */
  tLProb         a;        /**< Log-probability related to the
			      acoustic (and lexicon) models. */
  tLProb         l;        /**< Log-probability related to the
			      language model. */
  tLWGState     *from;     /**< Source state. */
  tLWGState     *to;       /**< Destination state. */
  tLWGSegment   *segs;     /**< Word segmentation, can be NULL. */
  size_t         nsegs;    /**< Number of segments. */
  
};

/** List node. Node used to build lists of word graph edges. */
struct tLWGList
{
  
  tLWGList *next;    /**< Next node of the list. */
  tLWGEdge *edge;    /**< Edge. */
  
};

/** Word graph. A word graph is a collection of hypotheses represented
 *  as a graph, where states represent instants of time and the edges
 *  represent recognized words. Additionally, the probability of each
 *  word is provided, divided into language model probability and
 *  acoustic model probability.
 */
typedef struct
{

  tLFloat      gsf;       /**< The grammar scale factor which was used
			     to generate this word graph. */
  tLFloat      wip;       /**< The word insertion penalty which was
			     used to generate this word graph. */
  size_t       NS;        /**< Number of states. */
  size_t       NE;        /**< Number of edges. */
  size_t       NSG;       /**< Number of segments. */
  tLWGState   *start;     /**< Initial state. */
  tLWGState   *end;       /**< Final state. */
  tLWGState   *states;    /**< States of the word graph. */
  tLWGEdge    *edges;     /**< Edges of the word graph. */
  tLWGList    *_nodes;
  tLWGSegment *_segs;
  size_t       _maxNS;
  size_t       _maxNE;
  size_t       _maxNSG;
  
} tLWordGraph;

/** Frees memory. Frees the memory allocated for the word graph.
 *
 *  @param wg The word graph.
 */
TL_API_FUNC(void)
tL_wordgraph_free (
		   tLWordGraph *wg
		   );

/** Loads a word graph. Loads a word graph from a text description.
 *
 *  @param wg Pointer to the memory address where the word graph will
 *  be loaded. If it points to NULL, a new word graph will be created.
 *  @param from File where the text description is stored.
 *  @param words Dictionary containing the words.
 *  @param syms Dictionary containing the symbols. Only required in
 *  the case of words graphs with segments.
 *  @param end_sym Token used to represent the special final word.
 *  NULL is equivalent to "</s>".
 *  @param err Pointer to string variable. If not NULL, an error
 *  message is allocated in the variable in case of error.
 *  @return -1 in case of error, 0 otherwise.
 */
TL_API_FUNC(int)
tL_wordgraph_load (
		   tLWordGraph  **wg,
		   gzFile         from,
		   const tLDict  *words,
		   const tLDict  *syms,
		   const char    *end_sym,
		   char         **err
		   );

/** Prints the word graph. This function writes in the given file a
 *  text representation of the word graph. The provided word
 *  dictionary is assumed to contain all needed words, otherwise an
 *  unexpected error could happen.
 *
 *  @param wg The word graph.
 *  @param to File to which the word graph is written.
 *  @param words Dictionary containing the words.
 *  @param syms Dictionary containing the symbols. Not required if the
 *  word graph doesn't contain any segments.
 *  @param end_sym Token used to print the special final word. NULL
 *  is equivalent to "</s>".
 */
TL_API_FUNC(void)
tL_wordgraph_print (
		    const tLWordGraph *wg,
		    FILE              *to,
		    const tLDict      *words,
		    const tLDict      *syms,
		    const char        *end_sym
		    );

/** Prints the word graph using the HTK format. This function writes
 *  to the given file a text representation of the word graph using
 *  the HTK format. The provided word dictionary is assumed to
 *  contain all needed words, otherwise an unexpected error
 *  could happen.
 *
 *  @param wg The word graph.
 *  @param to File to which the word graph is written.
 *  @param words Dictionary containing the words.
 *  @param syms Dictionary containing the symbols. Not required if the
 *  word graph doesn't contain any segments.
 *  @param feaname A string with the name of the feature related to
 *  the word graph.
 *  @param start_sym Token used to print the special initial word. NULL
 *  is equivalent to "<s>".
 *  @param end_sym Token used to print the special final word. NULL
 *  is equivalent to "</s>".
 */
TL_API_FUNC(void)
tL_wordgraph_print_htk (
			const tLWordGraph *wg,
			FILE              *to,
			const tLDict      *words,
			const tLDict      *syms,
			const char        *feaname,
			const char        *start_sym,
			const char        *end_sym
			);

/** @} */

#endif /* __LIBTL_WORDNET_H__ */
