/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  wordnet.c - Implementació de 'wordnet.h'.
 *
 *  Adrià Giménez Pastor, 2009,2011.
 *
 */


#include <math.h>
#include <limits.h>
#include <stddef.h>
#include <stdlib.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "../intl.h"

#include "wordnet.h"
#include "common.h"
#include "error.h"
#include "mem.h"
#include "text.h"




/*********/
/* TIPUS */
/*********/

struct lnode
{
  
  size_t next;
  size_t ed;
  
};


struct edlist
{
  
  size_t ed;
  size_t ned;
  
};




/**********************/
/* FUNCIONS PÚBLIQUES */
/**********************/

TL_API_FUNC(tLLM *)
tL_lm_new_from_wordnet_file (
			     gzFile         from,
			     tLDict        *vocab,
			     const tLBool   emptyvocab,
			     const char    *begin_sym,
			     const char    *end_sym,
			     char         **err
			     )
{
  
  tLLM *new;
  tLBuffer *buffer;
  int ok, ret;
  char *begin, *end;
  size_t line, state, n, to, NE, p, ed, endstate,
    *counts, aux;
  struct edlist *edlists;
  struct lnode *lnodes;
  tLLMState *states;
  tLLMEdge *edges, *aux_edges;
  
  const tLLMState state0= {{NULL,NULL},
			    {NULL,NULL},
			    ULONG_MAX,
			    -HUGE_VAL};
  
  
  /* Inicialitza. */
  new= mem_alloc ( tLLM, 1 );
  new->states= NULL;
  new->edges= NULL;
  new->backoff= FALSE;
  buffer= tL_buffer_new ();
  aux_edges= NULL;
  edlists= NULL;
  lnodes= NULL;
  counts= NULL;
  
  /* Capçalera. */
  while ( (ok= tL_gline ( from, buffer )) &&
          (buffer->v[0] == '\0' ||
           buffer->v[0] == '#' ||
           !strcmp ( buffer->v, "VERSION=1.0" )) );
  if ( !ok )
    {
      tL_msgerror ( err, _( "Unable to read the header" ) );
      goto error;
    }
  if ( tL_gtokenstr ( buffer->v, &begin, &end ) ) goto errorheadf;
  *end= '\0';
  if ( *begin != 'N' || *(begin+1) != '=' ) goto errorheadf;
  new->NS= atos ( begin+2 );
  if ( new->NS == 0 )
    {
      tL_msgerror ( err, _( "The number of states cannot be 0" ) );
      goto error;
    }
  tL_gtokenstr ( end+1, &begin, &end );
  if ( begin == NULL || *begin == '\0' ||
       *begin != 'L' || *(begin+1) != '=' ) goto errorheadf;
  new->NE= atos ( begin+2 );
  if ( new->NE == 0 )
    {
      tL_msgerror ( err, _( "The number of edges cannot be 0" ) );
      goto error;
    }
  states= new->states= mem_alloc ( tLLMState, new->NS );
  for ( n= 0; n < new->NS; ++n ) states[n]= state0;
  edges= new->edges= mem_alloc ( tLLMEdge, new->NE );
  
  /* Estats. */
  line= 0;
  while ( (ok= tL_gline ( from, buffer )) && line < new->NS )
    {
      if ( buffer->v[0] == '\0' ||
           buffer->v[0] == '#' ) continue;
      if ( tL_gtokenstr ( buffer->v, &begin, &end ) ) goto errorstatesf;
      *end= '\0';
      if ( *begin != 'I' || *(begin+1) != '=' ) goto errorstatesf;
      state= atos ( begin+2 );
      if ( state >= new->NS )
	{
	  tL_msgerror ( err, _( "State %lu is out of range" ) , _PS(state) );
	  goto error;
	}
      tL_gtokenstr ( end+1, &begin, &end );
      if ( begin == NULL || *begin == '\0' ||
           *begin != 'W' || *(begin+1) != '=' ||
	   *(begin+2) == '\0' ) goto errorstatesf;
      *end= '\0'; begin+= 2;
      if ( strcmp ( begin, "!NULL" ) &&
	   strcmp ( begin, begin_sym ) &&
	   strcmp ( begin, end_sym ) )
	{
	  if ( emptyvocab ) tL_dict_insert ( vocab, begin, &aux );
	  else
	    {
	      aux= tL_dict_find ( vocab, begin );
	      if ( aux == vocab->N )
		{
		  tL_msgerror ( err, _( "%s is out of vocabulary" ) , begin );
		  goto error;
		}
	    }
	  states[state].word= (unsigned long) aux;
        }
      ++line;
    }
  if ( !ok )
    {
      if ( line < new->NS ) tL_msgerror ( err, _( "Unable to read all states" ) );
      else tL_msgerror ( err, _( "Unable to read the edges" ) );
      goto error;
    }
  
  /* Arestes. Les fica primer en un vector desordenat
     per a després ordenarles. */
  aux_edges= mem_alloc ( tLLMEdge, new->NE );
  lnodes= mem_alloc ( struct lnode, new->NE+1 );
  edlists= mem_alloc ( struct edlist, new->NS );
  for ( n= 0; n < new->NS; ++n )
    edlists[n].ed= edlists[n].ned= 0;
  line= 0; n= 1;
  while ( ok && line < new->NE )
    {
      if ( buffer->v[0] == '\0' ||
           buffer->v[0] == '#' ) continue;
      if ( tL_gtokenstr ( buffer->v, &begin, &end ) ) goto erroredgesf;
      if ( tL_gtokenstr ( end+1, &begin, &end ) ) goto erroredgesf;
      *end= '\0';
      if ( *begin != 'S' || *(begin+1) != '=' ) goto erroredgesf;
      state= atos ( begin+2 );
      if ( state >= new->NS )
	{
	  tL_msgerror ( err, _( "The outgoing state %lu in the %luth edge is"
			" out of range" ) , _PS(state), _PS(line)+1 );
	  goto error;
	}
      ret= tL_gtokenstr ( end+1, &begin, &end );
      if ( begin == NULL || *begin == '\0' ||
           *begin != 'E' || *(begin+1) != '=' ) goto erroredgesf;
      *end= '\0';
      to= atos ( begin+2 );
      if ( to >= new->NS )
	{
	  tL_msgerror ( err, _( "The target state %lu in the %luth edge is"
			" out of range" ) , _PS(to), _PS(line)+1 );
	  goto error;
	}
      if ( states[to].word == ULONG_MAX )
	{
	  lnodes[n].next= edlists[state].ned;
	  edlists[state].ned= n;
	}
      else
	{
	  lnodes[n].next= edlists[state].ed;
	  edlists[state].ed= n;
	}
      lnodes[n].ed= line;
      aux_edges[line].to= to;
      aux_edges[line].p= 0.0;
      if ( !ret )
	{
	  tL_gtokenstr ( end+1, &begin, &end );
	  if ( begin != NULL )
	    {
	      if ( *begin == '\0' || *begin != 'l' ||
		   *(begin+1) != '=' ) goto erroredgesf;
	      aux_edges[line].p= (tLProb) atof ( begin+2 );
	    }
	}
      ++line; ++n;
      ok= tL_gline ( from, buffer );
    }
  if ( line != new->NE )
    {
      tL_msgerror ( err, _( "The number of expected edges (%lu) and read"
		    " edges (%lu) mismatch" ) , new->NE, _PS(line) );
      goto error;
    }
  
  /* Busca l'estat final i l'elimina. */
  for ( n= 0; n < new->NS && 
	  (edlists[n].ed != 0 || edlists[n].ned != 0);
	++n );
  if ( n == new->NS )
    {
      tL_msgerror ( err, _( "There is not a final state in the wordnet" ) );
      goto error;
    }
  endstate= n++;
  for ( ; n < new->NS && 
	  (edlists[n].ed != 0 || edlists[n].ned != 0);
	++n );
  if ( n < new->NS )
    {
      tL_msgerror ( err, _( "Both states %lu and %lu are final states" ) ,
		    _PS(endstate), _PS(n) );
      goto error;
    }
  --(new->NS);
  if ( endstate != new->NS )
    {
      edlists[endstate]= edlists[new->NS];
      states[endstate].word= states[new->NS].word;
      for ( n= 0; n < new->NE; ++n )
	if ( aux_edges[n].to == endstate )
	  aux_edges[n].to= new->NS;
	else if ( aux_edges[n].to == new->NS )
	  aux_edges[n].to= endstate;
    }
  
  /* Ordena les arestes. */
  NE= 0;
  for ( n= 0; n < new->NS; ++n )
    {
      states[n].edges.begin= edges+NE;
      for ( p= edlists[n].ed; p != 0; p= lnodes[p].next )
	{
	  ed= lnodes[p].ed;
	  if ( aux_edges[ed].to == new->NS )
	    {
	      states[n].fp= aux_edges[ed].p;
	      --(new->NE);
	    }
	  else
	    edges[NE++]= aux_edges[ed];
	}
      states[n].edges.end=
	states[n].null_edges.begin= edges+NE;
      for ( p= edlists[n].ned; p != 0; p= lnodes[p].next )
	{
	  ed= lnodes[p].ed;
	  if ( aux_edges[ed].to == new->NS )
	    {
	      states[n].fp= aux_edges[ed].p;
	      --(new->NE);
	    }
	  else
	    edges[NE++]= aux_edges[ed];
	}
      states[n].null_edges.end= edges+NE;
    }
  free ( aux_edges ); aux_edges= NULL;
  free ( edlists ); edlists= NULL;
  free ( lnodes ); lnodes= NULL;
  
  /* Localitza l'estat inicial. */
  counts= mem_alloc ( size_t, new->NS );
  for ( n= 0; n < new->NS; ++n ) counts[n]= 0;
  for ( n= 0; n < new->NE; ++n ) ++counts[edges[n].to];
  for ( n= 0; n < new->NS && counts[n] != 0; ++n );
  if ( n == new->NS )
    {
      tL_msgerror ( err, _( "There is not an initial state in the wordnet" ) );
      goto error;
    }
  new->start= n++;
  for ( ; n < new->NS && counts[n] != 0; ++n );
  if ( n != new->NS )
    {
      tL_msgerror ( err, _( "Both states %lu and %lu are initial states" ) ,
		     _PS(new->start), _PS(n) );
      goto error;
    }
  free ( counts );
  
  /* Allibera memòria. */
  tL_buffer_free ( buffer );
  
  return new;
  
  
 erroredgesf:
  tL_msgerror ( err, _( "Wrong format in the %luth edge" ) , _PS(line)+1 );
  goto error;
 errorstatesf:
  tL_msgerror ( err, _( "Wrong format in the %luth state" ) , _PS(line)+1 );
  goto error;
 errorheadf:
  tL_msgerror ( err, _( "Wrong header format" ) );
 error:
  tL_buffer_free ( buffer );
  tL_lm_free ( new );
  if ( aux_edges != NULL ) free ( aux_edges );
  if ( edlists != NULL ) free ( edlists );
  if ( lnodes != NULL ) free ( lnodes );
  if ( counts != NULL ) free ( counts );
  return NULL;
  
} /* end tL_lm_new_from_wordnet_file */
