/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  lmformat.c - Transforma un model de llenguatge d'un format
 *               estàndard al format de mhmm.
 *
 *  Adrià Giménez Pastor, 2009,2011,2013.
 *
 */


#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 1




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  tLBool      binary;
  const char *output_fn;
  const char *input_fn;
  const char *begin_sym;
  const char *end_sym;
  const char *ninf;
  
};


struct args
{
  
  const char *format;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{

  fprintf ( to, "%s [%s] [arpa|wordnet]\n\n", prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    arpa: %s\n", _("ARPA standard format") );
  fprintf ( to, "    wordnet: %s\n\n", _("HTK format") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -b,--begin-sym <string>: %s\n",
	    _("Symbol used as sentence start symbol. Default is '<s>'") );
  fprintf ( to, "      -e,--end-sym <string>: %s\n",
	    _("Symbol used as sentence end symbol. Default is '</s>'") );
  fprintf ( to, "      -i,--input <string>: %s\n",
	    _("File containing the input language model."
	      " Defaults to the standard input") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the resulting language model will be written") );
  fprintf ( to, "      --minus-inf <string>: %s\n",
	    _("Symbol used in the ARPA format"
	      " to designate the value -inf. Default is -99") );
  fprintf ( to, "      --noraw: %s\n",
            _("Store the model in text format") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n, noraw;
  const char *prog;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_value ( cmdl, 'b', "begin-sym", &(opts->begin_sym) );
  add_cmdl_value ( cmdl, 'e', "end-sym", &(opts->end_sym) );
  add_cmdl_value ( cmdl, 'i', "input", &(opts->input_fn) );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "minus-inf", &(opts->ninf) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "noraw", &noraw );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( opts->begin_sym == NULL )
    opts->begin_sym= "<s>";
  else if ( opts->begin_sym[0] == '\0' )
    opts->begin_sym= NULL;
  if ( opts->end_sym == NULL )
    opts->end_sym= "</s>";
  if ( opts->ninf == NULL )
    opts->ninf= "-99";
  opts->binary= (noraw==0);
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->format= argv[0];
  
} /* end usage */


static tLLM *
lmformat (
	  gzFile      input,
	  tLDict     *words,
	  const char *format,
	  const char *begin_sym,
	  const char *end_sym,
	  const char *ninf
	  )
{
  
  tLLM *lm;
  char *err;
  
  
  lm= NULL;
  if ( !strcmp ( format, "arpa" ) )
    lm= tL_lm_new_from_arpa_file ( input, words, TRUE, begin_sym,
				   end_sym, ninf, &err );
  else if ( !strcmp ( format, "wordnet" ) )
    lm= tL_lm_new_from_wordnet_file ( input, words, TRUE, begin_sym,
				      end_sym, &err );
  else tL_error ( _("Unknown input format '%s'"),
		  format );
  if ( lm == NULL ) tL_error ( err );
  
  return lm;
  
} /* end lmformat */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *words;
  gzFile input;
  FILE *output;
  tLLM *lm;
  char *err;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Prepara. */
  if ( opts.input_fn != NULL )
    {
      input= gzopen ( opts.input_fn, "r" );
      if ( input == NULL )
	tL_error ( _("Could not open the file '%s'"),
		   opts.input_fn );
    }
  else input= gzdopen(fileno(stdin),"r");
  words= tL_dict_new ();
  
  /* Cos. */
  if ( opts.output_fn != NULL )
    {
      output= fopen ( opts.output_fn, "w" );
      if ( output == NULL )
	tL_error ( _("Could not open the file '%s' for writing"),
		   opts.output_fn );
    }
  else output= stdout;
  if ( !strcmp ( args.format, "arpa" ) ) /* Special case for ARPA files. */
    {
      if ( tL_arpa2tlkformat ( input, output,
			       opts.begin_sym, opts.end_sym,
			       opts.ninf, opts.binary, &err ) != 0 )
	tL_error ( err );
    }
  else
    {
      lm= lmformat ( input, words, args.format,
		     opts.begin_sym, opts.end_sym,
		     opts.ninf );
      tL_lm_print ( lm, output, words, opts.binary );
      tL_lm_free ( lm );
    }
  
  /* Allibera recursos. */
  fclose ( output );
  tL_dict_free ( words );
  if ( opts.input_fn != NULL )  gzclose ( input );
  
  return EXIT_SUCCESS;
  
}
