/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  make_proto.c - Crea un model de caràcters amb N=1, un estat i
 *                 inicialitzat amb totes les mostres d'entrenament.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_pbm_lst.h"
#include "write_amodel.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 2




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  const char *output_fn;
  tLFloat     smooth;
  tLBool      binary;
  
};


struct args
{

  const char   *type;
  const char   *pbmlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{

  fprintf ( to, "%s [%s] <type> <sample-lst> ...\n\n", prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <type> %s\n", _("Type of acoustic model") );
  fprintf ( to, "    <sample-lst> %s\n\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    ... %s:\n", _("Extra arguments, according to the type") );
  fprintf ( to, "        Bernoulli,DGaussian: <D>\n" );
  fprintf ( to, "            <D>: %s\n", _("Dimension of the vectors") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -B,--binary: %s\n",
	    _("Print the model in binary format") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the initialized model will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -s,--smooth <double>: %s\n",
	    _("Smoothing factor. Default is 1e-6") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int         *argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n, i, cbinary;
  const char *smooth, *prog;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_counter ( cmdl, 'B', "binary", &(cbinary) );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 's', "smooth", &smooth );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, *argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( smooth != NULL )
    {
      opts->smooth= atof ( smooth );
      if ( opts->smooth < 0.0 || opts->smooth > 1.0 )
	tL_error ( _("The smoothing factor must be in the range [0,1]") );
    }
  else opts->smooth= 1e-6;
  opts->binary= (cbinary != 0);
  
  /* Arguments. */
  if ( n < NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->type= argv[0];
  args->pbmlst_fn= argv[1];
  
  for ( i= NUM_ARGS; i < n; ++i )
    argv[i-NUM_ARGS]= argv[i];
  *argc= n - NUM_ARGS;
  
} /* end usage */


static unsigned int
get_D (
       int         argc,
       const char *argv[]
       )
{
  
  unsigned int D;
  
  
  if ( argc < 1 )
    tL_error ( _("<D> has not been specified") );
  D= (unsigned int) atoi ( argv[0] );
  if ( D == 0 )
    tL_error ( _("The value of D cannot be 0") );
  
  return D;
  
} /* end get_D */


static tLAModel *
create_proto (
	      const char   *type,
	      const tLDict *syms,
	      int           argc,
	      const char   *argv[]
	      )
{
  
  unsigned int D;
  
  
  if ( !strcmp ( type, "Bernoulli" ) )
    {
      D= get_D ( argc, argv );
      return tL_amodel_bernoulli_new ( syms, 1, D );
    }
  else if ( !strcmp ( type, "DGaussian" ) )
    {
      D= get_D ( argc, argv );
      return tL_amodel_dgaussian_new ( syms, 1, D );
    }
  else tL_error ( _("Type '%s': not implemented in 'make_proto'"), type );
  
  return NULL; /* CALLAT! */
  
} /* end create_proto */


static tLDict *
get_syms ()
{
  
  tLDict *new;
  
  
  new= tL_dict_new ();
  tL_dict_insert ( new, "#", NULL );
  
  return new;
  
} /* end get_syms */


static void
make_proto (
	    const char   *type,
	    tLFileList   *pbm_lst,
	    const char   *output,
	    int           argc,
	    const char   *argv[],
	    const tLBool  binary,
	    const int     verbose
	    )
{
  
  tLAModel *amodel;
  tLDict *syms;
  tLAMiCounts *counts;
  tLFea *fea;
  size_t n;
  tLHMMInitializer *ic_aux;
  char *err;
  
  
  syms= get_syms ();
  amodel= create_proto ( type, syms, argc, argv );
  counts= tL_amicounts_new ( amodel );
  fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  tL_amicounts_init ( counts, amodel );
  ic_aux= tL_hmm_initializer_new ( amodel, &err );
  if ( ic_aux == NULL ) tL_error ( err );
  for ( n= 0; n < pbm_lst->N; ++n )
    {
      if ( tL_fea_load ( fea, pbm_lst->files[n], &err ) == -1 )
	tL_error ( err );
      if ( verbose )
	fprintf ( stderr, _("Processing sample '%s'"
			    " [cols:%d rows:%d]\n"),
		  pbm_lst->files[n], fea->cols, fea->rows );
      if ( !tL_amodel_test ( amodel, fea, &err ) )
	{
	  tL_warning ( _("The sample '%s' will be ignored: %s"),
		       pbm_lst->files[n], err );
	  free ( err );
	  continue;
	}
      tL_init_hmm ( ic_aux, counts, 0, fea );
    }
  tL_amicounts_update_amodel ( counts, amodel );
  write_amodel ( output, amodel, syms, binary );
  
  tL_hmm_initializer_free ( ic_aux );
  tL_fea_free ( fea );
  tL_amicounts_free ( counts, amodel );
  tL_dict_free ( syms );
  tL_amodel_free ( amodel );
  
} /* end make_proto */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLFileList *pbm_lst;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( &argc, argv, &opts, &args );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  make_proto ( args.type, pbm_lst,
	       opts.output_fn, argc,
	       argv, opts.binary, opts.verbose );
  tL_filelist_free ( pbm_lst );
  
  return EXIT_SUCCESS;
  
}
