/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mllr.c - Adapta un model amb MLLR.
 *
 *  Adrià Giménez Pastor, 2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#include "cmdl.h"
#include "get_amodel.h"
#include "get_pbm_lst.h"
#include "get_words.h"
#include "itertrain.h"
#include "macros.h"
#include "write_amodel.h"

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 3

#define abs(A) (((A)<0.0) ? -(A) : A)




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int           verbose;
  const char   *output_fn;
  tLFloat       cfactor;
  unsigned int  max_iters;
  tLBool        nowords;
  int           skiplll;
  const char   *rcs_fn;
  unsigned int  blocksize;
  const char   *tarfn;
  
};


struct args
{
  
  const char *amodel_fn;
  const char *wdlst_fn;
  const char *pbmlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <amodel> <transc-lst> <sample-lst>\n\n",
	    prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Initial acoustic model") );
  fprintf ( to, "    <transc-lst> %s\n",
	    _("List of the files containing the transcriptions") );
  fprintf ( to, "    <sample-lst> %s\n\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -b,--block-size <uint>: %s\n",
	    _("Number of samples in memory at each point. Default is 1") );
  fprintf ( to, "      -c,--cfactor <double>: %s\n",
	    _("Convergence factor. Default is 0.0001") );
  fprintf ( to, "      -m,--max-iters <uint>: %s\n",
	    _("Maximum number of iterations. Default is 20") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the adapted model will be written."
	      " Defaults to no output") );
  fprintf ( to, "      -r,--regression-classes <string>: %s\n",
	    _("File containing the assignment of states to regression classes."
	      " Initially, all states are assigned to class 0.") );
  /*
	      " This file is expected to be a sequence of tuples,"
	      " each tuple being 'symbol_name state class', or"
	      " 'state_name class' if we work with a TiedStates model") );
  */
  fprintf ( to, "      --show-last-ll: %s\n",
	    _("Log probability will be shown for the last iteration") );
  fprintf ( to, "      --tar <string>: %s\n",
	    _("Samples will be read from this file") );
  fprintf ( to, "      --words: %s\n",
	    _("Transcriptions will be expected to be words") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, cwords, showlll, n;
  const char *prog, *cfactor, *max, *blocksize;


  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'b', "block-size", &(blocksize) );
  add_cmdl_value ( cmdl, 'c', "cfactor", &cfactor );
  add_cmdl_value ( cmdl, 'm', "max-iters", &max );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 'r', "regression-classes", &(opts->rcs_fn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "show-last-ll", &showlll );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "tar", &(opts->tarfn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "words", &cwords );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( cfactor != NULL )
    {
      opts->cfactor= atof ( cfactor );
      if ( opts->cfactor <= 0.0 )
        tL_error ( _("The convergence factor must be greater than 0") );
    }
  else opts->cfactor= 0.0001;
  if ( max != NULL )
    {
      opts->max_iters= 
        (unsigned int) strtol ( max, (char **) NULL, 10 );
    }
  else opts->max_iters= 20;
  opts->nowords= (cwords == 0);
  if ( blocksize == NULL ) opts->blocksize= 1;
  else
    {
      opts->blocksize= atoui ( blocksize );
      if ( opts->blocksize == 0 )
        tL_error ( _("Block size cannot be 0") );
    }
  opts->skiplll= !showlll;
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->amodel_fn= argv[0];
  args->wdlst_fn= argv[1];
  args->pbmlst_fn= argv[2];
  
} /* end usage */


static void
mllr_alg (
	  tLAModel           *adaptmodel,
	  tLMLLR             *mllr,
	  tLTrainer          *trainer,
	  tLAMlogCounts      *counts,
	  tLSeqs             *words,
	  tLFileList         *pbm_lst,
	  unsigned int        max_iters,
	  tLFloat             cfactor,
	  int                 skiplll,
	  const tLDict       *syms,
	  const unsigned int  blocksize,
	  const char         *tarfn,
	  int                 verbose
	  )
{
  
  int i;
  tLProb lv, lvp;
  tLFloat inc;
  tLBool conv;
  feablock_t *feablock;
  tLTar *tar;
  char *err;
  
  
  conv= FALSE; /* CALLA!! */
  
  if ( verbose )
    fprintf ( stderr, _("Starting MLLR\n") );
  
  if ( tarfn == NULL ) tar= NULL;
  else
    {
      tar= tL_tar_new ( tarfn, &err );
      if ( tar == NULL ) tL_error ( _("Error while reading '%s': %s"), tarfn, err );
    }
  feablock= create_feablock ( blocksize, trainer->amodel );
  inc= HUGE_VAL;
  i= 0;
  lv= itertrain ( trainer, feablock, pbm_lst, tar, verbose );
  if ( verbose )
    {
      fprintf ( stderr, "  NºIter\t\tLL\t\tINC\n" );
      fprintf ( stderr, "%c %d\t\t%g\t\t%g\n",
		lv==0.0?'.':'*', i, _DB(lv), _DB(inc) );
    }
  if ( lv == 0 ) goto quit; /* No es toca el model. */
  while ( lv != -HUGE_VAL && i < max_iters && (i == 0 || !conv) )
    {
      
      /* Trainer té un punter al model que torna get_adapated_model. */
      tL_mllr_update_matrices ( mllr, counts, adaptmodel );
      tL_mllr_get_adapted_model ( mllr, adaptmodel );
      ++i;
      if ( skiplll && i == max_iters ) break;
      
      lvp= lv;
      lv= itertrain ( trainer, feablock, pbm_lst, tar, verbose );
      inc= (lv - lvp) / abs(lvp);
      conv= (inc < cfactor);
      if ( verbose )
        fprintf ( stderr, "%c %d\t\t%g\t\t%g\n",
                  conv ? '.' : '*',
                  i, _DB(lv), _DB(inc) );
      
    }
 quit:
  if ( tar != NULL ) tL_tar_free ( tar );
  free_feablock ( feablock );
  
} /* end mllr_alg */


static void
load_rcs (
	  const char      *fn,
	  const tLAModel  *amodel,
	  const tLDict    *syms,
	  tLMLLRrc       **rcs,
	  unsigned int    *rcs_length,
	  const int        verbose
	  )
{
  
  gzFile f;
  tLBuffer *buffer;
  unsigned int size;
  tLMLLRrc rc;
  tLBool istied;
  const tLDict *names;
  size_t aux;
  
  
  if ( verbose )
    fprintf ( stderr,
	      _("Reading assignment of states to regression classes\n") );
  
  /* Prepara. */
  f= gzopen ( fn, "r" );
  if ( f == NULL ) tL_error ( _("Could not open '%s'"), fn );
  buffer= tL_buffer_new ();
  *rcs= mem_alloc ( tLMLLRrc, 1 );
  *rcs_length= 0;
  size= 1;
  names= tL_amodel_tiedstates_get_names ( amodel );
  istied= (names!=NULL);
  
  /* Llig. */
  while ( tL_gtoken ( f, buffer ) )
    {
      
      /* nom (c/q). */
      if ( istied )
	{
	  rc.c= 0;
	  aux= (unsigned int) tL_dict_find ( names, buffer->v );
	  if ( aux == names->N )
	    tL_error ( _("The name '%s', read from '%s', doesn't correspond"
			 " to any valid state"), fn, buffer->v );
	  rc.q= (unsigned int) aux;
	}
      else
	{
	  aux= (unsigned int) tL_dict_find ( syms, buffer->v );
	  if ( aux == syms->N )
	    tL_error ( _("The name '%s', read from '%s', doesn't correspond"
			 " to any valid symbol"), fn, buffer->v );
	  rc.c= (unsigned int) aux;
	}
      
      /* q */
      if ( !istied )
	{
	  if ( !tL_gtoken ( f, buffer ) )
	    tL_error ( _("Could not read from '%s' the state for '%s'"),
		       fn, tL_dict_get_token ( syms, rc.c ) );
	  rc.q= atoui ( buffer->v );
	}
      
      /* rc */
      if ( !tL_gtoken ( f, buffer ) )
	tL_error ( _("Could not read from '%s' the regression class for '%s'"),
		   fn, tL_dict_get_token ( istied?names:syms, rc.c ) );
      rc.rc= atoui ( buffer->v );
      
      if ( *rcs_length == size )
	{
	  size*= 2;
	  *rcs= mem_realloc ( tLMLLRrc, *rcs, size );
	}
      (*rcs)[(*rcs_length)++]= rc;
      
    }
  
  if ( size > *rcs_length )
    *rcs= mem_realloc ( tLMLLRrc, *rcs, *rcs_length );
  
  /* Allibera memòria. */
  tL_buffer_free ( buffer );
  gzclose ( f );

} /* end load_rcs */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel, *adaptmodel;
  tLFileList *pbm_lst;
  tLSeqs *words;
  tLAMlogCounts *counts;
  tLTrainer *trainer;
  tLMLLR *mllr;
  char *err;
  tLMLLRrc *rcs;
  unsigned int rcs_length;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  words= get_words ( args.wdlst_fn, syms, opts.nowords, NULL, opts.verbose );
  if ( words->N != pbm_lst->N )
    tL_error ( _("The sample list contains %lu entries,"
		 " but the word list contains %lu entries"),
               _PS(pbm_lst->N), _PS(words->N) );
  rcs= NULL; rcs_length= 0;
  if ( opts.rcs_fn != NULL )
    load_rcs ( opts.rcs_fn, amodel, syms, &rcs, &rcs_length, opts.verbose );
  
  /* Cos. */
  mllr= tL_mllr_new ( amodel, rcs, rcs_length, &err );
  if ( mllr == NULL ) tL_error ( err );
  counts= tL_amlogcounts_new_mllr ( amodel );
  adaptmodel= tL_mllr_get_adapted_model ( mllr, NULL );
  trainer= tL_trainer_new ( adaptmodel, counts, words, NULL, &err );
  if ( trainer == NULL ) tL_error ( err );
  mllr_alg ( adaptmodel, mllr, trainer, counts, words, pbm_lst,
	     opts.max_iters, opts.cfactor, opts.skiplll,
	     syms, opts.blocksize, opts.tarfn, opts.verbose );
  write_amodel ( opts.output_fn, trainer->amodel, syms, FALSE );
  
  /* Allibera memòria. */
  if ( rcs != NULL ) free ( rcs );
  tL_trainer_free ( trainer );
  tL_amlogcounts_free ( counts, amodel );
  tL_mllr_free ( mllr );
  tL_seqs_free ( words );
  tL_filelist_free ( pbm_lst );
  tL_amodel_free ( adaptmodel );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
