/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mumix.c - Multiplicador de componets de mixtura.
 *
 *  Adrià Giménez Pastor, 2011,2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "macros.h"
#include "write_amodel.h"

#include "sysinfo.h"




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  const char *input_fn;
  const char *output_fn;
  tLBool      binary;
  tLFloat     smooth;
  const char *mcounts_fn;
  tLFloat     min_occ;
  
};


struct args
{
  
  unsigned int  newI;
  const char   *sym;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{

  fprintf ( to, "%s [%s] <newI> [<sym>]\n\n", prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <newI> %s\n",
	    _("Expected number of mixture components per state."
	      " 0 means that every component will be split in two") );
  fprintf ( to, "    <sym> %s\n",
	    _("Split the mixture components of the specified symbol") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -B,--binary: %s\n",
	    _("Print the model in binary format") );
  fprintf ( to, "      -c,--use-mixcounts <string>: %s\n",
	    _("Mixture coefficient counts will be read from this file") );
  fprintf ( to, "      -i,--input <string>: %s\n",
	    _("File from which the model to be transformed will be read."
	      " Defaults to the standard input") );
  fprintf ( to, "      -m,--min-occ <double>: %s\n",
	    _("Minimum occupancy of each mixture component. This value"
	      " is only used if the option '-c' has been used"
	      " to specify a file containing the mixture coefficient counts."
	      " Default is 10") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the resulting model will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -s,--smooth <double>: %s\n",
	    _("Smoothing factor. Smoothing will not be applied to transitions."
	      " Default is 1e-6") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct args *args,
       struct opts *opts
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n, cbinary;
  const char *prog, *smooth, *cmin_occ;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_counter ( cmdl, 'B', "binary", &(cbinary) );
  add_cmdl_value ( cmdl, 'c', "use-mixcounts", &(opts->mcounts_fn) );
  add_cmdl_value ( cmdl, 'i', "input", &(opts->input_fn) );
  add_cmdl_value ( cmdl, 'm', "min-occ", &cmin_occ );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 's', "smooth", &smooth );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  opts->binary= (cbinary != 0);
  if ( smooth != NULL )
    {
      opts->smooth= atof ( smooth );
      if ( opts->smooth < 0.0 || opts->smooth > 1.0 )
        tL_error ( _("The smoothing factor must be in the range [0,1]") );
    }
  else opts->smooth= 1e-6;
  opts->min_occ= (cmin_occ!=NULL) ? atof ( cmin_occ ) : 10.0;
  
  /* Arguments. */
  if ( n == 2 ) args->sym= argv[1];
  else if ( n != 1 )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  else args->sym= NULL;
  args->newI= atoui ( argv[0] );
  
} /* end usage */


static tLMixCounts *
read_mixcounts (
		const char     *fn,
		const tLAModel *amodel
		)
{
  
  gzFile f;
  tLMixCounts *mcounts;
  char *err;
  
  
  if ( fn == NULL ) return NULL;
  f= gzopen ( fn, "r" );
  if ( f == NULL ) tL_error ( _("Could not open '%s'"), fn );
  mcounts= tL_mixcounts_read ( amodel, f, &err );
  if ( mcounts == NULL ) tL_error ( err );
  gzclose ( f );
  
  return mcounts;
  
} /* end read_mixcounts */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{

  struct args args;
  struct opts opts;
  tLDict *syms;
  tLAModel *amodel;
  tLMixtCompMult *mcmu;
  char *err;
  unsigned int n, q;
  tLMixCounts *mcounts;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &args, &opts );
  syms= tL_dict_new ();
  amodel= get_amodel ( opts.input_fn, syms, opts.verbose );
  mcounts= read_mixcounts ( opts.mcounts_fn, amodel );
  
  if ( opts.verbose )
    {
      fprintf ( stderr, _("Transforming to %u components per state\n"),
		args.newI );
    }
  mcmu= tL_mixtcompmult_new ( amodel, mcounts, &err );
  if ( mcmu == NULL ) tL_error ( err );
  mcmu->min_occ= opts.min_occ;
  /* Mixtures */
  if ( !strcmp ( "Mixture", amodel->type ) )
    {
      if ( args.sym == NULL )
	for ( n= 0; n < amodel->N; ++n )
	  for ( q= 0; q < amodel->hmms[n].Q; ++q )
	    tL_mu_mixture_components ( mcmu, n, q, args.newI );
      else
	{
	  n= tL_dict_find ( syms, args.sym );
	  if ( n == syms->N )
	    tL_error ( _("Cannot find symbol '%s' in the acoustic model"),
		       args.sym );
	  for ( q= 0; q < amodel->hmms[n].Q; ++q )
	    tL_mu_mixture_components ( mcmu, n, q, args.newI );
	}
    }
  /* Tiedstates */
  else
    {
      if ( tL_mu_tmixture_components ( mcmu, args.newI, args.sym, &err ) == -1 )
	tL_error ( err );
    }
  tL_mixtcompmult_free ( mcmu );
  if ( opts.smooth > 0.0 )
    tL_amodel_smooth ( amodel, opts.smooth, FALSE );
  
  write_amodel ( opts.output_fn, amodel, syms, opts.binary );
  if ( mcounts != NULL ) tL_mixcounts_free ( mcounts, amodel );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
