/*
 *  Copyright 2014 transLectures-UPV Team
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  mustd.c - Mean variance normalization.
 */


#include <assert.h>
#include <libTLK/TLK.h>
#include <limits.h>
#include <locale.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cmdl.h"
#include "intl.h"
#include "macros.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 1




/*********/
/* TYPES */
/*********/

struct opts
{
  
  int         verbose;
  tLBool      binary;
  const char *output_fn;
  const char *print_fn;
  const char *mustd_fn;;
  
};


struct args
{
  
  const char *input_fn;
  
};




/*************/
/* FUNCTIONS */
/*************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{

  fprintf ( to, "%s [%s] <input-lst>\n", prog, _("options") );
  fprintf ( to, "\n" );
  if ( !details ) return;
  fprintf ( to, "    <input-lst> %s\n",
	    _("List of the input sample files") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -m,--mustd <string>: %s\n",
	    _("Use the mean and standard deviation of the specified file."
	      " By default, mean and standard deviation are calculated using"
	      " the input samples") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("List of the files to which the normalized samples"
	      " will be written. By default, the input samples are overwritten") );
  fprintf ( to, "      -p,--print <string>: %s\n",
	    _("Print the mean and standard deviation into the specified file") );
  fprintf ( to, "      --noraw: %s\n",
	    _("Store features in text format") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, n, noraw;
  const char *prog;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'm', "mustd", &(opts->mustd_fn) );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 'p', "print", &(opts->print_fn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "noraw", &noraw );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  opts->binary= (noraw==0);
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->input_fn= argv[0];
  
} /* end usage */


static tLFileList *
load_list (
	   const char   *fname,
	   const tLBool  is_input,
	   const int     verbose
	   )
{
  
  tLFileList *fl;
  size_t n;
  
  
  if ( verbose )
    fprintf ( stderr,
	      is_input ? 
	      _("Loading the input sample list\n") :
	      _("Loading the output sample list\n") );
  fl= tL_filelist_new_from_file ( fname );
  if ( fl == NULL ) tL_error ( _("Unable to open '%s'"), fname );
  
  if ( verbose > 1 )
    {
      fprintf ( stderr,
		is_input ? _("Input samples:\n") : _("Output samples:\n") );
      for ( n= 0; n < fl->N; ++n )
	fprintf ( stderr, "  - %s\n", fl->files[n] );
    }
  
  return fl;
  
} /* end load_list */


static void
accum_sample (
	      tLFea       *fea,
	      const char  *fn,
	      tLMuStd    **mustd,
	      const int    verbose
	      )
{
  
  int ret;
  char *err;
  
  
  /* Read file. */
  if ( verbose > 1 ) fprintf ( stderr, _("  Processing '%s'\n"), fn );
  ret= tL_fea_load ( fea, fn, &err );
  if ( ret )
    {
      tL_warning ( _("'%s' is empty, it will not be used for mean"
		     " and variance normalization"), fn );
      free ( err );
      return;
    }
  
  /* Accumulate counts */
  if ( *mustd == NULL ) *mustd= tL_mustd_new ( fea->rows );
  else if ( fea->rows != (*mustd)->D )
    tL_error ( _("An unexpected error occurred while estimating"
		 " the mean and variance normalization for sample"
		 " '%s': dimensions differ"), fn );
  tL_mustd_accum ( *mustd, fea );
  
} /* end accum_sample */


static void
norm_sample (
	     const char    *infn,
	     const char    *outfn,
	     tLFea         *fea,
	     const tLMuStd *mustd,
	     const tLBool   binary,
	     const int      verbose
	     )
{
  
  char *err;
  int ret, t;
  unsigned int d;
  float **frames, *frame;
  
  
  if ( verbose > 1 ) fprintf ( stderr, _("  Normalizing '%s'\n"), infn );
  ret= tL_fea_load ( fea, infn, &err );
  if ( ret )
    {
      tL_warning ( _("'%s' will not be normalized, an error occurred"
		     " while reading it: %s"), infn, err );
      free ( err );
      return;
    }
  if ( fea->rows != mustd->D )
    tL_error ( _("An unexpected error occurred while"
		 " normalizing sample the sample"
		 " '%s': dimensions differ"), infn );
  frames= tL_fea_get_real_data ( fea );
  for ( t= 0; t < fea->cols; ++t )
    {
      frame= frames[t];
      for ( d= 0; d < mustd->D; ++d )
        frame[d]= (frame[d]-mustd->mu[d])/mustd->std[d];
    }
  ret= tL_fea_write ( fea, outfn, binary, &err );
  if ( ret ) tL_error ( _("An error occurred while writing '%s': %s"),
			outfn, err );
  
} /* end norm_sample */


static void
print_mustd (
	     const char    *fn,
	     const tLMuStd *mustd
	     )
{
  
  FILE *f;
  
  
  f= fopen ( fn, "w" );
  if ( f == NULL )
    tL_error ( _("Could not open file '%s'"), fn );
  tL_mustd_print ( mustd, f );
  fclose ( f );
  
} /* end print_mustd */


tLMuStd *
get_mustd (
	   const char *fn,
	   const int   verbose
	   )
{
  
  tLMuStd *mustd;
  gzFile f;
  char *err;
  
  
  f= gzopen ( fn, "r" );
  if ( f == NULL )
    tL_error ( _("Could not open file '%s'"), fn );
  mustd= tL_mustd_new_from_file ( f, &err );
  if ( mustd == NULL ) tL_error ( "MUSTD: %s.", err );
  gzclose ( f );
  
  if ( verbose )
    {
      fprintf ( stderr, _("Mean and standard deviation have been read\n") );
      if ( verbose > 2 )
        tL_mustd_print ( mustd, stderr );
    }
  
  return mustd;
  
} /* end get_mustd */


/* Returns a tLMuStd struct if mean and variance normalization is
   applied. */
static void
norm (
      const tLFileList  *in_lst,
      const tLFileList  *out_lst,
      const struct opts *opts
      )
{
  
  size_t n;
  tLFea *fea;
  tLMuStd *mustd;
  
  
  /* Prepare. */
  fea= tL_fea_new ( TL_FEA_REAL );
  
  /* Get mustd. */
  if ( opts->mustd_fn == NULL )
    {
      if ( opts->verbose )
	fprintf ( stderr, _("Estimating mean and standard deviation\n") );
      mustd= NULL;
      for ( n= 0; n < in_lst->N; ++n )
	accum_sample ( fea, in_lst->files[n], &mustd, opts->verbose );
      tL_mustd_calc ( mustd );
    }
  else mustd= get_mustd ( opts->mustd_fn, opts->verbose );
  
  /* Mu-std normalization. */
  if ( opts->verbose ) fprintf ( stderr, _("Normalizing samples\n") );
  for ( n= 0; n < out_lst->N; ++n )
    norm_sample ( in_lst->files[n], out_lst->files[n], fea, mustd,
		  opts->binary, opts->verbose );
  
  /* Print mustd */
  if ( opts->print_fn != NULL ) print_mustd ( opts->print_fn, mustd );
  
  /* Free memory. */
  if ( mustd != NULL ) tL_mustd_free ( mustd );
  tL_fea_free ( fea );
  
} /* end norm */




/********/
/* MAIN */
/********/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLFileList *in_lst, *out_lst;
  const char *out_fn;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  in_lst= load_list ( args.input_fn, TRUE, opts.verbose );
  out_fn= opts.output_fn!=NULL ? opts.output_fn : args.input_fn;
  out_lst= load_list ( out_fn, FALSE, opts.verbose );
  if ( in_lst->N != out_lst->N )
    tL_error ( _("The number of samples in '%s' and '%s' differs"),
	       args.input_fn, out_fn );
  norm ( in_lst, out_lst, &opts );
  tL_filelist_free ( in_lst );
  tL_filelist_free ( out_lst );
  
  return EXIT_SUCCESS;
  
}
