/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  recognise.c - Reconeixedor de text (o el que es vullga
 *                interpretar).
 *
 *  Adrià Giménez Pastor, 2009,2010,2011,2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "get_lexicon.h"
#include "get_lm.h"
#include "get_pbm_lst.h"
#include "macros.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 3




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int             verbose;
  const char     *output_fn;
  const char     *lexicon_fn;
  unsigned long   nmaxas;
  tLFloat         gsf;
  tLProb          wip;
  tLFloat         beam;
  tLFloat         lm_beam;
  tLRecAlignType  align;
  tLBool          print_all;
  tLBool          print_words;
  tLBool          print_probs;
  tLBool          viterbi_states;
  unsigned int    nhyp;
  const char     *wglst_fn;
  tLBool          htk_wgs;
  unsigned int    la_cap;
  unsigned int    la_depth;
  const char     *sil;
  
};


struct args
{
  
  const char *lm_fn;
  const char *amodel_fn;
  const char *pbmlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <lm> <amodel> <sample-lst>\n\n", prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <lm> %s\n", _("Language model") );
  fprintf ( to, "    <amodel> %s\n", _("Acoustic model") );
  fprintf ( to, "    <sample-lst> %s\n\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -b,--beam <float>: %s\n",
	    _("Beam. Default value is inf") );
  fprintf ( to, "      -e,--word-end-pruning <float>: %s\n",
	    _("Word-End Pruning Beam. Default value is inf") );
  fprintf ( to, "      -g,--grammar-scale-factor <float>: %s\n",
	    _("Grammar Scale Factor. Default is 1.0") );
  fprintf ( to, "      -l,--lexicon <string>: %s\n",
	    _("File containing the lexicon."
	      " By default, it is created from <lm>,"
	      " assuming that each word is a UTF-8 word") );
  fprintf ( to, "      -m,--nmax-astates <uint>: %s\n",
	    _("Maximum number of active states."
	      " By default, there is no limit") );
  fprintf ( to, "      -N,--wordgraph-density <uint>: %s\n",
	    _("Maximum number of hypotheses saved in each state."
	      " Default is 10") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the transcriptions will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -s,--sil <string>: %s\n",
	    _("Symbol used to represent silence."
	      " By default, there is no silence") );
  fprintf ( to, "      -w,--word-insertion-penalty <float>: %s\n",
	    _("Word Insertion Penalty (it is a log probability)."
	      " Default is 0.0") );
  fprintf ( to, "      -W,--wordgraph-list <string>: %s\n",
	    _("List of the files to which the word graphs will be written."
	      " One file per sample must be specified."
	      " By default, no word graphs will be generated") );
  fprintf ( to, "      --htk-wgs: %s\n",
	    _("Word graphs will be printed using an HTK-like format") );
  fprintf ( to, "      --lookahead-capacity <uint>: %s\n",
	    _("Cache memory capacity"
	      " for the lookahead algorithm."
	      " 0 disables the algorithm. By default, it is disabled") );
  fprintf ( to, "      --lookahead-depth <uint>: %s\n",
	    _("The depth of the n-grams used in lookahead. By default,"
	      " the depth of the language model is used") );
  fprintf ( to, "      --print-all: %s\n",
	    _("Print words, symbols and probabilities") );
  fprintf ( to, "      --print-probs: %s\n",
	    _("Print to the output the probability"
	      " in addition to the transcription") );
  fprintf ( to, "      --print-syms: %s\n",
	    _("Print symbols instead of words") );
  fprintf ( to, "      --sym-align: %s\n", _("Symbol-level segmentation") );
  fprintf ( to, "      --viterbi-states: %s\n",
	    _("Summations in state functions are replaced by max operators") );
  fprintf ( to, "      --word-align: %s\n", _("Word-level segmentation") );
  fprintf ( to, "\n" );

} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, cwalign, csalign, chtkwgs, cprintall, cprintsyms,
    cprintprobs, cvstates, n;
  const char *prog, *snmaxas, *gsf, *wip, *beam, *nhyp, *clacap, *cladepth,
    *lmbeam;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'b', "beam", &beam );
  add_cmdl_value ( cmdl, 'e', "word-end-pruning", &lmbeam );
  add_cmdl_value ( cmdl, 'g', "grammar-scale-factor", &gsf );
  add_cmdl_value ( cmdl, 'l', "lexicon", &(opts->lexicon_fn) );
  add_cmdl_value ( cmdl, 'm', "nmax-astates", &snmaxas );
  add_cmdl_value ( cmdl, 'N', "wordgraph-density", &nhyp );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 's', "sil", &(opts->sil) );
  add_cmdl_value ( cmdl, 'w', "word-insertion-penalty", &wip );
  add_cmdl_value ( cmdl, 'W', "wordgraph-list", &(opts->wglst_fn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "htk-wgs", &chtkwgs );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "lookahead-capacity", &clacap );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "lookahead-depth", &cladepth );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "print-all", &cprintall );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "print-probs", &cprintprobs );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "print-syms", &cprintsyms );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "sym-align", &csalign );  
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "viterbi-states", &cvstates );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "word-align", &cwalign );  
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );
  
  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( gsf == NULL ) opts->gsf= 1.0;
  else
    {
      opts->gsf= atof ( gsf );
      if ( opts->gsf == 0.0 )
        tL_error ( _("The Grammar Scale Factor cannot be 0.0") );
    }
  opts->nmaxas= (snmaxas == NULL) ? ULONG_MAX : atol ( snmaxas );
  if ( wip == NULL ) opts->wip= 0.0;
  else opts->wip= (tLProb) atof ( wip );
  opts->print_all= (cprintall!=0);
  opts->print_words= ((cprintsyms==0) || opts->print_all);
  opts->print_probs= ((cprintprobs!=0) || opts->print_all);
  opts->align= TL_REC_ALIGN_NONE;
  if ( cwalign ) opts->align= TL_REC_ALIGN_WORDS;
  if ( csalign ) opts->align= TL_REC_ALIGN_SYMS;
  opts->viterbi_states= (cvstates!=0);
  if ( beam != NULL ) opts->beam= atof ( beam );
  else opts->beam= HUGE_VAL;
  if ( lmbeam != NULL ) opts->lm_beam= atof ( lmbeam );
  else opts->lm_beam= HUGE_VAL;
  if ( nhyp != NULL )
    {
      opts->nhyp= (unsigned int) atoi ( nhyp );
      if ( opts->nhyp == 0 )
	tL_error ( _("Word graph density cannot be 0") );
    }
  else opts->nhyp= 10;
  opts->htk_wgs= (chtkwgs!=0);
  opts->la_cap= (clacap==NULL) ? 0 : atoui ( clacap );
  opts->la_depth= (cladepth==NULL) ? 0 : atoui ( cladepth );
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->lm_fn= argv[0];
  args->amodel_fn= argv[1];
  args->pbmlst_fn= argv[2];

} /* end usage */


/* RECOGNISE ******************************************************************/
static void
recognise_sample (
		  tLRecogniser  *reco,
		  FILE          *out,
		  tLFea         *fea,
		  const char    *pbm,
		  const char    *wg_fn,
		  const tLDict  *syms,
		  const tLDict  *words,
		  const tLBool   print_all,
		  const tLBool   print_probs,
		  const tLBool   print_words,
		  const tLBool   htk_wgs,
		  tLRecWord    **rec,
		  size_t        *rec_size,
		  tLWordGraph  **wg,
		  const int      verbose
		  )
{
  
  tLRecWord *p;
  char *err;
  FILE *f;
  tLRecStats stats;
  
  
  /* Prepara. */
  if ( tL_fea_load ( fea, pbm, &err ) == -1 )
    tL_error ( err );
  if ( !tL_amodel_test ( reco->amodel, fea, &err ) )
    {
      tL_warning ( _("The sample '%s' is not compatible with the acoustic"
		     " model and will not be classified: %s"), pbm, err );
      free ( err );
      return;
    }
  if ( verbose > 1 )
    fprintf ( stderr, _("Recognizing the sample '%s'\n"), pbm );
  
  /* Reconeix. */
  tL_recognise ( reco, fea, rec, rec_size, wg, &stats );
  if ( reco->align != TL_REC_ALIGN_NONE || print_all )
    fprintf ( out, "\"%s\"\n", pbm );
  if ( stats.prob == -HUGE_VAL )
    tL_warning ( _("The sample '%s' could not be recognized"), pbm );
  else
    {
      if ( wg_fn != NULL )
	{
	  f= fopen ( wg_fn, "w" );
	  if ( f == NULL )
	    tL_error ( _("Could not open file '%s' for writing"),
		       wg_fn );
	  if ( htk_wgs )
	    tL_wordgraph_print_htk ( *wg, f, words, syms, pbm, NULL, NULL );
	  else
	    tL_wordgraph_print ( *wg, f, words, syms, NULL );
	  fclose ( f );
	}
      if ( print_probs ) fprintf ( out, "%g", _DB(stats.prob) );
      if ( reco->align == TL_REC_ALIGN_WORDS )
	{
	  if ( print_probs ) fputc ( '\n', out );
	  for ( p= *rec; p->cs != NULL; ++p )
	    {
	      fprintf ( out, "%d %d ", p->pos, p->length );
	      if ( print_all )
		{
		  fprintf ( out, "%s ", tL_dict_get_token ( words, p->w ) );
		  tL_seq_print_wildcards ( p->cs, out, syms, ' ',
					   p->sym, p->sympos );
		}
	      else if ( print_words )
		fprintf ( out, "%s", tL_dict_get_token ( words, p->w ) );
	      else
		tL_seq_print_wildcards ( p->cs, out, syms, ' ',
					 p->sym, p->sympos );
	      fputc ( '\n', out );
	    }
	}
      else if ( reco->align == TL_REC_ALIGN_SYMS )
	{
	  if ( print_probs ) fputc ( '\n', out );
	  for ( p= *rec; p->cs != NULL; ++p )
	    {
	      fprintf ( out, "%d %d %s", p->pos, p->length,
			tL_dict_get_token ( syms, p->sym ) );
	      if ( print_words )
		{
		  if ( p->sympos == (size_t) -1 )
		    fprintf ( out, " -1" );
		  else fprintf ( out, " %lu", _PS(p->sympos) );
		  fprintf ( out, " %s", tL_dict_get_token ( words, p->w ) );
		}
	      fputc ( '\n', out );
	    }
	}
      else /* TL_REC_ALIGN_NONE */
	{
	  if ( print_all )
	    {
	      fputc ( '\n', out );
	      for ( p= *rec; p->cs != NULL; ++p )
		{
		  fprintf ( out, "%s ", tL_dict_get_token ( words, p->w ) );
		  tL_seq_print_wildcards ( p->cs, out, syms, ' ',
					   p->sym, p->sympos );
		  fputc ( '\n', out );
		}
	    }
	  else if ( print_words )
	    {
	      
	      fprintf ( out, print_probs?" %s":"%s",
			tL_dict_get_token ( words, (*rec)[0].w ) );
	      for ( p= (*rec)+1; p->cs != NULL; ++p )
		fprintf ( out, " %s", tL_dict_get_token ( words, p->w ) );
	    }
	  else
	    {
	      if ( print_probs ) fputc ( ' ', out );
	      tL_seq_print_wildcards ( (*rec)[0].cs, out, syms, ' ',
				       (*rec)[0].sym, (*rec)[0].sympos );
	      for ( p= (*rec)+1; p->cs != NULL; ++p )
		{
		  fputc ( ' ', out );
		  tL_seq_print_wildcards ( p->cs, out, syms, ' ',
					   p->sym, p->sympos );
		}
	    }
	}
    }
  if ( reco->align != TL_REC_ALIGN_NONE || print_all ) fprintf ( out, ".\n" );
  else                                                 fputc ( '\n', out );
  fflush ( out );
  
  if ( verbose > 1 )
    {
      fprintf ( stderr, "ASTATES -> MIN: %lu MAX: %lu MEAN: %g\n",
		stats.min_astates, stats.max_astates,
		stats.total_astates / (double) fea->cols );
      fprintf ( stderr, _("LOOKAHEAD CACHE FAILS: %lu\n"),
		stats.lookahead_fails );
    }
  
} /* end recognise_sample */


static void
recognise (
	   const tLAModel       *amodel,
	   const tLLexicon      *lexicon,
	   const tLLM           *lm,
	   const tLDict         *syms,
	   const tLFileList     *pbm_lst,
	   const tLFileList     *wg_lst,
	   const char           *output,
	   unsigned long         nmaxstates,
	   const tLRecAlignType  align,
	   const tLFloat         beam,
	   const tLFloat         lm_beam,
	   const tLBool          print_all,
	   const tLBool          print_probs,
	   const tLBool          print_words,
	   const tLBool          htk_wgs,
	   const tLBool          viterbi_states,
	   const unsigned int    nhyp,
	   const unsigned int    la_cap,
	   const unsigned int    la_depth,
	   const char           *sil,
	   const int             verbose
	   )
{
  
  tLFea *fea;
  size_t n, rec_size;
  FILE *to;
  tLRecogniser *reco;
  char *err;
  tLRecWord *rec;
  tLWordGraph *wg;
  size_t silsym;
  clock_t t1,t2;
  
  
  if ( verbose )
    fprintf ( stderr, _("Recognizing\n") );
  
  /* Reserva memòria. */
  fea= tL_fea_new ( tL_amodel_get_fea_type ( amodel ) );
  if ( sil != NULL )
    {
      silsym= tL_dict_find ( syms, sil );
      if ( silsym >= amodel->N )
	tL_error ( _("%s is not in the acoustic model"), sil );
    }
  else silsym= amodel->N;
  wg= NULL;
  reco= tL_recogniser_new ( amodel, lexicon, lm,
			    viterbi_states, wg_lst==NULL ? 0 : nhyp,
			    la_cap, la_depth, silsym, &err );
  if ( reco == NULL ) tL_error ( err );
  reco->nmaxstates= nmaxstates;
  reco->beam= beam;
  reco->lm_beam= lm_beam;
  reco->align= align;
  if ( output == NULL ) to= stdout;
  else
    {
      to= fopen ( output, "w" );
      if ( to == NULL )
        tL_error ( _("Could not open '%s'"), output );
    }
  rec= NULL;
  
  /* Reconeix. */
  for ( n= 0; n < pbm_lst->N; ++n )
    {
      t1= clock ();
      recognise_sample ( reco, to, fea, pbm_lst->files[n],
			 wg_lst==NULL ? NULL : wg_lst->files[n],
			 syms, lexicon->words, print_all, print_probs,
			 print_words, htk_wgs, &rec, &rec_size, &wg, verbose );
      t2= clock ();
      if ( verbose )
	fprintf ( stderr, _("Elapsed time: %f s\n"),
		  ((double) t2-t1)/CLOCKS_PER_SEC  );
    }
  
  /* Allibera memòria. */
  if ( wg != NULL ) tL_wordgraph_free ( wg );
  if ( rec != NULL ) free ( rec );
  tL_recogniser_free ( reco );
  if ( to != stdout ) fclose ( to );
  tL_fea_free ( fea );
  
} /* end recognise */


tLFileList *
get_wg_lst (
            const char *name,
            size_t      N,
            int         verbose
            )
{
  
  tLFileList *fl;
  size_t n;
  
  
  fl= tL_filelist_new_from_file ( name );
  if ( fl == NULL ) tL_error ( _("Could not open '%s'"), name );
  if ( N != fl->N )
    tL_error ( _("The file list for the word graphs contains %lu entries,"
		 " but %lu were expected"),
	       _PS(fl->N), _PS(N) );
  if ( verbose )
    {
      fprintf ( stderr, _("The file list for the word graphs"
			  " has been read\n") );
      if ( verbose > 2 )
        {
          for ( n= 0; n < fl->N; ++n )
            fprintf ( stderr, "  - %lu: \"%s\"\n",
                      _PS(n), fl->files[n] );
        }
    }
  
  return fl;
  
} /* end get_wg_lst */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLLexicon *lexicon;
  tLAModel *amodel;
  tLLM *lm;
  tLFileList *pbm_lst, *wg_lst;
  clock_t t1,t2;
  
  
  /* i18n functions */
  setlocale (LC_MESSAGES, "");
  setlocale (LC_CTYPE, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  /*bind_textdomain_codeset(PACKAGE, "utf-8");*/
  textdomain (PACKAGE);
  
  usage ( argc, argv, &opts, &args );
  if ( opts.verbose )
    {
      fprintf ( stderr, "GSF: %g\n", _DB(opts.gsf) );
      fprintf ( stderr, "NMAXAS: %lu\n", opts.nmaxas );
      fprintf ( stderr, "WIP: %g\n", _DB(opts.wip) );
      fprintf ( stderr, "ALIGN: %s\n",
		(opts.align==TL_REC_ALIGN_NONE) ? _("No") :
		((opts.align==TL_REC_ALIGN_WORDS)?
		 _("Words") : _("Words+Symbols")));
      fprintf ( stderr, "BEAM: %g\n", _DB(opts.beam) );
      fprintf ( stderr, "WORD-END PRUN.: %g\n", _DB(opts.lm_beam) );
      fprintf ( stderr, "VITERBI STATES: %s\n",
		opts.viterbi_states ? _("Yes") : _("No") );
      fprintf ( stderr, "WORDGRAPHS: %s\n",
		opts.wglst_fn == NULL ? _("No") : _("Yes") );
      fprintf ( stderr, "WORDGRAPH DENSITY: %u\n", opts.nhyp );
      fprintf ( stderr, "LOOKAHEAD CAPACITY: %u\n", opts.la_cap );
      fprintf ( stderr, "LOOKAHEAD DEPTH: %u\n", opts.la_depth );
      if ( opts.sil != NULL )
	fprintf ( stderr, _("SILENCE: Yes (%s)\n"), opts.sil );
      else fprintf ( stderr, _("SILENCE: No\n") );
    }
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  lexicon= get_lexicon ( opts.lexicon_fn, syms, opts.verbose );
  lm= get_lm ( args.lm_fn, lexicon, opts.lexicon_fn == NULL,
               syms, opts.verbose );
  if ( opts.gsf != 1.0 ) tL_lm_set_gsf ( lm, opts.gsf );
  /* L'ordre és important. Primer GSF després WIP. */
  if ( opts.wip != 0.0 ) tL_lm_set_wip ( lm, opts.wip );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  wg_lst= opts.wglst_fn==NULL ?
    NULL :
    get_wg_lst ( opts.wglst_fn, pbm_lst->N, opts.verbose );
  
  /* Cos. */
  t1= clock ();
  recognise ( amodel, lexicon, lm, syms, pbm_lst, wg_lst,
	      opts.output_fn, opts.nmaxas, opts.align,
	      opts.beam, opts.lm_beam, opts.print_all, opts.print_probs,
	      opts.print_words, opts.htk_wgs,
	      opts.viterbi_states, opts.nhyp, opts.la_cap, opts.la_depth,
	      opts.sil, opts.verbose );
  t2= clock ();
  if ( opts.verbose )
    fprintf ( stderr, _("Total elapsed timed: %f s\n"),
	      ((double) (t2-t1))/CLOCKS_PER_SEC );
  
  /* Allibera memòria. */
  if ( wg_lst != NULL ) tL_filelist_free ( wg_lst );
  tL_filelist_free ( pbm_lst );
  tL_lm_free ( lm );
  tL_lexicon_free ( lexicon );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
