/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2008,2009,2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  train_amodel.c - Entrena un model de caràcters mitjançant
 *                   EM.
 *
 *  Adrià Giménez Pastor, 2008,2009,2010,2011,2012,2013.
 *
 */

#include <libTLK/TLK.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "get_pbm_lst.h"
#include "get_words.h"
#include "itertrain.h"
#include "itertrain_viterbi.h"
#include "macros.h"
#include "print_mixcounts.h"
#include "write_amodel.h"
#include "write_counts.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 3




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int           verbose;
  const char   *output_fn;
  tLFloat       smooth;
  tLFloat       cfactor;
  unsigned int  max_iters;
  tLBool        nowords;
  int           skiplll;
  tLBool        smooth_trans;
  tLBool        viterbi;
  tLBool        print_counts;
  const char   *occs_fn;
  const char   *mixcounts_fn;
  tLBool        binary;
  unsigned int  blocksize;
  const char   *tarfn;
  const char   *tartransfn;
  
};


struct args
{
  
  const char *amodel_fn;
  const char *wdlst_fn;
  const char *pbmlst_fn;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <amodel> <transc-lst> <sample-lst>\n\n",
	    prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Initial acoustic model") );
  fprintf ( to, "    <transc-lst> %s\n",
	    _("List of the files containing the transcriptions") );
  fprintf ( to, "    <sample-lst> %s\n\n",
	    _("List of the files containing the samples") );
  fprintf ( to, "    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -b,--block-size <uint>: %s\n",
	    _("Number of samples in memory at each point. Default is 1") );
  fprintf ( to, "      -B,--binary: %s\n",
	    _("Print the files in binary format") );
  fprintf ( to, "      -c,--cfactor <double>: %s\n",
	    _("Convergence factor. Default is 0.0001") );
  fprintf ( to, "      -m,--max-iters <uint>: %s\n",
	    _("Maximum number of iterations. Default is 20") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the trained model will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -s,--smooth <double>: %s\n",
	    _("Smoothing factor. Default is 1e-6") );
  fprintf ( to, "      --print-counts: %s\n",
	    _("Only one iteration will be performed"
	      " and counts will be printed to the output") );
  fprintf ( to, "      --print-mixcounts <string>: %s\n",
	    _("Print to this file the mixture coefficient counts") );
  fprintf ( to, "      --print-occs <string>: %s\n",
	    _("Print to this file state occupations at the last iteration") );
  fprintf ( to, "      --raw-transitions: %s\n",
	    _("Smoothing will not be applied to transition probabilities") );
  fprintf ( to, "      --show-last-ll: %s\n",
	    _("Log probability will be shown for the last iteration") );
  fprintf ( to, "      --tar <string>: %s\n",
	    _("Samples will be read from this file") );
  fprintf ( to, "      --tar-trans <string>: %s\n",
	    _("Transcriptions will be read from this file") );
  fprintf ( to, "      --viterbi: %s\n",
	    _("Use Viterbi training instead of the EM algorithm") );
   fprintf ( to, "      --words: %s\n",
	    _("Transcriptions will be expected to be words") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          argc,
       const char  *argv[],
       struct opts *opts,
       struct args *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, cwords, n, csmooth_trans, 
    cviterbi, cpcounts, cbinary, showlll;
  const char *smooth, *prog, *cfactor, *max, *blocksize;


  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_value ( cmdl, 'b', "block-size", &(blocksize) );
  add_cmdl_counter ( cmdl, 'B', "binary", &(cbinary) );
  add_cmdl_value ( cmdl, 'c', "cfactor", &cfactor );
  add_cmdl_value ( cmdl, 'm', "max-iters", &max );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 's', "smooth", &smooth );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "show-last-ll", &showlll );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "raw-transitions", &csmooth_trans );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "viterbi", &cviterbi );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "print-counts", &cpcounts );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "print-mixcounts",
		   &(opts->mixcounts_fn) );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "print-occs", &(opts->occs_fn) );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "tar", &(opts->tarfn) );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "tar-trans", &(opts->tartransfn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "words", &cwords );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, argc, argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( cfactor != NULL )
    {
      opts->cfactor= atof ( cfactor );
      if ( opts->cfactor <= 0.0 )
	tL_error ( _("The convergence factor must be greater than 0") );
    }
  else opts->cfactor= 0.0001;
  if ( max != NULL )
    {
      opts->max_iters= 
	(unsigned int) strtol ( max, (char **) NULL, 10 );
    }
  else opts->max_iters= 20;
  if ( smooth != NULL )
    {
      opts->smooth= atof ( smooth );
      if ( opts->smooth < 0.0 || opts->smooth > 1.0 )
        tL_error ( _("The smoothing factor must be in the range [0,1]") );
    }
  else opts->smooth= 1e-6;
  opts->nowords= (cwords == 0);
  opts->smooth_trans= (csmooth_trans == 0);
  opts->viterbi= (cviterbi != 0);
  opts->print_counts= (cpcounts != 0);
  opts->binary= (cbinary != 0);
  if ( blocksize == NULL ) opts->blocksize= 1;
  else
    {
      opts->blocksize= atoui ( blocksize );
      if ( opts->blocksize == 0 )
	tL_error ( _("Block size cannot be 0") );
    }
  opts->skiplll= !showlll;
  
  /* Arguments. */
  if ( n != NUM_ARGS )
    {
      help ( stderr, prog, FALSE );
      tL_error ( _("Unknown option or wrong number of arguments;"
		   " type --help for help") );
    }
  args->amodel_fn= argv[0];
  args->wdlst_fn= argv[1];
  args->pbmlst_fn= argv[2];
  
} /* end usage */


static void
train (
       tLAModel           *amodel,
       tLTrainer          *trainer,
       tLAMlogCounts      *logcounts,
       tLVTrainer         *vtrainer,
       tLAMiCounts        *icounts,
       tLSeqs             *words,
       tLFileList         *pbm_lst,
       unsigned int        max_iters,
       tLFloat             cfactor,
       tLFloat             smooth,
       int                 skiplll,
       const tLBool        smooth_trans,
       const tLBool        viterbi,
       const tLBool        print_counts,
       const char         *mixcounts_fn,
       const tLDict       *syms,
       const unsigned int  blocksize,
       const char         *tarfn,
       int                 verbose
       )
{
  
  int i;
  tLProb lv, lvp;
  tLFloat inc;
  tLBool conv;
  feablock_t *feablock;
  tLTar *tar;
  char *err;
  
  
  conv= FALSE; /* CALLA!! */
  
  if ( verbose )
    fprintf ( stderr, _("Starting training\n") );
  
  if ( tarfn == NULL ) tar= NULL;
  else
    {
      tar= tL_tar_new ( tarfn, &err );
      if ( tar == NULL ) tL_error ( _("Error while reading '%s': %s"), tarfn, err );
    }

  feablock= create_feablock ( blocksize, amodel );
  inc= HUGE_VAL;
  i= 0;
  if ( viterbi )
    lv= itertrain_viterbi ( vtrainer, feablock->feas[0], pbm_lst, verbose );
  else
    lv= itertrain ( trainer, feablock, pbm_lst, tar, verbose );
  if ( verbose )
    {
      if ( print_counts ) fprintf ( stderr, "LL: %g\n", _DB(lv) );
      else
	{
	  fprintf ( stderr, "  NºIter\t\tLL\t\tINC\n" );
	  fprintf ( stderr, "%c %d\t\t%g\t\t%g\n",
		    lv==0.0?'.':'*', i, _DB(lv), _DB(inc) );
	}
    }
  if ( lv == 0 || print_counts ) goto quit; /* No es toca el model. */
  while ( lv != -HUGE_VAL && i < max_iters && (i == 0 || !conv) )
    {
      
      if ( viterbi ) tL_amicounts_update_amodel ( icounts, amodel );
      else tL_amlogcounts_update_amodel ( logcounts, amodel );
      if ( smooth > 0.0 )
	tL_amodel_smooth ( amodel, smooth, smooth_trans );
      ++i;
      if ( mixcounts_fn != NULL && i == max_iters )
	print_mixcounts ( mixcounts_fn, amodel, syms,
			  logcounts, icounts, viterbi );
      if ( skiplll && i == max_iters ) break;
      
      
      lvp= lv;
      if ( viterbi )
	lv= itertrain_viterbi ( vtrainer, feablock->feas[0], pbm_lst, verbose );
      else
	lv= itertrain ( trainer, feablock, pbm_lst, tar, verbose );
      inc= (lv - lvp) / -lvp;
      conv= (inc < cfactor);
      if ( verbose )
	fprintf ( stderr, "%c %d\t\t%g\t\t%g\n",
		  conv ? '.' : '*',
		  i, _DB(lv), _DB(inc) );
      
    }
 quit:
  if ( mixcounts_fn != NULL && i != max_iters )
    print_mixcounts ( mixcounts_fn, amodel, syms, logcounts, icounts, viterbi );
  if ( tar != NULL ) tL_tar_free ( tar );
  free_feablock ( feablock );
  
} /* end train */


static void
write_occs (
	    const char      *fn,
	    tLFloat        **occs,
	    const tLAModel  *amodel,
	    const tLDict    *syms
	    )
{
  
  FILE *f;
  
  
  f= fopen ( fn, "w" );
  if ( f == NULL ) tL_error ( _("Could not open '%s'"), fn );
  tL_occupancies_print ( occs, f, amodel, syms );
  fclose ( f );
  
} /* end write_occs */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel;
  tLFileList *pbm_lst;
  tLSeqs *words;
  tLAMlogCounts *logcounts;
  tLTrainer *trainer;
  tLAMiCounts *icounts;
  tLVTrainer *vtrainer;
  char *err;
  tLFloat **occs;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( argc, argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  pbm_lst= get_pbm_lst ( args.pbmlst_fn, opts.verbose );
  words= get_words ( args.wdlst_fn, syms, opts.nowords, opts.tartransfn, opts.verbose );
  if ( words->N != pbm_lst->N )
    tL_error ( _("The sample list contains %lu entries,"
		 " but the word list contains %lu entries"),
	       _PS(pbm_lst->N), _PS(words->N) );
  
  /* Cos. */
  if ( opts.viterbi && opts.occs_fn != NULL )
    {
      tL_warning ( _("'print-occs' is not implemented for Viterbi training") );
      opts.occs_fn= NULL;
    }
  if ( opts.viterbi )
    {
      icounts= tL_amicounts_new ( amodel );
      vtrainer= tL_vtrainer_new ( amodel, icounts, words, &err );
      trainer= NULL;
      logcounts= NULL;
      occs= NULL;
    }
  else
    {
      occs= opts.occs_fn==NULL ? NULL : tL_occupancies_new ( amodel );
      logcounts= tL_amlogcounts_new ( amodel );
      trainer= tL_trainer_new ( amodel, logcounts, words, occs, &err );
      vtrainer= NULL;
      icounts= NULL;
    }
  if ( trainer == NULL && vtrainer == NULL ) tL_error ( err );
  train ( amodel, trainer, logcounts, vtrainer, icounts, words,
	  pbm_lst, opts.max_iters, opts.cfactor, opts.smooth,
	  opts.skiplll, opts.smooth_trans, opts.viterbi,
	  opts.print_counts, opts.mixcounts_fn, syms, opts.blocksize,
	  opts.tarfn, opts.verbose );
  if ( opts.print_counts )
    write_counts ( opts.output_fn, amodel, logcounts,
		   icounts, opts.binary, opts.viterbi );
  else write_amodel ( opts.output_fn, amodel, syms, opts.binary );
  if ( occs != NULL ) write_occs ( opts.occs_fn, occs, amodel, syms );
  
  /* Allibera memòria. */
  tL_seqs_free ( words );
  if ( trainer != NULL )
    {
      tL_trainer_free ( trainer );
      tL_amlogcounts_free ( logcounts, amodel );
    }
  else
    {
      tL_vtrainer_free ( vtrainer );
      tL_amicounts_free ( icounts, amodel );
    }
  if ( occs != NULL ) tL_occupancies_free ( occs );
  tL_filelist_free ( pbm_lst );
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
