/*
 *  Copyright 2012,2013,2014 transLectures-UPV Team
 *  Copyright 2010,2011 Adrià Giménez Pastor and Alfons Juan Císcar
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
/*
 *  update.c - Acumula comptadors i actualitza el model.
 *
 *  Adrià Giménez Pastor, 2010,2011,2012,2013.
 *
 */


#include <libTLK/TLK.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/* i18n includes and defines */
#include <locale.h>
#include "intl.h"

#include "cmdl.h"
#include "get_amodel.h"
#include "print_mixcounts.h"
#include "write_amodel.h"
#include "write_counts.h"

#include "sysinfo.h"




/**********/
/* MACROS */
/**********/

#define NUM_ARGS 2




/*********/
/* TIPUS */
/*********/

struct opts
{
  
  int         verbose;
  const char *output_fn;
  int         from_stdin;
  tLFloat     smooth;
  tLBool      smooth_trans;
  tLBool      binary;
  const char *mixcounts_fn;
  const char *counts_fn;
  tLBool      enable_eps;
  tLFloat     eps;
  
};


struct args
{
  
  const char *amodel_fn;
  enum
    {
      EM,
      VITERBI
    }         alg;
  
};




/************/
/* FUNCIONS */
/************/

static void
help (
      FILE       *to,
      const char *prog,
      tLBool      details
      )
{
  
  fprintf ( to, "%s [%s] <amodel> <alg> [<file1>"
	    " <file2> ...]\n\n", prog, _("options") );
  if ( !details ) return;
  fprintf ( to, "    <amodel> %s\n", _("Initial acoustic model") );
  fprintf ( to, "    <alg> %s\n",
	    _("Training algorithm that was used"
	      " to generate the count vectors:") );
  fprintf ( to, "        - em: Baum-Welch\n" );
  fprintf ( to, "        - viterbi: %s\n", _("Viterbi training") );
  fprintf ( to, "\n    %s:\n\n", _("Options") );
  fprintf ( to, "      -h,--help: %s\n", _("Show this help message") );
  fprintf ( to, "      -V,--version: %s\n", _("Version") );
  fprintf ( to, "      -v,--verbose: %s\n", _("Verbose mode") );
  fprintf ( to, "      -B,--binary: %s\n", _("Work in binary format") );
  fprintf ( to, "      -e,--epsilon <double>: %s\n",
	    _("EM algorithm: counts will be interpolated using this value"
	      " each time a pair of counts are accumulated."
	      " By default, no interpolation") );
  fprintf ( to, "      -o,--output <string>: %s\n",
	    _("File to which the trained model will be written."
	      " Defaults to the standard output") );
  fprintf ( to, "      -s,--smooth <double>: %s\n",
	    _("Smoothing factor. Default is 1e-6") );
  fprintf ( to, "      --print-counts <string>: %s\n",
	    _("Print to this file the accumulated counts") );
  fprintf ( to, "      --print-mixcounts <string>: %s\n",
	    _("Print to this file the mixture coefficient counts."
	      " Only valid for em and viterbi") );
  fprintf ( to, "      --raw-transitions: %s\n",
	    _("Smoothing will not be applied to transition probabilities") );
  fprintf ( to, "      --stdin <int>: %s\n",
	    _("Counts will be read from the standard input."
	      " '<int>' specifies how many files should be read."
	      " Default is 0") );
  fprintf ( to, "\n" );
  
} /* end help */


static void
usage (
       int          *argc,
       const char  **argv[],
       struct opts  *opts,
       struct args  *args
       )
{
  
  cmdl_t cmdl;
  int chelp, cversion, csmooth_trans, n, cbinary;
  const char *prog, *cstdin, *smooth, *eps;
  
  
  /* Crea opcions. */
  cmdl= create_cmdl ();
  add_cmdl_counter ( cmdl, 'h', "help", &chelp );
  add_cmdl_counter ( cmdl, 'V', "version", &cversion );
  add_cmdl_counter ( cmdl, 'v', "verbose", &(opts->verbose) );
  add_cmdl_counter ( cmdl, 'B', "binary", &(cbinary) );
  add_cmdl_value ( cmdl, 'e', "epsilon", &eps );
  add_cmdl_value ( cmdl, 'o', "output", &(opts->output_fn) );
  add_cmdl_value ( cmdl, 's', "smooth", &smooth );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "print-counts", &(opts->counts_fn) );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "print-mixcounts",
		   &(opts->mixcounts_fn) );
  add_cmdl_counter ( cmdl, CMDL_NOSHORT, "raw-transitions", &csmooth_trans );
  add_cmdl_value ( cmdl, CMDL_NOSHORT, "stdin", &cstdin );
  
  /* Parseja opcions. */
  n= parse_cmdl ( cmdl, *argc, *argv, &prog );
  free_cmdl ( cmdl );

  /* Opcions. */
  if ( chelp )
    {
      help ( stdout, prog, TRUE );
      exit ( EXIT_SUCCESS );
    }
  if ( cversion )
    {
      printf ( Version );
      exit ( EXIT_SUCCESS );
    }
  if ( smooth != NULL )
    {
      opts->smooth= atof ( smooth );
      if ( opts->smooth < 0.0 || opts->smooth > 1.0 )
        tL_error ( _("The smoothing factor must be in the range [0,1]") );
    }
  else opts->smooth= 1e-6;
  opts->from_stdin= cstdin==NULL ? 0 : atoi ( cstdin );
  if ( opts->from_stdin < 0 )
    tL_error ( _("The value for the 'stdin' option cannot be less than 0") );
  opts->smooth_trans= (csmooth_trans == 0);
  opts->binary= (cbinary != 0);
  if ( eps == NULL ) opts->enable_eps= FALSE;
  else
    {
      opts->enable_eps= TRUE;
      opts->eps= atof ( eps );
      if ( opts->eps < 0.0 || opts->eps > 1.0 )
 	tL_error ( "The value for the 'eps' option"
		   " must be in the range [0,1]" );
    }
  
  /* Arguments. */
  if ( n < NUM_ARGS ) goto quit;
  args->amodel_fn= (*argv)[0];
  if ( !strcmp ( (*argv)[1], "em" ) ) args->alg= EM;
  else if ( !strcmp ( (*argv)[1], "viterbi" ) ) args->alg= VITERBI;
  else tL_error ( _("Unknown algorithm: %s"), (*argv)[1] );
  if ( opts->from_stdin == 0 && n == 2 )
    tL_error ( _("No files could be found containing the counts") );
  *argc= n-2; *argv+= 2;
  
  return;
  
 quit:
  help ( stderr, prog, FALSE );
  tL_error ( _("Unknown option or wrong number of arguments;"
	       " type --help for help") );
  
} /* end usage */


static void
update_viterbi (
		const int     argc,
		const char   *argv[],
		tLAModel     *amodel,
		const tLDict *syms,
		const int     from_stdin,
		const tLBool  binary,
		const char   *counts_fn,
		const char   *mixcounts_fn,
		const int     verbose
		)
{
  
  int N, n;
  tLBool read_stdin;
  tLAMiCounts *counts;
  gzFile f;
  char *err;
  
  
  /* Inicialitza. */
  if ( verbose )
    fprintf ( stderr, _("Starting count accumulation"
			" for Viterbi training\n") );
  counts= tL_amicounts_new ( amodel );
  
  /* Cos. */
  N= (read_stdin= (from_stdin!=0)) ? from_stdin : argc;
  if ( read_stdin ) f= gzdopen ( fileno ( stdin ), "r" );
  else              f= NULL;
  for ( n= 0; n < N; ++n )
    {
      if ( !read_stdin )
	{
	  f= gzopen ( argv[n], "r" );
	  if ( f == NULL ) tL_error ( _("Could not open '%s'"), argv[n] );
	  if ( verbose > 1 )
	    fprintf ( stderr, _("  Reading counts file '%s'\n"),
		      argv[n] );
	}
      else if ( verbose > 1 )
	fprintf ( stderr,
		  _("  Reading counts file No. %d\n"),
		  n+1 );
      if ( n == 0 )
	{
	  if ( tL_amicounts_read ( counts, amodel, f, binary, &err ) == -1 )
	    tL_error ( err );
	}
      else
	{
	  if ( tL_amicounts_accum ( counts, amodel, f, binary, &err ) == -1 )
	    tL_error ( err );
	}
      if ( !read_stdin ) gzclose ( f );
    }
  tL_amicounts_update_amodel ( counts, amodel );
  if ( counts_fn != NULL )
    write_counts ( counts_fn, amodel, NULL, counts, binary, TRUE );
  if ( mixcounts_fn != NULL )
    print_mixcounts ( mixcounts_fn, amodel, syms, NULL, counts, TRUE );
  
  /* Allibera memòria. */
  tL_amicounts_free ( counts, amodel );
  
} /* end update_viterbi */


static void
update_em (
	   const int     argc,
	   const char   *argv[],
	   tLAModel     *amodel,
	   const tLDict *syms,
	   const int     from_stdin,
	   const tLBool  binary,
	   const char   *counts_fn,
	   const char   *mixcounts_fn,
	   const tLBool  enable_eps,
	   const tLFloat eps,
	   const int     verbose
	   )
{
  
  int N, n;
  tLBool read_stdin;
  tLAMlogCounts *counts;
  gzFile f;
  char *err;
  
  
  /* Inicialitza. */
  if ( verbose )
    fprintf ( stderr, _("Starting count accumulation"
			" for EM-based training\n") );
  counts= tL_amlogcounts_new ( amodel );
  tL_amlogcounts_init ( counts, amodel );
  
  /* Cos. */
  N= (read_stdin= (from_stdin!=0)) ? from_stdin : argc;
  if ( read_stdin ) f= gzdopen(fileno(stdin),"r");
  else              f= NULL;
  for ( n= 0; n < N; ++n )
    {
      if ( !read_stdin )
	{
	  f= gzopen ( argv[n], "r" );
	  if ( f == NULL ) tL_error ( _("Could not open '%s'"), argv[n] );
	  if ( verbose > 1 )
	    fprintf ( stderr, _("  Reading counts file '%s'\n"),
		      argv[n] );
	}
      else if ( verbose > 1 )
	fprintf ( stderr,
		  _("  Reading counts file No. %d\n"),
		  n+1 );
      if ( n == 0 )
	{
	  if ( tL_amlogcounts_read ( counts, amodel, f, binary, &err ) == -1 )
	    tL_error ( err );
	}
      else
	{
	  if ( enable_eps )
            {
              if ( tL_amlogcounts_accum_eps ( counts, amodel, f, eps,
                                              binary, &err ) == -1 )
                tL_error ( err );
            }
          else
            {
	      if ( tL_amlogcounts_accum ( counts, amodel, f,
					  binary, &err ) == -1 )
		tL_error ( err );
            }
	}
      if ( !read_stdin ) gzclose ( f );
    }
  if ( counts_fn != NULL )
    write_counts ( counts_fn, amodel, counts, NULL, binary, FALSE );
  tL_amlogcounts_update_amodel ( counts, amodel );
  if ( mixcounts_fn != NULL )
    print_mixcounts ( mixcounts_fn, amodel, syms, counts, NULL, FALSE );
  
  /* Allibera memòria. */
  tL_amlogcounts_free ( counts, amodel );
  
} /* end update_em */




/**********************/
/* PROGRAMA PRINCIPAL */
/**********************/

int main ( int argc, const char *argv[] )
{
  
  struct opts opts;
  struct args args;
  tLDict *syms;
  tLAModel *amodel;
  
  
  /* i18n functions */
  setlocale ( LC_MESSAGES, "" );
  setlocale ( LC_CTYPE, "" );
  bindtextdomain ( PACKAGE, LOCALEDIR );
  textdomain ( PACKAGE );
  
  usage ( &argc, &argv, &opts, &args );
  
  /* Carrega fitxers. */
  syms= tL_dict_new ();
  amodel= get_amodel ( args.amodel_fn, syms, opts.verbose );
  
  /* Cos. */
  if ( args.alg == VITERBI )
    update_viterbi ( argc, argv,  amodel, syms, opts.from_stdin,
		     opts.binary, opts.counts_fn, opts.mixcounts_fn,
		     opts.verbose );
  else if ( args.alg == EM )
    update_em ( argc, argv, amodel, syms, opts.from_stdin,
		opts.binary, opts.counts_fn, opts.mixcounts_fn,
		opts.enable_eps, opts.eps, opts.verbose );
  if ( opts.smooth > 0.0 )
    tL_amodel_smooth ( amodel, opts.smooth, opts.smooth_trans );
  write_amodel ( opts.output_fn, amodel, syms, opts.binary );
  
  /* Allibera memòria. */
  tL_amodel_free ( amodel );
  tL_dict_free ( syms );
  
  return EXIT_SUCCESS;
  
}
